/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "Adc64ConfigConverter.h"

#include "base/ConfigConverterUtil.h"
#include "base/DeviceIndexConverter.h"

#include "Adc64Config.h"
#include "Adc64SystemConfig.h"
#include "device-modules/DeserializersControl/AdcSerDesDelayConvertor.h"


namespace {
constexpr const char *FIR_SEPARATOR = "; ";

// ad5622operation
constexpr const char *k_data_baseline = "baseline";
constexpr const char *k_data_level = "level";

// SparseParams
constexpr const char *k_en = "en";
constexpr const char *k_sparseAllowed = "sparseAllowed";
constexpr const char *k_readCellNum = "readCellNum";
constexpr const char *k_offset = "offset";
constexpr const char *k_period = "period";
constexpr const char *k_pointNumber = "pointNumber";

// FirParams

constexpr const char *k_presetKey = "presetKey";
constexpr const char *k_roundoff = "roundoff";
constexpr const char *k_coef = "coef";

// DspParams
constexpr const char *k_fir = "fir";
constexpr const char *k_blc_thr = "blc_thr";
constexpr const char *k_maf_enable = "maf_enable";
constexpr const char *k_maf_tap_sel = "maf_tap_sel";
constexpr const char *k_test_enable = "test_enable";

constexpr const char *k_chEn = "chEn";

// DominoSettings
constexpr const char *k_chBaseline = "chBaseline";
constexpr const char *k_chTrigEn = "chTrigEn";
constexpr const char *k_chTrigThr = "chTrigThr";
constexpr const char *k_chZsThr = "chZsThr";

constexpr const char *k_devIndex = "devIndex";
constexpr const char *k_enabled = "enabled";

constexpr const char *k_nch = "nch";
constexpr const char *k_run = "run";
constexpr const char *k_skipRestartRun = "skipRestartRun";
constexpr const char *k_readCellNumber = "readCellNumber";
constexpr const char *k_trigDelay = "trigDelay";
constexpr const char *k_invertInput = "invertInput";
constexpr const char *k_zs_en = "zs_en";
constexpr const char *k_invert_thr_trig = "invert_thr_trig";
constexpr const char *k_invert_zs_thr = "invert_zs_thr";
constexpr const char *k_softwareZs = "softwareZs";
constexpr const char *k_trigsetup_all = "trigsetup_all";
constexpr const char *k_reg_pca_all = "reg_pca";
constexpr const char *k_ad5622op = "ad5622op";
constexpr const char *k_amp_board_set = "amp_board_set";
constexpr const char *k_sparseParams = "sparseParams";
constexpr const char *k_dspParams = "dspParams";
constexpr const char *k_mstreamEnable = "mstreamEnable";
constexpr const char *k_mtuSize = "mtuSize";
constexpr const char *k_periodicalAdcDelayCheck = "periodicalAdcDelayCheck";
constexpr const char *k_key = "key";
constexpr const char *k_serDesDelayPreset = "serDesDelayPreset";

// Adc64Config
constexpr const char *k_preamplifier = "preamplifier";
constexpr const char *k_showSparseMarkers = "showSparseMarkers";
constexpr const char *k_showPcbCh = "showPcbCh";
constexpr const char *k_adcDelayResultDir = "adcDelayResultDir";
constexpr const char *k_fir_map = "fir_map";

// Adc64SystemConfig
constexpr const char *k_allow_remote_control = "allow_remote_control";
constexpr const char *k_write_enable = "write_enable";
constexpr const char *k_storage_path = "storage_path";
constexpr const char *k_show_details = "show_details";
constexpr const char *k_event_number_limit = "event_number_limit";
constexpr const char *k_mtu_size = "mtu_size";
constexpr const char *k_trig_on_xoff_stop = "trig_on_xoff_stop";
constexpr const char *k_device_configs = "device_configs";
}

// -------------- FirPreset
template<>
bool AbstractConfigConverter<FirPreset>::classIsObject()
{
    return false;
}

template<>
QJsonObject AbstractConfigConverter<FirPreset>::toJsonObject(const FirPreset &)
{Q_UNREACHABLE();}

template<>
FirPreset AbstractConfigConverter<FirPreset>::fromJsonObject(const QJsonObject &)
{Q_UNREACHABLE();}

template<>
QJsonValue AbstractConfigConverter<FirPreset>::toJsonValue(const FirPreset &in)
{
    QStringList res;
    for(auto &v : in)
        res << QString::number(v);
    return res.join(FIR_SEPARATOR);
}

template<>
FirPreset AbstractConfigConverter<FirPreset>::fromJsonValue(const QJsonValue &in)
{
    FirPreset res;
    const auto &list = in.toString().split(FIR_SEPARATOR);
    res.reserve(list.size());
    for(auto &v : list)
        res.push_back(v.toShort());
    return res;
}

template<>
void AbstractConfigConverter<FirPreset>::registerSubTypes()
{
}
// -------------- ad5622operation
template<>
bool AbstractConfigConverter<ad5622operation>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<ad5622operation>::toJsonObject(const ad5622operation &in)
{
    QVariantMap m;
    m[k_data_baseline] = QVariant::fromValue(in.data_baseline);
    m[k_data_level] = QVariant::fromValue(in.data_level);
    return fromVariantMap(m);
}

template<>
ad5622operation AbstractConfigConverter<ad5622operation>::fromJsonObject(const QJsonObject &in)
{
    ad5622operation out;
    update_value(in, k_data_baseline, &out.data_baseline);
    update_value(in, k_data_level, &out.data_level);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<ad5622operation>::toJsonValue(const ad5622operation &in)
{
    return toJsonObject(in);
}

template<>
ad5622operation AbstractConfigConverter<ad5622operation>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<ad5622operation>::registerSubTypes()
{
}

// -------------- SparseParams
template<>
bool AbstractConfigConverter<SparseParams>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<SparseParams>::toJsonObject(const SparseParams &in)
{
    QVariantMap m;
    m[k_en] = QVariant::fromValue(in.en);
    m[k_sparseAllowed] = QVariant::fromValue(in.sparseAllowed); // doesn't stored
    m[k_readCellNum] = QVariant::fromValue(in.readCellNum);
    m[k_offset] = QVariant::fromValue(in.offset);
    m[k_period] = QVariant::fromValue(in.period);
    m[k_pointNumber] = QVariant::fromValue(in.pointNumber);
    return fromVariantMap(m);
}

template<>
SparseParams AbstractConfigConverter<SparseParams>::fromJsonObject(const QJsonObject &in)
{
    SparseParams out;
    update_value(in, k_en, &out.en);
    update_value(in, k_sparseAllowed, &out.sparseAllowed); // doesn't stored
    update_value(in, k_readCellNum, &out.readCellNum);
    update_value(in, k_offset, &out.offset);
    update_value(in, k_period, &out.period);
    update_value(in, k_pointNumber, &out.pointNumber);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<SparseParams>::toJsonValue(const SparseParams &in)
{
    return toJsonObject(in);
}

template<>
SparseParams AbstractConfigConverter<SparseParams>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<SparseParams>::registerSubTypes()
{}

// -------------- FirParams
template<>
bool AbstractConfigConverter<FirParams>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<FirParams>::toJsonObject(const FirParams &in)
{
    QVariantMap m;
    m[k_presetKey] = QVariant::fromValue(in.presetKey);
    m[k_en] = QVariant::fromValue(in.en);
    m[k_roundoff] = QVariant::fromValue(in.roundoff);
    m[k_coef] = QVariant::fromValue(in.coef); // dsp.fir.coef doesn't stored
    return fromVariantMap(m);
}

template<>
FirParams AbstractConfigConverter<FirParams>::fromJsonObject(const QJsonObject &in)
{
    FirParams out;
    update_value(in, k_presetKey, &out.presetKey);
    update_value(in, k_en, &out.en);
    update_value(in, k_roundoff, &out.roundoff);
    update_value(in, k_coef, &out.coef); // dsp.fir.coef doesn't stored
    return out;
}

template<>
QJsonValue AbstractConfigConverter<FirParams>::toJsonValue(const FirParams &in)
{
    return toJsonObject(in);
}

template<>
FirParams AbstractConfigConverter<FirParams>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<FirParams>::registerSubTypes()
{
    AbstractConfigConverter<FirPreset>::registerTypes();
}
// -------------- DspParams
template<>
bool AbstractConfigConverter<DspParams>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<DspParams>::toJsonObject(const DspParams &in)
{
    QVariantMap m;
    m[k_fir] = QVariant::fromValue(in.fir); // fir.coef doesn't stored
    m[k_en] = QVariant::fromValue(in.en);
    m[k_blc_thr] = QVariant::fromValue(in.blc_thr);
    m[k_maf_enable] = QVariant::fromValue(in.maf_enable);
    m[k_maf_tap_sel] = QVariant::fromValue(in.maf_tap_sel);
    m[k_test_enable] = QVariant::fromValue(in.test_enable); // doesn't stored
    return fromVariantMap(m);
}

template<>
DspParams AbstractConfigConverter<DspParams>::fromJsonObject(const QJsonObject &in)
{
    DspParams out;
    update_value(in, k_fir, &out.fir); // fir.coef doesn't stored
    update_value(in, k_en, &out.en);
    update_value(in, k_blc_thr, &out.blc_thr);
    update_value(in, k_maf_enable, &out.maf_enable);
    update_value(in, k_maf_tap_sel, &out.maf_tap_sel);
    update_value(in, k_test_enable, &out.test_enable); // doesn't stored
    return out;
}

template<>
QJsonValue AbstractConfigConverter<DspParams>::toJsonValue(const DspParams &in)
{
    return toJsonObject(in);
}

template<>
DspParams AbstractConfigConverter<DspParams>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<DspParams>::registerSubTypes()
{
    AbstractConfigConverter<FirParams>::registerTypes();
}

// -------------- DominoSettings
template<>
bool AbstractConfigConverter<DominoSettings>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<DominoSettings>::toJsonObject(const DominoSettings &in)
{
    QVariantMap m;
    m[k_chEn] = ConfigConverterUtil::toJsonValue(in.chEn);
    m[k_chBaseline] = ConfigConverterUtil::toJsonValue(in.chBaseline);
    m[k_chTrigEn] = ConfigConverterUtil::toJsonValue(in.chTrigEn);
    m[k_chTrigThr] = ConfigConverterUtil::toJsonValue(in.chTrigThr);
    m[k_chZsThr] = ConfigConverterUtil::toJsonValue(in.chZsThr);

    m[k_devIndex] = QVariant::fromValue(in.devIndex);
    m[k_enabled] = QVariant::fromValue(in.enabled);

    m[k_nch] = QVariant::fromValue(in.nch); // doesn't stored
    m[k_run] = QVariant::fromValue(in.run); // doesn't stored
    m[k_skipRestartRun] = QVariant::fromValue(in.skipRestartRun); // doesn't stored used in phasingStop
    m[k_readCellNumber] = QVariant::fromValue(in.readCellNumber);
    m[k_trigDelay] = QVariant::fromValue(in.trigDelay);
    m[k_invertInput] = QVariant::fromValue(in.invertInput);
    m[k_zs_en] = QVariant::fromValue(in.zs_en);
    m[k_invert_thr_trig] = QVariant::fromValue(in.invert_thr_trig);
    m[k_invert_zs_thr] = QVariant::fromValue(in.invert_zs_thr);
    m[k_softwareZs] = QVariant::fromValue(in.softwareZs);
    m[k_serDesDelayPreset] = ConfigConverterUtil::toJsonValue(in.serDesDelayPreset);
    m[k_trigsetup_all] = QVariant::fromValue(in.trigsetup.all);
    m[k_reg_pca_all] = QVariant::fromValue(in.reg_pca.all);
    m[k_ad5622op] = QVariant::fromValue(in.ad5622op);
    m[k_amp_board_set] = QVariant::fromValue(in.amp_board_set);
    m[k_sparseParams] = QVariant::fromValue(in.sparseParams);
    m[k_dspParams] = QVariant::fromValue(in.dspParams);
    m[k_mstreamEnable] = QVariant::fromValue(in.mstreamEnable); // doesn't stored
    m[k_mtuSize] = QVariant::fromValue(in.mtuSize); // doesn't stored
    m[k_periodicalAdcDelayCheck] = QVariant::fromValue(in.periodicalAdcDelayCheck); // doesn't stored
    m[k_key] = QVariant::fromValue(in.key); // doesn't stored
    return fromVariantMap(m);
}

template<>
DominoSettings AbstractConfigConverter<DominoSettings>::fromJsonObject(const QJsonObject &in)
{
    DominoSettings out;
    out.chEn = ConfigConverterUtil::toMap<int, bool>(in[k_chEn]);
    out.chBaseline = ConfigConverterUtil::toMap<int, int>(in[k_chBaseline]);
    out.chTrigEn = ConfigConverterUtil::toMap<int, bool>(in[k_chTrigEn]);
    out.chTrigThr = ConfigConverterUtil::toMap<int, int>(in[k_chTrigThr]);
    out.chZsThr = ConfigConverterUtil::toMap<int, int>(in[k_chZsThr]);

    update_value(in, k_devIndex, &out.devIndex);
    update_value(in, k_enabled, &out.enabled);

    update_value(in, k_nch, &out.nch); // doesn't stored
    update_value(in, k_run, &out.run); // doesn't stored
    update_value(in, k_skipRestartRun, &out.skipRestartRun); // doesn't stored used in phasingStop
    update_value(in, k_readCellNumber, &out.readCellNumber);
    update_value(in, k_trigDelay, &out.trigDelay);
    update_value(in, k_invertInput, &out.invertInput);
    update_value(in, k_zs_en, &out.zs_en);
    update_value(in, k_invert_thr_trig, &out.invert_thr_trig);
    update_value(in, k_invert_zs_thr, &out.invert_zs_thr);
    update_value(in, k_softwareZs, &out.softwareZs);
    out.serDesDelayPreset = ConfigConverterUtil::toMap<int, AdcSerDesDelay>(in[k_serDesDelayPreset]);
    update_value(in, k_trigsetup_all, &out.trigsetup.all);
    update_value(in, k_reg_pca_all, &out.reg_pca.all);
    update_value(in, k_ad5622op, &out.ad5622op);
    update_value(in, k_amp_board_set, &out.amp_board_set);
    update_value(in, k_sparseParams, &out.sparseParams);
    update_value(in, k_dspParams, &out.dspParams);
    update_value(in, k_mstreamEnable, &out.mstreamEnable); // doesn't stored
    update_value(in, k_mtuSize, &out.mtuSize); // doesn't stored
    update_value(in, k_periodicalAdcDelayCheck, &out.periodicalAdcDelayCheck); // doesn't stored
    update_value(in, k_key, &out.key); // doesn't stored
    return out;
}

template<>
QJsonValue AbstractConfigConverter<DominoSettings>::toJsonValue(const DominoSettings &in)
{
    return toJsonObject(in);
}

template<>
DominoSettings AbstractConfigConverter<DominoSettings>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<DominoSettings>::registerSubTypes()
{
    AbstractConfigConverter<DeviceIndex>::registerTypes();
    AbstractConfigConverter<ad5622operation>::registerTypes();
    AbstractConfigConverter<SparseParams>::registerTypes();
    AbstractConfigConverter<DspParams>::registerTypes();
    AbstractConfigConverter<DevSerDesDelay>::registerTypes();
}

// -------------- Adc64Config
template<>
bool AbstractConfigConverter<Adc64Config>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<Adc64Config>::toJsonObject(const Adc64Config &in)
{
    QVariantMap m;
    m[k_run] = QVariant::fromValue(in.run); // doesn't stored
    m[k_preamplifier] = QVariant::fromValue(in.preamplifier);
    m[k_showSparseMarkers] = QVariant::fromValue(in.showSparseMarkers);
    m[k_showPcbCh] = QVariant::fromValue(in.showPcbCh);
    m[k_adcDelayResultDir] = QVariant::fromValue(in.adcDelayResultDir);
    m[k_fir_map] = ConfigConverterUtil::toJsonValue(in.fir_map);
    m[k_device_configs] = ConfigConverterUtil::toJsonValue(in.device_configs);
    return fromVariantMap(m);
}

template<>
Adc64Config AbstractConfigConverter<Adc64Config>::fromJsonObject(const QJsonObject &in)
{
    Adc64Config out;
    update_value(in, k_run, &out.run); // doesn't stored
    update_value(in, k_preamplifier, &out.preamplifier);
    update_value(in, k_showSparseMarkers, &out.showSparseMarkers);
    update_value(in, k_showPcbCh, &out.showPcbCh);
    update_value(in, k_adcDelayResultDir, &out.adcDelayResultDir);
    out.fir_map = ConfigConverterUtil::toMap<QString, FirPreset>(in[k_fir_map]);
    out.device_configs = ConfigConverterUtil::toMap<DeviceIndex, DominoSettings>(in[k_device_configs]);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<Adc64Config>::toJsonValue(const Adc64Config &in)
{
    return toJsonObject(in);
}

template<>
Adc64Config AbstractConfigConverter<Adc64Config>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<Adc64Config>::registerSubTypes()
{
    AbstractConfigConverter<FirPreset>::registerTypes();
    AbstractConfigConverter<DeviceIndex>::registerTypes();
    AbstractConfigConverter<DominoSettings>::registerTypes();
}

// -------------- Adc64SystemConfig
template<>
bool AbstractConfigConverter<Adc64SystemConfig>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<Adc64SystemConfig>::toJsonObject(const Adc64SystemConfig &in)
{
    QVariantMap m;
    m[k_run] = QVariant::fromValue(in.run); // doesn't stored
    m[k_allow_remote_control] = QVariant::fromValue(in.allow_remote_control);
    m[k_write_enable] = QVariant::fromValue(in.write_enable);
    m[k_storage_path] = QVariant::fromValue(in.storage_path);
    m[k_show_details] = QVariant::fromValue(in.show_details);
    m[k_event_number_limit] = jsonValueFromUInt64(in.event_number_limit);
    m[k_mtu_size] = QVariant::fromValue(in.mtu_size);
    m[k_trig_on_xoff_stop] = QVariant::fromValue(in.trig_on_xoff_stop);
    m[k_device_configs] = ConfigConverterUtil::toJsonValue(in.device_configs);

    return fromVariantMap(m);
}

template<>
Adc64SystemConfig AbstractConfigConverter<Adc64SystemConfig>::fromJsonObject(const QJsonObject &in)
{
    Adc64SystemConfig out;
    update_value(in, k_run, &out.run); // doesn't stored
    update_value(in, k_allow_remote_control, &out.allow_remote_control);
    update_value(in, k_write_enable, &out.write_enable);
    update_value(in, k_storage_path, &out.storage_path);
    update_value(in, k_show_details, &out.show_details);
    out.event_number_limit = jsonValueToUInt64(in[k_event_number_limit]);
    update_value(in, k_mtu_size, &out.mtu_size);
    update_value(in, k_trig_on_xoff_stop, &out.trig_on_xoff_stop);
    out.device_configs = ConfigConverterUtil::toMap<DeviceIndex, DominoSettings>(in[k_device_configs]);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<Adc64SystemConfig>::toJsonValue(const Adc64SystemConfig &in)
{
    return toJsonObject(in);
}

template<>
Adc64SystemConfig AbstractConfigConverter<Adc64SystemConfig>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<Adc64SystemConfig>::registerSubTypes()
{
    AbstractConfigConverter<DeviceIndex>::registerTypes();
    AbstractConfigConverter<DominoSettings>::registerTypes();
}
