/*
**    Copyright 2010-2013 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "dominosettings.h"

#include <QDebug>
#include <QSettings>

namespace {
const int FIR_ROUNDOFF_DEFAULT = 0x0001;
const QString SECTION_PREFIX_FIR = "FIR_";
const QString SECTION_PREFIX_SETUP = "Setup_";
}

FirParams::FirParams()
    : roundoff(FIR_ROUNDOFF_DEFAULT)
{
    coef.resize(16);
    coef[0] = 32767;
}

void FirParams::setRoundoff(int value)
{
    if (value < 0)
        value = 0;
    if (value > getRoundoffMax())
        value = getRoundoffMax();
    roundoff = value;
}

DominoSettings::DominoSettings()
    // EVADC only
    // writecellnumber(1024),
    // analogDelay(1600)
{
    trigsetup.all = 0;
    reg_pca.all = 0;
}

bool DominoSettings::operator==(const DominoSettings &r) const
{
    return true &&
            chEn == r.chEn &&
    chBaseline == r.chBaseline &&
    chTrigEn == r.chTrigEn &&
    chTrigThr == r.chTrigThr &&
    chZsThr == r.chZsThr &&

    devIndex == r.devIndex &&
    enabled == r.enabled &&

    nch == r.nch &&
    run == r.run &&
    skipRestartRun == r.skipRestartRun &&
    readCellNumber == r.readCellNumber &&
    trigDelay == r.trigDelay &&
    invertInput == r.invertInput &&
    zs_en == r.zs_en &&
    invert_thr_trig == r.invert_thr_trig &&
    invert_zs_thr == r.invert_zs_thr &&
    softwareZs == r.softwareZs &&
    serDesDelayPreset == r.serDesDelayPreset &&
    trigsetup.all == r.trigsetup.all &&
    reg_pca.all == r.reg_pca.all &&
    ad5622op == r.ad5622op &&
    amp_board_set == r.amp_board_set &&
    sparseParams == r.sparseParams &&
    dspParams == r.dspParams &&
    mstreamEnable == r.mstreamEnable &&
    mtuSize == r.mtuSize &&
    periodicalAdcDelayCheck == r.periodicalAdcDelayCheck &&
    key == r.key;
}


void DominoSettings::copyFrom(const DominoSettings &ds)
{
    DominoSettings tmp = ds;
    tmp.devIndex = devIndex;
    tmp.chBaseline = chBaseline;
    tmp.enabled = enabled;
    tmp.nch = nch;

    *this = tmp;
}

void DominoSettings::setNCh(int n)
{
    nch = n;
    for(int i=0; i<n; i++) {
        if (!chEn.contains(i))
            chEn[i] = true;
        if (!chBaseline.contains(i))
            chBaseline[i] = 0;
        if (!chTrigEn.contains(i))
            chTrigEn[i] = true;
        if (!chTrigThr.contains(i))
            chTrigThr[i] = 100;
        if (!chZsThr.contains(i))
            chZsThr[i] = -0x8000;
    }
}

void DominoSettings::readFirPreset(const QString &firPreset, QSettings &s)
{
    const QString sectionKey = QString(SECTION_PREFIX_FIR).append(firPreset);

    if(!firPreset.isEmpty() && s.childGroups().contains(sectionKey)){
        dspParams.fir.presetKey = firPreset;
        s.beginGroup(sectionKey);
        for(int i=0;;i++) {
            QString key = QString("coef.%1").arg(i);
            if (!s.contains(key)) break;
            dspParams.fir.coef[i] = s.value(key).toInt();
        }
        s.endGroup();
    } else {
        // Set fir coef by default
        dspParams.fir.presetKey = "";
        FirParams fp;
        dspParams.fir.coef = fp.coef;
    }
}

void DominoSettings::set_en_all(bool state)
{
    for(int i=0; i<nch; i++)
        chEn[i] = state;
}

void DominoSettings::setInvert(bool invert)
{
    if(invertInput == invert)
        return;

    invertInput = invert;
    invert_thr_trig = !invert_thr_trig;
    invert_zs_thr = !invert_zs_thr;
    foreach (int ch, chBaseline.keys())
        chBaseline[ch] = chBaseline[ch]*(-1);
    foreach (int ch, chTrigThr.keys())
        chTrigThr[ch] = chTrigThr[ch]*(-1);
    foreach (int ch, chZsThr.keys())
        chZsThr[ch] = chZsThr[ch]*(-1);
}

QString DominoSettings::getSettingsKey() const
{
    return QString("%1%2_%3").arg(SECTION_PREFIX_SETUP)
            .arg(devIndex.getSerial(), 8, 16, QChar('0'))
            .arg(devIndex.getDevId(), 2, 16, QChar('0'));
}

DeviceIndex DominoSettings::getIndexBySettingsKey(const QString &key)
{
    QStringList subStr = key.split('_');
    if(subStr.size()==3 && subStr.first() == "Setup"){
        bool ok;
        quint64 ser = subStr[1].toULongLong(&ok, 16);
        if(!ok) return DeviceIndex();
        quint16 id = subStr[2].toUShort(&ok, 16);
        if(ser && id && ok) {
            return DeviceIndex(static_cast<DeviceId>(id), ser);
        }
    }
    return DeviceIndex();
}

QStringList DominoSettings::getFirPresets(QSettings &s)
{
    QStringList presets;
    foreach (const QString &section, s.childGroups()) {
        if(section.startsWith(SECTION_PREFIX_FIR))
            presets << QString(section).remove(0, SECTION_PREFIX_FIR.length());
    }
    return presets;
}

QString DominoSettings::toString() const
{
    QStringList lines;
    QString chMaskStr;
    QPair<int,int> range;
    range.first = -1;
    int enabledChCnt=0;
    foreach (int ch, chEn.keys()) {
        if(chEn[ch]) {
            ++enabledChCnt;
//            chMask.append(" ").append(QString::number(ch));
            if(range.first == -1) {
                //first range
                range.first=ch;
                range.second=ch;
            } else if(range.second+1 == ch) {
                // continuation of range
                range.second=ch;
            } else {
                // new range
                if(range.first == range.second)
                    chMaskStr.append(QString(" %1").arg(range.first));
                else
                    chMaskStr.append(QString(" %1-%2").arg(range.first).arg(range.second));
                range.first=ch;
                range.second=ch;
            }
        }
    }
    // Flush last range if exists
    if(range.first != -1) {
        if(range.first == range.second)
            chMaskStr.append(" ").append(QString::number(range.first));
        else
            chMaskStr.append(QString(" %1-%2").arg(range.first).arg(range.second));
    }

//    qDebug()<<QString("number of ch:%1").arg(dominoSettingnch);
    lines<<QString("number of ch:%1").arg(nch);

//    qDebug()<<"chMask:"<<chMask;
    lines<<QString("chMask:%1; total:%2 channels").arg(chMaskStr).arg(enabledChCnt);

    QString trig("trigger:");
    if(trigsetup.bit.ttl_en)
        trig.append("ttl ");
    if(trigsetup.bit.threshold_en) {
        QString thigCh;
        foreach (int ch, chEn.keys()) {
            if(chEn[ch] && chTrigEn[ch]) {
                thigCh.append(QString("%1:%2 ").arg(ch).arg(chTrigThr[ch]));
            }
        }
        trig.append(QString("thr (%1)").arg(thigCh.trimmed()));
    }
    if(trigsetup.bit.timer_en)
        trig.append("timer ");
//    qDebug()<<trig;
    lines<<trig;

    switch (amp_board_set) {
    case AMP_BOARD_NONE:
//        qDebug()<<"amp: NONE";
        lines<<"amp: NONE";
        break;
    case AMP_BOARD_PCA:
//        qDebug()<<"amp: PCA";
        lines<<"amp: PCA";
        break;
    case AMP_BOARD_AD5622:
        qDebug()<<"amp: AD5622";
        lines<<"amp: AD5622";
//        qDebug()<<QString("ad5622 setup: baseline=%1; level=%2")
//                  .arg(dominoSettingad5622op.data_baseline)
//                  .arg(dominoSettingad5622op.data_level);
        lines<<QString("ad5622 setup: baseline=%1; level=%2")
                  .arg(ad5622op.data_baseline)
                  .arg(ad5622op.data_level);
        break;
    case AMP_BOARD_AD5622v2:
//        qDebug()<<"amp: AD5622v2";
//        qDebug()<<QString("ad5622 setup: baseline=%1; level=%2")
//                  .arg(dominoSettingad5622op.data_baseline)
//                  .arg(dominoSettingad5622op.data_level);
        lines<<"amp: AD5622v2";
        lines<<QString("ad5622 setup: baseline=%1; level=%2")
                  .arg(ad5622op.data_baseline)
                  .arg(ad5622op.data_level);
        break;
    default:
        break;
    }

//    qDebug()<<"readcellnumber:"<<dominoSettingreadcellnumber;
    lines<<QString("read cellNumber:%1; trigDelay:%2")
           .arg(sparseParams.en ?
                    QString::number(sparseParams.readCellNum).append("(sparse)")
                  : QString::number(readCellNumber))
           .arg(trigDelay);

    QString sparseStr("Sparse:");
    if(sparseParams.en){
        sparseStr.append(QString("offset:%1; period:%2; pointNum:%3")
                         .arg(sparseParams.offset)
                         .arg(sparseParams.period)
                         .arg(sparseParams.pointNumber));
    } else
        sparseStr.append("disabled");
    lines<<sparseStr;

    QString dspStr("DSP:");
    if(dspParams.en){
        const QString mafStr = QString("tapSel:%1,blcThr:%2").arg(dspParams.maf_tap_sel).arg(dspParams.blc_thr);
        dspStr.append(QString("roundoff:%1; fir:%2; maf:%3; %4")
                         .arg(dspParams.fir.roundoff)
                         .arg(dspParams.fir.en?dspParams.fir.presetKey:"disabled")
                      .arg(dspParams.maf_enable?mafStr : "disabled"));
        if(dspParams.test_enable)
            dspStr.append(" (test enabled)");
    } else
        dspStr.append("disabled");
    lines<<dspStr;

    lines<<QString("periodicalAdcDelayCheck:%1").arg(periodicalAdcDelayCheck);

    return lines.join("\r");
}
