/*
**    Copyright 2010-2013 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef DOMINOSETTINGS_H
#define DOMINOSETTINGS_H

#include <QMap>
#include <QMetaType>
#include <QVariant>
#include <QVector>

#include "base/DeviceIndex.h"
#include "device-modules/DeserializersControl/AdcSerDesDelay.h"

class QSettings;

typedef union reg_trig_ctrl {
    quint16 all = {};
    struct {
        unsigned int timer_en: 1;
        unsigned int threshold_en: 1;
        unsigned int ttl_en: 1;
        unsigned int unused2: 13;
    } bit;
} reg_trig_ctrl;

typedef union reg_pca_struct {
    quint16 all = {};
    struct {
        unsigned int polarity1: 1;
        unsigned int shaper1_1: 1;
        unsigned int shaper1_2: 1;
        unsigned int shaper1_3: 1;
        unsigned int gain1_1: 1;
        unsigned int gain1_2: 1;
        unsigned int preamp_en1: 1;
        unsigned int unused1: 1;
        unsigned int polarity2: 1;
        unsigned int shaper2_1: 1;
        unsigned int shaper2_2: 1;
        unsigned int shaper2_3: 1;
        unsigned int gain2_1: 1;
        unsigned int gain2_2: 1;
        unsigned int preamp_en2: 1;
        unsigned int unused2: 1;
    } bit;
} reg_pca_struct;


struct ad5622operation {
    quint16 data_baseline = 0;
    quint16 data_level = 0;
    bool operator==(const ad5622operation &r) const {
        return data_baseline == r.data_baseline &&
                data_level == r.data_level;
    }
/*
    unsigned int power_down_modes: 2;
    unsigned int reserved1: 2;
    unsigned int read_operation: 1;
    unsigned int hw_address: 2;
    unsigned int reserved2: 5;
    unsigned int unused: 8;
*/
};
Q_DECLARE_METATYPE(ad5622operation)

struct SparseParams {
    bool en = false;
    bool sparseAllowed = true; // on false readout all readCellNumber (in adc64 app only)
    quint16 readCellNum = 4;
    quint16 offset = 0;
    quint16 period = 2;
    quint8 pointNumber = 0;
    bool operator==(const SparseParams &r) const {
        return true &&
                en == r.en &&
                sparseAllowed == r.sparseAllowed &&
                readCellNum == r.readCellNum &&
                offset == r.offset &&
                period == r.period &&
                pointNumber == r.pointNumber
                ;
    }
};
Q_DECLARE_METATYPE(SparseParams)

enum {
    AMP_BOARD_NONE       = 0x0,
    AMP_BOARD_PCA        = 0x1, // deprecated
    AMP_BOARD_AD5622     = 0x2,
    AMP_BOARD_AD5622v2   = 0x3
};

typedef QVector<qint16> FirPreset;
Q_DECLARE_METATYPE(FirPreset)

class FirParams
{
public:
    FirParams();
    int getRoundoffMax() const {return 3;}
    void setRoundoff(int value);

    bool operator==(const FirParams &r) const {
        return true &&
                presetKey == r.presetKey &&
                en == r.en &&
                roundoff == r.roundoff &&
                coef == r.coef
                ;
    }
    QString presetKey;
    bool en = false;
    int roundoff = 1;
    FirPreset coef;
};
Q_DECLARE_METATYPE(FirParams)

class DspParams
{
public:
    bool getFirEnable() const { return en && fir.en; }
    bool getMafEnable() const { return en && maf_enable; }
    bool getTestEnable() const { return en && test_enable; }
    uint getMafType() const { uint v=0; if(maf_enable) v+=1; if(test_enable) v+=2; return v; }

    bool operator==(const DspParams &r) const {
        return true &&
                fir == r.fir &&
                en == r.en &&
                blc_thr == r.blc_thr &&
                maf_enable == r.maf_enable &&
                maf_tap_sel == r.maf_tap_sel &&
                test_enable == r.test_enable
                ;
    }
    FirParams fir;
    bool en = false;
    int blc_thr = 100;
    bool maf_enable = false;
    int maf_tap_sel = 2;
    bool test_enable = false;
};
Q_DECLARE_METATYPE(DspParams)

class DominoSettings
{
public:
    DominoSettings();
    bool operator==(const DominoSettings &r) const;

    void copyFrom(const DominoSettings &ds);
    void setNCh(int n);
    void readFirPreset(const QString &firPreset, QSettings &s);
    void set_en_all(bool state);
    void setInvert(bool invert);
    QString getSettingsKey() const;
    static DeviceIndex getIndexBySettingsKey(const QString &key);
    static QStringList getFirPresets(QSettings &s);
    QString toString() const;

    QMap<int, bool> chEn;
    QMap<int, int> chBaseline;  //device serial id - channel number - baseline
    QMap<int, bool> chTrigEn;
    QMap<int, int> chTrigThr;
    QMap<int, int> chZsThr;

    DeviceIndex devIndex;
    bool enabled = false;

    int nch = 0; // FIXME: remove nch
    bool run = false;
    bool skipRestartRun = false;
    int readCellNumber = 1024;
    int trigDelay = 5;
    bool invertInput = false;
    bool zs_en = false;
    bool invert_thr_trig = false;
    bool invert_zs_thr = false;
    bool softwareZs = false;
    DevSerDesDelay serDesDelayPreset;
    reg_trig_ctrl trigsetup = {};
    reg_pca_struct reg_pca = {};
    ad5622operation ad5622op; // rename to preamplifier
    quint16 amp_board_set = 0; // to enum, move to ad5622operation
    SparseParams sparseParams;
    DspParams dspParams;
    bool mstreamEnable = false;
    bool periodicalAdcDelayCheck = false;
    int mtuSize = 1500;
    int key = 0; // to distinguish evolution of setup
};
Q_DECLARE_METATYPE(DominoSettings)
Q_DECLARE_METATYPE(DeviceId) // TODO: toDel???

#endif // DOMINOSETTINGS_H
