/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceTableWidget.h"

#include <QDateTime>
#include <QHeaderView>
#include <QHostAddress>

#include "adc64-common/AdcStatus.h"
#include "Adc64SystemConfig.h"
#include "mldiscover/LldpDiscovery.h"

DeviceTableWidget::DeviceTableWidget(QWidget *parent) :
    QTableWidget(parent),
    config(new Adc64SystemConfig)
{
    setSortingEnabled(true);
    setTableLabels();
}

DeviceTableWidget::~DeviceTableWidget()
{
    delete config;
}

void DeviceTableWidget::clear()
{
    setRowCount(0);
}

void DeviceTableWidget::insertRow(const DeviceIndex &index)
{
    if(getRowByIndex(index)!= -1)
        return;

    blockSignals(true);

    int row = rowCount();
    setRowCount(row+1);
    for(int col=0; col<COLUMN_NUMBER; ++col) {
        setItem(row, col, new QTableWidgetItem());
        Qt::ItemFlags eFlags = item(row, col)->flags();
        eFlags &= ~Qt::ItemIsEditable;
        item(row, col)->setFlags(eFlags);
    }

    // Pre-connection setup
    item(row, COLUMN_DEV_ID_STR)->setText(getDeviceTypeName(index.getDevId()));
    item(row, COLUMN_SERIAL_STR)->setText(index.getSerialStr());
    item(row, COLUMN_DEV_SER)->setText(QString::number(index.getSerial()));
    item(row, COLUMN_DEV_ID)->setText(QString::number(index.getDevId()));

    adjustTableHeight();

    blockSignals(false);
}

void DeviceTableWidget::removeRow(const DeviceIndex &index)
{
    int row = getRowByIndex(index);
    if(row==-1)
        return;
    QTableWidget::removeRow(row);
}

void DeviceTableWidget::setConfig(const Adc64SystemConfig &newConfig)
{
    if (*config == newConfig)
        return;
    *config = newConfig;

    if (config->show_details) {
        showColumn(COLUMN_UPTIME);
        showColumn(COLUMN_LM_RESET);
    } else {
        hideColumn(COLUMN_UPTIME);
        hideColumn(COLUMN_LM_RESET);
    }
}

DeviceIndex DeviceTableWidget::getIndexByRow(int row) const
{
    auto id = item(row, COLUMN_DEV_ID)->text().toUInt();
    auto serial = item(row, COLUMN_DEV_SER)->text().toULongLong();
    return DeviceIndex(static_cast<DeviceId>(id), serial);
}

int DeviceTableWidget::getRowByIndex(const DeviceIndex &index) const
{
    for(int row=0; row < rowCount(); ++row){
        if(getIndexByRow(row)==index)
            return row;
    }
    return -1;
}

void DeviceTableWidget::configDbSyncResult(const ConfigDbSyncResult &synchResult)
{
    for (auto it = synchResult.begin(); it != synchResult.end(); ++it) {
        auto index = it.key();
        int row = getRowByIndex(index);
        if (!synchResult[index]) {
            item(row, COLUMN_SYNCH_STATUS)->setText("Error");
            item(row, COLUMN_SYNCH_STATUS)->setBackground(QColor(255, 153, 153));
        } else{
            item(row, COLUMN_SYNCH_STATUS)->setText("Ok");
            item(row, COLUMN_SYNCH_STATUS)->setBackground(QColor(102, 204, 102));
        }
    }
}

void DeviceTableWidget::updateAdcStatus(const DeviceIndex &index, const AdcStatus &status)
{
    const int row = getRowByIndex(index);
    if(row==-1)
        return;

    const QColor bgColorDefault = style()->standardPalette().light().color();
    const QColor bgColorWarning = Qt::yellow;
    const QColor bgColorError = Qt::red;

    QTableWidget * const table = this;
    setUpdatesEnabled(false);
    // read temperature sensors
    {
        QStringList sl;
        if (status.temp > -255)
            sl += QString("ADC:%1").arg((int)status.temp);
        QMap<int, QString> name;
        name[0] = "PS";
        name[1] = "AMP";
        for (int i=0; i<status.bmStatus.sensors.size(); i++) {
            double t = status.bmStatus.sensors[i].temp;
            if (t > -255)
                sl += QString("%1:%2").arg(name[i]).arg((int)t);
        }
        table->item(row, COLUMN_TEMP)->setText(sl.join(" "));
    }
    table->item(row, COLUMN_EVENT)->setText(QString::number(status.evNum));

    table->item(row, COLUMN_ADC_STATE_OK)->setText(status.adcSelfTestOk?"Ok":QString("Error: %1").arg(status.adcDesStatus,8,2,QChar('0')));
    table->item(row, COLUMN_ADC_STATE_OK)->setBackground(status.adcSelfTestOk ? bgColorDefault : bgColorError);

    table->item(row, COLUMN_WR_LINK_ERROR)->setText(QString::number(status.wrEpStatus.wrLinkError));
    table->item(row, COLUMN_WR_SYNC_LOST)->setText(QString::number(status.wrEpStatus.wrSyncFail));

    bool wrTimeValid = false;
    const WrEpStatus &wrEpStatus = status.wrEpStatus;
    if (!wrEpStatus.hasWrEpStatus) {
        wrTimeValid = status.adcStatus & 0x2;
        table->item(row, COLUMN_WR_TIME)->setText(wrTimeValid?"Ok":"Invalid");
        table->item(row, COLUMN_WR_TIME)->setBackground(wrTimeValid ? bgColorDefault : bgColorError);
        table->item(row, COLUMN_WR_RX_CODE_ERROR)->setText("N/A");
        table->item(row, COLUMN_WR_TIME_OFFSET)->setText("N/A");
        table->item(row, COLUMN_WR_TIME_OFFSET)->setBackground(bgColorDefault);
    } else {
        wrTimeValid = status.wrEpStatus.timeValid();
        const bool wrTimeNull = status.wrEpStatus.timeNull();
        if (wrTimeNull) {
            table->item(row, COLUMN_WR_TIME)->setText(QString());
            table->item(row, COLUMN_WR_TIME)->setBackground(bgColorWarning);
            table->item(row, COLUMN_WR_TIME_OFFSET)->setText(QString());
            table->item(row, COLUMN_WR_TIME_OFFSET)->setBackground(bgColorWarning);
        } else {
            QDateTime wrTime = QDateTime::fromMSecsSinceEpoch(status.wrEpStatus.mSecsSinceEpoch());

            int wrTimeCheck = status.wrEpStatus.wrTimeCheck();
            double wrTimeOffsetMs = status.wrEpStatus.mSecsSinceEpoch() - status.wrEpStatus.acquiredAtMs;
            QColor offsetBgColor = bgColorDefault;
            if (wrTimeCheck == 1)
                offsetBgColor = bgColorWarning;
            if (wrTimeCheck > 1)
                offsetBgColor = bgColorError;

            table->item(row, COLUMN_WR_TIME)->setText(wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss")); // add ".zzz" for ms
            table->item(row, COLUMN_WR_TIME)->setBackground(wrTimeValid ? bgColorDefault : bgColorError);
            QString strOffset;
            if (wrTimeOffsetMs/1000. > -60 && wrTimeOffsetMs/1000. < 60)
                strOffset = QString(" %1").arg(wrTimeOffsetMs * 1e-3, 6, 'f', 3);
            table->item(row, COLUMN_WR_TIME_OFFSET)->setText(strOffset);
            table->item(row, COLUMN_WR_TIME_OFFSET)->setBackground(offsetBgColor);
        }
        table->item(row, COLUMN_WR_RX_CODE_ERROR)->setText(QString::number(status.wrEpStatus.wrCodeError));
        table->item(row, COLUMN_UPTIME)->setText(QString::number(status.wrEpStatus.wrUptime));
    }
    table->item(row, COLUMN_LM_RESET)->setText(QString::number(status.liveMagicResetCount));

    table->item(row, COLUMN_TRIG_ON_XOFF)->setText(QString::number(status.trigOnXOff));

    setUpdatesEnabled(true);
}

void DeviceTableWidget::lldpCacheUpdated(const LldpInfoCache &cache)
{
    for (auto it = cache.begin(); it != cache.end(); ++it) {
        updateDeviceDiscoverInfo(it.key(), it.value());
    }
}

void DeviceTableWidget::updateDeviceDiscoverInfo(const DeviceIndex &index, const LldpInfo &lldpInfo)
{
    int row = getRowByIndex(index);
    if(row == -1){
        return;
    }

    // Q_ASSERT(config->device_configs.contains(index));
    QHostAddress newHost(lldpInfo.ip_addr);
    QString prevHostStr = item(row, COLUMN_IP)->text();
    QString newHostStr = newHost.toString();

    if(prevHostStr != newHostStr) {
        if(prevHostStr.isEmpty())
            qDebug() << QString("[%1] host address set %2")
                        .arg(index.getIdent()).arg(newHostStr);
        else
            qInfo() << QString("[%1] host changed from %2 to %3")
                       .arg(index.getIdent()).arg(prevHostStr).arg(newHostStr);

        item(row, COLUMN_IP)->setText(newHostStr);
    }
    item(row, COLUMN_SLOT)->setText(QString::number(lldpInfo.chassis_slot));
}

void DeviceTableWidget::setTableLabels()
{
    using T = QPair<Columns, QString>;
    QList<T> columns({
                         {COLUMN_DEV_ID_STR, tr("Type")},
                         {COLUMN_DEV_ID_STR, tr("Type")},
                         {COLUMN_SERIAL_STR, tr("Serial")},
                         {COLUMN_SLOT, tr("Slot")},
                         {COLUMN_IP, tr("IP Address")},
                         {COLUMN_TEMP, tr("t, %1C").arg(QChar(0x02da))},
                         {COLUMN_EVENT, tr("Event")},
                         {COLUMN_TRIG_ON_XOFF, tr("Trig on XOff")},
                         {COLUMN_SYNCH_STATUS, tr("Synch to ADC64")},
                         {COLUMN_ADC_STATE_OK, tr("ADC Status")},
                         {COLUMN_WR_TIME, tr("WR Time")},
                         {COLUMN_WR_TIME_OFFSET, tr("Offset")},
                         {COLUMN_WR_LINK_ERROR, tr("Link")},
                         {COLUMN_WR_SYNC_LOST, tr("Unlock")},
                         {COLUMN_WR_RX_CODE_ERROR, tr("RX Err")},
                         {COLUMN_MS_EV_RATE, tr("Event rate")},
                         {COLUMN_MS_DATA_RATE, tr("Data rate")},
                         {COLUMN_DEV_SER, tr("Serial Num")},
                         {COLUMN_DEV_ID, tr("type")},
                         {COLUMN_UPTIME, tr("up time")},
                         {COLUMN_LM_RESET, tr("LM resets")}
                     });
       setColumnCount(columns.size());
    for (auto i : columns)
        setHorizontalHeaderItem(i.first, new QTableWidgetItem(i.second));
    hideColumn(COLUMN_DEV_SER);
    hideColumn(COLUMN_DEV_ID);
    if(!config->show_details){
        hideColumn(COLUMN_UPTIME);
        hideColumn(COLUMN_LM_RESET);
    }

//    setHorizontalHeaderLabels(columnHeaders);
#if QT_VERSION >= 0x050000
    horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
}

void DeviceTableWidget::adcStateChanged(const DeviceIndex &index, bool online)
{
    const int row = getRowByIndex(index);
    Q_ASSERT(row!=-1);
    const QColor bgColorDefault = style()->standardPalette().light().color();
    for(int col=0;col<COLUMN_NUMBER;++col){
        item(row, col)->setBackground(online ? bgColorDefault : Qt::lightGray);
    }
}


void DeviceTableWidget::adjustTableHeight()
{
    int height = horizontalHeader()->height();
    for (int i=0; i < rowCount(); i++) {
        height += rowHeight(i);
    }
    height += 5;
    setMinimumHeight(height);
    QRect geo = geometry();
    geo.setHeight(height);
    setGeometry(geo);
}
