/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2010-2020 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "Adc64SystemConfig.h"
#include "Adc64SystemConfigDb.h"
#include "Adc64SystemConfigDispatcher.h"
#include "DeviceSettingsDialog.h"
#include "DeviceTableWidget.h"
#include "MainWindow_p.h"
#include "StopConditionsDialog.h"
#include "WrDaqDialog.h"
#include "WrDaqDialog.h"
#include "adc64-common/AdcDeviceProxy.h"
#include "adc64-common/DeviceStateController.h"
#include "adc64-common/DominoDeviceFactory.h"
#include "adc64-common/dominodevice.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "device-discover/DiscoverDialog.h"
#include "m-stream-output/StatisticOutput.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mongo/StartupOptions.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "remote-control-server/RunControlCommands.h"
#include "fsm/AdcDeviceMgrFsm.h"

#include <cstdlib>
#include <QDateTime>
#include <QDebug>
#include <QDialogButtonBox>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QFileInfo>
#include <QHBoxLayout>
#include <QHostAddress>
#include <QHostInfo>
#include <QInputDialog>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QMessageBox>
#include <QNetworkInterface>
#include <QScrollBar>
#include <QShortcut>
#include <QSignalMapper>
#include <QSpinBox>
#include <QStringList>
#include <QThread>
#include <QTimer>

namespace {

const int REMOTE_CONTROL_PORT = 0; // 33392 adc64-system
const int WHOLE_TIMEOUT_MS = 15000;

const QMap<QString, quint64 MStreamStat::*> MS_WARN_MAP =
{
    {"fragDroped", &MStreamStat::fragDroped},
    {"fragMissed", &MStreamStat::fragMissed},
    {"missedEvents", &MStreamStat::missedEvents},
    {"missedEventsRanges", &MStreamStat::missedEventsRanges}
};
}

Q_DECLARE_METATYPE(Qt::GlobalColor)
Q_DECLARE_METATYPE(QHostAddress)
Q_DECLARE_METATYPE(QtMsgType)

MainWindow::MainWindow(const ConfigSelector &cs, QWidget *parent) :
    BaseMainWindow(parent),
    d_ptr(new MainWindowPrivate),
    ui(new Ui::MainWindow),
    labelFsmState(new QLabel(this)),
    labelRunState(new QLabel(this)),
    wrDaqDialog(new WrDaqDialog(this)),
    signalRouter(new Adc64SystemSignalRouter()),
    configDb(new Adc64SystemConfigDb(cs, this)),
    configDispatcher(new Adc64SystemConfigDispatcher(this)),
    deviceStateController(new DeviceStateController(this)),
    configuration_manager(new ConfigurationManager(this, cs.getType())),
    deviceDispatcher(new DeviceDispatcher(this)),
    remoteControl(new RemoteControlServer(this, REMOTE_CONTROL_PORT)),
    pnpServer(new PNPServer(this)),
    discoverDialog(new DiscoverDialog(this)),
    setupKey(1),
    statFlushTimer(new QTimer(this)),
    warnFlushTimer(new QTimer(this)),
    deviceSettingsDialog(new DeviceSettingsDialog(this)),
    evNumChecker(new EvNumChecker(this)),
    fsm(new AdcDeviceMgrFsm(this))
{
    qRegisterMetaType<MStreamStat>();
    qRegisterMetaType<QtMsgType>();
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<AdcStatus>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<DominoSettings>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<RcRunId>();
    qsrand(QDateTime::currentDateTime().toTime_t());

    signalRouter->setMainWindow(this);
    signalRouter->setConfigDb(configDb);
    signalRouter->setPnpServer(pnpServer);
    signalRouter->setRemoteControlServer(remoteControl);
    signalRouter->setConfigDispatcher(configDispatcher);
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->connectAll();

    connect(fsm, &BaseFsm::stateChanged, this, &MainWindow::onFsmStateChange);
    connect(fsm, &BaseFsm::stateChanged, remoteControl, &RemoteControlServer::onFsmStateChange);

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);
    statusBar()->addWidget(labelFsmState);
    statusBar()->addWidget(labelRunState);

    configuration_manager->setDefaultConfig(Adc64SystemConfig());

    objects_to_block.append(ui->actionAllow_Remote_Control);
    objects_to_block.append(ui->actionEnable_write);
    objects_to_block.append(ui->actionStop_run_on_TrigOnXOff);
    objects_to_block.append(ui->actionShow_details);

    configDispatcher->subscribe(wrDaqDialog);
    configDispatcher->subscribe(ui->table);
    configDispatcher->subscribe(ui->adcStatusWidget);

    connect(deviceStateController, &DeviceStateController::deviceStateChanged,
            this, &MainWindow::adcStateChanged);
    connect(deviceStateController, &DeviceStateController::deviceStateChanged,
            ui->table, &DeviceTableWidget::adcStateChanged);

    new ThreadController(remoteControl, this);

    QSet<DeviceId> devices;
    devices << DEVICE_ID_ADC64
            << DEVICE_ID_ADC64S2
            << DEVICE_ID_ADC64VE
            << DEVICE_ID_ADC64VE_XGE
            << DEVICE_ID_ADC64VE2
            << DEVICE_ID_ADC64ECAL
            << DEVICE_ID_ADC64WR
            << DEVICE_ID_ADC8BEDRS
            << DEVICE_ID_ADC8E
            << DEVICE_ID_EVADC
            << DEVICE_ID_TQDC16VS_DIG
            << DEVICE_ID_ADCM163
            << DEVICE_ID_ADC8E;

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter(devices);
    lldpDiscovery.subscribe(discoverDialog);
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(ui->table);
    lldpDiscovery.subscribe(ui->adcStatusWidget);

    deviceDispatcher->subscribeDevices(this);
    deviceDispatcher->subscribeDevices(discoverDialog);
    deviceDispatcher->subscribeDevices(ui->adcStatusWidget);
    deviceDispatcher->subscribeDevices(fsm);

#ifndef QT_DEBUG
    ui->adcStatusWidget->hide();
#endif

//    connect(discoverDialog, &DiscoverDialog::devicesUpdated,
//            this, &MainWindow::selectDevices);

    new ThreadController(pnpServer, this);

    connect(statFlushTimer, &QTimer::timeout,
            this, &MainWindow::flushMStreamStats);
    statFlushTimer->start(200);
    connect(warnFlushTimer, &QTimer::timeout,
            this, &MainWindow::flushWarnings);
    warnFlushTimer->setInterval(1000);
    warnFlushTimer->setSingleShot(true);

    updateProgramDescription();
    updateGui();
}

MainWindow::~MainWindow()
{
    delete signalRouter;
    emit deleteAll();
    QElapsedTimer timer;
    timer.start();

    for(QThread *thread : adcThreads){
        if(!threadQuiting(thread, WHOLE_TIMEOUT_MS, &timer)){
            qDebug()<<"adc thread failed to quit quietly";
        }
    }

    for(QThread *thread : dumpersThread){
        if(!threadQuiting(thread, WHOLE_TIMEOUT_MS, &timer)){
            qDebug()<<"dumpers thread to quit quietly";
        }
    }
//    delete discoverDialog;
    delete configDb;
    delete ui;
    delete d_ptr;
}

void MainWindow::updateGui()
{
    Q_D(MainWindow);

    deviceSettingsDialog->setMStreamWrite(d->config.storage_path);
    deviceSettingsDialog->setMtuSize(d->config.mtu_size);
    deviceSettingsDialog->setStopConditions(d->config.event_number_limit);

    block_gui_signals(true);

    ui->actionWR_DAQ->setVisible(Globals::getDebugMode());
    ui->actionAllow_Remote_Control->setChecked(d->config.allow_remote_control);
    emit allowRemoteControl(d->config.allow_remote_control);
    ui->actionEnable_write->setChecked(d->config.write_enable);
    ui->actionShow_details->setChecked(d->config.show_details);
    ui->actionStop_run_on_TrigOnXOff->setChecked(d->config.trig_on_xoff_stop);
    block_gui_signals(false);
}

void MainWindow::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    labelFsmState->setText("FSM: " + fsmStateToString(state));

    if(oldState == FsmState::Stopping) {
        updateProgramDescription();
    }

    switch (state) {
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:{}
        }
        break;
    }
    case FsmState::Starting :
    {
        setRunState(true);
        sendResetAll(true, true);
        emit clearStatistics(DeviceIndex());
        fsm->goState(FsmState::Run);
        break;
    }
    case FsmState::Run :
    {
        Q_D(MainWindow);
        qInfo() << QString("Start completed successfully");

        const auto &j = QVariant::fromValue(d->config).value<QJsonObject>();
        emit config_transfer_to_rc(QJsonDocument(j).toJson());
        updateProgramDescription();
        d->runId = RcRunId();
        break;
    }
    case FsmState::Stopping :
    {
        bool success = true;
        //TODO: deal with phasingStop();
//        success &= phasingStop();
//        if(!success)
//            qWarning() << "phasingStop failed";
        //TODO: deal with waitDevEvNum();
//        success &= waitDevEvNum(timer, 1000);
//        if(!ok)
//            qWarning() << "check for lastEvNum failed";
        setRunState(false);
        sendResetAll(true, false);
//        fsm->goState(success ? FsmState::Ready : FsmState::Error);
        fsm->goState(FsmState::Ready);
        break;
    }
    case FsmState::Calibrate :
    {
        Q_UNREACHABLE(); // will be after adc64/adc64-system merging
        incrementSetupKey();
        break;
    }
    case FsmState::FindAdcDelay :
    {
        Q_UNREACHABLE(); // will be adc64/adc64-system merging
        incrementSetupKey();
        break;
    }
    case FsmState::Error :
    {
        if(oldState != FsmState::Init)
            qWarning("Error while executing %s",
                     fsmStateToString(oldState).toStdString().c_str());
        setRunState(false);
        sendResetAll();
        break;
    }
    default:{}
    }
}

DeviceIndexEnabledMap configToDieMap(const Adc64SystemConfig &config)
{
    DeviceIndexEnabledMap map;
    for (const auto &i : config.device_configs) {
        map.insert(i.devIndex, DeviceIndexEnabled(i.devIndex, i.enabled));
    }
    return map;
}

void MainWindow::setConfig(const Adc64SystemConfig &config)
{
    Q_D(MainWindow);

    if (config == d->config)
        return;
    const auto oldMap = configToDieMap(d->config);
    const auto newMap = configToDieMap(config);

    d->config = config;
    if (oldMap != newMap) {
        emit deviceListChanged(newMap);
    }
    updateProgramDescription();
    updateGui();

    configDb->synchronize_with_adc64(&d->config);
    emit configChanged(config);
}

void MainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    const auto &cs = ConfigSelector(DaqModule(DaqModule::Type::Adc64System), conf);
    const auto &title = cs.toWindowTitle();
    if(windowTitle() != title)
        setWindowTitle(title);
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void MainWindow::on_actionSynchronize_configs_with_adc64_triggered()
{
    Q_D(MainWindow);

    configDb->synchronize_with_adc64(&d->config);
    emit configChanged(d->config);
}

void MainWindow::configDbSyncResult(const ConfigDbSyncResult &result)
{
    synch_result = result;
}

void MainWindow::addLog(QtMsgType msgType, const QString &log, const QString &threadName)
{
    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    Qt::GlobalColor color = Qt::black;
    switch (msgType) {
    case QtFatalMsg:;
        break;
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:;
        color = Qt::blue;
        break;
    case QtDebugMsg:;
        break;
    }
    tf.setForeground(QBrush(color));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    QString str = log;
    if(!threadName.isEmpty())
        str.prepend(QString("[thread:%1] ").arg(threadName));
    str.prepend(QString("[%1] ").arg(QTime::currentTime().toString()));
    ui->plainTextEdit->appendPlainText(str);
}


void MainWindow::on_actionWR_DAQ_triggered()
{
    wrDaqDialog->show();
}

DeviceIndexEnabledMap dievFromConfigs(const QMap<DeviceIndex, DominoSettings> &configs)
{
    DeviceIndexEnabledMap container;
    for(auto it = configs.begin(); it != configs.end(); ++it) {
        DeviceIndexEnabled item(it.key(), it.value().enabled);
        container.insert(item.index, item);
    }
    return container;
}

DeviceIndexSet indicesFromList(const QList<DeviceIndex> &list)
{
    return makeDeviceIndexSet(list);
}

DeviceIndexSet indicesFromDiev(const DeviceIndexEnabledMap &v)
{
    DeviceIndexSet set;
    for (const auto &i : v) {
        set.insert(i.index);
    }
    return set;
}

DeviceIndexEnabledMap deviceListFromConfig(const Adc64SystemConfig &config)
{
    DeviceIndexEnabledMap result;
    for (const auto &it : config.device_configs) {
        result.insert(it.devIndex, DeviceIndexEnabled(it.devIndex, it.enabled));
    }
    return result;
}

void MainWindow::setDeviceList(const DeviceIndexEnabledMap &newList)
{
    Q_D(MainWindow);
    static DeviceIndexEnabledMap oldDevices;
    if (newList == oldDevices)
        return;

    const DeviceIndexSet old_set(indicesFromDiev(oldDevices));
    const DeviceIndexSet new_set(indicesFromDiev(newList));
    oldDevices = newList;

    const auto &toRemove = old_set - new_set;
    const auto &toAdd = new_set - old_set;

    // 1. stop device threads
    removeDevices(toRemove);

    // 2. remove config entries
    for (auto i : toRemove)
        d->config.device_configs.remove(i);

    // 3. add config entries
    for (auto i : toAdd) {
        if(d->config.device_configs.contains(i))
            continue;
        auto &ds =d->config.device_configs[i] = DominoSettings();
        ds.devIndex = i;
    }

    // 4. pull config entry from adc64, if exists
    configDb->synchronize_with_adc64(&d->config);

    // 5. overwrite enabled flags
    for (auto i : newList)
        d->config.device_configs[i.index].enabled = i.enabled;

    emit configChanged(d->config);

    // 6. create device threads
    addDevices();
}

void MainWindow::on_actionFindDevices_triggered()
{
    Q_D(const MainWindow);

    const DeviceIndexSet old_set(indicesFromList(d->config.device_configs.keys()));
    const DeviceIndexEnabledMap old_v(dievFromConfigs(d->config.device_configs));
    discoverDialog->setDeviceList(old_v);
    discoverDialog->show();
}

void MainWindow::on_actionClear_statistics_triggered()
{
    emit clearStatistics(DeviceIndex()); // to All
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool checked)
{
    Q_D(MainWindow);

    d->config.allow_remote_control = checked;
    emit configChanged(d->config);
    emit allowRemoteControl(d->config.allow_remote_control);
    updateProgramDescription();
}

void MainWindow::on_actionEnable_write_toggled(bool enabled)
{
    Q_D(MainWindow);

    if(enabled){
        QString dir = QFileDialog::getExistingDirectory(this, tr("Storage Directory"),
                                                     d->config.storage_path,
                                                     QFileDialog::ShowDirsOnly
                                                     | QFileDialog::DontResolveSymlinks);
        if(dir.isEmpty()){
            ui->actionEnable_write->setChecked(false);
            return;
        }
        d->config.storage_path = dir;
        deviceSettingsDialog->setMStreamWrite(d->config.storage_path);
        qInfo() << QString("Changed storage dir to:").append(d->config.storage_path);
    }
    d->config.write_enable = enabled;
    emit configChanged(d->config);
    qInfo() << QString("MStream write %1.").arg(d->config.write_enable? "enabled" : "disabled");
    updateProgramDescription();
}

void MainWindow::on_actionShow_details_triggered(bool show)
{
    Q_D(MainWindow);

    d->config.show_details = show;
    emit configChanged(d->config);
}

void MainWindow::on_actionSet_stop_conditions_triggered()
{
    Q_D(MainWindow);

    StopConditionsDialog dialog(&d->config, this);
    if(dialog.exec() == QDialog::Accepted){
        deviceSettingsDialog->setStopConditions(d->config.event_number_limit);
        emit configChanged(d->config);
    }
}

void MainWindow::on_actionStop_run_on_TrigOnXOff_toggled(bool trig_on_xoff_stop)
{
    Q_D(MainWindow);

    d->config.trig_on_xoff_stop = trig_on_xoff_stop;
    emit configChanged(d->config);
}

void MainWindow::on_actionShow_device_settings_triggered()
{
    deviceSettingsDialog->show();
}

void MainWindow::on_actionSet_MTU_size_triggered()
{
    Q_D(MainWindow);

    bool ok;
    int val = QInputDialog::getInt(this, "Set MTU size...", "MTU size:", d->config.mtu_size, 1500, 8000, 100, &ok);
    if(!ok)
        return;

    val &= ~3;
    if(d->config.mtu_size != val) {
        qInfo() << QString("MTU size changed from %1 to %2")
                       .arg(d->config.mtu_size).arg(val);
        d->config.mtu_size = static_cast<quint16>(val);
        emit configChanged(d->config);
        deviceSettingsDialog->setMtuSize(d->config.mtu_size);
        sendResetAll(false);
    }
}

void MainWindow::on_radioButtonStart_toggled(bool checked)
{
    const auto &expectedState = checked ? FsmState::Ready : FsmState::Run;
    if(fsm->getState() != expectedState) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(expectedState))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        fsm->goState(FsmState::Error);
        return;
    }

    fsm->goState(checked ? FsmState::Starting : FsmState::Stopping);
}

void MainWindow::onRemoteStartRequest(const RcRunId &runId, quint64 start_event_number)
{
//    Q_D(const MainWindow);
    labelRunState->setText("Run:"+runId.toString());

    ui->plainTextEdit->appendPlainText("");
    qInfo() << QString("Got remote request to start %1").arg(runId.toString());

    static bool warn_start_event_number = true;
    if(start_event_number && warn_start_event_number) {
        qWarning() << "Remote start with start_event_number not implemented yet.";
        warn_start_event_number = false;
    }

    if(rcRunId.runIndex == runId.runIndex &&
            rcRunId.runNumber > runId.runNumber) {
        const auto &msg = QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                              .arg(rcRunId.runNumber)
                              .arg(runId.runNumber);
        qWarning() << msg;
        emit runWarning(rcRunId, msg);
        return;
    }

    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Ready))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit runWarning(rcRunId, msg);
        return;
    }

    rcRunId = runId;
    fsm->goState(FsmState::Starting);

//    if(success) {
//        DevFwRevMap adcFwRev;
//        for(DeviceIndex &i : lastAdcStatus.keys()) {
//            const AdcStatus &status = lastAdcStatus[i];
//            adcFwRev[i] = status.fw_ver;
//        }
////      success &= dbOk;
//    }
}

void MainWindow::onRemoteStopRequest(const RcRunId &runId, quint64 stop_event_number)
{
    static bool warnStopEvNum = true;
    if(stop_event_number && warnStopEvNum) {
        warnStopEvNum = false;
        qWarning() << "Stopping with 'stop_event_number' should be fixed";
    }
    bool success=true;
    if(rcRunId.runIndex != runId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                     .arg(rcRunId.runIndex)
                     .arg(runId.runIndex);
        success = false;
    }
    if(success && rcRunId.runNumber != runId.runNumber){
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                     .arg(rcRunId.runNumber)
                     .arg(runId.runNumber);
        success = false;
    }


    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit runWarning(rcRunId, msg);
        if(fsm->getState() != FsmState::Ready && fsm->getState() != FsmState::Error)
            fsm->goState(FsmState::Error);
        return;
    }
    fsm->goState(FsmState::Stopping);
}

void MainWindow::adcConnected(const DeviceIndex &index)
{
    if(index.isNull())
        return;
    adcStateChanged(index, true);
}

void MainWindow::adcDisconnected(const DeviceIndex &index)
{
    if(index.isNull())
        return;
    adcStateChanged(index, false);
}

void MainWindow::adcStateChanged(const DeviceIndex &index, bool online)
{
    if(online) {
        const QString serialStr = index.getSerialStr(); // QString("0x%1").arg(index.getSerial(), 8, 16, QChar('0'));
        qInfo() << QString("[%1] Device online").arg(serialStr);

//        sendAdcSetup(index);
        emit mstream_send_connect(index);
    } else {
        deviceSetupKeys.remove(index);
        updateRemoteStatus();
        auto mess = QString("[%1] Device offline").arg(index.getSerialStr());
        sendRunError(mess);
    }
}


void MainWindow::updateAdcStatus(const DeviceIndex &index, const AdcStatus &status)
{
    Q_D(const MainWindow);
    bool statusAfterReset = status.setup_key == setupKey;

    if (!devices.contains(index)) return;
//    if(devices[index] != sender()) {
//        // Got status from previous instance of hw
//        // DominoDevice was recreated
//        return;
//    }

    if(!status.liveMagicCorrect && d->config.run) {
        sendRunError("liveMagic incorrect");
    }
    if(statusAfterReset && !status.adcSelfTestOk && d->config.run) {
        sendRunError("AdcSelfTest failed");
    }
    if(statusAfterReset && d->config.trig_on_xoff_stop && status.trigOnXOff && d->config.run) {
        sendRunError("Trig on XOff detected");
        qWarning() << QString("[%1] trigOnXOff=%2").arg(index.getIdent()).arg(status.trigOnXOff);
    }

    //    if(hw->isConnected() && !status.liveMagicCorrect){
    //        AdcSetup &setup = systemSetup.adc[cfgIndex];
    //        qWarning() << QString("[0x%2] Warning: Device Adc-%1 failed magic word check. Resetting latest configuration. (%3)")
    //                      .arg(cfgIndex).arg(setup.serialId, 8, 16, QChar('0'))
    //                      .arg(setup.host.toString());
    //        addProcessLog(row, "==== RESET ====");
    //        dumpers[row]->restart();
    //    }

    /* REWRITE THIS SHIT
    if(d->config.run && d->config.event_number_limit){
        bool stop = true;
        for(int row=0; row < table->rowCount(); ++row){
            bool ok;
            quint64 evCnt = table->item(row, COLUMN_EVENT)->text().toULongLong(&ok);
            stop &= evCnt > d->config.event_number_limit;
            stop &= ok;
        }
        if(stop){
            qInfo() << QString("Stopping by stop condition: evNumLimit=%1").arg(d->config.event_number_limit);
            ui->radioButtonStop->setChecked(true);
        }
    }
    */

    const AdcStatus prevStatus = lastAdcStatus.value(index);
    const bool dumperRestart =
            prevStatus.mstreamVer2 !=status.mstreamVer2 ||
            prevStatus.msMultiAck !=status.msMultiAck ||
            prevStatus.hwBufSize !=status.hwBufSize;
    lastAdcStatus[index] = status;
    if(dumpers.contains(index)){
        if(dumperRestart) {
            resetMStream(index);
        }
    } else {
        addMStreamDump(index);
    }
}

void MainWindow::adcSetupWritten(const DominoSettings &s)
{
    const auto &index = s.devIndex;
    if(index.isNull())
        return;

    fsm->setDevSetupKey(index, s.key);

    const bool send = (!deviceSetupKeys.contains(index) && s.key==setupKey) ||
            ((deviceSetupKeys[index]!=setupKey) == ((s.key!=setupKey)));
    deviceSetupKeys[index] = s.key;
    if(send)
        updateRemoteStatus();
}

void MainWindow::addDevice(const DeviceIndex &index)
{
    Q_D(MainWindow);

    if (devices.contains(index))
        return;

    const DominoSettings &ds = d->config.device_configs[index];

    // Add device
    QThread *separateThread = new QThread();
    adcThreads[ds.devIndex] = separateThread;
    deviceSetupKeys.remove(ds.devIndex);
    separateThread->setObjectName(ds.devIndex.getIdent());
    separateThread->start(QThread::LowPriority);

    DominoDevice *hw = DominoDeviceFactory::create(ds.devIndex);
    deviceDispatcher->subscribeHostAddressUpdates(index, hw);
    connect(hw, &DominoDevice::deviceConnected,
            deviceStateController, &DeviceStateController::deviceConnected);
    connect(hw, &DominoDevice::deviceDisconnected,
            deviceStateController, &DeviceStateController::deviceDisconnected);
    connect(hw, &DominoDevice::statusUpdated,
            this, &MainWindow::updateAdcStatus);
    connect(hw, &DominoDevice::statusUpdated,
            ui->table, &DeviceTableWidget::updateAdcStatus);
    connect(this, &MainWindow::deleteHardware,
            hw, &DominoDevice::deleteHardware);
    connect(this, &MainWindow::deleteAll,
            hw, &DominoDevice::deleteLater);
    connect(this, &MainWindow::clearStatistics,
            hw, &DominoDevice::clearStatistics);
    connect(this, &MainWindow::writeAdcSetup,
            hw, &DominoDevice::on_setupChanged);
    connect(hw, &DominoDevice::setupWritten,
            this, &MainWindow::adcSetupWritten);
    connect(hw, &DominoDevice::destroyed,
            separateThread, &QThread::quit);
    connect(evNumChecker, &EvNumChecker::evNumRequest,
            hw, &DominoDevice::evNumRequest);
    connect(hw, &DominoDevice::evNumResponse,
            evNumChecker, &EvNumChecker::gotDevEvNum);
    const QPointer<AdcDeviceProxy> p = new AdcDeviceProxy(ds.devIndex, hw, this);
    devices.insert(ds.devIndex, p);
    hw->moveToThread(separateThread);
    qDebug() << QString("[%1] started device thread")
                .arg(ds.devIndex.getIdent());

    // print what will be default setup for adc64
    qDebug().noquote() << QString("[%1] ==== Default configuration on start-up ====\n    %2")
                .arg(ds.devIndex.getIdent())
                .arg(ds.toString().replace("\r", "\n    "));
    emit writeAdcSetup(ds);
}

void MainWindow::addDevices()
{
    Q_D(MainWindow);

    ui->actionWR_DAQ->setEnabled(!d->config.run && d->config.hasEnabledDevice());

    ui->table->blockSignals(true);

    for (const DeviceIndex &index : d->config.device_configs.keys())
    {
        DominoSettings &ds = d->config.device_configs[index];
        ds.setNCh(DominoDevice::getNumberOfChannels(index.getDevId()));

        // qDebug() << __FUNCTION__ << index << (ds.enabled ? "enabled" :  "disabled");
        if (!ds.enabled) continue;

        // Insert a row to table
        ui->table->insertRow(index);
        ui->statistic_widget->add_device(index);

        addDevice(index);
    }
    deviceSettingsDialog->deviceListUpdated(d->config.device_configs);
    sendResetAll();

    updateRemoteStatus();
//    evNumChecker->setupUpdated(&config);
    QVector<quint64> serials;
    for(const auto &ds : d->config.device_configs){
        if(ds.enabled)
            serials.append(ds.devIndex.getSerial());
    }
    evNumChecker->set_dev_serials(serials);
    updateRemoteStatus();
}

void MainWindow::removeDevices(const DeviceIndexSet &setToDel)
{
    qDebug() << "Removing" << setToDel;

    if(setToDel.isEmpty())
        return;

    for(const auto &i : setToDel) {
        Q_ASSERT(!i.isNull());
        emit deleteHardware(i);
    }

    QElapsedTimer timer;
    timer.start();

    for(const auto &i : setToDel) {
        auto adcIt = adcThreads.find(i);
        auto dumpIt = dumpersThread.find(i);
        if(adcIt == adcThreads.end() || dumpIt == dumpersThread.end()) {
            Q_ASSERT(!devices.contains(i));
            continue;
        }

        if(!threadQuiting(*adcIt, WHOLE_TIMEOUT_MS, &timer)){
            qWarning() << QString("Thread %1 failed to quit").arg((*adcIt)->objectName());
        }
        delete *adcIt;

        if(!threadQuiting(*dumpIt, WHOLE_TIMEOUT_MS, &timer)){
            qWarning() << QString("Thread %1 failed to quit").arg((*dumpIt)->objectName());
        }
        delete *dumpIt;

        devices.remove(i);

        adcThreads.remove(i);
        deviceSetupKeys.remove(i);

        dumpers.remove(i);
        dumpersThread.remove(i);
        lastAdcStatus.remove(i);

        ui->table->removeRow(i);
        ui->statistic_widget->remove_device(i);
    }
}

void MainWindow::removeAllDevices()
{
    Q_D(MainWindow);

    DeviceIndexSet toDel = makeDeviceIndexSet(d->config.device_configs.keys());
    removeDevices(toDel);
}

void MainWindow::setRunState(bool run)
{
    Q_D(MainWindow);

    d->config.setRun(run);
    if(run) {
        d->run_time_str = QDateTime::currentDateTime().toString("yyyyMMdd_hhmmss");
    }
    emit configChanged(d->config);
    emit set_run_state(run);
    ui->radioButtonStart->blockSignals(true);
    if(run)
        ui->radioButtonStart->setChecked(true);
    else
        ui->radioButtonStop->setChecked(true);
    ui->radioButtonStart->blockSignals(false);

    ui->actionWR_DAQ->setEnabled(!run && d->config.hasEnabledDevice());
    ui->actionFindDevices->setEnabled(!run);
    ui->actionAllow_Remote_Control->setEnabled(!run);
    ui->actionEnable_write->setEnabled(!run);
    ui->actionSet_stop_conditions->setEnabled(!run);
    ui->actionSet_MTU_size->setEnabled(!run);
    updateProgramDescription();
}

void MainWindow::updateRemoteStatus()
{
    Q_D(const MainWindow);

    QStringList readyStr, notReadyStr;
    for(DominoSettings &ds : d->config.device_configs.values()) {
        if(!ds.enabled)
            continue;
        const bool devOk = deviceSetupKeys.value(ds.devIndex)==setupKey
                && lastAdcStatus.contains(ds.devIndex)
                && lastAdcStatus[ds.devIndex].adcSelfTestOk;
        if(devOk) {
            readyStr << QString("0x%1 ").arg(ds.devIndex.getSerial(), 8, 16, QChar('0'));
        } else {
            notReadyStr << QString("0x%1 ").arg(ds.devIndex.getSerial(), 8, 16, QChar('0'));
        }
    }

    QString stateStr;
    if(!notReadyStr.isEmpty()) {
        auto str = notReadyStr.join(" ");
        if(str.size() > 40) {
            stateStr.append("Not ready:").append(QString("%1 device(s)").arg(notReadyStr.count()));
        } else {
            stateStr.append("Not ready:").append(str);
        }
    }
    if(!readyStr.isEmpty()) {
        if(!stateStr.isEmpty())
            stateStr.append("| ");
        auto str = readyStr.join(" ");
        if(str.size() > 40) {
            stateStr.append("Ready:").append(QString("%1 device(s)").arg(readyStr.count()));
        } else {
            stateStr.append("Ready:").append(str);
        }
    }
    if(stateStr.isEmpty())
        stateStr = "No devices";
    emit sendRemoteState(stateStr);
}

void MainWindow::remoteStatusUpdated(const RemoteControlServerStatus &status)
{
    Q_D(const MainWindow);

    if(!program_description.containsProgramInterface(ProgramInterfaceRemoteControl, 0)) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        interface.port = 0;
        interface.enabled = d->config.allow_remote_control;
        interface.isFree = false;
        program_description.interfaces.push_back(interface);
    }

    ProgramInterface &i = program_description.getProgramInterface(ProgramInterfaceRemoteControl);
    ProgramInterface prevIt = i;
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !status.portInUse;

    if(status.portInUse) {
        if(labelRunState->text().isEmpty())
            labelRunState->setText("waiting remote command");
        i.setPeer(status.peerHost, status.peerPort);
    } else {
        if(!labelRunState->text().isEmpty())
            labelRunState->setText("");
        i.peers.clear();
        const auto &curState = fsm->getState();
        if(curState != FsmState::Ready && curState != FsmState::Error) {
            qWarning() << "RemoteControlServer disconnected: fsm -> Error";
            fsm->goState(FsmState::Error);
        }
    }

    if(prevIt != i)
        updateRemoteStatus();
    ui->groupBoxRunControl->setEnabled(!status.portInUse);
    updateProgramDescription();
}

void MainWindow::sendRunError(QString mess)
{
    Q_D(const MainWindow);

    qWarning() << mess;
    if(d->config.run && d->config.allow_remote_control)
        emit runError(rcRunId, mess);
}

void MainWindow::sendAdcSetup(const DeviceIndex &index)
{
    Q_D(MainWindow);

    if(!d->config.device_configs.contains(index))
        return;
    DominoSettings &ds = d->config.device_configs[index];
    ds.mtuSize = d->config.mtu_size;
    ds.run = d->config.run;
    ds.key = setupKey;

    emit writeAdcSetup(ds);
}

void MainWindow::sendResetAll(bool doResetMStream, bool doStatClear)
{
    Q_D(MainWindow);

    incrementSetupKey();

    if(d->config.run && doResetMStream) {
        resetAllMStream(doStatClear); // TODO: do we should reseat MStream?
        //todo: cheange for handshake
        if(!devices.isEmpty())
            QThread::currentThread()->msleep(500);
    }

    for(DeviceIndex index : devices.keys()) {
        sendAdcSetup(index);
    }

    if(!d->config.run && doResetMStream) {
        resetAllMStream(doStatClear);
    }

    updateRemoteStatus();
}

void MainWindow::resetAllMStream(bool doStatClear) const
{
    for(const auto &devProxy : devices)
        resetMStream(devProxy->getIndex(), doStatClear);
}


void MainWindow::updateProgramDescription()
{
    Q_D(const MainWindow);

    program_description.type = configDb->getType().toString();
    program_description.index = configDb->getIndex();
    program_description.name = program_description.getName(); // FIXME

#ifdef GIT_DESCR
    program_description.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    program_description.ver_date = GIT_LAST_COMMIT;
#endif

    program_description.options.clear();
    const QString run_str = d->config.run ? "true" : "false";
    program_description.options["run"] = run_str;
    program_description.options["run_index"] = rcRunId.runIndex;
    program_description.options["run_number"] = QString::number(rcRunId.runNumber);
    QString write_enable_str = d->config.write_enable ? "true" : "false";
    program_description.options["write_enable"] = write_enable_str;

    int i=0;
    for(const DeviceIndex &index : d->config.device_configs.keys()){
        if(!d->config.device_configs[index].enabled)
            continue;
        program_description.options[QString("adc_%1_id").arg(i)] = QString("0x%1")
                .arg(index.getDevId(), 0, 16);
        program_description.options[QString("adc_%1_serial").arg(i)] = QString("0x%1")
                .arg(index.getSerial(), 0, 16);
        ++i;
    }

    emit programDescriptionUpdated(program_description);
}

bool MainWindow::readyForRun(bool printErr) const
{
    bool ready = true;
    for(DeviceIndex &index : devices.keys()){
        if(deviceSetupKeys.value(index, 0) != setupKey){
            if(printErr)
                qWarning() << QString("[%1] failed to write setup")
                              .arg(index.getIdent());
            ready = false;
        }
    }
    return ready;
}

bool MainWindow::waitDevSetupKey(const QElapsedTimer &timer, int timeout)
{
    Q_D(const MainWindow);

    bool ok = false;
    while(!ok && !timer.hasExpired(timeout)) {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 50);
        ok=true;
        for(const DominoSettings &ds : d->config.device_configs.values()) {
            if(!ds.enabled) continue;
            if(deviceSetupKeys.value(ds.devIndex, 0) != setupKey) {
                ok=false;
                break;
            }
        }
    }
    return ok;
}

bool MainWindow::waitDevEvNum(const QElapsedTimer &timer, int timeout)
{
    Q_D(const MainWindow);

    bool ok = false;
    while(!ok && !timer.hasExpired(timeout)) {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 50);
        ok=true;
        for(const DominoSettings &ds : d->config.device_configs.values()) {
            if(!ds.enabled) continue;

            const DeviceIndex &i = ds.devIndex;
            QString serial = i.getSerialStr();
            QMap<DeviceIndex, MStreamStat> stats = ui->statistic_widget->getStat();
            if(!lastAdcStatus.contains(i) ||
               !stats.contains(i) ||
               stats[i].lastEvNum != lastAdcStatus[i].evNum ||
               !ok) {
                ok=false;
                break;
            }
        }
    }
    return ok;
}

bool MainWindow::phasingStop()
{
    Q_D(MainWindow);

    QMap<DeviceIndex, quint16> trigsetup_cache;
    QMap<DeviceIndex, bool> skipRestartRun_cache;

    auto &device_configs = d->config.device_configs;
    for (auto dsIt=device_configs.begin(); dsIt!=device_configs.end(); ++dsIt) {
        if(!dsIt->enabled)
            continue;
        trigsetup_cache[dsIt.key()] = dsIt->trigsetup.all;
        skipRestartRun_cache[dsIt.key()] = dsIt->skipRestartRun;

        dsIt->trigsetup.all = 0;
        dsIt->skipRestartRun = true;
    }
    emit configChanged(d->config);

    sendResetAll(false, true);

    for (const DeviceIndex index : d->config.device_configs.keys()) {
        DominoSettings &ds = d->config.device_configs[index];
        if(!ds.enabled)
            continue;
        ds.trigsetup.all = trigsetup_cache[index];
        ds.skipRestartRun = skipRestartRun_cache[index];
    }
    emit configChanged(d->config);
    return true;
}


void MainWindow::addMStreamDump(const DeviceIndex &index)
{
    MStreamDumpParameters dump_parameters;
    auto adcStatus = lastAdcStatus.value(index);
    dump_parameters.deviceIndex = index;
    const auto dd = LldpDiscovery::instance().lookup(index);
    if(dd.ip_addr){
        dump_parameters.deviceAddress = QHostAddress(dd.ip_addr);
    }
    Q_ASSERT(dd.ip_addr);
    dump_parameters.fileName = getMStreamFileName(index);
    dump_parameters.bigFragId = adcStatus.mstreamVer2;
    dump_parameters.hwBufSize = adcStatus.hwBufSize;
    dump_parameters.ackSizeLimit = adcStatus.msMultiAck ? adcStatus.hwBufSize : 1;
    dump_parameters.preRunInit();

    MStreamDump *dump = dumpers[index] = new MStreamDump(dump_parameters, false);
    connect(dump, &MStreamDump::closing,
            pnpServer, &PNPServer::closeProgram);
    connect(dump, &MStreamDump::updateProgramDescription,
            pnpServer, &PNPServer::updateProgramDescription);
    connect(dump, &MStreamDump::stat_updated,
            ui->statistic_widget, &StatisticOutput::stat_updated);
    LldpDiscovery::instance().subscribe(dump);
    connect(dump, &MStreamDump::stat_from_receiver_updated,
            ui->statistic_widget, &StatisticOutput::stat_from_receiver_updated);
    connect(this, &MainWindow::msdump_reset,
            dump, &MStreamDump::msdump_reset);
    connect(this, &MainWindow::mstream_send_connect,
            dump, &MStreamDump::send_connect_to_dev);
    connect(this, &MainWindow::set_run_state,
            ui->statistic_widget, &StatisticOutput::set_run_state);

    QThread *thread = dumpersThread[index] = new QThread(this);
    thread->setObjectName("MSD_"+index.getSerialStr());
    connect(dump, &MStreamDump::destroyed,
            thread, &QThread::quit);
    connect(this, &MainWindow::deleteHardware,
            dump, &MStreamDump::delete_dump);
    connect(this, &MainWindow::deleteAll,
            dump, &MStreamDump::deleteLater);
    dump->moveToThread(thread);
    thread->start();
}

void MainWindow::resetMStream(const DeviceIndex &index, bool doStatClear) const
{
    Q_D(const MainWindow);

    if(!d->config.device_configs.contains(index))
        return;
    const DominoSettings &ds = d->config.device_configs[index];
    if(!ds.enabled)
        return;

    auto adcStatus = lastAdcStatus.value(index);

    QString fn = getMStreamFileName(index);
    emit msdump_reset(index, adcStatus.mstreamVer2, adcStatus.hwBufSize,
                     adcStatus.msMultiAck ? adcStatus.hwBufSize : 1, fn, doStatClear);
}

void MainWindow::flushMStreamStats()
{
    QJsonDocument mainDoc;
    QFile mainFile(QString("%1%2%3_list.txt")
                   .arg(QDir::tempPath())
                   .arg(QDir::separator())
                   .arg(QApplication::applicationName()) );
    QJsonArray mainArray;
    for(DeviceIndex &index : mstreamStats.keys()) {
        mainArray.push_back(QJsonValue(index.getSerialStr()));

        QFile devFile(QString("%1%2%3_%4.txt")
                       .arg(QDir::tempPath())
                       .arg(QDir::separator())
                       .arg(QApplication::applicationName())
                      .arg(index.getSerialStr()) );
        if(!devFile.open(QIODevice::WriteOnly)){
            qWarning()<<"Failed to open temp file:"<<devFile.fileName();
            continue;
        }
        QJsonDocument devDocument;
        QJsonObject devContent;
        for(QString &key : mstreamStats[index].keys()) {
            devContent.insert(key, mstreamStats[index][key]);
        }
        devDocument.setObject(devContent);
        devFile.write(devDocument.toJson(QJsonDocument::Indented));
        devFile.close();
    }
    mainDoc.setArray(mainArray);
    if(!mainFile.open(QIODevice::WriteOnly)){
        qWarning()<<"Failed to open temp file:"<<mainFile.fileName();
        return;
    }
    mainFile.write(mainDoc.toJson(QJsonDocument::Indented));
    mainFile.close();
}

void MainWindow::flushWarnings()
{
    auto stats = ui->statistic_widget->getStat();
    QStringList warnStr;
    for(auto serial : stats.keys()) {
        auto stat = stats[serial];
        QString devWarnStr;
        for(auto key : MS_WARN_MAP.keys()) {
            quint64 val = stat.*MS_WARN_MAP[key];
            if(val==0)
                continue;
            if(devWarnStr.isEmpty()) devWarnStr = QString("[%1] ").arg(serial.getSerialStr());
            warnStr.append(QString("%1=%2 ").arg(key).arg(val));
        }
    }

    emit runWarning(rcRunId, warnStr.join(' '));
    warnFlushTimer->start();
}

void MainWindow::block_gui_signals(bool block)
{
    for(int i=0; i<objects_to_block.size(); ++i){
        objects_to_block[i]->blockSignals(block);
    }
}

QString MainWindow::getMStreamFileName(DeviceIndex index) const
{
    Q_D(const MainWindow);

    QString fn;
    if(d->config.run && d->config.write_enable) {
        QString folderPrefix = d->config.storage_path;
        if(!folderPrefix.isEmpty())
            folderPrefix.append(QDir::separator());

        fn = folderPrefix.append(QString("%2_%3.data")
                                         .arg(index.getSerial(),8,16,QChar('0'))
                                         .arg(d->run_time_str));
    }
    return fn;
}

bool MainWindow::waitFsmState(const FsmState &state, const qint64 timeout)
{
    QElapsedTimer timer;
    timer.start();
    while (! timer.hasExpired(timeout)) {
        if (fsm->getState() == state)
            return true;
        if (fsm->getState() == FsmState::Error)
            return false;
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
    }
    return false;
}

void MainWindow::incrementSetupKey()
{
    deviceSetupKeys.clear();
    if(++setupKey==0) ++setupKey;
    fsm->setMainSetupKey(setupKey);
}
