/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2010-2020 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QLabel>
#include <QList>
#include <QString>
#include <QStringList>
#include <QWidgetAction>

#include "Adc64SystemConfig.h"
#include "Adc64SystemConfigDb.h"
#include "Adc64SystemSignalRouter.h"
#include "adc64-common/Adc64StartupOptions.h"

#include "base/app/BaseMainWindow.h"
#include "base/RcRunId.h"
#include "configuration-manager/ConfigurationManager.h"
#include "fsm/FsmState.h"
#include "m-stream-output/StatisticOutput.h"
#include "mstream-lib/MStreamDump.h"
#include "pnp-server/PNPMessage.h"

class QTableWidget;
class QTimer;
class QElapsedTimer;

class Adc64SystemSignalRouter;
class Adc64SystemConfigDispatcher;
class AdcDeviceProxy;
class AdcStatus;
class AdcDeviceMgrFsm;
class ConfigSelector;
class DeviceDispatcher;
class DeviceIndex;
class DeviceIndexEnabledMap;
class DeviceSettingsDialog;
class DeviceStateController;
class DiscoverDialog;
class DominoDevice;
class EvNumChecker;
class LldpInfo;
class MainWindowPrivate;
class PNPServer;
class RemoteControlServer;
class RemoteControlServerStatus;
class WrDaqDialog;
struct RcRunId;

namespace Ui {
class MainWindow;
}

class MainWindow : public BaseMainWindow
{
    Q_OBJECT
public:
    explicit MainWindow(const ConfigSelector &cs, QWidget *parent = nullptr);
    ~MainWindow();

public slots:
    void setConfig(const Adc64SystemConfig &config);
    void setDeviceList(const DeviceIndexEnabledMap &newList);
    void configSelectorChanged(const ConfigIndexName &conf);
    void sendRunError(QString mess=QString());
    void onRemoteStartRequest(const RcRunId &runId, quint64 start_event_number=0);
    void onRemoteStopRequest(const RcRunId &runId, quint64 stop_event_number=0);
    void remoteStatusUpdated(const RemoteControlServerStatus &status);
    void configDbSyncResult(const ConfigDbSyncResult &result);

protected slots:
    void addLog(QtMsgType msgType, const QString &log, const QString &threadName) override;

private slots:
    void updateGui();
    void onFsmStateChange(const FsmState &state, const FsmState &oldState);

    // Gui actions
    void on_actionWR_DAQ_triggered();
    void on_actionFindDevices_triggered();
    void on_actionClear_statistics_triggered();
    void on_actionAllow_Remote_Control_toggled(bool checked);
    void on_actionEnable_write_toggled(bool enabled);
    void on_actionShow_details_triggered(bool show);
    void on_actionSet_stop_conditions_triggered();
    void on_actionStop_run_on_TrigOnXOff_toggled(bool trig_on_xoff_stop);
    void on_actionShow_device_settings_triggered();
    void on_actionSet_MTU_size_triggered();

    void on_radioButtonStart_toggled(bool checked);

    // Other actions
    void adcConnected(const DeviceIndex &index);
    void adcDisconnected(const DeviceIndex &);
    void adcStateChanged(const DeviceIndex &index, bool online);
    void updateAdcStatus(const DeviceIndex &index, const AdcStatus &status);

    void adcSetupWritten(const DominoSettings &s);
    void updateRemoteStatus();

    void flushMStreamStats();
    void flushWarnings();

    void on_actionConfiguration_manager_triggered();
    void on_actionSynchronize_configs_with_adc64_triggered();

signals:
    void switch_to_config(const ConfigIndexName &conf) const;
    void deleteHardware(const DeviceIndex &index);
    void deleteAll();
    void clearStatistics(const DeviceIndex &index);
    void writeAdcSetup(const DominoSettings &setup);
    void sendRemoteState(const QString &state);
    void runError(const RcRunId &runId, QString errStr);
    void runWarning(const RcRunId &runId, QString warnStr);
    void allowRemoteControl(bool);
    void programDescriptionUpdated(const ProgramDescription &);
    void closeProgram(const ProgramDescription &);
    void config_transfer_to_rc(QByteArray data);
    void msdump_reset(DeviceIndex index,
                      bool bigFragId,
                      quint16 hwBufSize,
                      quint16 ackSize,
                      QString fn=QString(),
                      bool doStatClear = false) const;
    void mstream_send_connect(DeviceIndex index);
    void set_run_state(bool run);
    void configChanged(const Adc64SystemConfig &);
    void deviceListChanged(const DeviceIndexEnabledMap &);

private:
    void addDevice(const DeviceIndex &index);
    void addDevices();
    void removeDevices(const DeviceIndexSet &setToDel);
    void removeAllDevices();
    void setRunState(bool run);

    void addMStreamDump(const DeviceIndex &index);
    void resetMStream(const DeviceIndex &index, bool doStatClear = false) const;
    void sendAdcSetup(const DeviceIndex &index);
    void sendResetAll(bool doResetMStream = true, bool doStatClear = false);
    void resetAllMStream(bool doStatClear = false) const;
    QStringList hostStrList() const;
    void updateProgramDescription();
    bool readyForRun(bool printErr=false) const;
    bool waitDevSetupKey(const QElapsedTimer &timer, int timeout=1000);
    bool waitDevEvNum(const QElapsedTimer &timer, int timeout=1000);
    bool phasingStop();
    void block_gui_signals(bool block);
    QString getMStreamFileName(DeviceIndex index) const;
    bool waitFsmState(const FsmState &state, const qint64 timeout);
    void incrementSetupKey();

    Q_DECLARE_PRIVATE(MainWindow)
    MainWindowPrivate * const d_ptr;
    Ui::MainWindow * const ui;
    QLabel * const labelFsmState;
    QLabel * const labelRunState;
    WrDaqDialog * const wrDaqDialog;
    Adc64SystemSignalRouter * const signalRouter;
    Adc64SystemConfigDb * const configDb;
    Adc64SystemConfigDispatcher * const configDispatcher;
    DeviceStateController * const deviceStateController;
    QMap<DeviceIndex, bool> synch_result;
    ConfigurationManager * const configuration_manager;
    DeviceDispatcher * const deviceDispatcher;
    QList<QObject*> objects_to_block;
    QMap<DeviceIndex, QPointer<AdcDeviceProxy> > devices; // access to objects in non-gui thread
    QMap<DeviceIndex, int> deviceSetupKeys;
    QMap<DeviceIndex, QThread *> adcThreads;
    QMap<DeviceIndex, int> rowIndexMap; // [cfgIndex] = deviceIndex
    QMap<DeviceIndex, MStreamDump *> dumpers;
    QMap<DeviceIndex, QThread *> dumpersThread;
    QMap<DeviceIndex, AdcStatus> lastAdcStatus;
    RemoteControlServer * const remoteControl;
    PNPServer * const pnpServer;
    ProgramDescription program_description;
    DiscoverDialog * const discoverDialog;
    int setupKey;
    RcRunId rcRunId;
    QMap<DeviceIndex, QMap<QString, QString> > mstreamStats;
    QTimer * const statFlushTimer;
    QTimer * const warnFlushTimer;
    DeviceSettingsDialog * const deviceSettingsDialog;
    EvNumChecker * const evNumChecker;
    AdcDeviceMgrFsm * const fsm;
};


#endif // MAINWINDOW_H
