#include "CopyConfigDialog.h"

#include <QCheckBox>
#include <QPushButton>
#include <QRadioButton>

#include "ui_CopyConfigDialog.h"
#include "Adc64Config.h"
#include "base/gui_utils.h"

CopyConfigDialog::CopyConfigDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::CopyConfigDialog),
    config(new Adc64Config())
{
    ui->setupUi(this);
}

CopyConfigDialog::~CopyConfigDialog()
{
    delete config;
    delete ui;
}

void CopyConfigDialog::restore_gui_values()
{
    QVBoxLayout *formFrom = ui->verticalLayoutCopyFrom;
    QVBoxLayout *formTo = ui->verticalLayoutTo;
    const auto &all_keys = config->device_configs.keys();

    for(auto it=devButtons.begin(); it!=devButtons.end(); ) {
        const auto index = it.key();
        if(!all_keys.contains(index)) {
            it.value().first->deleteLater();
            it.value().second->deleteLater();
            ++it;
            devButtons.remove(index);
        } else {
            ++it;
        }
    }

    bool fromKeyIsEmpty = from_key.isNull() || !all_keys.contains(from_key);

    for(int i=all_keys.size()-1; i>=0; --i) {
        const DeviceIndex &index = all_keys.at(i);
        QRadioButton *rb = nullptr;
        QCheckBox *cb = nullptr;
        auto devButIt = devButtons.find(index);
        if(devButIt == devButtons.end()) {
            rb = new QRadioButton(index.getSerialStr(), this);
            connect(rb, &QRadioButton::toggled,
                    this, &CopyConfigDialog::radioButtonToggled, Qt::UniqueConnection);
            formFrom->addWidget(rb);
            if(fromKeyIsEmpty) {
                from_key = index;
                fromKeyIsEmpty = false;
            }
            if(from_key == index) {
                WidgetSignalBlocker b(rb);
                rb->setChecked(true);
            }

            cb = new QCheckBox(index.getSerialStr(), this);
            connect(cb, &QCheckBox::toggled,
                    this, &CopyConfigDialog::checkBoxToggled, Qt::UniqueConnection);
            formTo->addWidget(cb);
            {
                WidgetSignalBlocker b(cb);
                cb->setChecked(true);
            }

            devButtons[index] = {rb, cb};
        } else {
            rb = qobject_cast<QRadioButton *>(devButIt.value().first);
            cb = qobject_cast<QCheckBox *>(devButIt.value().second);
        }

        QString colorStr = config->device_configs[index].enabled ? "darkGreen" : "#505050";
        rb->setStyleSheet(QString("QRadioButton {background-color:transparent; color:%1;}").arg(colorStr));
        cb->setStyleSheet(QString("QCheckBox { background-color:transparent; color:%1; }").arg(colorStr));
    }
    updateCopyButton();
    guiValid = true;
}

void CopyConfigDialog::setConfig(const Adc64Config &r)
{
    bool doRestore = !guiValid;
    if(!doRestore)
        doRestore &= config->device_configs.keys() != r.device_configs.keys();
    if(!doRestore) {
        for(auto ds : r.device_configs) {
            doRestore &= config->device_configs[ds.devIndex].enabled != ds.enabled;
            if(doRestore)
                break;
        }
    }

    *config = r;

    if (doRestore) {
        if(isVisible())
            restore_gui_values();
        else
            guiValid = false;
    }
}

void CopyConfigDialog::setSelectedDevice(const DeviceIndex &index)
{
    if(index == from_key)
        return;

    if(!from_key.isNull()) {
        auto it = devButtons.find(from_key);
        if(it!=devButtons.end()) {
            it.value().first->setChecked(false);
        }
    }
    if(!index.isNull()) {
        auto it = devButtons.find(index);
        if(it!=devButtons.end()) {
            it.value().first->setChecked(true);
        }
    }
    from_key = index;
}

void CopyConfigDialog::setVisible(bool visible)
{
    if(visible && !guiValid)
        restore_gui_values();

    QDialog::setVisible(visible);
}

void CopyConfigDialog::on_pushButtonCopy_clicked()
{
    auto fromIt = config->device_configs.constFind(from_key);
    if(fromIt == config->device_configs.cend()) {
        addLog("Source config is missing\n", true);
        return;
    }

    bool targetIsEmpty = true;
    for(auto devIt=devButtons.cbegin(); devIt!=devButtons.cend(); ++devIt) {
        const auto index = devIt.key();
        if(index == from_key)
            continue;
        targetIsEmpty = false;

        auto toIt=config->device_configs.find(index);
        if(toIt == config->device_configs.end()) {
            addLog(QString("Target device (%1) is missing in config")
                   .arg(index.getSerialStr()),
                   true);
        } else {
            toIt.value().copyFrom(fromIt.value());
            addLog(QString("Copying from %1 to %2 successfully")
                   .arg(from_key.getSerialStr()).arg(index.getSerialStr()));
        }
    }

    if(targetIsEmpty)
        addLog("Copy target is empty", true);
    else
        emit configChanged(*config);

    addLog(" ");
}

void CopyConfigDialog::radioButtonToggled(bool checked)
{
    if(!checked) return;

    for(auto it=devButtons.cbegin(); it!=devButtons.cend(); ++it) {
        if(it->first == sender()) {
            from_key = it.key();
            return;
        }
    }

    from_key = DeviceIndex();
}

void CopyConfigDialog::checkBoxToggled(bool)
{
    updateCopyButton();
}

void CopyConfigDialog::addLog(QString msg, bool isErr)
{
    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    tf.setForeground(QBrush(isErr ? Qt::red : Qt::darkGreen));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    ui->plainTextEdit->appendPlainText(msg);
}

void CopyConfigDialog::updateCopyButton() const
{
    bool copyEn = false;
    for(auto p : devButtons) {
        if(p.second->isChecked()) {
            copyEn = true;
            break;
        }
    }
    ui->pushButtonCopy->setEnabled(copyEn);
}
