/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceTableWidget.h"
#include "ui_DeviceTableWidget.h"

#include <QDateTime>

#include "adc64-common/AdcStatus.h"
#include "base/DeviceIndexEnabled.h"

static const double TEMP_WARN_LIMIT = 60;

DeviceTableWidget::DeviceTableWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::DeviceTableWidget)
{
    ui->setupUi(this);
}

DeviceTableWidget::~DeviceTableWidget()
{
    delete ui;
}

void DeviceTableWidget::initTable() const
{
    QTableWidget * const table = ui->tableWidgetDevices;
    table->horizontalHeader()->hide();
    table->verticalHeader()->hide();
    table->setColumnCount(COL_TOTAL);
    table->hideColumn(COL_ID);
    table->hideColumn(COL_SERIAL);
    table->hideColumn(COL_WR_TIME);
    table->hideColumn(COL_CH);
    table->hideColumn(COL_ADC_TEST);
    table->hideColumn(COL_ADC_DELAY_CHECK);
}

void DeviceTableWidget::setDeviceList(const DeviceIndexEnabledMap &devices_)
{
    const DeviceIndexSet &newDeviceSet = devices_.getEnabled();

    if(curDeviceSet == newDeviceSet)
        return;

    const auto &toDel = curDeviceSet-newDeviceSet;
    removeDevices(toDel);

    const auto &toAdd = newDeviceSet-curDeviceSet;

    ui->tableWidgetDevices->hideColumn(COL_WR_TIME);
    ui->tableWidgetDevices->setSelectionBehavior(QAbstractItemView::SelectRows);

    bool first_key = ui->tableWidgetDevices->rowCount()==0;
    for(const auto &index : toAdd) {
        addDeviceRow(index);
        if(first_key) {
            ui->tableWidgetDevices->selectRow(0);
            first_key = false;
        }
    }
    updateTableSize();
}

void DeviceTableWidget::addDeviceRow(const DeviceIndex &index)
{
    if(getDeviceRow(index) != -1){
        return;
    }
    QTableWidget * const table = ui->tableWidgetDevices;
    table->setSortingEnabled(false);
    int row = table->rowCount();
    table->insertRow(row);
    table->setItem(row, COL_ID, new QTableWidgetItem(QString::number(index.getDevId())));
    table->setItem(row, COL_SERIAL, new QTableWidgetItem(QString::number(index.getSerial())));
    table->setItem(row, COL_ID_STR, new QTableWidgetItem(getDeviceTypeName(index.getDevId())));
    table->setItem(row, COL_SERIAL_STR, new QTableWidgetItem(index.getSerialStr()));
    table->setItem(row, COL_FW, new QTableWidgetItem());
    table->setItem(row, COL_WR_TIME, new QTableWidgetItem());
    table->setItem(row, COL_CH, new QTableWidgetItem());
    table->setItem(row, COL_TEMP, new QTableWidgetItem());
    table->setItem(row, COL_STATUS, new QTableWidgetItem("Creating..."));
    table->setItem(row, COL_ADC_LOCK, new QTableWidgetItem());
    table->setItem(row, COL_ADC_TEST, new QTableWidgetItem());
    table->setItem(row, COL_ADC_DELAY_CHECK, new QTableWidgetItem());
    table->setSortingEnabled(true);
    if(table->rowCount() > 1)
        table->showColumn(COL_CH);
    updateTableSize();
    onDeviceDisconnect(index);
    curDeviceSet << index;
    updateWrTimeCol();
}

void DeviceTableWidget::removeDevices(const DeviceIndexSet &setToDel)
{
    QTableWidget * const table = ui->tableWidgetDevices;
    for(const auto &i : setToDel) {
        const int &row = getDeviceRow(i);
        if(row!=-1)
            table->removeRow(row);
        curDeviceSet.remove(i);
    }
    updateWrTimeCol();
    table->setColumnHidden(COL_CH, table->rowCount()<2);
}

void DeviceTableWidget::updateWrTimeCol() const
{
    bool colHiden = true;
    for(auto &i : curDeviceSet)
        if(i.getDevId() == DEVICE_ID_ADC64WR) {
            colHiden = false;
            break;
        }
    ui->tableWidgetDevices->setColumnHidden(COL_WR_TIME, colHiden);
}

void DeviceTableWidget::updateTableSize() const
{
    QTableWidget * const table = ui->tableWidgetDevices;
    table->resizeColumnsToContents();
    table->resizeRowsToContents();
    int height = table->verticalHeader()->length()+4;
    if(height<20)
        height = 20;
    table->setFixedSize(table->horizontalHeader()->length()+4, height);
    table->updateGeometry();
}

void DeviceTableWidget::on_tableWidgetDevices_itemSelectionChanged()
{
    QTableWidget *table = ui->tableWidgetDevices;
    QList<QTableWidgetSelectionRange> ranges = table->selectedRanges();
    DeviceIndex newDevice;
    if(!ranges.empty()){
        const int row = table->currentRow();
        if(ranges.first().leftColumn() != COL_ID_STR || ranges.first().rightColumn() != COL_SERIAL_STR ||
                ranges.first().bottomRow() != ranges.first().topRow()){
            table->blockSignals(true);
            table->clearSelection();
            table->setRangeSelected(QTableWidgetSelectionRange(row, COL_ID_STR, row, COL_SERIAL_STR), true);
            table->blockSignals(false);
        }
        newDevice = getDeviceIndex(row);
    }
    if(selectedDevice != newDevice){
        selectedDevice = newDevice;
        emit selectedDeviceChanged(selectedDevice);
    }
}

void DeviceTableWidget::deviceStatusUpdated(const DeviceIndex &index, const QString &s) const
{
    int row = getDeviceRow(index);
    if(row == -1)
        return;
    QTableWidgetItem *item = ui->tableWidgetDevices->item(row, COL_STATUS);
    item->setText(s);
    updateTableSize();
}

void DeviceTableWidget::adcLockStatusUpdated(const DeviceIndex &index, bool ok, quint16 locked) const
{
    int row = getDeviceRow(index);
    if(row == -1)
        return;
    QTableWidgetItem *item = ui->tableWidgetDevices->item(row, COL_ADC_LOCK);
    if(ok){
        item->setText("Adc lock Ok");
        item->setTextColor(QColor());
    } else {
        item->setText(QString("Adc lock %1").arg(locked, 8, 2, QChar('0')));
        item->setTextColor(Qt::red);
    }
    updateTableSize();
}

void DeviceTableWidget::adcRampStatusUpdated(const DeviceIndex &index, bool ok, quint8 rampError) const
{
    ui->tableWidgetDevices->showColumn(COL_ADC_TEST);
    int row = getDeviceRow(index);
    if(row == -1)
        return;
    QTableWidgetItem *item = ui->tableWidgetDevices->item(row, COL_ADC_TEST);
    if(ok){
        item->setText("Adc test Ok");
        item->setTextColor(QColor());
    } else {
        item->setText(QString("Adc test %1").arg(rampError, 8, 2, QChar('0')));
        item->setTextColor(Qt::red);
    }
    updateTableSize();
}

void DeviceTableWidget::adcDelayCheckUpdated(const DeviceIndex &index,
                                             QSet<int> failedAdcIdNull,
                                             QSet<int> failedAdcId) const
{
    int row = getDeviceRow(index);
    if(row == -1)
        return;
    QTableWidgetItem *item = ui->tableWidgetDevices->item(row, COL_ADC_DELAY_CHECK);
    bool hasErr = true;
    QColor itemCol;
    QString itemText;

    if(!failedAdcIdNull.isEmpty()){
        itemCol = Qt::red;
        QStringList failedAdcStrList;
        for(auto id : failedAdcIdNull) failedAdcStrList << QString::number(id);
        failedAdcStrList.sort();
        itemText = QString("DelayChk Failed adcId:{%1}")
                .arg(failedAdcStrList.join(", "));
    } else if(!failedAdcId.isEmpty()){
        itemCol = QColor(255,140,0);
        QStringList failedAdcStrList;
        for(auto id : failedAdcId) failedAdcStrList << QString::number(id);
        failedAdcStrList.sort();
        itemText = QString("DelayChk warn adcId:{%1}")
                .arg(failedAdcStrList.join(", "));
    } else {
        itemText = "DelayChk Ok";
        hasErr = false;
    }

    item->setText(itemText);
    item->setData(Qt::UserRole, hasErr);
    item->setForeground(QBrush(itemCol));


    bool colVisible = false;
    for(int row=0; row< ui->tableWidgetDevices->rowCount(); ++row){
        if(ui->tableWidgetDevices->item(row, COL_ADC_DELAY_CHECK)->data(Qt::UserRole).toBool()) {
            colVisible = true;
            break;
        }
    }
    ui->tableWidgetDevices->setColumnHidden(COL_ADC_DELAY_CHECK, !colVisible);

    updateTableSize();
}

int DeviceTableWidget::getDeviceRow(const DeviceIndex &index) const
{
    for(int row=0; row< ui->tableWidgetDevices->rowCount(); ++row){
        if(getDeviceIndex(row) == index){
            return row;
        }
    }
    return -1;
}

DeviceIndex DeviceTableWidget::getDeviceIndex(int row) const
{
    auto id = ui->tableWidgetDevices->item(row, COL_ID)->text().toUShort();
    auto serial = ui->tableWidgetDevices->item(row, COL_SERIAL)->text().toULongLong();
    return DeviceIndex(static_cast<DeviceId>(id), serial);
}

void DeviceTableWidget::updateAdcStatus(const DeviceIndex &index, const AdcStatus &status)
{
    int row = getDeviceRow(index);
    if(row == -1)
        return;

    // read temperature sensors
    {
        QTableWidgetItem *item = ui->tableWidgetDevices->item(row, COL_TEMP);
        QStringList sl;
        double maxTemp = -255;
        if (status.temp > maxTemp) maxTemp = status.temp;
        if (status.temp > -255)
            sl += QString("ADC:%1").arg((int)status.temp);
        QMap<int, QString> name;
        name[0] = "PS";
        name[1] = "AMP";
        for (int i=0; i<status.bmStatus.sensors.size(); i++) {
            double t = status.bmStatus.sensors[i].temp;
            if (t > maxTemp) maxTemp = t;
            if (t > -255)
                sl += QString("%1:%2").arg(name[i]).arg((int)t);
        }
        item->setText(sl.join(" "));
        item->setBackgroundColor(maxTemp > TEMP_WARN_LIMIT ? Qt::red : Qt::transparent);
    }

    bool wrTimeValid = false;
    QTableWidgetItem *wrItem = ui->tableWidgetDevices->item(row, COL_WR_TIME);
    const QColor bgColorDefault = ui->tableWidgetDevices->style()->standardPalette().light().color();
    const QColor bgColorWarning = Qt::yellow;
    const QColor bgColorError = Qt::red;
    QStringList strListToolTip;
    strListToolTip << "WhiteRabbit sync";
    strListToolTip << QString("wrLinkError:%1").arg(status.wrEpStatus.wrLinkError);
    strListToolTip << QString("wrCodeError:%1").arg(status.wrEpStatus.wrSyncFail);
    QColor itemColor = bgColorDefault;
    if (status.wrEpStatus.hasWrEpStatus) {
        wrTimeValid = status.wrEpStatus.timeValid();
        const bool wrTimeNull = status.wrEpStatus.timeNull();
        if (wrTimeNull) {
            wrItem->setText(QString());
            itemColor = bgColorWarning;
        } else {
            QDateTime wrTime = QDateTime::fromMSecsSinceEpoch(status.wrEpStatus.mSecsSinceEpoch());

            int wrTimeCheck = status.wrEpStatus.wrTimeCheck();
            qint64 wrTimeOffsetMs = status.wrEpStatus.mSecsSinceEpoch() - status.wrEpStatus.acquiredAtMs;
            if(!wrTimeValid || wrTimeCheck > 1)
                itemColor = bgColorError;
            else if (wrTimeCheck == 1)
                itemColor = bgColorWarning;

            strListToolTip << QString("wrTime:%1").arg(wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss"));
//            wrItem->setText(wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss")); // add ".zzz" for ms
            QString strOffset;
            if (wrTimeOffsetMs/1000. > -60 && wrTimeOffsetMs/1000. < 60)
                strOffset = QString(" %1").arg(wrTimeOffsetMs * 1e-3, 6, 'f', 3);
            strListToolTip << QString("Offset:%1").arg(strOffset.trimmed());
        }
        strListToolTip << QString("wrCodeError:%1").arg(status.wrEpStatus.wrCodeError);
        strListToolTip << QString("wrUptime:%1").arg(status.wrEpStatus.wrUptime);
    } else {
            wrTimeValid = status.adcStatus & 0x2;
            itemColor = wrTimeValid ? bgColorDefault : bgColorError;
    }
    if(index.getDevId() == DEVICE_ID_ADC64WR) {
        wrItem->setText(wrTimeValid?"WR Ok":"WR Invalid");
        wrItem->setBackgroundColor(itemColor);
        wrItem->setToolTip(strListToolTip.join('\n'));
    } else {
        wrItem->setText(" - ");
        wrItem->setToolTip("Device doesn't support WhiteRabbit");
    }
    QTableWidgetItem *fwItem = ui->tableWidgetDevices->item(row, COL_FW);
    fwItem->setText(status.fwStr);
    updateTableSize();
}

void DeviceTableWidget::onDeviceConnect(const DeviceIndex &i)
{
    const int row = getDeviceRow(i);
    if(row != -1) {
        // ui->tableWidgetDevices->item(row, COL_FW)->setText(fw);
        for(int col = 0; col < ui->tableWidgetDevices->columnCount(); ++col)
            ui->tableWidgetDevices->item(row, col)->setBackgroundColor(Qt::transparent);
    }
}

void DeviceTableWidget::onDeviceDisconnect(const DeviceIndex &i)
{
    const int row = getDeviceRow(i);
    if(row != -1){
        for(int col = 0; col < ui->tableWidgetDevices->columnCount(); ++col)
            ui->tableWidgetDevices->item(row, col)->setBackgroundColor(Qt::lightGray);
    }
}

void DeviceTableWidget::updateChannelOffset(const QMap<DeviceIndex, QPair<int, int> > &map)
{
    if(devChRange == map)
        return;
    devChRange = map;
    auto table = ui->tableWidgetDevices;
    for(int r=0; r<table->rowCount(); ++r) {
        auto i = getDeviceIndex(r);
        auto it = devChRange.find(i);
        const auto &text = it == devChRange.cend()
                ? "-" : QString("%1-%2").arg(it->first).arg(it->second);
        table->item(r, COL_CH)->setText(text);
    }
}
