//
//    Copyright 2011 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "DominoDeviceManager.h"

#include <QDebug>
#include <QFile>
#include <QFileInfo>
#include <QList>
#include <QMap>
#include <QMetaType>
#include <QPointer>
#include <QStringList>
#include <QThread>
#include <QTimer>

#include "Adc64Config.h"
#include "Adc64SignalRouter.h"
#include "DeviceEntry.h"
#include "DeviceProperties.h"
#include "adc64-common/AdcDeviceProxy.h"
#include "adc64-common/DominoDeviceFactory.h"
#include "EventData.h"
#include "adc64-common/dominodevice.h"
#include "mldiscover/DeviceDescription.h"
#include "base/threadQuit.h"
#include "base/DeviceIndexEnabled.h"
#include "util/QDateTimeCompat.h"
#include "config/Adc64Config.h"

namespace {
const int monitorEventTimeMs = 300;
const int eventTimeoutMs = 500;
}

DominoDeviceManager::DominoDeviceManager(QObject *parent) :
    QObject(parent),
    eventData(new EventData),
    monitorEvent(new EventData),
//    properties(new DeviceProperties())
    eventTimeoutTimer(new QTimer(this))
{
    qRegisterMetaType<EventData>();
    qRegisterMetaType<AdcClkStatus>();
    qRegisterMetaType<QSet<int> >();
    supportedDevices
            << DEVICE_ID_TQDC16VS_DIG
            << DEVICE_ID_ADC64
            << DEVICE_ID_ADC64S2
            << DEVICE_ID_ADC64WR
            << DEVICE_ID_ADC64VE
            << DEVICE_ID_ADC64VE2
            << DEVICE_ID_ADC64ECAL
            << DEVICE_ID_ADC64VE_XGE
            << DEVICE_ID_ADC8E
//            << DEVICE_ID_AnyDevice
            << DEVICE_ID_ADCM163;
    eventTimeoutTimer->setSingleShot(true);
    eventTimeoutTimer->setInterval(eventTimeoutMs);
    connect(eventTimeoutTimer, &QTimer::timeout, this, &DominoDeviceManager::eventTimeout);
    qRegisterMetaType<AdcStatus>();
}

DominoDeviceManager::~DominoDeviceManager()
{
    removeDevices(makeDeviceIndexSet(devThreads.keys()));
    qDebug()<<"DominoDeviceManager destroyed";
}

void DominoDeviceManager::setHostAddress(const DeviceIndex &index, const QHostAddress &address)
{
    if (! devices.contains(index))
        return;
    devices[index]->setHostAddress(index, address);
}

void DominoDeviceManager::setDevices(const DeviceIndexEnabledMap &deviceList)
{
    const DeviceIndexSet &newDeviceSet = deviceList.getEnabled();
    const DeviceIndexSet &curSet = makeDeviceIndexSet(devices.keys());
    if(curSet == newDeviceSet)
        return;

    const auto &toDel = curSet-newDeviceSet;

    const auto &toAdd = newDeviceSet-curSet;
    for(const auto &index : toAdd) {
        DominoDevice * const dev = DominoDeviceFactory::create(index);
        QThread * thread;

        if(!devThreads.contains(index))
            devThreads[index] = new QThread(this);
        thread = devThreads[index];

        const QPointer<AdcDeviceProxy> p = new AdcDeviceProxy(index, dev, this);
//        connect(dev, &DominoDevice::deviceConnected, p, &AdcDeviceProxy::deviceConnected);
//        connect(dev, &DominoDevice::deviceDisconnected, p, &AdcDeviceProxy::deviceDisconnected);

        devices.insert(index, p);

        connect(this, &DominoDeviceManager::setupChanged, dev, &DominoDevice::on_setupChanged);
        connect(this, &DominoDeviceManager::clearEventNumber, dev, &DominoDevice::clearEventNumber);
        connect(dev, &DominoDevice::deviceConnected, this, &DominoDeviceManager::deviceConnected);
        connect(dev, &DominoDevice::deviceDisconnected, this, &DominoDeviceManager::deviceDisconnected);
        connect(dev, &DominoDevice::TempChanged, this, &DominoDeviceManager::deviceTempChanged);
        connect(dev, &DominoDevice::statusUpdated, this, &DominoDeviceManager::statusUpdated);
        connect(dev, &DominoDevice::deviceInfoUpdated, this, &DominoDeviceManager::deviceInfoUpdated);
        connect(dev, &DominoDevice::dataAcquired,
                this, &DominoDeviceManager::onDeviceDataAcquired);
        connect(this, &DominoDeviceManager::deleteDevice, dev, &DominoDevice::deleteHardware);
        connect(dev, &DominoDevice::destroyed, thread, &QThread::quit);
        connect(dev, &QtMregDevice::setStatus, this, &DominoDeviceManager::deviceStatusUpdated);
        connect(dev, &DominoDevice::adcLockStatusUpdated,
                this, &DominoDeviceManager::adcLockStatusUpdated);
        connect(dev, &DominoDevice::adcRampStatusUpdated,
                this, &DominoDeviceManager::adcRampStatusUpdated);
        connect(dev, &DominoDevice::adcClkDelayCheckUpdated,
                this, &DominoDeviceManager::adcClkDelayCheckUpdated);
        connect(this, &DominoDeviceManager::startAdcClockDelay, dev, &DominoDevice::FindAdcClkDelay);
        connect(this, &DominoDeviceManager::hardPatternTest, dev, &DominoDevice::setHardPatternTest);
        connect(dev, &DominoDevice::adcClkDelayUpdated, this, &DominoDeviceManager::adcClkDelayUpdated);

        Adc64SignalRouter::instance().connect_DominoDevice(dev);
        dev->moveToThread(thread);
        thread->setObjectName(index.getSerialStr());
        thread->start();
    }

    removeDevices(toDel);
    //    qDebug() << "DominoDeviceManager::setHostAddressList:" << devices.size() << "devices created";
}

//void DominoDeviceManager::clearEventNumber()
//{
//    Q_FOREACH(DeviceEntry ent, devices)
//        QMetaObject::invokeMethod(ent.dev.data(), "clearEventNumber");
//}

bool DominoDeviceManager::isAllConnected() const
{
    if (devices.isEmpty())
        return false;
    for (auto it = devices.begin(); it != devices.end(); ++it)
        if (! it.value()->isConnected())
            return false;
    return true;
}

//QtMregDevice::DeviceId DominoDeviceManager::getDeviceId() const
//{
//}

QString DominoDeviceManager::getDeviceIdStr() const
{
    QStringList sl;
    for (auto it = devices.begin(); it != devices.end(); ++it)
        sl.push_back(it.value()->getIndex().getIdent());
    return sl.join("\n");
}

void DominoDeviceManager::clearFullEventNumber()
{
    eventData->clearFullEventNumber();
    emit clearEventNumber();
}

void DominoDeviceManager::eventTimeout()
{
    if(eventData->timestamp.elapsed() > eventTimeoutMs)
        qWarning() << QString("EventData timed out, number of event blocks mismatch: %3 blocks, %4 devices")
                      .arg(eventData->size()).arg(devices.size());
}

void DominoDeviceManager::findAdcClockDelay(bool hardMode, const QString &dirPath)
{
    delayStatusMap.clear();
    findAdcClockDir = dirPath;
    emit startAdcClockDelay(hardMode);
}

void DominoDeviceManager::adcClkDelayUpdated(const DeviceIndex &index, const AdcClkStatus &s)
{
    if(!findAdcClockDir.isNull()) {
        QFile f;
        int loopCnt = 0;
        do {
            const auto &filePath = QString("%1/adcClkDelay_%2_t%3%4.txt")
                    .arg(findAdcClockDir)
                    .arg(index.getIdent().replace(' ', '_'))
                    .arg(s.temp, 2, 'f', 0, QChar('0'))
                    .arg(loopCnt ? QString("_%1").arg(loopCnt) : "");

            f.setFileName(filePath);

            if(++loopCnt > 100) {
                qWarning() << "Failed to save adcClkDelay; last file attempt:"
                           << filePath;
                break;
            }
        }while(!f.open(QIODevice::WriteOnly|QIODevice::NewOnly));

        if(f.isOpen()) {
            QTextStream stream(&f);
            stream << QString("=== %1 (f/w = %2; temp = %3) ===")
                      .arg(index.getIdent())
                      .arg(s.fwStr)
                      .arg(s.temp, 0, 'f', 1);
            stream << "\n";
            stream << s.toString();
            f.close();
        }
    }

    delayStatusMap[index] = s;
    qInfo() << "Got status from:" << index.getIdent()
            << " " << delayStatusMap.size() << " of " << devices.size();
    // check if all devices present in map
    if (devices.size() != delayStatusMap.size())
        return;

    qInfo().noquote() << AdcClkStatus::getLegend();
    for (DeviceIndex i : delayStatusMap.keys()) {
        const AdcClkStatus &devStatus = delayStatusMap[i];
        qInfo().noquote() << "";
        qInfo().noquote() << QString("=== %1 (f/w = %2; temp = %3) ===")
                             .arg(i.getIdent())
                             .arg(devStatus.fwStr)
                             .arg(devStatus.temp, 0, 'f', 1);
        qInfo().noquote() << devStatus.toString();

        DevSerDesDelay devDelay;
        for(auto adcId : devStatus.data.keys()) {
            const auto &wPoint = devStatus.getOptimalDelays(adcId);
            if(!wPoint.status)
                continue;
            devDelay[adcId] = {wPoint.y, wPoint.x};
        }
        emit devSerDesDelayChanged(i, devDelay);
    }
    bool multi = delayStatusMap.size() > 1;
    if (multi) {
        const auto &totalStatus = AdcClkStatus::aggregate(delayStatusMap.values());
        qInfo().noquote() << "";
        qInfo().noquote() << QString("=== %1 ===").arg("Aggregated status");
        qInfo().noquote() << totalStatus.toString();
    }
}

void DominoDeviceManager::removeDevices(const DeviceIndexSet &setToDel)
{
    if(setToDel.isEmpty())
        return;
    QElapsedTimer quitTimer;
    quitTimer.start();

    for(const auto &i : setToDel) {
        Q_ASSERT(!i.isNull());
        emit deleteDevice(i);
        if(devices.contains(i))
            devices[i]->setDeleteFlag();
    }

    for(const auto &i : setToDel) {
        auto threadIt = devThreads.find(i);
        if(threadIt == devThreads.end()) {
            Q_ASSERT(!devices.contains(i));
            continue;
        }

        if(!threadQuiting(*threadIt, 100, &quitTimer)) {
            qWarning() << QString("Failed to stop device thread %1").arg(i.getIdent());
            (*threadIt)->terminate();
        }
        devices.remove(i);
        devThreads.remove(i);
    }
}

void DominoDeviceManager::maybeSendMonitorEvent(const EventData &eventData)
{
    const qint64 dt = monitorEvent->timestamp.msecsTo(eventData.timestamp);
    if (!monitorEvent->timestamp.isValid() || dt > monitorEventTimeMs || dt < 0) {
        *monitorEvent = eventData;
        emit monitorEventChanged(*monitorEvent);
    }
}

void DominoDeviceManager::onDeviceDataAcquired(const DeviceIndex &index, const DominoData &dominoData)
{
    quint64 key = dominoData.serial_id;

//    if (eventData->find(key) != eventData->end()) {
//    } else {
//    }
    if (eventData->size() > devices.size()) {
        qWarning() << "Event demultiplexer: Too much different id numbers, event buffer reset";
        eventData->clear();
    }
    if(eventData->isEmpty()) {
        eventTimeoutTimer->start();
        eventData->timestamp.start();
    }
    (*eventData)[key] = dominoData;

//    bool match = eventData->timestampTrigMatch();
//    // emit data only when all event numbers match
//    if (!match) {
//        qWarning() << "Timestamp mismatch";
//        return;
//    }

    if (eventData->size() == devices.size()) {
        eventTimeoutTimer->stop();
        eventData->FullEventNumber++;
        maybeSendMonitorEvent(*eventData);
        emit dataAcquired(*eventData);
        eventData->clear();
        //        qDebug() << "Event number Ok";
    }
}
