/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2010-2020 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "mainwindow.h"

#include <stdexcept>

#include <QDebug>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QFileDialog>
#include <QHostAddress>
#include <QLabel>
#include <QMap>
#include <QMessageBox>
#include <QSettings>
#include <QSplitter>
#include <QStatusBar>
#include <QStringList>
#include <QStyleOptionGroupBox>
#include <QTextCodec>
#include <QTextEdit>
#include <QThread>
#include <QTimer>
#include <QUrl>
#include <QVBoxLayout>

#include "config/Adc64ConfigConverter.h"
#include "Adc64SignalRouter.h"
#include "AddFirPresetDialog.h"
#include "CalibrationDialog.h"
#include "Adc64ConfigDb.h"
#include "CopyConfigDialog.h"
#include "DominoDeviceManager.h"
#include "EventData.h"
#include "EventWriter.h"
#include "PreamplifierDialog.h"
#include "ReadoutWindowDialog.h"
#include "SparseDialog.h"
#include "Adc64ConfigDispatcher.h"
#include "dominodevice.h"
#include "dominosettings.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/gui_utils.h"
#include "device-discover/DiscoverDialog.h"
#include "device-modules/AdcCore/AdcCoreStatus.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "qwt_plot_marker.h"
#include "qxw/WaveformPlot.h"
#include "regio/regiomlink.h"
#include "ui_mainwindow.h"
#include "util/QDateTimeCompat.h"
#include "waitcursor.h"

namespace  {

const int chMapV2[7][9] = {
    {5,2,1,4,3,29,26,31,30},
    {25,28,27,24,23,22,21,18,17},
    {14,13,10,9,12,11,8,7,6},
    {20,19,64,63,62,61,58,57,60},
    {39,38,37,34,33,36,35,16,15},
    {59,56,55,54,53,49,50,51,52},
    {48,47,46,45,42,41,44,43,40}
};
// <X, <guiCh, pcbCh> >
const QMap<int, QMap<int,int> > chMapV4X
( { {1, QMap<int, int>({{14, 28},{07, 29},{15, 30},{ 8, 31},{16, 32},{26, 33},{19, 34},{25, 35},{18, 36} })},
    {2, QMap<int, int>({{27, 59},{20, 58},{28, 57},{21, 56},{29, 55},{22, 54},{30, 53},{23, 52},{31, 51} })},
    {3, QMap<int, int>({{02, 19},{10, 20},{03, 21},{11, 22},{04, 23},{12, 24},{05, 25},{13, 26},{06, 27} })},
    {4, QMap<int, int>({{24, 50},{32, 49},{49, 48},{57, 47},{50, 46},{58, 45},{51, 44},{59, 43},{52, 42} })},
    {5, QMap<int, int>({{45, 10},{38, 11},{46, 12},{39, 13},{47, 14},{40, 15},{48, 16},{ 1, 17},{ 9, 18} })},
    {6, QMap<int, int>({{60, 41},{53, 40},{61, 39},{54, 38},{62, 37},{63, 63},{55, 62},{64, 61},{56, 60} })},
    {7, QMap<int, int>({{33, 01},{41, 02},{34, 03},{42, 04},{35, 05},{43, 06},{36, 07},{44,  8},{37,  9} })},
    {0, QMap<int, int>({{17, 64}})}
  } );
// <gui, pcbCh>
QMap<int,int> chMapV4;

const QString CH_PREFIX_OCTAL = "Octal";
const QString CH_PREFIX_AMP_X = "Amp X";

const int CH_AMP_X_GROUP_SIZE = 9;
typedef QPair<DeviceIndex, int> AmpGroupIndex;
}

Q_DECLARE_METATYPE(AmpGroupIndex)

MainWindow::MainWindow(const Adc64StartupOptions &startup_options, QWidget *parent) :
    BaseMainWindow(parent),
    ui(new Ui::MainWindow),
    copyConfigDialog(new CopyConfigDialog(this)),
    config(),
    configDb(new Adc64ConfigDb(startup_options.cs, this)),
    configDispatcher(new Adc64ConfigDispatcher(this)),
    configuration_manager(new ConfigurationManager(this, startup_options.cs.getType())),
    deviceDispatcher(new DeviceDispatcher(this)),
    writerLabel(new QLabel(this)),
    evLabel(new QLabel(this)),
    deviceMgr(new DominoDeviceManager(nullptr)),
    lastEventData(new EventData),
    eventWriter (new EventWriter),
    deviceMngThread(new QThread(this)),
    writerThread(new QThread(this)),
    channelsSetupForm(new ChannelsSetupForm(&config, this)),
    sparseDialog(new SparseDialog(&config, this)),
    preamplifier_dialog(new PreamplifierDialog(&config, this)),
    discoverDialog(new DiscoverDialog(this)),
    pcbSwitchingEnabled(startup_options.pcb_ch)
  , adcSerDesCalDialog(new AdcSerDesCalDialog(this))
{
    AbstractConfigConverter<Adc64Config>::registerTypes();
    qRegisterMetaType<DominoSettings>();
    qRegisterMetaType<DominoData>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<DeviceIndexEnabledMap>();

    auto signalRouter = &Adc64SignalRouter::instance();
    signalRouter->setMainWindow(this);
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->setDominoDeviceManager(deviceMgr);
    signalRouter->setConfigDispatcher(configDispatcher);
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setConfigDb(configDb);
    signalRouter->setDiscoverDialog(discoverDialog);
    signalRouter->setCopyConfigDialog(copyConfigDialog);
    signalRouter->setAdcSerDesCalDialog(adcSerDesCalDialog);
    signalRouter->connectAll();

    {
        std::map<int,int> tmpChMap2;
        for(const auto& m : chMapV4X.values()){
            const auto tmpChMap = m.toStdMap();
            tmpChMap2.insert(tmpChMap.begin(), tmpChMap.end());
        }
        chMapV4 = QMap<int,int>(tmpChMap2);
    }

    if(Globals::getDebugMode()){
        QMap<int, bool> allCh;
        for(int x=0;x<64/CH_AMP_X_GROUP_SIZE;++x){
            for(int y=0; y<CH_AMP_X_GROUP_SIZE; ++y){
                int ch = chMapV2[x][y];
                Q_ASSERT(!allCh.contains(ch));
                allCh[ch] = true;
            }
        }
        Q_ASSERT(allCh.size() == 64/CH_AMP_X_GROUP_SIZE*CH_AMP_X_GROUP_SIZE);
        allCh.clear();
    }

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    if(startup_options.pcb_ch){
        enabledPcbSwitching();
    }

    statusBar()->addPermanentWidget(writerLabel, 10);
    statusBar()->addPermanentWidget(evLabel, 2);
    ui->actionShow_PCB_ch->setVisible(pcbSwitchingEnabled);

    const auto &filter = deviceMgr->getSupportedDevices();
    LldpDiscovery::instance().setDeviceIdFilter(filter);


    deviceDispatcher->subscribeDevices(ui->deviceTableWidget);
//    deviceDispatcher->subscribeDevices(channelsSetupForm);
    connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
            ui->deviceTableWidget, &DeviceTableWidget::setDeviceList);


    // configDispatcher->subscribe(wrDaqDialog);
    // configDispatcher->subscribe(ui->deviceTableWidget);

    deviceMngThread->start();
    deviceMngThread->setObjectName("Adc64DevMng");
    deviceMgr->moveToThread(deviceMngThread);
    connect(deviceMgr, &DominoDeviceManager::destroyed, deviceMngThread, &QThread::quit);
    connect(this, &MainWindow::destroyAll, deviceMgr, &QObject::deleteLater);

    connect(deviceMgr, &DominoDeviceManager::deviceConnected,
            ui->deviceTableWidget, &DeviceTableWidget::onDeviceConnect);
    connect(deviceMgr, &DominoDeviceManager::deviceDisconnected,
            ui->deviceTableWidget, &DeviceTableWidget::onDeviceDisconnect);
    connect(deviceMgr, &DominoDeviceManager::deviceStatusUpdated,
            ui->deviceTableWidget, &DeviceTableWidget::deviceStatusUpdated);
    connect(deviceMgr, &DominoDeviceManager::statusUpdated,
            ui->deviceTableWidget, &DeviceTableWidget::updateAdcStatus);
    connect(deviceMgr, &DominoDeviceManager::adcLockStatusUpdated,
            ui->deviceTableWidget, &DeviceTableWidget::adcLockStatusUpdated);
    connect(deviceMgr, &DominoDeviceManager::adcRampStatusUpdated,
            ui->deviceTableWidget, &DeviceTableWidget::adcRampStatusUpdated);
    connect(deviceMgr, &DominoDeviceManager::adcClkDelayCheckUpdated,
            ui->deviceTableWidget, &DeviceTableWidget::adcDelayCheckUpdated);

    connect(this, &MainWindow::findAdcClockDelay,
            deviceMgr, &DominoDeviceManager::findAdcClockDelay);
    connect(this, &MainWindow::hardPatternTest,
            deviceMgr, &DominoDeviceManager::hardPatternTest);

    connect(this, &MainWindow::updateChannelsSetupFormSettings,
            channelsSetupForm, &ChannelsSetupForm::onSetSettings);
    connect(ui->deviceTableWidget, &DeviceTableWidget::selectedDeviceChanged,
            channelsSetupForm, &ChannelsSetupForm::onDeviceChange);
    connect(ui->deviceTableWidget, &DeviceTableWidget::selectedDeviceChanged,
            this, &MainWindow::onDeviceChange);

    connect(this, &MainWindow::channelOffsetUpdated,
            ui->deviceTableWidget, &DeviceTableWidget::updateChannelOffset);

    connect(ui->deviceTableWidget, &DeviceTableWidget::selectedDeviceChanged,
            sparseDialog, &SparseDialog::onDeviceChange);
    connect(ui->deviceTableWidget, &DeviceTableWidget::selectedDeviceChanged,
            adcSerDesCalDialog, &AdcSerDesCalDialog::onDeviceChange);
    connect(channelsSetupForm, &ChannelsSetupForm::channelsSetupChanged,
            this, &MainWindow::channelsSetupFormChanged);
    connect(sparseDialog, &SparseDialog::setupChanged,
            this, &MainWindow::channelsSetupFormChanged);
    connect(sparseDialog, &SparseDialog::setupChanged,
            this, &MainWindow::updateSparseMarkers);

    connect(ui->deviceTableWidget, &DeviceTableWidget::selectedDeviceChanged,
            preamplifier_dialog, &PreamplifierDialog::device_changed);
    connect(preamplifier_dialog, &PreamplifierDialog::config_changed,
            this, &MainWindow::channelsSetupFormChanged);

    block_signal_widgets_list.append(ui->checkBoxInvertSignal);
    block_signal_widgets_list.append(ui->spinBoxSize);
    block_signal_widgets_list.append(ui->spinBoxLatency);
    block_signal_widgets_list.append(ui->checkBoxTrigTimer);
    block_signal_widgets_list.append(ui->checkBoxTrigThreshold);
    block_signal_widgets_list.append(ui->comboBoxThrTrig);
    block_signal_widgets_list.append(ui->checkBoxTrigLemo);
    block_signal_widgets_list.append(ui->groupBoxDsp);
    block_signal_widgets_list.append(ui->spinBoxFirRoundoff);
    block_signal_widgets_list.append(ui->groupBoxTailCancellation);
    block_signal_widgets_list.append(ui->comboBoxFirPreset);
    block_signal_widgets_list.append(ui->comboBoxDspMafType);
    block_signal_widgets_list.append(ui->comboBoxDspMafSel);
    block_signal_widgets_list.append(ui->spinBoxDspBlcThr);
    block_signal_widgets_list.append(ui->groupBoxZS);
    block_signal_widgets_list.append(ui->comboBoxZSThr);
    block_signal_widgets_list.append(ui->checkBoxSoftwareZS);
    block_signal_widgets_list.append(ui->checkBoxSSM);
    block_signal_widgets_list.append(ui->deviceTableWidget);
    block_signal_widgets_list.append(ui->groupBoxMaf);

    ui->widget_settings->layout()->setSpacing(20);

//    QSplitter *splitter_vertical = new QSplitter();
//    splitter_vertical->setOrientation(Qt::Vertical);
//    splitter_vertical->addWidget(ui->widget_settings);
//    splitter_vertical->addWidget(ui->deviceTableWidget);
//    splitter_vertical->addWidget(ui->widget_plot);
//    splitter_vertical->setCollapsible(0, false);
//    splitter_vertical->setCollapsible(1, true);
//    splitter_vertical->setCollapsible(2, false);
//    splitter_vertical->setStretchFactor(0,1);
//    splitter_vertical->setStretchFactor(1,1);
//    splitter_vertical->setStretchFactor(2,10);

//    QGridLayout *layout_main = new QGridLayout();
//    layout_main->addWidget(splitter_vertical);
//    ui->centralWidget->setLayout(layout_main);

    QString gbox_style_1 = (
    "   QGroupBox::title{                   "
    "       background: white;              "
    "       border: 1px solid gray;         "
    "       border-radius: 3px;             "
    "       padding: 2px;                   "
    "   }                                   "
    "   QGroupBox{                          "
    "       background: rgb(216, 231, 235); "
    "       border: 1px solid gray;         "
    "       border-radius: 3px;             "
    "       margin-top: 6px;                "
    "   }                                   "
    );

    QString gbox_style_2 = (
    "   QGroupBox::title{                   "
    "       background: white;              "
    "       border: 1px solid gray;         "
    "       border-radius: 3px;             "
    "       padding: 2px;                   "
    "   }                                   "
    "   QGroupBox{                          "
    "       border: 1px solid gray;         "
    "       border-radius: 3px;             "
    "       margin-top: 6px;                "
    "   }                                   "
    );

    ui->groupBoxTailCancellation->setStyleSheet(gbox_style_1);
    ui->groupBoxMaf->setStyleSheet(gbox_style_1);
    ui->groupBoxDsp->setStyleSheet(gbox_style_2);

    ui->groupBox_ReadoutWindow->setStyleSheet(gbox_style_1);
    ui->groupBox_ReadoutWindow->setStyleSheet(gbox_style_1);
    ui->groupBoxTrigger->setStyleSheet(gbox_style_1);
    ui->groupBoxZS->setStyleSheet(gbox_style_1);

    QStringList edge = (QStringList() << "Rising" << "Falling");
    block_gui_signals(true);
    ui->comboBoxThrTrig->addItems(edge);
    ui->comboBoxZSThr->addItems(edge);

    ui->comboBoxDspMafType->addItem("None", 0);
    ui->comboBoxDspMafType->addItem("MAF", 1);
    ui->comboBoxDspMafType->addItem("Test", 2);
    ui->comboBoxDspMafType->addItem("MAF+Test", 3);

    block_gui_signals(false);

    QString styleLargeFont = "font-size: 18pt";
    ui->pushButtonStart->setStyleSheet(styleLargeFont);
    ui->pushButtonStop->setStyleSheet(styleLargeFont);
    ui->pushButtonStart->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButtonStop->setIcon(style()->standardIcon(QStyle::SP_MediaStop));

    ui->deviceTableWidget->initTable();

    configuration_manager->setDefaultConfig(Adc64Config());

    Q_EMIT updateChannelsSetupFormSettings();

    if(!Globals::getDebugMode()){
        ui->comboBoxDspMafType->hide();
        ui->label_9->hide();
        ui->actionHardPatternTest->setVisible(false);

        for(DeviceIndex index : config.device_configs.keys()){
            config.device_configs[selectedDevice].dspParams.test_enable = false; // FIXME: selectedDevice -> index
        }
    }
    emit configChanged(config);

    enableRunControlButtons();

    writerThread->start();
    eventWriter->moveToThread(writerThread);
    connect(eventWriter, &QObject::destroyed, writerThread, &QThread::quit);
    connect(eventWriter, &EventWriter::writerStatus,
            this, &MainWindow::writerStatusUpdated);
    connect(this, &MainWindow::destroyAll, eventWriter, &QObject::deleteLater);

}

MainWindow::~MainWindow()
{
    emit destroyAll();
    QElapsedTimer t;
    t.start();
    do {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);
    } while( (deviceMngThread->isRunning() || writerThread->isRunning())
            && (t.elapsed()<5000) );

    if(deviceMngThread->isRunning()){
        deviceMngThread->terminate();
        if (!deviceMngThread->wait(1000))
            qWarning() << "Device thread aborted";
    }
    if(writerThread->isRunning()){
        writerThread->terminate();
        if (!writerThread->wait(1000))
            qWarning() << "Writer thread aborted";
    }
    delete ui;
}

void MainWindow::setConfig(const Adc64Config &_config)
{
    // no need to check for modifications here, see ConfigDispatcher
    if(config == _config)
        return;
    config = _config;

    update_fir_combo_box();

    DeviceIndexEnabledMap devices;

    for(const auto &ds : config.device_configs) {
        devices.insert(ds.devIndex, DeviceIndexEnabled(ds.devIndex, ds.enabled));
    }

    emit deviceListChanged(devices);

    if(config.device_configs.isEmpty()){
        restoreFormValues();
    }
    updateSparseMarkers();

    setupChanged();
}

void MainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    const auto &cs = ConfigSelector(DaqModule(DaqModule::Type::Adc64), conf);
    const auto &title = cs.toWindowTitle();
    if(windowTitle() != title)
        setWindowTitle(title);
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}


void MainWindow::enabledPcbSwitching()
{
    pcbSwitchingEnabled = true;
    ui->actionShow_PCB_ch->setVisible(pcbSwitchingEnabled);
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    QMainWindow::closeEvent(event);
}

void MainWindow::channelsSetupFormChanged()
{
    setupChanged(); // TODO: move to setConfig
}

void MainWindow::onDeviceConnect(const DeviceIndex &i)
{
    enableRunControlButtons();
}

void MainWindow::onDeviceDisconnect(const DeviceIndex &i)
{
    enableRunControlButtons();
}

void MainWindow::deviceInfoUpdated(DeviceIndex index, int chNum, int maxSampleCnt, bool blockOffset)
{
    if(!config.device_configs.contains(index)){
        qWarning() << "There is no device in settings with index:" << hex << index;
//        return;
    }
    config.device_configs[index].setNCh(chNum);
    emit configChanged(config);

    sparseDialog->setOffsetMaximum(index, maxSampleCnt);
    sparseDialog->blockOffset(index, blockOffset);
    bool deviceMwStatusUpdated = updateDeviceMetaStatus(index, chNum, maxSampleCnt, blockOffset);

    if(deviceMwStatusUpdated) {
        restoreFormValues();
        totalChNum = 0;
        QMap<DeviceIndex, QPair<int, int> > devChOffset;
        for(const DominoSettings &ds : config.device_configs.values()) {
            if(!ds.enabled)
                continue;
            auto devStIt = devicesMetaInfo.find(ds.devIndex);
            if(devStIt != devicesMetaInfo.end()) {
                if(devStIt->nch) {
                    devChOffset[ds.devIndex] = devStIt->chRange
                            = {totalChNum, totalChNum+devStIt->nch-1};
                    totalChNum += devStIt->nch;
                }
            }
        }
        emit channelOffsetUpdated(devChOffset);
    }
    updateAdcChGroups();

//    ui->spinBoxWriteCNT->setMaximum(channelMaxSampleCount);
    if(selectedDevice==index && maxSampleCnt)
        ui->spinBoxSize->setMaximum(maxSampleCnt);

    if(ui->spinBoxSize->minimum()<4)
        ui->spinBoxSize->setMinimum(4);
    Q_EMIT updateChannelsSetupFormSettings();
}

void MainWindow::onMonitorEvent(const EventData &eventData)
{
    DrawScope(eventData);
}

void MainWindow::DrawScope(const EventData &eventData, bool force)
{
    if (!force && scopeDisplayTime.isValid() && scopeDisplayTime.elapsed() < 20) // update too fast
        return;

    if (!force && eventData.timestamp.isValid() && eventData.timestamp.elapsed() > 1000) { // data too old
        qWarning() << QString("DrawScope: event %1 dropped (old)").arg(eventData.getEventNumberStr());
        return;
    }

    *lastEventData = eventData;

    WaveformDataAll mvv;
    int adcOctetNum = ui->comboBoxOctalADC->currentIndex()-1;
    const bool displayAll = (adcOctetNum == -1);
//    QVector<double> trig_tai_ts;
    for(const DominoData &dominoData : eventData) {
        const DominoRawData &dd = dominoData.rawData;
        DominoSettings devSettings;
        for(const DominoSettings &ds : config.device_configs) {
            if(ds.devIndex.getSerial()==dominoData.serial_id){
                devSettings = ds;
                break;
            }
        }
        if(devSettings.devIndex.isNull()) continue; // Data from removed device
        Q_ASSERT(devSettings.devIndex.getSerial() == dominoData.serial_id);
//        trig_tai_ts.push_back(dominoData.trigStatus.trig_tai_ts.bit.toSeconds());

//        std::cout << QString("%1").arg(dominoData.trigStatus.trig_in_delay_ns).toStdString() << std::endl;

        int channelOffset = 0;
        auto devStIt = devicesMetaInfo.constFind(devSettings.devIndex);
        if(devStIt != devicesMetaInfo.constEnd())
            channelOffset = devStIt->chRange.first;
        for(int i : dd.keys()) {
            if (!displayAll && !displayEn[channelOffset+i])
                continue;
            if (dd[i].empty())
                continue;
            WaveformData wd;
            wd.c = channelOffset+i;
            if(pcbSwitchingEnabled && config.isShowPcbCh())
                wd.name = QString("pcb%1").arg(chMapV4.value(i+1, 0)).toStdString();
            else
                wd.name = QString("ch%1").arg(channelOffset+i+1).toStdString();

            wd.p.resize(dd[i].size());
            for(int j=0; j < dd[i].size(); j++)
            {
                wd.p[j].y = dd[i][j]-devSettings.chBaseline[i];
                wd.p[j].x = j;
//                double x_ns = j*dominoData.stepNs - dominoData.trigStatus.trig_in_delay_ns;
//                wd.p[j].x = x_ns / 1e3;
            }
            mvv.push_back(wd);
        }

        channelOffset += dd.size();
//        qDebug() << "channelOffset=" << channelOffset;
    }
//    // print timestamp delta
//    if (trig_tai_ts.size() > 1) {
//        for (int i=1; i<trig_tai_ts.size(); i++) {
//            double delta = trig_tai_ts[i] - trig_tai_ts[0];
//            std::cout << QString("%1").arg(delta, 0, 'f', 15).toStdString() << std::endl;
//        }
//    }
    ui->xPlot->setAllData(mvv);
//    evLabel->setText(QString("Ev %1").arg(eventData.getEventNumberStr()));
    evLabel->setText(QString("Ev %1").arg(eventData.FullEventNumber));
    //    dominoData.PrintRMS();
    scopeDisplayTime.start();
}

void MainWindow::writerStatusUpdated(const QString &st)
{
    writerLabel->setText(st);
}

void MainWindow::deviceUpdated(const DeviceIndex &index, const DeviceDescription &dd)
{
    if(!config.device_configs.contains(index)) return;

    updateDisplayedChannels();
}

void MainWindow::updateSparseMarkers()
{
    bool wasAutoReplot = ui->xPlot->autoReplot();
    ui->xPlot->setAutoReplot(false);

    foreach (QwtPlotMarker *m, sparseMarkers) {
        delete m;
    }
    sparseMarkers.clear();

    int period=0;
    int readCellNum=0;
    foreach (const DominoSettings &setting, config.device_configs.values()) {
        if(setting.enabled && setting.sparseParams.en) {
            period=setting.sparseParams.period;
            readCellNum = setting.sparseParams.readCellNum;
            readCellNum = qMin(readCellNum, setting.readCellNumber);
            break;
        }
    }

//    ui->checkBoxSSM->setEnabled(readCellNum);
    if(!readCellNum){
        ui->checkBoxSSM->hide();
    }
    else{
        ui->checkBoxSSM->show();
    }
    if(!config.isShowSparseMarkers())
        readCellNum=0;

    if(readCellNum == 0){
        ui->xPlot->replot();
    }

    for(int i=0; i<readCellNum; ++i) {
        auto * marker = new QwtPlotMarker();
        marker->setLineStyle(QwtPlotMarker::VLine);
        marker->setXValue(i*period);
        marker->label().font().italic();
        QwtText label(QString("sparse: %1").arg(i*period));
        QFont font = label.font();
        font.setItalic(true);
        label.setFont(font);
        marker->setLabel(label);
        marker->setLabelOrientation(Qt::Vertical);
        marker->setLabelAlignment(Qt::AlignRight | Qt::AlignBottom);
        marker->attach(ui->xPlot);
        sparseMarkers.push_back(marker);
    }
    ui->xPlot->setAutoReplot(wasAutoReplot);
}

void MainWindow::setupChanged()
{
    emit configChanged(config);
    for(const auto &setting : config.device_configs) {
        emit setupChangedSignal(setting);
    }
}

void MainWindow::enableRunControlButtons()
{
    ui->pushButtonStart->setEnabled(!config.isRunning() && deviceMgr->isAllConnected());
    ui->pushButtonStop->setEnabled(config.isRunning() && deviceMgr->isAllConnected());
    ui->actionCalibrate->setEnabled(!config.isRunning() && deviceMgr->isAllConnected());
}

void MainWindow::restoreFormValues()
{
    block_gui_signals(true);

    setWindowTitle(configDb->getWindowTitle());

    const bool present = config.device_configs.contains(selectedDevice);
    ui->actionCopyConfig->setEnabled(present);
    if(!present){
        block_gui_signals(false);
        return;
    }

    const DominoSettings &settings = config.device_configs[selectedDevice];
    ui->spinBoxSize->setMinimum(4);
    auto devStatusIt = devicesMetaInfo.find(selectedDevice);
    if(devStatusIt!=devicesMetaInfo.cend() && devStatusIt->maxSampleCnt){
        ui->spinBoxSize->setMaximum(devStatusIt->maxSampleCnt);
    }
    else{
        ui->spinBoxSize->setMaximum(settings.readCellNumber);
    }
    ui->spinBoxSize->setValue(settings.readCellNumber);
    int latMin = settings.trigDelay;
    int latMax = settings.trigDelay;

    if(devStatusIt != devicesMetaInfo.cend()) {
        latMin = devStatusIt->negLat ? -511:1;
        latMax = devStatusIt->negLat ? 512:65535;
    }

    ui->spinBoxLatency->setMinimum(latMin);
    ui->spinBoxLatency->setMaximum(latMax);
    ui->spinBoxLatency->setValue(settings.trigDelay);
    ui->checkBoxTrigTimer->setChecked(settings.trigsetup.bit.timer_en);
    ui->checkBoxTrigLemo->setChecked(settings.trigsetup.bit.ttl_en);
    ui->checkBoxTrigThreshold->setChecked(settings.trigsetup.bit.threshold_en);
    updateThresholdTrigger(settings.trigsetup.bit.threshold_en);

    ui->groupBoxZS->setChecked(settings.zs_en);

    ui->comboBoxThrTrig->setCurrentIndex(settings.invert_thr_trig ? 1 : 0);
    ui->comboBoxZSThr->setCurrentIndex(settings.invert_zs_thr ? 1 : 0);
    ui->checkBoxSoftwareZS->setChecked(settings.softwareZs);
    ui->checkBoxInvertSignal->setChecked(settings.invertInput);

    ui->actionPreamplifier->setEnabled(true);

    ui->groupBoxDsp->setChecked(settings.dspParams.en);
    ui->groupBoxTailCancellation->setChecked(settings.dspParams.fir.en);

    if(Globals::getDebugMode()){
        ui->groupBoxMaf->setCheckable(false);
    }
    else{
        ui->groupBoxMaf->setChecked(settings.dspParams.maf_enable);
    }
    ui->spinBoxFirRoundoff->setMinimum(0);
    ui->spinBoxFirRoundoff->setMaximum(settings.dspParams.fir.getRoundoffMax());
    ui->spinBoxFirRoundoff->setValue(settings.dspParams.fir.roundoff);
    const int mafTypeIndex = ui->comboBoxDspMafType->findData(settings.dspParams.getMafType());
    Q_ASSERT(mafTypeIndex != -1);

    ui->comboBoxDspMafType->setCurrentIndex(mafTypeIndex);


    ui->spinBoxDspBlcThr->setMaximum(32767);
    ui->spinBoxDspBlcThr->setValue(settings.dspParams.blc_thr);
    ui->spinBoxDspBlcThr->setMinimum(0);
    if (!ui->comboBoxDspMafSel->count()) {
        QStringList sl;
        sl += "2";
        sl += "4";
        sl += "8";
        sl += "16";
        ui->comboBoxDspMafSel->insertItems(0, sl);
    }
    ui->comboBoxDspMafSel->setCurrentIndex(settings.dspParams.maf_tap_sel);
    int firIndex = ui->comboBoxFirPreset->findText(settings.dspParams.fir.presetKey);
    if(firIndex == -1)
        firIndex = 0;
    ui->comboBoxFirPreset->setCurrentIndex(firIndex);
    ui->checkBoxSSM->setChecked(config.isShowSparseMarkers());

    ui->actionPeriodical_Adc_delay_check->setChecked(settings.periodicalAdcDelayCheck);

    block_gui_signals(false);
}

void MainWindow::updateThresholdTrigger(bool enabled)
{
    ui->comboBoxThrTrig->setEnabled(enabled);
//    ui->labelThresholdTrigger->setEnabled(enabled);
}

void MainWindow::updateDisplayedChannels()
{
    displayEn.clear();

    bool displayOctet = false;
    int adcOctetNum = 0;

    bool displayAmpX = false;
    QVector<int> chGroupVec(CH_AMP_X_GROUP_SIZE);

    QString str = ui->comboBoxOctalADC->currentText();
    if(str.startsWith(CH_PREFIX_OCTAL)){
        QStringList var = str.remove(0, CH_PREFIX_OCTAL.size()).trimmed().split(' ');
        adcOctetNum = QString(var[0]).toInt(&displayOctet);
    } else if(str.startsWith(CH_PREFIX_AMP_X)){
        AmpGroupIndex ampGroupIndex = ui->comboBoxOctalADC->currentData().value<AmpGroupIndex>();

        displayAmpX = config.device_configs.contains(ampGroupIndex.first);
        if(displayAmpX) {
            const int xChGroup = ampGroupIndex.second;
            chGroupVec = QVector<int>::fromList(chMapV4X[xChGroup].keys());

            // const int hwVer = config.device_configs[ampGroupIndex.first].getHwVer();
            // switch (hwVer) {
            // case 2:
            //     qCopy(chMapV2[xChGroup-1], chMapV2[xChGroup-1]+CH_AMP_X_GROUP_SIZE, chGroupVec.begin()); // FIXME
            //     break;
            // case 4:
            //    chGroupVec = QVector<int>::fromList(chMapV4X[xChGroup].keys());
            //     break;
            // default:
            //     displayAmpX = false;
            //     break;
            // }

            // Shift ch in chGroupVec
            int chOffset = 0;
            foreach (const DominoSettings &ds, config.device_configs) {
                if(ds.devIndex == ampGroupIndex.first) {
                    if(chOffset) {
                        for(int i=0; i<CH_AMP_X_GROUP_SIZE; ++i) {
                            chGroupVec[i] += chOffset;
                        }
                    }
                    break;
                } else
                    chOffset += ds.nch;
            }
        }
    }

    for(int ch=0; ch<totalChNum; ++ch){
        bool v = true;
        if(displayOctet)
            // Octal
            v = ((ch/8)==(adcOctetNum-1));
        else if(displayAmpX)
            // Amp X
            v = chGroupVec.contains(ch+1);
        else
            // All
            v = true;

        displayEn[ch] = v;
    }
}

void MainWindow::updateAdcChGroups()
{
    QMap<int, QString> channelGroups;
    for(int i=0; i<totalChNum/8; i++)
        channelGroups[i+1] = QString("%1 %2 (%3-%4)")
                .arg(CH_PREFIX_OCTAL).arg(i+1)
                .arg(i*8+1).arg((i+1)*8);

    ui->comboBoxOctalADC->clear();
    int id = 0;
    ui->comboBoxOctalADC->insertItem(id++, QString("All channels (%1-%2)").arg(totalChNum>0?1:0).arg(totalChNum), 0);
    for(int i : channelGroups.keys())
        ui->comboBoxOctalADC->insertItem(id++, channelGroups[i], i);

    for(const DominoSettings &ds : config.device_configs) {
        if(ds.enabled) {
            for(int xGroupIndex=1; xGroupIndex<=ds.nch/CH_AMP_X_GROUP_SIZE; ++xGroupIndex){
                ui->comboBoxOctalADC->insertItem(id++, QString("%1%2 %3")
                                                 .arg(CH_PREFIX_AMP_X).arg(xGroupIndex)
                                                 .arg(ds.devIndex.getSerialStr()),
                                                 QVariant::fromValue(AmpGroupIndex(ds.devIndex, xGroupIndex)));
            }
        }
    }
}

void MainWindow::update_fir_combo_box()
{
    ui->comboBoxFirPreset->blockSignals(true);
    const QStringList firPresets = config.fir_map.keys();
    ui->comboBoxFirPreset->clear();
    ui->comboBoxFirPreset->insertItem(0, "None", "");
    for(const QString &preset : firPresets) {
        if(!preset.isEmpty()){
            ui->comboBoxFirPreset->insertItem(firPresets.size(), preset, preset);
        }
    }
    ui->comboBoxFirPreset->blockSignals(false);
}

bool MainWindow::updateDeviceMetaStatus(DeviceIndex index, int chNum, int maxSampleCnt, bool blockOffset)
{
    bool ret = !devicesMetaInfo.contains(index);
    if(!ret) {
        auto &v = devicesMetaInfo[index];
        ret |=
                v.maxSampleCnt != maxSampleCnt ||
                v.negLat != blockOffset ||
                v.nch != chNum;
    }
    if(ret) {
        auto &v = devicesMetaInfo[index];
        v.maxSampleCnt = maxSampleCnt;
        v.negLat = blockOffset;
        v.nch = chNum;
    }
    return ret;
}

void MainWindow::block_gui_signals(bool block)
{
    for(int i=0; i<block_signal_widgets_list.size(); ++i){
        block_signal_widgets_list[i]->blockSignals(block);
    }
}

//--- GUI slots
//--- options
void MainWindow::on_actionChange_device_triggered()
{
    on_pushButtonFindDevice_clicked();
}

void MainWindow::on_actionCalibrate_triggered()
{
    if(config.isRunning()){
        on_pushButtonStop_clicked();
     }

    // FIXME: use temporary copy of config
    QMap<DeviceIndex, unsigned int> timer_en_map;
    QMap<DeviceIndex, int> read_cell_number_map;
    QMap<DeviceIndex, int> trig_delay_map;
    QMap<DeviceIndex, bool> zs_en;
    QMap<DeviceIndex, QMap<int, bool>> ch_en_map;

    foreach (const DeviceIndex &index, config.device_configs.keys()){
        timer_en_map[index] = config.device_configs[index].trigsetup.bit.timer_en;
        read_cell_number_map[index] = config.device_configs[index].readCellNumber;
        trig_delay_map[index] = config.device_configs[index].trigDelay;
        zs_en[index] = config.device_configs[index].zs_en;
        ch_en_map[index] = config.device_configs[index].chEn;
    }

    foreach (const DeviceIndex &index, config.device_configs.keys()) {
        config.device_configs[index].trigsetup.bit.timer_en = 1;
        config.device_configs[index].readCellNumber = 1000;
        config.device_configs[index].trigDelay = 0;
        config.device_configs[index].zs_en = false;
        config.device_configs[index].set_en_all(true);
    }

    on_pushButtonStart_clicked();

    CalibrationDialog dialog(&config, this);
    connect(deviceMgr, &DominoDeviceManager::dataAcquired,
            &dialog, &CalibrationDialog::onDataAcquired);
    if(dialog.exec() == QDialog::Accepted) {
        foreach (const DeviceIndex &index, config.device_configs.keys())
            if(config.device_configs[index].enabled){
                config.device_configs[index].chBaseline = dialog.getBaseLineByIndex(index);
             }
        emit configChanged(config);
    }

    on_pushButtonStop_clicked();

    foreach (const DeviceIndex &index, config.device_configs.keys()) {
        config.device_configs[index].trigsetup.bit.timer_en = timer_en_map[index];
        config.device_configs[index].readCellNumber = read_cell_number_map[index];
        config.device_configs[index].trigDelay = trig_delay_map[index];
        config.device_configs[index].zs_en = zs_en[index];
        config.device_configs[index].chEn = ch_en_map[index];
    }

    emit configChanged(config);
    Q_EMIT updateChannelsSetupFormSettings();
}

void MainWindow::on_actionADC_Ser_Des_calibration_triggered()
{
    adcSerDesCalDialog->show();
}

void MainWindow::on_actionPeriodical_Adc_delay_check_toggled(bool val)
{
    config.setPeriodicalAdcDelayCheck(val);
    setupChanged();
}

void MainWindow::on_actionHardPatternTest_triggered(bool val)
{
    ui->actionPeriodical_Adc_delay_check->setEnabled(!val);
    if(Globals::getDebugMode()) {
        hardPatternTestEnabled = val;
        emit hardPatternTest(hardPatternTestEnabled);
    }
//    else if(val) {
//        ui->actionHardPatternTest->setChecked(false);
//        qWarning() << __func__ << "is blocked in non-debug mode";
//    }
}

void MainWindow::on_actionSet_read_out_window_triggered()
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];

    ReadoutWindowDialog dialog(settings.readCellNumber, settings.trigDelay, this);
    if(dialog.exec() == QDialog::Accepted){
        settings.readCellNumber = dialog.getReadCount();
        settings.trigDelay = dialog.getTrigDelay();
        setupChanged();
        restoreFormValues();
    }
}

void MainWindow::on_actionSparse_parameters_triggered()
{
    sparseDialog->show();
}

void MainWindow::on_actionCopyConfig_triggered()
{
    copyConfigDialog->show();
}

void MainWindow::on_actionShow_PCB_ch_toggled(bool val)
{
    config.setShowPcbCh(val);
    emit configChanged(config);
}

void MainWindow::on_actionPreamplifier_triggered()
{
    preamplifier_dialog->show();
}

void MainWindow::on_actionAdd_FIR_preset_triggered()
{
    AddFirPresetDialog fir_dialog(&config, this);
    if(fir_dialog.exec() == QDialog::Accepted) {
        update_fir_combo_box();
        emit configChanged(config);
    }
}

void MainWindow::on_actionRemove_all_presets_triggered()
{
    QDialog dialog;
    QVBoxLayout layout(&dialog);
    QLabel label("Remove all FIR presets?");
    layout.addWidget(&label);
    QDialogButtonBox * button_box = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    connect(button_box, &QDialogButtonBox::accepted, &dialog, &QDialog::accept);
    connect(button_box, &QDialogButtonBox::rejected, &dialog, &QDialog::reject);
    layout.addWidget(button_box);

    if(dialog.exec() == QDialog::Accepted)
    {
        config.fir_map.clear();
        for(int i=ui->comboBoxFirPreset->count(); i>0; --i){
            ui->comboBoxFirPreset->removeItem(i);
        }
        emit configChanged(config);
//        setupChanged(); //todo: ???
    }
}
//--about
void MainWindow::on_actionDocs_triggered()
{
    QDesktopServices::openUrl(QUrl("https://afi-project.jinr.ru/projects/adc64-sw/wiki/"));
}

void MainWindow::on_actionChannel_map_triggered()
{
    if(!channelMapDialog){
        channelMapDialog = new QDialog(this);
        channelMapDialog->setWindowTitle("Channel map help");
        auto *view = new QTextEdit(channelMapDialog);
        view->setReadOnly(true);
        QFile htmlFile(":/channelMap.html");
        QString html;
        if(htmlFile.open(QIODevice::ReadOnly)){
            html = QString(htmlFile.readAll());
            htmlFile.close();
            if(html.isEmpty())
                html = "<b>Channel map file is empty.</b>";
        } else
            html = "<b>There is no channel map file.</b>";
        view->setText(html);
        channelMapDialog->setLayout(new QVBoxLayout(channelMapDialog));
        channelMapDialog->layout()->addWidget(view);
        channelMapDialog->resize(800, 600);
    }
    channelMapDialog->show();
}

void MainWindow::on_actionCheckForUpdates_triggered()
{
    QDesktopServices::openUrl(QUrl("http://afi.jinr.ru/Downloads/ADC64"));
}

//--readout window
void MainWindow::on_spinBoxSize_valueChanged(int newval)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
//#ifdef QT_NO_DEBUG
//    // pass only even value if not debug
//    if(newval & 1) {
//        ui->spinBoxReadCNT->setValue(newval+1);
//        return;
//    }
//#endif
    settings.readCellNumber = newval;
    setupChanged();
    updateSparseMarkers();
}

void MainWindow::on_spinBoxLatency_valueChanged(int newval)
{
    if(!config.device_configs.contains(selectedDevice))
        return;
    config.device_configs[selectedDevice].trigDelay=newval;
    setupChanged();
}

//--trigger
void MainWindow::on_checkBoxTrigTimer_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.trigsetup.bit.timer_en = checked;
    setupChanged();
}


void MainWindow::on_checkBoxTrigThreshold_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.trigsetup.bit.threshold_en = checked ? 1:0 ;
    updateThresholdTrigger(checked);
    Q_EMIT updateChannelsSetupFormSettings();
    setupChanged();
}

void MainWindow::on_comboBoxThrTrig_currentIndexChanged(const QString &arg1)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.invert_thr_trig = (arg1 != "Rising");
    setupChanged();
}

void MainWindow::on_checkBoxTrigLemo_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.trigsetup.bit.ttl_en = checked ? 1:0 ;
    setupChanged();
}

//--dsp roundoff
void MainWindow::on_groupBoxDsp_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.dspParams.en = checked;
    setupChanged();
//   Q_EMIT updateChannelsSetupFormSettings();
}

void MainWindow::on_checkBoxInvertSignal_toggled(bool invert)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.setInvert(invert);
    setupChanged();
    restoreFormValues();
    Q_EMIT updateChannelsSetupFormSettings();
}

void MainWindow::on_spinBoxFirRoundoff_valueChanged(int value)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.dspParams.fir.roundoff = value;
    setupChanged();
}
//--dsp tail
void MainWindow::on_groupBoxTailCancellation_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.dspParams.fir.en = checked;
    setupChanged();
}

void MainWindow::on_comboBoxFirPreset_currentIndexChanged(int index)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    const QString firPreset = ui->comboBoxFirPreset->itemData(index).toString();
    config.device_configs[selectedDevice].dspParams.fir.presetKey = firPreset;
    setupChanged();
}

//--dsp maf
void MainWindow::on_groupBoxMaf_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    if (config.device_configs[selectedDevice].dspParams.maf_enable != checked) {
        config.device_configs[selectedDevice].dspParams.maf_enable = checked;
        setupChanged();
    }
}

void MainWindow::on_comboBoxDspMafType_currentIndexChanged(int index)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    const uint v = ui->comboBoxDspMafType->itemData(index).toUInt();
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.dspParams.maf_enable = v & 1;
    settings.dspParams.test_enable = v & 2;
    setupChanged();
}

void MainWindow::on_comboBoxDspMafSel_activated(int index)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.dspParams.maf_tap_sel = index;
    setupChanged();
}

void MainWindow::on_spinBoxDspBlcThr_valueChanged(int value)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.dspParams.blc_thr = value;
    setupChanged();
}

//--zero suppr
void MainWindow::on_groupBoxZS_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.zs_en = checked;
    setupChanged();
    Q_EMIT updateChannelsSetupFormSettings();
}

void MainWindow::on_comboBoxZSThr_currentIndexChanged(const QString &arg1)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.invert_zs_thr = (arg1 != "Rising");
    setupChanged();
}

void MainWindow::on_checkBoxSoftwareZS_toggled(bool checked)
{
    if(!config.device_configs.contains(selectedDevice)) return;
    DominoSettings &settings = config.device_configs[selectedDevice];
    settings.softwareZs = checked;
    setupChanged();
}

//--find device
void MainWindow::on_pushButtonFindDevice_clicked()
{
    discoverDialog->show();
}


void MainWindow::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    QMap<DeviceIndex, DominoSettings> oldSettings = config.device_configs;
    config.device_configs.clear();

    foreach (const auto &dd, devices) {
        DominoSettings &ds = config.device_configs[dd.index];
        if(oldSettings.contains(dd.index)){
            ds = oldSettings.take(dd.index);
        } else {
            ds.devIndex = dd.index;
        }
        ds.enabled = dd.enabled;
    }
    config.blockSparseSetup();
    emit configChanged(config);
    if(Globals::getDebugMode()) {
        emit hardPatternTest(hardPatternTestEnabled);
    }
    setupChanged();

    restoreFormValues();
}

void MainWindow::onDeviceChange(const DeviceIndex &index)
{
    selectedDevice = index;
    restoreFormValues();
    emit selectedDeviceChanged(selectedDevice);
}

//--start/stop
void MainWindow::on_pushButtonStart_clicked()
{
    if(config.isRunning())
        return;
    config.setRun(true);
    emit clearEventNumber();
    setupChanged();
    enableRunControlButtons();
}

void MainWindow::on_pushButtonStop_clicked()
{
    if (!config.isRunning())
        return;
    config.setRun(false);
    setupChanged();
    enableRunControlButtons();
}

void MainWindow::on_checkBoxW2D_toggled(bool checked)
{
    if (checked)
    {
        QString fileName = QFileDialog::getSaveFileName(this,
                                                        tr("Save File"),
                                                        eventWriter->dataFileName,
                                                        tr("Text files *.txt;;All files *.*"));
        if (!fileName.isNull()) {
            QMetaObject::invokeMethod(eventWriter, "setFileName", Q_ARG(QString, fileName));
            connect(deviceMgr, &DominoDeviceManager::dataAcquired,
                    eventWriter, &EventWriter::SaveData);
        }
        else ui->checkBoxW2D->setChecked(false);
    } else {
        QMetaObject::invokeMethod(eventWriter, "close");
        disconnect(deviceMgr, &DominoDeviceManager::dataAcquired, eventWriter, nullptr);
    }
}

void MainWindow::on_pushButtonChannelsSetup_clicked()
{
    channelsSetupForm->show();
}

void MainWindow::on_comboBoxOctalADC_currentIndexChanged(int)
{
    updateDisplayedChannels();

    DrawScope(*lastEventData, true);
}

void MainWindow::on_checkBoxSSM_toggled(bool checked)
{
    config.setShowSparseMarkers(checked);
    emit configChanged(config);
    updateSparseMarkers();
}
