/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2010-2020 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QElapsedTimer>

#include "Adc64Config.h"
#include "PreamplifierDialog.h"
#include "adc64-common/Adc64StartupOptions.h"
#include "adc64-common/DeviceMetaInfo.h"
#include "adc64-common/dominodata.h"
#include "base/app/BaseMainWindow.h"
#include "channelssetupform.h"
#include "configuration-manager/ConfigurationManager.h"

namespace Ui {
    class MainWindow;
}

class AdcStatus;
class CopyConfigDialog;
class Adc64ConfigDb;
class Adc64ConfigDispatcher;
class AdcSerDesCalDialog;
class DeviceDescription;
class DeviceDispatcher;
class DeviceIndexEnabledMap;
class DiscoverDialog;
class DominoData;
class DominoDeviceManager;
class EventData;
class EventWriter;
class LldpInfo;
class QDialog;
class QLabel;
class QThread;
class QwtPlotMarker;
class SparseDialog;

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(const Adc64StartupOptions &startup_options, QWidget *parent = nullptr);
    ~MainWindow() override;
    void enabledPcbSwitching();

Q_SIGNALS:
    void switch_to_config(const ConfigIndexName &conf) const;
    void setupChangedSignal(const DominoSettings &) const;
    void updateChannelsSetupFormSettings() const;
    void destroyAll() const;
    void clearEventNumber() const;
//    void setDeviceIdFilter(const DeviceIdSet &r);
    void channelOffsetUpdated(const QMap<DeviceIndex, QPair<int, int> > &) const;
    void configChanged(const Adc64Config &) const;
    void deviceListChanged(const DeviceIndexEnabledMap &) const;
    void findAdcClockDelay(bool hardMode, const QString &dirPath) const;
    void selectedDeviceChanged(const DeviceIndex &index) const;
    void hardPatternTest(bool en);

public slots:
    void setConfig(const Adc64Config &_config);
    void setDeviceList(const DeviceIndexEnabledMap &devices);
    void configSelectorChanged(const ConfigIndexName &conf);
    void onDeviceConnect(const DeviceIndex &i);
    void onDeviceDisconnect(const DeviceIndex &i);
    void deviceInfoUpdated(DeviceIndex index, int chNum, int maxSampleCnt, bool blockOffset);
    void onMonitorEvent(const EventData &);

protected:
    void changeEvent(QEvent *e) override;
    void closeEvent(QCloseEvent *event) override;

private Q_SLOTS:
    //--Menu bar slots
    //options
    void on_actionChange_device_triggered();
    void on_actionCalibrate_triggered();
    void on_actionADC_Ser_Des_calibration_triggered();
    void on_actionPeriodical_Adc_delay_check_toggled(bool val);
    void on_actionHardPatternTest_triggered(bool val);
    void on_actionSet_read_out_window_triggered();
    void on_actionSparse_parameters_triggered();
    void on_actionPreamplifier_triggered();
    void on_actionCopyConfig_triggered();
    void on_actionShow_PCB_ch_toggled(bool val);
    void on_actionAdd_FIR_preset_triggered();
    void on_actionRemove_all_presets_triggered();
    void on_actionConfiguration_manager_triggered();
    //about
    void on_actionDocs_triggered();
    void on_actionChannel_map_triggered();
    void on_actionCheckForUpdates_triggered();

    // Other slots
    void channelsSetupFormChanged();
    void writerStatusUpdated(const QString &st);
    void deviceUpdated(const DeviceIndex &index, const DeviceDescription &dd);
    void updateSparseMarkers();

    //--GUI slots
    void on_checkBoxInvertSignal_toggled(bool invert);

    //--readout window
    void on_spinBoxSize_valueChanged(int newval);
    void on_spinBoxLatency_valueChanged(int newval);

    //--trigger
    void on_checkBoxTrigTimer_toggled(bool checked);
    void on_checkBoxTrigThreshold_toggled(bool checked);
    void on_comboBoxThrTrig_currentIndexChanged(const QString &arg1);
    void on_checkBoxTrigLemo_toggled(bool checked);

    //--dsp roundoff
    void on_groupBoxDsp_toggled(bool checked);
    void on_spinBoxFirRoundoff_valueChanged(int value);
    //--dsp tail
    void on_groupBoxTailCancellation_toggled(bool checked);
    void on_comboBoxFirPreset_currentIndexChanged(int index);
    //--dsp maf
    void on_groupBoxMaf_toggled(bool checked);
    void on_comboBoxDspMafType_currentIndexChanged(int index);
    void on_comboBoxDspMafSel_activated(int index);
    void on_spinBoxDspBlcThr_valueChanged(int value);

    //--zero
    void on_groupBoxZS_toggled(bool checked);
    void on_comboBoxZSThr_currentIndexChanged(const QString &arg1);
    void on_checkBoxSoftwareZS_toggled(bool checked);

    //--find device
    void on_pushButtonFindDevice_clicked();
    void onDeviceChange(const DeviceIndex &index);

    //--start/stop
    void on_pushButtonStart_clicked();
    void on_pushButtonStop_clicked();
    void on_checkBoxW2D_toggled(bool checked);
    void on_pushButtonChannelsSetup_clicked();
    void on_comboBoxOctalADC_currentIndexChanged(int);
    void on_checkBoxSSM_toggled(bool checked);

private:
    void DrawScope(const EventData &, bool force = false);
    void enableRunControlButtons();
    void setupChanged();
    void block_gui_signals(bool block);
    void restoreFormValues();
    void updateThresholdTrigger(bool enabled);
    void updateDisplayedChannels();
    void updateAdcChGroups();
    void update_fir_combo_box();
    bool updateDeviceMetaStatus(DeviceIndex index, int chNum, int maxSampleCnt, bool blockOffset);

    Ui::MainWindow * const ui;
    CopyConfigDialog * const copyConfigDialog;
    Adc64Config config;
    Adc64ConfigDb * const configDb;
    Adc64ConfigDispatcher * const configDispatcher;
    ConfigurationManager * const configuration_manager;
    DeviceDispatcher * const deviceDispatcher;
    QList<QWidget*> block_signal_widgets_list;
    QLabel * const writerLabel;
    QLabel * const evLabel;
    DominoDeviceManager * const deviceMgr;
    EventData * const lastEventData;
    QElapsedTimer scopeDisplayTime;
    EventWriter * const eventWriter;
    QThread * const deviceMngThread;
    QThread * const writerThread;
    ChannelsSetupForm * const channelsSetupForm;
    SparseDialog * const sparseDialog;
    PreamplifierDialog * preamplifier_dialog;
    DiscoverDialog * const discoverDialog;
    QDialog * channelMapDialog = nullptr;
    DeviceIndex selectedDevice;
    int totalChNum = 0;
    QMap<int, bool> displayEn; // which channels we should draw
    QVector<QwtPlotMarker *> sparseMarkers;
    bool pcbSwitchingEnabled;
    QMap<DeviceIndex, DeviceMetaInfo> devicesMetaInfo;
    bool hardPatternTestEnabled = false;
    AdcSerDesCalDialog * const adcSerDesCalDialog;
};

#endif // MAINWINDOW_H
