/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2014-2020 Ilia Slepnev
**    Copyright 2014-2015 Nikita Tarasov
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <limits>
#include <math.h>

#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QSettings>
#include <QStatusBar>
#include <QTimer>

#include "device-discover/DiscoverDialog.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mregdev/device_id.h"
#include "mregdev/qtmregdevice.h"

namespace {
enum {COL_ID, COL_SERIAL, COL_STATE, COL_FIRMWARE, COL_TEMP, COL_NUM};
}

MainWindow::MainWindow(QWidget *parent):
    BaseMainWindow(parent),
    ui(new Ui::MainWindow),
    deviceDispatcher(new DeviceDispatcher(this)),
    discoverDialog(new DiscoverDialog(this))
{
    qRegisterMetaType<DeviceIndex>();

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, nullptr);

    ui->mainToolBar->hide();

    config.loadSettings();

    ui->comboBoxRegType->addItem("16-bit", Config::REG_TYPE_16);
    ui->comboBoxRegType->addItem("32-bit", Config::REG_TYPE_32);
    ui->comboBoxRegType->addItem("64-bit", Config::REG_TYPE_64);

    ui->comboBoxRegAddr->addItem("MLink::CTRL 16-bit", 0x40);
    ui->comboBoxRegAddr->addItem("MLink::Temp 16-bit", 0x54);
    ui->comboBoxRegAddr->addItem("MLink::Temp2 16-bit", 0x4b);
    ui->comboBoxRegAddr->addItem("MLink::DevId 16-bit", 0x42);
    ui->comboBoxRegAddr->addItem("MLink::OneWireID 64-bit", 0x48);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(discoverDialog);

    deviceDispatcher->subscribeDevices(this);
    deviceDispatcher->subscribeDevices(discoverDialog);
    // deviceDispatcher->subscribeHostAddressUpdates(dev);

    discoverDialog->setSingleMode(true);
    DeviceIndexEnabledMap map;
    if (! config.deviceIndex.isNull()) {
        map[config.deviceIndex] = DeviceIndexEnabled(config.deviceIndex, true);
        emit deviceListChanged(map);
    }
    restoreGui();
}

MainWindow::~MainWindow()
{
    saveConfig();

    QElapsedTimer timer;
    timer.start();
    if(dev) {
        delete dev;
    }
    delete ui;
}

void MainWindow::on_pushButtonChange_clicked()
{
    discoverDialog->show();
}

void MainWindow::on_comboBoxRegAddr_currentIndexChanged(int)
{
    ui->radioButtonPreDefReg->setChecked(true);
}

void MainWindow::on_lineEditRegAddr_textEdited(const QString)
{
    ui->radioButtonCustomReg->setChecked(true);
}

void MainWindow::on_comboBoxRegType_currentIndexChanged(int)
{
    updateLineEditWriteLen();
}

void MainWindow::on_pushButtonRead_clicked()
{
    if(!saveConfig(false)) {
        qWarning() << "Can't read. Some field has incorrect value";
        return;
    }

    if(!dev || !dev->isConnected()) {
        if(dev)
            qWarning() << "Can't read. Device disconnected";
        else
            qWarning() << "Can't read. Device undefined";
        return;
    }

    quint64 val=0;

    switch (config.regType) {
    case Config::REG_TYPE_16:
        val = dev->regRead(config.regAddr);
        break;
    case Config::REG_TYPE_32:
        val = dev->regRead32(config.regAddr);
        break;
    case Config::REG_TYPE_64:
        val = dev->regRead64(config.regAddr);
        break;
    }
    int size = 2 * (2<<config.regType);
    ui->labelRegVal->setText(QString("%L1 (0x%2)").arg(val).arg(val, size, 16, QChar('0')));

    qInfo() << QString("Read reg value at 0x%1: %2 (0x%3)")
               .arg(config.regAddr, 0, 16)
               .arg(val)
               .arg(val, 0, 16);
}

void MainWindow::on_pushButtonWrite_clicked()
{
    if(!saveConfig()) {
        qWarning() << "Can't read. Some field has incorrect value";
        return;
    }

    if(!dev || !dev->isConnected()) {
        if(dev)
            qWarning() << "Can't read. Device disconnected";
        else
            qWarning() << "Can't read. Device undefined";
        return;
    }

    switch (config.regType) {
    case Config::REG_TYPE_16:
        dev->regWrite(config.regAddr, static_cast<quint16>(config.regWriteVal));
        break;
    case Config::REG_TYPE_32:
        dev->regWrite32(config.regAddr, static_cast<quint32>(config.regWriteVal));
        break;
    case Config::REG_TYPE_64:
        dev->regWrite64(config.regAddr, config.regWriteVal);
        break;
    }

    qInfo() << QString("Set reg value at 0x%1 to %2 (0x%3)")
               .arg(config.regAddr, 0, 16)
               .arg(config.regWriteVal)
               .arg(config.regWriteVal, 0, 16);
}


void MainWindow::deviceConnected()
{
    if(dev && dev==sender()){
        ui->labelDev->setStyleSheet("QLabel { color : green; }");
    }
}

void MainWindow::deviceDisconnected()
{
    if(dev && dev==sender()){
        ui->labelDev->setStyleSheet("QLabel { color : red; }");
    }
}

void MainWindow::deviceTempChanged(const DeviceIndex &, double temp)
{
    qInfo() << QString("%1%2").arg(temp,4,'f',1).arg("C");
}

void MainWindow::deviceStatusUpdated(const DeviceIndex &, QString s)
{
    statusBar()->showMessage(s);
    qInfo() << "StatusUpdated:" << s;
}

QtMregDevice *MainWindow::createDevice(const DeviceIndex &index)
{
    QtMregDevice *d = new QtMregDevice(index, this);
    deviceDispatcher->subscribeHostAddressUpdates(d->getDeviceIndex(), d);
    d->setBlockLiveMagic();

    connect(d, &QtMregDevice::deviceConnected,
            this, &MainWindow::deviceConnected);

    connect(d, &QtMregDevice::deviceDisconnected,
            this, &MainWindow::deviceDisconnected);

    connect(d, &QtMregDevice::setStatus,
            this, &MainWindow::deviceStatusUpdated);

    connect(d, &QtMregDevice::TempChanged,
            this, &MainWindow::deviceTempChanged);

    return d;
}

void MainWindow::addDevice(const DeviceIndex &index)
{
    if (dev)
        delete dev;

    dev = createDevice(index);
    config.deviceIndex = index;
    config.saveSettings();
    ui->labelDev->setText(dev->getIdent());
    qInfo() << QString("Added device %1").arg(dev->getIdent());
}

void MainWindow::removeDevice(const DeviceIndex &index)
{
    if (!dev)
        return;
    if (index != dev->getDeviceIndex())
        return;
    qInfo() << "Removed device" << index;
    delete dev;
    dev = nullptr;
    config.deviceIndex = DeviceIndex();
}

DeviceIndexSet getDeviceIndexSetFromMap(const DeviceIndexEnabledMap &devices)
{
    DeviceIndexSet set;
    for (auto i : devices)
        if (i.enabled)
            set.insert(i.index);
    return set;
}

void MainWindow::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    static DeviceIndexEnabledMap deviceListCache;
    if (devices == deviceListCache)
        return;
    const auto old_set = getDeviceIndexSetFromMap(deviceListCache);
    const auto new_set = getDeviceIndexSetFromMap(devices);
    deviceListCache = devices;
    const auto toRemove = old_set - new_set;
    const auto toAdd = new_set - old_set;
    for (auto i : toRemove)
        removeDevice(i);
    for (auto i : toAdd)
        addDevice(i);
}

bool MainWindow::saveConfig(bool checkWriteVal)
{
    bool addrOk, valOk;
    config.isCustom = ui->radioButtonCustomReg->isChecked();

    if(config.isCustom) {
        int regAddr = ui->lineEditRegAddr->text().toInt(&addrOk);
        if(!addrOk) {
            regAddr = ui->lineEditRegAddr->text().toInt(&addrOk, 16);
        }

        if(addrOk) {
            addrOk = regAddr>=0 && regAddr < std::numeric_limits<quint16>::max();
            config.regAddr = addrOk ? static_cast<quint16>(regAddr) : 0;
        }

        if(!addrOk)
            qWarning() << "Bad custom reg address:"<<ui->lineEditRegAddr->text();
    } else {
        config.regAddr = static_cast<quint16>(ui->comboBoxRegAddr->currentData().toUInt(&addrOk));
        addrOk &= config.regAddr != 0;
    }

    config.regType = static_cast<Config::REG_TYPE>(ui->comboBoxRegType->currentData().toInt());

    quint64 val = ui->lineEditRegVal->text().toULongLong(&valOk);
    if(valOk) {
        config.regWriteVal = val;
    } else {
        val = ui->lineEditRegVal->text().toULongLong(&valOk, 16);
        if(valOk) {
            config.regWriteVal = val;
            ui->lineEditRegVal->setText(QString("0x%1").arg(config.regWriteVal,0,16));
        }
        else if(checkWriteVal)
            qWarning() << "Bad writting value:"<<ui->lineEditRegVal->text();
    }

    config.saveSettings();
    return addrOk & (valOk || !checkWriteVal);
}

void MainWindow::restoreGui()
{
    if(config.isCustom)
        ui->radioButtonCustomReg->setChecked(true);
    else
        ui->radioButtonPreDefReg->setChecked(true);

    ui->lineEditRegAddr->setText(QString("0x%1").arg(config.regAddr, 0, 16));
    ui->lineEditRegVal->setText(QString("0x%1").arg(config.regWriteVal, 0, 16));
    ui->comboBoxRegType->setCurrentIndex(static_cast<int>(log2(static_cast<int>(config.regType))));
    updateLineEditWriteLen();
}

void MainWindow::updateLineEditWriteLen()
{
    bool ok;
    int len = ui->comboBoxRegType->currentData().toInt(&ok);
//    ui->lineEditRegVal->setMaxLength(2+len*4);
}
