#include "Client.h"
#include "params.h"
#include "Analizer.h"
#include "RingBuffer.h"
#include "TcpClient.h"
#include "base/threadQuit.h"
#include "base/RcRunId.h"

#include <QCoreApplication>
#include <QElapsedTimer>
#include <QThread>

namespace {
int getPreferedBufSize(quint64 inputBufSize) {
    int preferedBufSize = inputBufSize ? inputBufSize : DATA_RING_SIZE;
    if(inputBufSize/EVENT_RING_SIZE < EVENT_SIZE_MIN)
        preferedBufSize = EVENT_SIZE_MIN*EVENT_RING_SIZE;
    return preferedBufSize;
}
}
Client::Client(const EvBClientConfig &cc,
                             int _evNumIndex,
                             QObject *parent) :
    QObject(parent),
    index(cc.index),
    enabled(cc.enabled),
    evNumIndex(_evNumIndex),
    dataRingBuff( new RingBuffer<char>(getPreferedBufSize(cc.inputBufSize))),
    evRingBuff(new RingBuffer<EventMetaInfo>(EVENT_RING_SIZE)),
    tcpClientThread(new QThread(this)),
    analizerThread(new QThread(this)),
    tcpClient(new TcpClient(index, dataRingBuff)),
    analizer(new Analizer(index, dataRingBuff, evRingBuff))
  , fsm(new ClientFsm(index, this))
{
    progDescription.type = index.getProgramType().toString();
    progDescription.index = index.getProgramIndex().toString();

    qRegisterMetaType<ProgramInterface> ();
    tcpClientThread->setObjectName(QString("tcpClient #%1").arg(index.toString()));
    analizerThread->setObjectName(QString("analizer #%1").arg(index.toString()));
    tcpClient->moveToThread(tcpClientThread);
    tcpClientThread->start();

    analizerThread->start();
    analizer->moveToThread(analizerThread);

    connect(this, &Client::stopReadOut, tcpClient, &TcpClient::stopReadOut);
    connect(this, &Client::startReadOut, tcpClient, &TcpClient::startReadOut);

    connect(this, &Client::starting, analizer, &Analizer::startAnalize);
    connect(this, &Client::starting, tcpClient, &TcpClient::startRun);

    connect(this, &Client::stoping, analizer, &Analizer::stopAnalize);
    connect(this, &Client::stoping, tcpClient, &TcpClient::stopRun);

    connect(analizer, &Analizer::updateRunState, fsm, &ClientFsm::setAnalizerInRun);
    connect(tcpClient->getFsm(), &BaseFsm::stateChanged, fsm, &ClientFsm::setTcpState);

    connect(tcpClient, &TcpClient::sendStatus, this, &Client::sendStatus);
    connect(tcpClient, &TcpClient::sendStatistic, this, &Client::sendStatistic);

    connect(this, &Client::destroyingClient, analizer, &Analizer::deleteLater);
    connect(this, &Client::destroyingClient, tcpClient, &TcpClient::deleteLater);

    connect(analizer, &Analizer::destroyed, analizerThread, &QThread::quit);
    connect(tcpClient, &TcpClient::destroyed, tcpClientThread, &QThread::quit);

    connect(this, QOverload<const ProgramDescription &,
            const ProgramInterface &>::of(&Client::setTcpAddress),
            tcpClient, &TcpClient::setAddress);

    connect(this, &Client::clientDescrChanged, analizer, &Analizer::clientDescrChanged);
    connect(this, &Client::clientDescrChanged, tcpClient, &TcpClient::clientDescrChanged);

    connect(analizer, &Analizer::statUpdated, this, &Client::statUpdated);
    connect(analizer, &Analizer::statUpdated, tcpClient, &TcpClient::analyzerStatUpdated);

    connect(tcpClient, &TcpClient::inputBufferSizeUpdated, this, &Client::inputBufferSizeUpdated);

    connect(fsm, &ClientFsm::stateChanged, this, &Client::onFsmStateChanged);
    connect(this, &Client::printFsmInput, fsm, &BaseFsm::printFsmInput);
}

Client::~Client()
{
    emit destroyingClient();
    QElapsedTimer timer;
    timer.start();

    if(!threadQuiting(analizerThread, 1000, &timer))
        qWarning()<<"Analizer's thread isn't stopped for client #"<<index.toString();

    if(!threadQuiting(tcpClientThread, 1000, &timer))
        qWarning()<<"TcpClient's thread isn't stopped for client #"<<index.toString();

    delete dataRingBuff;
    delete evRingBuff;
}

void Client::setProgDescription(const ProgramDescription &descr)
{
    if(progDescription.getName() != descr.getName())
        emit clientDescrChanged(descr.getName());

    progDescription = descr;
    ProgramInterface newInterface;
    for(const ProgramInterface &interface : descr.interfaces) {
        if(interface.type == ProgramInterface::getTypeName(ProgramInterfaceDataFlow)
                && interface.id == evNumIndex){
            newInterface = interface;
            break;
        }
    }
//    qDebug() << QString("[ClientMgr #%1] connect to interface '%2' (id:%3)").arg(index)
//                .arg(newInterface.type).arg(newInterface.id);
    if(!newInterface.host.isNull() && newInterface.port)
        emit setTcpAddress(descr, newInterface);
    tcpInterface = newInterface;
}

void Client::startRun()
{
    if(!enabled)
        return;
    fsm->goState(fsm->getState() == FsmState::Ready ? FsmState::Starting
                                                    : FsmState::Error);
}

void Client::stopRun()
{
    if(!enabled)
        return;
    if(fsm->getState() != FsmState::Run) {
        fsm->goState(FsmState::Error);
        return;
    }
    emit stoping();
    fsm->goState(FsmState::Ready);
}

void Client::destroyClient(ClientIndex i)
{
    if(!i.isEmpty() && index!=i)
        return;
    deleteLater();
}

void Client::updateClientReadout(ClientIndex i, bool readout)
{
    if(!i.isEmpty() && index!=i)
        return;
    enabled = readout;
    if(enabled){
        emit startReadOut();
    } else
        emit stopReadOut();
}

void Client::onFsmStateChanged(const FsmState &state)
{
    emit clientStateUpdated(index, state);

    switch (state) {
    case FsmState::Error :
        emit stoping();
        break;
    case FsmState::Starting :
    {
        qDebug() << QString("ClientManager for #%1 start resetting")
                    .arg(getName());

        emit starting();
        fsm->goState(FsmState::Run);
        break;
    }
    default:{}
    }
}
