#include "Combiner.h"

#include "params.h"
#include "Analizer.h"
#include "EvbSignalRouter.h"
#include "FileWriter.h"
#include "OutputTcpServer.h"
#include "RingBuffer.h"
#include "TcpClient.h"
#include "base/threadQuit.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QElapsedTimer>
#include <QSemaphore>
#include <QSharedPointer>
#include <QThread>
#include <QTimer>
#include "CombinerFsm.h"

namespace {
    const int OUTPUT_TCP_PORT = 0; // 33399 or 0 for auto-detect
    const quint64 LOOP_TIMEOUT = 100;

    const int START_STOP_TIMEOUT_MS = 1000;
    const int SUPPRESS_LOG_LAGGING_LIMIT = 10;
    const int prevEventIncompletLIMIT = 10;

    struct TLV_Hdr {
        TLV_Hdr(quint32 _type, quint32 _len=0):
            type(_type),
            len(_len)
        {}
        int totSize() const { return len+2*sizeof(quint32); }
        virtual QByteArray getData() {
            QByteArray a;
            a.append(reinterpret_cast<const char*>(&type), sizeof(type));
            a.append(reinterpret_cast<const char*>(&len), sizeof(len));
            return a;
        }
        quint32 type;
        quint32 len;
    };

    struct RunNumberRecord : public TLV_Hdr {
        RunNumberRecord(quint32 runNum):
            TLV_Hdr(0x236E7552, sizeof(quint32)),
            runNumber(runNum)
        {}
        virtual QByteArray getData() {
            QByteArray a = TLV_Hdr::getData();
            a.append(reinterpret_cast<const char*>(&runNumber), sizeof(runNumber));
            return a;
        }
        quint32 runNumber;
    };

    struct RunIndexRecord : public TLV_Hdr {
        RunIndexRecord(QString index):
            TLV_Hdr(0x78646E49),
            runIndexLatin1(index.toLatin1())
        {
            // complete 32-bit format
            while(runIndexLatin1.size()%sizeof(quint32)){
                runIndexLatin1.append((char)0);
            }
            len += runIndexLatin1.size();
        }
        virtual QByteArray getData() {
            QByteArray a = TLV_Hdr::getData();
            a.append(runIndexLatin1);
            return a;
        }

        QByteArray runIndexLatin1;
    };

    struct SartStopRecord : public TLV_Hdr {
        SartStopRecord(bool start, quint32 runNumber, QString runIndex):
            TLV_Hdr(start ? MPD_SYNC_RUN_START : MPD_SYNC_RUN_STOP),
            runNumberRecord(runNumber),
            runIndexRecord(runIndex)
        {
            len += runNumberRecord.totSize();
            len += runIndexRecord.totSize();
        }

        QByteArray getData() {
            QByteArray a = TLV_Hdr::getData();
            a.append(runNumberRecord.getData());
            a.append(runIndexRecord.getData());
            return a;
        }

        RunNumberRecord runNumberRecord;
        RunIndexRecord runIndexRecord;
    };
}

Q_DECLARE_METATYPE(QVector<Analizer*>)

Combiner::Combiner() :
    QObject(0),
    fsm(new CombinerFsm(this)),
    initFlag(true),
    enabled(false),
    eventStarted(false),
    clientNumbers(0),
    startEvNumber(0),
    evNumberStep(1),
    expectedEvNum(startEvNumber),
    workLoopTimer(new QTimer(this)),
    statusTimer(new QTimer(this)),
    outputBufferSize(0),
    writerRingBuf(new RingBuffer<char>(outputBufferSize)),
    fileWriter(new FileWriter(writerRingBuf)),
    fileWriterThread(new QThread(this)),
    outToFile(true),
    outToTcpServer(false),
    noFreeSpaceError(false),
    eventsWithDebug(0),
    finishId(1),
    prevEventIncomplet(0),
    suppressLogLagging(0),
    timeDistrEn(false),
    waitingResize(false),
    preferedBufSize(0)
{
    qRegisterMetaType<CombinerStats>();
    qRegisterMetaType<COMBINER_OUT>();
    qRegisterMetaType<OutputTcpStatus>();
    qRegisterMetaType< QVector<Analizer*> >();
    qRegisterMetaType< QMap<int, QString> >();
    connect(workLoopTimer, &QTimer::timeout, this, &Combiner::workLoop);
    workLoopTimer->setSingleShot(true);
    workLoopTimer->setInterval(0); // it's ok, work can't completed faster then workTimeoutMs

    fileWriterThread->setObjectName("fileWriterThread");
    fileWriter->moveToThread(fileWriterThread);
    fileWriterThread->start();
    connect(fileWriter, &FileWriter::destroyed, fileWriterThread, &QThread::quit);
    connect(fileWriter, &FileWriter::statusUpdated, this, &Combiner::fileWriterStatsUpdated);
    connect(fileWriter, &FileWriter::fileOpened, this, &Combiner::writerFileOpened);
    connect(fileWriter, &FileWriter::error, this, &Combiner::outputError);
    connect(this, &Combiner::clearTcp, fileWriter, &FileWriter::clear);
    connect(this, &Combiner::deleteFileWriter, fileWriter, &FileWriter::deleteLater);
    connect(this, &Combiner::storageChanged, fileWriter, &FileWriter::storageChanged);
    connect(this, &Combiner::startingFile, fileWriter, &FileWriter::openFileToWrite);
    connect(this, &Combiner::closeFile, fileWriter, &FileWriter::closeFile);

    connect(fsm, &BaseFsm::stateChanged, this, &Combiner::onFsmStateChanged);
    connect(fsm, &BaseFsm::stateChanged, this, &Combiner::fsmStateChanged);
    connect(fileWriter, &FileWriter::fileClosed, fsm, &CombinerFsm::setFileClosed);

    connect(statusTimer, &QTimer::timeout, this, &Combiner::sendStatus);
    statusTimer->setSingleShot(false);
    statusTimer->setInterval(1000);

    auto evbSignalRouter = &EvbSignalRouter::instance();
    evbSignalRouter->setCombiner(this);
    evbSignalRouter->setCombinerFsm(fsm);

    QTimer::singleShot(0, this, SLOT(init()));
}

Combiner::~Combiner()
{
    if(enabled)
        stop();

    QElapsedTimer timer;
    timer.start();
    emit deleteFileWriter();
    emit deleteTcp(0);

    bool ok=true;
    for (int index : outputTcpServers.keys()) {
        if(!threadQuiting(outputTcpServerThreads[index], 1000, &timer)){
            qWarning() << QString("Failed to stop output TcpServer id=%1").arg(index);
            ok=false;
        }
    }
    if(!threadQuiting(fileWriterThread, 1000, &timer)) {
        qWarning() << "Failed to stop FileWriter thread";
        ok=false;
    }
    if(ok)
        qDebug() << "All parts of Combiner were stoped.";
    else
        qWarning() << "Some output TcpServers failed to stop.";
    delete writerRingBuf;
}

void Combiner::initOutputBuffers(int size, bool outToTcp, int outTcpNum)
{
    if(size<0)
        size=0;
    preferedBufSize = size ? size : FILE_WRITER_BUFFER_SIZE;
}

void Combiner::onFsmStateChanged(const FsmState &cur, const FsmState &prev)
{
    switch (cur) {
    case FsmState::Starting:
    {
        onStartingAction();
        break;
    }
    case FsmState::Run:
    {
        writeStartStopRecrd(true);
        stats.newPeriod();
        break;
    }
    case FsmState::PrepareStopping:
    {
        break;
    }
    case FsmState::Stopping:
    {
        onStoppingAction();
        break;
    }
    case FsmState::Error:
    {
        if(prev == FsmState::Stopping && outToFile) {
            QString errStr = QString("Can't wait end of work of file writer; there are some data pending (%1 bytes)")
                    .arg(writerRingBuf->getContinuosDataSize(0));
            qWarning() << errStr;
            emit error(errStr);
        }
        if(prev == FsmState::Run)
            onStoppingAction();
        break;
    }
    default:;
    }
}

void Combiner::start_run(const RcRunId &newRunId)
{
    runId = newRunId;
    qDebug() << QString("Combiner starting run %1").arg(runId.toString());
    fsm->goState(FsmState::Starting);
}

void Combiner::init()
{
    stats.newPeriod();
    int monitorId = addTcpServer();
    Q_ASSERT(monitorId==0);
    monitorTcpServer = outputTcpServers.first();
    resizeBuffer();
    emit startTcpServer(monitorId, true);
    // tcpStatusUpdated from monitor TCP will invoke this method
//    sendStatus();
}

void Combiner::onStartingAction()
{
    reset();

    if(outToTcpServer) {
        emit startingTcpWork();
    } else if(outToFile) {
        emit startingFile(runId, static_cast<int>(startEvNumber%evNumberStep));
    }

    enabled = true;
    workLoopTimer->start();
    qDebug() << "Combiner is started.";

    statusTimer->start();
    sendStatus();
}

void Combiner::prepareForStop(quint32 stopEvNum)
{
    fsm->setStopEvNum(stopEvNum);
    fsm->goState(FsmState::PrepareStopping);
    checkStopEvNum();
}

void Combiner::stop()
{
    fsm->goState(FsmState::Stopping);
}

void Combiner::reset()
{
    prevEventIncomplet=0;
    events.clear();
//    events.resize(clientNumbers);
    emit clearTcp();

#ifdef RESET_ROUGH
    // All work with RingBuffers performed in Analizer and TcpClient
#else
    for (RingBuffer<EventMetaInfo> *rb : eventRingBuffers) {
        int eventsPending = rb->getDataSize();
        rb->trashData(eventsPending);
    }
#endif
    for(ClientIndex &index : eventRingBuffersDataSpace.keys()){
        eventRingBuffersDataSpace[index] = 0;
    }

    initFlag=true;
    Q_ASSERT(!enabled); // enabled = false;
    eventStarted=false;
    expectedEvNum=startEvNumber;
    stats.clear();
    lastMissingEvFromClient.clear();

    sendStatus();
}


void Combiner::destroyClient(const ClientIndex &clientId)
{
    if(!clientId.isEmpty()) {
        events.remove(clientId);
        eventRingBuffers.remove(clientId);
        eventRingBuffersDataSpace.remove(clientId);
        clientIds.remove(clientId);
    }
}

void Combiner::setOutput(COMBINER_OUT combinerOut, int tcpNum)
{
    fsm->setOutputList(combinerOut, tcpNum);

    outToFile = combinerOut==COMBINER_FILE;
    outToTcpServer = combinerOut==COMBINER_TCP;

    if(combinerOut != COMBINER_TCP) tcpNum = 0;

    for(int i=outputTcpServers.size(); i<tcpNum+1; ++i) {
        int id = addTcpServer();
        emit startTcpServer(id, true);
        qInfo() << QString("Add out TCP id:%1").arg(id);
    }
    for(int i=outputTcpServers.size()-1; i>tcpNum; --i ) {
        qInfo() << QString("Remove out TCP id:%1").arg(i);
        emit deleteTcp(i);
        if(!threadQuiting(outputTcpServerThreads[i], 1000)){
            qWarning() << QString("failed to stop output TcpServer id=%1").arg(i);
            outputTcpServerThreads[i]->terminate();
        }
        stats.tcpStatus.remove(i);
        outputTcpServers.remove(i);
        outputTcpServerThreads.remove(i);
    }

    resizeBuffer();
    sendStatus();
}

void Combiner::workLoop()
{
    if(eventsWithDebug)
        qInfo()<<"[Combiner] start of workLoop";

    if(waitingResize)
        resizeBuffer();

    if(!enabled)
        return;

    QElapsedTimer eventMinTime, loopTimer, workTimer;
    const qint64 workTimeoutMs = 1000;
    const qint64 loopTimeoutMs = 100;
    qint64 leftTimeMs;
    bool loopExpired, eventExpired;
    int gotClients;

    workTimer.start();
    if(initFlag){
        processInit();
        if(initFlag) {
            // process other pending signals and try again
            workLoopTimer->start();
            return;
        }
    }
    if(stats.isNextPeriod(workTimer)) {
        int newEvSize = stats.periodSize/stats.periodCnt;
        newEvSize = qMin(newEvSize, EVENT_SIZE_MAX);
        newEvSize = qMax(newEvSize, EVENT_SIZE_MIN);
        int newBufSize = newEvSize * EVENT_RING_SIZE;
        if(waitingResize) {
            preferedBufSize = newBufSize;
        } else {
            const int curBufSize = preferedBufSize;
            double factor = 1.*newBufSize/curBufSize;
            if(factor<0.5 || factor>1.1) {
                waitingResize = true;
                preferedBufSize = newBufSize;
                qInfo() << QString("WaitingResize up to %1MB")
                           .arg(1e-6*preferedBufSize);
            }
        }
        if(waitingResize) {
            emit outputBufferSizeUpdated(preferedBufSize);
            resizeBuffer();
        }

        stats.newPeriod();
    }

    while(!initFlag && !workTimer.hasExpired(workTimeoutMs)){
        if(eventsWithDebug)
            qInfo()<<"[Combiner] while loop";

        eventMinTime.invalidate();
        loopTimer.start();
        loopExpired = false;
        eventExpired = false;

        //quick check eventMinTime
        for(auto id : clientIds) {
            while(updateClientsEventInfo(id)) {
                EventMetaInfo* curEv = events[id];
                if(curEv->evNum<expectedEvNum) {
                    if(eventsWithDebug) {
                        qInfo() << "trash with lower evNumber=" << curEv->evNum
                                << " expectedEvNum:" << expectedEvNum
                                << "clientId:" << id;
                    }
                    if(suppressLogLagging < SUPPRESS_LOG_LAGGING_LIMIT) {
                        qWarning()<<QString("1.Trash lagging event evNum(%1)<expected(%2) from %3")
                                    .arg(curEv->evNum).arg(expectedEvNum).arg(id.toString());
                        if(++suppressLogLagging == SUPPRESS_LOG_LAGGING_LIMIT)
                            qWarning() << "suppressLogLagging";
                    }
                    events[id] = nullptr;
                    eventRingBuffers[id]->trashData();
                    --eventRingBuffersDataSpace[id];
                    stats.trashedEvents[id]++;
                    continue;
                }
                if(!eventMinTime.isValid() || curEv->createTime.msecsTo(eventMinTime)>0){
                    eventMinTime = curEv->createTime;
                    if(eventsWithDebug)
                        qInfo()<<"The earliest ev from "<<id.toString();
                }
                break;
            }
        }

        // At this step eventMinTime (if valid) contains min time of event which evNum >= expectedEvNum
        // Timing loop
        gotClients=0;
        EventMetaInfo* curEv;
        for(auto id : clientIds) {
            leftTimeMs=0;
            while(true) {
                curEv = nullptr;
                if(updateClientsEventInfo(id, leftTimeMs)){
                    curEv = events[id];
                } else {
                    if(leftTimeMs) {
                        loopExpired = true;
                        break; // try to get event from following client
                    }
                }

                if(curEv) {
                    if(curEv->evNum < expectedEvNum) {
                        addTrashStat(id, curEv->evNum);
                        events[id] = nullptr;
                        eventRingBuffers[id]->trashData();
                        --eventRingBuffersDataSpace[id];
                        stats.trashedEvents[id]++;
                        // try to receive correct ev in loop timeout
                        continue;
                    }

                    // curEv->evNum >= expectedEvNum
                    if(curEv->evNum == expectedEvNum) {
                        ++gotClients;
                    } else if(eventsWithDebug)
                        qInfo()<<QString("Got event with event numer (%1) greater that expected evNum (%2) from %3")
                              .arg(curEv->evNum).arg(expectedEvNum).arg(id.toString());

                    if(!eventMinTime.isValid() || curEv->createTime.msecsTo(eventMinTime)>0) {
                        eventMinTime = curEv->createTime;
                    }
                    break;
                }

                if(loopExpired) {
                    leftTimeMs = 0;
                } else {
                    qint64 timeToLoopExpire = loopTimeoutMs-loopTimer.elapsed();
                    if(eventMinTime.isValid()) {
                        qint64 timeToEvExpire = EVENT_EXPIRE_TIMEOUT_MS-eventMinTime.elapsed();
                        if(timeToEvExpire < 0)
                            eventExpired = true;
                        leftTimeMs = qMin(timeToEvExpire, timeToLoopExpire);
                    } else
                        leftTimeMs = timeToLoopExpire;
//                    qInfo()<<leftTimeMs<<timeToEvExpire<<timeToLoopExpire;
                    if(leftTimeMs < 0){
                        loopExpired = true;
                        leftTimeMs = 0;
                    }
                }
                if(leftTimeMs==0)
                    break;
            }
        }

        // At this moment we either collected all clients with proper evNum,
        // either workLoop timeout expired
        // or some client expired (eventMinTime)
        if(gotClients) {
            if(gotClients==clientIds.size() ||
               eventExpired                 ||
               eventMinTime.hasExpired(EVENT_EXPIRE_TIMEOUT_MS)) {
                if(gotClients==clientIds.size()) {
                    if(prevEventIncomplet) {
//                        prevEventIncomplet=false;
                        prevEventIncomplet=0;
                        qWarning() << "prevEventIncomplet suppresed canceled; curEvNum="<<expectedEvNum;
//                        qInfo() << "Resume to complete event; curEvNum="<<expectedEvNum;
                    }
                } else {
//                    if(!prevEventIncomplet) {
                    if(prevEventIncomplet < prevEventIncompletLIMIT) {
//                        EventMetaInfo* curEv;
                        QString msg = QString("Send incomplete event; curEvNum=%1; missing:").arg(expectedEvNum);
                        for(auto id : clientIds) {
                            curEv = events[id];
                            if(curEv==nullptr || curEv->evNum!=expectedEvNum) {
                                msg.append(id.toString()).append(',');
                            }
                        }
                        qWarning() << msg;
//                        prevEventIncomplet=true;
                        if(++prevEventIncomplet == prevEventIncompletLIMIT)
                            qWarning() << "prevEventIncomplet suppresed";
                    }
                }

                ++stats.clientNumDistr[gotClients];
                flushEvent();
            } else {
                // try to complete event in next work loop
            }
//            eventMinTime.invalidate();
        } else if(eventMinTime.isValid() && eventMinTime.hasExpired(EVENT_EXPIRE_TIMEOUT_MS)) {
            //todo is this a error?
            qWarning() << QString("Got expired event with EvNum greater than expectedEvNum %1; start initializing...")
                          .arg(expectedEvNum);
            initFlag=true;
//            eventsWithDebug=50;
//            qInfo()<<"\tInit eventsWithDebug";
        }
    }

    workLoopTimer->start();
}

void Combiner::sendStatus()
{
//    QString mess;
//    if(!enabled)
//        mess.append("stopped; ");
//    QString clientsStr;
//    for (int i : stats.lastEventClients) {
//        clientsStr.append(QString("%1 ").arg(i));
//    }
//    mess.append(QString("reg:%1;skip:%2;eob:%3;clients:%4;nextEv:%5")
//                .arg(stats.regularEvents)
//                .arg(stats.skipped)
//                .arg(stats.eob)
//                .arg(clientsStr)
//                .arg(expectedEvNum));
//    for(int i=0;i<events.size(); ++i){
//        const EventMetaInfo* ev = events[i];
//        if(ev == nullptr) continue;
//        mess.append(QString("; #%1=%2").arg(i).arg(events[i]->evNum));
//    }
//    emit status(mess);
    stats.expectedEvNum = expectedEvNum;
    for(auto i : clientIds) {
        EventMetaInfo * curEv = events[i];
        if(curEv && curEv->type==MPD_SYNC_NORM)
            stats.hasPendingEvent = true;
        stats.freeEvPart[i] = eventRingBuffers[i]->getFreeSize();
        stats.preparedEvPart[i] = eventRingBuffers[i]->getDataSize();
        stats.flushedEvPart[i] = eventRingBuffers[i]->getTrashSize();
    }
    if(outToTcpServer) {
        stats.fwFullSize = 0;
        stats.fwDataPart = 0;
        for (const OutputTcpServer *tcpServer : outputTcpServers.values()) {
            if(tcpServer->getIndex()==0)
                continue;
            stats.fwFullSize += tcpServer->getRingBuffer()->getAllSize();
            stats.fwDataPart += tcpServer->getRingBuffer()->getDataSize();
        }
        if(stats.fwFullSize)
            stats.fwDataPart /= stats.fwFullSize;
    } else if(outToFile) {
        stats.fwDataPart = 1.*writerRingBuf->getDataSize()/writerRingBuf->getAllSize();
        stats.fwFullSize = writerRingBuf->getAllSize();
    } else {
        stats.fwFullSize = 0;
        stats.fwDataPart = 0;
    }

    stats.fwPrefSize = preferedBufSize;
    stats.timeStamp.start();
    emit statusRaw(stats);

    for (ClientIndex index : trashStat.keys()) {
        const ClientTrashStat &ts = trashStat[index];
        for (quint64 _expectedEvNum : ts.keys()) {
            const EventsRange &range = ts[_expectedEvNum];
            if(range.first == range.second)
                qWarning() << QString("Lagging event evNum(%1) was trashed from %2; expected(%3)")
                            .arg(range.first).arg(index.toString()).arg(_expectedEvNum);
            else
                qWarning() << QString("Lagging events evNum(%1-%2) were trashed from %3; expected(%4)")
                            .arg(range.first).arg(range.second).arg(index.toString()).arg(_expectedEvNum);
        }
    }
    trashStat.clear();
}


void Combiner::tcpStatusUpdated(const OutputTcpStatus &s)
{
    bool immediateSend = true;
    if(stats.tcpStatus.contains(s.index)) {
        OutputTcpStatus &prevS = stats.tcpStatus[s.index];
        immediateSend = s.listening!=prevS.listening
                || s.port!=prevS.port || s.isFree!=prevS.isFree;
    }

    stats.tcpStatus[s.index] = s;
    fsm->setOutputTcpConnected(s.index, !s.isFree);
    fsm->setOutputTcpDone(s.index, s.isEmpty);
    if(immediateSend)
        sendStatus();
}

void Combiner::tcpFinished(int tcpId, uint fId) const
{
    Q_UNUSED(fId)
    fsm->setOutputTcpDone(tcpId, true);
}

void Combiner::writerFileOpened(const RcRunId &runId) const
{
    Q_UNUSED(runId)
    fsm->setFileCreated(true);
}

void Combiner::outputError(int index, const QString &errStr) const
{
    if(index == -1){
        // FileWriter
        qWarning() << "Got error in FileWriter:" << errStr;
    } else if(index){
        // Tcp Data Output
        qWarning() << QString("Got error in Tcp Server #%1 for data output:")
                      .arg(index) << errStr;
    } else {
        // Monitor
        qWarning() << "Got error in Tcp Server for monitor data" << errStr;
    }
    // emit error for all exept monitor
    if(index)
        emit error(errStr);
}

void Combiner::enableTimeDistr(bool en)
{
    timeDistrEn = en;
}

void Combiner::processInit()
{
    QElapsedTimer eventMinTimer;
    quint32 minNumber=startEvNumber;
    qint64 timeoutMs = LOOP_TIMEOUT/2;
    qint64 loopTimeout;
    eventMinTimer.invalidate();

    for(auto i : clientIds) {
        loopTimeout = (i).isEmpty() ? 0 : timeoutMs;
        if(updateClientsEventInfo((i), loopTimeout)) {
            EventMetaInfo* curEv = events[(i)];

            eventMinTimer = curEv->createTime;
            minNumber = curEv->evNum;
            break;
        }
    }

    if(!eventMinTimer.isValid()){
        // There is no any pending event and updateClientsEventInfo timed out
        return;
    }

    for(auto i : clientIds) {
        loopTimeout = timeoutMs-eventMinTimer.elapsed();
        loopTimeout = qMax((qint64)0, loopTimeout);
        if(updateClientsEventInfo((i), loopTimeout)){
            const EventMetaInfo* curEv = events[(i)];
            if(curEv->createTime.msecsTo(eventMinTimer)>0)
                eventMinTimer = curEv->createTime;
            if(minNumber > curEv->evNum)
                minNumber = curEv->evNum;
        }
    }

    if(eventMinTimer.isValid()){
        initFlag = false;
        expectedEvNum = minNumber;
        qInfo() << QString("Initialied: minNumber=%1").arg(expectedEvNum);
    }
}

void Combiner::flushEvent(const ClientIndex &singleIndex)
{
    mpdEventHdr.evNum = expectedEvNum;
    mpdEventHdr.length = 0;
    QMap<ClientIndex, qint64> timeDistr;
    qint64 maxTime=0, dT;
    bool completEvent = true;

    // 1st step: Calculate total event size
    if(singleIndex.isEmpty()) {
        QSharedPointer<QElapsedTimer> curElTimer;
        for(auto i : clientIds) {
            const EventMetaInfo* ev = events[i];
            if(ev == nullptr) {
                completEvent = false;
                if(!lastMissingEvFromClient.contains(i) || expectedEvNum-lastMissingEvFromClient[i]>1000){
//                    qWarning()<<"There is no event from "<<getClientNameByIndex(index)<<" evNum="<<expectedEvNum;
                    lastMissingEvFromClient[i] = expectedEvNum;
                }
            } else if(ev->evNum != expectedEvNum) {
                completEvent = false;
                qWarning()<<QString("%1 has event with wrong evNum(%2) expectedEvNum(%3)")
                            .arg(i.toString()).arg(ev->evNum).arg(expectedEvNum);
            } else {
                mpdEventHdr.length += ev->size1 + ev->size2-sizeof(mpdEventHdr);
                if(timeDistrEn && completEvent) {
                    if(curElTimer.isNull()){
                        curElTimer = QSharedPointer<QElapsedTimer>(new QElapsedTimer());
                        curElTimer->start();
                    }

                    dT = ev->createTime.msecsTo(*curElTimer);
                    maxTime = qMax(maxTime, dT);
                    timeDistr[i] = dT;
                }
            }
        }
    } else {
        const EventMetaInfo* ev = events[singleIndex];
        if(ev != nullptr)
            mpdEventHdr.length += ev->size1 + ev->size2-sizeof(mpdEventHdr);
    }

    if(mpdEventHdr.length != 0) {
        // 2nd step: Deside to what RingBuffers we should write a data
        const int allSize = mpdEventHdr.length+sizeof(MpdEventHdr);
        RingBuffer<char> *primaryBuffer = nullptr;
        if(outToFile)
            primaryBuffer = writerRingBuf;
        else if(outToTcpServer)
            primaryBuffer = outputTcpServers[expectedEvNum%(outputTcpServers.size()-1)+1]->getRingBuffer();
        RingBuffer<char> *monitorBuffer = outputTcpServers[0]->getRingBuffer();
        const bool monitorSend = !stats.tcpStatus[0].isFree
                && (monitorBuffer->getFreeSize() >= allSize);

        if(primaryBuffer && !primaryBuffer->waitFreeSpace(allSize, EVENT_EXPIRE_TIMEOUT_MS)) {
            if(!noFreeSpaceError)
                qCritical() << "Can't write event, there is no free space in buffer";
            noFreeSpaceError = true;
            return;
        }
        noFreeSpaceError = false;
        // 3rd step: write a data
        if(singleIndex.isEmpty()) {
//        QString mess = QString("write ev:%1;").arg(mpdDataHdr.evNum);
            // 3-1 step: write ev header
            if(primaryBuffer)
                  writeData(primaryBuffer, (char*)&mpdEventHdr, sizeof(MpdEventHdr));
            if(monitorSend)
                writeData(monitorBuffer, (char*)&mpdEventHdr, sizeof(MpdEventHdr));

            // 3-2 step: write ev device blocks
            stats.lastEventClients.clear();
            bool hasMissingClient=false;
            for(auto i : clientIds) {
                if(completEvent) {
                    const qint64 dt = maxTime-timeDistr[i];
                    stats.timeDistr[i] += dt;
                    stats.timeDistrFull[i] += timeDistr[i];
                    stats.addEvent(allSize);
                }
                const EventMetaInfo *ev = events[i];
                if(ev == nullptr || ev->evNum != expectedEvNum){
                    hasMissingClient=true;
                    continue;
                }
                Q_ASSERT(!stats.lastEventClients.contains(i));
                stats.lastEventClients.insert(i);
                //            mess.append(QString("%1 ").arg(clientId));

                stats.flushedSize[i] += ev->size1+ev->size2;
                int headerSize1 = qMin(ev->size1, (int)sizeof(MpdEventHdr));
                if(ev->size1-headerSize1>0) {
                    if(primaryBuffer)
                        writeData(primaryBuffer, ev->data1+headerSize1, ev->size1-headerSize1);
                    if(monitorSend)
                        writeData(monitorBuffer, ev->data1+headerSize1, ev->size1-headerSize1);
                } else
                    Q_ASSERT(ev->size1-headerSize1==0);

                if(ev->size2) {
                    int headerSize2 = sizeof(MpdEventHdr)-headerSize1;
                    Q_ASSERT(headerSize2>=0 && headerSize2<sizeof(MpdEventHdr));
                    if(primaryBuffer)
                        writeData(primaryBuffer, ev->data2+headerSize2, ev->size2-headerSize2);
                    if(monitorSend)
                        writeData(monitorBuffer, ev->data2+headerSize2, ev->size2-headerSize2);
                }

                RingBuffer<EventMetaInfo> *rb = eventRingBuffers[i];
                //            qDebug()<<"trash flushed event="<<ev->evNum;
                //            qDebug()<<"Combiner: anal="<<clientIds[index]<<" trashData "
                //                   << "event pos="<<rb->getReaderPos()
                //                   << " evNum="<<rb->getReadPointer()->evNum
                //                   << ev->evNum;

                rb->trashData();
                --eventRingBuffersDataSpace[i];
                events[i] = nullptr;
            }
            if(completEvent)
                ++stats.timeDistrCnt;
            //        qDebug()<<mess;
            ++stats.regularEvents;
            if(hasMissingClient)
                ++stats.partedEvents;
            else {
                ++stats.fullEvents;
                if(suppressLogLagging) {
                    suppressLogLagging = 0;
                    qWarning() << "suppressLogLagging canceled";
                }
            }
        } else {
            // 3-1 step: write full ev block for single client (EOB case)
            const EventMetaInfo *ev = events[singleIndex];
            stats.flushedSize[singleIndex] += ev->size1+ev->size2;

            if(primaryBuffer)
                writeData(primaryBuffer, ev->data1, ev->size1);
            if(monitorSend)
                writeData(monitorBuffer, ev->data1, ev->size1);
            if(ev->size2) {
                if(primaryBuffer)
                    writeData(primaryBuffer, ev->data2, ev->size2);
                if(monitorSend)
                    writeData(monitorBuffer, ev->data2, ev->size2);
            }
            if(ev->type == MPD_SYNC_EOB)
                ++stats.eob;
            events[singleIndex] = nullptr;
            eventRingBuffers[singleIndex]->trashData();
            --eventRingBuffersDataSpace[singleIndex];
//            qInfo() << "EOB: curEvNum="<<expectedEvNum;
        }
        if(monitorSend){
            ++stats.monitorEvCount;
//            sendLog(QString("send evNum=%1").arg(mpdDataHdr.evNum));
        }
    } else {
        qWarning() << QString("Missed ev:").append(expectedEvNum);
        ++stats.skipped;
    }
    if(eventsWithDebug) {
        qInfo()<<"flushed:"<<expectedEvNum;
        --eventsWithDebug;
    }
    if(singleIndex.isEmpty())
        expectedEvNum += evNumberStep;
    checkStopEvNum();
//    if(expectedEvNum >= 0xFFFFFFu){
//        expectedEvNum %= evNumberStep;
//    }
}

void Combiner::flushSingleEvent(const ClientIndex &index)
{
    const EventMetaInfo *ev = events[index];
//    const int &clientIndex = clientIds[index];
    Q_ASSERT(ev);
    Q_ASSERT(ev->size1);

//    mpdDataHdr.sync = ev->type;
//    mpdDataHdr.evNum = ev->evNum;
//    mpdDataHdr.length = ev->size1 + ev->size2;

//    QString mess = QString("write single ev:%1; sync:0x%2 index=%3")
//            .arg(ev->evNum)
//            .arg(ev->type, 8, 16, QChar('0'))
//            .arg(clientIndex);
//    qDebug()<<mess;
    if(ev->type == MPD_SYNC_EOB)
        ++stats.eob;
    stats.flushedSize[index] += ev->size1+ev->size2;

//    sendData((char*)&mpdDataHdr, sizeof(MpdEventHdr));
    sendData(ev->data1, ev->size1);
    if(ev->size2)
        sendData(ev->data2, ev->size2);

//    qDebug() << "trash flushed event="<<ev->evNum;
//    qDebug() << "Combiner: anal="<<clientIndex<<" trashData "
//             << "event pos="<<eventRingBuffers[index]->getReaderPos(0)
//             << " evNum="<<eventRingBuffers[index]->getReadPointer()->evNum
//             << ev->evNum;
    events[index] = nullptr;
    eventRingBuffers[index]->trashData();
    --eventRingBuffersDataSpace[index];
//    mpdDataHdr.sync = MPD_SYNC_NORM;
}

void Combiner::sendData(char *src, int size, bool addToOutEvent)
{
}

void Combiner::writeData(RingBuffer<char> *buffer, char *src, int size)
{
    if(buffer->getFreeSize()<size)
        qWarning() << "Start Combiner::writeData without enough space";
    Q_ASSERT_X(buffer->getFreeSize()>=size,
               "writeData",
               QThread::currentThread()->objectName().toStdString().c_str());
    const int part1=qMin(size, buffer->getContinuosFreeSize());
    memcpy(buffer->getWritePointer(), src, part1);
    buffer->writeData(part1);
    const int part2=size-part1;
    if(part2){
        memcpy(buffer->getWritePointer(), src+part1, part2);
        buffer->writeData(part2);
    }
}

bool Combiner::updateClientsEventInfo(const ClientIndex &index, int timeout)
{
//    Q_ASSERT( eventRingBuffers.size()>index
//              && eventRingBuffers[index] != nullptr
//            && eventRingBuffersDataSpace.size()>index);
    EventMetaInfo * curEv = events[index];
    if(curEv != nullptr)
        return true;
    RingBuffer<EventMetaInfo> * rb = eventRingBuffers[index];
    int &rdDataSpace = eventRingBuffersDataSpace[index];

    bool ret = false;
    int loopTimeout=0;

    // 1st loop with 0 timeout; following with incoming 'timeout'
    while(true){
        if(rdDataSpace < 1){
            // update evMetaInfo cache
            if(rb->waitDataSpace(1, loopTimeout))
                rdDataSpace = rb->getDataSize();
        }
        if(rdDataSpace >= 1){
            curEv = rb->getReadPointer();
            events[index] = curEv;
            stats.dataEvents[index]++;
            if(curEv->type != MPD_SYNC_NORM){
                flushSingleEvent(index);
                continue;
            }
            curEv->evNum &= 0xFFFFFFu; // trunk to single mask
            // Restore real evNum with expectedEvNum
//            if((curEv->evNum & 0xF00000u)==0 && expectedEvNum) {
//                // if ev in the begining
//                curEv->evNum += expectedEvNum & 0xFF000000u;
//            } else {
            quint32 shiftEvNum = expectedEvNum & 0xFF000000u;
            const quint32 lowerEvNumLimit = 0xFFFFFFu;
            const quint32 upperEvNumLimit = 0xFFFFFFu;
            if((expectedEvNum&0xFFFFFFu) < lowerEvNumLimit && (curEv->evNum > upperEvNumLimit)) {
                    if(shiftEvNum) {
                        shiftEvNum -= 0x1000000u;
                        qInfo() << "shiftEvNum -= 0x1000000u;";
                    } else
                        qWarning() << "Wrong work with ev number compensation";
            } else if((expectedEvNum&0xFFFFFFu) > upperEvNumLimit && (curEv->evNum < lowerEvNumLimit)) {
                shiftEvNum += 0x1000000u;
                qInfo() << "shiftEvNum += 0x1000000u;";
            }

            curEv->evNum += shiftEvNum;
//            }
            if(evNumberStep > 1) {
                // EvB works in parallel
                if(curEv->evNum % evNumberStep != expectedEvNum%evNumberStep){
                    qWarning()<<QString("Flush event which evNum(%1) % evNumStep(%2) != EvB_index(%3)")
                                .arg(curEv->evNum).arg(evNumberStep).arg(expectedEvNum%evNumberStep);
                    flushEvent(index);
                    continue;
                }
            }
            stats.lastEventNum[index] = curEv->evNum;
            if(eventsWithDebug)
                qInfo() << "updateClientsEventInfo " << index.toString()
                        << " evNumber=" << curEv->evNum;

            ret = true;
            break;
        }

        if(loopTimeout){
            if(updateClientsTimer.hasExpired(timeout))
                break;
        } else {
            if(timeout) {
                loopTimeout = timeout;
                updateClientsTimer.start();
            } else
                break;
        }
    }

    return ret;
}

int Combiner::addTcpServer()
{
    const int id = outputTcpServers.size();
    OutputTcpServer *server = new OutputTcpServer(id);
    QThread *thread = new QThread(this);
    thread->setObjectName(id ? QString("outTcpServers #%1").arg(id) : "monitor Tcp");
    server->moveToThread(thread);
    thread->start();
    connect(this, &Combiner::startTcpServer, server, &OutputTcpServer::startListen);
    connect(this, &Combiner::deleteTcp, server, &OutputTcpServer::deleteTcp);
    connect(this, &Combiner::clearTcp, server, &OutputTcpServer::clear);
    connect(server, QOverload<OutputTcpStatus>::of(&OutputTcpServer::statusUpdated),
            this, &Combiner::tcpStatusUpdated);
    connect(server, &OutputTcpServer::destroyed, thread, &QThread::quit);
    connect(this, &Combiner::startingTcpWork, server, &OutputTcpServer::startWork);
    connect(this, &Combiner::closeTcpWork, server, &OutputTcpServer::finishWork);
    connect(server, &OutputTcpServer::workFinished, this, &Combiner::tcpFinished);
    if(id)
        connect(server, &OutputTcpServer::error, this, &Combiner::outputError);
    else
        connect(this, &Combiner::setMonitorPortNum, server, &OutputTcpServer::setMonitorPortNum);
    outputTcpServers[id] = server;
    outputTcpServerThreads[id] = thread;
    return id;
}

void Combiner::writeStartStopRecrd(bool isStart)
{
    Q_ASSERT(enabled);

    SartStopRecord record(isStart, runId.runNumber, runId.runIndex);

    RingBuffer<char> *primaryBuffer = nullptr;

    if(outToFile)
        primaryBuffer = writerRingBuf;
    else if(outToTcpServer)
        primaryBuffer = outputTcpServers[expectedEvNum%(outputTcpServers.size()-1)+1]->getRingBuffer();
    RingBuffer<char> *monitorBuffer = outputTcpServers[0]->getRingBuffer();
    const bool monitorSend = !stats.tcpStatus[0].isFree
            && (monitorBuffer->getFreeSize() >= record.totSize());

    if(primaryBuffer) {
        if(!primaryBuffer->waitFreeSpace(record.totSize(), EVENT_EXPIRE_TIMEOUT_MS)){
            if(!noFreeSpaceError)
                qCritical() << QString("Can't write %1 record, there is no free space in buffer")
                              .arg(isStart ? "start" : "stop");
            noFreeSpaceError = true;
            return;
        } else {
//            qInfo() << (isStart?"Start":"Stop")<<"record; totSize:"<<record.totSize()
//                    << (monitorSend?"monitorSend":"skip monitorSend");
            writeData(primaryBuffer, record.getData().data(), record.totSize());
        }
    }

    if(monitorSend)
        writeData(monitorBuffer, record.getData().data(), record.totSize());
}

void Combiner::addTrashStat(ClientIndex index, quint32 trashEvNum)
{
    ClientTrashStat &ts = trashStat[index];
    if(ts.contains(expectedEvNum)){
        EventsRange &range = ts[expectedEvNum];
        if(range.second+1 != trashEvNum)
            qWarning() << QString("Gap in trashing events (%1-%2) for same expected event number (%3)")
                          .arg(range.second).arg(trashEvNum).arg(expectedEvNum);
        range.second = trashEvNum;
    } else
        ts[expectedEvNum] = EventsRange(trashEvNum, trashEvNum);
}

void Combiner::resizeBuffer()
{
    Q_ASSERT(preferedBufSize>0);
    bool allResized=true;
    if(outToTcpServer) {
        int nextTcpSize = preferedBufSize/(outputTcpServers.size()-1);
        nextTcpSize = qMax(nextTcpSize, 300000000);
        for (int id : outputTcpServers.keys()) {
            if(id==0)
                continue;
            RingBuffer<char> *tcpRb = outputTcpServers[id]->getRingBuffer();
            if(tcpRb->getAllSize() == preferedBufSize)
                continue;

            bool ok=false;
            if(tcpRb->getFreeSize()==tcpRb->getAllSize()) {
                int oldSize = tcpRb->getAllSize();
                ok = tcpRb->resize(nextTcpSize);
                if(ok)
                    qInfo() << QString("TCP-buffer(%1) resized from %2MB to %3MB")
                               .arg(id).arg(1e-6*oldSize,0,'f',3)
                               .arg(1e-6*nextTcpSize,0,'f',3);
            }
            allResized &= ok;
        }
    } else {
        if(writerRingBuf->getFreeSize()==writerRingBuf->getAllSize()) {
            int oldSize = writerRingBuf->getAllSize();
            if(oldSize != preferedBufSize) {
                if(preferedBufSize < 1e5)
                    qInfo() << "Small size";
                int nextTcpSize = qMax(preferedBufSize, 300000000);
                allResized = writerRingBuf->resize(nextTcpSize);
                if(allResized)
                    qInfo() << QString("File-buffer resized from %2MB to %3MB")
                               .arg(1e-6*oldSize,0,'f',3)
                               .arg(1e-6*nextTcpSize,0,'f',3);
            } else
                allResized = true;

        } else
            allResized = false;
    }
    if(allResized) {
        waitingResize = false;
        return;
    }
}

void Combiner::checkStopEvNum() const
{
    if(fsm->getState() != FsmState::PrepareStopping)
        return;
    fsm->setExpectedEvNum(expectedEvNum);
}

void Combiner::onStoppingAction()
{
    workLoopTimer->stop();
    writeStartStopRecrd(false);
    enabled = false;
    if(outToFile) {
        emit closeFile();
    } else if(outToTcpServer) {
        ++finishId;
        // 0 reserved for error state
        if(finishId==0)
            ++finishId;
        emit closeTcpWork(finishId);
    }

    statusTimer->stop();
    sendStatus();
}

void Combiner::updateAnalizerList(const ClientIndexSet &clients, const QVector<Analizer *> &analizers)
{
    stats.clear();

    events.clear();
    clientIds = clients;
    eventRingBuffers.clear();
    eventRingBuffersDataSpace.clear();
    for (int i=0;i<analizers.size();++i) {
        const Analizer * a = analizers[i];
        eventRingBuffers[a->getClientIndex()] = a->getEventRingBuff();
        eventRingBuffersDataSpace[a->getClientIndex()] = a->getEventRingBuff()->getDataSize();
    }
    qDebug() << QString("clientNumbers=%1").arg(clients.size());
}

void Combiner::setStartEvNumber(quint32 startEvNum, quint32 step)
{
    startEvNumber = startEvNum;
    expectedEvNum = startEvNum;
    evNumberStep = step;
    qInfo() << QString("Set expected EvNum:%1; step:%2")
               .arg(expectedEvNum).arg(evNumberStep);
}

