#ifndef COMBINER_H
#define COMBINER_H

#include <QMap>
#include <QObject>
#include <QSet>

#include "base/RcRunId.h"
#include "fsm/FsmState.h"
#include "lib-common/MpdRawTypes.h"

#include "CombinerStats.h"
#include "EventMetaInfo.h"
#include "EvBConfig.h"
#include "FileWriter.h"
#include "params.h"
#include "OutputTcpServer.h"

class Analizer;
class CombinerFsm;
class FileWriter;
template<class T> class RingBuffer;

class QSemaphore;
class QTimer;
class OutputTcpServer;

typedef QPair<quint64, quint64> EventsRange;
typedef QMap<quint64, EventsRange> ClientTrashStat;


class Combiner : public QObject
{
    Q_OBJECT
public:
    explicit Combiner();
    ~Combiner();
    void initOutputBuffers(int size, bool outToTcp, int outTcpNum);

signals:
    void status(const QString &) const;
    void statusRaw(const CombinerStats &) const;
//    void clientStateUpdated(ClientIndex index, FsmState state); // from combiner and clients
    void fsmStateChanged(FsmState state) const;
    void error(const QString &) const;
    void startTcpServer(int index, bool state) const;
    void deleteTcp(int) const;
    void clearTcp() const;
    void startingFile(const RcRunId &runId, int writerIndex) const;
    void startingTcpWork() const;
    void closeFile() const;
    void closeTcpWork(uint) const;
    void storageChanged(const QString &dir) const;
    void deleteFileWriter() const;
    void fileWriterStatsUpdated(const FileWriterStats &s) const;
    void outputBufferSizeUpdated(quint64 size) const;
    void setMonitorPortNum(quint16 port) const;

public slots:
    void onFsmStateChanged(const FsmState &cur, const FsmState &prev);
    void start_run(const RcRunId &newRunId);
    void prepareForStop(quint32 stopEvNum);
    void stop();
    void destroyClient(const ClientIndex &clientId);
    void setOutput(COMBINER_OUT combinerOut, int tcpNum);
    void updateAnalizerList(const ClientIndexSet &clients,
                            const QVector<Analizer*> &analizers);
    void enableTimeDistr(bool en);
    void setStartEvNumber(quint32 startEvNum, quint32 step);

private slots:
    void init();
    void workLoop();
    void sendStatus();
    void tcpStatusUpdated(const OutputTcpStatus &s);
    void tcpFinished(int tcpId, uint fId) const;
    void writerFileOpened(const RcRunId &runId) const;
    void outputError(int index, const QString &errStr) const;

private:
    void processInit();
    void reset();
    void flushEvent(const ClientIndex &singleIndex= ClientIndex());
    void flushSingleEvent(const ClientIndex &index);
    void sendData(char *src, int size, bool addToOutEvent=false);
    void writeData(RingBuffer<char> *buffer, char *src, int size);
    bool updateClientsEventInfo(const ClientIndex &index, int timeout=0);
    int addTcpServer();
    void writeStartStopRecrd(bool isStart);
    void addTrashStat(ClientIndex index, quint32 evNum);
//    QString getClientNameByIndex(int index) const { return clientNames[clientIds[index]]; }
    void resizeBuffer();
    void checkStopEvNum() const;
    void onStartingAction();
    void onStoppingAction();

    CombinerFsm * const fsm;
    QSet<ClientIndex> clientIds; // client id by index of enabled client
//    QMap<int, QString> clientNames; // by client id
//    QVector<EventMetaInfo*> events; // Events pending to be summed
//    QVector<RingBuffer<EventMetaInfo> *> eventRingBuffers;
//    QVector<int> eventRingBuffersDataSpace;
    QMap<ClientIndex, EventMetaInfo*> events; // Events pending to be summed
    QMap<ClientIndex, RingBuffer<EventMetaInfo> *> eventRingBuffers;
    QMap<ClientIndex, int> eventRingBuffersDataSpace;
    QMap<ClientIndex, ClientTrashStat> trashStat;
    bool initFlag;
    bool enabled;
    bool eventStarted;
    int clientNumbers;
    int readoutClientNumbers;
    quint32 startEvNumber;
    quint32 evNumberStep;
    quint32 expectedEvNum;
    CombinerStats stats;
    RcRunId runId;

    MpdEventHdr mpdEventHdr;
    QTimer * const workLoopTimer;
    QTimer * const statusTimer;
    QString predefStorage;
    QMap<int, OutputTcpServer *> outputTcpServers;
    QMap<int, QThread *> outputTcpServerThreads;
    OutputTcpServer * monitorTcpServer;
    int outputBufferSize;
    RingBuffer<char> * const writerRingBuf;
    FileWriter * const fileWriter;
    QThread * const fileWriterThread;
    QByteArray outputEvent;
    bool outToFile;
    bool outToTcpServer;
    bool noFreeSpaceError; // for single error string
    uint eventsWithDebug;
    uint finishId;
//    bool prevEventIncomplet;
    int prevEventIncomplet;
    QElapsedTimer updateClientsTimer;
    uint curStateIndex;
    QMap<ClientIndex, quint32> lastMissingEvFromClient;
    int suppressLogLagging;
    bool timeDistrEn;

    bool waitingResize;
    int preferedBufSize;
};

#endif // COMBINER_H
