#include "CombinerFsm.h"

#include <QDebug>
#include <QLoggingCategory>

#include "EvBConfig.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.EvB.Combiner")
}

CombinerFsm::CombinerFsm(QObject *parent):
    BaseFsm(parent)
{
    setObjectName("CombinerFsm");
    recreateInput(new CombinerFsmInput());
}

void CombinerFsm::setOutputList(COMBINER_OUT combinerOut, int tcpNum)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << combinerOut << tcpNum;

    QSet<int> outputs;
    switch (combinerOut) {
    case COMBINER_FILE:
        outputs << 0;
        break;
    case COMBINER_TCP:
        for(int i=0; i<tcpNum; ++i)
            outputs << i+1;
    default:;
    }
    getInput()->combinerOutType = combinerOut;
    getInput()->outputList = outputs;
    evaluate();
}

void CombinerFsm::setOutputTcpConnected(int id, bool connected)
{
    auto input = getInput();
    if(!input->outputList.contains(id))
        return;

    if(input->outputConnected.value(id,false) != connected) {
        input->outputConnected[id] = connected;
        evaluate();
    }
}

void CombinerFsm::setOutputTcpDone(int id, bool rbIsEmpty)
{
    auto input = getInput();
    if(!input->outputList.contains(id))
        return;

    if(input->outputDone.value(id, false) != rbIsEmpty) {
        input->outputDone[id] = rbIsEmpty;
        if(getState() == FsmState::Stopping)
            evaluate();
    }
}

void CombinerFsm::setStopEvNum(quint64 stopEvNum)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << stopEvNum;
    getInput()->stopEvNum = stopEvNum;
    if(getState() == FsmState::PrepareStopping)
        evaluate();
}

void CombinerFsm::setExpectedEvNum(quint64 expectedEvNum)
{
    getInput()->expectedEvNum = expectedEvNum;
    if(getState() == FsmState::PrepareStopping) {
        qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << expectedEvNum;
        evaluate();
    }
}

void CombinerFsm::setFileCreated(bool created)
{
    getInput()->fileCreated = created;
    evaluate();
}

void CombinerFsm::setFileClosed(bool closed)
{
    getInput()->fileClosed = closed;
    evaluate();
}

bool CombinerFsm::evaluate()
{
    if(BaseFsm::evaluate())
        return true;

    if(!allOutputConnected()) {
        setState(FsmState::Error);
        return true;
    }

    switch (getState()) {
    case FsmState::Error:
        if (allOutputConnected()) {
            setState(FsmState::Ready);
            return true;
        }
        break;
    case FsmState::Ready:
        if (getRequestState() == FsmState::Starting) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::Starting:
        if(checkStart()) {
            setState(FsmState::Run);
            return true;
        }
        break;
    case FsmState::Run:
        if (getRequestState() == FsmState::PrepareStopping) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::PrepareStopping:
        if(checkStopEvNum()) {
            setState(FsmState::Stopping);
            return true;
        }
        break;
    case FsmState::Stopping:
        if(checkStop()) {
            setState(FsmState::Ready);
            return true;
        }
        break;
    default:;
    }

    return false;
}

void CombinerFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 CombinerFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Starting:
//        return 1000;
    case FsmState::PrepareStopping:
//        return 10000;
    case FsmState::Stopping:
//        return 100;
        return 1e4;
    default:
        return BaseFsm::stateTimeoutMs(state);
    }
}

bool CombinerFsm::checkStart() const
{
    const auto input = getInput();
    if(input->combinerOutType == COMBINER_FILE) {
        return input->fileCreated;
    } else if(input->combinerOutType == COMBINER_TCP) {
        return allOutputConnected();
    } else
        return true;
}

bool CombinerFsm::checkStop() const
{
    const auto input = getInput();
    if(input->combinerOutType == COMBINER_FILE) {
        return input->fileClosed;
    } else if(input->combinerOutType == COMBINER_TCP) {
        return allOutputDone();
    } else
        return true;
}

bool CombinerFsm::allOutputConnected() const
{
    const auto &i = getInput();
    if(i->combinerOutType != COMBINER_TCP)
        return true;

    for(const auto &id : i->outputList)
        if(i->outputConnected.value(id, false) == false)
            return false;
    return true;
}

bool CombinerFsm::allOutputDone() const
{
    for(const auto &id : getInput()->outputList)
        if(getInput()->outputDone.value(id, false) == false)
            return false;
    return true;
}

bool CombinerFsm::checkStopEvNum() const
{
    return (getInput()->stopEvNum == 0) ||
            ((getInput()->stopEvNum+1) == getInput()->expectedEvNum);
}

QDebug operator<<(QDebug dbg, const CombinerFsmInput &in)
{
    dbg << BaseFsmInput(in);
    dbg << "setupKey:" << in.setupKey << ";";
    dbg << "fileCreated:" << in.fileCreated << ";";
    dbg << "fileClosed:" << in.fileClosed << ";";
    dbg << "stopEvNum:" << in.stopEvNum << ";";
    dbg << "expectedEvNum:" << in.expectedEvNum << ";";
    dbg << "combinerOutType:" << in.combinerOutType << ";";
    dbg << "outputList:" << in.outputList << ";";
    dbg << "outputConnected:" << in.outputConnected << ";";
    dbg << "outputDone:" << in.outputDone << ";";
    return dbg;
}
