﻿//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MainWindow.h"
#include "ui_MainWindow.h"
#include "Client.h"
#include "Combiner.h"
#include "CombinerFsm.h"
#include "EvBConfigDispatcher.h"
#include "EvbSignalRouter.h"
#include "TcpClient.h"
#include "StatisticDialog.h"
#include "ClientManagerFsm.h"

#include "add-pnp-dialog/AddPnpDialog.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "base/ClientIndexConverter.h"
#include "base/SyslogMessageSender.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "configuration-manager/ConfigurationManager.h"
#include "metric-sender/MetricSender.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/RemoteControlServer.h"
#include "remote-control-server/RunControlCommands.h"

#include <QActionGroup>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QJsonDocument>
#include <QJsonObject>
#include <QHostAddress>
#include <QInputDialog>
#include <QMessageBox>
#include <QMetaType>
#include <QThread>
#include <QTimer>
#include <QSemaphore>
#include <QSettings>
#include <QScreen>
#include <time.h>
#include <QTime>
#include <iostream>

namespace {
enum COLUMNS{ COL_READOUT_CB, COL_PROG_TYPE, COL_PROG_INDEX,
              COL_CONNECTION_STATE, COL_PROG_HOST, COL_STATUS,
              COL_EV_NUM, COL_TRASHED_EVENTS, COL_ALL_EVENTS,
              COL_FLUSHED_SIZE, COL_TIME_DISTR, COL_TCP_BUFF,
              COL_EV_BUFF, COL_TOTAL};

QString getTcpName(int tcpId) {
    return tcpId ? QString("Tcp #%1").arg(tcpId-1) : QString("Mon");
}
const QList<DaqModule> EVB_SEARCH_TARGETS {
            DaqModule(DaqModule::Type::EventBuilder),
            DaqModule(DaqModule::Type::Generator),
            DaqModule(DaqModule::Type::MStream),
            DaqModule(DaqModule::Type::VmeDaq),
            DaqModule(DaqModule::Type::Land)
};

void updateFsmLabel(QLabel *label, const QString &prefix, FsmState st)
{
    label->setText(prefix + ": " + fsmStateToString(st));
    auto p = label->palette();
    auto b = p.brush(label->foregroundRole());
    switch (st) {
    case FsmState::Error:
        b.setColor(Qt::red);
        break;
    case FsmState::Ready:
        b.setColor(QColor());
        break;
    case FsmState::Run:
        b.setColor(Qt::darkGreen);
        break;
    default:
        b.setColor(Qt::darkYellow);
    }
    p.setBrush(label->foregroundRole(), b);
    label->setPalette(p);
}
}

using namespace std;

Q_DECLARE_METATYPE(Qt::GlobalColor)

MainWindow::MainWindow(const ConfigSelector &cs, QWidget *parent) :
    BaseMainWindow(parent)
    , ui(new Ui::MainWindow)
    , configuration_manager (new ConfigurationManager(this, cs.getType()))
    , configDb(new EvBConfigDb(cs, this))
    , configDispatcher(new EvBConfigDispatcher(this))
    , evb_fsm(new EventBuilderFsm(this))
    , client_manager_fsm(new ClientManagerFsm(this))
    , combiner(new Combiner())
    , combinerThread(new QThread(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , pnpServer(new PNPServer(this))
    , run(false)
    , labelFileErrSize(new QLabel(this))
    , clientDistrTime(new QElapsedTimer())
    , statisticDialog(new StatisticDialog(this))
    , backPreasure(false)
    , labelEvbFsmState(new QLabel(this))
    , labelCombinerFsmState(new QLabel(this))
    , labelClientManagerFsmState(new QLabel(this))
    , labelCombinerStatus(new QLabel(this))
{
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<AnalyzerStat>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<ClientIndex>();
    qRegisterMetaType<ClientIndexEnabledMap>();
    qRegisterMetaType<ClientIndexSet>();
    AbstractConfigConverter<ClientIndex>::registerTypes();

    ui->setupUi(this);
    ui->mainToolBar->hide();
    labelEvbFsmState->setMinimumWidth(120);
    labelCombinerFsmState->setMinimumWidth(150);
    labelClientManagerFsmState->setMinimumWidth(180);
    ui->statusBar->addWidget(labelEvbFsmState);
    ui->statusBar->addWidget(labelCombinerFsmState);
    ui->statusBar->addWidget(labelClientManagerFsmState);
    ui->statusBar->addWidget(labelCombinerStatus);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    auto evbSignalRouter = &EvbSignalRouter::instance();
    evbSignalRouter->setConfigurationManager(configuration_manager);
    evbSignalRouter->setEvBConfigDb(configDb);
    evbSignalRouter->setMainWindow(this);
    evbSignalRouter->setPnpServer(pnpServer);
    evbSignalRouter->setRemoteControlServer(remoteControl);
    evbSignalRouter->setWidgetRunControl(ui->widgetRunControl);
    evbSignalRouter->setEvbFsm(evb_fsm);
    evbSignalRouter->setClientManagerFsm(client_manager_fsm);

    evbSignalRouter->connectAll();

    QThread::currentThread()->setObjectName("GUI thread");
    combinerThread->setObjectName("combinerThread");

    QActionGroup *dataOutActionGroup = new QActionGroup(this);
    dataOutActionGroup->addAction(ui->actionOutToFile);
    dataOutActionGroup->addAction(ui->actionOutToTcp);
    dataOutActionGroup->addAction(ui->actionCombinerIdle);
    connect(dataOutActionGroup, &QActionGroup::triggered,
            this, &MainWindow::data_out_changed);

    pnpServer->startSearch(EVB_SEARCH_TARGETS, false);

    new ThreadController(pnpServer, this);
    new ThreadController(remoteControl, this);

    configDispatcher->subscribeCS(this);
    configDispatcher->subscribeCS(configDb);
    configDispatcher->subscribeCS(configuration_manager);
    configDispatcher->subscribeEvbConfig(this);
    configDispatcher->subscribeEvbConfig(configDb);
    configDispatcher->subscribeClients(this);
    configDispatcher->subscribeClients(client_manager_fsm);

    configuration_manager->setDefaultConfig(EvBConfig());
    init();
}

void MainWindow::init()
{
    setupTable();

    ui->statusBar->addWidget(labelFileErrSize);
//    statusBar()->addPermanentWidget(labelFileErrSize);
    statusBar()->setToolTip("reg:\tNumber of regular events\n"
                            "skip:\tNumber of skipped events from clients\n"
                            "eob:\tNumber of 'end of burst' events\n"
                            "clients:\tClient's ids from last combined event\n"
                            "nextEv:\tEvent number of next expected event\n"
                            "monitor:\tNumber of sent events to monitor");
    labelFileErrSize->hide();

    initCombiner();
    updateOutTcpLabelList();
    delete clientDistrTime;
}

void MainWindow::setConfig(const EvBConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;

    ui->actionAdvanced_counters->setChecked(config.advancedCounters);
    ui->actionAllow_Remote_Control->setChecked(config.isRemoteEnabled);
    ui->actionEnable_client_time_distribution->setChecked(config.timeDistrEn);

    //TODO: implement check that all parts are corresponding to Config
    emit allowRemoteControl(config.isRemoteEnabled);
    update_combiner_to_config();
    setup_clients();
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void MainWindow::onEvbFsmStateChanged(const FsmState &cur, const FsmState &prev)
{
    updateFsmLabel(labelEvbFsmState, "EvbFsm", cur);
    switch (cur) {
    case FsmState::Starting:
    {
        emit starting(loc_runId);
        client_manager_fsm->goState(FsmState::Starting);
        break;
    }
    case FsmState::Run:
    {
        auto jObj = QVariant::fromValue(config).value<QJsonObject>();
        auto cfg_byteArray = QJsonDocument(jObj).toJson();
        emit config_transfer_to_rc(cfg_byteArray);
        break;
    }
    case FsmState::PrepareStopping:
    {
        evb_fsm->goState(FsmState::Stopping);
        break;
    }
    case FsmState::Stopping:
    {
        emit stoping();
        evb_fsm->goState(FsmState::Ready);
        client_manager_fsm->goState(FsmState::Ready);
        break;
    }
    case FsmState::Error:
    {
        if(prev == FsmState::Starting || prev == FsmState::Run)
            emit stoping();
        break;
    }
    default:;
    }

    updatePushButtonState();
}

void MainWindow::onCombinerFsmStateChanged(const FsmState &cur_state)
{
    updateFsmLabel(labelCombinerFsmState, "CombinerFsm", cur_state);
}

void MainWindow::onClientManagerFsmStateChanged(const FsmState &cur_state)
{
    updateFsmLabel(labelClientManagerFsmState, "ClientManagerFsm", cur_state);
}

void MainWindow::startLocalRun()
{
    if(evb_fsm->getState() == FsmState::Ready)
        evb_fsm->goState(FsmState::Starting);
}

void MainWindow::stopLocalRun()
{
    if(evb_fsm->getState() == FsmState::Run)
        evb_fsm->goState(FsmState::Stopping);
}

void MainWindow::configSelectorChanged(const ConfigIndexName& conf)
{
    auto t = ConfigSelector(DaqModule(DaqModule::Type::EventBuilder),
                            conf).toWindowTitle();
    if(windowTitle() != t)
        setWindowTitle(t);

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.evb." + shortHostName + "."
                               + conf.getIndex().toString());

    initProgramDescription(conf);
}

MainWindow::~MainWindow()
{
    QElapsedTimer timer;
    timer.start();

    disconnect(remoteControl, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);

    emit closeProgram(progDesc);

//    Q_ASSERT(ret);

    emit destroyAll();
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"destroyClient(0);";
    //todo: wait all clients are destroyed

    timer.start();

    //Give time to proceed destroyed signal from Combiner
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"threadQuiting(combinerThread";
    if(!threadQuiting(combinerThread, 5000, &timer)){
        qWarning() << "Combiner's thread isn't stopped";
//        qInfo()<<"[MainWindow] "<<timer.elapsed()<<"Combiner's thread isn't stopped";
    }

    disconnect(&SyslogMessageSender::instance(), &SyslogMessageSender::newMsg,
               nullptr, nullptr);
    delete ui;
}

void MainWindow::on_pushButtonAdd_clicked()
{
    QVector<ProgramDescription> freeProgs = getFreeProgs();
    if(!freeProgs.size())
        return;

    AddPnpDialog dialog(freeProgs, true, this);
    connect(pnpServer, &PNPServer::programClosed, &dialog, &AddPnpDialog::gotClosedProgram);
    connect(pnpServer, &PNPServer::programDiscovered, &dialog, &AddPnpDialog::gotProgramDescription);

    dialog.set_function_filter(std::bind(&MainWindow::isFreeProg, this, std::placeholders::_1));

    auto clients_map = config.getClientIndexEnabledMap();
    if(dialog.exec() == QDialog::Accepted) {
        for(auto pd : dialog.getProgDescr()) {     
            ClientIndex index(pd.getDaqModule(), pd.index);
            clients_map[index] = ClientIndexEnabled(index, true);
        }
        emit clientListChanged(clients_map);
    }
}

void MainWindow::on_pushButtonRemove_clicked()
{
    int row = getSelectedRow();
    ClientIndex index;
    if(row >= 0) {
        index = getClientIndexByRow(row);
    }

    auto clients_map = config.getClientIndexEnabledMap();
    if(index.isEmpty() || !clients_map.contains(index)) {
        qWarning() << "Don't know what client was selected; removing canceled";
        return;
    }

    QString message = QString("Are you shure to remove client %1?")
        .arg(index.toString());
    int ret = QMessageBox::warning(this, tr("Confirm removing"), message,
                                   QMessageBox::Yes | QMessageBox::No,
                                   QMessageBox::No);
    if(ret == QMessageBox::No) {
        qDebug() << "Client's removing canceled";
        return;
    }

    clients_map.remove(index);
    emit clientListChanged(clients_map);

    qInfo() << QString("Client deleted %1").arg(index.toString());
}

void MainWindow::setupTable()
{
    QTableWidget *table = ui->tableWidget;
    table->setColumnCount(COL_TOTAL);
#if QT_VERSION >= 0x050000
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    table->setHorizontalHeaderItem(COL_READOUT_CB, new QTableWidgetItem("readout"));
    table->setHorizontalHeaderItem(COL_PROG_TYPE, new QTableWidgetItem("type"));
    table->setHorizontalHeaderItem(COL_PROG_INDEX, new QTableWidgetItem("index"));
    table->setHorizontalHeaderItem(COL_CONNECTION_STATE, new QTableWidgetItem("Online"));
    table->setHorizontalHeaderItem(COL_PROG_HOST, new QTableWidgetItem("host"));
    table->setHorizontalHeaderItem(COL_STATUS, new QTableWidgetItem("state"));
    table->setHorizontalHeaderItem(COL_EV_NUM, new QTableWidgetItem("evNum"));
    table->setHorizontalHeaderItem(COL_TRASHED_EVENTS, new QTableWidgetItem("trashed"));
    table->setHorizontalHeaderItem(COL_ALL_EVENTS, new QTableWidgetItem("all"));
    table->setHorizontalHeaderItem(COL_FLUSHED_SIZE, new QTableWidgetItem("flushedSize"));

    table->setHorizontalHeaderItem(COL_TIME_DISTR, new QTableWidgetItem("Time distr"));
    table->setHorizontalHeaderItem(COL_TCP_BUFF, new QTableWidgetItem("Raw data buff"));
    table->setHorizontalHeaderItem(COL_EV_BUFF, new QTableWidgetItem("Ev. meta-info"));
}

void MainWindow::on_tableWidget_itemSelectionChanged()
{
    updatePushButtonState();
    updateTablePalette();
}

void MainWindow::updateClientStatus(ClientIndex index, bool isConnected, QString status)
{
    if(!config.clients.contains(index))
        return;
    QTableWidget *table = ui->tableWidget;
    bool connectionChanged = !lastConnectionState.contains(index)
            || lastConnectionState[index] != isConnected;
    lastConnectionState[index] = isConnected;
    if(connectionChanged)
        updateTablePalette();
    const int row = getTableRowByClientIndex(index);
    if(row != -1) {
        table->item(row, COL_STATUS)->setText(status);
        QLabel *label = qobject_cast<QLabel*>(table->cellWidget(row, COL_CONNECTION_STATE));
        QPixmap p;
        if (isConnected){
            label->setToolTip("Device sends discover packets");
            p = QApplication::style()->standardPixmap(QStyle::SP_DialogApplyButton);
        } else {
            if(run && config.clients.contains(index) && config.clients[index].enabled){
                QString errStr = QString("%1 disconnected").arg(index.toString());
                qWarning() << errStr;
                // todo handle mstream restart period
                if(lastCombinerStats.hasPendingEvent && lastCombinerStats.regularEvents)
                    deviceError(index, errStr);
            }
            label->setToolTip("Device doesn't send discover packets");
            p = QApplication::style()->standardPixmap(QStyle::SP_DialogCloseButton);
        }
        if(connectionChanged)
            updateTableTextColor(row, index);

        label->setPixmap(p);
        updateTableSize();
        return;
    }
}

void MainWindow::updateTcpStatistic(ClientIndex index, const TcpStatistic &statistic)
{
    QTableWidget *table = ui->tableWidget;
    const int row = getTableRowByClientIndex(index);
    // Row not found
    if(row == -1)
        return;
    // Row found
    QString text = QString("%1/%2/%3")
            .arg( statistic.freePart, 4, 'f', 1)
            .arg( statistic.dataPart, 4, 'f', 1)
            .arg(statistic.trashPart, 4, 'f', 1);
    table->item(row, COL_TCP_BUFF)->setText(text);
    updateTableSize();
}

void MainWindow::setup_clients()
{
    for(ClientIndex index : clients.keys()) {
        if(!config.clients.contains(index)) {
            emit destroyClient(index);
            clients.remove(index);
            int row = getTableRowByClientIndex(index);
            QWidget *chButton = ui->tableWidget->cellWidget(row, COL_READOUT_CB);
            QWidget *stateLabels = ui->tableWidget->cellWidget(row, COL_CONNECTION_STATE);
            Q_ASSERT(chButton);
            Q_ASSERT(stateLabels);
            delete chButton;
            delete stateLabels;
            ui->tableWidget->removeRow(getTableRowByClientIndex(index));
            lastConnectionState.remove(index);
        }
    }

    for(ClientIndex index : config.clients.keys()) {
        if(!clients.contains(index))
            addClient(config.clients[index]);
    }
    updateTableSize();

    updateAnalizerList();
}

bool MainWindow::addClient(const EvBClientConfig &client_config)
{
    Q_ASSERT(evb_fsm->getState()==FsmState::Init ||
             evb_fsm->getState()==FsmState::Ready ||
             evb_fsm->getState()==FsmState::Error);

    ClientIndex index = client_config.index;
    if (index.isEmpty())
        return false;
    if (clients.contains(index))
        return false;

    Client *clientManager = new Client(client_config, config.getEvNumIndex(), this);
    makeConnections(clientManager);
    clients[index] = clientManager;
    addTableRow(client_config);
    emit updateClientReadout(index, client_config.enabled);
    updateTableSize();
    qDebug() << QString("Added client %1").arg(index.toString());
    return true;
}

void MainWindow::makeConnections(Client *cm) const
{
    connect(cm, &Client::sendStatus, this, &MainWindow::updateClientStatus);
    connect(cm, &Client::sendStatistic, this, &MainWindow::updateTcpStatistic);
    connect(this, &MainWindow::starting, cm, &Client::startRun);
    connect(this, &MainWindow::stoping, cm, &Client::stopRun);
    connect(cm, &Client::clientStateUpdated, client_manager_fsm, &ClientManagerFsm::setClientState);
    connect(this, &MainWindow::destroyClient, cm, &Client::destroyClient);
    connect(this, &MainWindow::destroyAll, cm, &Client::deleteLater);
    connect(this, &MainWindow::updateClientReadout, cm, &Client::updateClientReadout);

    connect(cm, &Client::sendStatistic, statisticDialog, &StatisticDialog::updateTcpStatistic);
    connect(cm, &Client::statUpdated, statisticDialog, &StatisticDialog::analyzerStatUpdated);

    connect(cm, &Client::inputBufferSizeUpdated, this, &MainWindow::inputBufferSizeUpdated);
    connect(this, &MainWindow::printFsmInput, cm, &Client::printFsmInput);
}

void MainWindow::addTableRow(const EvBClientConfig &cc) const
{
    QTableWidget *table = ui->tableWidget;
    int row = table->rowCount();
    table->setRowCount(row+1);
    QCheckBox *cb = new QCheckBox(table);
    cb->setChecked(cc.enabled);
    connect(cb, &QCheckBox::toggled, this, &MainWindow::setClientReadOut);
    QLabel *label = new QLabel(table);
    label->setAlignment(Qt::AlignCenter);

    QTableWidgetItem *item = nullptr;
    table->setCellWidget(row, COL_READOUT_CB, cb);
    table->setCellWidget(row, COL_CONNECTION_STATE, label);
    table->setItem(row, COL_PROG_TYPE, new QTableWidgetItem(cc.index.program_type.toString()));
    table->setItem(row, COL_PROG_INDEX, item = new QTableWidgetItem(cc.index.program_index.toString()));
    item->setData(Qt::UserRole, QVariant::fromValue(cc.index));
    table->setItem(row, COL_PROG_HOST, new QTableWidgetItem(""));
    table->setItem(row, COL_STATUS, new QTableWidgetItem("creating..."));
    table->setItem(row, COL_EV_NUM, new QTableWidgetItem());
    table->setItem(row, COL_TRASHED_EVENTS, new QTableWidgetItem());
    table->setItem(row, COL_ALL_EVENTS, new QTableWidgetItem());
    table->setItem(row, COL_FLUSHED_SIZE, new QTableWidgetItem());
    table->setItem(row, COL_TIME_DISTR, new QTableWidgetItem());
    table->setItem(row, COL_TCP_BUFF, new QTableWidgetItem());
    table->setItem(row, COL_EV_BUFF, new QTableWidgetItem());

    table->item(row, COL_TIME_DISTR)->setToolTip("Aver. time diff for earliest/\nAver. time full, ms");
    table->item(row, COL_TCP_BUFF)->setToolTip("free/data/trash");
    table->item(row, COL_EV_BUFF)->setToolTip("free/prepared/flushed");
}

void MainWindow::updatePushButtonState()
{
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    bool isStopped = (evb_fsm->getState()==FsmState::Ready) || (evb_fsm->getState()== FsmState::Error);
    if (ranges.empty() || !isStopped) {
        ui->pushButtonRemove->setEnabled(false);
    } else {
        Q_ASSERT(ranges.size()==1);
        ui->pushButtonRemove->setEnabled(true);
    }

    ui->pushButtonAdd->setEnabled(isStopped && !getFreeProgs().isEmpty());
    ui->pushButtonReset->setEnabled(isStopped);
    ui->tableWidget->setEnabled(isStopped);
    ui->actionSet_EvB_index->setEnabled(isStopped);
    ui->actionOutToFile->setEnabled(isStopped);
    ui->actionOutToTcp->setEnabled(isStopped);
    ui->actionCombinerIdle->setEnabled(isStopped);
    ui->actionChange_storage_dir->setEnabled(isStopped);
    ui->actionReset_to_defaults->setEnabled(isStopped);
}

void MainWindow::updateTableSize()
{
    int width = ui->tableWidget->verticalHeader()->width();
    for(int i=0;i<ui->tableWidget->columnCount();++i){
        if(!ui->tableWidget->isColumnHidden(i))
            width += ui->tableWidget->columnWidth(i);
    }
    width += 5;

    QScreen *screen = QGuiApplication::primaryScreen();
    QRect rec = screen->geometry();
    int s_width = rec.width();

    if(width < (s_width*0.66))
        ui->tableWidget->setMinimumWidth(width);
}

void MainWindow::updateAnalizerList()
{
    QVector<Analizer *> analizers;
    ClientIndexSet clients_;
    for(ClientIndex index : clients.keys()) {
        if(clients[index]->isReadoutEnabled()) {
            analizers.push_back(clients[index]->getAnalizer());
            clients_.insert(clients[index]->getIndex());
        }
    }
    enabledClientNum = analizers.size();
    emit analizerListUpdated(clients_, analizers);
    updateRemoteStatus();
    updateProgramDescription();
}

void MainWindow::updateRemoteStatus()
{
    QString stateStr;
    int clientsEnabled=0;
    for(const Client *mgr : clients.values()){
        if(mgr->isReadoutEnabled())
            ++clientsEnabled;
    }
    if(clientsEnabled)
        stateStr = QString::number(clientsEnabled).append(" clients");
    else
        stateStr = "no clients";
    emit sendRemoteState(stateStr);
}

void MainWindow::initProgramDescription(const ConfigIndexName& conf)
{
    progDesc.type = DaqModule(DaqModule::Type::EventBuilder).toString();
    progDesc.index = conf.getIndex().toString();
    progDesc.name = progDesc.getName();
}

void MainWindow::updateProgramDescription()
{
    bool hasRcInterface = progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl);
    ProgramInterface rcInterface;
    if(hasRcInterface)
        rcInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    progDesc.clear();
    if(hasRcInterface)
        progDesc.interfaces.push_back(rcInterface);

//    for (ClientIndex index : lastCombinerStats.tcpStatus.keys()) {
    for (int i=0; i<lastCombinerStats.tcpStatus.size(); ++i) {
        const OutputTcpStatus &tcpStatus = lastCombinerStats.tcpStatus[i];
        if(!tcpStatus.listening)
            continue;
        ProgramInterface tcpInterface;
        tcpInterface.setType(i==0 ? ProgramInterfaceMonitorData : ProgramInterfaceDataFlow);
        tcpInterface.enabled = true;
//        tcpInterface.id = id.i ? id-1 : 0;
        tcpInterface.port = tcpStatus.port;
        tcpInterface.isFree = tcpStatus.isFree;
        if(!tcpStatus.isFree){
            tcpInterface.peers << ProgramInterfacePeer(tcpStatus.peer_host, tcpStatus.peer_port);
        }
        progDesc.interfaces.push_back(tcpInterface);
    }

    for (Client *manager : clients) {
        if(manager->isReadoutEnabled()) {
            progDesc.options[QString("Client %1").arg(manager->getIndex().toString())] =
                    manager->getProgDescr().getName();
        }
    }

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif
    progDesc.options["run"] = QVariant(run).toString();
    progDesc.options["runIndex"] = loc_runId.runIndex;
    progDesc.options["runNumber"] = QString::number(loc_runId.runNumber);
    progDesc.options["output"] = config.getOutputStr();

    emit changeProgramDescription(progDesc);
}

void MainWindow::updateTcpPortLabels()
{
    for (OutputTcpStatus st : lastCombinerStats.tcpStatus.values()) {
        QLabel *l = getTcpLabel(st.index);
        Q_ASSERT(l);
        QString text = QString("Tcp %1: %2")
                .arg(st.index)
                .arg(st.port);
//        if(st.dataPart>0.1){
            text.append(QString("(%1%)").arg((int)(100*st.dataPart)));
//        }
        l->setText(text);
        l->setStyleSheet(QString("QLabel { color : %1; }")
                         .arg(st.isFree ? "red" : "green"));
    }
}

void MainWindow::setOutTcpServerNumber(int tcpNum)
{
    if(tcpNum < 1)
        tcpNum = 1;
    if(config.combinerOutTcpNum != tcpNum){
        config.combinerOutTcpNum = tcpNum;
        emit configChanged(config);
    }

    updateOutTcpLabelList();

    emit setOutput(config.combinerOut, config.combinerOutTcpNum);
}

QVector<ProgramDescription> MainWindow::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;
    for (const ProgramDescription &descr : progDescrs) {
        if(isFreeProg(descr))
            freeProgs.push_back(descr);
    }
    return freeProgs;
}

bool MainWindow::isFreeProg(const ProgramDescription &pd) const
{
    for (auto client : clients.values()) {
        if(pd.sameProgram(client->getProgDescr())){
            return false;
        }
    }

    int interface_id=config.startEvNumber % config.evNumberStep;
    for (const ProgramInterface &i : pd.interfaces) {
        if(i.type == ProgramInterface::getTypeName(ProgramInterfaceDataFlow)
                && i.enabled && i.isFree && i.id == interface_id) {
            return true;
        }
    }

    return false;
}

int MainWindow::getTableRowByClientIndex(const ClientIndex &index) const
{
    QTableWidget *table = ui->tableWidget;
    for(int row =0; row < table->rowCount(); ++row) {
        if(index == getClientIndexByRow(row))
            return row;
    }
    return -1;
}

int MainWindow::getSelectedRow() const
{
    int row=-1;
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    if(!ranges.isEmpty()){
        QTableWidgetSelectionRange firstRange = ranges[0];
        row = firstRange.topRow();
    }
    return row;
}

ClientIndex MainWindow::getClientIndexByRow(int row) const
{
    const auto &item = ui->tableWidget->item(row, COL_PROG_INDEX);
    return item ? item->data(Qt::UserRole).value<ClientIndex>() : ClientIndex();
}

void MainWindow::updateTableTextColor(int row, ClientIndex index) const
{
    if(index.isEmpty())
        return;

    const bool &isOk = !config.clients[index].enabled || lastConnectionState[index];
    QColor color = isOk ? QColor() : Qt::red;
    QTableWidgetItem *i = ui->tableWidget->item(row, COL_STATUS);
    if(i->textColor() != color)
        i->setTextColor(color);
}

void MainWindow::combinerOutUpdated()
{
    if(config.isOutTcp() && config.combinerOutTcpNum==0){
        setOutTcpServerNumber(1);
        emit configChanged(config);
    }

    ui->actionSet_number_of_outputTcpServers->setEnabled(config.isOutTcp());
    ui->labelDataFileName->setVisible(config.isOutFile());
    ui->labelCombinerIdle->setVisible(config.isOutIdle());
    ui->widgetOutTcp->setVisible(config.isOutTcp());
    emit setOutput(config.combinerOut, config.combinerOutTcpNum);
}

QLabel *MainWindow::getTcpLabel(int index) const
{
    bool ok;
    for(int i=0; i<ui->horizontalLayoutOutTcp->count(); ++i) {
        QLayoutItem *item = ui->horizontalLayoutOutTcp->itemAt(i);
        if(item) {
            QWidget *w = item->widget();
            if(w) {
                QLabel *l = qobject_cast<QLabel *>(w);
                if(l) {
                    if(l->property("id").toInt(&ok)==index && ok)
                        return l;
                }
            }
        }
    }
    return nullptr;
}


void MainWindow::initCombiner()
{
    combiner->initOutputBuffers(config.outputBufSize, config.combinerOut, config.combinerOutTcpNum);

    combiner->moveToThread(combinerThread);
    combinerThread->start();
    connect(this, &MainWindow::starting, combiner, &Combiner::start_run);
    connect(this, &MainWindow::prepareForStop, combiner, &Combiner::prepareForStop);
    connect(this, &MainWindow::stoping, combiner, &Combiner::stop);
    connect(this, &MainWindow::destroyClient, combiner, &Combiner::destroyClient);
    connect(this, &MainWindow::storageChanged, combiner, &Combiner::storageChanged);
    connect(this, &MainWindow::setOutput, combiner, &Combiner::setOutput);
    connect(this, &MainWindow::analizerListUpdated, combiner, &Combiner::updateAnalizerList);
    connect(this, &MainWindow::analizerListUpdated, statisticDialog, &StatisticDialog::analizerListUpdated);
    connect(this, &MainWindow::enableTimeDistr, combiner, &Combiner::enableTimeDistr);
    connect(this, &MainWindow::setEvNumberStep, combiner, &Combiner::setStartEvNumber);

//    connect(combiner, &Combiner::clientStateUpdated, fsm, &EventBuilderFsm::setClientState);
    connect(combiner, &Combiner::error, this, &MainWindow::combinerError);

    connect(combiner, &Combiner::destroyed, combinerThread, &QThread::quit);
    connect(combiner, &Combiner::status, this, &MainWindow::updateStatus);
    connect(combiner, &Combiner::statusRaw, this, &MainWindow::updateCombinerStats);
    connect(combiner, &Combiner::statusRaw, statisticDialog, &StatisticDialog::updateCombinerStats);
    connect(combiner, &Combiner::fileWriterStatsUpdated, this, &MainWindow::fileWriterStatsUpdated);
    connect(combiner, &Combiner::outputBufferSizeUpdated, this, &MainWindow::outputBufferSizeUpdated);
    connect(this, &MainWindow::setMonitorPortNum, combiner, &Combiner::setMonitorPortNum);
}

void MainWindow::update_combiner_to_config()
{
    emit storageChanged(config.storageDir);
    updateAnalizerList();
    ui->actionOutToFile->setChecked(config.isOutFile());
    ui->actionOutToTcp->setChecked(config.isOutTcp());
    ui->actionCombinerIdle->setChecked(config.isOutIdle());
    combinerOutUpdated();
    setOutTcpServerNumber(config.combinerOutTcpNum);


    ui->actionEnable_client_time_distribution->setChecked(config.timeDistrEn);
    emit setEvNumberStep(config.startEvNumber, config.evNumberStep);
    if(config.monitorPort)
        emit setMonitorPortNum(config.monitorPort);
}

void MainWindow::updateOutTcpLabelList()
{
    QLayout *layout = ui->horizontalLayoutOutTcp;

    const auto &tcpNum = config.combinerOutTcpNum;
    for(int index=0; index < layout->count(); ) {
        bool ok;
        QWidget *w = layout->itemAt(index)->widget();
        if(w->property("id").toInt(&ok) > tcpNum && ok) {
            layout->removeWidget(w);
            delete w;
        } else {
            ++index;
        }
    }

    for(int tcpId=0; tcpId <= tcpNum; ++tcpId) {
        if(getTcpLabel(tcpId)) continue;
        QLabel *label = new QLabel(QString("%1: creating...").arg(getTcpName(tcpId)), ui->widgetOutTcp);
//        if(tcpId == 0) label->setVisible(false);
        label->setProperty("id", tcpId);
        layout->addWidget(label);
        label->setAlignment(Qt::AlignHCenter);
    }
}

void MainWindow::combinerError(const QString &errStr)
{
    qCritical() << "Got error from combiner" << errStr;
    if(run)
        emit sendRunError(loc_runId, errStr);
}

void MainWindow::deviceError(const ClientIndex &index, const QString &errStr)
{
    QString msg = QString("Got error from %1").arg(index.toString());
    qCritical() << msg << errStr;
    if(run)
        emit sendRunError(loc_runId, errStr);
}

void MainWindow::updateStatus(const QString &status)
{
    labelCombinerStatus->setText(status);
    emit sendRemoteState(status);
}

void MainWindow::updateCombinerStats(const CombinerStats &ss)
{
    if(backPreasure){
        if(ss.fwDataPart > 0.9) {
            backPreasure = false;
            QString mess = "Enable backPressure";
            qWarning() << mess;
            emit sendRunWarning(loc_runId, mess);
            emit sendBackPreasure(loc_runId, true);
        }
    } else {
        if(ss.fwDataPart < 0.5) {
            backPreasure = true;
            qInfo() << "BackPreasure disabled";
            emit sendBackPreasure(loc_runId, false);
        }
    }
    bool outDataInterfaceChanged = false;
    bool outDataPartsChanged = false;
    for (int id : ss.tcpStatus.keys()) {
        if(!lastCombinerStats.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
        const OutputTcpStatus &prevS = lastCombinerStats.tcpStatus[id];
        const OutputTcpStatus &nextS = ss.tcpStatus[id];
        outDataInterfaceChanged |= prevS.isFree!=nextS.isFree || prevS.port!=nextS.port;
        outDataPartsChanged |= !qFuzzyCompare(prevS.dataPart, nextS.dataPart);
    }
    for (int id : lastCombinerStats.tcpStatus.keys()) {
        if(!ss.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
    }
    lastCombinerStats = ss;
//    fsm->setExpectedEvNum(ss.expectedEvNum);
    QTableWidget * table = ui->tableWidget;
    const bool tableUpdatesEnabled = table->updatesEnabled();
    table->setUpdatesEnabled(false);
    if(outDataInterfaceChanged)
        updateProgramDescription();
    if(outDataInterfaceChanged || outDataPartsChanged)
        updateTcpPortLabels();

    if(ss.lastEventNum.isEmpty()) {
        table->setUpdatesEnabled(tableUpdatesEnabled);
        return;
    }
    for(int row=0; row<table->rowCount(); ++row){
        ClientIndex index = getClientIndexByRow(row);
        if(index.isEmpty())
            continue;
        QString str;
        if(ss.lastEventNum.contains(index))
            str = QString::number(ss.lastEventNum[index]);
        table->item(row, COL_EV_NUM)->setText(str);
        table->item(row, COL_TRASHED_EVENTS)->setText(QString::number(ss.trashedEvents[index]));
        table->item(row, COL_ALL_EVENTS)->setText(QString::number(ss.dataEvents[index]));
        const quint64 flushedSize = ss.flushedSize[index];
        table->item(row, COL_FLUSHED_SIZE)->setText(humanReadableSize(flushedSize));
        table->item(row, COL_FLUSHED_SIZE)->setToolTip(QString::number(flushedSize));
        if(config.timeDistrEn){
            if(ss.timeDistrCnt && ss.timeDistr.contains(index) && ss.timeDistrFull.contains(index))
                table->item(row, COL_TIME_DISTR)->
                        setText(QString("%1/%2")
                                .arg(ss.timeDistr[index]/ss.timeDistrCnt)
                                .arg(ss.timeDistrFull[index]/ss.timeDistrCnt));
            else
                table->item(row, COL_TIME_DISTR)->setText("-");
        }

        if(ss.freeEvPart.contains(index)) {
            QString evBuff = QString("%1/%2/%3")
                    .arg(ss.freeEvPart[index])
                    .arg(ss.preparedEvPart[index])
                    .arg(ss.flushedEvPart[index]);
            table->item(row, COL_EV_BUFF)->setText(evBuff);
        }

        // metrics
        QString clientName = QString("%1").arg(index.toString()/*, 8, 16, QChar('0')*/);
        if (clients.contains(index) && clients[index]) {
            const Client * cm = clients[index];
            const ProgramDescription &pd = cm->getProgDescr();
            clientName = pd.index;
        }
        QString metricPrefix = QString("in.%1.").arg(clientName);
        MetricSender::send(metricPrefix+"subevents.drop", ss.trashedEvents[index]);
        MetricSender::send(metricPrefix+"subevents.pass", ss.dataEvents[index] - ss.trashedEvents[index]);
    }

    {
        QString mess;
//        if(!enabled)
//            mess.append("stopped; ");
        QString clientsStr;
        int first=-1;
        int last=-1;
        for(int r=0; r<ui->tableWidget->rowCount(); ++r){
            const auto &i = getClientIndexByRow(r);
            if(!ss.lastEventClients.contains(i))
                continue;
            if(first==-1) {
                first = r;
            } else if(last+1!=r) {
                if(first == last)
                    clientsStr.append(QString("%1 ").arg(first));
                else
                    clientsStr.append(QString("%1-%2 ").arg(first).arg(last));
                first = r;
            }
            last = r;
        }

        if(first!=-1){
            if(first == last)
                clientsStr.append(QString("%1 ").arg(first));
            else
                clientsStr.append(QString("%1-%2 ").arg(first).arg(last));
        }

        mess.append(QString("reg:%1;skip:%2;eob:%3;clients:%4;nextEv:%5  monitor:%6")
                    .arg(ss.regularEvents)
                    .arg(ss.skipped)
                    .arg(ss.eob)
                    .arg(clientsStr)
                    .arg(ss.expectedEvNum)
                    .arg(ss.monitorEvCount)
                    );
//        for(int i=0;i<events.size(); ++i){
//            const EventMetaInfo* ev = events[i];
//            if(ev == nullptr) continue;
//            mess.append(QString("; #%1=%2").arg(i).arg(events[i]->evNum));
//        }
        updateStatus(mess);
    }

    labelFileErrSize->hide();
    labelFileErrSize->setText(QString("fileErr:%1B").arg(ss.missedFileSize));
    if(ss.missedFileSize){
        labelFileErrSize->show();
        labelFileErrSize->setStyleSheet("QLabel { color : red;}");
    }
    labelFileErrSize->setStyleSheet(QString("QLabel { color : %1;}")
                                    .arg(ss.missedFileSize?"red" : "black"));
    updateTableSize();

    // Periodic print of client's number distribution
//    if(!clientDistrTime->isValid() || clientDistrTime->elapsed()>1000) {
//        QString str("clientNumInEvents \t");
//        quint64 sum=0;
//        QStringList disrStrList;
//        for(int n : ss.clientNumDistr.keys()) {
//            disrStrList << QString("{%1:%2}").arg(n).arg(ss.clientNumDistr[n]);
//            sum += ss.clientNumDistr[n]*n;
//        }
//        str += disrStrList.join(", ");
//        qInfo().noquote()<<str;
//        qInfo().noquote().nospace()
//                <<"Sum:"<<sum<<"; skipped:"<<ss.skipped<<"; Sum+skipped:"<<sum+ss.skipped;
//        int enabledClients = config.getClientIndexEnabledMap().getEnabled().size();
//        qInfo().noquote().nospace()
//                <<"Exp:"<<ss.expectedEvNum<<"; clients:"<<enabledClients<<"; Exp*clients:"<<ss.expectedEvNum*enabledClients;
//        clientDistrTime->start();
//    }

    table->setUpdatesEnabled(tableUpdatesEnabled);
}

void MainWindow::fileWriterStatsUpdated(const FileWriterStats &fwStat)
{
    QString dataFileNameText;
    if(!fwStat.fileName.isEmpty()){
        dataFileNameText = QString("%1 (%2)")
                .arg(fwStat.fileName)
                .arg(humanReadableSize(fwStat.writtenFileSize));
        if(fwStat.missedFileSize)
            dataFileNameText.append(QString(" (missed: %1b)").arg(fwStat.missedFileSize));
    } else {
        dataFileNameText = QString(fwStat.predefStorage).append("...");
    }
    ui->labelDataFileName->setText(dataFileNameText);
}

void MainWindow::inputBufferSizeUpdated(ClientIndex id, quint64 size)
{
    qInfo() << "Save in cgf input resize:" << id << size;
    if(size<=0) {
        qWarning() << "Got inputBufferSizeUpdated with incorrect size:"<<size;
        return;
    }
    if(!config.clients.contains(id)) {
        qWarning() << "Got inputBufferSizeUpdated for missing clientId:"<<id;
        return;
    }

    config.clients[id].inputBufSize = size;
    emit configChanged(config);
}

void MainWindow::outputBufferSizeUpdated(quint64 size)
{
    qInfo() << "Save in cgf output resize:" << size;
    if(size<=0) {
        qWarning() << "Got outputBufferSizeUpdated with incorrect size:"<<size;
        return;
    }
    config.outputBufSize = size;
    emit configChanged(config);
}

void MainWindow::remoteStartRequest(const RcRunId &rcRunId, quint64 startEvNumber)
{
    ui->plainTextEdit->appendPlainText("");
    qInfo() << QString("Got remote request to start %1").arg(rcRunId.toString());
    emit configChanged(config);

    static bool warn_start_event_number = true;
    if(startEvNumber && warn_start_event_number) {
        qWarning() << "Remote start with start_event_number not implemented yet.";
        warn_start_event_number = false;
    }

    if(loc_runId.runIndex == rcRunId.runIndex &&
            loc_runId.runNumber > rcRunId.runNumber) {
        const auto &msg = QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                              .arg(loc_runId.runNumber)
                              .arg(rcRunId.runNumber);
        qWarning() << msg;
        emit sendRunWarning(loc_runId, msg);
        return;
    }

    if(evb_fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Ready))
                              .arg(fsmStateToString(evb_fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(rcRunId, msg);
        return;
    }

    loc_runId = rcRunId;
    evb_fsm->goState(FsmState::Starting);
}

void MainWindow::remoteStopRequest(const RcRunId &rcRunId, quint64 stopEvNum)
{
    qDebug() << __PRETTY_FUNCTION__ << rcRunId << stopEvNum;

    static bool warnStopEvNum = true;
    if(stopEvNum && warnStopEvNum) {
        warnStopEvNum = false;
        qWarning() << "Stopping with 'stop_event_number' should be fixed";
    }
    bool success=true;
    if(loc_runId.runIndex != rcRunId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                     .arg(loc_runId.runIndex)
                     .arg(rcRunId.runIndex);
        success = false;
    }
    if(success && loc_runId.runNumber != rcRunId.runNumber){
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                     .arg(loc_runId.runNumber)
                     .arg(rcRunId.runNumber);
        success = false;
    }


    if(evb_fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(evb_fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(loc_runId, msg);
        if(evb_fsm->getState() != FsmState::Ready && evb_fsm->getState() != FsmState::Error)
            evb_fsm->goState(FsmState::Error);
        return;
    }

//    if(stopEvNum) {
//        fsm->setStopEvNum(stopEvNum);
//        fsm->goState(FsmState::PrepareStopping);
//        emit prepareForStop(stopEvNum);
//    } else
        evb_fsm->goState(FsmState::Stopping);
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &rcSerStat)
{
    if(!progDesc.containsProgramInterface(ProgramInterfaceRemoteControl,0)){
        ProgramInterface remoteControlInterface;
        remoteControlInterface.setType(ProgramInterfaceRemoteControl);
        remoteControlInterface.id = 0;
        progDesc.interfaces.push_back(remoteControlInterface);
    }

    ProgramInterface &remoteInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    remoteInterface.port = rcSerStat.tcpPort;
    remoteInterface.enabled = rcSerStat.listening;
    remoteInterface.isFree = !rcSerStat.portInUse;

    if(rcSerStat.portInUse) {
        remoteInterface.setPeer(rcSerStat.peerHost, rcSerStat.peerPort);
        updateRemoteStatus();
        qInfo() << "Remote control connected.";
    } else {
        qWarning() << "Remote control disconnected.";
        loc_runId = RcRunId();

        if(evb_fsm->getState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            evb_fsm->goState(evb_fsm->getState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }

    if(ui->widgetRunControl->isEnabled() == rcSerStat.portInUse){
        ui->widgetRunControl->setEnabled(!rcSerStat.portInUse);
    }
    updateProgramDescription();
}

void MainWindow::addLogHistory(QString log, Qt::GlobalColor color)
{
//    qDebug()<<QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);
    int repeat=0;
    int now = time(nullptr);
    bool found=false;
    for(int i=0;i<logHistory.size();++i) {
        LogHistorEntry &e = logHistory[i];
        if(log == e.mess){
            found=true;
            repeat = e.repeatCnt++;
            if(now - e.time > 5){
                e.time = now;
                logHistory.move(i,logHistory.size()-1);
            } else
                return;
        }
    }
    if(!found){
        if(logHistory.size() >= (3*clients.size()+3))
            logHistory.removeAt(0);
        LogHistorEntry e;
        e.mess = log;
        e.repeatCnt = 0;
        e.time = now;
        logHistory.append(e);
    }

    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    tf.setForeground(QBrush(color));

    SyslogMessageSender::log(color==Qt::red ? LOG_PRI_WARNING : LOG_PRI_NOTICE, log.toStdString().c_str());
    if(repeat)
        log.append(QString(" (repeat: %1 times)").arg(repeat));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    ui->plainTextEdit->appendPlainText(QString("[%1] %2").arg(QTime::currentTime().toString(), log));
}

void MainWindow::addLog(QtMsgType severity, const QString &log, const QString &thrName)
{
    QString msg = QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);

    Qt::GlobalColor color;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    default:
        color = Qt::black;
    }
    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    ui->plainTextEdit->appendPlainText(QString("[%1] [%2] %3")
                                       .arg(QTime::currentTime().toString(), thrName, log));
}

void MainWindow::on_actionChange_storage_dir_triggered()
{
//    QFileDialog::QFileDialog ( QWidget * parent = 0, const QString & caption = QString(), const QString & directory = QString(), const QString & filter = QString() )
//    QFileDialog dialog(this, "Set predefined storage for mpd data file", "", "*.data");
//    dialog.setOption(QFileDialog::ShowDirsOnly);
//    dialog.exec();
//    qDebug() << dialog.selectedFiles();

    QString dir = QFileDialog::getExistingDirectory( this, "Choose storage directory...",
                                                     config.storageDir, QFileDialog::ShowDirsOnly );
    if(!dir.isEmpty() && config.storageDir != dir){
        config.storageDir = dir;
        emit configChanged(config);
        emit storageChanged(config.storageDir);
    }
}

void MainWindow::on_actionReset_to_defaults_triggered()
{
    config = EvBConfig();
    emit configChanged(config);

    emit storageChanged(config.storageDir);
//    emit allowRemoteControl(config.isRemoteEnabled);
}

void MainWindow::on_actionAdvanced_counters_toggled(bool showColumns)
{
    if(config.advancedCounters != showColumns) {
        config.advancedCounters = showColumns;
        emit configChanged(config);
    }

    QTableWidget *table = ui->tableWidget;
//    table->setColumnHidden(COL_EV_NUM, !showColumns);
//    table->setColumnHidden(COL_ID, !showColumns);
    table->setColumnHidden(COL_TCP_BUFF, !showColumns);
    table->setColumnHidden(COL_EV_BUFF, !showColumns);
    updateTableSize();
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool checked)
{
    if(config.isRemoteEnabled != checked) {
        config.isRemoteEnabled = checked;
        emit configChanged(config);
    }
    emit allowRemoteControl(config.isRemoteEnabled);
}

void MainWindow::data_out_changed(QAction *a)
{
    if(a == ui->actionCombinerIdle) config.combinerOut = COMBINER_IDLE;
    else if(a == ui->actionOutToTcp) config.combinerOut = COMBINER_TCP;
    else if(a == ui->actionOutToFile) config.combinerOut = COMBINER_FILE;
    else {
        qCritical() << "Data out was switch to unknown state.";
    }

    combinerOutUpdated();
    emit configChanged(config);
    updateProgramDescription();
}

void MainWindow::on_actionSet_number_of_outputTcpServers_triggered()
{
    bool ok;
    const int tcpNum = QInputDialog::getInt(this, tr("TCP servers for output"),
                                 tr("Set number of TCP servers for output:"), config.combinerOutTcpNum, 1, 20, 1, &ok);
    if (ok && config.combinerOutTcpNum != tcpNum)
        setOutTcpServerNumber(tcpNum);
}

void MainWindow::on_actionSet_EvB_index_triggered()
{
    bool ok;
    quint32 startNumber = QInputDialog::getInt(this, tr("Event builder index"),
                                 tr("Set 1st event number (EvB index):"), config.startEvNumber, 0, 19, 1, &ok);
    if(!ok)
        startNumber = config.startEvNumber;
    quint32 stepNumber = QInputDialog::getInt(this, tr("EvB event numbet step"),
                                 tr("Set event number step:"), config.evNumberStep, 1, 20, 1, &ok);
    if(!ok)
        startNumber = config.evNumberStep;

    if (config.startEvNumber != startNumber || config.evNumberStep != stepNumber) {
        int evbIndexOld = config.getEvNumIndex();
        config.startEvNumber = startNumber;
        config.evNumberStep = stepNumber;
        emit setEvNumberStep(config.startEvNumber, config.evNumberStep);
        emit configChanged(config);
        if(evbIndexOld != config.getEvNumIndex()){
            QMessageBox msgBox;
            msgBox.setText("Event Builder index was changed, configuration was saved.\n"
                           "Needed manual restart\n"
                           "Closing...");
            msgBox.exec();
            close();
        }
    }
}

void MainWindow::on_actionShow_statistics_triggered()
{
    statisticDialog->show();
}

void MainWindow::on_actionEnable_client_time_distribution_toggled(bool en)
{
    if(config.timeDistrEn != en) {
        config.timeDistrEn = en;
        emit configChanged(config);
    }

    emit enableTimeDistr(config.timeDistrEn);
    QTableWidget *table = ui->tableWidget;
    table->setColumnHidden(COL_TIME_DISTR, !config.timeDistrEn);
    updateTableSize();
}

void MainWindow::on_actionSet_monitor_port_number_triggered()
{
    bool ok;
    const int MIN_PORT_NUMBER = 0;
    const int MAX_PORT_NUMBER = 32999;
    const int portNum = QInputDialog::getInt(this, tr("Set monitor port number"),
                                             QString("Set monitor port number (in range %1-%2):")
                                             .arg(MIN_PORT_NUMBER).arg(MAX_PORT_NUMBER),
                                             config.monitorPort ? config.monitorPort : MAX_PORT_NUMBER, MIN_PORT_NUMBER, MAX_PORT_NUMBER, 1, &ok);
    if (ok && config.monitorPort != portNum) {
        config.monitorPort = portNum;
        emit configChanged(config);
        emit setMonitorPortNum(static_cast<quint16>(portNum));
//        QMessageBox msgBox;
//        msgBox.setText("Monitor port number was changed, configuration was saved.\n"
//                       "Needed manual restart\n"
//                       "Closing...");
//        msgBox.exec();
//        close();
    }
}

void MainWindow::on_actionPrint_FSM_input_triggered()
{
    emit printFsmInput();
}

void MainWindow::setClientReadOut(bool readout)
{
    QTableWidget *table = ui->tableWidget;
    for(int row=0; row<table->rowCount(); ++row){
        if(table->cellWidget(row, COL_READOUT_CB)==reinterpret_cast<QWidget *>(sender())){
//            int clientId = table->item(row, COL_ID)->text().toInt(&ok);
//            qInfo() << QString("client at row=%1 clientId=%2 was %3abled")
//                       .arg(row).arg(clientId).arg(readout?"en":"dis");
//            Q_ASSERT(config.clients.contains(clientId));
            QString type = table->item(row, COL_PROG_TYPE)->text();
            QString index = table->item(row, COL_PROG_INDEX)->text();
            ClientIndex client_index(type, index);
            Q_ASSERT(!client_index.isEmpty());
            Q_ASSERT(config.clients.contains(client_index));
            if(config.clients[client_index].enabled != readout) {
                config.clients[client_index].enabled = readout;
                emit configChanged(config);
                Q_ASSERT(clients.contains(client_index));
                updateTableTextColor(row, client_index);
                updateTablePalette();
            }
            emit updateClientReadout(client_index, readout);
            updateAnalizerList();
            emit configChanged(config);
            return;
        }
    }
}

void MainWindow::gotProgramDescription(const ProgramDescription &newDescr)
{
    if(!EVB_SEARCH_TARGETS.contains(newDescr.getDaqModule()))
       return;

    bool missing = true;

    for (int i=0; i < progDescrs.size(); ++i) {
        ProgramDescription &descr = progDescrs[i];
        if(newDescr.sameProgram(descr)) {
            missing = false;
            if(descr == newDescr)
                return; // Got dublicate
            descr = newDescr;
            break;
        }
    }

    if(missing)
        progDescrs.push_back(newDescr);

    for(ClientIndex index : clients.keys()) {
        Client *client = clients[index];
        if(newDescr.sameProgram(client->getProgDescr())) {
            client->setProgDescription(newDescr);
            const int row = getTableRowByClientIndex(client->getIndex());

            if(row != -1) {
                const QString h = newDescr.hostName.split('.').first();
                if(ui->tableWidget->item(row, COL_PROG_HOST)->text() != h)
                    ui->tableWidget->item(row, COL_PROG_HOST)->setText(h);
            }

            break;
        }
    }
    updatePushButtonState();
}

void MainWindow::gotClosedProgram(const ProgramDescription &newDescr)
{
    for(int i=0; i< progDescrs.size(); ++i) {
        if(progDescrs[i].uuid == newDescr.uuid){
            progDescrs.remove(i);
            break;
        }
    }
    updatePushButtonState();
}

void MainWindow::updateTablePalette()
{
    int row = getSelectedRow();
    if(row == -1)
        return;
    ClientIndex clientId = getClientIndexByRow(row);
    if(clientId.isEmpty())
        return;
    const bool &isConn = lastConnectionState.contains(clientId) ? lastConnectionState[clientId] : false;
    const bool &isOk = isConn || !config.clients[clientId].enabled;

    QPalette p = ui->tableWidget->palette();
    QColor color = isOk ? QColor() : Qt::red;
    if(p.color(QPalette::HighlightedText) != color) {
        p.setColor(QPalette::HighlightedText, color);
        ui->tableWidget->setPalette(p);
    }
}

QString MainWindow::humanReadableSize(double size)
{
    QStringList list;
    list << "KB" << "MB" << "GB" << "TB";

    QStringListIterator i(list);
    QString unit("bytes");

    while(size >= 1024.0 && i.hasNext())
    {
        unit = i.next();
        size /= 1024.0;
    }
    return QString().setNum(size,'f',2)+" "+unit;
}

// --- GARBAGE ---
//QString MainWindow::getClientNameByIndex(ClientIndex index) const
//{
//    if(index)
//        return config.clients.contains(index) ?
//                    config.clients[index].toString() :
//                    QString("Client #%1").arg(index);
//    else
//        return "Combiner";
//}
