#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "Combiner.h"
#include "EvBConfigDb.h"
#include "RunControl.h"

#include "base/app/BaseMainWindow.h"
#include "base/RcRunId.h"
#include "base/ClientIndexEnabled.h"
#include "fsm/FsmState.h"
#include "configuration-manager/ConfigurationManager.h"
#include "pnp-server/PNPServer.h"
#include "EventBuilderFsm.h"

#include <QMainWindow>
#include <QMap>

namespace Ui {
class MainWindow;
}

class Analizer;
class Client;
class ConfigSelector;
class EvBConfigDispatcher;
class EvBStartupOptions;
class MongoConfig;
class PNPServer;
class RemoteControlServer;
class RemoteControlServerStatus;
class StatisticDialog;
class TcpStatistic;
class ClientManagerFsm;

class QElapsedTimer;
class QLabel;
class QSemaphore;
class QThread;

struct LogHistorEntry{
    int time;
    QString mess;
    int repeatCnt;
};

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(const ConfigSelector &cs, QWidget * parent = nullptr);
    void init();
    ~MainWindow() override;

public slots:
    void setConfig(const EvBConfig &newConfig);
    void setClientList(const ClientIndexEnabledMap &) {}
    void remoteStartRequest(const RcRunId &rcRunId, quint64 startEvNumber);
    void remoteStopRequest(const RcRunId &rcRunId, quint64 stopEvNum);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &rcSerStat);
    void gotProgramDescription(const ProgramDescription &newDescr);
    void gotClosedProgram(const ProgramDescription &newDescr);
    void configSelectorChanged(const ConfigIndexName& conf);
    void onEvbFsmStateChanged(const FsmState &cur, const FsmState &prev);
    void onCombinerFsmStateChanged(const FsmState &cur_state);
    void onClientManagerFsmStateChanged(const FsmState &cur_state);
    void startLocalRun();
    void stopLocalRun();

protected slots:
    void addLog(QtMsgType severity, const QString &log, const QString &thrName) override;

private slots:
    void on_pushButtonAdd_clicked();
    void on_pushButtonRemove_clicked();
    void on_tableWidget_itemSelectionChanged();
    void updateClientStatus(ClientIndex index, bool isConnected, QString status);
    void updateTcpStatistic(ClientIndex index, const TcpStatistic &statistic);
    void combinerError(const QString &errStr);
    void deviceError(const ClientIndex &index, const QString &errStr); // from clients
    void updateStatus(const QString&);
    void updateCombinerStats(const CombinerStats&);
    void fileWriterStatsUpdated(const FileWriterStats &fwStat);
    void inputBufferSizeUpdated(ClientIndex index, quint64 size);
    void outputBufferSizeUpdated(quint64 size);

    void addLogHistory(QString log, Qt::GlobalColor color=Qt::black);

    void on_actionChange_storage_dir_triggered();
    void on_actionReset_to_defaults_triggered();
    void on_actionAdvanced_counters_toggled(bool showColumns);
    void on_actionAllow_Remote_Control_toggled(bool checked);
    void data_out_changed(QAction *);
    void on_actionSet_number_of_outputTcpServers_triggered();
    void on_actionSet_EvB_index_triggered();
    void on_actionShow_statistics_triggered();
    void on_actionEnable_client_time_distribution_toggled(bool en);
    void on_actionSet_monitor_port_number_triggered();
    void on_actionPrint_FSM_input_triggered();
    void setClientReadOut(bool readout);

private slots:
    void updateTablePalette();
//    bool clientsDestroying(int client_id, int timeout, QElapsedTimer *timer);
    void on_actionConfiguration_manager_triggered();

signals:
    void configChanged(const EvBConfig &config) const;
    void clientListChanged(const ClientIndexEnabledMap &clients_map) const;
    void switch_to_config(const ConfigIndexName& conf) const;
    void stoping() const;
    void starting(const RcRunId &runId) const;
    void destroyClient(const ClientIndex &id) const;
    void destroyAll() const;
    void sendRemoteState(const QString &state);
    void sendRunError(const RcRunId &runId, QString errStr);
    void sendRunWarning(const RcRunId &runId, QString warnStr);
    void sendBackPreasure(const RcRunId &runId, bool backPreasure);
    void storageChanged(const QString &storage);
    void updateClientReadout(ClientIndex index, bool readout);
    void analizerListUpdated(ClientIndexSet clients, QVector<Analizer *> list);
    void setOutput(COMBINER_OUT combinerOut, int tcpNum);
    void setEvNumberStep(quint32 startEvNum, quint32 step);
    void allowRemoteControl(bool);
    void changeProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void enableTimeDistr(bool en) const;
    void resizeOutputBuffer(int outputBufSize) const;
    void setMonitorPortNum(quint16 portNum);
    void config_transfer_to_rc(QByteArray data) const;
    void prepareForStop(quint32 stopEvNum) const;
    void printFsmInput() const;

private:
    void initCombiner();
    void update_combiner_to_config();
    void updateOutTcpLabelList();
    void setupTable();
    void setup_clients();
    bool addClient(const EvBClientConfig &client_config);
    void makeConnections(Client *cm) const;
    void addTableRow(const EvBClientConfig &cc) const;
    void updatePushButtonState();
    void updateTableSize();
    void updateAnalizerList();
    void updateRemoteStatus();
    void initProgramDescription(const ConfigIndexName &conf);
    void updateProgramDescription();
    void updateTcpPortLabels();
    void setOutTcpServerNumber(int tcpNum);
    QVector<ProgramDescription> getFreeProgs() const;
    bool isFreeProg(const ProgramDescription &pd) const;
    int getTableRowByClientIndex(const ClientIndex &index) const;
    int getSelectedRow() const;
    ClientIndex getClientIndexByRow(int row) const;
    void updateTableTextColor(int row, ClientIndex index = ClientIndex()) const;
    void combinerOutUpdated();
    QLabel *getTcpLabel(int index) const;
    QString humanReadableSize(double size);

    Ui::MainWindow *ui;
    ConfigurationManager * const configuration_manager;
    EvBConfig config;
    EvBConfigDb * const configDb;
    EvBConfigDispatcher * const configDispatcher;
    EventBuilderFsm * const evb_fsm;
    ClientManagerFsm * const client_manager_fsm;
    QMap<ClientIndex, Client*> clients;
    QMap<ClientIndex, bool> lastConnectionState; // by clientId
    Combiner *combiner;
    QThread *combinerThread;
    bool allOk;
    RemoteControlServer * const remoteControl;
    PNPServer * const pnpServer;
    ProgramDescription progDesc;
    bool run;
    RcRunId loc_runId;
    QList<LogHistorEntry> logHistory;
    QLabel * const labelFileErrSize;
    QVector<ProgramDescription> progDescrs;
    CombinerStats lastCombinerStats;
    QElapsedTimer * const clientDistrTime;
    uint enabledClientNum;
    StatisticDialog * const statisticDialog;
    bool backPreasure;
    QLabel * const labelEvbFsmState;
    QLabel * const labelCombinerFsmState;
    QLabel * const labelClientManagerFsmState;
    QLabel * const labelCombinerStatus;
};

#endif // MAINWINDOW_H
