#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "DeviceConfig.h"
#include "DeviceWidget.h"
#include "MscConfig.h"
#include "MscConfigDb.h"
#include "MscConfigDispatcher.h"
#include "MscDevice.h"
#include "base/DeviceIndexEnabled.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "mldiscover/DeviceDescription.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mregdev/device_id.h"
#include "pnp-server/PNPServer.h"

#include <QDebug>
#include <QFileDialog>
#include <QInputDialog>
#include <QLabel>
#include <QMessageBox>
#include <QPlainTextEdit>
#include <QScrollBar>
#include <QShortcut>
#include <QSpacerItem>
#include <QSplitter>
#include <QTableWidgetItem>
#include <QTextCharFormat>
#include <QThread>
#include <QTime>
#include <QTimer>

static const int DataRoleId = Qt::UserRole;
static const int DataRoleSerial = Qt::UserRole+1;

enum { COL_SER_TXT, COL_FW, COL_TEMP, COL_STATUS, COL_TOTAL};

MainWindow::MainWindow(const ConfigSelector &cs, QWidget *parent) :
    BaseMainWindow(parent),
    ui(new Ui::MainWindow),
    configDb(new MscConfigDb(cs, parent)),
    configDispatcher(new MscConfigDispatcher(this)),
    configuration_manager(new ConfigurationManager(this, cs.getType())),
    discoverDialog(new DiscoverDialog(this)),
    deviceDispatcher(new DeviceDispatcher(this)),
    pnp_server(new PNPServer(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    qRegisterMetaType<DeviceStatus>();

    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<DeviceConfig>();

    configDispatcher->subscribe(configDb);
    configDispatcher->subscribe(this);

    configDispatcher->subscribeCS(this);
    configDispatcher->subscribeCS(configuration_manager);
    configDispatcher->subscribeCS(configDb);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter({DEVICE_ID_MSC16VE_TM});
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(discoverDialog);

    deviceDispatcher->subscribeDevices(discoverDialog);
    deviceDispatcher->subscribeDevices(this);

    init();
}

MainWindow::~MainWindow()
{
    emit delete_all();
    delete configDb;
    delete ui;
}

void MainWindow::init()
{
    init_gui();

    configuration_manager->setDefaultConfig(config);

    connect(this, &MainWindow::change_program_description,
            pnp_server, &PNPServer::changeProgramDescription);
    connect(this, SIGNAL(delete_all()), pnp_server, SLOT(deleteLater()));

    // discover_dialog->setSingleMode(false);

    update_program_description();

    DeviceIndexEnabledMap device_list;
    for (auto it = config.device_configs.begin(); it != config.device_configs.end(); ++it) {
        device_list.insert(it.key(), DeviceIndexEnabled(it.key(), it.value().enabled));
    }
    discoverDialog->setDeviceList(device_list);
}

void MainWindow::init_gui()
{
    set_window_title();

    ui->tableWidgetFindDevice->horizontalHeader()->setStretchLastSection(false);

    //--- log
    QPalette Pal(palette());
    Pal.setColor(QPalette::Background, Qt::gray);
    ui->widgetLogButton->setAutoFillBackground(true);
    ui->widgetLogButton->setPalette(Pal);
    ui->pushButtonLog->setChecked(config.show_log);
    ui->pushButtonLog->setText(config.show_log ? "Hide log" : "Show log");
    ui->plainTextEditLog->setHidden(!config.show_log);


    //--- splitter
    QSplitter *splitter = ui->split_horizontal;

    int index_find_device = splitter->indexOf(ui->tableWidgetFindDevice);
    int index_main = splitter->indexOf(ui->widgetMain);
    int index_log = splitter->indexOf(ui->plainTextEditLog);
    int index_log_button = splitter->indexOf(ui->widgetLogButton);

    splitter->setCollapsible(index_find_device, true);
    splitter->setCollapsible(index_log_button, true);

    splitter->setStretchFactor(index_find_device, 1);
    splitter->setStretchFactor(index_main, 30);
    splitter->setStretchFactor(index_log, 1);
    splitter->setStretchFactor(index_log_button, 1);

    populate_devices_table();
}

void MainWindow::setConfig(const MscConfig &newConfig)
{
    if (newConfig == config)
        return;
    config = newConfig;
    init_gui();
    DeviceIndexEnabledMap devicesAll;

    for (const auto &dc: config.device_configs)
        devicesAll[dc.device_index] = DeviceIndexEnabled(dc.device_index, dc.enabled);

    emit deviceListChanged(devicesAll);
}

void MainWindow::create_device_widget(DeviceConfig device_config)
{
    DeviceWidget *device_widget = new DeviceWidget(device_config, deviceDispatcher, this);
    connect(device_widget, SIGNAL(update_device_config(DeviceConfig)), this, SLOT(update_device_config(DeviceConfig)));
    connect(device_widget, SIGNAL(update_status(DeviceStatus)), this, SLOT(update_status(DeviceStatus)));

    connect(ui->actionFind_Baseline, SIGNAL(triggered()), device_widget, SIGNAL(find_base_line()));
    connect(ui->actionSoft_clear, SIGNAL(triggered()), device_widget, SIGNAL(soft_clear()));
    device_widgets[device_config.device_index] = device_widget;
    ui->layout_main->addWidget(device_widget);
}

void MainWindow::populate_devices_table()
{
    QTableWidget *table = ui->tableWidgetFindDevice;

    table->setColumnCount(COL_TOTAL);

    table->setHorizontalHeaderItem(COL_SER_TXT, new QTableWidgetItem("Serial"));
    table->setHorizontalHeaderItem(COL_FW, new QTableWidgetItem("f/w"));
    table->setHorizontalHeaderItem(COL_TEMP, new QTableWidgetItem("Temp"));
    table->setHorizontalHeaderItem(COL_STATUS, new QTableWidgetItem("Status"));

    table->setRowCount(0);

    int row_count=0;
    for(DeviceIndex device_index : device_widgets.keys()){
        if(config.device_configs[device_index].enabled){
            ++row_count;
        }
    }

    table->setRowCount(row_count);
    int row_numder = 0;
    for(DeviceIndex device_index : device_widgets.keys()){
        DeviceConfig device_config = device_widgets[device_index]->get_device_config();

        QTableWidgetItem *serial_item = new QTableWidgetItem(device_config.device_index.getSerialStr());
        serial_item->setData(DataRoleId,     device_config.device_index.getDevId());
        serial_item->setData(DataRoleSerial, device_config.device_index.getSerial());
        table->setItem(row_numder, COL_SER_TXT, serial_item);
        table->setItem(row_numder, COL_FW, new QTableWidgetItem());
        table->setItem(row_numder, COL_TEMP, new QTableWidgetItem());
        table->setItem(row_numder, COL_STATUS, new QTableWidgetItem("disconnected"));

        table->item(row_numder, COL_STATUS)->setTextColor(Qt::red);
        table->item(row_numder, COL_SER_TXT)->setTextColor(Qt::red);
        ++row_numder;
    }
    table->resizeColumnsToContents();
    table->resizeRowsToContents();

    update_table_size(table);
}

void MainWindow::setDeviceList(const DeviceIndexEnabledMap &map)
{
    const auto old_key_list = config.device_configs.keys();
    const auto new_key_list = map.keys();
    const auto old_indices = makeDeviceIndexSet(old_key_list);
    const auto new_indices = makeDeviceIndexSet(new_key_list);
    const auto toRemove = old_indices - new_indices;
    const auto toAdd = new_indices - old_indices;

    // remove old config
    for (auto i : toRemove)
        config.device_configs.remove(i);

    // add new config
    for (auto i : toAdd) {
        config.device_configs.insert(i, DeviceConfig(i));
    }

    // update enable state
    for (auto i : new_indices) {
        config.device_configs[i].enabled = map[i].enabled;
    }

    // remove widget config
    for (auto i : toRemove) {
        if (!config.device_configs.contains(i) || !config.device_configs[i].enabled) {
            delete  device_widgets[i];
            device_widgets.remove(i);
        }
    }

    // add widget config
    for (auto i : toAdd) {
        if (!device_widgets.contains(i) && config.device_configs[i].enabled) {
            create_device_widget(config.device_configs[i]);
        }
    }

    populate_devices_table();
    emit configChanged(config);
}

void MainWindow::update_device_config(const DeviceConfig &device_config)
{
    config.device_configs[device_config.device_index] = device_config;
    emit configChanged(config);
}

void MainWindow::update_status(const DeviceStatus &device_status)
{
    QTableWidget *table = ui->tableWidgetFindDevice;
    int row = get_device_row(device_status.index);
    if(row == -1){
        return;
    }

    table->item(row, COL_FW)->setText(device_status.fwStr);
    table->item(row, COL_STATUS)->setText(device_status.status);
    table->item(row, COL_STATUS)->setTextColor(device_status.connected ? QColor() : Qt::red);
    table->item(row, COL_SER_TXT)->setTextColor(device_status.connected ? Qt::darkGreen : Qt::red);
    table->item(row, COL_TEMP)->setText(QString::number(device_status.temp, 'f', 1)
                                        .append(QChar(0x02da)).append('C'));
    table->resizeColumnsToContents();
    update_table_size(table);
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void MainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    const auto &cs = ConfigSelector(DaqModule(DaqModule::Type::Msc16VeTm), conf);
    const auto &title = cs.toWindowTitle();
    if(title != windowTitle())
        setWindowTitle(title);
}

void MainWindow::on_actionFind_device_triggered()
{
    discoverDialog->show();
}

void MainWindow::on_actionSet_hit_width_extension_triggered()
{
    bool ok;
    int val = QInputDialog::getInt(this, "Set hit width", "Hit extension in clocks:", config.hit_extension, 0, 16, 1, &ok);
    if(ok) {
        config.hit_extension = val;
        emit configChanged(config);
    }
}

void MainWindow::on_actionLut_c_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile("lut.html");
    QString html;

    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
    } else
        html = "<b>There is no lut.c example.</b>";

    QMessageBox::about (this, "lut.c", html);
}

void MainWindow::on_pushButtonLog_clicked(bool checked)
{
    ui->plainTextEditLog->setHidden(!checked);
    config.show_log = checked;
    if(checked){
        ui->pushButtonLog->setText("Hide log");
    }
    else{
        ui->pushButtonLog->setText("Show log");
    }
    ui->plainTextEditLog->setHidden(!config.show_log);
    emit configChanged(config);
}

int MainWindow::get_device_row(const DeviceIndex &index)
{
    for(int row=0; row<ui->tableWidgetFindDevice->rowCount(); ++row){
        auto i = get_index_by_row(row);
        if(index == i){
            return row;
        }
    }
    return -1;
}

DeviceIndex MainWindow::get_index_by_row(int row)
{
    auto item = ui->tableWidgetFindDevice->item(row, COL_SER_TXT);
    return DeviceIndex(static_cast<DeviceId>(item->data(DataRoleId).toUInt()),
                       item->data(DataRoleSerial).toULongLong());
}

void MainWindow::update_program_description()
{
    program_description.options.clear();
    program_description.type = configDb->getType().toString();
    program_description.index = configDb->getIndex();
    emit change_program_description(program_description);
}

void MainWindow::add_log(QtMsgType severity, const QString &log, const QString &thrName)
{
    Qt::GlobalColor color = Qt::black;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    case QtFatalMsg:
        qInfo()<<"";
    }
    QTextCharFormat tf = ui->plainTextEditLog->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEditLog->setCurrentCharFormat(tf);
    QString str = log;
    if(!thrName.isEmpty())
        str.prepend(QString("[%1] ").arg(thrName));
    str.prepend(QString("[%1] ").arg(QTime::currentTime().toString()));
    ui->plainTextEditLog->appendPlainText(str);
}

void MainWindow::set_window_title()
{
    setWindowTitle(configDb->getWindowTitle());
}

void MainWindow::update_table_size(QTableWidget *t)
{
    int width=t->verticalHeader()->width();
    for(int i=0;i<t->columnCount();++i){
        if(t->isColumnHidden(i)) continue;
        width += t->columnWidth(i);
    }
    width += 5;
    if(t->verticalScrollBar()->isVisible())
        width += t->verticalScrollBar()->width();
    t->setMinimumWidth(width);
    t->updateGeometry();

    int height=t->horizontalHeader()->height();
    for(int i=0;i<t->rowCount();++i){
        if(t->isRowHidden(i)) continue;
        height += t->rowHeight(i);
    }
    height += 5;
    if(t->horizontalScrollBar()->isVisible())
        height += t->horizontalScrollBar()->height();
    t->setMinimumHeight(height);
}
