//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "RcConfigDb.h"
#include "RcConfigDispatcher.h"
#include "RcCoreApp.h"
#include "RcSignalRouter.h"
#include "SlowControlServer.h"
#include "TrigCore.h"
#include "TrigDevice.h"
#include "TrigDeviceController.h"
#include "VmeTrigsCore.h"
#include "VmeTrigsController.h"
#include "base/SyslogMessageSender.h"
#include "base/ThreadController.h"
#include "base/config/ConfigSelector.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mongo/StartupOptions.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "run-control-client-manager/ClientManagerWidget.h"
#include "run-control-client-manager/RunStates.h"

#include <QDateTime>
#include <QDebug>
#include <QDesktopWidget>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QHostAddress>
#include <QJsonDocument>
#include <QJsonObject>
#include <QLabel>
#include <QMessageBox>
#include <QSettings>
#include <QShortcut>
#include <QStatusBar>
#include <QStyle>
#include <QTabBar>
#include <QThread>
#include <QTimer>

Q_DECLARE_METATYPE(Qt::GlobalColor)
Q_DECLARE_METATYPE(TrigStatus)

static const char *styleLargeFont = "font-size: 18pt";

MainWindow::MainWindow(const ConfigSelector &cs, QWidget *parent) :
    BaseMainWindow(parent),
    ui(new Ui::MainWindow),
    labelFsmStateRc(new QLabel(this)),
    labelFsmStateTrig(new QLabel(this)),
    labelFsmStateClient(new QLabel(this)),
    labelFsmStateVme(new QLabel(this)),
    labelDevState(new QLabel(this)),
    labelSlowControl(new QLabel(this)),
    labelDevice(new QLabel(this)),
    labelFirmware(new QLabel(this)),
    labelTemp(new QLabel(this)),
    discoverDialog(new DiscoverDialog(this)),
    rcCoreApp(new RcCoreApp(this)),
    configDb(new RcConfigDb(cs, this)),
    remoteControlStatusCache(new RemoteControlServerStatus)
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    ui->tabWidget->setCurrentIndex(0); // always start with main controls view

    discoverDialog->setSingleMode(true);
    discoverDialog->setObjectName("MainDiscoverDialog");

    qRegisterMetaType<RUN_STATES>();
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<TrigStatus>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<RcTrigConfig>();

    setWindowTitle(configDb->getWindowTitle());

    // setup VmeTrigsWidget
    auto vmeTrigsWidget = ui->vmeTrigsWidget;
    auto vmeTrigsCore = rcCoreApp->getVmeTrigsCore();
    auto vmeTrigsController = vmeTrigsCore->getVmeTrigsController();
    auto vmeDeviceDispatcher = vmeTrigsCore->getDeviceDispatcher();
    vmeTrigsWidget->setController(vmeTrigsController);
    vmeTrigsWidget->setDeviceDispatcher(vmeDeviceDispatcher);

    auto rcConfigDispatcher = rcCoreApp->getConfigDispatcher();
    ui->trigMainForm->setConfigDispatcher(rcConfigDispatcher);

    auto trigCore = rcCoreApp->getTrigCore();
    auto trigDeviceController = trigCore->getTrigDeviceController();
    auto trigDeviceDispatcher = trigCore->getDeviceDispatcher();
    trigDeviceDispatcher->subscribeDevices(discoverDialog);
    trigDeviceDispatcher->subscribeDevices(this);

    LldpDiscovery::instance().subscribe(discoverDialog, trigDeviceController->getDeiveIdFilter());

    // broadcast SlaveMessage via RcCoreApp
    connect(this, &MainWindow::sendSlaveMessage,
            rcCoreApp, &RcCoreApp::sendSlaveMessage);

    connect(rcCoreApp, &RcCoreApp::sendSlaveMessage,
            this, &MainWindow::slaveMessageReceived);

    init(cs);
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    discoverDialog->close();

    rcCoreApp->quit();

//    if(trigDevice){
//        emit deleteHardware(DeviceIndex(trigDevice->getDeviceId(), trigDevice->getDeviceSerialNumber()));
//    }

    emit destroyAll();
    BaseMainWindow::closeEvent(event);
}

MainWindow::~MainWindow()
{
    delete ui->trigMainForm;
    delete ui->clientManagerWidget;
    delete ui->vmeTrigsWidget;
    delete ui;
    delete remoteControlStatusCache;
}

void MainWindow::init(const ConfigSelector &cs)
{
    qInfo() << QString("Program index is: '%1'").arg(configDb->getIndex());

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.rc." + shortHostName + "." + configDb->getIndex());

    ui->groupBoxWarn->hide();
    ui->labelWarTrigDev->hide();
    ui->labelWarClients->hide();
    ui->labelWarTmwr->hide();
    ui->labelWarnTrigDev->hide();
    ui->labelWarnClients->hide();
    ui->labelWarnTmwr->hide();
    ui->labelWarnTrigDev->setStyleSheet("QLabel { color : orange;}");
    ui->labelWarnClients->setStyleSheet("QLabel { color : orange;}");
    ui->labelWarnTmwr->setStyleSheet("QLabel { color : orange;}");

    ui->mainToolBar->hide();

    labelFsmStateRc->setMinimumWidth(150);
    labelFsmStateRc->setStyleSheet("font-weight: bold;");
    labelFsmStateClient->setMinimumWidth(150);
    labelFsmStateTrig->setMinimumWidth(150);
    labelFsmStateVme->setMinimumWidth(150);

    statusBar()->addWidget(labelFsmStateRc);
    statusBar()->addWidget(labelFsmStateTrig);
    statusBar()->addWidget(labelFsmStateClient);
    statusBar()->addWidget(labelFsmStateVme);
    statusBar()->addPermanentWidget(labelSlowControl);
    statusBar()->addPermanentWidget(labelDevState);
    statusBar()->addPermanentWidget(labelDevice);
    statusBar()->addPermanentWidget(labelTemp);
    statusBar()->addPermanentWidget(labelFirmware);

//    ui->groupBoxRunControl->setStyleSheet();
    ui->pushButtonRestartAll->setStyleSheet(styleLargeFont);
    ui->pushButtonStopAll->setStyleSheet(styleLargeFont);
    ui->pushButtonRestartAll->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButtonStopAll->setIcon(style()->standardIcon(QStyle::SP_MediaStop));
//  labelState->setStyleSheet("QLabel {color: red}");

    connect(ui->actionCheck_LVDS_parity, &QAction::toggled,
            ui->trigMainForm, &TrigMainForm::checkLvdsParityToggled);

    configuration_manager = new ConfigurationManager(this, cs.getType());

    connect(configuration_manager.data(), &ConfigurationManager::switch_to_config,
            this, &MainWindow::configSelectorChanged,
            Qt::UniqueConnection);

    auto &router = RcSignalRouter::instance();
    router.addClientManagerWidget(ui->clientManagerWidget);
    router.addTrigMainForm(ui->trigMainForm);
    router.addMainWindow(this);
    router.addConfigurationManager(configuration_manager);
    router.addRcConfigDb(configDb);
    router.connectAll();

    configuration_manager->setDefaultConfig(RcConfig());

    loadSettings();

//    if(config.rcTrigConfig.deviceIndex) {
//        quint32 ipAddr = 0;
//        if(!config.rcTrigConfig.hostStr.isEmpty()){
//            QHostAddress host(config.rcTrigConfig.hostStr);
//            if(!host.isNull())
//                ipAddr = host.toIPv4Address();
//        }
//        DeviceDescription dd(config.rcTrigConfig.rcDeviceId,
//                             config.rcTrigConfig.deviceIndex,
//                             ipAddr);
//        connectToTrig(dd);
//    }
}

void MainWindow::setConfig(const RcConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;
    // update GUI elements that depend on config
    updateStatStopButton();
    loadSettings();
}

void MainWindow::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    updateDeviceLabel();
}

static QString map_fsm_to_ui_style(const FsmState &state)
{
    QString color;
    switch (state) {
    case FsmState::Init:;
    case FsmState::Error:
        color = "red";
        break;
    case FsmState::Ready:;
        break;
    case FsmState::Run:;
        color = "lightgreen";
        break;
    case FsmState::Starting:;
    case FsmState::Stopping:
        color = "orange";
        break;
    default:
    {}
    }
    return QString("QLabel { %1 font-size: 14pt}")
            .arg(color.isNull() ? "" : "background-color:" + color + ";");
};

void MainWindow::logRcFsmStateChange(const FsmState &previousState, const FsmState &state)
{
    auto RunIdStr = QString(config.pc.runId);
    auto previousStateStr = fsmStateToString(previousState);

    ui->labelRunState->setStyleSheet(map_fsm_to_ui_style(state));

    if (state == FsmState::Starting) {
        ui->labelRunState->setText(QString("Starting Run %1...").arg(RunIdStr));
        ui->plainTextEditLog->appendPlainText("");
    }

    if (state == FsmState::Stopping) {
        ui->labelRunState->setText(QString("Stopping Run %1...").arg(RunIdStr));
    }

    if (state == FsmState::Run) {
        QString txt = QString("Run started: %1").arg(RunIdStr);
        ui->labelRunState->setText(txt);
    }
    if (state == FsmState::Ready) {
        ui->labelRunState->setText(QString("Previous Run %1").arg(RunIdStr));
    }
    if (state == FsmState::Error) {
        switch (previousState) {
        case FsmState::Starting:;
        case FsmState::Stopping:;
        case FsmState::Run: {
            QString txt = QString("%1 '%2' failed").arg(previousStateStr, RunIdStr);
            ui->labelRunState->setText(txt);
            break;
        }
        default:;
        }
    }
}

void MainWindow::onRcFsmStateChange(const FsmState &state, const FsmState &previousState)
{
    logRcFsmStateChange(curRcFsmState, state);
    if(state == FsmState::Starting) {
    // run has been started
        ui->trigMainForm->clearStats();
    }
    curRcFsmState = state;

    labelFsmStateRc->setText("RcFsm: " + fsmStateToString(state));
    updateStatStopButton();
    blockGuiControl(state == FsmState::Starting ||
                    state == FsmState::Stopping);
}

void MainWindow::trigManagerRunStateChanged(const FsmState &state)
{
    subFsmStateChanged(TAB_TRIG_DEV, state);
}

void MainWindow::clientManagerRunStateChanged(const FsmState &state)
{
    subFsmStateChanged(TAB_CLIENTS, state);
}

void MainWindow::vmeTrigsRunStateChanged(const FsmState &state)
{
    subFsmStateChanged(TAB_VME_TRIGS, state);
}

void MainWindow::client_config_receiver(QMap<int, ClientConfig> client_config_)
{
    config.clients = client_config_;
    emit configChanged(config);
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = remoteControlStatusCache->portInUse != status.portInUse;
    *remoteControlStatusCache = status;
    if(changed)
        updateStatStopButton();
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void MainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    const ConfigSelector cs(DaqModule(DaqModule::Type::RunControl), conf);
    setWindowTitle(cs.toWindowTitle());
}

void MainWindow::clearRunWarning()
{
    ui->labelWarClients->hide();
    ui->labelWarnClients->clear();
    ui->labelWarnClients->hide();

    ui->labelWarTrigDev->hide();
    ui->labelWarnTrigDev->hide();
    ui->labelWarnTrigDev->clear();

    ui->labelWarTmwr->hide();
    ui->labelWarnTmwr->hide();
    ui->labelWarnTmwr->clear();
    updateGroupBoxWarn();
}

void MainWindow::onSlowCntStatusUpdated(quint16 tcpPort, bool listening, bool portInUse)
{
    QString txt = "SlowControl: ";

    if(listening)
        if(portInUse)
            txt.append("connected");
        else
            txt.append("disconnected");
    else
        txt.append("disabled");

    labelSlowControl->setStyleSheet(QString("QLabel { color : %1; }")
                                    .arg(listening ? (portInUse?"green":"red") : "gray"));

    if(tcpPort)
        labelSlowControl->setToolTip(QString("Listening on %1 port").arg(tcpPort));
    labelSlowControl->setText(txt);
}

void MainWindow::subFsmStateChanged(int tabIndex, const FsmState &state)
{
    const bool isOk = state!=FsmState::Error;
    setTabState(state!=FsmState::Error, tabIndex);
    QLabel *label = nullptr;
    QString prefix;
    switch (tabIndex) {
    case TAB_TRIG_DEV:
        label = labelFsmStateTrig;
        prefix = "TrigFsm";
        break;
    case TAB_CLIENTS:
        label = labelFsmStateClient;
        prefix = "ClientFsm";
        break;
    case TAB_VME_TRIGS:
        label = labelFsmStateVme;
        prefix = "VmeTrigFsm";
        break;
    default:;
    }
    if(label) {
        label->setText(prefix + ": " + fsmStateToString(state));
        auto palette = label->palette();
        palette.setColor(label->foregroundRole(), isOk ? QColor() : Qt::red);
        label->setPalette(palette);
    }
}

void MainWindow::setTabState(bool isOk, int tabIndex)
{
    QTabBar *bar = ui->tabWidget->tabBar();
    const auto curTabCol = bar->tabTextColor(tabIndex);
    if(curTabCol.isValid() == isOk)
        bar->setTabTextColor(tabIndex, isOk ? QColor() : Qt::red);
}

void MainWindow::blockGuiControl(bool block)
{
    const bool enabled = !block;
    emit sendControlEnable(enabled);
    ui->menuOptions->setEnabled(enabled);
}

/*
bool MainWindow::waitVmeTrigsStarted(const QElapsedTimer &startTime)
{
    bool result = true;
    // Get started state from Vme Modules
    while (! startTime.hasExpired(START_TIMEOUT)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if (vmeTrigsController()->get_runId()==config.pc.runId &&
                vmeTrigsController()->getRunState() != RS_STARTING )
            // Break on error or started
            break;
    }
    if (vmeTrigsController()->getRunState() != RS_STARTED){
        qCritical() << "Some vme trigger failed to reset EvNumber";
        result = false;
    }
    return result;
}
*/
/*
bool MainWindow::waitClientsStarted(const QElapsedTimer &startTime)
{
    bool result = true;
    // Get started state from Tcp Clients (sub-programms)
    while (! startTime.hasExpired(START_TIMEOUT)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if(ui->clientManagerWidget->get_run_id()==config.pc.runId &&
                ui->clientManagerWidget->getRunState() != RS_STARTING)
            // Break on error or started
            break;
    }
    auto sttt = ui->clientManagerWidget->getRunState();
    if(sttt != RS_STARTED){
        qDebug() << "Some client failed to start";
        result = false;
    }
    return result;
}
*/


void MainWindow::updateDeviceLabel()
{
    auto deviceIndex = config.rcTrigConfig.deviceIndex;
    bool devicePresent = ! deviceIndex.isNull();
    if (devicePresent) {
        QString devLabelStr;
        const QString serialStr = deviceIndex.getSerialStr();
        devLabelStr = QString("%1")
            .arg(deviceIndex.getIdent());
//        if(config.rcTrigConfig.hostStr.isEmpty())
//            devLabelStr.append(QString(", IP %3").arg(config.rcTrigConfig.hostStr));
//        else
//            devLabelStr.append(", no IP");
        labelDevice->setText(devLabelStr);
    } else {
        labelDevice->clear();
        labelDevState->setText("Device status: unknown");
        updateDeviceStatusLabel(TrigStatus());
    }
}

void MainWindow::updateDeviceStatusLabel(const TrigStatus &status)
{
    labelTemp->setText(QString::number(status.temp, 'f', 1).append(QChar(0x02da)).append('C'));
    labelFirmware->setText(QString("f/w:%1").arg(status.fwStr));
    labelDevice->setStyleSheet(status.connected ? "" : "background-color : red");
}

void MainWindow::trigStatusUpdated(const TrigStatus &status)
{
    updateDeviceStatusLabel(status);
    if(ui->actionCheck_LVDS_parity->isEnabled() != status.supportLvdsParity)
        ui->actionCheck_LVDS_parity->setEnabled(status.supportLvdsParity);

    static bool oldStatusConnected = false;
    if (status.connected != oldStatusConnected) {
        oldStatusConnected = status.connected;
        updateDeviceLabel();
    }
}

void MainWindow::setTrigStatus(const DeviceIndex &, QString status)
{
    labelDevState->setText("Device status: " + status);
}

SlaveType MainWindow::mapSenderToSlaveType(QObject *sender)
{
    if (sender == ui->clientManagerWidget) {
        return SlaveType::TcpClient;
    } else if(sender == ui->trigMainForm) {
        return SlaveType::Trig;
    } else if(sender == ui->vmeTrigsWidget) {
        return SlaveType::VmeTrig;
    } else {
        Q_ASSERT(false);
        return SlaveType::Unknown;
    }
}

void MainWindow::onRunWarning(const RcRunId &runId, QString warnStr)
{
    const auto slaveType = mapSenderToSlaveType(sender());
    const MessageTreeItem m(RcSlaveId(slaveType),
                           runId,
                           SeverityMessage(SlaveMessageSeverity::Warning, warnStr));
    emit sendSlaveMessage(m);
}

void MainWindow::onRunError(const RcRunId &runId, QString warnStr)
{
    const auto slaveType = mapSenderToSlaveType(sender());
    const MessageTreeItem m(RcSlaveId(slaveType),
                           runId,
                           SeverityMessage(SlaveMessageSeverity::Error, warnStr));
    emit sendSlaveMessage(m);
}

void MainWindow::slaveMessageReceived(const MessageTreeItem &m)
{
    if (!m.isEmpty() &&
            (!config.pc.runEnable || config.pc.runId != m.runId) )
        return;

    QLabel *lab=nullptr;
    QLabel *label=nullptr;
    switch (m.type()) {
    case SlaveType::TcpClient:
        lab = ui->labelWarClients;
        label = ui->labelWarnClients;
        break;
    case SlaveType::Trig:
        lab = ui->labelWarTrigDev;
        label = ui->labelWarnTrigDev;
        break;
    case SlaveType::VmeTrig:
        lab = ui->labelWarTmwr;
        label = ui->labelWarnTmwr;
        break;
    default:
        return;
    }

    QString warnStr = m.text();
    const int WARN_SIZE_LIMIT= 90;
    if(warnStr.size()>WARN_SIZE_LIMIT) {
        warnStr = warnStr.left(WARN_SIZE_LIMIT-3)+"...";
    }
    Q_ASSERT(label);
    Q_ASSERT(lab);
    if (label && lab) {
        label->setText(warnStr);
        label->setHidden(warnStr.isEmpty());
        lab->setHidden(label->isHidden());
    }

    updateGroupBoxWarn();
}

void MainWindow::addLogMessage(QtMsgType severity, const QString &category, const QString &log, const QString &thrName)
{
    if (!ui)
        return;
    Qt::GlobalColor color = Qt::black;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    case QtFatalMsg:
        qInfo()<<"";
    }

    if((ui->tabWidget->tabBar()->currentIndex() != TAB_LOGS)
            && (severity==QtCriticalMsg || severity==QtWarningMsg)) {
        setTabState(false, TAB_LOGS);
    }

    QTextCharFormat tf = ui->plainTextEditLog->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEditLog->setCurrentCharFormat(tf);
    QString str = log;
    if (!category.isEmpty() && category != "default")
        str.prepend(QString("%1: ").arg(category));
    if(!thrName.isEmpty())
        str.prepend(QString("[%1] ").arg(thrName));
    str.prepend(QString("[%1] ").arg(QTime::currentTime().toString()));
    ui->plainTextEditLog->appendPlainText(str);
//    ui->plainTextEditLog->appendPlainText(QString("[%1] [%2] %3")
//                                          .arg(QTime::currentTime().toString(), thrName, log));
}

void MainWindow::loadSettings()
{
    updateDeviceLabel();

    ui->clientManagerWidget->init(config.clients);

//    ui->trigMainForm->init(&config);
//    ui->vmeTrigsWidget->init(&config);

    bool oldState1 = ui->actionAllow_Remote_Control->blockSignals(true);
    ui->actionAllow_Remote_Control->setChecked(config.pc.allowRemoteControl);
    ui->actionAllow_Remote_Control->blockSignals(oldState1);
    ui->actionEnableStatusDump->setChecked(config.pc.enableStatusDump);
    ui->actionCheck_LVDS_parity->setChecked(config.rcTrigConfig.checkLvdsParity);
    emit setStatusDumpEnable(config.pc.enableStatusDump);
    emit setStatusDumpDirName(config.pc.statusDumpDirName);

    if(configDb->getIndex() == "Glob") { // FIXME: hardcoded data
        emit allowSlowControl(true);
    } else {
#ifdef QT_DEBUG
//        emit allowSlowControl(true);
#endif
    }
//    updateProgramDescription();
}

void MainWindow::updateStatStopButton()
{
    bool enableStart = curRcFsmState == FsmState::Ready;
    bool enableStop = curRcFsmState == FsmState::Run;
    bool underRemote = remoteControlStatusCache->portInUse;
//    bool isRunning = config.pc.runEnable;
    ui->pushButtonRestartAll->setEnabled(enableStart && !underRemote);
    ui->pushButtonStopAll->setEnabled(enableStop && !underRemote);
}

void MainWindow::updateGroupBoxWarn()
{
    QStringList warnStrList;
    bool hidden = true;
    if(!ui->labelWarnTrigDev->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnTrigDev->text();
    }
    if(!ui->labelWarnClients->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnClients->text();
    }
    if(!ui->labelWarnTmwr->isHidden()) {
        hidden = false;
        warnStrList << ui->labelWarnTmwr->text();
    }

    ui->groupBoxWarn->setHidden(hidden);
    if(!hidden)
        emit sendRunWarn(config.pc.runId, warnStrList.join("; "));
}

void MainWindow::updateRemoteBlocking()
{
    // TODO
    //    ui->trigMainForm->updateRemoteBlocking(remoteConnected());
}

void MainWindow::on_pushButtonRestartAll_clicked()
{
    // TODO
//    clients_config.clear();

    rcCoreApp->perfomeLocalStart();
}

void MainWindow::on_pushButtonStopAll_clicked()
{
    rcCoreApp->perfomeStop(false);
}

void MainWindow::on_actionFindDevice_triggered()
{
    discoverDialog->show();
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool val)
{
    config.pc.allowRemoteControl = val;
    emit configChanged(config);
}

void MainWindow::on_actionEnableStatusDump_toggled(bool val)
{
    config.pc.enableStatusDump = val;
    emit configChanged(config);
    emit setStatusDumpEnable(config.pc.enableStatusDump);
}

void MainWindow::on_actionSetStatusDumpDirectory_triggered()
{
    QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"),
                                                    config.pc.statusDumpDirName,
                                                    QFileDialog::ShowDirsOnly
                                                    | QFileDialog::DontResolveSymlinks);
    if (dir.isNull())
        return;
    config.pc.statusDumpDirName = dir;
    emit configChanged(config);
    emit setStatusDumpDirName(dir);
}

void MainWindow::on_tabWidget_currentChanged(int index)
{
    if(index == TAB_LOGS)
        setTabState(true, TAB_LOGS);
}

void MainWindow::on_actionPrint_Fsm_input_triggered()
{
    emit printFsmInput();
}
