//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "CoreTypes.h"
#include "RcConfig.h"
#include "RcFsm.h"
#include "TrigStatus.h"
#include "app/BaseMainWindow.h"
#include "pnp-server/PNPMessage.h"

#include <QMainWindow>
#include <QMap>
#include <QPointer>
#include <QTime>

class ConfigIndexName;
class ConfigSelector;
class ConfigurationManager;
class DeviceDescription;
class DeviceDispatcher;
class DeviceIndex;
class DeviceIndexEnabledMap;
class DiscoverDialog;
class LldpDiscovery;
class PNPServer;
class QHostAddress;
class QLabel;
class QTimer;
class RcConfigDb;
class RcCoreApp;
class RcSignalRouter;
class RemoteControlServer;
class RemoteControlServerStatus;
class SlowControlServer;
class TrigCore;
class TrigDevice;
class TrigDeviceProxy;
class TrigStatus;
class VmeTrigsController;
class VmeTrigsCore;

namespace Ui {
class MainWindow;
}

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

    friend class RcSignalRouter;

public:
    enum TAB_INDEX { TAB_TRIG_DEV=0, TAB_CLIENTS=1, TAB_VME_TRIGS=2, TAB_LOGS=3};

    explicit MainWindow(const ConfigSelector &cs, QWidget *parent = nullptr);
    virtual ~MainWindow() override;

public Q_SLOTS:
    void setConfig(const RcConfig &newConfig);
    void configSelectorChanged(const ConfigIndexName &conf);
    void setDeviceList(const DeviceIndexEnabledMap &devices);

    void onRcFsmStateChange(const FsmState &state, const FsmState &previousState);
    void trigManagerRunStateChanged(const FsmState &state);
    void clientManagerRunStateChanged(const FsmState &state);
    void vmeTrigsRunStateChanged(const FsmState &state);
    void client_config_receiver(QMap<int, ClientConfig> client_config_);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);

protected slots:
    void addLogMessage(QtMsgType severity, const QString &category, const QString &log, const QString &thrName) override;

private Q_SLOTS:
    void init(const ConfigSelector &cs);

    void trigStatusUpdated(const TrigStatus &status); // GUI
    void setTrigStatus(const DeviceIndex &, QString); // GUI

    void onRunWarning(const RcRunId &runId, QString warnStr);
    void onRunError(const RcRunId &runId, QString warnStr);
    void on_pushButtonRestartAll_clicked();
    void on_pushButtonStopAll_clicked();
    void on_actionFindDevice_triggered();
    void on_actionAllow_Remote_Control_toggled(bool val);
    void on_actionEnableStatusDump_toggled(bool val);

    void on_actionSetStatusDumpDirectory_triggered();

    void on_tabWidget_currentChanged(int index);

    void onSlowCntStatusUpdated(quint16 tcpPort, bool listening, bool portInUse); // GUI
    void blockGuiControl(bool block); // GUI

    void on_actionConfiguration_manager_triggered();
    void slaveMessageReceived(const MessageTreeItem &m);

    void on_actionPrint_Fsm_input_triggered();

signals:
    void configChanged(const RcConfig &config);
    void switch_to_config(const ConfigIndexName &conf) const;
    void deviceListChanged(const DeviceIndexEnabledMap &); // unused

//    void deleteHardware(const DeviceIndex &) const;
    void destroyAll() const;
    void allowSlowControl(bool) const;
    void sendRunWarn(const RcRunId &, QString) const;
    void deleteAll() const;
    // void deviceDiscovered(const DeviceDescription &) const;
    void sendControlEnable(bool block) const;
    void setStatusDumpDirName(const QString &);
    void setStatusDumpEnable(bool);
    __attribute__((deprecated("todel")))
    void sendSlaveMessage(const MessageTreeItem &);
    void printFsmInput();

protected:
    void closeEvent(QCloseEvent *event) override;

private:
    void subFsmStateChanged(int tabIndex, const FsmState &state);
    void setTabState(bool isOk, int tabIndex); // GUI
    void logRcFsmStateChange(const FsmState &previousState, const FsmState &state);
//    bool waitVmeTrigsStarted(const QElapsedTimer &startTime);
//    bool waitClientsStarted(const QElapsedTimer &startTime);
    void updateDeviceLabel();
    void updateDeviceStatusLabel(const TrigStatus &status);
    // unused int addClient(QHostAddress host, int port, int trig, int clientIndex=-1);
    void loadSettings();
    void updateStatStopButton();
    void updateGroupBoxWarn();
    void updateRemoteBlocking();
    void clearRunWarning();
    SlaveType mapSenderToSlaveType(QObject *sender);

    Ui::MainWindow *ui;

    QLabel * const labelFsmStateRc;
    QLabel * const labelFsmStateTrig;
    QLabel * const labelFsmStateClient;
    QLabel * const labelFsmStateVme;

    QLabel * const labelDevState;
    QLabel * const labelSlowControl;
    QLabel * const labelDevice;
    QLabel * const labelFirmware;
    QLabel * const labelTemp;
    DiscoverDialog * const discoverDialog;

    RcCoreApp * const rcCoreApp;
    RcConfig config;
    RcConfigDb * const configDb;
    QPointer<ConfigurationManager> configuration_manager;

    __attribute__((deprecated("change to specified BaseFsm state")))
    QMap<int, bool> tabOk;
    FsmState curRcFsmState = FsmState::Init;
    RemoteControlServerStatus * const remoteControlStatusCache;
};

#endif // MAINWINDOW_H
