/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcConfig.h"

#include <QDataStream>
#include <QDebug>
#include <QStringList>

DeviceIndex RcConfig::findVmeDeviceIndexBySerial(quint64 serial) const
{
    for (auto it = vmeTrigs.begin(); it != vmeTrigs.end(); ++it) {
        if (it.key().getSerial() == serial)
            return it.key();
    }
    return DeviceIndex();
}

QDebug operator<<(QDebug dbg, const VmeTrigConfig &r)
{
    dbg.noquote().nospace()
        << "VmeTrigConfig("
        << "enabled: " << r.enabled
        << "evNumReadReset: " << r.evNumReadReset
        << "checkWr: " << r.checkWr
    << ")";
    return dbg;
}

QDebug operator<<(QDebug dbg, const RcTrigConfig &r)
{
    dbg.noquote().nospace()
        << "RcTrigConfig("
        << ", deviceIndex: " << r.deviceIndex
        << ", checkLvdsParity: " << r.checkLvdsParity
        << ", checkBeforeProtection: " << r.checkBeforeProtection
        << ", checkAfterProtection: " << r.checkAfterProtection
        << ", bapBlock: " << r.bapBlock
        << ", trigRejectDelay: " << r.trigRejectDelay
        << ", beforeProtWidth: " << r.beforeProtWidth
        << ", afterProtWidth: " << r.afterProtWidth
        << ", lvdsMask: " << r.lvdsMask
        << ", trigWordMask: " << r.trigWordMask
        << ", trig: " << r.trig
        << ", spill: " << r.spill
        << ", plotSrc: " << r.plotSrc
        << ")";
    return dbg;
}

QDebug operator<<(QDebug dbg, const RcProgramConfig &r)
{
    dbg.noquote().nospace()
        << "RcProgramConfig("
        << "runEnable: " << r.runEnable
        << ", runId: " << r.runId
        << ", localRunNumber: " << r.localRunNumber
        << ", allowRemoteControl: " << r.allowRemoteControl
        << ", enableStatusDump: " << r.enableStatusDump
        << ", statusDumpDirName: " << r.statusDumpDirName
        << ")";
    return dbg;
}

QDebug operator<<(QDebug dbg, const RcConfig &r)
{
    dbg.noquote().nospace()
        << "RcConfig("
        << "pc: " << r.pc
        << ", rcTrigConfig: " << r.rcTrigConfig
        << ", clients: " << r.clients
        << ", channelDescr: " << r.channelDescr
        << ", vmeTrigs: " << r.vmeTrigs
        << ")";
    return dbg;
}

QDataStream &operator<<(QDataStream &out, const VmeTrigConfig &r)
{
    out
        << r.enabled
        << r.evNumReadReset
        << r.checkWr;
    return out;
}

QDataStream &operator>>(QDataStream &in, VmeTrigConfig &r)
{
    in
        >> r.enabled
        >> r.evNumReadReset
        >> r.checkWr;
    // TODO: error checking
    return in;
}

QDataStream &operator<<(QDataStream &out, const RcProgramConfig &r)
{
    out
        << r.runEnable
        << r.runId
        << r.localRunNumber
        << r.allowRemoteControl
        << r.enableStatusDump
        << r.statusDumpDirName;
    return out;
}

QDataStream &operator>>(QDataStream &in, RcProgramConfig &r)
{
    in
        >> r.runEnable
        >> r.runId
        >> r.localRunNumber
        >> r.allowRemoteControl
        >> r.enableStatusDump
        >> r.statusDumpDirName;
    // TODO: error checking
    return in;
}

QDataStream &operator<<(QDataStream &out, const TrigConfig &r)
{
    out
        << r.xOffMask
        << r.trigOpen
        << r.trigSrc
        << r.trigTimerFreq
        << r.trigRandomFreq
        << r.trigRandomDeadTimeUs;
    return out;
}

QDataStream &operator>>(QDataStream &in, TrigConfig &r)
{
    in
        >> r.xOffMask
        >> r.trigOpen
        >> r.trigSrc
        >> r.trigTimerFreq
        >> r.trigRandomFreq
        >> r.trigRandomDeadTimeUs;
    // TODO: error checking
    return in;
}

QDataStream &operator<<(QDataStream &out, const SpillConfig &r)
{
    out
        << r.durPhase1
        << r.durPhase2
        << r.durPhase3
        << r.durPhase4
        << r.durPhase5
        << static_cast<int>(r.source)
        << r.softSpill
        << r.timerPeriod
        << r.spillPlotStep
        << r.trigTimeStep;
    return out;
}

QDataStream &operator>>(QDataStream &in, SpillConfig &r)
{
    int source;
    in
        >> r.durPhase1
        >> r.durPhase2
        >> r.durPhase3
        >> r.durPhase4
        >> r.durPhase5
        >> source
        >> r.softSpill
        >> r.timerPeriod
        >> r.spillPlotStep
        >> r.trigTimeStep;
    // TODO: error checking
    r.source = static_cast<SpillConfig::spill_sel_t>(source);
    return in;
}

QDataStream &operator<<(QDataStream &out, const RcTrigConfig &r)
{
    out
        << r.deviceIndex
        << r.checkLvdsParity
        << r.checkBeforeProtection
        << r.checkAfterProtection
        << r.bapBlock
        << r.trigRejectDelay
        << r.beforeProtWidth
        << r.afterProtWidth
        << r.lvdsMask
        << r.trigWordMask
        << r.trig
        << r.spill
        << r.plotSrc
        ;
    return out;
}

QDataStream &operator>>(QDataStream &in, RcTrigConfig &r)
{
    in
        >> r.deviceIndex
        >> r.checkLvdsParity
        >> r.checkBeforeProtection
        >> r.checkAfterProtection
        >> r.bapBlock
        >> r.trigRejectDelay
        >> r.beforeProtWidth
        >> r.afterProtWidth
        >> r.lvdsMask
        >> r.trigWordMask
        >> r.trig
        >> r.spill
        >> r.plotSrc
        ;
    // TODO: error checking
    return in;
}

QDataStream &operator<<(QDataStream &out, const RcConfig &r)
{
    out
        << r.pc
        << r.rcTrigConfig
        << r.clients
        << r.channelDescr
        << r.vmeTrigs;
    return out;
}

QDataStream &operator>>(QDataStream &in, RcConfig &r)
{
    in
        >> r.pc
        >> r.rcTrigConfig
        >> r.clients
        >> r.channelDescr
        >> r.vmeTrigs;
    // TODO: error checking
    return in;
}
