/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef RCCONFIG_H
#define RCCONFIG_H

#include <QDateTime>
#include <QDebug>
#include <QMap>
#include <QString>
#include <QVector>
#include <QMetaType>

#include "SpillConfig.h"
#include "TrigConfig.h"
#include "base/ClientConfig.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "base/RcRunId.h"

class QDataStream;

struct VmeTrigConfig
{
    bool enabled = true;
    bool evNumReadReset = true;
    bool checkWr = true;
};
Q_DECLARE_METATYPE(VmeTrigConfig)

inline bool operator==(const VmeTrigConfig &l, const VmeTrigConfig &r)
{
    return l.enabled == r.enabled &&
           l.evNumReadReset == r.evNumReadReset &&
           l.checkWr == r.checkWr;
}

QDebug operator<<(QDebug dbg, const VmeTrigConfig &r);
QDataStream &operator<<(QDataStream &out, const VmeTrigConfig &r);
QDataStream &operator>>(QDataStream &in, VmeTrigConfig &r);

struct RcTrigConfig{
    bool isSpillGate() const {
        return spill.source == SpillConfig::SPILL_GATE;
    }

    DeviceIndex deviceIndex;
    bool checkLvdsParity = false;
    bool checkBeforeProtection = false;
    bool checkAfterProtection = false;
    bool bapBlock = false; // Logic Before/After protection (block or check)
    quint16 trigRejectDelay = 0; // in 16ns steps
    quint16 beforeProtWidth = 2000/16; // in 16ns steps
    quint16 afterProtWidth = 2000/16; // in 16ns steps
    quint16 lvdsMask = 0;
    quint16 trigWordMask = 0;
    TrigConfig trig;
    SpillConfig spill;
    quint8 plotSrc = 14;
};

Q_DECLARE_METATYPE(RcTrigConfig)

inline bool operator==(const RcTrigConfig &l, const RcTrigConfig &r)
{
    return
           l.deviceIndex == r.deviceIndex &&
           l.checkLvdsParity == r.checkLvdsParity &&
           l.checkBeforeProtection == r.checkBeforeProtection &&
           l.checkAfterProtection == r.checkAfterProtection &&
           l.bapBlock == r.bapBlock &&
           l.trigRejectDelay == r.trigRejectDelay &&
           l.beforeProtWidth == r.beforeProtWidth &&
           l.afterProtWidth == r.afterProtWidth &&
           l.lvdsMask == r.lvdsMask &&
           l.trigWordMask == r.trigWordMask &&
           l.trig == r.trig &&
           l.spill == r.spill &&
           l.plotSrc == r.plotSrc;
}
inline bool operator!=(const RcTrigConfig &l, const RcTrigConfig &r)
{
    return !(l==r);
}

QDebug operator<<(QDebug dbg, const RcTrigConfig &r);

struct RcProgramConfig
{
    bool runEnable = false;
    RcRunId runId;
    int localRunNumber = 0;
    bool allowRemoteControl = true;
    bool enableStatusDump = false;
    QString statusDumpDirName;
};

Q_DECLARE_METATYPE(RcProgramConfig)

inline bool operator==(const RcProgramConfig &l, const RcProgramConfig &r)
{
    return l.runEnable == r.runEnable &&
           l.runId == r.runId &&
           l.localRunNumber == r.localRunNumber &&
           l.allowRemoteControl == r.allowRemoteControl &&
           l.enableStatusDump == r.enableStatusDump &&
           l.statusDumpDirName == r.statusDumpDirName;
}

QDebug operator<<(QDebug dbg, const RcProgramConfig &r);

class QSettings;

using RcClientMap = QMap<int, ClientConfig>;
using RcXoffChannelDescr = QMap<int, QString>;
using RcVmeTrigs = QMap<DeviceIndex, VmeTrigConfig>;

class RcConfig
{
public:


    DeviceIndex findVmeDeviceIndexBySerial(quint64 serial) const;

    bool hasVme() const {
        for (const auto &cc : clients) {
            if (cc.readout && cc.hasVme())
                return true;
        }
        return false;
    }
    inline bool trigInMoreThatOne() const
    {
        return rcTrigConfig.deviceIndex.getDevId() == DEVICE_ID_UT24VE;
    }

    RcProgramConfig pc;
    RcTrigConfig rcTrigConfig;
    RcClientMap clients;
    RcXoffChannelDescr channelDescr;
    RcVmeTrigs vmeTrigs;
};

Q_DECLARE_METATYPE(RcConfig)

inline bool operator==(const RcConfig &l, const RcConfig &r)
{
    return l.pc == r.pc &&
           l.rcTrigConfig == r.rcTrigConfig &&
           l.clients == r.clients &&
           l.channelDescr == r.channelDescr &&
           l.vmeTrigs == r.vmeTrigs;
}

QDebug operator<<(QDebug dbg, const RcConfig &r);
QDataStream &operator<<(QDataStream &out, const RcConfig &r);
QDataStream &operator>>(QDataStream &in, RcConfig &r);

#endif // RCCONFIG_H
