/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcCoreApp.h"

#include <QJsonDocument>
#include <QThread>
#include <QTimer>

#include "RcConfigDb.h"
#include "RcConfigDispatcher.h"
#include "RcFsm.h"
#include "RcSignalRouter.h"
#include "SlowControlServer.h"
#include "TrigCore.h"
#include "TrigDevice.h"
#include "TrigDeviceController.h"
#include "VmeTrigsController.h"
#include "VmeTrigsCore.h"
#include "base/SyslogMessageSender.h"
#include "base/ThreadController.h"
#include "base/config/ConfigSelector.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mongo/StartupOptions.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "run-control-client-manager/ClientManagerWidget.h"
#include "run-control-client-manager/RunStates.h"

static const int CONFIG_TIMEOUT = 4000; // time to collect configuration from all clients
static const int START_TIMEOUT = 3e3;
static const int STOP_TIMEOUT = 10e3;

static const QMap<FsmState, QString> map_fsm_to_sc_state = {
    {FsmState::Init, SC_STATE_STANDBY},
    {FsmState::Error, SC_STATE_FAULT},
    {FsmState::Ready, SC_STATE_STANDBY},
//    {FsmState::Configure, SC_STATE_INIT},
    {FsmState::Run, SC_STATE_RUNNING},
    {FsmState::Starting, SC_STATE_INIT},
    {FsmState::Stopping, SC_STATE_INIT},
};

RcCoreApp::RcCoreApp(QObject *parent) :
    AbstractConfigurable(parent)
    , rcConfigDispatcher(new RcConfigDispatcher(this))
    , trigCore(new TrigCore(this))
    , vmeTrigsCore(new VmeTrigsCore(this))
    , slowControl(new SlowControlServer(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , pnpServer(new PNPServer(this))
    , evNumChecker(new EvNumChecker(this))
    , config_timer(new QTimer(this))
    , fsm(new RcFsm(this))
{
    vmeTrigsCore->setConfigDispatcher(rcConfigDispatcher);

    auto trigDeviceController = trigCore->getTrigDeviceController();
    trigDeviceController->setConfigDispatcher(rcConfigDispatcher);

    connect(this, &RcCoreApp::sendSlaveMessage,
            this, &RcCoreApp::slaveMessageReceived);

    progDesc.type = DaqModule(DaqModule::Type::RunControl).toString();

    auto &router = RcSignalRouter::instance();
    router.addRcCoreApp(this);
    router.addRcConfigDispatcher(rcConfigDispatcher);
    router.addVmeTrigsController(vmeTrigsController());
    router.addTrigDeviceController(trigCore->getTrigDeviceController());
    router.addRemoteControlServer(remoteControl);
    router.addSlowControlServer(slowControl);
    router.addPnpServer(pnpServer);
    router.addEvNumChecker(evNumChecker);
    router.addRcFsm(fsm);
    // router.connectCore();

    //--- configurations transfer
    connect(config_timer, &QTimer::timeout,
            this, &RcCoreApp::config_timeout);
    config_timer->setSingleShot(true);
    config_timer->setInterval(CONFIG_TIMEOUT);

    new ThreadController(remoteControl, this);
    new ThreadController(slowControl, this);

    QList<DaqModule> searchTargs;
    searchTargs << DaqModule(DaqModule::Type::Adc64System);
    searchTargs << DaqModule(DaqModule::Type::EventBuilder);
    searchTargs << DaqModule(DaqModule::Type::HrbSystem);
    searchTargs << DaqModule(DaqModule::Type::RunControl);
    searchTargs << DaqModule(DaqModule::Type::VmeDaq);
    searchTargs << DaqModule(DaqModule::Type::Land);
    searchTargs << DaqModule(DaqModule::Type::Ttvxs);
    searchTargs << DaqModule(DaqModule::Type::Generator);
    searchTargs << DaqModule(DaqModule::Type::TdcControl);
    searchTargs << DaqModule(DaqModule::Type::TQDC);
    pnpServer->startSearch(searchTargs, false);
}

RcCoreApp::~RcCoreApp()
{
}

DeviceDispatcher *RcCoreApp::getRcDeviceDispatcher()
{
    return trigCore->getDeviceDispatcher();
}

VmeTrigsController *RcCoreApp::vmeTrigsController()
{
    return vmeTrigsCore->getVmeTrigsController();
}

void RcCoreApp::quit()
{
    emit closeProgram(progDesc);

    RcSignalRouter::instance().removeRemoteControlServer();

    deleteLater();
}

void RcCoreApp::configSelectorChanged(const ConfigIndexName &conf)
{
    localProgramIndex = conf.getIndex();
    progDesc.index = conf.getIndex().toString();
    emit changeProgramDescription(progDesc);
}

void RcCoreApp::setConfig(const RcConfig &newConfig)
{
    if(config == newConfig)
        return;

    const auto oldConfig = config;
    config = newConfig;
    emit allowRemoteControl(config.pc.allowRemoteControl);
    updateProgramDescription();
    updateRemoteState();
    if (oldConfig.hasVme() != config.hasVme()) {
        emit sendVmeFlag(config.hasVme());
    }
}

void RcCoreApp::actionOnStateRun()
{
    QCoreApplication::processEvents(QEventLoop::AllEvents, 2000); // why?
    QVector<quint64> devSerials;
    if (!megamaster_flg)
        devSerials << config.rcTrigConfig.deviceIndex.getSerial();
    evNumChecker->set_dev_serials(devSerials);
    if (megamaster_flg)
        evNumChecker->startCheck();

    config_timer->start();
}

void RcCoreApp::actionPostRun()
{
//    incrementRunNumber();
}

void RcCoreApp::incrementRunNumber()
{
    config.pc.localRunNumber++;
    emit configChanged(config);
}

void RcCoreApp::onRcFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    logRcFsmStateChange(oldState, state);

    switch (state) {
    case FsmState::CloseGlobTrig: {
        emit closeGlobTrig();
        const bool waitSpillEnd = megamaster_flg && config.hasVme();
        fsm->goState(waitSpillEnd ? FsmState::WaitSpillEnd : FsmState::Starting);
        break;
    }
    case FsmState::OpenGlobTrig: {
        emit openGlobTrig();
        fsm->goState(FsmState::Run);
        break;
    }
    case FsmState::Starting:
        emit runStart(config.pc.runId);
        fsm->goState(megamaster_flg ? FsmState::OpenGlobTrig : FsmState::Run);
        break;
    case FsmState::Run:
        actionOnStateRun();
        break;
    case FsmState::PrepareStopping:
        Q_ASSERT(megamaster_flg);
        emit prepareStop();
        fsm->goState(FsmState::Stopping);
        break;
    case FsmState::Stopping:
    {
        quint64 stopEvNum=0;
        if(megamaster_flg) {
            evNumChecker->stopCheck();
            stopEvNum = getTrigCore()->getTrigDeviceController()->getStopEvNum();
        }
        emit runStop(stopEvNum);
        config.pc.runEnable = false;
        emit configChanged(config);
        break;
    }
    case FsmState::Error:
        if(megamaster_flg && config.pc.runEnable){
            setTrigState(false);
            config.pc.runEnable = false;
            emit configChanged(config);

            evNumChecker->stopCheck();
            emit runStop();
        }
        break;
    default:
    {}
    }

    if (oldState == FsmState::Run || state != FsmState::Run) {
        actionPostRun();
    }
    emit slowControlStateUpdated(map_fsm_to_sc_state[state], "");
}

void RcCoreApp::logRcFsmStateChange(const FsmState &previousState, const FsmState &state)
{
    auto RunIdStr = QString(config.pc.runId);
    auto previousStateStr = fsmStateToString(previousState);

    if (state == FsmState::Starting) {
        qInfo() << QString("!!! Preparing Run %1... !!!").arg(RunIdStr);
    }

    if (state == FsmState::Stopping) {
        qInfo() << QString("!!! Stopping Run %1... !!!").arg(RunIdStr);
    }

    if (state == FsmState::Run) {
        qInfo() << "!!! Run started !!!";
    }
    if (state == FsmState::Ready) {
        if (previousState == FsmState::Stopping) {
            qInfo() << "!!! Run stopped !!!";
        }
    }
    if (state == FsmState::Error) {
        switch (previousState) {
        case FsmState::Starting:;
        case FsmState::Stopping:;
        case FsmState::Run: {
            qCritical() << QString("!!! %1 failed !!!").arg(previousStateStr);
            break;
        }
        default:;
        }
    }
}

void RcCoreApp::trigManagerRunStateChanged(const FsmState &state)
{
    fsm->setTrigDeviceMgrState(state);
}

void RcCoreApp::clientManagerRunStateChanged(const FsmState &state)
{
    fsm->setClientManagerRunState(state);
}

void RcCoreApp::vmeTrigsRunStateChanged(const FsmState &state)
{
    fsm->setVmeTrigsRunState(state);
}

void RcCoreApp::setTrigState(bool state)
{
    config.rcTrigConfig.trig.trigOpen = state;
    emit configChanged(config);
}

void RcCoreApp::config_transfer_from_clients(int id, const QJsonObject &json)
{
    for(int id_tmp : config.clients.keys()){
        if(config.clients[id_tmp].id == id){
            clients_config[id] = json;
        }
    }

    bool all_clients_config = true;
    for(int id_tmp : config.clients.keys()) {
        if(!clients_config.contains(id_tmp)) {
            all_clients_config = false;
        }
    }

    if(all_clients_config) {
        run_root_prepare();
        config_timer->stop();
    }
}

void RcCoreApp::run_root_prepare()
{
//    RootConfig my_config = getRootConfig();

//    for(int id_tmp : clients_config.keys()){
//        RootConfig root_client = clients_config[id_tmp];
//        QString program_type = root_client.get_data<QString>("program_type");
//        QString program_index = root_client.get_data<QString>("program_index");
//        QString run_number = QString::number(root_client.get_data<int>("run_number"));
//        QString item_name = program_type + "_" + program_index + "_" + run_number;

//        RootConfig my_child;
//        QString child_name;
//        for(int i=1; i<=clients_config.size(); ++i){
//            child_name = QString("Client_%1").arg(i);
//            my_child = my_config.child(child_name);
//            QString program_type_tmp = my_child.get_data<QString>("program_type");
//            QString program_index_tmp = my_child.get_data<QString>("program_index");
//            if((program_type == program_type_tmp)&&(program_index == program_index_tmp)){
//                my_child.insert(item_name, root_client);
//                my_config.remove_child(child_name);
//                my_config.insert(child_name, my_child);
//                break;
//            }
//        }
//    }

    if(config.clients.isEmpty()) {
        const auto &json = QVariant::fromValue(config).value<QJsonObject>();
        // FIXME configDb->write_history_config(run_root);
        emit write_history_config(config.pc.runId, json, megamaster_flg);
    }

    if(megamaster_flg) {
//        auto run_root = RootConfigUtil::createRootConfig(getConfigSelector().getType(), config.pc.runId);
//        run_root.insert(RootConfigUtil::makeRootItemName(getConfigSelector(), config.pc.runId),
//                        my_config);
        // FIXME configDb->write_history_config(run_root);
    } else {
        const auto &json = QVariant::fromValue(config).value<QJsonObject>();
        emit config_transfer_to_rc(QJsonDocument(json).toJson());
    }
}

void RcCoreApp::config_timeout()
{
    run_root_prepare();
}

void RcCoreApp::onRemoteStartRequest(const RcRunId &runId, quint64 startEvNumber)
{
    Q_UNUSED(startEvNumber)
    megamaster_flg = false;
    clients_config.clear();
    bool success = perfomeStart(runId);
    if(success && config.clients.isEmpty()){
        const auto &json = QVariant::fromValue(config).value<QJsonObject>();
        emit config_transfer_to_rc(QJsonDocument(json).toJson());
    }

    updateProgramDescription();
}

void RcCoreApp::onRemoteStopRequest(const RcRunId &runId, quint64 stopEvNumber)
{
    Q_UNUSED(stopEvNumber)
    perfomeStop(false);
    updateProgramDescription();
}

void RcCoreApp::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDesc.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit changeProgramDescription(progDesc);

    const bool doStop = config.pc.runEnable && isRemoteConnected() && !status.portInUse;
    remoteConnected = status.portInUse;

    if(doStop) {
        qWarning() << "Stop run due to remoteControl disconnect";
        perfomeStop(true);
    }
    // TODO updateStatStopButton();
}

void RcCoreApp::trigStatusUpdated(const TrigStatus &status)
{
    lastTrigStatus = status;
    fsm->setSpillState(status.spillState.reg.bit.phase == 4);
    if(!status.connected && config.pc.runEnable) {
        auto msg = SeverityMessage(SlaveMessageSeverity::Warning, QString(status.ident + " disconnected"));

        const MessageTreeItem m(RcSlaveId(SlaveType::Trig, status.ident),
                               config.pc.runId,
                               msg);
        emit slaveMessageReceived(m);
    }
}

void RcCoreApp::onSlaveErrorMessageReceived(const MessageTreeItem &m)
{
    if (isRemoteConnected()) {
        QString mess = QString("Retransmit runError to the remoteControl in Run %1")
                           .arg(QString(config.pc.runId));
        if (!m.isEmpty())
            mess.append(": ").append(m.text());
        qWarning() << mess;
        emit sendRunError(m.runId, m.text());
    } else {
        QString mess = QString("Errors in Run %1").arg(QString(config.pc.runId));
        if (!m.isEmpty())
            mess.append(": ").append(m.text());
        qWarning() << mess;

        perfomeStop(true);
    }
}

void RcCoreApp::slaveMessageReceived(const MessageTreeItem &m)
{
    if (!config.pc.runEnable || config.pc.runId != m.runId) {
        // qWarning() << QString("Got RunError from %1 run").arg(config.pc.runEnable?"previous":"stopped");
        return;
    }
    switch (m.severity()) {
    case SlaveMessageSeverity::Error:
        onSlaveErrorMessageReceived(m);
        break;
    case SlaveMessageSeverity::Warning:
        break;
    }
    // append and send to upstream

}

bool RcCoreApp::perfomeLocalStart()
{
    incrementRunNumber();
    const RcRunId newRunId(localProgramIndex.toString(),
                           config.pc.localRunNumber,
                           QUuid::createUuid());
    megamaster_flg = true;
    return perfomeStart(newRunId);
}

bool RcCoreApp::perfomeStart(const RcRunId &runId)
{
    config.pc.runId = runId;
    config.pc.runEnable = true;
    emit configChanged(config);

    fsm->goState(megamaster_flg ? FsmState::CloseGlobTrig : FsmState::Starting);
    return true;
}

void RcCoreApp::perfomeStop(bool isError)
{
    qInfo() << __PRETTY_FUNCTION__ << "isError:" << isError << " state:" << fsm->getState();
    isError = fsm->getState() != FsmState::Run;
    fsm->goState(isError ? FsmState::Error
                         : (megamaster_flg ? FsmState::PrepareStopping
                                           : FsmState::Stopping));

//    // TODO: Smart wait
//    if (remoteConnected()) {
//        QCoreApplication::processEvents(QEventLoop::AllEvents, 1000);
//    } else {
//        QTime t;
//        t.start();
//        int elapsedTime;
//        while((elapsedTime=t.elapsed())<1000){
//            QCoreApplication::processEvents(QEventLoop::AllEvents, 1000-elapsedTime);
//            QThread::msleep(100);
//        }
//    }
//    emit runStop();

//    QElapsedTimer timer;
//    timer.start();

//    isOk &= waitFsmState(FsmState::Ready, STOP_TIMEOUT);
/*
    // Get stoped state from Tcp Clients (sub-programms)
    while(!timer.hasExpired(10000)) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
        if(ui->clientManagerWidget->get_run_id().runNumber==config.pc.runId.runNumber &&
            // FIXME: compare runId, not number
            ui->clientManagerWidget->getRunState() != RS_STOPPING )
            break;
    }
    if(ui->clientManagerWidget->getRunState() != RS_STOPPED){
        qDebug() << "Some client failed to stop";
        isOk = false;
    }
*/
//    if(isOk) {
//    } else {
        // emit runStop(); // Again???
//    }
//    emit slowControlStateUpdated(hasRunError ? SC_STATE_FAULT : SC_STATE_STANDBY, "Run stopped");
//    return isOk;
}

void RcCoreApp::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(config.pc.runEnable && config.pc.runId == runId) {
        if (isRemoteConnected()) {
            QString mess = QString("Retransmit backPressure(%2) to the remoteControl in Run %1")
                    .arg(QString(config.pc.runId))
                    .arg(backPressure);
            if(backPressure)
                qWarning() << mess;
            else
                qInfo() << mess;
            emit sendBackPressure(runId, backPressure);
        } else {
            if(backPressure) {
                setTrigState(false);
                qWarning()<< "BackPressure enabled";
            } else {
                setTrigState(true);
                qInfo()<< "BackPressure disabled";

            }
        }
    }
}

void RcCoreApp::onVmeFlagUpdated()
{
    emit sendVmeFlag(config.hasVme());
}

QMap<QString, QString> RcCoreApp::createProgramDescriptionOptions(const RcConfig &config)
{
    QMap<QString, QString> options;

    options["runEnable"] = QVariant(config.pc.runEnable).toString();
    options["run_index"] = config.pc.runId.runIndex;
    options["run_number"] = QVariant(config.pc.runId.runNumber).toString();

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    for (const ClientConfig &c: config.clients.values()) {
        /*
        options[QString("client_%1_progType").arg(c.id)] = c.progType;
        options[QString("adc_%1_progIndex").arg(c.id)] = c.progIndex;
        options[QString("adc_%1_readout").arg(c.id)] = QVariant(c.readout).toString();
        */
        options[QString("Client-%1").arg(c.id)] = c.daq_module.toString() +" "+ c.program_index +" , "+ QVariant(c.readout).toString();
    }

    int n=0;
    for (auto it = config.vmeTrigs.begin(); it != config.vmeTrigs.end(); ++it) {
        const auto index = it.key();
        const auto &vme = it.value();
        if(!vme.enabled)
            continue;
        options[QString("vmeTrigModule_%1_serial").arg(n)] = index.getSerialStr();
        options[QString("vmeTrigModule_%1_evNumReadReset").arg(n)] = QVariant(vme.evNumReadReset).toString();
        options[QString("vmeTrigModule_%1_checkWr").arg(n)] = QVariant(vme.checkWr).toString();
    }
    return options;
}

QString RcCoreApp::createRemoteState(const QMap<int, ClientConfig> &clients, const QMap<DeviceIndex, VmeTrigConfig> &vmeTrigs)
{
    QStringList remoteState;

    for (const ClientConfig &c: clients) {
        if(c.readout)
            remoteState << QString("%1 #%2").arg(c.daq_module.toString(),
                                                 c.program_index);
    }

    for (auto it = vmeTrigs.begin(); it != vmeTrigs.end(); ++it) {
        const auto index = it.key();
        const auto &vme = it.value();
        if (vme.enabled && (vme.evNumReadReset || vme.evNumReadReset))
            remoteState << QString("tmwr:%1").arg(index.getIdent());
    }
    return remoteState.join("; ");
}

void RcCoreApp::updateProgramDescription()
{
    const auto newOptions = createProgramDescriptionOptions(config);
    progDesc.options = newOptions;
    if (newOptions != progDesc.options) {
        progDesc.options = newOptions;
        emit changeProgramDescription(progDesc);
    }
}

void RcCoreApp::updateRemoteState()
{
    static QString oldState;
    const auto &state = createRemoteState(config.clients, config.vmeTrigs);
    if (state == oldState)
        return;
    oldState = state;
    emit sendRemoteState(state);
}

ConfigSelector RcCoreApp::getConfigSelector() const
{
    return ConfigSelector(); // TODO
}
