/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef RCCOREAPP_H
#define RCCOREAPP_H

#include <QObject>
#include <QPointer>

#include "AbstractConfigurable.h"
#include "CoreTypes.h"
#include "RcConfig.h"
#include "fsm/FsmState.h"
#include "run-control-client-manager/RunStates.h"
#include "TrigStatus.h"
#include "pnp-server/PNPMessage.h"
#include "base/ProgramIndex.h"

class ConfigIndexName;
class ConfigSelector;
class DeviceDescription;
class DeviceDispatcher;
class DeviceIndex;
class DeviceIndexEnabledMap;
class DiscoverDialog;
class EvNumChecker;
class LldpDiscovery;
class PNPServer;
class QHostAddress;
class QLabel;
class QTimer;
class RcConfigDispatcher;
class RcCoreApp;
class RcFsm;
class RcSignalRouter;
class RemoteControlServer;
class RemoteControlServerStatus;
class SlowControlServer;
class StartupOptions;
class TrigCore;
class TrigDevice;
class TrigDeviceProxy;
class TrigStatus;
class VmeTrigsController;
class VmeTrigsCore;


class RcCoreApp : public AbstractConfigurable
{
    Q_OBJECT
public:
    explicit RcCoreApp(QObject *parent = nullptr);
    virtual ~RcCoreApp() override;
    RcConfigDispatcher *getConfigDispatcher() { return rcConfigDispatcher; }
    DeviceDispatcher *getRcDeviceDispatcher();
    TrigCore *getTrigCore() { return trigCore; }
    VmeTrigsCore *getVmeTrigsCore() { return vmeTrigsCore; }

signals:
    void runStart(const RcRunId &runId) const;
    void closeGlobTrig() const;
    void openGlobTrig() const;
    void prepareStop() const;
    void runStop(quint64 stopEvNum=0) const;
    void slowControlStateUpdated(QString, QString) const;
    void clientManagerRunStateChanged_(const FsmState &); //proxy
    void vmeTrigsRunStateChanged_(const FsmState &); //proxy
    void trigManagerRunStateChanged_(const FsmState &); //proxy
    void changeProgramDescription(const ProgramDescription &) const;
    void sendRemoteState(const QString &) const;
    void sendVmeFlag(bool) const;
    void sendBackPressure(const RcRunId &, bool) const;
    void closeProgram(const ProgramDescription &) const;
    void sendRunError(const RcRunId &, QString) const;
    void allowRemoteControl(bool);
    void config_transfer_to_rc(QByteArray data);
    void sendSlaveMessage(const MessageTreeItem &m);
    void write_history_config(const RcRunId &, const QJsonObject &, bool megamaster_flg) const;
    void switch_to_config(const ConfigIndexName &conf) const;
    void printFsmInput() const;

public slots:
    void quit();
    void configSelectorChanged(const ConfigIndexName &conf);
    void setConfig(const RcConfig &newConfig) override;

    void onRcFsmStateChange(const FsmState &state, const FsmState &oldState);
    void setTrigState(bool state);
    void trigManagerRunStateChanged(const FsmState &state);
    void clientManagerRunStateChanged(const FsmState &state);
    void vmeTrigsRunStateChanged(const FsmState &state);

    void onBackPressure(const RcRunId &runId, bool backPressure);
    void onVmeFlagUpdated();

    void config_transfer_from_clients(int id, const QJsonObject &json);

    void onRemoteStartRequest(const RcRunId &runId, quint64 startEvNumber=0);
    void onRemoteStopRequest(const RcRunId &runId, quint64 stopEvNumber=0);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);
    void trigStatusUpdated(const TrigStatus &status);
    void onSlaveErrorMessageReceived(const MessageTreeItem &m);
    void slaveMessageReceived(const MessageTreeItem &m);

private slots:
    void config_timeout();

public: // temporarily public
    bool perfomeLocalStart();
    bool perfomeStart(const RcRunId &runId);
    void perfomeStop(bool isError);

private:
    void actionOnStateRun();
    void actionPostRun();
    void incrementRunNumber();
    VmeTrigsController *vmeTrigsController();
    void run_root_prepare();
    inline bool isRemoteConnected() const { return remoteConnected; }
    void logRcFsmStateChange(const FsmState &previousState, const FsmState &state);
    void updateProgramDescription();

    static QMap<QString, QString> createProgramDescriptionOptions(const RcConfig &config);
    static QString createRemoteState(const QMap<int,
                                     ClientConfig> &clients,
                                     const QMap<DeviceIndex,
                                     VmeTrigConfig> &vmeTrigs);
    void updateRemoteState();
    ConfigSelector getConfigSelector() const;

    using ClientsJsonConfig = QMap<int, QJsonObject>;

    RcConfig config;
    ProgramIndex localProgramIndex;
    RcConfigDispatcher * const rcConfigDispatcher;
    TrigCore * const trigCore;
    VmeTrigsCore * const vmeTrigsCore;
    SlowControlServer * const slowControl;
    RemoteControlServer * const remoteControl;
    bool remoteConnected = false;
    PNPServer * const pnpServer;
    ProgramDescription progDesc;
    EvNumChecker * const evNumChecker;
    QTimer * const config_timer;
    ClientsJsonConfig clients_config;
    bool megamaster_flg = false;
    RcFsm * const fsm;
    TrigStatus lastTrigStatus;
    // RcSlaveStatus slaveStatus;
};

#endif // RCCOREAPP_H
