/*
**    Copyright 2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcSignalRouter.h"

#include "MainWindow.h"
#include "RcConfigDb.h"
#include "RcConfigDispatcher.h"
#include "RcCoreApp.h"
#include "SlowControlServer.h"
#include "TrigDevice.h"
#include "TrigDeviceController.h"
#include "TrigMainForm.h"
#include "VmeTrigsController.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "run-control-client-manager/ClientManagerWidget.h"

RcSignalRouter::RcSignalRouter()
    : QObject(nullptr)
{

}

void RcSignalRouter::connectAll()
{
    Q_ASSERT(rcConfigDispatcher);
    Q_ASSERT(vmeTrigsController);

    connect_ClientManagerWidget();
    connect_Configuration();
    connect_EvNumChecker();
    connect_MainWindow();
    connect_PnpServer();
    connect_RemoteControlServer();
    connect_SlowControlServer();
    connect_VmeTrigsController();
    connect_TrigMainForm();
    connect_TrigDeviceController();
}

void RcSignalRouter::connect_MainWindow()
{
    Q_ASSERT(clientManagerWidget);
    Q_ASSERT(mainWindow);
    Q_ASSERT(remoteControlServer);

    connect(clientManagerWidget, &ClientManagerWidget::config_transfer,
            rcCoreApp, &RcCoreApp::config_transfer_from_clients,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::config_transfer_to_rc,
            remoteControlServer, &RemoteControlServer::config_transfer,
            Qt::UniqueConnection);
    connect(mainWindow, &MainWindow::printFsmInput,
            rcCoreApp, &RcCoreApp::printFsmInput);
}

void RcSignalRouter::connect_PnpServer()
{
    Q_ASSERT(clientManagerWidget);
    Q_ASSERT(mainWindow);
    Q_ASSERT(pnpServer);
    Q_ASSERT(vmeTrigsController);
    Q_ASSERT(rcFsm);

    connect(pnpServer, &PNPServer::programDiscovered,
            clientManagerWidget, &ClientManagerWidget::programDiscovered,
            Qt::UniqueConnection);

    connect(pnpServer, &PNPServer::programClosed,
            clientManagerWidget, &ClientManagerWidget::programClosed,
            Qt::UniqueConnection);

    connect(pnpServer, &PNPServer::programDiscovered,
            vmeTrigsController, &VmeTrigsController::gotProgramDescription,
            Qt::UniqueConnection);

    connect(pnpServer, &PNPServer::programClosed,
            vmeTrigsController, &VmeTrigsController::gotClosedProgram,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::changeProgramDescription,
            pnpServer, &PNPServer::changeProgramDescription,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::closeProgram,
            pnpServer, &PNPServer::closeProgram,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::deleteAll,
            pnpServer, &QObject::deleteLater,
            Qt::UniqueConnection);

    connect(rcFsm, &RcFsm::stateChanged,
            rcCoreApp, &RcCoreApp::onRcFsmStateChange);
    connect(rcCoreApp, &RcCoreApp::printFsmInput,
            rcFsm, &RcFsm::printFsmInput);
}

void RcSignalRouter::connect_TrigDevice(TrigDevice *trigDevice)
{
    if (!trigDevice)
        return;

//    Q_ASSERT(deviceDispatcher);
    Q_ASSERT(mainWindow);
    Q_ASSERT(slowControlServer);
    Q_ASSERT(trigDevice);
    Q_ASSERT(trigMainForm);
    Q_ASSERT(rcCoreApp);

    connect(trigDevice, &TrigDevice::firmwareBuildParamsChanged,
            trigMainForm, &TrigMainForm::setFirmwareBuildParams,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::setStatus,
            mainWindow, &MainWindow::setTrigStatus,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::updateStatus,
            mainWindow, &MainWindow::trigStatusUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::updateStatus,
            rcCoreApp, &RcCoreApp::trigStatusUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::updateStatus,
            trigMainForm, &TrigMainForm::trigStatusUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::updateStatus,
            slowControlServer, &SlowControlServer::trigStatusUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::spillPlotUpdated,
            trigMainForm, &TrigMainForm::spillPlotUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::spillPlotUpdated,
            slowControlServer, &SlowControlServer::spillPlotUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::triggerHistogramUpdated,
            trigMainForm, &TrigMainForm::triggerHistogramUpdated,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::triggerHistogramUpdated,
            slowControlServer, &SlowControlServer::triggerHistogramUpdated,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::setStatusDumpEnable,
            trigDevice, &TrigDevice::setStatusDumpEnable,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::setStatusDumpDirName,
            trigDevice, &TrigDevice::setStatusDumpDirName,
            Qt::UniqueConnection);

//    connect(trigMainForm, &TrigMainForm::trigSetupUpdated,
//            trigDevice, &TrigDevice::writeSetup,
//            Qt::UniqueConnection);

//    connect(trigDevice, &TrigDevice::setupWritten,
//            mainWindow, &MainWindow::writeSetup,
//            Qt::UniqueConnection);

    connect(trigMainForm, &TrigMainForm::clearTrigStats,
            trigDevice, &TrigDevice::clearStats,
            Qt::UniqueConnection);

    connect(trigMainForm, &TrigMainForm::resetAll,
            trigDevice, &TrigDevice::resetAll,
            Qt::UniqueConnection);

    connect_EvNumChecker(trigDevice);
//    deviceDispatcher->subscribeHostAddressUpdates(trigDevice->getDeviceIndex(), trigDevice);
}

void RcSignalRouter::connect_RemoteControlServer()
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(remoteControlServer);

    connect(remoteControlServer, &RemoteControlServer::startRequest,
            rcCoreApp, &RcCoreApp::onRemoteStartRequest,
            Qt::UniqueConnection);

    connect(remoteControlServer, &RemoteControlServer::stopRequest,
            rcCoreApp, &RcCoreApp::onRemoteStopRequest,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::sendRemoteState,
            remoteControlServer, &RemoteControlServer::sendState,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::sendVmeFlag,
            remoteControlServer, &RemoteControlServer::onVmeFlag,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::sendRunError,
            remoteControlServer, &RemoteControlServer::onRunError,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::sendRunWarn,
            remoteControlServer, &RemoteControlServer::onRunWarning,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::sendBackPressure,
            remoteControlServer, &RemoteControlServer::onBackPressure,
            Qt::UniqueConnection);

    connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
            rcCoreApp, &RcCoreApp::onRemoteStatusUpdated,
            Qt::UniqueConnection);
    connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
            mainWindow, &MainWindow::onRemoteStatusUpdated,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::allowRemoteControl,
            remoteControlServer, &RemoteControlServer::startListen,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::changeProgramDescription,
            remoteControlServer, &RemoteControlServer::changeProgramDescription,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::destroyAll,
            remoteControlServer, &RemoteControlServer::deleteLater,
            Qt::UniqueConnection);

    connect(rcFsm, &RcFsm::stateChanged,
            remoteControlServer, &RemoteControlServer::onFsmStateChange);
}

void RcSignalRouter::removeRemoteControlServer()
{
    disconnect(remoteControlServer, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControlServer, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControlServer, &RemoteControlServer::stateRequest, nullptr, nullptr);
    disconnect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);
    remoteControlServer = nullptr;
}

void RcSignalRouter::connect_ClientManagerWidget()
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(clientManagerWidget);

    connect(clientManagerWidget, &ClientManagerWidget::runWarn,
            mainWindow, &MainWindow::onRunWarning,
            Qt::UniqueConnection);

    connect(clientManagerWidget, &ClientManagerWidget::runError,
            mainWindow, &MainWindow::onRunError,
            Qt::UniqueConnection);

    connect(clientManagerWidget, &ClientManagerWidget::client_config_changed,
            mainWindow, &MainWindow::client_config_receiver,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::runStart,
            clientManagerWidget, &ClientManagerWidget::onRestartAll,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::sendControlEnable,
            clientManagerWidget, &ClientManagerWidget::setControlEnable,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::runStop,
            clientManagerWidget, &ClientManagerWidget::onStopAll,
            Qt::UniqueConnection);

    connect(clientManagerWidget, &ClientManagerWidget::backPressure,
            rcCoreApp, &RcCoreApp::onBackPressure,
            Qt::UniqueConnection);

    connect(clientManagerWidget, &ClientManagerWidget::vmeFlagUpdated,
            rcCoreApp, &RcCoreApp::onVmeFlagUpdated,
            Qt::UniqueConnection);

    connect(clientManagerWidget, &ClientManagerWidget::runStateChanged,
            rcCoreApp, &RcCoreApp::clientManagerRunStateChanged);

    connect(clientManagerWidget, &ClientManagerWidget::runStateChanged,
            rcCoreApp, &RcCoreApp::clientManagerRunStateChanged_);

    connect(rcCoreApp, &RcCoreApp::printFsmInput,
            clientManagerWidget, &ClientManagerWidget::printFsmInput);
}

void RcSignalRouter::connect_Configuration()
{
    rcConfigDispatcher->subscribe(mainWindow);
    rcConfigDispatcher->subscribeCS(mainWindow);

    rcConfigDispatcher->subscribe(rcCoreApp);
    rcConfigDispatcher->subscribeCS(rcCoreApp);

    rcConfigDispatcher->subscribeCS(configurationManager);

    rcConfigDispatcher->subscribe(configDb);
    rcConfigDispatcher->subscribeCS(configDb);
}

void RcSignalRouter::connect_EvNumChecker()
{
    if(!evNumChecker)
        return;

    if(vmeTrigsController) {
        connect(evNumChecker, &EvNumChecker::evNumRequest,
                vmeTrigsController, &VmeTrigsController::evNumRequest,
                Qt::UniqueConnection);

        connect(vmeTrigsController, &VmeTrigsController::evNumResponse,
                evNumChecker, &EvNumChecker::gotVmeTrigEvNum,
                Qt::UniqueConnection);

        connect(vmeTrigsController, &VmeTrigsController::vmeMapUpdated,
                evNumChecker, &EvNumChecker::updateVmeMap,
                Qt::UniqueConnection);
    }

    if(remoteControlServer) {
        connect(remoteControlServer, &RemoteControlServer::evNumCheckRequest,
                evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest,
                Qt::UniqueConnection);

        connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
                remoteControlServer, &RemoteControlServer::evNumCheckResponse,
                Qt::UniqueConnection);
    }

    if(clientManagerWidget) {
        connect(evNumChecker, &EvNumChecker::evNumRequest,
                clientManagerWidget, &ClientManagerWidget::evNumRequest,
                Qt::UniqueConnection);

        connect(clientManagerWidget, &ClientManagerWidget::evNumResponse,
                evNumChecker, &EvNumChecker::gotClientEvNum,
                Qt::UniqueConnection);

        connect(clientManagerWidget, &ClientManagerWidget::clientMapUpdated,
                evNumChecker, &EvNumChecker::updateClientMap,
                Qt::UniqueConnection);
    }
}

void RcSignalRouter::connect_EvNumChecker(const TrigDevice *trigDevice)
{
    if(!evNumChecker || !trigDevice)
        return;

    connect(evNumChecker, &EvNumChecker::getTrigEvNum,
            trigDevice, &TrigDevice::evNumRequest,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::trigEvNumResponse,
            evNumChecker, &EvNumChecker::gotTopLevelEvNum,
            Qt::UniqueConnection);

    connect(trigDevice, &TrigDevice::devEvNumResponse,
            evNumChecker, &EvNumChecker::gotDevEvNum,
            Qt::UniqueConnection);
}



void RcSignalRouter::connect_SlowControlServer()
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(slowControlServer);

    connect(rcCoreApp, &RcCoreApp::runStart,
            slowControlServer, &SlowControlServer::onRunStart,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::destroyAll,
            slowControlServer, &SlowControlServer::deleteLater,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::allowSlowControl,
            slowControlServer, &SlowControlServer::startListen,
            Qt::UniqueConnection);

    connect(slowControlServer, &SlowControlServer::connectionStatusUpdated,
            mainWindow, &MainWindow::onSlowCntStatusUpdated,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::slowControlStateUpdated,
            slowControlServer, &SlowControlServer::slowControlStateUpdated,
            Qt::UniqueConnection);
}

void RcSignalRouter::connect_TrigMainForm()
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(trigMainForm);

    connect(mainWindow, &MainWindow::sendControlEnable,
            trigMainForm, &TrigMainForm::setControlEnable,
            Qt::UniqueConnection);

    if(rcFsm)
        connect(rcFsm, &RcFsm::stateChanged,
                mainWindow, &MainWindow::onRcFsmStateChange);
    connect(rcCoreApp, &RcCoreApp::clientManagerRunStateChanged_,
            mainWindow, &MainWindow::clientManagerRunStateChanged);
    connect(rcCoreApp, &RcCoreApp::vmeTrigsRunStateChanged_,
            mainWindow, &MainWindow::vmeTrigsRunStateChanged);
    connect(rcCoreApp, &RcCoreApp::trigManagerRunStateChanged_,
            mainWindow, &MainWindow::trigManagerRunStateChanged);
}

void RcSignalRouter::connect_VmeTrigsController()
{
    Q_ASSERT(mainWindow);
    Q_ASSERT(vmeTrigsController);

    connect(vmeTrigsController, &VmeTrigsController::sendSlaveMessage,
            rcCoreApp, &RcCoreApp::sendSlaveMessage,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::runStart,
            vmeTrigsController, &VmeTrigsController::onRunStart,
            Qt::UniqueConnection);

    connect(mainWindow, &MainWindow::sendControlEnable,
            vmeTrigsController, &VmeTrigsController::setControlEnable,
            Qt::UniqueConnection);

    connect(rcCoreApp, &RcCoreApp::runStop,
            vmeTrigsController, &VmeTrigsController::onRunStop,
            Qt::UniqueConnection);

    connect(vmeTrigsController, &VmeTrigsController::stateChanged,
            rcCoreApp, &RcCoreApp::vmeTrigsRunStateChanged);

    connect(vmeTrigsController, &VmeTrigsController::stateChanged,
            rcCoreApp, &RcCoreApp::vmeTrigsRunStateChanged_); //proxy

    connect(rcCoreApp, &RcCoreApp::printFsmInput,
            vmeTrigsController, &VmeTrigsController::printFsmInput);
}

void RcSignalRouter::connect_TrigDeviceController()
{
    Q_ASSERT(trigDeviceController);

    connect(rcCoreApp, &RcCoreApp::runStart,
            trigDeviceController, &TrigDeviceController::onRunStart,
            Qt::UniqueConnection);
    connect(rcCoreApp, &RcCoreApp::prepareStop,
            trigDeviceController, &TrigDeviceController::onPrepareStop,
            Qt::UniqueConnection);
    connect(rcCoreApp, &RcCoreApp::runStop,
            trigDeviceController, &TrigDeviceController::onRunStop,
            Qt::UniqueConnection);
    connect(rcCoreApp, &RcCoreApp::closeGlobTrig,
            trigDeviceController, &TrigDeviceController::onCloseGlobTrig,
            Qt::UniqueConnection);
    connect(rcCoreApp, &RcCoreApp::openGlobTrig,
            trigDeviceController, &TrigDeviceController::onOpenGlobTrig,
            Qt::UniqueConnection);
    connect(trigDeviceController, &TrigDeviceController::stateChanged,
            rcCoreApp, &RcCoreApp::trigManagerRunStateChanged);
    connect(trigDeviceController, &TrigDeviceController::stateChanged,
            rcCoreApp, &RcCoreApp::trigManagerRunStateChanged_);
    connect(mainWindow, &MainWindow::printFsmInput,
            trigDeviceController, &TrigDeviceController::printFsmInput);
}
