/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigDeviceController_p.h"
#include "TrigDeviceController.h"

#include "TrigDeviceMgrFsm.h"

TrigDeviceControllerPrivate::TrigDeviceControllerPrivate(TrigDeviceController *p, QObject *parent)
    : QObject(parent), q_ptr(p), fsm(new TrigDeviceMgrFsm(this))
{
    connect(fsm, &DeviceMgrFsm::stateChanged,
            q_ptr, &TrigDeviceController::stateChanged);
    connect(fsm, &DeviceMgrFsm::stateChanged,
            this, &TrigDeviceControllerPrivate::onFsmStateChange);
    connect(this, &TrigDeviceControllerPrivate::printFsmInput,
            fsm, &DeviceMgrFsm::printFsmInput);
}

void TrigDeviceControllerPrivate::onFsmStateChange(const FsmState &state)
{
    switch (state) {
    case FsmState::Error:
        setTrigState(false);
    case FsmState::Ready:
        megamaster_flg = false;
        break;
    case FsmState::Starting:
    {
        if(megamaster_flg) {
            setTrigState(false);
            emit preRunReset();
        } else {
            emit preRunReset();
            setTrigState(true);
        }
        qInfo() << __PRETTY_FUNCTION__ << state;
        fsm->goState(megamaster_flg ? FsmState::CloseGlobTrig : FsmState::Run);
        break;
    }
    case FsmState::CloseGlobTrig:
//        fsm->goState(FsmState::Run);
        break;
    case FsmState::OpenGlobTrig:
        setTrigState(true);
        fsm->goState(FsmState::Run);
        break;
    case FsmState::Stopping:
        setTrigState(false);
        fsm->goState(FsmState::Ready);
        break;
    default:
    {}
    }
}

void TrigDeviceControllerPrivate::setTrigStatus(const TrigStatus &status)
{
    Q_Q(TrigDeviceController);
    trigStatus[status.devIndex] = status;

    if(lookingForStopEvNum) {
        qDebug() << "StopEvNum detected as:" << status.globEvNum;
        stopEvNumber = status.globEvNum;
        fsm->markStopEvNumber();
        lookingForStopEvNum = false;
    }
}

void TrigDeviceControllerPrivate::setupWritten(const DeviceIndex &index, int setupSeqNum)
{
    fsm->setDevSetupKey(index, setupSeqNum);
    if(fsm->getState() == FsmState::Stopping)
        lookingForStopEvNum = true;
}

void TrigDeviceControllerPrivate::incrementSetupKey()
{
    if(++setupKey<100) setupKey = 100;
    fsm->setMainSetupKey(setupKey);
}

void TrigDeviceControllerPrivate::onTrigDevConfigUpdate()
{
    incrementSetupKey();
    emit writeSetup(config.rcTrigConfig, setupKey);
}

void TrigDeviceControllerPrivate::updateDeviceSet()
{
//    DeviceIndexSet set = makeDeviceIndexSet(trigThreads.keys());
//    DeviceIndexEnabledMap map;
    //    fsm->setDeviceList(map);
}

void TrigDeviceControllerPrivate::onOpenGlobTrig()
{
    Q_ASSERT(megamaster_flg);

    setTrigState(true);
    fsm->goState(fsm->getState()==FsmState::CloseGlobTrig ? FsmState::OpenGlobTrig : FsmState::Error);
}

void TrigDeviceControllerPrivate::setTrigState(bool open)
{
    bool hasEnabledModules = !config.rcTrigConfig.deviceIndex.isNull();
    if(hasEnabledModules) {
        if(config.rcTrigConfig.trig.trigOpen != open) {
            config.rcTrigConfig.trig.trigOpen = open;
            onTrigDevConfigUpdate();
            Q_Q(TrigDeviceController);
            emit q->configChanged(config);
        }
    }
    if(fsm->getState()==FsmState::Stopping) {
        if(hasEnabledModules && megamaster_flg)
            lookingForStopEvNum = true;
        else {
            qDebug() << "Mark StopEvNum as zero.";
            stopEvNumber = 0;
            fsm->markStopEvNumber();
        }
    }
}
