#include "TrigDeviceMgrFsm.h"

#include <QDebug>
#include <QLoggingCategory>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.TrigDeviceMgrFsm")

struct Transaction
{
    FsmState from;
    FsmState to;
    QString reason;
    QString action;
};

// MM - megamaster flag
const QList<Transaction> TransactionMap = {
    {FsmState::Init, FsmState::Error, "", ""},
    {FsmState::Error, FsmState::Ready, "cfg written to device", ""},
    {FsmState::Ready, FsmState::Starting, "[!MM] on remote start request (external slot)", "open trigger in cfg and write to dev"},
    {FsmState::Ready, FsmState::Starting, "[MM] on local start request (external slot)", "close trigger in cfg and write to dev"},
    {FsmState::Starting, FsmState::Run, "[!MM] on cfg (with opened trig) written to device", ""},
    {FsmState::Starting, FsmState::CloseGlobTrig, "[MM] on cfg (with closed trig) written to device", "then wait while all client will be in run state"},
    {FsmState::CloseGlobTrig, FsmState::OpenGlobTrig, "[MM] all system is ready to accept triggers (external slot)", "open trigger in cfg and write to dev"},
    {FsmState::OpenGlobTrig, FsmState::Run, "[MM] on cfg (with opened trig) written to device", ""},
    {FsmState::Run, FsmState::Stopping, "[!MM] on remote stop request", "close trigger in cfg and write to dev"},
    {FsmState::Stopping, FsmState::Ready, "[!MM] on cfg (with closed trig) written to device", ""},
    {FsmState::Unknown, FsmState::Error, "devSet is empty", ""},
    {FsmState::Unknown, FsmState::Error, "", "close trig"},
};
}

TrigDeviceMgrFsm::TrigDeviceMgrFsm(QObject *p):
    DeviceMgrFsm(p)
{
    setObjectName("TrigDeviceMgrFsm");
    recreateInput(new TrigDeviceMgrFsmInput());
}

bool TrigDeviceMgrFsm::evaluate()
{
    if(getInput()->devSet.isEmpty()) {
        setState(FsmState::Error);
        return true;
    }

    if(DeviceMgrFsm::evaluate())
        return true;

    switch (getState()) {
    case FsmState::Starting:
        if(allDevicesWritten() && getRequestState() == FsmState::CloseGlobTrig) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::CloseGlobTrig:
        if (getRequestState() == FsmState::Stopping) {
            acceptRequestedState();
            return true;
        }
        if (allDevicesWritten() &&
                getRequestState() == FsmState::OpenGlobTrig) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::OpenGlobTrig:
        if(allDevicesWritten() &&
                getRequestState() == FsmState::Run) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::Stopping:
        if(allDevicesWritten() && getInput()->hasStopEvNumber &&
                getRequestState() == FsmState::Ready) {
            acceptRequestedState();
            return true;
        }
        break;
    default:
    {}
    }

    return false;
}

void TrigDeviceMgrFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

void TrigDeviceMgrFsm::markStopEvNumber()
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName();
    getInput()->hasStopEvNumber = true;
    evaluate();
}

qint64 TrigDeviceMgrFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
//    case FsmState::CloseGlobTrig: return 20e3;
    case FsmState::OpenGlobTrig: return DEVICE_CONFIG_WRITE_TIMEOUT;
    default:
        return DeviceMgrFsm::stateTimeoutMs(state);
    }
}
