﻿#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <QBitArray>
#include <QBoxLayout>
#include <QClipboard>
#include <QFileDialog>
#include <QJsonDocument>
#include <QHostAddress>
#include <QPlainTextEdit>
#include <QScrollBar>
#include <QSemaphore>
#include <QSettings>
#include "QShortcut"
#include <QSpinBox>
#include <QSplitter>
#include <QStandardPaths>
#include <QTableWidgetItem>
#include <QThread>
#include <QTimer>
#include <QTreeWidgetItem>
#ifdef __linux__
#include <signal.h>
#endif

#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "device-discover/DiscoverDialog.h"
#include "fsm/DeviceMgrFsm.h"
#include "lib-common/UIStyles.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mregdev/device_id.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "remote-control-server/RunControlCommands.h"
#include "TdcSignalRouter.h"

#include "LogDialog.h"
#include "AdjustDialog.h"
#include "OrderDialog.h"
#include "TdcConfig.h"
#include "TdcConfigDb.h"
#include "TdcConfigDispatcher.h"
#include "FeLinkDialog/FeLinkStatusDialog.h"

namespace {
    enum {COL_DEV_SERIAL, COL_DEV_SLOT, COL_DEV_FW, COL_DEV_MCU_VER, COL_DEV_PLL_UNLOCK, COL_DEV_PLL_STATUS, COL_DEV_TEMP, COL_DEV_POWER, COL_DEV_TIME,
          COL_DEV_RUN_TIME, COL_DEV_CUR_EV_NUM, COL_DEV_UNBLOCK, COL_DEV_BLOCK, COL_DEV_MS_ERR,
          COL_DEV_STATUS, COL_HPTDC_STATUS,
          COL_BUSY_TIME, COL_BUSY_STATUS,COL_TRIG_QUEUE,
          COL_CLOCK_STATUS, COL_DEV_STREAM, COL_ERR_CNT, COL_DEV_TOTAL};
    enum {
        tree_tdc_name, // moduleName or modCh
        tree_tdc_num_cols
    };
    const QString BUSY_STATUS_TOOL_TIP = "Current Busy Status\n"
                                         "1x - TDC busy\n"
                                         "x1 - ReadOut busy";
    const QString BUSY_TIME_TOOL_TIP = "Moment Busy time\n"
                                       "1) TDC busy\n"
                                       "2) ReadOut busy\n"
                                       "3) Total busy";
    const QString DEV_FW_TOOL_TIP = "Doubli-click copies product-status information to clipboard";
    const int REMOTE_CONTROL_PORT = 0;

    void setItemTextColor(QTableWidgetItem* const item, QColor col, bool isVertical=false) {
        if(!item)
            return;
        if(isVertical) {
            item->setTextColor(col);
        } else {
            auto brush = item->foreground();
            brush.setColor(col);
            item->setForeground(brush);
        }
    }
    void setItemBgColor(QTableWidgetItem* const item, QColor col) {
        if(!item)
            return;
        auto brush = item->background();
        brush.setColor(col);
        brush.setStyle(Qt::SolidPattern);
        item->setBackground(brush);
    }
    void setTableItem(QTableWidget *tableDev, int row, int col) {
        bool blockSignals = tableDev->blockSignals(true);
        if(col==COL_CLOCK_STATUS) {
            auto label = new QLabel(tableDev);
            label->setAlignment(Qt::AlignCenter);
            tableDev->setCellWidget(row, col, label);
        } else {
            auto item = new QTableWidgetItem();
                item->setTextAlignment(Qt::AlignCenter);
            if(col == COL_DEV_MS_ERR) {
                setItemTextColor(item, Qt::red);
            }
            tableDev->setItem(row, col, item);
        }
        tableDev->blockSignals(blockSignals);
    }

    class TableSortLocker {
    public:
        TableSortLocker(QTableWidget *t) :
            table(t),
            sort_en(t->isSortingEnabled())
        {
            if(sort_en) table->setSortingEnabled(false);
        }
        ~TableSortLocker() {
            if(sort_en) table->setSortingEnabled(true);

        }
        QTableWidget * const table;
        const bool sort_en;
    };

    class TableUpdateLocker {
    public:
        TableUpdateLocker(QWidget *t) :
            w(t),
            up_en(t->updatesEnabled())
        {
            if(up_en) w->setUpdatesEnabled(false);
        }
        ~TableUpdateLocker() {
            if(up_en) w->setUpdatesEnabled(true);

        }
        QWidget * const w;
        const bool up_en;
    };

    void initTableItemAtRow(QTableWidget *tableDev, int row) {
        auto sortLocker = TableSortLocker(tableDev);
        for(int col=0; col<COL_DEV_TOTAL; ++col) {
            setTableItem(tableDev, row, col);
        }
        tableDev->setVerticalHeaderItem(row, new QTableWidgetItem());
        tableDev->item(row, COL_BUSY_STATUS)->setToolTip(BUSY_STATUS_TOOL_TIP);
        tableDev->item(row, COL_BUSY_TIME)->setToolTip(BUSY_TIME_TOOL_TIP);
        tableDev->item(row, COL_DEV_FW)->setToolTip(DEV_FW_TOOL_TIP);
        if(row==0) {
            tableDev->item(row, COL_DEV_SERIAL)->setText("No device selected");
            tableDev->verticalHeaderItem(row)->setText("No device selected");
        }
    }

    QString getDevMetricPrefix(DeviceIndex i) {
        return QString("%1.").arg(i.getSerialStr());
    }

    void setLabelColor(QLabel * const lab, QColor col) {
        if(!lab) return;
        QPalette p = lab->palette();
        p.setColor(lab->foregroundRole(), col);
        lab->setPalette(p);
    }
}

MainWindow::MainWindow(StartupOptions &start_options, QWidget *parent) :
      BaseMainWindow(parent)
    , ui(new Ui::MainWindow)
    , fsm(new DeviceMgrFsm())
    , signalRouter(new TdcSignalRouter())
    , current_cs(start_options.cs)
    , configuration_manager(new ConfigurationManager(this, current_cs.getType()))
    , tdcConfigDb(new TdcConfigDb(start_options.cs, this))
    , tdcConfigDispatcher(new TdcConfigDispatcher())
    , write_config_timer(new QTimer(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , discoverDialog(new DiscoverDialog(this))
    , feLinkStatusDialog(new FeLinkStatusDialog(this))
    , remoteControl(new RemoteControlServer(this, REMOTE_CONTROL_PORT))
    , pnpServer(new PNPServer(this))
    , evNumChecker(new EvNumChecker(this))
    , logDialog(new LogDialog(this))
    , adjustDialog(Globals::getDebugMode() ? new AdjustDialog(this) : nullptr)
    , rootTreeItem(new QTreeWidgetItem(QStringList("All")))
    , deviceStatusTimer(new QTimer(this))
    , labelFsmState(new QLabel(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<TdcStatus>();
    qRegisterMetaType<QMap<int, FeLinkStatus> >();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<QMap<int,double> >();
    qRegisterMetaType<QMap<int,int> >();
    qRegisterMetaType<AdjustConfig>();
    qRegisterMetaType<TdcConfig>();
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    configuration_manager->setDefaultConfig(TdcConfig());
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->setMainWindow(this);
    signalRouter->setPnpServer(pnpServer);
    signalRouter->setRemoteControlServer(remoteControl);
    signalRouter->setConfigDb(tdcConfigDb);
    signalRouter->setConfigDispatcher(tdcConfigDispatcher);
    signalRouter->setEvNumChecker(evNumChecker);
    signalRouter->setFsm(fsm);

    tdcConfigDispatcher->subscribe(tdcConfigDb);
    tdcConfigDispatcher->subscribe(this);

    new ThreadController(remoteControl, this);
    new ThreadController(pnpServer, this);

    const auto &device_filter_list = makeDeviceIdSet({DEVICE_ID_TDC72VXS, DEVICE_ID_TDC64VHLE});
    LldpDiscovery::subscribe(this, device_filter_list);
    LldpDiscovery::subscribe(deviceDispatcher, device_filter_list);

    discoverDialog->setSingleMode(false);
    LldpDiscovery::subscribe(discoverDialog, device_filter_list);

    deviceStatusTimer->setSingleShot(true);
    deviceStatusTimer->start(500);

    write_config_timer->setSingleShot(true);

    initConnections();
    initTable();
    initTree();
    initShortcut();
    init_gui();
    updateControlState();
}

MainWindow::~MainWindow()
{
    emit closeProgram(program_description);

    //todo move write_config_timer to ...
    if(write_config_timer->isActive()) {
        write_config_timer->stop();
        emit configChanged(tdcConfig);
    }
    for (const DeviceIndex &i: devGlob.keys()) {
        removeDevice(i);
    }

    emit closeAll();
    QElapsedTimer closeTimer;
    closeTimer.start();
    for (const DeviceIndex &i: devGlob.keys()) {
        if(!threadQuiting(devGlob[i]->devThread, 2000, &closeTimer)) {
            qWarning() << "Can't stop device thread";
        }
    }

    delete ui;
}

//=== public slots:
void MainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    ConfigSelector cs(DaqModule(DaqModule::Type::TdcControl), conf);
    setWindowTitle(cs.toWindowTitle());

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.tdc2-ctrl." + shortHostName + "."
                               + cs.getIndex().toString());
}

void MainWindow::setConfig(const TdcConfig &newConfig)
{
    if (newConfig == tdcConfig)
        return;
    tdcConfig = newConfig;

    block_widgets_signal(true);
    ui->actionon_board_PLL_bypass->setChecked(tdcConfig.default_device_config.pll_bypass);
    if(adjustDialog) {
        adjustDialog->setConfig(tdcConfig.default_device_config.adjust_set);
    }
    updateControlState();

    ui->actionAllow_remote_control->setChecked(tdcConfig.allow_remote_control);
    emit allowRemoteControl(tdcConfig.allow_remote_control);
    block_widgets_signal(false);

    if(tdcConfig.sort_column >= 0) {
        ui->tableWidgetDevice->sortByColumn(
            tdcConfig.sort_column,
            tdcConfig.sort_asc ? Qt::AscendingOrder : Qt::DescendingOrder);
    }
    ui->actionConfiguration_window->setChecked(tdcConfig.show_cfg_widget);

    for (const DeviceIndex &i: devGlob.keys()) {
        if(!tdcConfig.known_configs.contains(i) ||
                !tdcConfig.known_configs[i].enabled)
            removeDevice(i);
    }

    DeviceIndexEnabledMap devicesAll;
    for (const TdcDeviceConfig &ds: tdcConfig.known_configs.values()) {
        devicesAll[ds.index] = DeviceIndexEnabled(ds.index, ds.enabled);
        if(ds.enabled && !devGlob.contains(ds.index))
            addDevice(ds.index);
    }

    configUpdated();
    updateDeviceOrder();

    emit deviceListChanged(devicesAll);
    (tdcConfig.global_mode ? ui->radio_button_global : ui->radio_button_separate)
            ->setChecked(true);
    ui->actionTrace_dev_info_changes->setChecked(tdcConfig.trace_device_info);
    ui->actionShow_power_current->setChecked(tdcConfig.show_power_current);

}

void MainWindow::lldpCacheUpdated(const LldpInfoCache &cache)
{
    lldpCache = cache;

    auto sortLocker = TableSortLocker(ui->tableWidgetDevice);

    for(const auto &lldp : lldpCache.values()) {
        const int row = getRowByDevice(DeviceIndex(lldp));
        if(row == -1)
            continue;

        QString txt = lldp.isNull() ? "-" :
                                      QString("%1").arg(lldp.chassis_slot, 2, 10, QChar('0'));

        auto itemSlot = ui->tableWidgetDevice->item(row, COL_DEV_SLOT);
        if(itemSlot && itemSlot->text() != txt) {
            itemSlot->setText(txt);
            updateModelList();
        }
    }
}

void MainWindow::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    labelFsmState->setText("FSM: " + fsmStateToString(state));
    auto label_remote_status = ui->label_remote_status;

    switch (state) {
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            if(tdcConfig.remote_start_request){
                qInfo("Remote stop completed successfully");
                label_remote_status->setText(QString("Run finished: %1")
                    .arg(tdcConfig.runId.toString()));
                setLabelColor(label_remote_status, Qt::black);
            }
            else
                qInfo("Stop completed successfully");
            break;
        default:{}
        }
        break;
    }
    case FsmState::Starting :
    {
        setLabelColor(label_remote_status, QColor("orange"));
        label_remote_status->setText("Preparing for start...");
        on_actionReset_MStream_triggered();
        perform_start();
        const auto &cfgJson = QVariant::fromValue(tdcConfig).value<QJsonObject>();
        emit config_transfer_to_rc(QJsonDocument(cfgJson).toJson());
        fsm->goState(FsmState::Run);
        break;
    }
    case FsmState::Run :
    {
        if(tdcConfig.remote_start_request){
            qInfo() << QString("Remote start completed successfully for run %1")
                       .arg(tdcConfig.runId.toString());
            label_remote_status->setText(QString("Running %1...")
                                  .arg(tdcConfig.runId.toString()));
            setLabelColor(label_remote_status, Qt::darkGreen);
        }
        else
            qInfo() << QString("Start completed successfully");

        break;
    }
    case FsmState::Stopping :
    {
        setLabelColor(label_remote_status, QColor("orange"));
        label_remote_status->setText("Preparing for stop...");
        perform_stop();
        fsm->goState(FsmState::Ready);
        break;
    }
    case FsmState::Error :
    {
        if(oldState == FsmState::Starting) {
            qWarning() << "Something was wrong while executing remote start";
            auto label_status = ui->label_remote_status;
            label_status->setText(QString("Failed to start run %1")
                .arg(tdcConfig.runId.toString()));
            setLabelColor(label_status, Qt::red);
        }
        if(oldState == FsmState::Stopping) {
            qWarning("Something was wrong while executing remote stop");
            auto label_status = ui->label_remote_status;
            label_status->setText(QString("Errors on run %1...")
                .arg(tdcConfig.runId.toString()));
            setLabelColor(label_status, Qt::red);
        }
    }

    default:
    {}
    }
}

void MainWindow::setDeviceList(const DeviceIndexEnabledMap &list)
{
    QMap<DeviceIndex, TdcDeviceConfig> oldSetups = tdcConfig.known_configs;
    QMap<DeviceIndex, TdcDeviceConfig> newSetups;
    for (const auto &die: list) {
        const DeviceIndex index = die.index;
        auto &setup = newSetups[index] = tdcConfig.getConfig(index);
        setup.enabled = die.enabled;
    }

    DeviceIndex selected_device;
    if(newSetups.contains(tdcConfig.curDev) && newSetups[tdcConfig.curDev].enabled)
        selected_device = tdcConfig.curDev;

    // clean-up
    for (const DeviceIndex &index: devGlob.keys()) {
        if(!newSetups.contains(index) || !newSetups[index].enabled)
            removeDevice(index);
    }

    tdcConfig.known_configs = newSetups;
    // add new
    for (const DeviceIndex &index: newSetups.keys()) {
        if(!newSetups[index].enabled || devGlob.contains(index))
            continue;

        addDevice(index);
        if(index==selected_device) {
            int row = getRowByDevice(selected_device);
            Q_ASSERT(row!=-1);
            ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->setSelected(true);
        }
    }

    write_config_request();
    configUpdated();
    updateDeviceOrder();
    updateControlState();
    updateProgramDescription();
}
void MainWindow::onRemoteStartRequest(const RcRunId &runId,
                                      quint64 start_event_number)
{
    Q_UNUSED(start_event_number)
    qInfo() << QString("Got remote request to start Run %1")
               .arg(runId.toString());
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(msg);
        return;
    }
    if(tdcConfig.runId.runIndex == runId.runIndex &&
            tdcConfig.runId.runNumber > runId.runNumber) {
        qWarning() <<
            QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                .arg(tdcConfig.runId.runNumber)
                .arg(runId.runNumber);
    }
    tdcConfig.runId = runId;
    tdcConfig.remote_start_request = true;
    fsm->goState(FsmState::Starting);
}

void MainWindow::onRemoteStopRequest(const RcRunId &runId,
                                     quint64 start_event_number)
{
    Q_UNUSED(start_event_number)
    auto label_status = ui->label_remote_status;
    label_status->setText("Stopping run...");
    setLabelColor(label_status, QColor("orange"));
    bool success=true;
    if(tdcConfig.runId.runIndex != runId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                     .arg(tdcConfig.runId.runIndex)
                     .arg(runId.runIndex);
        success = false;
    }
    if(success && tdcConfig.runId.runNumber != runId.runNumber){
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                     .arg(tdcConfig.runId.runNumber)
                     .arg(runId.runNumber);
        success = false;
    }

    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(msg);
    }

    fsm->goState(FsmState::Stopping);
}

void MainWindow::remoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(!program_description.containsProgramInterface(ProgramInterfaceRemoteControl, 0)) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        interface.port = 0;
        interface.enabled = tdcConfig.allow_remote_control;
        interface.isFree = false;
        program_description.interfaces.push_back(interface);
    }
    const bool &port_in_use = status.portInUse;
    on_remote = port_in_use && status.listening;

    ProgramInterface &i = program_description.getProgramInterface(ProgramInterfaceRemoteControl);
    ProgramInterface prevIt = i;
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !port_in_use;

    if(port_in_use) {
        i.setPeer(status.peerHost, status.peerPort);
    } else {
        i.peers.clear();
        if(tdcConfig.run)
            perform_stop();
    }

    updateRemoteStatus();
    enable_gui_controls();
    ui->label_remote_status->setVisible(port_in_use);

    // lock all control widgets
    updateProgramDescription();
}
void MainWindow::sendRunWarning(QString mess)
{
    if(tdcConfig.isRun() && tdcConfig.allow_remote_control)
        emit runWarning(tdcConfig.runId,
            QString("TDC #%1 %2").arg(current_cs.getIndex(), mess));
}


//=== private:
void MainWindow::initConnections()
{
    signalRouter->connectAll();

    deviceDispatcher->subscribeDevices(this);
    deviceDispatcher->subscribeDevices(discoverDialog);
    deviceDispatcher->subscribeDevices(feLinkStatusDialog);
    LldpInfoCache lldpCache;



    connect(this, &MainWindow::set_run_state,
            ui->m_stream_output, &StatisticOutput::set_run_state);


    auto hhView = ui->tableWidgetDevice->horizontalHeader();
    connect(hhView, &QHeaderView::sortIndicatorChanged,
            this, &MainWindow::sortIndicatorChanged);


    connect(deviceStatusTimer, &QTimer::timeout,
            this, &MainWindow::applyDeviceStatusUpdates);

    connect(write_config_timer, &QTimer::timeout,
            this, &MainWindow::write_config_request);

    if(Globals::getDebugMode()) {
        connect(adjustDialog, &AdjustDialog::adjust_setup_changed,
                this, &MainWindow::adjust_setup_changed);
    }
}

void MainWindow::initTable()
{
    ++blockControl;
    QTableWidget * const tableDev = ui->tableWidgetDevice;
    tableDev->setSortingEnabled(true);
    tableDev->setColumnCount(COL_DEV_TOTAL);

    tableDev->setHorizontalHeaderItem(COL_DEV_SERIAL, new QTableWidgetItem("Serial"));
    tableDev->setHorizontalHeaderItem(COL_DEV_SLOT, new QTableWidgetItem("Slot"));
    tableDev->setHorizontalHeaderItem(COL_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));

    tableDev->setHorizontalHeaderItem(COL_DEV_PLL_UNLOCK, new QTableWidgetItem("PLL unlock"));
    tableDev->setHorizontalHeaderItem(COL_DEV_PLL_STATUS, new QTableWidgetItem("PLL st."));

    tableDev->setHorizontalHeaderItem(COL_DEV_TEMP, new QTableWidgetItem("temp"));
    tableDev->setHorizontalHeaderItem(COL_DEV_POWER, new QTableWidgetItem("power"));
    tableDev->setHorizontalHeaderItem(COL_DEV_TIME, new QTableWidgetItem("time"));
    tableDev->setHorizontalHeaderItem(COL_DEV_RUN_TIME, new QTableWidgetItem("Run time"));
    tableDev->setHorizontalHeaderItem(COL_DEV_CUR_EV_NUM, new QTableWidgetItem("Cur evNum"));
    tableDev->setHorizontalHeaderItem(COL_DEV_UNBLOCK, new QTableWidgetItem("Accepted"));
    tableDev->setHorizontalHeaderItem(COL_DEV_BLOCK, new QTableWidgetItem("Blocked"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STATUS, new QTableWidgetItem("Dev. St."));
    tableDev->setHorizontalHeaderItem(COL_HPTDC_STATUS, new QTableWidgetItem("HPTDC St."));
    tableDev->setHorizontalHeaderItem(COL_BUSY_TIME, new QTableWidgetItem("Busy time"));
    tableDev->setHorizontalHeaderItem(COL_BUSY_STATUS, new QTableWidgetItem("Busy st."));
    tableDev->setHorizontalHeaderItem(COL_TRIG_QUEUE, new QTableWidgetItem("Trig queue"));
    tableDev->setHorizontalHeaderItem(COL_CLOCK_STATUS, new QTableWidgetItem("Clock"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STREAM, new QTableWidgetItem("Data rate"));
    tableDev->setHorizontalHeaderItem(COL_ERR_CNT, new QTableWidgetItem("Error Cnt"));
    tableDev->setRowCount(1);
    initTableItemAtRow(tableDev, 0);
    tableDev->verticalHeader()->hide();
    tableDev->hideColumn(COL_DEV_TIME);
    tableDev->hideColumn(COL_DEV_MS_ERR);
    tableDev->hideColumn(COL_DEV_PLL_UNLOCK);
    tableDev->hideColumn(COL_DEV_PLL_STATUS);
    tableDev->hideColumn(COL_ERR_CNT);
    tableDev->hideColumn(COL_DEV_STATUS);
    if(!Globals::getDebugMode()) {
        tableDev->hideColumn(COL_DEV_POWER);
        tableDev->hideColumn(COL_DEV_CUR_EV_NUM);
        tableDev->hideColumn(COL_BUSY_STATUS);
        tableDev->hideColumn(COL_TRIG_QUEUE);
    }

    tableDev->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    tableDev->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize(tableDev);

    connect(tableDev->horizontalHeader(), &QHeaderView::sectionClicked, this, &MainWindow::update_order_index);

    tableDev->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
    tableDev->verticalScrollBar()->setSingleStep(tableDev->rowHeight(0)/3);
    --blockControl;
}

void MainWindow::initTree()
{
    ++blockControl;

    QTreeWidget *tree = ui->treeWidgetChCfg;
    tree->setColumnCount(tree_tdc_num_cols);

    QStringList headerLabels;
    headerLabels << "name"
                 << "ch";
//    tree->hideColumn(tree_tdc_ch);
    tree->setHeaderLabels(headerLabels);

    rootTreeItem->setCheckState(tree_tdc_name, Qt::PartiallyChecked);
    tree->insertTopLevelItem(0, rootTreeItem);
    rootTreeItem->setExpanded(true);
    tree->resizeColumnToContents(tree_tdc_name);
    --blockControl;

    tree->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(tree, &QTreeWidget::customContextMenuRequested,
            this, &MainWindow::customContextMenuRequested);
}

void MainWindow::initShortcut()
{
    QShortcut *shortcut = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_L), this);
    connect(shortcut, &QShortcut::activated, logDialog, &QWidget::show);
    connect(this, &MainWindow::addLog, logDialog, &LogDialog::addLog);

    if(adjustDialog) {
        QShortcut *shortcut_adjust = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_A), this);
        connect(shortcut_adjust, &QShortcut::activated, adjustDialog, &QWidget::show);
        connect(shortcut_adjust, &QShortcut::activated, adjustDialog, &AdjustDialog::activate_dialog);
    }

    QShortcut *shortcut_change_device = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_F), this);
    connect(shortcut_change_device, &QShortcut::activated, this, &MainWindow::on_pushButtonChangeDevice_clicked);
}


void MainWindow::init_gui()
{
    widgets_to_block.append(ui->comboBoxEdges);
    widgets_to_block.append(ui->comboBoxPreset);
    widgets_to_block.append(ui->actionAllow_remote_control);
    widgets_to_block.append(ui->actionon_board_PLL_bypass);
    widgets_to_block.append(ui->actionAdd_HPTDC_event_header);
    widgets_to_block.append(ui->actionAdd_HPTDC_event_tralier);
    widgets_to_block.append(ui->groupBoxStat);
    widgets_to_block.append(ui->spinBoxStatPeriod);

    ui->statusBar->addWidget(labelFsmState);
    ui->label_remote_status->hide();

    ui->labelPreset->setVisible(Globals::getDebugMode());
    ui->comboBoxPreset->setVisible(Globals::getDebugMode());
    ui->actionReset_power_status->setVisible(false);

    ui->groupBoxTrigger->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxReadout->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxStat->setStyleSheet(UIStyles::gbox_fill_blue);

    auto debug_mode = Globals::getDebugMode();
    ui->actionAdjust_HPTDC->setVisible(debug_mode);
    ui->actionTrace_dev_info_changes->setVisible(debug_mode);
    ui->actionShow_power_current->setVisible(debug_mode);

    ui->splitter_vertical->setCollapsible(0, false);
    ui->splitter_vertical->setCollapsible(1, false);
    ui->splitter_vertical->setStretchFactor(0,1);
    ui->splitter_vertical->setStretchFactor(1,2);

    block_widgets_signal(true);
    ui->comboBoxEdges->addItem("Leading", 1);
    ui->comboBoxEdges->addItem("Trailing", 2);
    ui->comboBoxEdges->addItem("Lead+Trail", 3);

    for(auto preset : StatPresetFabric::getAllPresets()){
        ui->comboBoxPreset->addItem(preset.label, preset.type);
    }
    block_widgets_signal(false);

    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.beginGroup("splitter_vertical");
    ui->splitter_vertical->restoreGeometry(settings.value("geometry").toByteArray());
    ui->splitter_vertical->restoreState(settings.value("state").toByteArray());
    settings.endGroup();
    settings.beginGroup("splitter_bottom");
    ui->splitter_bottom->restoreGeometry(settings.value("geometry").toByteArray());
    ui->splitter_bottom->restoreState(settings.value("state").toByteArray());
    settings.endGroup();
    settings.endGroup();

    setWindowTitle(current_cs.toWindowTitle());
}

void MainWindow::read_config()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.beginGroup("splitter_vertical");
    ui->splitter_vertical->restoreGeometry(settings.value("geometry").toByteArray());
    ui->splitter_vertical->restoreState(settings.value("state").toByteArray());
    settings.endGroup();
    settings.beginGroup("splitter_bottom");
    ui->splitter_bottom->restoreGeometry(settings.value("geometry").toByteArray());
    ui->splitter_bottom->restoreState(settings.value("state").toByteArray());
    settings.endGroup();
    settings.endGroup();
}


void MainWindow::write_config()
{
    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.beginGroup("splitter_vertical");
    settings.setValue("geometry", ui->splitter_vertical->saveGeometry());
    settings.setValue("state", ui->splitter_vertical->saveState());
    settings.endGroup();
    settings.beginGroup("splitter_bottom");
    settings.setValue("geometry", ui->splitter_bottom->saveGeometry());
    settings.setValue("state", ui->splitter_bottom->saveState());
    settings.endGroup();
    settings.endGroup();

    emit configChanged(tdcConfig);
}


void MainWindow::updateModelList()
{
    QTreeWidget *tree = ui->treeWidgetChCfg;

    tree->blockSignals(true);
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        rootTreeItem->removeChild(i);
        delete i;
    }
    modTreeItem.clear();

    int modChOffest = 0;
    int devCnt = 0;
    for (const DeviceIndex &index: devGlob.keys()) {
        QStringList list;
        QTreeWidgetItem * const modItem = new QTreeWidgetItem(rootTreeItem, list);
        modTreeItem[index] = modItem;
        modItem->setText(tree_tdc_name, getDeviceNameForTree(index, devCnt++));
        modItem->setCheckState(tree_tdc_name, Qt::PartiallyChecked);
        modItem->setExpanded(false);

        const TdcDeviceConfig &ds = tdcConfig.getConfig(index);
        const int modNch = ds.getNCh();
        for(int ch=0; ch<modNch; ++ch){
            QStringList list;
            list << QString::number(ch)
                 << QString::number(modChOffest+ch);
            QTreeWidgetItem * const chItem = new QTreeWidgetItem(modItem, list);
            Qt::CheckState st = ds.chTdcEn[ch] ? Qt::Checked : Qt::Unchecked;
            chItem->setCheckState(tree_tdc_name, st);
        }

        modChOffest += modNch;
    }
    tree->blockSignals(false);
    updateTreeSize(tree);
    updateTreeItemStates();
}

void MainWindow::updateTreeItemStates()
{
    ui->treeWidgetChCfg->blockSignals(true);
    Qt::CheckState allSt = rootTreeItem->checkState(tree_tdc_name);
    Qt::CheckState st = Qt::PartiallyChecked;
    bool firstMod=true;
    for (QTreeWidgetItem *i: modTreeItem.values()) {
        for(int chIndex=0; chIndex<i->childCount(); ++chIndex){
            Qt::CheckState childState = i->child(chIndex)->checkState(tree_tdc_name);
            if(chIndex==0){
                st = childState;
                continue;
            }
            if(st!=childState){
                st = Qt::PartiallyChecked;
                break;
            }
        }
        i->setCheckState(tree_tdc_name, st);
        if(firstMod) {
            allSt = st;
            firstMod = false;
        } else if(allSt != st)
            allSt = Qt::PartiallyChecked;
    }

    rootTreeItem->setCheckState(tree_tdc_name, allSt);
    ui->treeWidgetChCfg->blockSignals(false);
}

void MainWindow::configUpdated()
{
    if(blockControl) return;
    increase_setup_index();
    emit set_run_state(tdcConfig.run);
    updateRemoteStatus();
    write_config_request();

    for (const DeviceIndex &index: devGlob.keys())
        emit setupChanged(tdcConfig.getConfig(index));
}

void MainWindow::updateGuiVal()
{
    ++blockControl;
    TdcDeviceConfig ds;
    if(tdcConfig.global_mode) {
        if(tdcConfig.curDev.isNull()) {
            --blockControl;
            return;
        }
        ds = tdcConfig.getCurrentConfig();
    } else {
        ds = tdcConfig.getConfig(tdcConfig.curDev);
    }

    ui->checkBoxTrigTimer->setChecked(ds.trigsetup.bit.timer_en);
    ui->spinBoxTimeTrig->setEnabled(ds.trigsetup.bit.timer_en);
    int rate = static_cast<int>(125e6/tdcConfig.getCurrentConfig().trigPeriod/3);
    ui->spinBoxTimeTrig->setValue(rate);
    updateRealRate();

    ui->checkBoxTrigTtl->setChecked(ds.trigsetup.bit.ttl_en);
    ui->checkBoxTrigTtc->setChecked(ds.trigsetup.bit.ttc_en);
    ui->checkBoxTrigVxs->setChecked(ds.trigsetup.bit.vxs_en);

    ui->spinBoxLatency->setValue(static_cast<int>(ds.latency));
    ui->spinBoxMatchWin->setMaximum(static_cast<int>(ds.latency));
    ui->spinBoxMatchWin->setValue(static_cast<int>(ds.matchWin));
    int edgeData = 0;
    if(ds.leadEdge) edgeData |= 1;
    if(ds.trailEdge) edgeData |= 2;
    ui->comboBoxEdges->setCurrentIndex(ui->comboBoxEdges->findData(edgeData));

    block_widgets_signal(true);
    ui->actionAdd_HPTDC_event_header->setChecked(ds.hptdcHdr);
    ui->actionAdd_HPTDC_event_tralier->setChecked(ds.hptdcTrl);
    ui->groupBoxStat->setChecked(ds.stat_readout);
    ui->spinBoxStatPeriod->setValue(ds.stat_period);
    block_widgets_signal(false);


    ui->comboBoxPreset->setCurrentText(tdcConfig.default_device_config.stat_preset.label);
    --blockControl;
    updateControlState();
}

void MainWindow::updateControlState()
{
    bool enableControl = (blockControl==0) && (!on_remote || Globals::getDebugMode());
    const bool isRun = tdcConfig.isRun();
    ui->pushButtonChangeDevice->setEnabled(enableControl && !isRun);

    enableControl &= !tdcConfig.curDev.isNull();
    ui->frame_configuration->setEnabled(enableControl);
    ui->treeWidgetChCfg->setEnabled(enableControl);
    ui->pushButtonStart->setEnabled(!on_remote && enableControl && !isRun);
    ui->pushButtonStop ->setEnabled(!on_remote && enableControl &&  isRun);
}

void MainWindow::removeDevice(const DeviceIndex &index)
{
    auto dgIt = devGlob.find(index);
    if(dgIt != devGlob.end()) {
        auto &dg = dgIt.value();
        sendLog(dg->dev->getDeviceInfoStr().prepend("Disconnect from:"));
        emit deleteDevice(index);
        if(!threadQuiting(dg->devThread, 2000)){
            qWarning() << "Can't stop dev. h/w thread for "<< index.getIdent();
        }
        if(!threadQuiting(dg->msThread, 2000)){
            qWarning() << "Can't stop MStream thread for "<< index.getIdent();
        }
        ui->m_stream_output->remove_device(index);

        /// TODO: Check for memory leak
        devGlob.erase(dgIt);
    } else {
        qCritical() << "disconnectFromDev: no dev in devGlob map";
    }

    const int row = getRowByDevice(index);
    if(row != -1) {
        QTableWidget *tableDev = ui->tableWidgetDevice;
        if(tableDev->rowCount() == 1) {
            Q_ASSERT(row==0);
            initTableItemAtRow(tableDev, 0);
            tdcConfig.curDev = DeviceIndex();
        } else {
            tableDev->removeRow(row);
            if(index == tdcConfig.curDev)
                tableDev->item(0, COL_DEV_SERIAL)->setSelected(true);
        }
        updateTableSize(tableDev);
//        ui->tableWidgetDevice->sortByColumn(0, Qt::AscendingOrder);
    } else {
        qCritical() << "disconnectFromDev: no row in table";
    }

    updateControlState();
    update_pll_unlock_visible();
    updateModelList();
    table_warn_updated();
}

void MainWindow::addDevice(const DeviceIndex &index)
{
//  qInfo() << "==== Begin of connectToDev:" << index.getSerialStr() << " ====\n";
    Q_ASSERT(tdcConfig.known_configs.contains(index));
    auto tableDev = ui->tableWidgetDevice;
    auto sortLocker = TableSortLocker(tableDev);
    if(devGlob.contains(index))
        removeDevice(index); // FIXME: ???
    int row = -1;
    if(tableDev->rowCount() == 1 && getDevIndexByRow(0).isNull()) {
        row = 0;
    }
    if(row == -1) {
        row = tableDev->rowCount();
        tableDev->setRowCount(row+1);
        initTableItemAtRow(tableDev, row);
    }

    devGlob.insert(index, new DeviceGlob(index, Globals::getDebugMode()));
    auto &dg = devGlob[index];
    connect(this, &MainWindow::deleteDevice, dg, &DeviceGlob::deleteDevice);
    dg->warn_flags.resize(COL_DEV_TOTAL);
    for(int col=0; col<COL_DEV_TOTAL; ++col)
        dg->warn_flags[col] = false;

    TDCDevice *dev = dg->dev;
    connect(evNumChecker, &EvNumChecker::evNumRequest, dev, &TDCDevice::evNumRequest);
    connect(dev, &TDCDevice::evNumResponse, evNumChecker, &EvNumChecker::gotDevEvNum);

    bool blockSignals = tableDev->blockSignals(true);
    tableDev->verticalHeaderItem(row)->setText(index.getSerialStr());
    tableDev->item(row, COL_DEV_SERIAL)->setText(index.getSerialStr());
    tableDev->verticalHeaderItem(row)->setData(Qt::UserRole, QVariant::fromValue(index));
    tableDev->item(row, COL_DEV_SERIAL)->setData(Qt::UserRole, QVariant::fromValue(index));
    tableDev->item(row, COL_DEV_STATUS)->setText("connecting...");
    tableDev->item(row, COL_DEV_STREAM)->setText("");
    tableDev->item(row, COL_ERR_CNT)->setText("");
    updateTableSize(tableDev);

    if(tdcConfig.curDev.isNull()) {
        tdcConfig.curDev = index;
    } else if(index == tdcConfig.curDev) {
        tableDev->item(0, COL_DEV_SERIAL)->setSelected(true);
    }

    updateGuiVal();

    try{
        ui->m_stream_output->add_device(index);

        connect(this, &MainWindow::closeAll, dev, &QObject::deleteLater);
        connect(this, &MainWindow::deleteDevice, dev, &QtMregDevice::deleteHardware);

        connect(this, &MainWindow::setupChanged, dev, &TDCDevice::setSetup);
        connect(this, &MainWindow::setupChangedForce, dev, &TDCDevice::writeSetupForce);
        connect(ui->actionForce_write_setup, &QAction::triggered, dev, &TDCDevice::writeSetupForce);
        connect(this, &MainWindow::softReset, dev, &TDCDevice::softReset);
        connect(this, &MainWindow::blockDeviceRun, dev, &TDCDevice::blockDeviceRun);
        connect(this, &MainWindow::reset_error_counters, dev, &TDCDevice::reset_error_counters);
        connect(dev, &QtMregDevice::setStatus, this, &MainWindow::setDeviceStatus);
        connect(dev, &TDCDevice::deviceInfoUpdated, this, &MainWindow::deviceInfoUpdated);
        connect(dev, &TDCDevice::deviceSetupWritten, this, &MainWindow::deviceSetupWritten);
        connect(dev, &TDCDevice::statusUpdated, this, &MainWindow::deviceStatusUpdated);
        connect(dev, &TDCDevice::feLinkStatusUpdated, feLinkStatusDialog, &FeLinkStatusDialog::setFeLinkStatus);
        connect(dev, &QtMregDevice::deviceConnected, this, &MainWindow::deviceConnected);
        connect(dev, &QtMregDevice::deviceDisconnected, this, &MainWindow::deviceDisconnected);

        dg->startThreads();

        // Connecting
        const auto &lldpInfo = lldpCache.value(index);
        QString slot_str = lldpInfo.isNull() ? "-"
                                             : QString::number(lldpInfo.chassis_slot);
        tableDev->item(row, COL_DEV_SLOT)->setText(slot_str);
    } catch (std::exception &e){
        sendLog(QString("Connect to device %1").arg(e.what()));
    }
    tableDev->blockSignals(blockSignals);
    deviceDispatcher->subscribeHostAddressUpdates(index, dev);

//    qInfo() << "==== End of connectToDev:" << index.getSerialStr() << " ====\n";
    updateModelList();

    QVector<quint64> serials;
    for(const auto &ds: tdcConfig.known_configs.values()) {
        if(ds.enabled)
            serials.append(ds.index.getSerial());
    }
    evNumChecker->set_dev_serials(serials);
}

void MainWindow::sendLog(QString l)
{
    emit addLog(l);
}

void MainWindow::logDevInfo(QTableWidgetItem *item, bool force)
{
    if(!tdcConfig.trace_device_info) return;

    const int col = item->column();
    const int row = item->row();

    if(!force) {
        if(col == COL_DEV_RUN_TIME || col == COL_DEV_CUR_EV_NUM || col == COL_DEV_UNBLOCK || col == COL_DEV_BLOCK ||
           col == COL_DEV_STREAM || col == COL_BUSY_STATUS || col == COL_TRIG_QUEUE ||
           col == COL_BUSY_TIME ||  col == COL_DEV_TEMP || col == COL_DEV_POWER)
            return;
    }
// Columns to log:
//    COL_DEV_SERIAL; COL_DEV_SLOT; COL_DEV_FW; COL_DEV_MCU_VER; COL_DEV_PLL_UNLOCK; COL_DEV_PLL_STATUS; COL_DEV_TIME;
//    COL_DEV_MS_ERR;
//    COL_DEV_STATUS; COL_HPTDC_STATUS;
//    COL_CLOCK_STATUS; COL_ERR_CNT;

    auto newTxt = item->text();
    DeviceIndex index = getDevIndexByRow(row);
    auto devGlobIt = devGlob.find(index);
    if(devGlobIt == devGlob.end())
        return;

    QString &txt = devGlobIt.value()->tbl_info_log[col];
    if(newTxt == txt)
        return;

    QString colStr = ui->tableWidgetDevice->horizontalHeaderItem(col)->text();
    if(colStr.isEmpty()) colStr = QString::number(col);
    QString runTime;
    const auto &devSt = devGlob[index]->pendingDevStatus;
    if(!devSt.index.isNull() && devSt.connected)
        runTime = QString(" [RunTime:%2]").arg(devSt.statistic.getRunTime());
    qInfo().noquote() << QDateTime::currentDateTime().toString("hh:mm:ss")
                      << QString("[%1]%2 Dev. info change col='%3': '%4' -> '%5'")
                         .arg(index.getSerialStr(), runTime, colStr, txt, newTxt);
    txt = newTxt;
}

void MainWindow::updateRealRate()
{
    if(tdcConfig.curDev.isNull()) {
        ui->labelRealRate->setText("-");
    } else {
        const auto &ds = tdcConfig.getCurrentConfig();
        int realRateHz = static_cast<int>(125e6/ds.trigPeriod/3);
        ui->labelRealRate->setText(QString("%1 Hz").arg(realRateHz));
        ui->labelRealRate->setToolTip(QString("Trig period: %1*24ns = %2ns")
                                      .arg(ds.trigPeriod).arg(ds.trigPeriod*24));
    }
}

void MainWindow::update_pll_unlock_visible()
{
    const int col = COL_DEV_PLL_UNLOCK;
    auto table = ui->tableWidgetDevice;
    bool hidden = true;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->item(row, col)->data(Qt::UserRole).toInt()) {
            hidden = false;
            break;
        }
    }
    table->setColumnHidden(col, hidden);
}

void MainWindow::updateDeviceOrder(bool from_dialog)
{
    auto table = ui->tableWidgetDevice;
    auto updateLocker = TableUpdateLocker(table);
    auto sortLocker = TableSortLocker(table);

    if(!from_dialog)
        update_order_index(1);

    QMap<DeviceIndex, QList<QTableWidgetItem*> > upperItems;
    int row;
    QList<TdcDeviceConfig> devices;

    //deattach
    for(row=0; row<table->rowCount(); ++row) {
        DeviceIndex i = getDevIndexByRow(row);
        if(i.isNull()) break;
        upperItems[i] = OrderDialog::takeRow(table, row);
        Q_ASSERT(tdcConfig.known_configs.contains(i));
        devices.append(tdcConfig.known_configs[i]);
        rootTreeItem->takeChild(0);
    }

    std::sort(devices.begin(),devices.end());

    //attach
    row=0;
    bool blockSignals=table->blockSignals(true);
    for(TdcDeviceConfig ds : devices) {
        DeviceIndex index = ds.index;
        if(!devGlob.contains(index)) {
            Q_ASSERT(!ds.enabled); continue;
        }
        OrderDialog::setRow(table, row, upperItems[index]);
        Q_ASSERT(modTreeItem.contains(index));
        auto treeItem = modTreeItem[index];
        treeItem->setText(tree_tdc_name, getDeviceNameForTree(index, row));
        rootTreeItem->insertChild(row, treeItem);
        ++row;
    }
    table->blockSignals(blockSignals);
}

void MainWindow::update_tree_order()
{
    if(tdcConfig.known_configs.isEmpty())
        return;
    QTreeWidgetItem *item_all = ui->treeWidgetChCfg->topLevelItem(0);

    QMap<int, QTreeWidgetItem *> order_treewitem;
    for(const TdcDeviceConfig &setup : tdcConfig.known_configs.values()){
        if(!setup.enabled) continue;
        auto treeItem = modTreeItem.value(setup.index);
        Q_ASSERT(treeItem);
        int tree_index = item_all->indexOfChild(treeItem);
        Q_ASSERT(tree_index != -1);
        int order = tdcConfig.known_configs[setup.index].order_index;
        order_treewitem[order] = treeItem;
    }

    for(int order : order_treewitem.keys()) {
        item_all->removeChild(order_treewitem[order]);
    }

    QList<QTreeWidgetItem *> list_to_append;
    for(int key : order_treewitem.keys()) {
        list_to_append.append(order_treewitem[key]);
    }

    item_all->addChildren(list_to_append);
    ui->treeWidgetChCfg->addTopLevelItem(item_all);

    for(int i=0; i < item_all->childCount(); ++i) {
        QString str = item_all->child(i)->text(0);
        QString index_str = str.split(" ").at(3);
        for(int k : order_treewitem.keys()) {
            if(order_treewitem[k] == item_all->child(i)){
                str.remove(0,1);
                str.prepend(QString::number(k+1));
                item_all->child(i)->setText(0, str);
            }
        }
    }
}
void MainWindow::updateProgramDescription()
{
    program_description.type = current_cs.getType().toString();
    program_description.index = current_cs.getIndex().toString();
    program_description.name = program_description.getName();

#ifdef GIT_DESCR
    program_description.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    program_description.ver_date = GIT_LAST_COMMIT;
#endif

    program_description.options.clear();
    program_description.options["run"] = tdcConfig.run ? "true" : "false";
    program_description.options["run_index"] = tdcConfig.runId.runIndex;
    program_description.options["run_number"] = QString::number(tdcConfig.runId.runNumber);

    int i=0;
    for(const auto &ds : tdcConfig.known_configs.values()){
        if(!ds.enabled)
            continue;
        program_description.options[QString("TDC_%1").arg(i)] = QString("0x%1;0x%2")
                .arg(ds.index.getDevId(), 2, 16, QChar('0'))
                .arg(ds.index.getSerial(), 8, 16, QChar('0'));
        ++i;
    }
    emit changeProgramDescription(program_description);
}

void MainWindow::enable_gui_controls()
{
    const bool en = !on_remote || Globals::getDebugMode();

    ui->actionReset_MStream->setEnabled(en);
    ui->actionSend_soft_reset->setEnabled(en);
    ui->actionForce_write_setup->setEnabled(en);
    ui->actionAdjust_HPTDC->setEnabled(en);
    ui->actionon_board_PLL_bypass->setEnabled(en);
    ui->menuHPTDC_Event_hdr_trl->setEnabled(en);
    ui->actionConfiguration_manager->setEnabled(en);
    updateControlState();
}

void MainWindow::updateRemoteStatus()
{
    if(!on_remote || !update_remote_enabled)
        return;
    QString readyStr, notReadyStr;
    bool ready4Run = true;

    auto dg_it = devGlob.begin();
    while(dg_it != devGlob.end()) {
        const auto &dg = dg_it.value();
        const auto &curSt = dg->pendingDevStatus;
        const bool devOk = dg->setup_index == setup_index;
        const auto devStr = QString("0x%1 ").arg(dg_it.key().getSerial(), 8, 16, QChar('0'));
        if(devOk && curSt.connected
//                && dg.msStat == QProcess::Running
                ) {
            readyStr.append(devStr);
        } else {
            notReadyStr.append(devStr);
            ready4Run = false;
        }
        ++dg_it;
    }

    QString stateStr;
    if(!notReadyStr.isEmpty())
        stateStr.append("Not ready:").append(notReadyStr);
    if(!readyStr.isEmpty()) {
        if(!stateStr.isEmpty())
            stateStr.append("| ");
        stateStr.append("Ready:").append(readyStr);
    }
    if(stateStr.isEmpty())
        stateStr = "No devices";
    emit sendRemoteState(stateStr, ready4Run);
}

void MainWindow::perform_start()
{
    if(!tdcConfig.has_enabled_device())
        return;

    update_remote_enabled = false;
    tdcConfig.run = true;
    tdcConfig.default_device_config.run = true;
    for (auto &ds : tdcConfig.known_configs) {
        if(ds.enabled) {
            ds.run = true;
        }
    }

    configUpdated();
    update_remote_enabled = true;
    updateProgramDescription();
}



bool MainWindow::perform_stop()
{
    increase_setup_index();
    bool noDevice = true;

    update_remote_enabled = false;
    tdcConfig.run = false;
    tdcConfig.default_device_config.run = false;
    for (const DeviceIndex &index: devGlob.keys()) {
        TdcDeviceConfig &ds  = tdcConfig.getConfig(index);
        if(ds.enabled) {
            ds.run = false;
            noDevice = false;
        }
    }
    if(noDevice)
        return true;
    configUpdated();

    const bool ok = wait_device_setup_written();
    update_remote_enabled = true;
    updateRemoteStatus();
    updateProgramDescription();
    return ok;
}

void MainWindow::block_widgets_signal(bool block)
{
    for(int i=0; i<widgets_to_block.size(); ++i)
        widgets_to_block[i]->blockSignals(block);
}

void MainWindow::table_row_color(QTableWidget *table, int row, QColor color)
{
    for(int i=0; i<COL_DEV_TOTAL; ++i) {
        if(i==COL_CLOCK_STATUS) {
            auto label = qobject_cast<QLabel*>(table->cellWidget(row, i));
            if(label) {
                auto palette = label->palette();
                label->setAutoFillBackground(true);
                palette.setColor(label->backgroundRole(), color);
                label->setPalette(palette);
            }
        } else {
            setItemBgColor(table->item(row, i), color);
        }
    }
}

bool MainWindow::wait_device_setup_written()
{
    QElapsedTimer timer;
    timer.start();
    bool ok=false;
    while (!ok && timer.elapsed()<1000) {
        ok=true;
        QCoreApplication::processEvents(QEventLoop::AllEvents, 10);
        auto dg_it = devGlob.begin();
        while(ok && (dg_it != devGlob.end())) {
            if(dg_it.value()->setup_index != setup_index) {
                ok = false;
                break;
            }
            ++dg_it;
        }
    }
    if(!ok)
        qWarning() << "Some devices failed to write setup";
    return ok;
}

void MainWindow::increase_setup_index()
{
    ++setup_index;
    if(setup_index<100)
        setup_index=100;
    fsm->setMainSetupKey(setup_index);

    for (const DeviceIndex &index: devGlob.keys()) {
        TdcDeviceConfig &ds  = tdcConfig.getConfig(index);
        if(ds.enabled) {
            ds.setup_index = setup_index;
        }
    }
}

bool MainWindow::set_dev_warn(DeviceIndex index, int col, bool warn_flag)
{
    bool warn_changes = false;
    auto dg_it = devGlob.find(index);
    if(dg_it != devGlob.end()) {
        auto &warn = dg_it.value()->warn_flags[col];
        if(warn != warn_flag) {
            warn = warn_flag;
            warn_changes = true;
        }
    }
    return warn_changes;
}

void MainWindow::table_warn_updated()
{
    auto table = ui->tableWidgetDevice;
    auto updateLocker = TableUpdateLocker(table);

    QSet<DeviceIndex> dev_warn;
    for(int col=0; col<COL_DEV_TOTAL; ++col) {
        bool col_warn_flag=false;
        auto dg_it = devGlob.begin();
        while(dg_it != devGlob.end()) {
            auto &dg = dg_it.value();
            bool item_warn = dg->warn_flags[col];
            if(item_warn)
                dev_warn.insert(dg->index);
            col_warn_flag |= item_warn;
            QColor fg_color = item_warn ? Qt::red : Qt::black;
            QColor bg_color = Qt::transparent;
            if(col==COL_HPTDC_STATUS || col==COL_DEV_POWER || col==COL_ERR_CNT) {
                fg_color = Qt::black;
                bg_color = item_warn ? Qt::red : Qt::green;
//            } else if(col==COL_DEV_MCU_VER) {
//                // specific logic
            }

            const int row = getRowByDevice(dg->index);
            auto item = table->item(row, col);
            if(item && col!=COL_DEV_MCU_VER) {
                setItemTextColor(item, fg_color);
                setItemBgColor(item, bg_color);
            }
            ++dg_it;
        }
        auto col_item = table->horizontalHeaderItem(col);
        QColor col_color = col_warn_flag ? Qt::red : Qt::black;
        if(1) {
            col_item->setTextColor(col_color);
        } else {
            // Setting color to horizontal headers doesn't work
            if(col==COL_DEV_POWER)
                qInfo() << col << col_item->text() << col_item->textColor();
            setItemTextColor(col_item, col_color);
            if(col==COL_DEV_POWER)
                qInfo() << col << col_item->text() << col_item->textColor() << col_color;
        }
        // Additional actions for group (column)
        switch (col) {
        case COL_DEV_POWER:
            criticalPowerMode=col_warn_flag;
            ui->actionReset_power_status->setVisible(criticalPowerMode);
            if(criticalPowerMode && col_warn_flag)
                ui->actionShow_power_current->setChecked(true);
            table->setColumnHidden(col, !Globals::getDebugMode() && !col_warn_flag);
            break;
        case COL_ERR_CNT:
            ui->actionReset_error_counters->setVisible(col_warn_flag);
            table->setColumnHidden(col, !col_warn_flag);
            break;
        }
    }
    auto dg_it = devGlob.begin();
    while(dg_it != devGlob.end()) {
        auto &dg = dg_it.value();
        dg->dev_warn = dev_warn.contains(dg->index);
        int row = getRowByDevice(dg->index);
        setItemTextColor(table->verticalHeaderItem(row), dg->getColor(), true);
        ++dg_it;
    }

    updateTableSize(table);
}

void MainWindow::table_warn_updated(DeviceIndex warn_index, int warn_col)
{
    Q_UNUSED(warn_index)
    Q_UNUSED(warn_col)
    table_warn_updated();
}

void MainWindow::deviceConnectStatueChanged(bool connected)
{
    auto dg_it = devGlob.begin();
    while(dg_it != devGlob.end()) {
        if(dg_it.value()->dev == sender())
            break;
        ++dg_it;
    }

    if(dg_it == devGlob.end()) return;
    dg_it.value()->dev_online = connected;
    auto table = ui->tableWidgetDevice;
    auto sortLocker = TableSortLocker(table);
    const int row = getRowByDevice(dg_it.key());
    if(row == -1) return;

    setItemTextColor(table->item(row, COL_DEV_SERIAL), connected ? Qt::darkGreen : Qt::darkGray);
    setItemTextColor(table->verticalHeaderItem(row), dg_it.value()->getColor(), true);
    table_row_color(table, row, connected ? Qt::white : Qt::gray);
    updateTableSize(table);
    if(tdcConfig.isRun() && !connected) {
        sendRunError(QString("%1 offline").arg(dg_it.key().getIdent()));
    }
}

DeviceIndex MainWindow::getDevIndexByRow(int row) const
{
    const auto &index_data = ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)
            ->data(Qt::UserRole);
    return index_data.value<DeviceIndex>();
}

int MainWindow::getRowByDevice(const DeviceIndex &i) const
{
    for(int row=ui->tableWidgetDevice->rowCount()-1; row>=0; --row) {
        if(getDevIndexByRow(row) == i) return row;
    }
    return -1;
}

void MainWindow::tdc_device_config_updated()
{
    for (DeviceIndex i: devGlob.keys()) {
        if(tdcConfig.global_mode || i == tdcConfig.curDev) {
            tdcConfig.getConfig(i).global_mode_copy(tdcConfig.default_device_config);
        }
    }
    if(!blockControl) {
        write_config_request();
        configUpdated();
    }
}

QString MainWindow::getDeviceNameForTree(DeviceIndex index, int devCnt) const
{
    const auto &lldpInfo = lldpCache.value(index);
    const auto &slot_sufix = lldpInfo.isNull() ? "" : QString(" slot:%1")
                                                 .arg(lldpInfo.chassis_slot);
    return QString("%1: %2 (s/n: %3%4)").arg(++devCnt)
            .arg(getDeviceTypeName(index.getDevId()))
            .arg(index.getSerialStr())
            .arg(slot_sufix);
}

void MainWindow::addMStreamDump(const DeviceIndex &index)
{
    auto dgIt = devGlob.find(index);
    if(dgIt == devGlob.end())
        return;

    auto &dg = dgIt.value();
    const auto &tdcStatus = dg->pendingDevStatus;

    MStreamDumpParameters dump_parameters;
    dump_parameters.deviceIndex = index;
    const auto &lldpInfo = lldpCache.value(index);
    if(!lldpInfo.isNull() && lldpInfo.ip_addr){
        dump_parameters.deviceAddress = QHostAddress(lldpInfo.ip_addr);
    }
    Q_ASSERT(lldpInfo.ip_addr);
    dump_parameters.bigFragId = true;
    dump_parameters.hwBufSize = tdcStatus.mstreamHwBufSize;
    dump_parameters.ackSizeLimit = tdcStatus.mstreamMultiAck ? tdcStatus.mstreamHwBufSize : 1;
    dump_parameters.preRunInit();

    MStreamDump *dump = dg->mstream = new MStreamDump(dump_parameters, false);
    connect(dump, &MStreamDump::closing, pnpServer, &PNPServer::closeProgram);
    connect(dump, &MStreamDump::updateProgramDescription,
            pnpServer, &PNPServer::updateProgramDescription);
    connect(dump, &MStreamDump::stat_updated,
            ui->m_stream_output, &StatisticOutput::stat_updated);
    connect(dump, &MStreamDump::stat_updated, this, &MainWindow::mstream_status_updated);
    LldpDiscovery::subscribe(dump, {DEVICE_ID_TDC72VXS, DEVICE_ID_TDC64VHLE});
    connect(&LldpDiscovery::instance(), &LldpDiscovery::cacheUpdated,
            dump, &MStreamDump::lldpCacheUpdated);
    connect(dump, &MStreamDump::stat_from_receiver_updated,
            ui->m_stream_output, &StatisticOutput::stat_from_receiver_updated);
    connect(this, &MainWindow::msdump_reset, dump, &MStreamDump::msdump_reset);
    connect(this, &MainWindow::mstream_send_connect, dump, &MStreamDump::send_connect_to_dev);

    QThread *thread = dg->msThread = new QThread(this);
    thread->setObjectName("MSD_"+index.getSerialStr());
    connect(dump, &MStreamDump::destroyed, thread, &QThread::quit);
    connect(this, &MainWindow::deleteDevice, dump, &MStreamDump::delete_dump);
    dump->moveToThread(thread);
    thread->start();
}

void MainWindow::resetMStream(const DeviceIndex &index, bool doStatClear)
{
    auto dgIt = devGlob.find(index);
    if(dgIt == devGlob.end())
        return;
    auto &dg = dgIt.value();
    auto &lastStatus = dg->pendingDevStatus;
    if(!lastStatus.connected)
        return;

    emit msdump_reset(index, true, lastStatus.mstreamHwBufSize,
                     lastStatus.mstreamMultiAck ? lastStatus.mstreamHwBufSize : 1, QString(), doStatClear);
}

//=== private slots:
void MainWindow::sendRunError(QString mess)
{
    if(tdcConfig.isRun() && tdcConfig.allow_remote_control)
        emit runError(tdcConfig.runId,
            QString("TDC #%1 %2").arg(current_cs.getIndex(), mess));
}

void MainWindow::customContextMenuRequested(const QPoint &pos)
{
    QMenu contextMenu(tr("Context menu"), this);

    QAction action1("Enabled", this);
    connect(&action1, &QAction::triggered, this, &MainWindow::customContextMenuAction);
    action1.setData(true);
    contextMenu.addAction(&action1);

    QAction action2("Disabled", this);
    action2.setData(false);
    connect(&action2, &QAction::triggered, this, &MainWindow::customContextMenuAction);
    contextMenu.addAction(&action2);

    contextMenu.exec(ui->treeWidgetChCfg->mapToGlobal(pos));
}

void MainWindow::customContextMenuAction()
{
    auto s = qobject_cast<QAction*>(QObject::sender());

    if(s == nullptr) return;
    bool enable = s->data().toBool();

    auto tree = ui->treeWidgetChCfg;
    auto updateLocker = TableUpdateLocker(tree);
    for(auto item : tree->selectedItems()) {
        item->setCheckState(tree_tdc_name, enable ? Qt::Checked : Qt::Unchecked);
    }
}

void MainWindow::adjust_setup_changed(AdjustConfig adjust_set)
{
    tdcConfig.default_device_config.adjust_set = adjust_set;
    // use for all  devices regardless of global_mode
    for(const DeviceIndex &index: tdcConfig.known_configs.keys()){
        tdcConfig.getConfig(index).adjust_set = adjust_set;
    }
    configUpdated();
}

void MainWindow::temperatureSet(const TdcStatus &status)
{
    const auto devMetricPrefix = getDevMetricPrefix(status.index)+"temp.";
    const QMap<int, double> &temps = status.temp;
    auto table = ui->tableWidgetDevice;
    auto sortLocker = TableSortLocker(table);
    auto updateLocker = TableUpdateLocker(table);
    int row = getRowByDevice(status.index);
    if(temps.isEmpty() || row==-1) return;
    QList<double> tempList = temps.values();
    double minTemp = *std::min_element(tempList.begin(), tempList.end());
    double maxTemp = *std::max_element(tempList.begin(), tempList.end());

    const bool tempWarn = maxTemp>=70 || minTemp<0;
    bool warn_changed = false;
    auto dg_it = devGlob.find(status.index);
    if(dg_it != devGlob.end()) {
        auto &warn = dg_it.value()->warn_flags[COL_DEV_TEMP];
        if(warn != tempWarn) {
            warn = tempWarn;
            warn_changed = true;
        }
        if(warn) {
            sendRunWarning(QString("%1 temp %2-%3")
                           .arg(dg_it.key().getSerialStr())
                           .arg(minTemp).arg(maxTemp));
        }
    }

    QString tempStr = QString::number(minTemp,'f',1);
    QString tempMaxStr = QString::number(maxTemp,'f',1);
    if(tempStr != tempMaxStr) tempStr += QString("-%2").arg(tempMaxStr);
    tempStr.append(QChar(0x02da));
    QStringList fullInfo;
    for (int tempId: temps.keys()) {
        MetricSender::send(devMetricPrefix + QString::number(tempId), temps[tempId]);
        QString suffixStr;
        if(status.index.getDevId() == DEVICE_ID_TDC64VHLE) {
            if(tempId) {
                char c = 'A';
                c += tempId-1;
                suffixStr = QString("HPTDC Chain %1").arg(c);
            } else
                suffixStr = "Board temp.";
        } else {
            switch (tempId) {
            case 0:
                suffixStr = "Board temp.";
                break;
            case 1:
                suffixStr = "Pll temp.";
                break;
            case 2:
            case 3:
            case 4:
                suffixStr = QString("Ch %1-%2").arg((tempId-2)*24).arg((tempId-1)*24-1);
                break;
            default:
                break;
            }
        }
        if(!suffixStr.isEmpty())
            fullInfo << QString("%1%2 - %3").arg(temps[tempId],0,'f',1)
                .arg(QChar(0x02da)).arg(suffixStr);
    }
    if(warn_changed)
        table_warn_updated(status.index, COL_DEV_TEMP);
    QTableWidgetItem * item = table->item(row, COL_DEV_TEMP);
    item->setText(tempStr);
    item->setToolTip(fullInfo.join('\n'));

    updateTableSize(table);
}

void MainWindow::setDeviceStatus(const DeviceIndex &index, const QString &status)
{
    return; // planing to delete this column
    if(!devGlob.contains(index))
        return;
    int row = getRowByDevice(index);
    if(row == -1) {
        row = getRowByDevice(index);
        Q_ASSERT(row!=-1);
        return;
    }
    ui->tableWidgetDevice->item(row, COL_DEV_STATUS)->setText(status);
}
void MainWindow::deviceStatusUpdated(const TdcStatus &status)
{
    auto dg_it = devGlob.find(status.index);
    if(dg_it == devGlob.end())
        return;

    auto &dg = dg_it.value();
    bool doMStreamReset =
        (dg->pendingDevStatus.mstreamHwBufSize != status.mstreamHwBufSize) ||
        (dg->pendingDevStatus.mstreamMultiAck != status.mstreamMultiAck);
    dg->pendingDevStatus = status;

    if(status.connected) {
        if(dg->mstream) {
            if(doMStreamReset)
                resetMStream(status.index);
        } else {
            addMStreamDump(status.index);
        }
    }

    if(status.bmc_status.power.isCriticalCurrent() ||
            status.isMStreamRestartNeeded(dg->lastDevStatus)) {
        applyDeviceStatusUpdates();
    }
}

void MainWindow::applyDeviceStatusUpdates()
{
    QTableWidget *table = ui->tableWidgetDevice;
    auto updateLocker = TableUpdateLocker(table);
    auto sortLocker = TableSortLocker(table);

    auto dgIt = devGlob.begin();
    while(dgIt != devGlob.end()) {
        if(!dgIt.value()->pendingDevStatus.processed)
            applySingleDeviceStatusUpdates(dgIt.key());
        ++dgIt;
    }

    if(updateLocker.up_en) {
        updateTableSize(table);
    }

    updateRemoteStatus();
    deviceStatusTimer->start(500);
}

void MainWindow::applySingleDeviceStatusUpdates(DeviceIndex index)
{
    auto dg_it = devGlob.find(index);
    if(dg_it == devGlob.end())
        return;

    auto &dg = dg_it.value();
    TdcStatus &status = dg->pendingDevStatus;
    if(status.processed)
        return;
    status.processed = true;
    TdcStatus &lastDevStatus = dg->lastDevStatus;

    QTableWidget *table = ui->tableWidgetDevice;
    const int row = getRowByDevice(status.index);
    if(row == -1) {
        lastDevStatus = status;
        return;
    }
    const auto &statistic = status.statistic;
    auto updateLocker = TableUpdateLocker(table);
    if(status.connected) {
        const auto devMetricPrefix = getDevMetricPrefix(index);
        table->item(row, COL_DEV_RUN_TIME)->setText(status.statistic.getRunTime());

        temperatureSet(status);
        table->item(row, COL_DEV_FW)->setText(status.fwStr);
        deviceBmcStatusUpdated(row, status);

//        table->item(row, COL_DEV_TIME)->setText(QDateTime::fromTime_t(status.wrTime).toString());
        table->item(row, COL_DEV_CUR_EV_NUM)->
            setText(QString("%L1").arg(status.curEvNum));
        MetricSender::send(devMetricPrefix + "curEvNum", status.curEvNum);
        table->item(row, COL_DEV_UNBLOCK)->
            setText(QString("%L1").arg(statistic.evNumAccepted));
        table->item(row, COL_DEV_BLOCK)->
            setText(QString("%L1").arg(statistic.getBlocked()));

        QString str;
        bool hptdcErr = status.hptdcChipErrorFlags;
        if(status.hptdcChipErrorFlags)
            str += QString("HPTDC Err flags: %1").arg(status.hptdcChipErrorFlags, 0, 2);
        for (int hptdcId: status.hptdcStatus.keys()) {
            const HptdcStatus &hptdcStatus = status.hptdcStatus[hptdcId];
            if(!hptdcStatus.getSummary().isEmpty()) {
                if(!str.isEmpty()) str += "\n";
                str += QString("HPTDC #%1:\n%2").arg(hptdcId).arg(hptdcStatus.getSummary());
            }
            hptdcErr |= hptdcStatus.hasError();
        }
        if(str.isEmpty()) {
            str = "Ok";
        }

        table->item(row, COL_HPTDC_STATUS)->setText(str);

        qobject_cast<QLabel*>(table->cellWidget(row, COL_CLOCK_STATUS))
                ->setText(status.getClockStatus());
        table->item(row, COL_BUSY_STATUS)->
            setText(QString("%1").arg(status.busyStatus, 2, 2, QChar('0')));
        table->item(row, COL_TRIG_QUEUE)->
            setText(QString("%1").arg(status.trigQueue));
        if(statistic.evNumAccepted != lastDevStatus.statistic.evNumAccepted){
            table->item(row, COL_BUSY_TIME)->
                setText(statistic.getBusyStr(Globals::getDebugMode(), lastDevStatus.statistic));
            MetricSender::send(devMetricPrefix+"time.busy.total",
                statistic.getBusyTotal(lastDevStatus.statistic));
            MetricSender::send(devMetricPrefix+"time.busy.tdc",
                statistic.getBusyTdc(lastDevStatus.statistic));
            MetricSender::send(devMetricPrefix+"time.busy.readout",
                statistic.getBusyReadout(lastDevStatus.statistic));
        }
        MetricSender::send(devMetricPrefix+"time.uptime",
            status.statistic.getRunTimeSec());
        QStringList errCntStr;
        auto errCnt = statistic.errCnt;
        auto &powerOnErrCnt = statistic.powerOnErrCnt;
        errCnt -= powerOnErrCnt;

        QString errStr = errCnt.toString();
        bool errCntFlag = !errStr.isEmpty();
        auto errCntItem = table->item(row, COL_ERR_CNT);
        errCntItem->setText(errStr);

        bool warn_changed = false;
        if(dg->warn_flags[COL_HPTDC_STATUS] != hptdcErr) {
            dg->warn_flags[COL_HPTDC_STATUS] = hptdcErr;
            warn_changed = true;
        }
        if(dg->warn_flags[COL_ERR_CNT] != errCntFlag) {
            dg->warn_flags[COL_ERR_CNT] = errCntFlag;
            warn_changed = true;
        }
        if(warn_changed)
            table_warn_updated(index, COL_HPTDC_STATUS);
    } else {
//        table->item(row, COL_DEV_STATUS)->setText("Offline");
//        table->item(row, COL_DEV_STATUS)->setTextColor(Qt::black);
    }
    if(updateLocker.up_en) {
        updateTableSize(table);
    }

    const bool doResetMStream = status.isMStreamRestartNeeded(lastDevStatus);
    lastDevStatus = status;
    if(doResetMStream) {
        resetMStream(index);
    }
}

void MainWindow::deviceBmcStatusUpdated(int row, const TdcStatus &tdcStatus)
{
    const BmcStatus &status = tdcStatus.bmc_status;
    const auto index = tdcStatus.index;
    if(row<0) return;
    QTableWidget *table = ui->tableWidgetDevice;
    Q_ASSERT(!table->isSortingEnabled());
    QColor status_col = status.get_overall_status_col();
    const bool mcu_ver_warn = status.overall_status==0b11;
    auto mcu_ver_item = table->item(row, COL_DEV_MCU_VER);
    mcu_ver_item->setText(status.getVerStr());
    auto bg_brush = mcu_ver_item->background();
    if(bg_brush.color() != status_col) {
        bg_brush.setColor(status_col);
        bg_brush.setStyle(Qt::SolidPattern);
        mcu_ver_item->setBackground(bg_brush);
    }
    mcu_ver_item->setToolTip(QString("overall - %1\npower - %2\nPLL- %3")
                             .arg(status.get_status_str(status.overall_status))
                             .arg(status.get_status_str(status.power_status))
                             .arg(status.get_status_str(status.pll_status)));

    auto mcu_pll_unlock_item = table->item(row, COL_DEV_PLL_UNLOCK);
    mcu_pll_unlock_item->setText(QString::number(status.pllUnlockCnt));
    bool prev_was_set;
    bool refresh_col = (mcu_pll_unlock_item->data(Qt::UserRole).toInt(&prev_was_set)==0) != (status.pllUnlockCnt==0);
    mcu_pll_unlock_item->setData(Qt::UserRole, status.pllUnlockCnt);
    if(refresh_col || !prev_was_set) update_pll_unlock_visible();
//    auto mcu_pll_st_item = table->item(row, COL_DEV_PLL_STATUS);
//    mcu_pll_st_item->setText(status.onboard_pll_ok() ? "Ok" : "Fail");
//    mcu_pll_st_item->setBackgroundColor(status.onboard_pll_ok() ? Qt::green : Qt::red);

    // Power
    const BmcPower &power = status.power;
    const auto powerItem = table->item(row, COL_DEV_POWER);
    const bool powerCritical = power.isCriticalCurrent();
    auto dg_it = devGlob.find(index);
    if(dg_it != devGlob.end()) {
        bool warn_changes = false;
        auto &warn_flags = dg_it.value()->warn_flags;
        if(powerCritical && warn_flags[COL_DEV_POWER]!=powerCritical) {
            warn_flags[COL_DEV_POWER]=powerCritical;
            warn_changes = true;
        }
        if(warn_flags[COL_DEV_MCU_VER]!=mcu_ver_warn) {
            warn_flags[COL_DEV_MCU_VER]=mcu_ver_warn;
            warn_changes = true;
        }
        if(warn_changes)
            table_warn_updated();
        if(powerCritical) {
            sendRunWarning(QString("%1 power crit.").arg(dg_it.key().getSerialStr()));
        }
    }

    if(Globals::getDebugMode() || powerCritical || criticalPowerMode) {
        powerItem->setText(power.getSummary(tdcConfig.show_power_current));
        if(powerCritical) {
            logDevInfo(powerItem, true);
        }
    }
}

void MainWindow::deviceConnected()
{
    deviceConnectStatueChanged(true);
}

void MainWindow::deviceDisconnected()
{
    deviceConnectStatueChanged(false);
}
void MainWindow::deviceInfoUpdated(DeviceIndex index, int chNum)
{
    Q_ASSERT(!index.isNull());
    if(!tdcConfig.known_configs.contains(index))
        return;
    if(tdcConfig.known_configs[index].getNCh() != chNum) {
        tdcConfig.known_configs[index].setNCh(chNum);
        write_config_request();
        updateModelList();
        configUpdated();
    }

    ui->spinBoxLatency->setMaximum(2048);
//    ui->spinBoxMatchWin -> setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
    updateGuiVal();
}

void MainWindow::deviceSetupWritten(DeviceIndex index, quint32 setup_index)
{
    auto dg_it=devGlob.find(index);
    if(dg_it != devGlob.end()) {
        dg_it.value()->setup_index = setup_index;
        fsm->setDevSetupKey(index, setup_index);
        updateRemoteStatus();
    }
}

void MainWindow::mstream_status_updated(const MStreamStat &status)
{
    auto table = ui->tableWidgetDevice;
    auto dg_it = devGlob.begin();

    DeviceIndex devIndex;
    while(dg_it != devGlob.end()) {
        if(dg_it.value()->mstream == sender()) {
            break;
        }
        ++dg_it;
    }
    if(dg_it == devGlob.end())
        return;
    auto &prevStatus = dg_it.value()->msStat;
    auto sortLocker = TableSortLocker(table);
    int devRow = getRowByDevice(dg_it.key());
    auto updateLocker = TableUpdateLocker(table);
    if(prevStatus.timestamp.isValid()) {
        qint64 milisecs = 0;
        milisecs = prevStatus.timestamp.msecsTo(status.timestamp);
        double receive_data_rate = 0;
        if(milisecs != 0) {
            receive_data_rate = 1.*(status.recvDataSize - prevStatus.recvDataSize)/milisecs;
        }
        table->item(devRow, COL_DEV_STREAM)->setText(QString("%L1 kB/s").arg(receive_data_rate,0,'f',2));
    }
    prevStatus = status;
    table->item(devRow, COL_DEV_MS_ERR)->setText(QString::number(status.badFormatPackage));
    auto &warn = dg_it.value()->warn_flags[COL_DEV_TEMP];
    if(warn != (status.badFormatPackage != 0)) {
        warn = status.badFormatPackage != 0;
        table_warn_updated(dg_it.key(), COL_DEV_MS_ERR);
    }
    updateTableSize(table);
}

void MainWindow::update_order_index(int)
{
    for(DeviceIndex index1 : tdcConfig.known_configs.keys()){
        for(int row = 0; row < ui->tableWidgetDevice->rowCount(); ++row){
            DeviceIndex index2 = getDevIndexByRow(row);
            if(index1 == index2){
                tdcConfig.known_configs[index1].order_index = row;
            }
        }
    }
    update_tree_order();
}

void MainWindow::write_config_request()
{
    if(write_config_timer == QObject::sender()){
        if(write_config_request_flag){
            write_config();
            write_config_request_flag = false;
            return;
        }
    }
    else{
        if(write_config_timer->isActive()){
            write_config_request_flag = true;
            return;
        }
        else{
            write_config();
            write_config_request_flag = false;
            write_config_timer->start(1000);
        }
    }
}

void MainWindow::sortIndicatorChanged(int logicalIndex, Qt::SortOrder order)
{
    tdcConfig.sort_column = logicalIndex;
    tdcConfig.sort_asc = order==Qt::AscendingOrder;
}

void MainWindow::updateTableSize(QTableWidget * const table, bool widthOnly)
{
    widthOnly = true;
    table->resizeRowsToContents();
    table->resizeColumnsToContents();
//    table->setColumnWidth(COL_RATE, 500);

    int width = 0;
    width += table->verticalScrollBar()->width();
    if(table->verticalHeader()->isVisible())
        width += table->verticalHeader()->width();
    bool test=false;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
        if(test)
            qInfo() << table->horizontalHeaderItem(col)->text() << table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    width = qMin(width, 1000);
    height = qMin(height, 200);
    if(widthOnly) {
        table->setMinimumWidth(width);
    } else {
        table->setMinimumWidth(width);
        table->setMinimumHeight(height);
    }
    table->updateGeometry();
    table->horizontalHeader()->setStretchLastSection(true);
}

void MainWindow::updateTreeSize(QTreeWidget * const tree)
{
    QString colInfo;
    int width = 0;
    colInfo += "fix:5;";
    for(int col=0; col<tree->columnCount(); ++col) {
        if(tree->isColumnHidden(col))            continue;
        tree->resizeColumnToContents(col);
        colInfo += QString("%1:%2;").arg(tree->headerItem()->text(col)).arg(tree->columnWidth(col));
        width += tree->columnWidth(col);
    }
}

//=== GUI slots
void MainWindow::on_pushButtonChangeDevice_clicked()
{
    discoverDialog->show();
}

void MainWindow::on_tableWidgetDevice_itemSelectionChanged()
{
    QTableWidget *table = ui->tableWidgetDevice;
    QList<QTableWidgetSelectionRange> ranges = table->selectedRanges();
    tdcConfig.curDev = DeviceIndex();
    if(ranges.isEmpty()) {
        updateGuiVal();
        return;
    }
    if(ranges.size()!=1) {
        updateGuiVal();
        return;
    }
    Q_ASSERT(ranges.size()==1);
    QTableWidgetSelectionRange range = ranges.first();
    Q_ASSERT(range.bottomRow() == range.topRow());
    int row = range.bottomRow();

    DeviceIndex index = getDevIndexByRow(row);
    if(!index.isNull() && devGlob.contains(index)
            && tdcConfig.known_configs.contains(index)
            && tdcConfig.known_configs[index].enabled) {
        tdcConfig.curDev = index;
        tdcConfig.default_device_config.global_mode_copy(tdcConfig.getConfig(index));
    }
    updateGuiVal();
}

void MainWindow::on_tableWidgetDevice_itemChanged(QTableWidgetItem *item)
{
    logDevInfo(item);
}

void MainWindow::on_tableWidgetDevice_cellDoubleClicked(int row, int column)
{
    if(column != COL_DEV_FW) return;
    DeviceIndex index = getDevIndexByRow(row);
    if(!devGlob.contains(index)) return;

    const auto &stat = devGlob[index]->lastDevStatus;

    const auto &power = stat.bmc_status.power;
    QStringList powerList;
    for(auto id : power.vol.keys()) {
        auto str = QString("%1-%2").arg(power.getIdStr(id)).arg(1e-3*power.vol[id],0,'f',3);
        if(str.contains("TDC",Qt::CaseInsensitive))
            powerList << str.remove("TDC_");
    }
    QString powerStr = powerList.join(' ').prepend("Long test passed. Voltages ");

    QStringList list;
    list << stat.fwStr.split('.').last();
    list << stat.bmc_status.getVerStr();
    list << QDateTime::currentDateTime().toString("dd.MM.yyyy");
    list << "Robot";
#ifdef GIT_DESCR
    list << GIT_DESCR;
#endif
    list << powerStr;

    QApplication::clipboard()->setText(list.join('\t'));
}
void MainWindow::on_actionAllow_remote_control_toggled(bool checked)
{
    tdcConfig.allow_remote_control = checked;
    write_config_request();
    emit allowRemoteControl(tdcConfig.allow_remote_control);
}

void MainWindow::on_actionon_board_PLL_bypass_toggled(bool pll_bypass)
{
    tdcConfig.default_device_config.pll_bypass = pll_bypass;
    tdc_device_config_updated();
}

void MainWindow::on_actionFeLinkStatusDialog_triggered()
{
    feLinkStatusDialog->show();
}

void MainWindow::on_pushButtonStart_clicked()
{
    tdcConfig.remote_start_request = false;
    fsm->goState(FsmState::Starting);
    updateControlState();
}

void MainWindow::on_pushButtonStop_clicked()
{
    fsm->goState(FsmState::Stopping);
    updateControlState();
}

void MainWindow::on_radio_button_global_toggled(bool checked)
{
    if(blockControl) return;
    tdcConfig.global_mode = checked;
    if(!tdcConfig.global_mode) return;

    for(DeviceIndex i : tdcConfig.known_configs.keys())
        tdcConfig.getConfig(i).global_mode_copy(tdcConfig.default_device_config);

    configUpdated();
}

void MainWindow::on_checkBoxTrigTimer_toggled(bool checked)
{
    tdcConfig.default_device_config.trigsetup.bit.timer_en = checked ? 1 : 0;
    tdc_device_config_updated();
    ui->spinBoxTimeTrig->setEnabled(checked);
}

void MainWindow::on_spinBoxTimeTrig_valueChanged(int rateHz)
{
    const quint32 period = static_cast<quint32>(125e6/rateHz/3); // period in 24ns
    tdcConfig.default_device_config.trigPeriod = period;
    tdc_device_config_updated();

    updateRealRate();
}

void MainWindow::on_checkBoxTrigTtl_toggled(bool checked)
{
    tdcConfig.default_device_config.trigsetup.bit.ttl_en = checked ? 1 : 0;
    tdc_device_config_updated();
}

void MainWindow::on_checkBoxTrigTtc_toggled(bool checked)
{
    tdcConfig.default_device_config.trigsetup.bit.ttc_en = checked ? 1 : 0;
    tdc_device_config_updated();
}

void MainWindow::on_checkBoxTrigVxs_stateChanged(int arg1)
{
    tdcConfig.default_device_config.trigsetup.bit.vxs_en = (Qt::Checked == arg1) ? 1 : 0;
    tdc_device_config_updated();
}

void MainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    ui->labelLatencyNs->setText(QString("%1 ns").arg(lat*24));
    ui->spinBoxMatchWin->setMaximum(lat);

    tdcConfig.default_device_config.latency = static_cast<quint32>(lat);
    tdc_device_config_updated();
}

void MainWindow::on_spinBoxMatchWin_valueChanged(int win)
{
    ui->labelMatchWinNs->setText(QString("%1 ns").arg(win*24));
    emit matchWinUpdated(win);

    tdcConfig.default_device_config.matchWin = static_cast<quint32>(win);
    tdc_device_config_updated();
}

void MainWindow::on_comboBoxEdges_currentIndexChanged(int index)
{
    bool ok;
    int v = ui->comboBoxEdges->itemData(index).toInt(&ok);
    Q_ASSERT(ok);

    tdcConfig.default_device_config.leadEdge = v&1;
    tdcConfig.default_device_config.trailEdge = v&2;
    tdc_device_config_updated();
}
void MainWindow::on_groupBoxStat_toggled(bool checked)
{
    tdcConfig.default_device_config.stat_readout = checked;
    tdc_device_config_updated();
}

void MainWindow::on_spinBoxStatPeriod_valueChanged(int period_ms)
{
    tdcConfig.default_device_config.stat_period = static_cast<quint16>(period_ms);
    tdc_device_config_updated();
}

void MainWindow::on_comboBoxPreset_currentIndexChanged(int index)
{
    int preset_type = ui->comboBoxPreset->itemData(index).toInt();
    StatPreset stat_preset = StatPresetFabric::getPreset(preset_type);

    tdcConfig.default_device_config.stat_preset = stat_preset;
    // use for all  devices regardless of global_mode
    for(const DeviceIndex &index: tdcConfig.known_configs.keys()){
        tdcConfig.getConfig(index).stat_preset = stat_preset;
    }
    configUpdated();
}

void MainWindow::on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column)
{
    updateTreeSize(ui->treeWidgetChCfg);

    if(blockControl)
        return;
    if(column != tree_tdc_name)
        return;
    Qt::CheckState state = item->checkState(tree_tdc_name);
    if(state == Qt::PartiallyChecked)
        return;
    bool checked = (state == Qt::Checked);
    bool modified = false;
    if(item==rootTreeItem) {
        for (const DeviceIndex &index: modTreeItem.keys()) {
            if(modTreeItem[index]->checkState(tree_tdc_name) != state) {
                modTreeItem[index]->setCheckState(tree_tdc_name, state);
                modified = true;
            }
        }
    } else {
        QTreeWidgetItem *p = item->parent();
        bool found=false;
        for (const DeviceIndex &index: modTreeItem.keys()){
            if(item==modTreeItem[index]){
                // Module item
                found = true;
                TdcDeviceConfig &ds = tdcConfig.getConfig(index);
                for (int ch: ds.chTdcEn.keys()) {
                    if(ds.chTdcEn[ch] != checked) {
                        ds.chTdcEn[ch] = checked;
                        modified = true;
                    }
                }
                if(modified) {
                    ui->treeWidgetChCfg->blockSignals(true);
                    for(int childIndex=0; childIndex<item->childCount(); ++childIndex) {
                        item->child(childIndex)->setCheckState(tree_tdc_name, state);
                    }
                    ui->treeWidgetChCfg->blockSignals(false);
                }
            } else if(p==modTreeItem[index]) {
                // Channel item
                found = true;
                bool ok;
                int modCh = item->text(tree_tdc_name).toInt(&ok);
                TdcDeviceConfig &ds = tdcConfig.getConfig(index);
                Q_ASSERT(ok);
                Q_ASSERT(modCh<=ds.getNCh());
                bool &chState = ds.chTdcEn[modCh];
                if(chState != checked) {
                    chState = checked;
                    modified = true;
                }
            }
            if(modified)
                updateTreeItemStates();
            if(found)
                break;
        }
        Q_ASSERT(found);
    }

    if(modified) {
        updateGuiVal();
        configUpdated();
    }
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void MainWindow::on_actionExit_triggered()
{
    close();
}
void MainWindow::on_actionConfiguration_window_toggled(bool show)
{
    tdcConfig.show_cfg_widget = show;
    write_config_request();
    ui->widget_bottom->setVisible(show);
}

void MainWindow::on_actionChange_device_order_triggered()
{
    OrderDialog dialog(&tdcConfig, knownDevDescr, this);
    if(dialog.exec() == QDialog::Accepted) {
        tdcConfig.known_configs = dialog.get_known_setups();
        write_config_request();
        updateDeviceOrder(true);
    }
}

void MainWindow::on_actionReset_MStream_triggered()
{
    for (DeviceIndex i: devGlob.keys()) {
        resetMStream(i);
    }
}

void MainWindow::on_actionSend_soft_reset_triggered()
{
    emit softReset();
}

void MainWindow::on_actionAdjust_HPTDC_triggered()
{
    if(!adjustDialog) return;
    adjustDialog->activate_dialog();
    adjustDialog->show();
}

void MainWindow::on_actionTrace_dev_info_changes_toggled(bool checked)
{
    tdcConfig.trace_device_info = checked;
    if(!tdcConfig.trace_device_info) {
        auto devIt = devGlob.begin();
        while(devIt != devGlob.end()) {
            devIt.value()->tbl_info_log.clear();
            ++devIt;
        }
    }
}

void MainWindow::on_actionShow_power_current_toggled(bool checked)
{
    tdcConfig.show_power_current = checked;
    if(!checked) {
        ui->actionReset_power_status->trigger();
    }
    write_config_request();
}

void MainWindow::on_actionReset_power_status_triggered()
{
    auto dg_it = devGlob.begin();
    while (dg_it != devGlob.end()) {
        dg_it.value()->warn_flags[COL_DEV_POWER] = false;
        ++dg_it;
    }

    table_warn_updated();
}

void MainWindow::on_actionReset_error_counters_triggered()
{
    // todo make slot for column auto-hide
//    ui->actionReset_error_counters->setVisible(false);
    emit reset_error_counters();

    auto dg_it = devGlob.begin();
    while (dg_it != devGlob.end()) {
        dg_it.value()->warn_flags[COL_ERR_CNT] = false;
        ++dg_it;
    }

    table_warn_updated();
}
void MainWindow::on_actionAdd_HPTDC_event_header_toggled(bool checked)
{
    tdcConfig.default_device_config.hptdcHdr = checked;
    tdc_device_config_updated();
}

void MainWindow::on_actionAdd_HPTDC_event_tralier_toggled(bool checked)
{
    tdcConfig.default_device_config.hptdcTrl = checked;
    tdc_device_config_updated();
}

void MainWindow::on_actionTxt_data_format_triggered()
{
    QString s;
    s += "<h2>TDC2 Txt Data format</h2>";
    s += "<p>File constructs with following lines:</p>";
    s += "<ul>";
    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
    s += "</ul>";
    s += "<p>where:</p>";
    s += "<dl>";
    s += "<dt><b>TAI time</b></dt>";
    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
    s += "<dt><b>ch</b></dt>";
    s += "<dd><i>Channel number in range 0-15</i></dd>";
    s += "<dt><b>tdcLeadTime</b></dt>";
    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>adcTime</b></dt>";
    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>sample</b></dt>";
    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
    s += "</dl>";

    QMessageBox::about (this, "Txt data format", s);
}

void MainWindow::on_actionRaw_data_format_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
    mainLayout->addWidget(&htmlLabel);
    d.setLayout(mainLayout);
    d.exec();
}
