#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QAbstractButton>
#include <QHostAddress>
#include <QProcess>
#include <QMessageBox>
#include <QMetaType>
#include <QTime>
#include <QMap>
#include <QBitArray>

#include "DeviceGlob.h"
#include "TdcDevice.h"
#include "TdcConfig.h"

#include "base/app/BaseMainWindow.h"
#include "configuration-manager/ConfigurationManager.h"
#include "mldiscover/LldpDiscovery.h"
#include "mongo/StartupOptions.h"
#include "pnp-server/PNPMessage.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"

namespace Ui {
class MainWindow;
}

class QSemaphore;
class QTableWidget;
class QTreeWidgetItem;

class DeviceDescription;
class DeviceMgrFsm;
class DiscoverDialog;
class LogDialog;
class AdjustDialog;
struct TdcStatus;
class TcpClient;
class PNPServer;
class ProgramDescription;
class QTreeWidget;
struct MStreamStatus;
class QTableWidgetItem;
class LldpDiscovery;
class DeviceDispatcher;
class DeviceIndexEnabledMap;
class TdcConfigDb;
class TdcConfigDispatcher;
class TdcSignalRouter;
class FeLinkStatusDialog;

Q_DECLARE_METATYPE(QHostAddress)

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(StartupOptions &start_options, QWidget *parent = nullptr);
    ~MainWindow() override;

signals:
    void switch_to_config(const ConfigIndexName& conf) const;
    void configChanged(const TdcConfig &config);
    void setupChanged(const TdcDeviceConfig &);
    void setupChangedForce();
    void deleteDevice(const DeviceIndex &);
    void addLog(QString);
    void matchWinUpdated(int matchWin);
    void softReset();
    void blockDeviceRun(bool isBlock);
    void reset_error_counters();
    void closeAll();

    //pnp server
    void changeProgramDescription(const ProgramDescription &);
    void closeProgram(const ProgramDescription &);

    //remote control
    void sendRemoteState(QString state, bool ready4run);
    void runError(const RcRunId &runId, QString errStr);
    void runWarning(const RcRunId &runId, QString warnStr);
    void allowRemoteControl(bool state);
    void config_transfer_to_rc(QByteArray data);
    void msdump_reset(DeviceIndex index, bool bigFragId, quint16 hwBufSize, quint16 ackSize, QString fn=QString(), bool doStatClear = false);
    void mstream_send_connect(DeviceIndex index);
    void set_run_state(bool isRun);
    void deviceListChanged(const DeviceIndexEnabledMap &);

public slots:
    void configSelectorChanged(const ConfigIndexName& conf);
    void setConfig(const TdcConfig &newConfig);
    void lldpCacheUpdated(const LldpInfoCache &cache);
    void onFsmStateChange(const FsmState &state, const FsmState &oldState);
    void setDeviceList(const DeviceIndexEnabledMap &list);
    //remote control
    void onRemoteStartRequest(const RcRunId &runId, quint64 start_event_number=0);
    void onRemoteStopRequest(const RcRunId &runId, quint64 start_event_number=0);
    void remoteStatusUpdated(const RemoteControlServerStatus &status);
    //evNumChecker
    void sendRunWarning(QString mess);

private:
    void initConnections();
    void initTable();
    void initTree();
    void initShortcut();
    void init_gui();
    void read_config();
    void write_config();
    void updateModelList();
    void updateTreeItemStates();
    void configUpdated();
    void updateGuiVal();
    void updateControlState();
    void removeDevice(const DeviceIndex &index);
    void addDevice(const DeviceIndex &index);
    void sendLog(QString l);
    void logDevInfo(QTableWidgetItem *item, bool force=false);
    void updateRealRate();
    void update_pll_unlock_visible();
    void updateDeviceOrder(bool from_dialog = false);
    void update_tree_order();
    void updateProgramDescription();
    void enable_gui_controls();
    void updateRemoteStatus();
    void perform_start();
    bool perform_stop();
    void block_widgets_signal(bool block);
    void table_row_color(QTableWidget *table, int row, QColor color);
    bool wait_device_setup_written();
    void increase_setup_index();
    bool set_dev_warn(DeviceIndex index, int col, bool warn);
    void table_warn_updated();
    void table_warn_updated(DeviceIndex warn_index, int warn_col);
    void deviceConnectStatueChanged(bool connected);
    DeviceIndex getDevIndexByRow(int row) const;
    int getRowByDevice(const DeviceIndex &i) const;
    void tdc_device_config_updated();
    QString getDeviceNameForTree(DeviceIndex index, int devCnt) const;
    void addMStreamDump(const DeviceIndex &index);
    void resetMStream(const DeviceIndex &index, bool doStatClear = false);
    static void updateTableSize(QTableWidget * const table, bool widthOnly=false);
    static void updateTreeSize(QTreeWidget * const tree);

private slots:
    //evNumChecker
    void sendRunError(QString mess);

    void customContextMenuRequested(const QPoint &);
    void customContextMenuAction();

    //--adjust HPTDC
    void adjust_setup_changed(AdjustConfig);

    void temperatureSet(const TdcStatus &status);
    void setDeviceStatus(const DeviceIndex &index, const QString &status);
    void deviceStatusUpdated(const TdcStatus &status);
    void applyDeviceStatusUpdates();
    void applySingleDeviceStatusUpdates(DeviceIndex index);
    void deviceBmcStatusUpdated(int row, const TdcStatus &tdcStatus);
    void deviceConnected();
    void deviceDisconnected();

    void deviceInfoUpdated(DeviceIndex index, int chNum);
    void deviceSetupWritten(DeviceIndex index, quint32 setup_index);
    void mstream_status_updated(const MStreamStat &status);
    void update_order_index(int i);
    void write_config_request();
    void sortIndicatorChanged(int logicalIndex, Qt::SortOrder order);

    //--GUI slots
    void on_pushButtonChangeDevice_clicked();
    void on_tableWidgetDevice_itemSelectionChanged();
    void on_tableWidgetDevice_itemChanged(QTableWidgetItem *item);
    void on_tableWidgetDevice_cellDoubleClicked(int row, int column);
    void on_actionAllow_remote_control_toggled(bool checked);
    void on_actionon_board_PLL_bypass_toggled(bool pll_bypass);
    void on_actionFeLinkStatusDialog_triggered();
    //--Start/stop button
    void on_pushButtonStart_clicked();
    void on_pushButtonStop_clicked();
    //--global/separate mode
    void on_radio_button_global_toggled(bool checked);
    //--Trig setup
    void on_checkBoxTrigTimer_toggled(bool checked);
    void on_spinBoxTimeTrig_valueChanged(int rateHz);
    void on_checkBoxTrigTtl_toggled(bool checked);
    void on_checkBoxTrigTtc_toggled(bool checked);
    void on_checkBoxTrigVxs_stateChanged(int arg1);
    //--Readout
    void on_spinBoxLatency_valueChanged(int lat);
    void on_spinBoxMatchWin_valueChanged(int win);
    void on_comboBoxEdges_currentIndexChanged(int index);
    //--Statistic
    void on_groupBoxStat_toggled(bool checked);
    void on_spinBoxStatPeriod_valueChanged(int period_ms);
    void on_comboBoxPreset_currentIndexChanged(int index);
    //--ch tree widget
    void on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column);
    // Menu File
    void on_actionConfiguration_manager_triggered();
    void on_actionExit_triggered();
    // Menu View
    void on_actionConfiguration_window_toggled(bool show);
    // Menu Options
    void on_actionChange_device_order_triggered();
    void on_actionReset_MStream_triggered();
    void on_actionSend_soft_reset_triggered();
    void on_actionAdjust_HPTDC_triggered();
    void on_actionTrace_dev_info_changes_toggled(bool checked);
    void on_actionShow_power_current_toggled(bool checked);
    void on_actionReset_power_status_triggered();
    void on_actionReset_error_counters_triggered();
    void on_actionAdd_HPTDC_event_header_toggled(bool checked);
    void on_actionAdd_HPTDC_event_tralier_toggled(bool checked);
    // Menu Help
    void on_actionTxt_data_format_triggered();
    void on_actionRaw_data_format_triggered();

private:
    Ui::MainWindow *ui;
    DeviceMgrFsm * fsm;
    std::unique_ptr<TdcSignalRouter> signalRouter;
    ConfigSelector current_cs;
    ConfigurationManager * const configuration_manager;
    TdcConfig tdcConfig;
    TdcConfigDb * const tdcConfigDb;
    TdcConfigDispatcher * const tdcConfigDispatcher;
    QTimer  *write_config_timer;
    bool write_config_request_flag = false;
    DeviceDispatcher * const deviceDispatcher;
    DiscoverDialog * const discoverDialog;
    FeLinkStatusDialog * const feLinkStatusDialog;
    RemoteControlServer * const remoteControl;
    PNPServer * const pnpServer;
    EvNumChecker * const evNumChecker;
    ProgramDescription program_description;

    QMap<DeviceIndex, DeviceGlob*> devGlob;
    QMap<DeviceIndex, QTreeWidgetItem *> modTreeItem;
    int blockControl = 0;
    LogDialog *const logDialog;
    AdjustDialog *const adjustDialog;
    QMap<DeviceIndex, DeviceDescription> knownDevDescr;
    QVector<QString> mstreamTags;
    QMap<QString, QString> dumperTags;
    QMap<QString, QString> recieverTags;

    QTreeWidgetItem * const rootTreeItem;
    bool msTagDublicate;
    QTimer * const deviceStatusTimer;
    bool criticalPowerMode = false;
    bool on_remote = false;
    QList<QObject*> widgets_to_block;
    quint32 setup_index = 0;
    bool update_remote_enabled = true;
    LldpInfoCache lldpCache;
    QLabel * const labelFsmState;
};

#endif // MAINWINDOW_H
