#include "CalibrationModel.h"
#include "DeviceChModelItem.h"
#include "TQDCDevice.h"

#include <QFont>
#include <QBrush>
#include <QDebug>

namespace {
const int RMS_LIMIT = 20; // in adc64 100

QString getThrText(int dacValue)
{
    return QString("%1 mV (%2)")
            .arg(TQDCDevice::dac_mV(dacValue), 0, 'f', 2)
            .arg(dacValue);
}
}

CalibrationModel::CalibrationModel(QObject *parent)
    :QAbstractItemModel(parent)
    , rootItem(new DeviceChModelItem())

{}

void CalibrationModel::clear()
{
    for(auto die : map){
        if(die.enabled) {
            auto cal = CalResult();
            cal.index = die.index;
            cal_config[die.index] = cal;
        }
    }
    resetItems();
    emit calConfigChanged(cal_config);
}

void CalibrationModel::removeDev(const DeviceIndex &index)
{
    auto newCalConfig = cal_config;
    newCalConfig.remove(index);
//    setCalConfig(newCalConfig);
}

QVariant CalibrationModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();

    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::ForegroundRole:
        return getForegroundData(index);
    case Qt::FontRole:
        return fontData(index);
    }
    return QVariant();
}

QVariant CalibrationModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return QVariant();
    const auto col = static_cast<TreeColumns>(section);
    if(role == Qt::DisplayRole) {
        switch (col) {
        case TreeColumns::NAME:
            return "Device / ch";
        case TreeColumns::BASELINE:
            return "Baseline";
        case TreeColumns::BASELINE_RMS:
            return "Baseline rms";
        case TreeColumns::BASELINE4:
            return "Baseline x4";
        case TreeColumns::BASELINE4_RMS:
            return "Baseline x4 rms";
        case TreeColumns::DAC:
            return "DAC offset, mV (dacValue)";
        default:
            return QVariant();
        }
    }
    if(role == Qt::ForegroundRole) {
        if(col == TreeColumns::BASELINE_RMS ||
                col == TreeColumns::BASELINE4_RMS)
            return getHeaderForeground(col);
    }

    return QVariant();
}

QModelIndex CalibrationModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const DeviceChModelItem *parentItem = parent.isValid() ?
                static_cast<DeviceChModelItem*>(parent.internalPointer()) :
                rootItem;
    DeviceChModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex CalibrationModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const DeviceChModelItem *childItem = static_cast<DeviceChModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    DeviceChModelItem *parentItem = childItem->parent;
    if(parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int CalibrationModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
                static_cast<DeviceChModelItem*>(parent.internalPointer()) : rootItem;
    return parentItem->childCount();
}

void CalibrationModel::setCalConfig(int config_key, const CalConfig &newCalConfig)
{
    Q_UNUSED(config_key);
    if(cal_config == newCalConfig)
        return;

    bool performReset = cal_config.keys() != newCalConfig.keys();
    if(!performReset) {
        for(auto &cr : newCalConfig)
            performReset |= cal_config[cr.index].nch != cr.nch;
    }

    cal_config = newCalConfig;

    if(performReset) {
        resetItems();
    } else {
        auto top = index(0,0);
        auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
        auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
        emit dataChanged(top, lastCh);
    }
}

void CalibrationModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    if(this->map == map)
        return;
    this->map  = map;
    resetItems();
}

QVariant CalibrationModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    switch (col) {
    case TreeColumns::NAME:
        if(item->dev.isNull()) {
            return "All";
        } else if(item->ch == -1) {
            return item->dev.getSerialStr();
        } else {
            return QString::number(item->ch);
        }
    case TreeColumns::DAC:
        if(item->dev.isNull() || item->ch == -1) {
            return QVariant();
        } else {
            auto v = cal_config.value(item->dev).dacOffset.value(item->ch, 0);
            return getThrText(v);
        }
    default:
    {
        if(item->ch == -1)
            return QVariant();

        QMap<int, double> CalResult::* ptr = nullptr;
        if(col == TreeColumns::BASELINE)
            ptr = &CalResult::adcVals;
        else if(col == TreeColumns::BASELINE_RMS)
            ptr = &CalResult::rms;
        else if(col == TreeColumns::BASELINE4)
            ptr = &CalResult::adcVals4;
        else if(col == TreeColumns::BASELINE4_RMS)
            ptr = &CalResult::rms4;
        if(!ptr)
            return QVariant();

        return QString::number((cal_config[item->dev].*ptr)[item->ch], 'f', 2);
    }
    }
}

QVariant CalibrationModel::getForegroundData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    if(!item->dev.isNull() && item->ch == -1) {
        bool ok = isDeviceRmsOk(item->dev);
        QBrush b;
        b.setColor(ok ? Qt::black : Qt::red);
        return b;
    }


    QMap<int, double> CalResult::* ptr = nullptr;
    if(col == TreeColumns::BASELINE_RMS)
        ptr = &CalResult::rms;
    else if(col == TreeColumns::BASELINE4_RMS)
        ptr = &CalResult::rms4;
    else
        return QVariant();

    if(!item->dev.isNull() && item->ch != -1) {
        Q_ASSERT(cal_config.contains(item->dev));
        const auto &map = cal_config[item->dev].*ptr;
//        Q_ASSERT(map.contains(item->ch)); // rms map can be empty
        bool ok = map.value(item->ch, 0) < RMS_LIMIT;
        QBrush b;
        b.setColor(ok ? Qt::black : Qt::red);
        return b;
    }
    return QVariant();
}

QVariant CalibrationModel::fontData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    QFont font;
//    if(item->dev.isNull()) {
//        font.setItalic(true);
//        font.setBold(true);
//    } else
        if(item->ch == -1) {
        font.setBold(true);
    } else {
//            fg.setItalic(true);
    }
        return font;
}

QVariant CalibrationModel::getHeaderForeground(CalibrationModel::TreeColumns col) const
{
    QMap<int, double> CalResult::* ptr = nullptr;
    if(col == TreeColumns::BASELINE_RMS)
        ptr = &CalResult::rms;
    else if(col == TreeColumns::BASELINE4_RMS)
        ptr = &CalResult::rms4;
    else
        return QVariant();
    bool ok = true;
    for(auto &cr : cal_config) {
        for(auto rms : cr.*ptr){
            if(rms > RMS_LIMIT) {
                ok = false;
                break;
            }
        }
        if(!ok)
            break;
    }
    return QBrush (ok ? Qt::black : Qt::red);
}

bool CalibrationModel::isDeviceRmsOk(const DeviceIndex &index) const
{
    Q_ASSERT(cal_config.contains(index));
    if(!cal_config.contains(index))
        return false;
    auto &cr = cal_config[index];
    for(auto v : cr.rms) {
        if(v>RMS_LIMIT)
            return false;
    }
    for(auto v : cr.rms4) {
        if(v>RMS_LIMIT)
            return false;
    }
    return true;
}

CalibrationModel::TreeColumns CalibrationModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<TreeColumns>(index.column());
}

DeviceChModelItem *CalibrationModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<DeviceChModelItem*>(index.internalPointer());
}

int CalibrationModel::getDevRow(const DeviceIndex &di)
{
    auto allItem = rootItem->children.first();
    if(allItem) {
        for(int r=0; r<allItem->children.size(); ++r)
            if(allItem->children[r]->dev == di)
                return r;
    }
    return -1;
}

void CalibrationModel::resetItems()
{
    beginResetModel();
    rootItem->children.clear();

    for(auto devIt=cal_config.begin();devIt!=cal_config.end(); ++devIt) {
        if(map.contains(devIt.key())){
            if(map[devIt.key()].enabled){
                auto devItem = new DeviceChModelItem(rootItem);
                devItem->dev = devIt.key();
                for(int ch=0; ch<devIt->nch; ++ch) {
                    auto chItem = new DeviceChModelItem(devItem);
                    chItem->ch = ch;
                    devItem->addChild(chItem);
                }
                rootItem->addChild(devItem);
            }
        }
    }
    endResetModel();
}
