#include "DeviceChModel.h"

#include <QDebug>
#include <QFont>
#include <QTimer>

namespace {
QString getRateStr(double rate) {
    QString mask;
    double downScale;
    if(rate<1e3) {
        mask = "%1 Hz";
        downScale = 1;
    } else if(rate<1e6) {
        mask = "%1 kHz";
        downScale = 1e3;
    } else {
        mask = "%1 MHz";
        downScale = 1e6;
    }
    return mask.arg(rate/downScale, 0, 'f', 2);
}
const QString TYPE_HERE = "-- type here --";
}

DeviceChModel::DeviceChModel(QObject *parent)
    :
      QAbstractItemModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(0);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &DeviceChModel::postCheckStateChange);

}

QVariant DeviceChModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::FontRole:
        return fontData(index);
    }
    return QVariant();
}

bool DeviceChModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    switch (role) {
    case Qt::CheckStateRole:
    {
        const auto item = getItemByIndex(index);
        if(item->ch == -1) {
            auto rows = rowCount(index);
            bool ok = true;
            for(int r=0; r<rows; ++r){
                ok &= setData(this->index(r,index.column(),index), value, role);
            }
            return ok;
        } else {
            Q_ASSERT(!item->dev.isNull());
            auto col = static_cast<TreeColumns>(index.column());
            QList<QMap<int,bool> DeviceConfig::*> ptrs;
            switch(col) {
            case TreeColumns::COL_NAME:
                ptrs << &DeviceConfig::chTdcEn;
                ptrs << &DeviceConfig::chAdcEn;
                break;
            case TreeColumns::COL_TDC_EN:
                ptrs << &DeviceConfig::chTdcEn;
                break;
            case TreeColumns::COL_ADC_EN:
                ptrs << &DeviceConfig::chAdcEn;
                break;
            case TreeColumns::COL_INVERT:
                ptrs << &DeviceConfig::chInv;
                break;
            case TreeColumns::COL_TRIG_EN:
                ptrs << &DeviceConfig::chTrigEn;
                break;
            default:
                return false;
            }
            Q_ASSERT(config.knownConfigs.contains(item->dev));
            if(!config.knownConfigs.contains(item->dev))
                return false;
            auto &ds = config.knownConfigs[item->dev];
            for(auto ptr : ptrs) {
                auto &map = (ds.*ptr);
                Q_ASSERT(map.contains(item->ch));
                if(!map.contains(item->ch))
                    return false;
                Q_ASSERT(value!=Qt::PartiallyChecked);
                map[item->ch] = (value==Qt::Checked) ? true : false;
                if(!delayedSendConfigTimer->isActive())
                    delayedSendConfigTimer->start();
            }
        }
        return true;
    }
    case Qt::EditRole:
    {
        const auto item = getItemByIndex(index);
        if(item->ch == -1) {
            if(index.parent() == QModelIndex()){
                config.setThreshold(value.toInt());
            }
            else{
                config.knownConfigs[item->dev].setThreshold(value.toInt());
            }
            emit configChanged(config);
            emit dataChanged(index, index, {Qt::DisplayRole});
            auto rows = rowCount(index);
            bool ok = true;
            for(int r=0; r<rows; ++r){
                ok &= setData(this->index(r,index.column(),index), value, role);
            }
            return ok;
        }
        config.knownConfigs[item->dev].chThr[item->ch] = value.toInt();
        emit configChanged(config);
        emit dataChanged(index, index, {Qt::DisplayRole});
    }
    }

    return false;
}

void DeviceChModel::setEnabled(bool enabled){
    this->enable = enabled;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}


Qt::ItemFlags DeviceChModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = static_cast<TreeColumns>(index.column());
    if(col == TreeColumns::COL_NAME || col == TreeColumns::COL_INVERT
            || col == TreeColumns::COL_TDC_EN || col == TreeColumns::COL_ADC_EN
            || col == TreeColumns::COL_TRIG_EN)
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
    if(col == TreeColumns::COL_THRESHOLD)
        f |= Qt::ItemIsEditable;

    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant DeviceChModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return QVariant();
    if(role != Qt::DisplayRole)
        return QVariant();
    switch (static_cast<TreeColumns>(section)) {
    case TreeColumns::COL_NAME:
        return "Name";
    case TreeColumns::COL_TDC_EN:
        return "TdcEn";
    case TreeColumns::COL_ADC_EN:
        return "AdcEn";
    case TreeColumns::COL_INVERT:
        return "Inv";
    case TreeColumns::COL_TRIG_EN:
        return "TrigEn";
    case TreeColumns::COL_THRESHOLD:
        return "Threshold";
    case TreeColumns::COL_RATE:
        return "Rate, Hz";
    default:
        return QVariant();
    }
}

QModelIndex DeviceChModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const DeviceChModelItem *parentItem = parent.isValid() ?
        static_cast<DeviceChModelItem*>(parent.internalPointer()) :
        &rootItem;
    DeviceChModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex DeviceChModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const DeviceChModelItem *childItem = static_cast<DeviceChModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    DeviceChModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int DeviceChModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<DeviceChModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

void DeviceChModel::setConfig(int config_key, const TqdcConfig &newConfig)
{
    Q_UNUSED(config_key);
    if(config == newConfig)
        return;
    bool performReset = config.knownConfigs.keys() != newConfig.knownConfigs.keys();
    if(!performReset) {
        for(auto &ds : config.knownConfigs) {
            performReset |= ds.getNCh() != newConfig.knownConfigs[ds.index].getNCh()
                    || ds.enabled != newConfig.knownConfigs[ds.index].enabled;
        }
    }

    config = newConfig;
    if(performReset) {
        beginResetModel();
        rootItem.children.clear();
        const auto &setups = config.knownConfigs;
        if(config.hasEnabledDevices()){
            DeviceChModelItem * allItem = new DeviceChModelItem(&rootItem);
            rootItem.addChild(allItem);

            for(auto devIt=setups.begin();devIt!=setups.end(); ++devIt) {
                if(!devIt->enabled)
                    continue;
                auto devItem = new DeviceChModelItem(allItem);
                devItem->dev = devIt.key();
                for(auto chIt=devIt->chTdcEn.begin();chIt!=devIt->chTdcEn.end();++chIt) {
                    auto chItem = new DeviceChModelItem(devItem);
                    chItem->ch = chIt.key();
                    devItem->addChild(chItem);
                }
                allItem->addChild(devItem);
            }
        }
        endResetModel();
    } else {
        sendDataChanged();
    }
}

void DeviceChModel::deviceStatusUpdated(const TqdcStatus &status)
{
    dev_status[status.index] = status;
    auto devIndex = index(getDevRow(status.index), 0, index(0,0));
    auto rows = rowCount(devIndex);
    if(rows) {
        auto top = index(0,static_cast<int>(TreeColumns::COL_RATE), devIndex);
        auto bottom = top.sibling(rows, top.column());
        emit dataChanged(top, bottom, {Qt::DisplayRole});
    }
}

void DeviceChModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void DeviceChModel::sendDataChanged(QVector<int> roles)
{
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh, roles);
}

QVariant DeviceChModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case TreeColumns::COL_NAME:
        if(item->dev.isNull()) {
            return "All";
        } else if(item->ch == -1) {
            return item->dev.getSerialStr();
        } else {
            return QString::number(item->ch);
        }
    case TreeColumns::COL_THRESHOLD:
    {
        const auto &val = editData(index);
        if(val.isValid())
            return QString("%1 mV").arg(val.toInt());
        break;
    }
    case TreeColumns::COL_RATE:
        if(item->dev.isNull() || item->ch == -1) {
            //todo: should we display sum(rate)?
            return "";
        } else {
            auto v = dev_status.value(item->dev).hitRate.value(item->ch, 0);
            return getRateStr(v);
        }
    default:
        return QVariant();
    }
}

QVariant DeviceChModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto &col = getColumnByIndex(index);
    if(col!=TreeColumns::COL_THRESHOLD)
        return QVariant();

    if(item->ch == -1) {
        if(index.parent() == QModelIndex()) {
            return config.getSimilarThreshold();
        } else {
            return config.knownConfigs[item->dev].getSimilarThreshold();
        }
    } else {
        return config.knownConfigs[item->dev].chThr[item->ch];
    }
}

QVariant DeviceChModel::checkStateData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    config.knownConfigs.value(item->dev).chTdcEn.value(item->ch);
    switch (getColumnByIndex(index)) {
    case TreeColumns::COL_NAME:
    {
        auto st1 = data(index.sibling(index.row(), index.column()+1),
                        Qt::CheckStateRole);
        auto st2 = data(index.sibling(index.row(), index.column()+2),
                        Qt::CheckStateRole);
        return st1 == st2 ? st1 : Qt::PartiallyChecked;
    }
    case TreeColumns::COL_TDC_EN:
    case TreeColumns::COL_ADC_EN:
    case TreeColumns::COL_INVERT:
    case TreeColumns::COL_TRIG_EN:
    {
        QMap<int, bool> DeviceConfig::* ptr = nullptr;
        switch (getColumnByIndex(index)) {
        case TreeColumns::COL_TDC_EN:
            ptr = &DeviceConfig::chTdcEn;
            break;
        case TreeColumns::COL_ADC_EN:
            ptr = &DeviceConfig::chAdcEn;
            break;
        case TreeColumns::COL_INVERT:
            ptr = &DeviceConfig::chInv;
            break;
        case TreeColumns::COL_TRIG_EN:
            ptr = &DeviceConfig::chTrigEn;
            break;
        default:
            ptr = nullptr;
        }
        if(item->dev.isNull()) {
            return getAllState(ptr);
        } else if(item->ch == -1){
            return getDevState(item->dev, ptr);
        } else {
            return (config.knownConfigs.value(item->dev).*ptr).value(item->ch) ?
                        Qt::Checked : Qt::Unchecked;
        }
    }
    default:
        return QVariant();
    }
}

QVariant DeviceChModel::fontData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    QFont font;
//    if(item->dev.isNull()) {
//        font.setItalic(true);
//        font.setBold(true);
//    } else
        if(item->ch == -1) {
        font.setItalic(true);
    } else {
//            fg.setItalic(true);
    }
    return font;
}

TreeColumns DeviceChModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<TreeColumns>(index.column());
}

DeviceChModelItem *DeviceChModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<DeviceChModelItem*>(index.internalPointer());
}

QVariant DeviceChModel::getAllState(QMap<int, bool> DeviceConfig::*ptr) const
{
    QVariant allSt;
    for(auto &s : config.knownConfigs) {
        auto st = getDevState(s.index, ptr);
        if(allSt.isNull())
            allSt = st;
        else if(allSt != st)
            return Qt::PartiallyChecked;

    }
    return allSt.isNull() ? Qt::Checked : allSt;
}

QVariant DeviceChModel::getDevState(const DeviceIndex &index,
                                        QMap<int, bool> DeviceConfig::*ptr) const
{
    QVariant allSt;
    const auto &map = config.knownConfigs.value(index).*ptr;
    for(auto &s : map) {
        auto chSt = s ? Qt::Checked : Qt::Unchecked;
        if(allSt.isNull())
            allSt = chSt;
        else if(allSt != chSt)
            return Qt::PartiallyChecked;

    }
    return allSt.isNull() ? Qt::Checked : allSt;
}

int DeviceChModel::getDevRow(const DeviceIndex &di) const
{
    DeviceChModelItem * allItem = nullptr;
    if(!rootItem.children.isEmpty())
        allItem = rootItem.children.first();
    if(allItem) {
        for(int r=0; r<allItem->children.size(); ++r)
            if(allItem->children[r]->dev == di)
                return r;
    }
    return -1;
}

