﻿#include "DeviceChTreeView.h"

#include <QHeaderView>
#include <QSettings>

#include "DeviceChModel.h"
#include "TreeColumns.h"
#include "SpinBoxDelegate.h"

const QString EXPAND_DEVICE_MAP = "expand_device_map";

DeviceChTreeView::DeviceChTreeView(QWidget *parent)
    : QTreeView(parent)
    , delegate(new SpinBoxDelegate(this))
{
    connect(this, &QTreeView::expanded, this, &DeviceChTreeView::expanded);
    connect(this, &QTreeView::collapsed, this, &DeviceChTreeView::collapsed);
    header()->setStretchLastSection(false);
    header()->setSectionResizeMode(QHeaderView::ResizeToContents);
}

void DeviceChTreeView::setModel(QAbstractItemModel *model)
{
    QTreeView::setModel(model);
    setItemDelegateForColumn(static_cast<int>(TreeColumns::COL_THRESHOLD), delegate);

    connect(model, &QAbstractItemModel::modelReset, this, &DeviceChTreeView::onModelReset);
}

void DeviceChTreeView::saveAppGuiSettings(QSettings &s) const
{
    s.beginGroup(EXPAND_DEVICE_MAP);
    QStringList devList;
    for(DeviceIndex index : expand_device_map.keys()){
        s.setValue(index.toString(), expand_device_map[index]);
        devList << index.toString();
    }
    s.setValue("devList", devList);
    s.endGroup();
}

void DeviceChTreeView::loadAppGuiSettings(const QSettings &s)
{
    expand_device_map.clear();
    const auto &devList = s.value(EXPAND_DEVICE_MAP+"/devList").toStringList();
    for(const QString &key : devList){
        DeviceIndex index = DeviceIndex::fromString(key);
        if(!index.isNull())
            expand_device_map[index] = s.value(EXPAND_DEVICE_MAP+"/"+key).toBool();
    }
}

void DeviceChTreeView::expanded(const QModelIndex &index)
{
   if(!getDeviceIndex(index).isNull()){
        expand_device_map[getDeviceIndex(index)] = true;
   }
}

void DeviceChTreeView::collapsed(const QModelIndex &index)
{
    if(!getDeviceIndex(index).isNull()){
        expand_device_map[getDeviceIndex(index)] = false;
    }
}

void DeviceChTreeView::onModelReset()
{
    expandTopItem();
    addDelegate();
    expandAll();
}

void DeviceChTreeView::expandTopItem()
{
    auto model = qobject_cast<const DeviceChModel*>(this->model());
    auto top = model->index(0,0);
    expand(top);

    for(DeviceIndex index : expand_device_map.keys()){
        auto row = model->getDevRow(index);
        if(expand_device_map[index])
            expand(model->index(row, 0, top));
        else
            collapse(model->index(row, 0, top));
    }

    for(int row=0; row<this->model()->rowCount(); ++row){
        DeviceIndex index = static_cast<DeviceChModelItem*>(model->index(row, 0, top).internalPointer())->dev;
        if(!expand_device_map.contains(index))
            expand_device_map.remove(index);
     }
}

DeviceIndex DeviceChTreeView::getDeviceIndex(const QModelIndex &index) const
{
    return static_cast<DeviceChModelItem*>(index.internalPointer())->dev;
}

void DeviceChTreeView::addDelegate(const QModelIndex &parent)
{
    const auto cols = {static_cast<int>(TreeColumns::COL_THRESHOLD)};
    if(parent.isValid()) {
        for(auto col : cols) {
            const auto &i = parent.siblingAtColumn(col);
            if(!isPersistentEditorOpen(i))
                openPersistentEditor(i);
        }
    }

    const auto &m = model();
    int r1 = m->rowCount(parent);
    for(int i=0; i<r1; ++i) {
        auto chIndex = m->index(i, 0, parent);
        addDelegate(chIndex);
    }
}
