#include "MainWindow.h"
#include "ui_MainWindow.h"
#include "CalibrationDialog.h"
#include "LogDialog.h"
#include "TcpClient.h"
#include "TQDCDevice.h"
#include "TqdcConfigDispatcher.h"
#include "TqdcSignalRouter.h"
#include "TqdcConfigDb.h"
#include "DeviceChModel.h"
#include "SpinBoxDelegate.h"
#include "TreeColumns.h"

#include "fsm/DeviceMgrFsm.h"
#include "base/Globals.h"
#include "base/config/ConfigSelector.h"
#include "base/DeviceIndexEnabled.h"
#include "base/threadQuit.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "lib-common/UIStyles.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "mldiscover/LldpDiscovery.h"
#include "mstream-lib/MStreamDump.h"
#include "remote-control-server/RemoteControlServer.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"

#include <QBoxLayout>
#include <QFileDialog>
#include <QHostAddress>
#include <QMetaType>
#include <QPlainTextEdit>
#include <QScrollBar>
#include <QSemaphore>
#include <QSettings>
#include <QShortcut>
#include <QSpinBox>
#include <QTimer>

Q_DECLARE_METATYPE(QHostAddress)

namespace {
const int STOP_EVENT_TIMEOUT_MS = 10e3;
const int STATUS_UPDATE_DELAY = 1000;
enum {COL_DEV_SERIAL, COL_DEV_SLOT, COL_DEV_FW, COL_DEV_MCU_VER, COL_DEV_TEMP, COL_DEV_BUSY,
      COL_DEV_UNBLOCK, COL_DEV_BLOCK,
      COLUMN_WR_TIME, COLUMN_WR_TIME_OFFSET,
      COLUMN_WR_LINK_ERROR, COLUMN_WR_SYNC_LOST, COLUMN_WR_RX_CODE_ERROR,
      COL_DEV_MS_ERR, COL_DEV_STATUS, COL_DEV_STREAM, COL_DEV_TOTAL};

constexpr const char *BUSY_TOOLTIP = "Current Busy Status\n"
                                     "1xxx - External input busy\n"
                                     "x1xx - ADC busy\n"
                                     "xx1x - TDC busy\n"
                                     "xxx1 - ReadOut busy";

void setWidgetWarn(QWidget *w, const QStringList &warnList)
{
    bool en = w->isEnabled();
    auto c = en ? (warnList.isEmpty() ? QColor()         : QColor(0xFF8800)) :
                  (warnList.isEmpty() ? QColor(0xbfbfbf) : QColor(0xFFCC44));
    auto p = w->palette();

    auto b = p.brush(w->foregroundRole());
    if(p.color(w->foregroundRole()) != c) {
        p.setColor(w->foregroundRole(), c);
        w->setPalette(p);
    }

    QString tt = warnList.join('\n');
    if(w->toolTip() != tt)
        w->setToolTip(tt);
}
}

class GuardCtrl
{
public:
    GuardCtrl(int *_guardCnt):
    guardCnt(_guardCnt)
    {
        (*guardCnt)++;
    }

    ~GuardCtrl() {(*guardCnt)--;}

    int * const guardCnt;
};

MainWindow::MainWindow(const ConfigSelector &cs, QWidget *parent) :
      BaseMainWindow(parent)
    , ui(new Ui::MainWindow)
    , configuration_manager(new ConfigurationManager(this, cs.getType()))
    , discoverDialog(new DiscoverDialog(this))
    , logDialog(new LogDialog(this))
    , calibrationDialog(new CalibrationDialog(this))
    , treeModel(new DeviceChModel(this))
    , status_update_timer(new QTimer(this))
    , labelFsmState(new QLabel(this))
    , mStreamStatisticDialog(new MStreamStatisticDialog(this))
    , adcSerDesCalDialog(new AdcSerDesCalDialog(this))
{ 
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    status_update_timer->setSingleShot(true);
    status_update_timer->setInterval(STATUS_UPDATE_DELAY);
    connect(status_update_timer, &QTimer::timeout,
            this, &MainWindow::deviceStatusUpdatedRequest);
    ui->statusBar->addWidget(labelFsmState);

    SpinBoxDelegate::initSpinBox(ui->spinBoxThreshold);

    ui->deviceTreeView->setModel(treeModel);

    discoverDialog->setSingleMode(true);
    init_connections();
    init_table();
    setWindowTitle(cs.toWindowTitle());

    configuration_manager->setDefaultConfig(TqdcConfig());

    init_main_window();
    updateControlState();

    auto signalRouter = &TqdcSignalRouter::instance();
    signalRouter->setMainWindow(this);
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->setAdcSerDesCalDialog(adcSerDesCalDialog);
}

MainWindow::~MainWindow()
{
    emit closeAll();
    delete ui;
}

TtlIoWidget *MainWindow::getWidgetTtl() const
{
    return ui->widgetTtl;
}

void MainWindow::init_connections()
{
    QShortcut *shortcut = new QShortcut(QKeySequence(Qt::CTRL+Qt::Key_L), this);
    connect(shortcut, &QShortcut::activated, logDialog, &LogDialog::show);
    connect(this, &MainWindow::addLog, logDialog, &LogDialog::addLog);

    connect(this, &MainWindow::mstream_stat_updated,
            this, &MainWindow::update_mstream_data_rate);

    connect(this, &MainWindow::mstream_stat_updated,
            mStreamStatisticDialog, &MStreamStatisticDialog::stat_updated);
    connect(this, &MainWindow::stat_from_receiver_updated,
            mStreamStatisticDialog, &MStreamStatisticDialog::stat_from_receiver_updated);
}

void MainWindow::init_main_window()
{
    ui->labelAdcPreWin->hide();
    ui->labelAdcPreWinNs->hide();
    ui->spinBoxAdcPreWin->hide();

    ui->groupBoxTrigger->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxAdc->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxTtl->setStyleSheet(UIStyles::gbox_fill_blue);
}

void MainWindow::update_mstream_data_rate(const MStreamStat &ms_stat)
{
    int row = getDevRow(ms_stat.index);
    if(row == -1) return;

    double receive_data_rate=0;

    auto prev_stat_it = prev_ms_stat.find(ms_stat.index);
    if(prev_stat_it == prev_ms_stat.end()) {
        prev_ms_stat.insert(ms_stat.index, ms_stat);
    } else {
        qint64 milisecs = 0;
        milisecs = prev_stat_it->timestamp.msecsTo(ms_stat.timestamp);
        if(milisecs != 0) {
            receive_data_rate = 1e3*
                    static_cast<double>(ms_stat.recvDataSize - prev_stat_it->recvDataSize)/
                    (static_cast<double>(milisecs)*1e6);
        }
        *prev_stat_it = ms_stat;
    }

    ui->tableWidgetDevice->item(row, COL_DEV_STREAM)
            ->setText(QString("%L1").arg(receive_data_rate));
}

void MainWindow::saveAppGuiSettings(QSettings &s) const
{
    ui->deviceTreeView->saveAppGuiSettings(s);
}

void MainWindow::loadAppGuiSettings(const QSettings &s)
{
    ui->deviceTreeView->loadAppGuiSettings(s);
}

void MainWindow::deviceStatusUpdated(const TqdcStatus &status)
{
    dev_status[status.index] = status;
    deviceStatusUpdatedRequest();
}

void MainWindow::configSelectorChanged(const ConfigIndexName &cin)
{
    auto t = ConfigSelector(DaqModule(DaqModule::Type::TQDC),
                            cin).toWindowTitle();
    if(windowTitle() != t)
        setWindowTitle(t);
}

void MainWindow::onFsmStateChange(const FsmState &state)
{
    labelFsmState->setText("FSM: " + fsmStateToString(state));
    this->state = state;
    updateControlState();
}

DeviceIndex MainWindow::getIndexByRow(int row)
{
    auto item = ui->tableWidgetDevice->item(row, COL_DEV_SERIAL);
    if(!item)
        return DeviceIndex();
    const auto &data = item->data(Qt::UserRole);
    if(data.isNull())
        return DeviceIndex();
    return data.value<DeviceIndex>();
}

int MainWindow::getDevRow(const DeviceIndex &index)
{
    for(int row=ui->tableWidgetDevice->rowCount()-1; row>=0; --row){
        if(getIndexByRow(row) == index){
            return row;
        }
    }
    return -1;
}

void MainWindow::temperatureSet(DeviceIndex index, QMap<int, double> temps)
{
    const int row = getDevRow(index);
    if(row == -1 || temps.isEmpty()) return;

    QList<double> tempList = temps.values();
    double minTemp = *std::min_element(tempList.begin(), tempList.end());
    double maxTemp = *std::max_element(tempList.begin(), tempList.end());
    QColor col = (maxTemp>=70 || minTemp<0) ? Qt::red : Qt::black;
    QString tempStr = QString("%1-%2%3").arg(minTemp,0,'f',1).arg(maxTemp,0,'f',1)
            .arg(QChar(0x02da));
    QStringList fullInfo;
    for (int tempId : temps.keys()) {
        QString suffixStr;
        switch (tempId) {
        case 0:
            suffixStr = "Board temp.";
            break;
        case 1:
            suffixStr = "Pll temp.";
            break;
        case 2:
            suffixStr = "Analog input temp.";
            break;
        default:
            break;
        }
        if(!suffixStr.isEmpty())
            fullInfo << QString("%1%2 - %3").arg(temps[tempId],0,'f',1)
                        .arg(QChar(0x02da)).arg(suffixStr);
    }
    QTableWidgetItem * item = ui->tableWidgetDevice->item(row, COL_DEV_TEMP);
    item->setText(tempStr);
    auto fg = item->foreground();
    fg.setColor(col);
    item->setForeground(fg);
    item->setToolTip(fullInfo.join('\n'));
    updateTableSize(ui->tableWidgetDevice);
}

void MainWindow::setTableDevRowCnt(int cnt)
{
    auto t = ui->tableWidgetDevice;
    int oldRowCnt = t->rowCount();
    t->setRowCount(cnt);
    for(int row=oldRowCnt; row<cnt; ++row) {
        for(int col=0; col<COL_DEV_TOTAL; ++col) {
            t->setItem(row, col, new QTableWidgetItem());
            if(col == COL_DEV_MS_ERR) {
                auto item = t->item(row, col);
                auto fg = item->foreground();
                fg.setColor(Qt::red);
                item->setForeground(fg);
            }
        }
    }
}

void MainWindow::deviceStatusUpdatedRequest()
{
    if(status_update_timer == QObject::sender()) {
        if(status_request_flag){
            deviceStatusUpdatedResponce();
            status_request_flag = false;
        }
    }
    else{
        if(status_update_timer->remainingTime() > 0){
            status_request_flag = true;
        }
        else{
            deviceStatusUpdatedResponce();
            status_request_flag = false;
            status_update_timer->start();
        }
    }
}

void MainWindow::deviceStatusUpdatedResponce()
{
    for(auto &s : dev_status.keys()){
        auto status = dev_status[s];
        if(!config.knownConfigs.contains(status.index) ||
                !config.knownConfigs.value(status.index).enabled) {
            dev_status.remove(s);
            continue;
        }

        treeModel->deviceStatusUpdated(status);
        const int row = getDevRow(status.index);
        if(row == -1) continue;

        if(status.connected) {
            temperatureSet(status.index, status.temp);
            ui->tableWidgetDevice->item(row, COL_DEV_FW)->setText(status.fwStr);
            ui->tableWidgetDevice->item(row, COL_DEV_MCU_VER)->setText(QString("%1.%2")
                .arg(status.mcuVer).arg(status.mcuRev));
            updateDevWrStatus(row, status.wrStatus);
            ui->tableWidgetDevice->item(row, COL_DEV_BUSY)->
                setText(QString("%1").arg(status.busyStatus, 4, 2, QChar('0')));
            ui->tableWidgetDevice->item(row, COL_DEV_UNBLOCK)->
                setText(QString("%L1").arg(status.statistic.evNumAccepted));
            if(status.statistic.evNumCandidateRun<status.statistic.evNumAccepted) {
                qint64 delta = (qint64)status.statistic.evNumCandidateRun-
                    status.statistic.evNumAccepted;
                ui->tableWidgetDevice->item(row, COL_DEV_BLOCK)->
                        setText(QString("%L1").arg(delta));
                ui->tableWidgetDevice->item(row, COL_DEV_BLOCK)->
                    setTextColor(Qt::red);
            } else {
            ui->tableWidgetDevice->item(row, COL_DEV_BLOCK)->
                    setText(QString::number(
                        status.statistic.evNumCandidateRun -
                        status.statistic.evNumAccepted));
                if(ui->tableWidgetDevice->item(row, COL_DEV_BLOCK)->textColor() == Qt::red)
                    ui->tableWidgetDevice->item(row, COL_DEV_BLOCK)->
                        setTextColor(Qt::black);
            }
#ifdef QT_DEBUG
//        qInfo() << QString("All:%1; Candidate:%2; Accepted:%3;")
//                   .arg(status.statistic.evNumAll)
//                   .arg(status.statistic.evNumCandidateRun)
//                   .arg(status.statistic.evNumAccepted);
#endif

            QString str;
            if(!status.hptdcStatus[0].getSummary().isEmpty()) {
                str += QString("HPTDC #1:\n%1").arg(status.hptdcStatus[0].getSummary());
            }
            if(!status.hptdcStatus[1].getSummary().isEmpty()) {
                if(!str.isEmpty()) str += "\n";
                str += QString("HPTDC #2:\n%1").arg(status.hptdcStatus[1].getSummary());
            }
            if(!status.adcSelfTestOk) {
                if(!str.isEmpty()) str += "\n";
                str += QString("DesStatus: %1").arg(status.adcStatus, 8, 2, QChar('0'));
            }
            if(str.isEmpty()) {
                str = "Ok";
            }

            bool err = status.hptdcStatus[0].hasError() || status.hptdcStatus[1].hasError()
                     || !status.adcSelfTestOk;
            ui->tableWidgetDevice->item(row, COL_DEV_STATUS)->setText(str);
            ui->tableWidgetDevice->item(row, COL_DEV_STATUS)->
                setTextColor(err?Qt::red : Qt::black);

            if(lookForStop && status.runFinished) {
                //TODO: move phasing stop logic to core
                //TODO: get stopEvNum in status after closing triggers
                if(status.statistic.evNumAccepted) {
                    stopEvNum = status.statistic.evNumAccepted-1;
                    emit prepareForStop(stopEvNum);
                } else {
                    stopEvNum = 0;
                }
                lookForStop = false;
                qInfo().noquote() <<
                    "[MainWindow]: Stop event number detected:" <<
                    status.statistic.evNumAccepted-1;
            }
        } else {
            auto itemStatus = ui->tableWidgetDevice->item(row, COL_DEV_STATUS);
            itemStatus->setText("Offline");
            auto fg = itemStatus->foreground();
            fg.setColor(Qt::black);
            itemStatus->setForeground(fg);
        }
    }
    updateTableSize(ui->tableWidgetDevice);
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    underRemoteCtrl = status.portInUse;
    updateControlState();
}

bool MainWindow::isControlBlocked() const
{
    return blockControl || underRemoteCtrl;
}

void MainWindow::updateMatchWinMaximum() const
{
    int v = config.matchWinLimitEn ? config.defaultSetup.latency : 670;
    ui->spinBoxMatchWin->setMaximum(std::min(v, 670));
}

void MainWindow::updateWdgetWarnings() const
{
    const auto &ds= config.defaultSetup;
    QStringList warnList;
    if(ds.zsEn && ((ds.matchWin*3 + ds.readCellNumber) >= (670*3)))
        warnList << "MatchWin+AdcWin > 16us";

    setWidgetWarn(ui->checkBoxAdcZs, warnList);
    setWidgetWarn(ui->labelAdcWinNs, warnList);

    if(ds.matchWin > ds.latency)
        warnList << "MatchWin > Latency";

    setWidgetWarn(ui->labelMatchWinNs, warnList);
}

void MainWindow::deviceConnected(const DeviceIndex &index)
{
    auto row = getDevRow(index);
    if(row == -1)
        return;
    ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->setTextColor(Qt::darkGreen);
    updateTableSize(ui->tableWidgetDevice);
}

void MainWindow::deviceDisconnected(const DeviceIndex &index)
{
    auto row = getDevRow(index);
    if(row == -1)
        return;
    ui->tableWidgetDevice->item(row, COL_DEV_SERIAL)->setTextColor(Qt::red);
}

void MainWindow::deviceInfoUpdated(DeviceIndex index, int chNum, int maxSampleCount)
{
    auto cfgIt = config.knownConfigs.find(index);
    if(cfgIt == config.knownConfigs.end())
        return;
    auto &cfg = cfgIt.value();
    if(cfg.getNCh() != chNum) {
        cfg.setNCh(chNum);
        emit configChanged(config);
    }
    ui->spinBoxLatency->setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
//    ui->spinBoxMatchWin -> setMaximum(maxSampleCount ? maxSampleCount/3 : 416);
    ui->spinBoxAdcWin->setMaximum(maxSampleCount ? maxSampleCount/2  : 625);
    updateGuiVal();
}

void MainWindow::setConfig(int config_key, const TqdcConfig &newConfig)
{
    Q_UNUSED(config_key);
    if(config == newConfig)
        return;

    config = newConfig;
    updateGuiVal();
    updateControlState();

    const auto &knownSetups = config.knownConfigs;
    //add devices
    for(DeviceIndex index : knownSetups.keys()) {
        auto row = getDevRow(index);
        if(knownSetups[index].enabled && row==-1) {
            addDevice(index);
        }
    }

    //remove from devices
    for(int r = ui->tableWidgetDevice->rowCount()-1; r>=0; --r) {
        const auto &i = getIndexByRow(r);
        if(i.isNull()) {
            Q_ASSERT(ui->tableWidgetDevice->rowCount()==1 && r==0);
            continue;
        }
        if(knownSetups.contains(i) && knownSetups[i].enabled)
            continue;

        disconnectFromDev(i);
    }

    updateTableSize(ui->tableWidgetDevice);
    updateGuiVal();
}

void MainWindow::lldpCacheUpdated(const LldpInfoCache &cache)
{
    lldpCache = cache;

    for(const auto &lldp : lldpCache.values()) {
        const int row = getDevRow(DeviceIndex(lldp));
        if(row == -1)
            continue;

        QString txt = lldp.isNull() ? "-" :
                                      QString("%1").arg(lldp.chassis_slot, 2, 10, QChar('0'));

        auto itemSlot = ui->tableWidgetDevice->item(row, COL_DEV_SLOT);
        if(itemSlot && itemSlot->text() != txt) {
            itemSlot->setText(txt);
        }
    }
}

void MainWindow::init_table()
{
    ++blockControl;

    QTableWidget * const tableDev = ui->tableWidgetDevice;
    tableDev->verticalHeader()->hide();
    tableDev->setColumnCount(COL_DEV_TOTAL);
    tableDev->setHorizontalHeaderItem(COL_DEV_SERIAL, new QTableWidgetItem("Serial"));
    tableDev->setHorizontalHeaderItem(COL_DEV_SLOT, new QTableWidgetItem("Slot"));
    tableDev->setHorizontalHeaderItem(COL_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));
    tableDev->setHorizontalHeaderItem(COL_DEV_TEMP, new QTableWidgetItem("temp"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_TIME, new QTableWidgetItem("WR Time"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_TIME_OFFSET, new QTableWidgetItem("Offset"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_LINK_ERROR, new QTableWidgetItem("Link"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_SYNC_LOST, new QTableWidgetItem("Unlock"));
    tableDev->setHorizontalHeaderItem(COLUMN_WR_RX_CODE_ERROR, new QTableWidgetItem("RX Err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_BUSY, new QTableWidgetItem("Busy"));
    tableDev->setHorizontalHeaderItem(COL_DEV_UNBLOCK, new QTableWidgetItem("Acepted"));
    tableDev->setHorizontalHeaderItem(COL_DEV_BLOCK, new QTableWidgetItem("Blocked"));
    tableDev->setHorizontalHeaderItem(COL_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STATUS, new QTableWidgetItem("Status"));
    tableDev->setHorizontalHeaderItem(COL_DEV_STREAM, new QTableWidgetItem("Data Stream, MB/s"));
    tableDev->hideColumn(COLUMN_WR_TIME_OFFSET);
    tableDev->hideColumn(COL_DEV_MS_ERR);

    setTableDevRowCnt(1);
    tableDev->item(0, COL_DEV_SERIAL)->setText("");
    tableDev->item(0, COL_DEV_STATUS)->setText("No device selected");

    tableDev->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    tableDev->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize(tableDev);

    tableDev->horizontalHeaderItem(COL_DEV_BUSY)->setToolTip(BUSY_TOOLTIP);
    --blockControl;
}

void MainWindow::copyAndSetupDevice()
{
    for(auto &ds : config.knownConfigs)
        ds.global_mode_copy(config.defaultSetup);
    setupDevice();
}

void MainWindow::setupDevice()
{
    if(blockControl) return;
    emit configChanged(config);
}

void MainWindow::updateGuiVal()
{
    ++blockControl;
    const DeviceConfig &ds = config.defaultSetup;

    const bool selfTrig = ds.trigsetup.bit.selftrig_en;
    ui->checkBoxSelfTrig->setChecked(selfTrig);
    ui->spinBoxSelfTrigDelay->setEnabled(selfTrig);
    ui->deviceTreeView->setColumnHidden(static_cast<int>(TreeColumns::COL_TRIG_EN), !selfTrig);

    ui->checkBoxTrigTimer->setChecked(ds.trigsetup.bit.timer_en);
    ui->spinBoxTimeTrig->setEnabled(ds.trigsetup.bit.timer_en);
    double rate = 125e6/ds.trigPeriod;
    ui->spinBoxTimeTrig->setValue(static_cast<int>(rate));
    ui->spinBoxSelfTrigDelay->setValue(ds.selfTrigDelay);
    ui->labelSelfTrigDelay->setText(QString("%1 ns").arg(ds.selfTrigDelay*8));

    ui->checkBoxTrigTtl->setChecked(ds.trigsetup.bit.ttl_en);
    ui->checkBoxTrigTtc->setChecked(ds.trigsetup.bit.ttc_en);

    ui->spinBoxAdcPreWin->setValue(ds.adcPreWin);
    ui->spinBoxAdcWin->setValue(ds.readCellNumber);
    ui->spinBoxLatency->setValue(ds.latency);
    updateMatchWinMaximum();
    ui->spinBoxMatchWin->setValue(ds.matchWin);
    updateWdgetWarnings();

//    ui->treeViewChCfg->setColumnHidden(static_cast<int>(DevChCfgTreeModel::TreeColumns::COL_NAME), config.separateChEn);
    for(auto c : {TreeColumns::COL_TDC_EN, TreeColumns::COL_ADC_EN})
        ui->deviceTreeView->setColumnHidden(static_cast<int>(c),
                                            !config.separateChEn);

    for(auto c : {COLUMN_WR_LINK_ERROR, COLUMN_WR_SYNC_LOST, COLUMN_WR_RX_CODE_ERROR})
        ui->tableWidgetDevice->setColumnHidden(c, !config.showAllWrInfo);

    Q_ASSERT(ds.getNCh());
    ui->spinBoxThreshold->setValue(config.getSimilarThreshold());

    ui->checkBoxAdcZs->setChecked(ds.zsEn);
    ui->spinBoxAdcWin->setEnabled(ds.zsEn);
    ui->spinBoxAdcPreWin->setEnabled(ds.zsEn);

    ui->radioButtonGain1->setChecked(!ds.isGain4());
    ui->radioButtonGain4->setChecked(ds.isGain4());
    ui->actionAllow_remote_control->setChecked(config.allowRemoteControl);
    ui->actionSeparate_channel_enable->setChecked(config.separateChEn);
    ui->actionShow_all_WhiteRabbit_info->setChecked(config.showAllWrInfo);
    ui->actionDisable_match_win_limitations->setChecked(!config.matchWinLimitEn);

    --blockControl;
    updateControlState();
}

void MainWindow::controlStateEnable(bool enable)
{
    ui->groupBoxTrigger->setEnabled(enable);
    ui->groupBoxAdc->setEnabled(enable);
    ui->groupBoxTtl->setEnabled(enable);
    treeModel->setEnabled(enable);
    updateWdgetWarnings();
}

void MainWindow::updateControlState()
{
    bool inStateErrReady = state == FsmState::Ready || state == FsmState::Error;

    ui->pushButtonStart->setEnabled(state == FsmState::Ready && !underRemoteCtrl);
    ui->pushButtonStop->setEnabled(state == FsmState::Run && !underRemoteCtrl);

    controlStateEnable((inStateErrReady || Globals::getDebugMode())
                       && state != FsmState::Calibrate);

    ui->pushButtonChangeDevice->setEnabled(inStateErrReady);
    ui->actionConfiguration_manger->setEnabled(inStateErrReady);
    ui->actionReset_MStream->setEnabled(inStateErrReady);
    ui->actionSend_soft_reset->setEnabled(inStateErrReady);

    ui->actionAllow_remote_control->setEnabled(!underRemoteCtrl ||
                                               state != FsmState::Run);
}

void MainWindow::disconnectFromDev(const DeviceIndex &index)
{
    dev_status.remove(index);
    const int row = getDevRow(index);
    auto t = ui->tableWidgetDevice;
    if(row==0 && t->rowCount() == 1){
        for(int i=0; i<COL_DEV_TOTAL; ++i){
            auto serial_item = t->item(row, i);
            serial_item->setText("");
            serial_item->setData(Qt::UserRole, QVariant());
        }
        t->item(row, COL_DEV_STATUS)->setText("No device selected");
    } else {
        t->removeRow(row);
    }
    updateControlState();
    calibrationDialog->reset();
    mStreamStatisticDialog->remove_device(index);
}

void MainWindow::addDevice(const DeviceIndex &index)
{
    calibrationDialog->reset();

    auto t = ui->tableWidgetDevice;
    int row = getIndexByRow(0).isNull() ? 0 : t->rowCount();
    setTableDevRowCnt(row+1);

    t->item(row, COL_DEV_SERIAL)->setText(index.getSerialStr());
    t->item(row, COL_DEV_SERIAL)->setData(Qt::UserRole, QVariant::fromValue(index));
    t->item(row, COL_DEV_STATUS)->setText("connecting...");
    const auto &lldpInfo = lldpCache.value(index);
    QString slot_str = lldpInfo.isNull() ? "-"
                                         : QString::number(lldpInfo.chassis_slot);
    t->item(row, COL_DEV_SLOT)->setText(slot_str);

    mStreamStatisticDialog->add_device(index);
}

void MainWindow::sendLog(QString l)
{
    emit addLog(l);
}

void MainWindow::updateTableSize(QTableWidget * const table, bool widthOnly)
{
    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

void MainWindow::updateDevWrStatus(int row, const WrStatus &wrStatus)
{
    const QColor bgColorDefault = style()->standardPalette().light().color();
    const QColor bgColorWarning = Qt::yellow;
    const QColor bgColorError = Qt::red;

    auto table = ui->tableWidgetDevice;
    bool wrTimeValid = false;
    const auto &sfpOk = wrStatus.statusCode.bit.sfpOk;
    if (wrStatus.hasWrEpStatus && sfpOk) {
        wrTimeValid = wrStatus.timeValid();
        const bool wrTimeNull = wrStatus.timeNull();
        const auto itemTime = table->item(row, COLUMN_WR_TIME);
        const auto itemTimeOffset = table->item(row, COLUMN_WR_TIME_OFFSET);

        if (wrTimeNull || !wrStatus.statusCode.bit.linkUp) {
            itemTime->setText(wrStatus.statusCode.bit.linkUp ? QString() : "No link");
            itemTime->setBackground(bgColorWarning);
            itemTimeOffset->setText(QString());
            itemTimeOffset->setBackground(bgColorWarning);
        } else {
            const auto &wrTime = QDateTime::fromMSecsSinceEpoch(wrStatus.mSecsSinceEpoch());
            itemTime->setText(wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss")); // add ".zzz" for ms
            itemTime->setBackground(wrTimeValid ? bgColorDefault : bgColorError);

            // logic from adc64-system with acquired time
//            int wrTimeCheck = wrStatus.wrTimeCheck();
//            double wrTimeOffsetMs = wrStatus.mSecsSinceEpoch() - wrStatus.acquiredAtMs;
//            QColor offsetBgColor = bgColorDefault;
//            if (wrTimeCheck == 1)
//                offsetBgColor = bgColorWarning;
//            if (wrTimeCheck > 1)
//                offsetBgColor = bgColorError;

//            QString strOffset;
//            if (wrTimeOffsetMs/1000. > -60 && wrTimeOffsetMs/1000. < 60)
//                strOffset = QString(" %1").arg(wrTimeOffsetMs * 1e-3, 6, 'f', 3);
//            itemTimeOffset->setText(strOffset);
//            itemTimeOffset->setBackground(offsetBgColor);
        }
        table->item(row, COLUMN_WR_LINK_ERROR)->setText(QString::number(wrStatus.linkError));
        table->item(row, COLUMN_WR_SYNC_LOST)->setText(QString::number(wrStatus.syncFail));
        table->item(row, COLUMN_WR_RX_CODE_ERROR)->setText(QString::number(wrStatus.codeError));
    } else {
        static const auto &WR_COLS = { COLUMN_WR_TIME,
                                       COLUMN_WR_TIME_OFFSET,
                                       COLUMN_WR_LINK_ERROR,
                                       COLUMN_WR_SYNC_LOST,
                                       COLUMN_WR_RX_CODE_ERROR};
        for(auto col : WR_COLS) {
            table->item(row, col)->setText(col==COLUMN_WR_TIME && !sfpOk ? "No SFP" : "N/A");
            table->item(row, col)->setBackground(bgColorDefault);
        }
    }
}

// ==== Slots from UI
void MainWindow::on_pushButtonChangeDevice_clicked()
{
    discoverDialog->show();
}

void MainWindow::on_actionConfiguration_manger_triggered()
{
    configuration_manager->show();
}

void MainWindow::on_actionAllow_remote_control_toggled(bool checked)
{
    if(config.allowRemoteControl != checked) {
        config.allowRemoteControl = checked;
        emit configChanged(config);
    }
}

void MainWindow::on_pushButtonStart_clicked()
{
    if(isControlBlocked()) return;
    if(!config.hasEnabledDevices()) {
        qWarning() << "There is no enabled device";
        return;
    }

    emit performLocalStart();
}

void MainWindow::on_pushButtonStop_clicked()
{
    if(isControlBlocked()) return;
    emit performLocalStop();
}

void MainWindow::on_radioButtonGain1_toggled(bool checked)
{
    on_radioButtonGain4_toggled(!checked);
}

void MainWindow::on_radioButtonGain4_toggled(bool checked)
{
    config.defaultSetup.setGain4(checked);
    for(auto &ds : config.knownConfigs)
        ds.setGain4(checked);
    copyAndSetupDevice();
}

void MainWindow::on_checkBoxAdcZs_toggled(bool zsEn )
{
    ui->spinBoxAdcWin->setEnabled(zsEn);
    ui->spinBoxAdcPreWin->setEnabled(zsEn);

    config.defaultSetup.zsEn = zsEn;
    copyAndSetupDevice();
    updateWdgetWarnings();
}

void MainWindow::on_checkBoxTrigTimer_toggled(bool checked)
{
    ui->spinBoxTimeTrig->setEnabled(checked);

    config.defaultSetup.trigsetup.bit.timer_en = checked ? 1 : 0;
    copyAndSetupDevice();
}

void MainWindow::on_spinBoxTimeTrig_valueChanged(int rateHz)
{
    quint32 period = static_cast<quint32>(125e6/rateHz);

    config.defaultSetup.trigPeriod = period;
    copyAndSetupDevice();
}

void MainWindow::on_checkBoxTrigTtl_toggled(bool checked)
{
    config.defaultSetup.trigsetup.bit.ttl_en = checked ? 1 : 0;
    copyAndSetupDevice();
}

void MainWindow::on_checkBoxTrigTtc_toggled(bool checked)
{
    config.defaultSetup.trigsetup.bit.ttc_en = checked ? 1 : 0;
    copyAndSetupDevice();
}

void MainWindow::on_checkBoxSelfTrig_toggled(bool checked)
{
    ui->spinBoxSelfTrigDelay->setEnabled(checked);
    config.defaultSetup.trigsetup.bit.selftrig_en = checked ? 1 : 0;
    copyAndSetupDevice();
    ui->deviceTreeView->setColumnHidden(
        static_cast<int>(TreeColumns::COL_TRIG_EN), !checked);
}

void MainWindow::on_spinBoxSelfTrigDelay_valueChanged(int delay)
{
    ui->labelSelfTrigDelay->setText(QString("%1 ns").arg(delay*8));
    config.defaultSetup.selfTrigDelay = delay;
    copyAndSetupDevice();
}

void MainWindow::on_spinBoxThreshold_valueChanged(int thr)
{
    config.defaultSetup.setThreshold(thr);
    for(auto &ds : config.knownConfigs) {
        ds.setThreshold(thr);
    }
    if(blockControl) return;
    updateGuiVal();
    setupDevice();
}

void MainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    ui->labelLatencyNs->setText(QString("%1 ns").arg(lat*24));
    auto &ds = config.defaultSetup;
    ds.latency = lat;
    if(config.matchWinLimitEn)
        updateMatchWinMaximum();
    updateWdgetWarnings();
    if(blockControl) return;
    copyAndSetupDevice();
}

void MainWindow::on_spinBoxMatchWin_valueChanged(int win)
{
    ui->labelMatchWinNs->setText(QString("%1 ns").arg(win*24));
    config.defaultSetup.matchWin = win;
    updateWdgetWarnings();
    if(blockControl) return;
    copyAndSetupDevice();
}

void MainWindow::on_spinBoxAdcWin_valueChanged(int val)
{
    ui->labelAdcWinNs->setText(QString("%1 ns").arg(val*8));

    config.defaultSetup.readCellNumber = val;
    if(blockControl) return;
    copyAndSetupDevice();
}

void MainWindow::on_spinBoxAdcPreWin_valueChanged(int preWin)
{
    ui->labelAdcPreWinNs->setText(QString("%1 ns").arg(preWin*8));
    config.defaultSetup.adcPreWin = preWin;
    if(blockControl) return;
    copyAndSetupDevice();
}

void MainWindow::on_actionCalibrationData_triggered()
{
    calibrationDialog->show();
}

void MainWindow::on_actionADC_Ser_Des_calibration_triggered()
{
    adcSerDesCalDialog->show();
}

void MainWindow::on_actionSeparate_channel_enable_toggled(bool checked)
{
    config.separateChEn = checked;
    emit configChanged(config);
    updateGuiVal();
}

void MainWindow::on_actionShow_all_WhiteRabbit_info_toggled(bool checked)
{
    config.showAllWrInfo = checked;
    emit configChanged(config);
    updateGuiVal();
}

void MainWindow::on_actionDisable_match_win_limitations_toggled(bool checked)
{
    config.matchWinLimitEn = !checked;
    emit configChanged(config);
    updateGuiVal();
}

void MainWindow::on_actionReset_MStream_triggered()
{
    emit mstream_reset_all(true);
}

void MainWindow::on_actionSend_soft_reset_triggered()
{
    emit softReset();
}

void MainWindow::on_actionPrint_FSM_inputs_triggered()
{
    emit printFsmInputs();
}

void MainWindow::on_actionMStream_statistic_triggered()
{
    mStreamStatisticDialog->show();
}

void MainWindow::on_actionTxt_data_format_triggered()
{
    QString s;
    s += "<h2>TQDC2 Txt Data format</h2>";
    s += "<p>File constructs with following lines:</p>";
    s += "<ul>";
    s += "<li>Ev <i>&lt;evNumber&gt;</i> <i>&lt;TAI_sec&gt;</i>.<i>&lt;TAI_ns&gt;</i></li>";
    s += "<li>Tdc <i>&lt;ch&gt;</i>: <i>&lt;tdcLeadTime&gt;</i></li>";
    s += "<li>Adc <i>&lt;ch&gt;</i>: <i>&lt;adcTime&gt;</i> <i>&lt;sample1&gt;</i> <i>&lt;sample2&gt;</i> ... <i>&lt;sampleN&gt;</i></li>";
    s += "</ul>";
    s += "<p>where:</p>";
    s += "<dl>";
    s += "<dt><b>TAI time</b></dt>";
    s += "<dd><i>Precise arrival time of the trigger</i></dd>";
    s += "<dt><b>ch</b></dt>";
    s += "<dd><i>Channel number in range 0-15</i></dd>";
    s += "<dt><b>tdcLeadTime</b></dt>";
    s += "<dd><i>HPTDC time in 0.024/1.024ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>adcTime</b></dt>";
    s += "<dd><i>Timestamp of the first sample in 8ns steps from trigger (TAI time)</i></dd>";
    s += "<dt><b>sample</b></dt>";
    s += "<dd><i>ADC sample in signed 16-bit integer</i></dd>";
    s += "</dl>";

    QMessageBox::about (this, "Txt data format", s);
}

void MainWindow::on_actionRaw_data_format_triggered()
{
    QDialog d;
    QPlainTextEdit scrollArea;
    QLabel htmlLabel;

    QFile htmlFile(":/rawDataFormat.html");
    QString html;
    if(htmlFile.open(QIODevice::ReadOnly)){
        html = QString(htmlFile.readAll());
        htmlFile.close();
        if(html.isEmpty())
            html = "<b>Channel map file is empty.</b>";
    } else
        html = "<b>There is no channel map file.</b>";

    scrollArea.appendHtml(html);
    htmlLabel.setText(html);

    QVBoxLayout *mainLayout = new QVBoxLayout(&d);
//    mainLayout->addWidget(&scrollArea);
    mainLayout->addWidget(&htmlLabel);
    d.setLayout(mainLayout);
    d.exec();

//    QMessageBox::about (this, "Raw data format", html);
}
