#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QElapsedTimer>
#include <QMainWindow>
#include <QMessageBox>

#include "TqdcConfig.h"
#include "TqdcStatus.h"
#include "CalibrationDialog.h"

#include "base/app/BaseMainWindow.h"
#include "fsm/FsmState.h"
#include "mldiscover/LldpDiscovery.h"

namespace Ui {
class MainWindow;
}

class QTableWidget;

class ConfigSelector;
class ConfigurationManager;
class DeviceIndexEnabledMap;
class DiscoverDialog;
class LogDialog;
class ConfigIndexName;
struct MStreamStat;
struct ReceiverStat;
class RemoteControlServerStatus;
class RcRunId;
class DeviceChModel;
class CalibrationModel;
class TtlIoWidget;
class MStreamStatisticDialog;
class AdcSerDesCalDialog;

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(const ConfigSelector &cs,
                        QWidget *parent = nullptr);
    ~MainWindow() override;
    DiscoverDialog * getDiscoverDialog() const
    { return discoverDialog; }
    DeviceChModel * getDeviceTreeModel() const
    { return treeModel; }
    TtlIoWidget *getWidgetTtl() const;

signals:
    void configChanged(const TqdcConfig &config) const;
    void switch_to_config(const ConfigIndexName &conf) const; // unused just to subscribe
    void setupChangedForce();
    void closeAll();
    void addLog(QString);
    void softReset();
    void requestInlCorr(const DeviceIndex index);
    void prepareForStop(quint32 stopEvNum);
    void mstream_reset_all(bool doStatClear);
    void mstream_stat_updated(const MStreamStat &ms_stat);
    void stat_from_receiver_updated(const ReceiverStat &stat);
    void sendRemoteState(QString state, bool ready4Run=false);
    void performLocalStart() const;
    void performLocalStop() const;
    void printFsmInputs() const;

public slots:
    void setConfig(int config_key, const TqdcConfig &newConfig);
    void lldpCacheUpdated(const LldpInfoCache &cache);
    void configSelectorChanged(const ConfigIndexName &cin);
    void onFsmStateChange(const FsmState &state);
    void deviceInfoUpdated(DeviceIndex index, int chNum, int maxSampleCount);
    void deviceStatusUpdated(const TqdcStatus &status);
    void deviceConnected(const DeviceIndex &index);
    void deviceDisconnected(const DeviceIndex &index);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);

private slots:
    void update_mstream_data_rate(const MStreamStat &ms_stat);

    void on_pushButtonChangeDevice_clicked();
    void on_actionConfiguration_manger_triggered();
    void on_actionAllow_remote_control_toggled(bool checked);
    void on_pushButtonStart_clicked();
    void on_pushButtonStop_clicked();
    void on_radioButtonGain1_toggled(bool checked);
    void on_radioButtonGain4_toggled(bool checked);
    void on_checkBoxAdcZs_toggled(bool checked);
    void on_checkBoxTrigTimer_toggled(bool checked);
    void on_spinBoxTimeTrig_valueChanged(int rateHz);
    void on_checkBoxTrigTtl_toggled(bool checked);
    void on_checkBoxTrigTtc_toggled(bool checked);
    void on_checkBoxSelfTrig_toggled(bool checked);
    void on_spinBoxSelfTrigDelay_valueChanged(int delay);
    void on_spinBoxThreshold_valueChanged(int thr);
    void on_spinBoxLatency_valueChanged(int lat);
    void on_spinBoxMatchWin_valueChanged(int win);
    void on_spinBoxAdcWin_valueChanged(int val);
    void on_spinBoxAdcPreWin_valueChanged(int preWin);
    void on_actionCalibrationData_triggered();
    void on_actionADC_Ser_Des_calibration_triggered();
    void on_actionSeparate_channel_enable_toggled(bool checked);
    void on_actionShow_all_WhiteRabbit_info_toggled(bool checked);
    void on_actionDisable_match_win_limitations_toggled(bool checked);
    void on_actionReset_MStream_triggered();
    void on_actionSend_soft_reset_triggered();
    void on_actionPrint_FSM_inputs_triggered();
    void on_actionMStream_statistic_triggered();
    void on_actionTxt_data_format_triggered();
    void on_actionRaw_data_format_triggered();

protected:
    virtual void saveAppGuiSettings(QSettings &s) const override;
    virtual void loadAppGuiSettings(const QSettings &s) override;

private:
    DeviceIndex getIndexByRow(int row);
    int getDevRow(const DeviceIndex &index);
    void init_connections();
    void init_main_window();
    void init_table();
    void copyAndSetupDevice();
    void setupDevice();
    void updateGuiVal();
    void controlStateEnable(bool enable);
    void updateControlState();
    void disconnectFromDev(const DeviceIndex &index);
    void addDevice(const DeviceIndex &index);
    void sendLog(QString l);
    void temperatureSet(DeviceIndex index, QMap<int, double> temps);
    void setTableDevRowCnt(int cnt);
    void deviceStatusUpdatedRequest();
    void deviceStatusUpdatedResponce();
    bool isControlBlocked() const;
    void updateMatchWinMaximum() const;
    void updateWdgetWarnings() const;

public:
    static void updateTableSize(QTableWidget * const table,
                                bool widthOnly=false);
    CalibrationDialog * const calibrationDialog;

private:
    void updateDevWrStatus(int row, const WrStatus &wrStatus);

    Ui::MainWindow *ui;
    TqdcConfig config;
    QMap<DeviceIndex, TqdcStatus> dev_status;
    QMap<DeviceIndex, MStreamStat> prev_ms_stat;
    ConfigurationManager * const configuration_manager;
    DiscoverDialog * const discoverDialog;
    int blockControl = 0;
    LogDialog *const logDialog;

    bool isDevBlocked = false;
    bool lookForStop = false;
    quint32 stopEvNum = 0;
    DeviceChModel * const treeModel;
    bool underRemoteCtrl = false;
    QTimer *status_update_timer;
    bool status_request_flag = false;
    QLabel * const labelFsmState;
    LldpInfoCache lldpCache;
    FsmState state;
    MStreamStatisticDialog * const mStreamStatisticDialog;
    AdcSerDesCalDialog * const adcSerDesCalDialog;
};

#endif // MAINWINDOW_H
