#include "TqdcCore.h"

#include <QElapsedTimer>
#include <QThread>

#include "MainWindow.h"
#include "CalResult.h"
#include "CalConfigDb.h"
#include "TqdcConfigDb.h"
#include "TqdcConfigDispatcher.h"
#include "TqdcSignalRouter.h"
#include "TQDCDevice.h"

#include "pnp-server/PNPServer.h"
#include "remote-control-server/RemoteControlServer.h"
#include "fsm/AdcDeviceMgrFsm.h"
#include "base/ConfigIndexName.h"
#include "base/DeviceIndexEnabled.h"
#include "base/threadQuit.h"
#include "mldiscover/DeviceDispatcher.h"

TqdcCore::TqdcCore(const ConfigSelector &cs, MainWindow *mw, QObject *parent)
    : QObject(parent)
    , fsm(new AdcDeviceMgrFsm(this))
    , remoteControlServer(new RemoteControlServer(this, 0))
    , tqdcConfigDb(new TqdcConfigDb(cs, this))
    , calConfigDb(new CalConfigDb(this))
    , pnpServer(new PNPServer(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , cal_key(1)
{
    qRegisterMetaType<CalResult>();
    qRegisterMetaType<CalConfig>();
    qRegisterMetaType<TqdcStatus>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<LldpInfoCache>();

    auto signalRouter = &TqdcSignalRouter::instance();
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setRemoteControlServer(remoteControlServer);
    signalRouter->setCalConfigDb(calConfigDb);
    signalRouter->setConfigDb(tqdcConfigDb);
    signalRouter->setPnpServer(pnpServer);
    signalRouter->setCore(this);
    signalRouter->setFsm(fsm);
    signalRouter->connectAll();

    initTqdcProgDescr();
}

TqdcCore::~TqdcCore()
{
    emit deleteDevice(DeviceIndex());
}

void TqdcCore::setConfig(int config_key, const TqdcConfig &newConfig)
{
    Q_UNUSED(config_key);
    if(config == newConfig)
        return;
    config = newConfig;

    DeviceIndexEnabledMap diem;
    QSet<DeviceIndex> toDel;
    for(auto it=devices.begin(); it!=devices.end(); ++it) {
        const auto cfgIt = config.knownConfigs.find(it.key());
        if(cfgIt == config.knownConfigs.cend() || !cfgIt.value().enabled) {
            toDel << it.key();
            emit deleteDevice(it.key());
        }
    }
    if(!toDel.isEmpty()) {
        QElapsedTimer timer;
        timer.start();
        for(auto index : toDel) {
            auto &dg = devices[index];
            if(!threadQuiting(dg->getDeviceThread(), 2000, &timer)) {
                qWarning() << "Can't stop dev. h/w thread for "<< index.getIdent();
            }
            devices.remove(index);
        }
    }

    for(auto cfgIt=config.knownConfigs.cbegin(); cfgIt!=config.knownConfigs.cend(); ++cfgIt) {
        const auto &cfg = cfgIt.value();
        diem[cfg.index] = {cfg.index, cfg.enabled};
        if(!cfg.enabled)
            continue;
        const auto it = devices.find(cfgIt.key());
        if(it == devices.cend())
            addDevice(cfg.index);
    }
//    cfgDispatcher->sendCache();
    updateTqdcProgDescrOptions();
    emit allowRemoteControl(config.allowRemoteControl);
}

void TqdcCore::configSelectorChanged(const ConfigIndexName &conf)
{
    // FIXME: make method of ProgramDescription
    progDescr.index = conf.getIndex().toString();
    emit changeProgramDescription(progDescr);
}

void TqdcCore::performLocalStart()
{
    fsm->goState(FsmState::Starting);
}

void TqdcCore::performLocalStop()
{
    fsm->goState(FsmState::Stopping);
}

void TqdcCore::remoteStartRequest(const RcRunId &runId_, quint64 startEvNum)
{
    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
        return;
    }

    runId = runId_;
    fsm->goState(FsmState::Starting);
}

void TqdcCore::remoteStopRequest(const RcRunId &runId_, quint64 stopEvNum)
{
    bool success=true;
    if(runId.runIndex != runId_.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(runId_.runIndex)
                     .arg(runId_.runIndex);
        success = false;
    }

    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(runId_, msg);
    }

    fsm->goState(FsmState::Stopping);
}

void TqdcCore::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDescr.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit updateProgramDescription(progDescr);

    if(status.portInUse) {
        QString ready4run_str = "Ready for run";
        for(DeviceIndex &index : devices.keys()){
            if(!devices[index]->getTqdcStatus().connected){
                ready4run_str = "Not ready for run";
            }
        }
//        emit sendRemoteState(ready4run_str);
    } else if(lastRcStatus.portInUse) {
        qWarning() << "Remote control disconnected.";

        if(fsm->getState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            fsm->goState(fsm->getState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }
    lastRcStatus = status;
}

void TqdcCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    switch (state) {
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:{}
        }
        break;
    }
    case FsmState::Starting :
    {
        config.setRunState(true);
        emit configChanged(config);
        fsm->goState(FsmState::Run);
        break;
    }
    case FsmState::Calibrate :
    case FsmState::FindAdcDelay :
    {
        if(++cal_key < 100)
            cal_key = 100;
        fsm->setMainCalKey(cal_key);
        if(state == FsmState::Calibrate)
            emit startCalibration(cal_key);
        else
            emit startAdcSrdDesCalibration(cal_key, adcSerDesCalHardMode, adcSerDesCalResDir);
        fsm->goState(FsmState::Ready);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        break;
    }
    case FsmState::Stopping :
    {
        config.setRunState(false);
        emit configChanged(config);
        fsm->goState(FsmState::Ready);
        break;
    }
    default:
    {}
    }
}

void TqdcCore::addDevice(const DeviceIndex &index)
{
    try {
        devices[index] = new TqdcDeviceGlobal(index);
    } catch (std::exception &e){
        qCritical() << QString("[%1] Error on connect to device: %2")
                       .arg(index.getDeviceIdName()).arg(e.what());
    }
}

void TqdcCore::deviceDisconnected(const DeviceIndex &index)
{
    if(config.isRun()){
        sendRunError(runId, "Device disconnected");
    }
}

void TqdcCore::deviceStatusUpdated(const TqdcStatus &status)
{
    dev_status[status.index] = status;
    bool device_disconnected = false;
    for(DeviceIndex &index : dev_status.keys()){
        if(!dev_status[index].connected)
            device_disconnected = true;
    }
    if(config.isRun() && device_disconnected){
        sendRunError(runId, "Device disconnected");
    }
}

void TqdcCore::startCalibration_()
{
    fsm->goState(FsmState::Calibrate);
}

void TqdcCore::requestAdcSerDesCalibration(bool hardMode, QString dir)
{
    if(fsm->getState() != FsmState::Ready){
        qWarning() << "AdcSerDesCalibration can be performed just in Ready state";
        return;
    }
    adcSerDesCalHardMode = hardMode;
    adcSerDesCalResDir = dir;
    fsm->goState(FsmState::FindAdcDelay);
}

void TqdcCore::initTqdcProgDescr()
{
    auto cs = tqdcConfigDb->getConfigSelector();
    progDescr.type = cs.getType().toString();
    progDescr.index = cs.getIndex().toString();
    progDescr.name = progDescr.getName();

    //todo: transfer it from libs.pro
#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    emit updateProgramDescription(progDescr);
}

void TqdcCore::updateTqdcProgDescrOptions()
{
    QStringList devList;
    for(const auto &dg : devices)
        devList << dg->getDeviceIndex().toString();
    progDescr.options["DevList"] = devList.join(";");
    progDescr.options["run"] = QVariant(config.isRun()).toString();

    emit updateProgramDescription(progDescr);
}
