#include "TqdcSignalRouter.h"

#include <QObject>

#include "configuration-manager/ConfigurationManager.h"
#include "fsm/AdcDeviceMgrFsm.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/RemoteControlServer.h"
#include "TqdcDeviceGlobal.h"
#include "TQDCDevice.h"
#include "DeviceChTreeView.h"

#include "CalConfigDb.h"
#include "MainWindow.h"
#include  "CalibrationModel.h"
#include "DeviceChModel.h"
#include "SpinBoxDelegate.h"
#include "TqdcCore.h"
#include "TqdcConfigDb.h"
#include "TqdcConfigDispatcher.h"
#include "pnp-server/PNPServer.h"
#include "mldiscover/LldpDiscovery.h"
#include "mldiscover/DeviceDispatcher.h"
#include "device-discover/DiscoverDialog.h"
#include "configuration-manager/ConfigurationManager.h"
#include "TtlIoWidget/TtlIoWidget.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"

namespace {
const auto &device_filter_list = makeDeviceIdSet({DEVICE_ID_TQDC16VS_E});

template <typename Func1, typename Func2>
inline void connect(const typename QtPrivate::FunctionPointer<Func1>::Object *sender, Func1 signal,
                    const typename QtPrivate::FunctionPointer<Func2>::Object *receiver, Func2 slot)
{
    QObject::connect(sender, signal, receiver, slot);
}
}

void TqdcSignalRouter::setMainWindow(MainWindow *p)
{
    // set mainWindow must be before core
    Q_ASSERT(!mainConnectPassed);
    mainWindow = p;
}

void TqdcSignalRouter::connectAll()
{
    Q_ASSERT(pnpServer);

    connect_ConfigurationDispatcher();
    connect_deviceDispatcher();
    connect_MainWindow();
    connect_PnpServer();
    connect_RemoteControlServer();
    connect_FSM();
    mainConnectPassed = true;
}

void TqdcSignalRouter::connect_ConfigurationDispatcher()
{
    auto tqdcConfigDispatcher = &TqdcConfigDispatcher::instance();

    if(core) {
        tqdcConfigDispatcher->subscribeConfig(core);
        tqdcConfigDispatcher->subscribeCS(core);
    }
    if(mainWindow) {
        tqdcConfigDispatcher->subscribeConfig(mainWindow);
        tqdcConfigDispatcher->subscribeConfig(mainWindow->getDeviceTreeModel());
        tqdcConfigDispatcher->subscribeCalConfig(mainWindow->calibrationDialog->getCalibrationModel());
        tqdcConfigDispatcher->subscribeCS(mainWindow);
        auto widgetTtl = mainWindow->getWidgetTtl();
        if(widgetTtl) {
            tqdcConfigDispatcher->subscribeTtlCfg(widgetTtl);
        }

    }
    if(tqdcConfigDb) {
        tqdcConfigDispatcher->subscribeConfig(tqdcConfigDb);
        tqdcConfigDispatcher->subscribeCS(tqdcConfigDb);
    }
    if(configurationManager)
        tqdcConfigDispatcher->subscribeCS(configurationManager);
    if(calConfigDb){
        tqdcConfigDispatcher->subscribeCalConfig(calConfigDb);
    }
    if(adcSerDesCalDialog)
        tqdcConfigDispatcher->subscribeSerDesDelay(adcSerDesCalDialog);
}

void TqdcSignalRouter::connect_deviceDispatcher()
{
    if(!deviceDispatcher)
        return;
    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter(device_filter_list);
    lldpDiscovery.subscribe(deviceDispatcher);

    if(mainWindow){
        lldpDiscovery.subscribe(mainWindow->getDiscoverDialog());
        deviceDispatcher->subscribeDevices(mainWindow->getDiscoverDialog());
        deviceDispatcher->subscribeDevices(mainWindow->calibrationDialog->getCalibrationModel());
    }
    deviceDispatcher->subscribeDevices(&TqdcConfigDispatcher::instance());
    if(fsm){
        deviceDispatcher->subscribeDevices(fsm);
     }
    if(adcSerDesCalDialog)
        deviceDispatcher->subscribeDevices(adcSerDesCalDialog);
}

void TqdcSignalRouter::connect_MainWindow()
{
    if(!core)
        return;

    if(adcSerDesCalDialog)
        connect(adcSerDesCalDialog, &AdcSerDesCalDialog::requestCalibration,
                core, &TqdcCore::requestAdcSerDesCalibration);

    if(!mainWindow)
        return;
    LldpDiscovery::subscribe(mainWindow, device_filter_list);

    connect(core, &TqdcCore::deviceInfoUpdated,
            mainWindow, &MainWindow::deviceInfoUpdated);

    connect(core, &TqdcCore::stat_from_receiver_updated,
            mainWindow, &MainWindow::stat_from_receiver_updated);

    connect(mainWindow, &MainWindow::closeAll,
            core, &TqdcCore::deleteLater);

    connect(mainWindow, &MainWindow::performLocalStart,
            core, &TqdcCore::performLocalStart);

    connect(mainWindow, &MainWindow::performLocalStop,
            core, &TqdcCore::performLocalStop);
}

void TqdcSignalRouter::connect_PnpServer()
{
    if(!pnpServer || !core)
        return;

    connect(core, &TqdcCore::changeProgramDescription,
            pnpServer, &PNPServer::changeProgramDescription);
    connect(core, &TqdcCore::closeProgram,
            pnpServer, &PNPServer::closeProgram);
    connect(core, &TqdcCore::updateProgramDescription,
            pnpServer, &PNPServer::updateProgramDescription);
    connect(core, &TqdcCore::closeAll,
            pnpServer, &PNPServer::deleteLater);
}

void TqdcSignalRouter::connect_RemoteControlServer()
{
    if(!remoteControlServer)
        return;

    if (core) {
        connect(core, &TqdcCore::changeProgramDescription,
                remoteControlServer, &RemoteControlServer::changeProgramDescription);

        connect(remoteControlServer, &RemoteControlServer::startRequest,
                core, &TqdcCore::remoteStartRequest);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                core, &TqdcCore::remoteStopRequest);
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                core, &TqdcCore::onRemoteStatusUpdated);

        connect(core, &TqdcCore::closeAll,
                remoteControlServer, &RemoteControlServer::deleteLater);
        connect(core, &TqdcCore::sendRunError,
                remoteControlServer,
                QOverload<const RcRunId &, QString>::of(&RemoteControlServer::onRunError));

        connect(core, &TqdcCore::sendRunWarning,
                remoteControlServer, &RemoteControlServer::onRunWarning);
        connect(core, &TqdcCore::allowRemoteControl,
                remoteControlServer, &RemoteControlServer::startListen);
    }
    if(mainWindow){
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                mainWindow, &MainWindow::onRemoteStatusUpdated);

        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                mainWindow->calibrationDialog, &CalibrationDialog::onRemoteStatusUpdated);
    }
}

void TqdcSignalRouter::connect_FSM()
{
    if(!fsm)
        return;

    if(remoteControlServer){
        connect(fsm, &BaseFsm::stateChanged,
                remoteControlServer, &RemoteControlServer::onFsmStateChange);
    }
    if(mainWindow){
        connect(fsm, &BaseFsm::stateChanged,
                mainWindow, &MainWindow::onFsmStateChange);

        connect(fsm, &BaseFsm::stateChanged,
                mainWindow->calibrationDialog, &CalibrationDialog::onFsmStateChange);

        connect(mainWindow, &MainWindow::printFsmInputs,
                fsm, &BaseFsm::printFsmInput);

        auto widgetTtl = mainWindow->getWidgetTtl();
        if(widgetTtl) {
            connect(fsm, &BaseFsm::stateChanged,
                    widgetTtl, &TtlIoWidget::onFsmStateChange);
        }
    }
    if(core){
        connect(fsm, &BaseFsm::stateChanged,
                core, &TqdcCore::onFsmStateChange);
    }

    connect(&TqdcConfigDispatcher::instance(), &TqdcConfigDispatcher::configKeyChanged,
            fsm, &AdcDeviceMgrFsm::setMainSetupKey);

    if(adcSerDesCalDialog)
        connect(fsm, &BaseFsm::stateChanged,
                adcSerDesCalDialog, &AdcSerDesCalDialog::onFsmStateChange);
}

void TqdcSignalRouter::connect_TqdcDeviceGlobal(const TqdcDeviceGlobal *devGlob)
{
    if(!devGlob)
        return;

    auto dev = devGlob->getDevice();

    connect(dev, &TQDCDevice::deviceInfoUpdated,
            core, &TqdcCore::deviceInfoUpdated);

    connect(dev, &TQDCDevice::statusUpdated,
            core, &TqdcCore::deviceStatusUpdated);

    connect(dev, &TQDCDevice::statusUpdated,
            devGlob, &TqdcDeviceGlobal::deviceStatusUpdated);

    connect(core, &TqdcCore::deleteDevice,
            devGlob, &TqdcDeviceGlobal::deleteDevice);
    connect(core, &TqdcCore::deleteDevice,
            dev, &TQDCDevice::deleteHardware);

    connect(dev, &TQDCDevice::deviceDisconnected,
            core, &TqdcCore::deviceDisconnected);

    if(mainWindow){
        connect(dev, &TQDCDevice::statusUpdated,
                mainWindow, &MainWindow::deviceStatusUpdated);

        connect(mainWindow, &MainWindow::softReset,
                dev, &TQDCDevice::softReset);

        connect(dev, &TQDCDevice::deviceConnected,
                mainWindow, &MainWindow::deviceConnected);

        connect(dev, &TQDCDevice::deviceDisconnected,
                mainWindow, &MainWindow::deviceDisconnected);

        connect(mainWindow, &MainWindow::softReset,
                dev, &TQDCDevice::softReset);

        connect(devGlob, &TqdcDeviceGlobal::stat_updated,
                mainWindow, &MainWindow::mstream_stat_updated);

        connect(mainWindow, &MainWindow::mstream_reset_all,
                devGlob, &TqdcDeviceGlobal::resetMStream);

        connect(mainWindow->calibrationDialog, &CalibrationDialog::startCalibration,
                core, &TqdcCore::startCalibration_);

        auto widgetTtl = mainWindow->getWidgetTtl();
        if(widgetTtl) {
            connect(dev, &TQDCDevice::ttlIoStartupCfg,
                    widgetTtl, &TtlIoWidget::setStartUpCfg);
        }
    }

    if(adcSerDesCalDialog) {
        connect(dev, &TQDCDevice::adcDesCountUpdated,
                adcSerDesCalDialog, &AdcSerDesCalDialog::adcDesCountUpdated);
        connect(dev, &TQDCDevice::devSerDesDelayChanged,
                adcSerDesCalDialog, &AdcSerDesCalDialog::setDevSerDesDelay);
        connect(dev, &TQDCDevice::adcSerDesProgressUpdated,
                adcSerDesCalDialog, &AdcSerDesCalDialog::progressUpdated);
    }

    connect(devGlob, &TqdcDeviceGlobal::stat_from_receiver_updated,
            core, &TqdcCore::stat_from_receiver_updated);

    connect(devGlob, &TqdcDeviceGlobal::closeProgram,
            core, &TqdcCore::closeProgram);

    connect(devGlob, &TqdcDeviceGlobal::updateProgramDescription,
            core, &TqdcCore::updateProgramDescription);

    connect(core, &TqdcCore::startCalibration,
            dev, &TQDCDevice::startCalibration);
    connect(core, &TqdcCore::startAdcSrdDesCalibration,
            dev, &TQDCDevice::startAdcSrdDesCalibration);

    connect(dev, &TQDCDevice::configWritten,
            fsm, &AdcDeviceMgrFsm::setDevSetupKey);

    connect(dev, &TQDCDevice::calibrateFinished,
            fsm, &AdcDeviceMgrFsm::setDevCalKey);

    deviceDispatcher->subscribeHostAddressUpdates(devGlob->getDeviceIndex(),
                                                  dev);
    auto tqdcConfigDispatcher = &TqdcConfigDispatcher::instance();
    tqdcConfigDispatcher->subscribeDeviceConfig(dev);
    tqdcConfigDispatcher->subscribeCalResult(dev);
}
