#include "CalConfig.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QSettings>

#include "base/DeviceIndex.h"
#include "mregdev/device_id.h"
#include "CalResult.h"

namespace {
const QString BASE_CAL_PREFIX = "base_cal";


inline QString getSectionStr(const DeviceIndex &devIndex, const QString &sectionType)
{
    return QString("%1-%2-%3")
            .arg(getDeviceTypeName(devIndex.getDevId()))
            .arg(devIndex.getSerial(), 8, 16, QChar('0'))
            .arg(sectionType);
}

inline QString getKeyName(QString type, int ch)
{
    return QString("%1_%2").arg(type).arg(ch);
}

double getDouble(QSettings &s, QString key, bool *ok)
{
    if(s.contains(key)) {
        return s.value(key).toDouble(ok);
    } else {
        *ok = false;
        return 0;
    }
}
int getInt(QSettings &s, QString key, bool *ok)
{
    if(s.contains(key)) {
        return s.value(key).toInt(ok);
    } else {
        *ok = false;
        return 1920;
    }
}

QString getCalSectionStr(const DeviceIndex &devIndex)
{
    return getSectionStr(devIndex, BASE_CAL_PREFIX);
}

QString getKeyStr_X1(int ch)
{
    return getKeyName("x1", ch);
}

QString getKeyStr_X4(int ch)
{
    return getKeyName("x4", ch);
}

QString getKeyStr_ThrOffset(int ch)
{
    return getKeyName("thrOffset", ch);
}

DeviceIndex getDevIndexByGroup(QString groupStr, QString &type)
{
    DeviceIndex index;
    if(groupStr.endsWith(BASE_CAL_PREFIX)){
        type = BASE_CAL_PREFIX;
    } else {
        return index;
    }
    QString devIndexStr = groupStr.left(groupStr.length()-type.length()-1); // "TQDC16VS_E-046F0C75"
    QString devType = devIndexStr.left(devIndexStr.length()-9); // "-046F0C75"
    QString devSerial = devIndexStr.right(8); // "046F0C75"
    bool ok;
    quint32 serial = devSerial.toUInt(&ok, 16);
    if(ok) {
        quint8 id=1;
        for(; id!=0; ++id) {
            if(devType == getDeviceTypeName(id)) {
                break;
            }
        }
        if(id)
            return DeviceIndex(static_cast<DeviceId>(id), serial);
    }
    return DeviceIndex();
}

}

void CalConfig::reloadCalibration()
{
    {
        QSettings s(QSettings::IniFormat, QSettings::UserScope,
                QCoreApplication::organizationName(), "tqdc2_cal");
        reloadCalibration(s);
    }
    {
        QSettings s;
        reloadCalibration(s);
    }
}

void CalConfig::writeAll()
{
    for(auto it = begin(); it != end(); ++it)
        updateBaseCal(it.value());
}

void CalConfig::updateBaseCal(const CalResult &calRes)
{
    const auto &index = calRes.index;
    operator[](index) = calRes;

    const QMap<int, double> baseline = calRes.adcVals;
    const QMap<int, double> baseline4 = calRes.adcVals4;
    const QMap<int, int> thr = calRes.dacOffset;

    QSettings s;
    s.remove(getCalSectionStr(index));
    s.beginGroup(getCalSectionStr(index));
    s.sync();
    s.beginWriteArray("temp");
    for(auto it=calRes.temps.begin(); it!=calRes.temps.end(); ++it){
        s.setArrayIndex(it.key());
        s.setValue("v", *it);
    }
    s.endArray();
    s.setValue("data", QDateTime::currentDateTime().toString());
    int maxNCh = -1;
    for(int ch : baseline.keys()) {
        s.setValue(getKeyStr_X1(ch), baseline[ch]);
        maxNCh = qMax(maxNCh, ch);
    }
    for(int ch : baseline4.keys()) {
        s.setValue(getKeyStr_X4(ch), baseline4[ch]);
        maxNCh = qMax(maxNCh, ch);
    }
    for(int ch : thr.keys()) {
        s.setValue(getKeyStr_ThrOffset(ch), thr[ch]);
        maxNCh = qMax(maxNCh, ch);
    }
    s.setValue("nch", calRes.nch);
    s.endGroup();
}

void CalConfig::reloadCalibration(QSettings &s)
{
    bool ok;
    double valD;
    int valI;
    for(const QString &group : s.childGroups()) {
        QString groupType;
        DeviceIndex devIndex = getDevIndexByGroup(group, groupType);
        if(devIndex.isNull())
            continue;

        auto &devCal = operator[](devIndex);
        devCal.index = devIndex;
        s.beginGroup(group);

        if(group.endsWith(BASE_CAL_PREFIX)) {
            devCal.nch = s.value("nch", -1).toInt(&ok);
            if(ok) {
                for(int ch=0; ch<devCal.nch; ++ch) {
                    valD = getDouble(s, getKeyStr_X1(ch), &ok);
                    if(ok)
                        devCal.adcVals[ch] = valD;
                    valD = getDouble(s, getKeyStr_X4(ch), &ok);
                    if(ok)
                        devCal.adcVals4[ch] = valD;
                    valI = getInt(s, getKeyStr_ThrOffset(ch), &ok);
                    if(ok)
                        devCal.dacOffset[ch] = valI;
                }
            }
        }

        s.endGroup();
    }
}
