#include "DeviceGlob.h"

#include <QElapsedTimer>
#include <QThread>

#include "TtvxsAppConfig.h"
#include "TtvxsDevice.h"

#include "base/threadQuit.h"
#include "mldiscover/LldpDiscovery.h"
#include "mstream-lib/MStreamDump.h"

DeviceGlob::DeviceGlob(const DeviceIndex &index_, QObject *parent):
    QObject(parent)
    , index(index_)
    , dev(new TtvxsDevice(index))
    , dev_thread(new QThread(this))
    , mstream_thread(new QThread(this))
{
    dev->moveToThread(dev_thread);
    dev_thread->start();
    dev_thread->setObjectName(QString("Device %1").arg(index.getSerialStr()));

    connect(this, &DeviceGlob::resetTimestamp, dev, &TtvxsDevice::resetTimestamp);
    connect(this, &DeviceGlob::clearStat, dev, &TtvxsDevice::clearStat);
    connect(this, &DeviceGlob::evNumReset, dev, &TtvxsDevice::evNumReset);
    connect(this, &DeviceGlob::softClear, dev, &TtvxsDevice::softClear);
    connect(this, &DeviceGlob::restartAll, dev, &TtvxsDevice::onRestartAll);

    connect(dev, &TtvxsDevice::configWritten, this, &DeviceGlob::configWritten);
    connect(this, &DeviceGlob::configChanged_, dev, &TtvxsDevice::writeConfig);
    connect(dev, &TtvxsDevice::statusUpdated, this, &DeviceGlob::statusUpdated);
    connect(dev, &TtvxsDevice::feLinkStatusUpdated, this, &DeviceGlob::feLinkStatusUpdated);
    connect(this, &DeviceGlob::setHostAddress, dev, &TtvxsDevice::setHostAddress);
    connect(this, &DeviceGlob::getTrigEvNum, dev, &TtvxsDevice::getTrigEvNum);
    connect(dev, &TtvxsDevice::trigEvNumResponse, this, &DeviceGlob::trigEvNumResponse);
    connect(dev, &TtvxsDevice::devEvNumResponse, this, &DeviceGlob::devEvNumResponse);

    connect(this, &DeviceGlob::deleteHardware_, dev, &TtvxsDevice::deleteHardware);
    connect(dev, &QObject::destroyed, dev_thread, &QThread::quit);
}

DeviceGlob::~DeviceGlob()
{
    qInfo() << "Disconnect from:" << dev->getDeviceInfoStr();
    QElapsedTimer closeTimer;
    closeTimer.start();
    if(!threadQuiting(dev_thread, 2000, &closeTimer)) {
        qWarning() << "Can't stop device thread";
    }
    if(!threadQuiting(mstream_thread, 2000, &closeTimer)) {
        qWarning() << "Can't stop mstream thread";
    }
}

void DeviceGlob::setConfig(int config_key, const TtvxsAppConfig &app_config)
{
    TtvxsConfig ttvxs_config_new = app_config.devicesCfg[index];
    ttvxs_config = ttvxs_config_new;
    emit configChanged_(config_key, ttvxs_config);
}

void DeviceGlob::deleteHardware(DeviceIndex index)
{
    emit deleteHardware_(index);
}

void DeviceGlob::statusUpdated(const TtvxsStatus &status)
{
    if(index != status.devIndex)
        return;
    emit statusUpdated_(status, prev_status);
    bool doMStreamReset =
            (prev_status.mstreamHwBufSize != status.mstreamHwBufSize) ||
            (prev_status.mstreamMultiAck != status.mstreamMultiAck);
    prev_status = status;
    if(status.isConnected) {
        if(mstream){
            if(doMStreamReset){
                quint16 ack = status.mstreamMultiAck ? status.mstreamHwBufSize : 1;
                emit msdump_reset(index, true, status.mstreamHwBufSize, ack,
                                  QString(), false);
            }
        } else{
            addMStreamDump();
        }
    }
}

void DeviceGlob::addMStreamDump()
{
    MStreamDumpParameters dump_parameters;
    dump_parameters.deviceIndex = index;
    const auto dd = LldpDiscovery::instance().lookup(index);
    if(dd.ip_addr){
        dump_parameters.deviceAddress = QHostAddress(dd.ip_addr);
    }
    Q_ASSERT(dd.ip_addr);
    dump_parameters.bigFragId = true;
    dump_parameters.hwBufSize = prev_status.mstreamHwBufSize;
    dump_parameters.ackSizeLimit = prev_status.mstreamMultiAck ?
                                   prev_status.mstreamHwBufSize : 1;
    dump_parameters.preRunInit();

    mstream = new MStreamDump(dump_parameters, false);
    mstream_thread->setObjectName("MStream "+index.getSerialStr());
    mstream->moveToThread(mstream_thread);
    mstream_thread->start();
    LldpDiscovery::instance().subscribe(mstream);
    connect(mstream, &MStreamDump::closing,
            this, &DeviceGlob::closeProgram);
    connect(mstream, &MStreamDump::updateProgramDescription,
            this, &DeviceGlob::updateProgramDescription);
    connect(mstream, &MStreamDump::stat_updated,
            this, &DeviceGlob::stat_updated);
    connect(mstream, &MStreamDump::stat_from_receiver_updated,
            this, &DeviceGlob::stat_from_receiver_updated);
    connect(this, &DeviceGlob::msdump_reset,
            mstream, &MStreamDump::msdump_reset);
    connect(this, &DeviceGlob::deleteHardware_,
            mstream, &MStreamDump::delete_dump);
    connect(mstream, &MStreamDump::destroyed,
            mstream_thread, &QThread::quit);
}
