#include "MainWindow.h"
#include "ui_MainWindow.h"

#include "DeviceGlob.h"
#include "RcFsm.h"
#include "SnmpDialog.h"
#include "TrigDeviceMgrFsm.h"
#include "TtvxsAppConfigConverter.h"
#include "TtvxsAppConfigDispatcher.h"
#include "TtvxsConfigDb.h"
#include "TtvxsParams.h"

#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "FeLinkDialog/FeLinkStatusDialog.h"
#include "lib-common/UIStyles.h"
#include "m-stream-output/MStreamOutput.h"
#include "metric-sender/MetricSender.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PNPServer.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"

#ifdef __linux__
#include <signal.h>
#endif

#include <QDateTime>
#include <QDebug>
#include <QJsonDocument>
#include <QHostInfo>
#include <QScrollBar>
#include <QShortcut>
#include <QStyle>
#include <QThread>
#include <QTextCharFormat>
#include <QTimer>

namespace {

const int MAX_SLOT_NUMBER = 21;
const int CRATE_SLOT_ZERO = 10; // slot number for ch:0
const int CONFIG_TIMEOUT = 4000;

enum TAB_INDEX { TAB_TRIG_DEV=0, TAB_CLIENTS=1, TAB_MSTREAM=2, TAB_LOGS=3};

enum {
    ROW_NAME,ROW_STATUS,
    ROW_DEV_FW, ROW_DEV_MCU_VER, ROW_DEV_TEMP,
    ROW_DEV_STREAM, ROW_DEV_MS_ERR, ROW_WR_TIME,
    ROW_SPACER_0,
    ROW_UP_TIME, ROW_CUR_EV_NUM, ROW_LEMO_CNT, ROW_LEMO_RATE, ROW_SFP_CNT, ROW_SFP_RATE,
    ROW_SPACER_1,
    ROW_RUN_TIME, ROW_RUN_TRIG_CANDID, ROW_RUN_TRIG_ACCEPTED, ROW_RUN_TRIG_RATE,
    ROW_SPACER_2,
    ROW_CH_SLOT_FIRST,
    ROW_CH_SLOT_LAST = ROW_CH_SLOT_FIRST+(TTVXS_CH_NUM-1)+3, // 3 = 1st slot + two in middle
    ROW_CH_CNT
};
const QSet<int> SPACER_ROWS = {ROW_SPACER_0,ROW_SPACER_1,ROW_SPACER_2};

int chToSlot(int ch)
{
    bool odd = ch&1;
    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;

    int slot = (ch&1) ? (slotBase+(ch>>1)) : (slotBase-(ch>>1));
    if(slot<0 || slot>MAX_SLOT_NUMBER) return -1;

    return slot;
}

int slotToCh(int slot)
{
    if(slot>MAX_SLOT_NUMBER) return -1;
    if(slot>CRATE_SLOT_ZERO && slot<CRATE_SLOT_ZERO+3) return -1;
    bool odd = slot>CRATE_SLOT_ZERO;

    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;
    int ch = qAbs(slotBase-slot);
    ch <<=1;
    if(odd)
        ch |= 1;
    return ch;
}
QString getDevMetricPrefix(DeviceIndex i) {
    return QString("%1.").arg(i.getSerialStr());
}
}

MainWindow::MainWindow(const ConfigSelector &cs, QWidget *parent) :
    BaseMainWindow(parent)
    , ui(new Ui::MainWindow)
    , configuration_manager(new ConfigurationManager(this, cs.getType()))
    , configDb(new TtvxsConfigDb(cs, this))
    , configDispatcher(new TtvxsAppConfigDispatcher(this))
    , discoverDialog(new DiscoverDialog(this))
    , feLinkStatusDialog(new FeLinkStatusDialog(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , pnpServer(new PNPServer(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , remoteConnected(false)
    , evNumChecker(new EvNumChecker(this))
    , writeRunHistoryTimer(new QTimer(this))
    , megamaster_flg(false)
    , rc_fsm(new RcFsm(this))
    , trig_fsm(new TrigDeviceMgrFsm(this))
    , labelFsmStateRc(new QLabel(this))
    , labelFsmStateTrig(new QLabel(this))
    , labelFsmStateClient(new QLabel(this))
{
    qRegisterMetaType<TtvxsStatus>();
    qRegisterMetaType<QMap<int, FeLinkStatus> >();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<TtvxsConfig>();
    qRegisterMetaType<RUN_STATES>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<MStreamStatus>();
    qRegisterMetaType<LldpInfoCache>();
    AbstractConfigConverter<TtvxsAppConfig>::registerTypes();

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);
    labelFsmStateRc->setMinimumWidth(150);
    labelFsmStateTrig->setMinimumWidth(150);
    labelFsmStateClient->setMinimumWidth(150);
    statusBar()->addWidget(labelFsmStateRc);
    statusBar()->addWidget(labelFsmStateTrig);
    statusBar()->addWidget(labelFsmStateClient);

    setWindowTitle(configDb->getWindowTitle());

    widgets_to_block.append(ui->checkBox_external_ttl);
    widgets_to_block.append(ui->checkBox_timer);
    widgets_to_block.append(ui->checkBox_random);
    widgets_to_block.append(ui->spinBox_timer);
    widgets_to_block.append(ui->spinBox_random);
    widgets_to_block.append(ui->doubleSpinBox_dead_time);
    widgets_to_block.append(ui->checkBox_aux_cnt_for_trig);
    widgets_to_block.append(ui->groupBoxStat);
    widgets_to_block.append(ui->spinBoxStatPeriod);
    widgets_to_block.append(ui->comboBoxPreset);
    widgets_to_block.append(ui->actionAllow_Remote_Control);
    widgets_to_block.append(ui->radioButton_open);

    init_GUI_style();
    init_GUI_default_settings();
    initTableDev();
    initStatPresets();

    writeRunHistoryTimer->setSingleShot(true);
    writeRunHistoryTimer->setInterval(CONFIG_TIMEOUT);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter({DEVICE_ID_TTVXS});
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(discoverDialog);
    lldpDiscovery.subscribe(this);

    deviceDispatcher->subscribeDevices(discoverDialog);
    deviceDispatcher->subscribeDevices(configDispatcher);
    deviceDispatcher->subscribeDevices(trig_fsm);
    deviceDispatcher->subscribeDevices(feLinkStatusDialog);

    configDispatcher->subscribe(this);
    configDispatcher->subscribeCS(this);
    configDispatcher->subscribe(configDb);
    configDispatcher->subscribeCS(configDb);
    configDispatcher->subscribeCS(configuration_manager);

    connect(rc_fsm, &RcFsm::stateChanged,
            this, &MainWindow::onRcFsmStateChange);
//    connect(configDispatcher, &TtvxsAppConfigDispatcher::deviceListChanged,
//            trig_fsm, &TrigDeviceMgrFsm::setDeviceList);
    connect(trig_fsm, &TrigDeviceMgrFsm::stateChanged,
            this, &MainWindow::onTrigFsmStateChange);
    connect(configDispatcher, &TtvxsAppConfigDispatcher::configChanged,
            trig_fsm, &DeviceMgrFsm::setMainSetupKey);
    connect(ui->clientManagerWidget, &ClientManagerWidget::runStateChanged,
            this, &MainWindow::clientManagerRunStateChanged);

    connect(ui->clientManagerWidget, &ClientManagerWidget::client_config_changed,
            this, &MainWindow::client_config_receiver);

    connect(ui->clientManagerWidget, &ClientManagerWidget::runWarn,
            this, &MainWindow::onRunWarning);
    connect(ui->clientManagerWidget, &ClientManagerWidget::runError,
            this, &MainWindow::onRunError);
    connect(ui->clientManagerWidget, &ClientManagerWidget::backPressure,
            this, &MainWindow::onBackPressure);
    connect(ui->clientManagerWidget, &ClientManagerWidget::config_transfer,
            this, &MainWindow::config_transfer_from_clients);

    connect(this, &MainWindow::restartAll,
            ui->clientManagerWidget, &ClientManagerWidget::onRestartAll);
    connect(this, &MainWindow::stopAll,
            ui->clientManagerWidget, &ClientManagerWidget::onStopAll);
    connect(this, &MainWindow::sendControlEnable,
            ui->clientManagerWidget, &ClientManagerWidget::setControlEnable);

    connect(pnpServer, &PNPServer::programDiscovered,
            ui->clientManagerWidget, &ClientManagerWidget::programDiscovered);
    connect(pnpServer, &PNPServer::programClosed,
            ui->clientManagerWidget, &ClientManagerWidget::programClosed);
    connect(this, &MainWindow::changeProgramDescription,
            pnpServer, &PNPServer::changeProgramDescription);
    connect(this, &MainWindow::closeProgram,
            pnpServer, &PNPServer::closeProgram);
    connect(this, &MainWindow::deleteAll,
            pnpServer, &PNPServer::deleteLater);

    QList<DaqModule> searchTargs;
    searchTargs
        << DaqModule(DaqModule::Type::Adc64System)
        << DaqModule(DaqModule::Type::EventBuilder)
        << DaqModule(DaqModule::Type::HrbSystem)
        << DaqModule(DaqModule::Type::RunControl)
        << DaqModule(DaqModule::Type::VmeDaq)
        << DaqModule(DaqModule::Type::Land)
        << DaqModule(DaqModule::Type::Generator)
        << DaqModule(DaqModule::Type::TdcControl)
        << DaqModule(DaqModule::Type::Ttvxs);
    pnpServer->startSearch(searchTargs);

    connect(evNumChecker, &EvNumChecker::evNumRequest,
            ui->clientManagerWidget, &ClientManagerWidget::evNumRequest);
    connect(evNumChecker, &EvNumChecker::checkFail, this, &MainWindow::onEvNumCheckFail);
    connect(ui->clientManagerWidget, &ClientManagerWidget::evNumResponse,
            evNumChecker, &EvNumChecker::gotClientEvNum);
    connect(ui->clientManagerWidget, &ClientManagerWidget::clientMapUpdated,
            evNumChecker, &EvNumChecker::updateClientMap);

    connect(this, &MainWindow::deleteAll, remoteControl,
            &RemoteControlServer::deleteLater);
    connect(remoteControl, &RemoteControlServer::startRequest,
            this, &MainWindow::onRemoteStartRequest);
    connect(remoteControl, &RemoteControlServer::stopRequest,
            this, &MainWindow::onRemoteStopRequest);
    connect(this, &MainWindow::sendRemoteState,
            remoteControl, &RemoteControlServer::sendState);
    connect(rc_fsm, &RcFsm::stateChanged,
            remoteControl, &RemoteControlServer::onFsmStateChange);
    connect(this, &MainWindow::sendRunError,
            remoteControl, &RemoteControlServer::onRunError);
    connect(this, &MainWindow::sendRunWarn,
            remoteControl, &RemoteControlServer::onRunWarning);
    connect(this, &MainWindow::sendBackPressure,
            remoteControl, &RemoteControlServer::onBackPressure);

    connect(remoteControl, &RemoteControlServer::connectionStatusUpdated,
            this, &MainWindow::onRemoteStatusUpdated);
    connect(this, &MainWindow::allowRemoteControl,
            remoteControl, &RemoteControlServer::startListen);
    connect(this, &MainWindow::changeProgramDescription,
            remoteControl, &RemoteControlServer::changeProgramDescription);

    connect(remoteControl, &RemoteControlServer::evNumCheckRequest,
            evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);
    connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
            remoteControl, &RemoteControlServer::evNumCheckResponse);

    connect(this, &MainWindow::config_transfer_to_rc,
            remoteControl, &RemoteControlServer::config_transfer);

    new ThreadController(remoteControl, this);

    updateGui();

    configuration_manager->setDefaultConfig(TtvxsAppConfig());
}

MainWindow::~MainWindow()
{
    disconnect(remoteControl, &RemoteControlServer::startRequest,
               nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest,
               nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stateRequest,
               nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated,
               nullptr, nullptr);

    emit deleteAll();
    disconnectFromDev(DeviceIndex());

    delete ui->clientManagerWidget;
    delete ui;
}

//=== Public slots:
void MainWindow::setConfig(int config_key, const TtvxsAppConfig &config_)
{
    Q_UNUSED(config_key);
    if(this->config == config_)
        return;
    config = config_;
    ui->clientManagerWidget->init(config.clients);
    emit allowRemoteControl(config.allowRemoteControl);
    updateGui();

    //remove from devices
    QSet<DeviceIndex> toDel;
    const auto &devicesCfg = config.devicesCfg;
    for(DeviceIndex index : devices.keys()){
        if(!devicesCfg[index].enabled ||
           !devicesCfg.contains(index))
            disconnectFromDev(index);
    }
    //add devices
    for(DeviceIndex index : config.devicesCfg.keys()){
        if(devicesCfg[index].enabled){
            addDevice(index);
            configDispatcher->sendCache();
        }
    }

}

void MainWindow::configSelectorChanged(const ConfigIndexName &conf)
{
    const auto &cs = ConfigSelector(DaqModule(DaqModule::Type::Ttvxs), conf);
    setWindowTitle(cs.toWindowTitle());

    localProgramIndex = conf.getIndex();

    // FIXME: make method of ProgramDescription
    progDesc.type = cs.getType().toString();
    progDesc.index = cs.getIndex().toString();
    progDesc.name = progDesc.getName();

    // Initialize MetricSender
    const QString shortHostName = QHostInfo::localHostName().section(".", 0, 0);
    MetricSender::setNamespace("daq.ttvxs-rc." + shortHostName + "."
                               + conf.getIndex().toString());
}

void MainWindow::lldpCacheUpdated(const LldpInfoCache &lldpCache)
{
    const auto lldp_keys = lldpCache.keys();
    const auto lldp_indices = makeDeviceIndexSet(lldp_keys);
    const auto my_keys = config.devicesCfg.keys();
    const auto my_indices = makeDeviceIndexSet(my_keys);
    const auto indices = lldp_indices & my_indices;

    for (const auto &i : indices)
        updateDeviceLldpInfo(i, lldpCache[i]);
}

void MainWindow::onRcFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    logRcFsmStateChange(oldState, state);
    labelFsmStateRc->setText("RcFsm: " + fsmStateToString(state));

    switch (state) {
    case FsmState::CloseGlobTrig:
        trig_fsm->goState(trig_fsm->getState()==FsmState::Ready ?
                          FsmState::Starting : FsmState::Error);
        rc_fsm->goState(FsmState::Starting);
        break;
    case FsmState::Starting:
        emit restartAll(config.runId);
        rc_fsm->goState(megamaster_flg ? FsmState::OpenGlobTrig : FsmState::Run);
        if(!megamaster_flg)
            trig_fsm->goState(FsmState::Starting);
        break;

    case FsmState::OpenGlobTrig:
        trig_fsm->goState(FsmState::OpenGlobTrig);
        rc_fsm->goState(FsmState::Run);
        break;

    case FsmState::Run:
        actionOnStateRun();
        break;
    case FsmState::PrepareStopping:
        Q_ASSERT(megamaster_flg);
        trig_fsm->goState(FsmState::Stopping);
        rc_fsm->goState(FsmState::Stopping);
        break;
    case FsmState::Stopping:
        actionOnStateStop();
        trig_fsm->goState(FsmState::Stopping);
        emit stopAll();
        break;
    case FsmState::Error:
        if(megamaster_flg && config.runEnable){
            setTrigState(false);
            config.runEnable = false;
            emit configChanged(config);
            evNumChecker->stopCheck();
            emit stopAll();
        }
        break;
    default:
    {}
    }
}

void MainWindow::onTrigFsmStateChange(const FsmState &state)
{
    setTabState(state!=FsmState::Error, TAB_TRIG_DEV);
    labelFsmStateTrig->setText("TrigFsm: " + fsmStateToString(state));
    rc_fsm->setTrigDeviceMgrState(state);

    switch (state) {
    case FsmState::Error:
    case FsmState::Ready:
        megamaster_flg = false;
        break;
    case FsmState::Starting:
        if(megamaster_flg)
            setTrigState(false);
        else
            setTrigState(true);
        trig_fsm->goState(megamaster_flg ? FsmState::CloseGlobTrig : FsmState::Run);
        break;
    case FsmState::CloseGlobTrig:
        break;
    case FsmState::OpenGlobTrig:
        setTrigState(true);
        trig_fsm->goState(FsmState::Run);
        break;
    case FsmState::Stopping:
        setTrigState(false);
        trig_fsm->goState(FsmState::Ready);
        break;
    default:
    {}
    }
}

void MainWindow::clientManagerRunStateChanged(const FsmState &state)
{
    setTabState(state!=FsmState::Error, TAB_CLIENTS);
    labelFsmStateClient->setText("ClientFsm: " + fsmStateToString(state));
    rc_fsm->setClientManagerRunState(state);
}


// Private methods
void MainWindow::init_GUI_style()
{
    QString styleLargeFont = "font-size: 18pt";
    ui->pushButton_start->setStyleSheet(styleLargeFont);
    ui->pushButton_stop->setStyleSheet(styleLargeFont);
    ui->labelRunState->setStyleSheet(styleLargeFont);
    ui->pushButton_start->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButton_stop->setIcon(style()->standardIcon(QStyle::SP_MediaStop));

    ui->groupBox_trigger_source->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBox_trigger_state->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxStat->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->plainTextEdit_log->setReadOnly(true);

    ui->spinBox_timer->setMaximum(10000000);
    ui->spinBox_random->setMaximum(1000000);
}

void MainWindow::init_GUI_default_settings()
{
    ui->pushButton_stop->setEnabled(false);

    labelWarnTtvxs = new QLabel();
    labelWarnClients = new QLabel();
    labelWarnOther = new QLabel();

    ui->statusBar->addWidget(labelWarnTtvxs);
    ui->statusBar->addWidget(labelWarnClients);
    ui->statusBar->addWidget(labelWarnOther);

    labelWarnTtvxs->hide();
    labelWarnClients->hide();
    labelWarnOther->hide();
}

void MainWindow::initTableDev()
{
    auto t = ui->tableWidgetDev;
    t->setRowCount(ROW_CH_CNT);

    t->setVerticalHeaderItem(ROW_NAME, new QTableWidgetItem("Name"));
    t->setVerticalHeaderItem(ROW_STATUS, new QTableWidgetItem("Status"));

    t->setVerticalHeaderItem(ROW_DEV_FW, new QTableWidgetItem("FPGA f/w"));
    t->setVerticalHeaderItem(ROW_DEV_MCU_VER, new QTableWidgetItem("MCU f/w"));
    t->setVerticalHeaderItem(ROW_DEV_TEMP, new QTableWidgetItem("Temp"));
    t->setVerticalHeaderItem(ROW_DEV_STREAM, new QTableWidgetItem("Data Stream, kB/s"));
    t->setVerticalHeaderItem(ROW_DEV_MS_ERR, new QTableWidgetItem("MStream err"));
    t->setVerticalHeaderItem(ROW_WR_TIME, new QTableWidgetItem("WR time"));

    t->setVerticalHeaderItem(ROW_UP_TIME, new QTableWidgetItem("Up time"));
    t->setVerticalHeaderItem(ROW_CUR_EV_NUM, new QTableWidgetItem("Cur evNum"));
    t->setVerticalHeaderItem(ROW_LEMO_CNT, new QTableWidgetItem("TTL pulses"));
    t->setVerticalHeaderItem(ROW_LEMO_RATE, new QTableWidgetItem("TTL rate"));
    t->setVerticalHeaderItem(ROW_SFP_CNT, new QTableWidgetItem("SFP pulses"));
    t->setVerticalHeaderItem(ROW_SFP_RATE, new QTableWidgetItem("SFP rate"));

    t->setVerticalHeaderItem(ROW_RUN_TIME, new QTableWidgetItem("Run time"));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_CANDID, new QTableWidgetItem("Trig run cand."));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_ACCEPTED, new QTableWidgetItem("Trig run accepted"));
    t->setVerticalHeaderItem(ROW_RUN_TRIG_RATE, new QTableWidgetItem("Run trig rate"));

    for(int row : SPACER_ROWS) {
        t->setVerticalHeaderItem(row, new QTableWidgetItem(" "));
    }

    for(int row=ROW_CH_SLOT_FIRST; row<=ROW_CH_SLOT_LAST; ++row) {
        auto item = new QTableWidgetItem(QString("Slot %2").arg(row-ROW_CH_SLOT_FIRST+1, 2, 10, QChar('0')));
        if(isSlotBlocked(row)) {
            item->setForeground(Qt::darkCyan);
        }
        t->setVerticalHeaderItem(row, item);
    }

    t->resizeRowsToContents();
    t->setWordWrap(false);
    t->setStyleSheet(":disabled {background-color:lightGray}");

    int spacer_height = 8;
    auto header = t->verticalHeader();
    header->setSectionResizeMode(QHeaderView::Fixed);
    header->setMinimumSectionSize(spacer_height);

    for(int row : SPACER_ROWS) {
        header->resizeSection(row, spacer_height);
        auto spItem = t->verticalHeaderItem(row);
        spItem->setBackground(Qt::lightGray);
    }
}

void MainWindow::initStatPresets()
{
    ui->comboBoxPreset->blockSignals(true);
    for(auto preset : StatPresetFabric::getAllPresets()){
        ui->comboBoxPreset->addItem(preset.label, preset.type);
    }
    ui->comboBoxPreset->blockSignals(false);
}

void MainWindow::addDevice(const DeviceIndex &index)
{
    Q_ASSERT(!index.isNull());
    if(devices.contains(index)) return;

    DeviceGlob * glob = new DeviceGlob(index);
    devices.insert(index, glob);
    connect(this, &MainWindow::deleteAll, glob, &DeviceGlob::deleteLater);
    connect(this, &MainWindow::deleteHardware_, glob, &DeviceGlob::deleteHardware);
    connect(this, &MainWindow::resetTimestamp, glob, &DeviceGlob::resetTimestamp);
    connect(this, &MainWindow::clearStat, glob, &DeviceGlob::clearStat);
    connect(this, &MainWindow::evNumReset, glob, &DeviceGlob::evNumReset);
    connect(this, &MainWindow::softClear, glob, &DeviceGlob::softClear);
    connect(this, &MainWindow::restartAll, glob, &DeviceGlob::restartAll);
    connect(glob, &DeviceGlob::statusUpdated_, this, &MainWindow::statusUpdated);
    connect(glob, &DeviceGlob::feLinkStatusUpdated, feLinkStatusDialog, &FeLinkStatusDialog::setFeLinkStatus);
    connect(glob, &DeviceGlob::configWritten, trig_fsm, &DeviceMgrFsm::setDevSetupKey);

    connect(evNumChecker, &EvNumChecker::getTrigEvNum,
            glob, &DeviceGlob::getTrigEvNum);
    connect(glob, &DeviceGlob::trigEvNumResponse,
            evNumChecker, &EvNumChecker::gotTopLevelEvNum);
    connect(glob, &DeviceGlob::devEvNumResponse,
            evNumChecker, &EvNumChecker::gotDevEvNum);

    ui->statistic_widget->add_device(index);
    connect(glob, &DeviceGlob::stat_updated,
            ui->statistic_widget, &StatisticOutput::stat_updated);
    connect(glob, &DeviceGlob::stat_from_receiver_updated,
            ui->statistic_widget, &StatisticOutput::stat_from_receiver_updated);
    connect(this, &MainWindow::set_run_state,
            ui->statistic_widget, &StatisticOutput::set_run_state);

    connect(glob, &DeviceGlob::closeProgram,
            pnpServer, &PNPServer::closeProgram);
    connect(glob, &DeviceGlob::updateProgramDescription,
            pnpServer, &PNPServer::updateProgramDescription);

    deviceDispatcher->subscribeHostAddressUpdates(index, devices[index]);
    configDispatcher->subscribe(glob);

    int colDev = -1;
    auto tableDev = ui->tableWidgetDev;

    if(colDev == -1){
        colDev = tableDev->columnCount();
        tableDev->setColumnCount(colDev+1);
        for(int row=0; row<tableDev->rowCount(); ++row) {
            auto item = new QTableWidgetItem();
            tableDev->setItem(row, colDev, item);
            if(row == ROW_NAME) {
                item->setFlags(item->flags() | Qt::ItemIsEditable);
                item->setText(config.devicesCfg[index].custom_name);
            } else {
                item->setFlags(item->flags() & ~Qt::ItemIsEditable);
            }
            if(SPACER_ROWS.contains(row)) {
                item->setBackground(Qt::lightGray);
                continue;
            }
            if(row < ROW_CH_SLOT_FIRST || row > ROW_CH_SLOT_LAST) continue;
            if(isSlotBlocked(row)){
                item->setBackground(Qt::darkCyan);
            } else {
                item->setCheckState(Qt::Unchecked);
            }
        }
    }

    const quint32 &vxsChMask = config.devicesCfg[index].vxsChMask;
    Q_ASSERT(colDev != -1);
    for(int row=ROW_CH_SLOT_FIRST; row<=ROW_CH_SLOT_LAST; ++row) {
        const int slot = row-ROW_CH_SLOT_FIRST+1;
        const int ch = slotToCh(slot);
        if(!isSlotBlocked(row)){
            tableDev->item(row, colDev)->setCheckState(vxsChMask&(1<<ch)?Qt::Checked:Qt::Unchecked);
        }
    }

    auto header_item = new QTableWidgetItem(index.getSerialStr());
    header_item->setData(Qt::UserRole, QList<QVariant>() << index.getDevId() << index.getSerial());
    header_item->setForeground(Qt::red);
    tableDev->setHorizontalHeaderItem(colDev, header_item);
    qInfo() << QString("%1 was added").arg(index.getIdent());
}

void MainWindow::disconnectFromDev(const DeviceIndex &index)
{
    if(!index.isNull() && !devices.contains(index)) return;
    emit deleteHardware_(index);
    devices.remove(index);
    int col = getChCfgCol(index);
    if(col!=-1) {
        ui->tableWidgetDev->removeColumn(col);
    }
    ui->statistic_widget->remove_device(index);

    if(!index.isNull()) return;
}

void MainWindow::sendLog(QString log)
{
    qInfo() << log;
}

void MainWindow::updateGui()
{
    block_widgets_signal(true);
    const auto &cfg = config.default_cfg;

    ui->checkBox_external_ttl->setChecked(cfg.trigSrc.bit.ttl);
    ui->checkBox_timer->setChecked(cfg.trigSrc.bit.fixed_pulser);
    ui->checkBox_random->setChecked(cfg.trigSrc.bit.random_pulser);
    ui->spinBox_timer->setValue(cfg.trigTimerFreq);
    ui->spinBox_random->setValue(cfg.trigRandomFreq);
    ui->doubleSpinBox_dead_time->setValue(cfg.trigRandomDeadTimeUs);

    ui->checkBox_aux_cnt_for_trig->setChecked(cfg.aux_cnt_for_trig);

    ui->groupBoxStat->setChecked(cfg.stat_readout);
    ui->spinBoxStatPeriod->setValue(cfg.stat_period);
    ui->comboBoxPreset->setCurrentText(cfg.stat_preset.label);

    ui->actionAllow_Remote_Control->setChecked(config.allowRemoteControl);
    ui->radioButton_open->setChecked(cfg.trig_open);
     block_widgets_signal(false);
}


int MainWindow::getChCfgCol(const DeviceIndex &i) const
{
    auto t = ui->tableWidgetDev;
    if(i.isNull()) return -1;
    for(int col=0; col<t->columnCount(); ++col) {
        if(getDevByTableItem(t->horizontalHeaderItem(col)) == i)
            return col;
    }
    return -1;
}

DeviceIndex MainWindow::getDevByTableItem(const QTableWidgetItem *item) const
{
    if(!item) return DeviceIndex();

    bool ok1,ok2;
    QList<QVariant> list = item->data(Qt::UserRole).toList();
    if(list.size()>=2) {
        auto id = static_cast<DeviceId>(list.first().toUInt(&ok1));
        quint64 ser = list.at(1).toULongLong(&ok2);
        if(ok1 && ok2)
            return DeviceIndex(id, ser);
    }
    return DeviceIndex();
}

DeviceIndex MainWindow::getDevByCol(int col) const
{
    auto item = ui->tableWidgetDev->horizontalHeaderItem(col);
    return getDevByTableItem(item);
}

bool MainWindow::isSlotBlocked(int row)
{
    static QSet<int> BLOCKED_ROWS{ROW_CH_SLOT_FIRST,
                ROW_CH_SLOT_FIRST+CRATE_SLOT_ZERO,
                ROW_CH_SLOT_FIRST+CRATE_SLOT_ZERO+1};
    return BLOCKED_ROWS.contains(row);
}

bool MainWindow::perfomeStart(const RcRunId &runId)
{
    qInfo() << __PRETTY_FUNCTION__ ;
    if(rc_fsm->getState() != FsmState::Ready) {
        qWarning() << __PRETTY_FUNCTION__ << "failed";
        return false;
    }

    config.runId = runId;
    config.runEnable = true;
    emit configChanged(config);
    rc_fsm->goState(megamaster_flg ? FsmState::CloseGlobTrig : FsmState::Starting);

    return true;
}

void MainWindow::openGlobTrig()
{
    if(!megamaster_flg)
        return;

    setTrigState(true);
    trig_fsm->goState(trig_fsm->getState()==FsmState::CloseGlobTrig ?
                          FsmState::OpenGlobTrig : FsmState::Error);
}

void MainWindow::closeGlobTrig()
{
    if(!megamaster_flg)
        return;
    setTrigState(false);
    trig_fsm->goState(trig_fsm->getState()==FsmState::Ready ?
                          FsmState::Starting : FsmState::Error);
}

void MainWindow::setTrigState(bool open)
{
    auto configNew = config;
    configNew.default_cfg.trig_open = open;
    configNew.unify_config();
    emit configChanged(configNew);
}

void MainWindow::actionOnStateRun()
{
    blockControl(true);
    ui->labelRunState->setText(QString("Starting Run %1...")
        .arg(config.runId.toString()));
    ui->labelRunState->setStyleSheet("QLabel { color : orange;}");
    ui->plainTextEdit_log->appendPlainText("");
    qInfo() << QString("Preparing Run %1... ")
        .arg(config.runId.toString());


    QCoreApplication::processEvents(QEventLoop::AllEvents, 200);
    QString txt = QString("Run started: %1")
        .arg(config.runId.toString());

    ui->labelRunState->setText(txt);
    qInfo() << "Run started !";
    ui->labelRunState->setStyleSheet("QLabel { color : green;}");

    ui->pushButton_stop->setEnabled(true);

    QVector<quint64> devSerials;
    if(!megamaster_flg) {
        for(auto cfg : config.devicesCfg) {
            if(cfg.enabled)
                devSerials << cfg.device_index.getSerial();
        }
    }
    evNumChecker->set_dev_serials(devSerials);
    if(megamaster_flg)
        evNumChecker->startCheck();


    updateStatStopButton();
    blockControl(false);
    return;
}

void MainWindow::actionOnStateStop()
{
    blockControl(true);

    ui->pushButton_stop->setEnabled(false);
    ui->labelRunState->setStyleSheet("QLabel { color : orange;}");
    ui->labelRunState->setText(QString("Stopping Run %1...")
        .arg(config.runId.toString()));
    QString msg = QString("!!! Stopping Run %1... !!! ")
        .arg(config.runId.toString());
        update();

    config.runEnable = false;
    emit configChanged(config);

    evNumChecker->stopCheck();
    // TODO: Smart wait that all clients finished it's work
    if(remoteConnected) {
        QCoreApplication::processEvents(QEventLoop::AllEvents, 1000);
    } else {
        QElapsedTimer t;
        t.start();
        int elapsedTime;
        while((elapsedTime=t.elapsed())<1000){
            QCoreApplication::processEvents(QEventLoop::AllEvents, 1000-elapsedTime);
            QThread::msleep(100);
        }
    }
    emit stopAll();

    ui->labelRunState->setText(QString("Run %1 %2")
                               .arg(config.runId.toString())
                               .arg("stopped"));
    qWarning() << "!!! Run stoped !!!";
    ui->labelRunState->setStyleSheet( "QLabel { color : green;}");

    updateStatStopButton();
    emit set_run_state(false);

    blockControl(false);
    return;
}

void MainWindow::perfomeStop(bool isError)
{
    qInfo() << __PRETTY_FUNCTION__ << "isError:" << isError << " state:" << rc_fsm->getState();
    isError = rc_fsm->getState() != FsmState::Run;
    rc_fsm->goState(isError ? FsmState::Error
                            : (megamaster_flg ? FsmState::PrepareStopping
                                              : FsmState::Stopping));
    return;
}

bool MainWindow::isReadForRun() const
{
    bool allOk=true;
    for(bool ok: tabOk.values())
        allOk &= ok;
    return allOk;
}

void MainWindow::updateProgramDescription()
{
    QString remoteState;

    progDesc.options.clear();
    progDesc.options["runEnable"] = QVariant(config.runEnable).toString();
    progDesc.options["run_index"] = config.runId.runIndex;
    progDesc.options["run_number"] = QVariant(config.runId.runNumber).toString();

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    bool readyForRun = isReadForRun();
    for(const ClientConfig &c: config.clients.values()) {
        remoteState += QString("%1 #%2; ").arg(c.daq_module.toString(), c.program_index);
        progDesc.options[QString("Client-%1").arg(c.id)] = c.daq_module.toString() +" "+ c.program_index +" , "+ QVariant(c.readout).toString();
    }
    emit changeProgramDescription(progDesc);
    emit sendRemoteState(remoteState);
}

void MainWindow::blockControl(bool block)
{
    const bool &isRun = config.runEnable;
    const bool enabled_main = !block && (!isRun || Globals::getDebugMode());
    const bool enabled_menu_and_trig = !block &&
        (!isRun || Globals::getDebugMode() || megamaster_flg);
    emit sendControlEnable(enabled_main);
    ui->menuOptions->setEnabled(enabled_menu_and_trig);

    auto table = ui->tableWidgetDev;
    for(int col=0; col<table->columnCount(); ++col){
        for(int row=ROW_CH_SLOT_FIRST; row<=ROW_CH_SLOT_LAST; ++row) {
            if(isSlotBlocked(row))
                continue;
            auto i = table->item(row, col);
            if(enabled_main) {
                i->setFlags(i->flags() | Qt::ItemIsUserCheckable | Qt::ItemIsEnabled);
            } else {
                i->setFlags(i->flags() & (~Qt::ItemIsUserCheckable) & (~Qt::ItemIsEnabled));
            }
        }
    }

    ui->checkBox_aux_cnt_for_trig->setEnabled(enabled_main);
    ui->groupBoxStat->setEnabled(enabled_main);
    ui->groupBox_trigger_source->setEnabled(enabled_main);
    ui->groupBox_trigger_state->setEnabled(enabled_menu_and_trig);
}

void MainWindow::updateStatStopButton()
{
    ui->pushButton_start->setEnabled(!config.runEnable && !remoteConnected);
    ui->pushButton_stop->setEnabled(config.runEnable && !remoteConnected);
}

void MainWindow::updateGroupBoxWarn()
{
    QStringList warnStrList;
    bool hidden = true;
    if(!labelWarnTtvxs->isHidden()) {
        hidden = false;
        warnStrList << labelWarnTtvxs->text();
    }
    if(!labelWarnClients->isHidden()) {
        hidden = false;
        warnStrList << labelWarnClients->text();
    }
    if(!labelWarnOther->isHidden()) {
        hidden = false;
        warnStrList << labelWarnOther->text();
    }

    if(!hidden)
        emit sendRunWarn(config.runId, warnStrList.join("; "));
}

void MainWindow::logRcFsmStateChange(const FsmState &previousState,
                                     const FsmState &state)
{
    auto RunIdStr = QString(config.runId.toString());
    auto previousStateStr = fsmStateToString(previousState);

    if (state == FsmState::Starting) {
        qInfo() << QString("!!! Preparing Run %1... !!!").arg(RunIdStr);
    }

    if (state == FsmState::Stopping) {
        qInfo() << QString("!!! Stopping Run %1... !!!").arg(RunIdStr);
    }

    if (state == FsmState::Run) {
        qInfo() << "!!! Run started !!!";
    }
    if (state == FsmState::Ready) {
        if (previousState == FsmState::Stopping) {
            qInfo() << "!!! Run stopped !!!";
        }
    }
    if (state == FsmState::Error) {
        switch (previousState) {
        case FsmState::Starting:;
        case FsmState::Stopping:;
        case FsmState::Run: {
            qCritical() << QString("!!! %1 failed !!!").arg(previousStateStr);
            break;
        }
        default:;
        }
    }
}

void MainWindow::block_widgets_signal(bool block)
{
    for(int i=0; i<widgets_to_block.size(); ++i)
        widgets_to_block[i]->blockSignals(block);
}

// Private slots
void MainWindow::statusUpdated(const TtvxsStatus &st, const TtvxsStatus &prev_st)
{
    const auto &index = st.devIndex;
    if(!devices.contains(index))
        return;

    auto tableDev = ui->tableWidgetDev;
    const int colCh = getChCfgCol(index);

    Q_ASSERT(colCh!=-1);
    if(colCh == -1) return;

    const auto &aux_cnt = st.aux_cnt;
    const auto &devCfg = config.devicesCfg[index];
    const auto devMetricPrefix = getDevMetricPrefix(index);
    auto vxsChMask = devCfg.vxsChMask;
    for(int ch=0; ch<TTVXS_CH_NUM; ++ch) {
        bool inputSt = st.inputState & (1<<ch);
        bool timeoutSt = st.timeoutState & (1<<ch);
        const int slot = chToSlot(ch);
        if(!prev_st.devIndex.isNull()) {
            if(timeoutSt != ((prev_st.timeoutState & (1<<ch)) != 0)) {
                QString baseStr = QString("[%3] %1 %4 timeout on slot=%2")
                        .arg(index.getIdent()).arg(slot)
                        .arg(QDateTime::currentDateTime().toString());
                if(timeoutSt) {
                    qWarning() << baseStr.arg("got");
                } else {
                    qInfo() << baseStr.arg("remove");
                }
            }
        }
        const int row = ROW_CH_SLOT_FIRST+slot-1;
        auto chItem = tableDev->item(row, colCh);
        QBrush brush_green;
        brush_green.setColor(Qt::green);
        QBrush brush_transparent;
        brush_transparent.setColor(Qt::transparent);
        chItem->setBackground(inputSt ? brush_green : brush_transparent);
        if(timeoutSt != (chItem->foreground() == Qt::red)) {
            chItem->setForeground(timeoutSt ? Qt::red : QColor());
            chItem->setText(timeoutSt ? "timeout" : "");
        }
        bool ch_en = vxsChMask&(1<<ch);

        tableDev->item(row, colCh)->setText(ch_en ? aux_cnt.getBusyTime(ch) : "");
        if(ch_en)
            MetricSender::send(QString(devMetricPrefix+"time.busy.slot.%1").arg(slot), aux_cnt.getBusyTimeUs(ch));
    }
    tableDev->item(ROW_UP_TIME, colCh)->setText(aux_cnt.getUpTime());
    MetricSender::send(devMetricPrefix+"time.uptime", aux_cnt.getUpTimeSec());
    tableDev->item(ROW_LEMO_CNT, colCh)->setText(aux_cnt.getTtlCnt());
    tableDev->item(ROW_SFP_CNT, colCh)->setText(aux_cnt.getSfpCnt());
    tableDev->item(ROW_CUR_EV_NUM, colCh)->setText(QString("%L1").arg(st.curEvNum));
    MetricSender::send(devMetricPrefix+"curEvNum", st.curEvNum);

    if(devCfg.slot){
        tableDev->item(ROW_CH_SLOT_FIRST-1+devCfg.slot, colCh)->setText(QString("ttvxs %1").arg(aux_cnt.getResultingBusyTime()));
        MetricSender::send(devMetricPrefix+"time.busy.total", aux_cnt.getResultingBusyTimeUs());
    }
    tableDev->item(ROW_RUN_TIME, colCh)->setText(aux_cnt.getRunTime());
    MetricSender::send(devMetricPrefix+"time.runtime", aux_cnt.getRunTimeSec());
    tableDev->item(ROW_RUN_TRIG_CANDID, colCh)->setText(aux_cnt.getTrigRunCandCnt());
    tableDev->item(ROW_RUN_TRIG_ACCEPTED, colCh)->setText(aux_cnt.getTrigRunAcceptCnt());

    tableDev->item(ROW_DEV_FW, colCh)->setText(st.fwStr);
    tableDev->item(ROW_DEV_TEMP, colCh)->setText(QString("%1 %2C").arg(st.temp, 0, 'f', 1).arg(QChar(0x2da)));

    auto wr_item = tableDev->item(ROW_WR_TIME, colCh);
    QDateTime wrTime = QDateTime::fromMSecsSinceEpoch(st.wrEpStatus.mSecsSinceEpoch());
    bool wrTimeValid = st.wrEpStatus.timeValid();
    wr_item->setText(st.wrEpStatus.timeValid() ? wrTime.toUTC().toString("yyyy-MM-dd hh:mm:ss") : "invalid"); // add ".zzz" for ms
    wr_item->setBackground(wrTimeValid ? Qt::green : Qt::lightGray);

    tableDev->horizontalHeaderItem(colCh)->setForeground(st.isConnected ? QColor() : Qt::gray);
    tableDev->item(ROW_STATUS, colCh)->setText(st.isConnected ? "Connected" : "Disconnected");
    tableDev->item(ROW_STATUS, colCh)->setForeground(st.isConnected ? Qt::darkGreen : Qt::red);
    auto mcu_ver_item = tableDev->item(ROW_DEV_MCU_VER, colCh);
    mcu_ver_item->setText(st.bmc.getVerStr());
    mcu_ver_item->setToolTip(QString("overall - %1\npower - %2\nPLL- %3")
                             .arg(st.bmc.get_overall_status_str())
                             .arg(st.bmc.get_power_status_str())
                             .arg(st.bmc.get_pll_status_str()));
//    QBrush mcu_ver_brush = mcu_ver_item->background();
//    mcu_ver_brush.setColor(QColor(255,0,0));
//    mcu_ver_item->setBackground(mcu_ver_brush);
    mcu_ver_item->setBackground(st.bmc.get_overall_status_col());

//    lastDevStatus = st;
    const auto &aux1 = st.aux_cnt;
    const auto &aux2 = prev_st.aux_cnt;
    tableDev->item(ROW_RUN_TRIG_RATE, colCh)->setText(aux1.getRunTrigRate(aux2));
    tableDev->item(ROW_LEMO_RATE, colCh)->setText(aux1.getTtlRate(aux2));
    tableDev->item(ROW_SFP_RATE, colCh)->setText(aux1.getSfpRate(aux2));

    if(!st.isConnected && config.runEnable) {
        onRunError(config.runId,
                   QString("TTVXS #%1 s/n:%2 offline")
                   .arg(configDb->getIndex().toString())
                   .arg(st.devIndex.getSerialStr()));
    }
}

void MainWindow::updateDeviceLldpInfo(const DeviceIndex &index, const LldpInfo &lldp)
{
    const int col = getChCfgCol(index);
    if(col == -1) return;
    auto ttvxsItem = ui->tableWidgetDev->item(ROW_CH_SLOT_FIRST+lldp.chassis_slot-1, col);
    if(!ttvxsItem)
        return;
//    Q_ASSERT(ttvxsItem);
    if(ttvxsItem->text() != "ttvxs") {
        ttvxsItem->setText("ttvxs");
        ui->tableWidgetDev->resizeColumnsToContents();
    }
}

void MainWindow::mstream_status_updated(QString serial, const MStreamStatus &status)
{
    int devCol = get_col_by_serial(serial);
    ui->tableWidgetDev->item(ROW_DEV_STREAM, devCol)->setText(status.recv_speed);
    ui->tableWidgetDev->item(ROW_DEV_MS_ERR, devCol)->setText(status.bad_format);
    bool hide_row = true;
    for(int col=0; col<ui->tableWidgetDev->columnCount(); ++col) {
        if(!ui->tableWidgetDev->item(ROW_DEV_MS_ERR, col)->text().isEmpty()) {
            hide_row = false;
            break;
        }
    }

    if(hide_row != ui->tableWidgetDev->isRowHidden(ROW_DEV_MS_ERR))
        ui->tableWidgetDev->setRowHidden(ROW_DEV_MS_ERR, hide_row);
    update_table_size(ui->tableWidgetDev);
}

int MainWindow::get_col_by_serial(QString serial)
{
    QTableWidget *table = ui->tableWidgetDev;
    for(int col=table->columnCount()-1; col>=0; --col) {
        if(getDevByTableItem(table->horizontalHeaderItem(col)).getSerialStr() == serial)
            return col;
    }
    return -1;
}

void MainWindow::update_table_size(QTableWidget * const table, bool widthOnly)
{
//    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width()+5;
    for(int col=0; col<table->columnCount(); ++col) {
        if(table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if(width<20)
        width = 20;

    int height = table->horizontalHeader()->height()+5;
    for(int row=0; row<table->rowCount(); ++row) {
        if(table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if(height<20)
        height = 20;
    if(widthOnly) {
        width += table->verticalScrollBar()->width();
        table->setMinimumWidth(width);
    } else
        table->setFixedSize(width, height);
    table->updateGeometry();
}

void MainWindow::onRunError(const RcRunId &runId, QString errStr)
{
    if(config.runEnable && config.runId == runId) {
        if(remoteConnected){
            QString mess = QString("Retransmit runError to the remoteControl in Run %1")
                    .arg(config.runId.toString());
            if(!errStr.isEmpty())
                mess.append(": ").append(errStr);
            qWarning() << mess;
            emit sendRunError(runId, errStr);
        } else {
            QString mess = QString("Errors in Run %1")
                    .arg(config.runId.toString());
            if(!errStr.isEmpty())
                mess.append(": ").append(errStr);
            qWarning()<< mess;

            perfomeStop(true);
        }
//    } else {
//        qWarning() << QString("Got RunError from %1 run").arg(config.runEnable?"previous":"stopped");
    }
}

void MainWindow::onRunWarning(const RcRunId &runId, QString warnStr)
{
    if(!warnStr.isEmpty() &&
            (!config.runEnable ||
             config.runId != runId) )
        return;

    QLabel *label=nullptr;
    if(sender() == ui->clientManagerWidget){
        label = labelWarnClients;
        warnStr.prepend("Clients: ");
    }
    else
        if(devices.values().contains(qobject_cast<DeviceGlob*>(sender()))){
            label = labelWarnTtvxs;
            warnStr.prepend("Ttvxs: ");
        }
        else{
            label = labelWarnOther;
            warnStr.prepend("Other: ");
        }
    const int WARN_SIZE_LIMIT= 90;
    if(warnStr.size()>WARN_SIZE_LIMIT) {
        warnStr = warnStr.left(WARN_SIZE_LIMIT-3)+"...";
    }
    label->setText(warnStr);
    label->setHidden(false);

    updateGroupBoxWarn();
}

void MainWindow::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(config.runEnable && config.runId == runId) {
        if(remoteConnected) {
            QString mess = QString("Retransmit backPressure(%2) to the remoteControl in Run %1")
                    .arg(config.runId.toString()).arg(backPressure);
            if(backPressure)
                qWarning() << mess;
            else
                qInfo() << mess;
            emit sendBackPressure(runId, backPressure);
        } else {
            if(backPressure) {
                setTrigState(false);
                qWarning()<< "BackPressure enabled";
            } else {
                setTrigState(true);
                qInfo()<< "BackPressure disabled";

            }
        }
    }
}

void MainWindow::onEvNumCheckFail(QString mess)
{
    onRunWarning(config.runId, mess);
}

void MainWindow::client_config_receiver(QMap<int, ClientConfig> client_config_)
{
    config.clients = client_config_;
    emit configChanged(config);
}

void MainWindow::setTabState(bool isOk, int tabIndex)
{
    QTabBar *bar = ui->tabWidget->tabBar();
    QColor c = bar->tabTextColor(tabIndex);
    if(c.isValid() == isOk) {
        bar->setTabTextColor(tabIndex, isOk ? QColor() : Qt::red);
    }
    bool prevOk = isReadForRun();
    tabOk[tabIndex] = isOk;
    bool newOk = isReadForRun();
    if(newOk != prevOk)
        updateProgramDescription();
}

void MainWindow::config_transfer_from_clients(int id, const QJsonObject &json)
{
    for(int id_tmp : config.clients.keys()) {
        if(config.clients[id_tmp].id == id) {
            clients_config[id] = json;
        }
    }

    bool all_clients_config = true;
    for(int id_tmp : config.clients.keys()) {
        if(!clients_config.contains(id_tmp)) {
            all_clients_config = false;
        }
    }

    if(all_clients_config) {
//        write_run_history();
    }
}

void MainWindow::onRemoteStartRequest(const RcRunId &runId, quint64 startEvNumber)
{
    qInfo() << __PRETTY_FUNCTION__;
    Q_UNUSED(startEvNumber)
    megamaster_flg = false;
    bool success = perfomeStart(runId);
    clients_config.clear();
    emit remoteStartCompleted(success, runId);
    if(success && config.clients.isEmpty()) {
        const auto &json = QVariant::fromValue(config).value<QJsonObject>();
        emit config_transfer_to_rc(QJsonDocument(json).toJson());
    }

    updateProgramDescription();
}

void MainWindow::onRemoteStopRequest(const RcRunId &runId, quint64 stopEvNumber)
{
    Q_UNUSED(stopEvNumber)
    perfomeStop(false);
    updateProgramDescription();
}

void MainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl) == 0) {
        ProgramInterface interface;
        interface.setType(ProgramInterfaceRemoteControl);
        progDesc.interfaces.push_back(interface);
    }

    ProgramInterface &i = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    i.port = status.tcpPort;
    i.enabled = status.listening;
    i.isFree = !status.portInUse;

    if(status.portInUse)
        i.setPeer(status.peerHost, status.peerPort);
    else
        i.peers.clear();

    updateProgramDescription();
    bool remoteDisconnectedOnRun = config.runEnable
            && remoteConnected && !status.portInUse;
    remoteConnected = status.portInUse;
    if(remoteDisconnectedOnRun) {
        qWarning() << "Stop run due to remoteControl disconnect";
        perfomeStop(true);
    }
    updateStatStopButton();
}

void MainWindow::addLog(QtMsgType severity, const QString &log, const QString &thrName)
{
    QString msg = QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);

    Qt::GlobalColor color = Qt::black;
    switch (severity) {
    case QtCriticalMsg:
        color = Qt::magenta;
        break;
    case QtWarningMsg:
        color = Qt::red;
        break;
    case QtInfoMsg:
        color = Qt::blue;
        break;
    case QtDebugMsg:
#ifndef QT_DEBUG
        return;
#endif
    case QtFatalMsg:
        qInfo()<< "";
    }
    QTextCharFormat tf = ui->plainTextEdit_log->currentCharFormat();
    tf.setForeground(QBrush(color));
    ui->plainTextEdit_log->setCurrentCharFormat(tf);
    ui->plainTextEdit_log->appendPlainText(QString("[%1] [%2] %3")
        .arg(QTime::currentTime().toString(), thrName, log));
}

//=== GUI SLOTS
void MainWindow::on_actionChange_devices_triggered()
{
    discoverDialog->show();
}

void MainWindow::on_actionReset_timestamp_triggered()
{
    emit resetTimestamp();
}

void MainWindow::on_actionSoftclear_triggered()
{
    emit softClear();
}

void MainWindow::on_actionClear_stats_triggered()
{
    emit clearStat();
}

void MainWindow::on_actionSNMP_dialog_triggered()
{
    SnmpDialog dialog(this);
    dialog.exec();
    close();
}

void MainWindow::on_actionEvent_number_reset_triggered()
{
    emit evNumReset();
}

void MainWindow::on_actionAllow_Remote_Control_toggled(bool val)
{
    config.allowRemoteControl = val;
    emit configChanged(config);
    emit allowRemoteControl(config.allowRemoteControl);
    updateProgramDescription();
}

void MainWindow::on_pushButton_start_clicked()
{
    if(devices.size() == 1 )
        megamaster_flg = true;
    else{
        qInfo() << "Start failed! More then one device.";
        return;
    }
    clients_config.clear();

    const RcRunId newRunId(localProgramIndex.toString(),
                           ++config.localRunNumber,
                           QUuid::createUuid());
    bool success = perfomeStart(newRunId);

    if(success && config.clients.isEmpty()) {
        const auto &json = QVariant::fromValue(config).value<QJsonObject>();
        emit write_history_config(newRunId, json, true);
    }
    emit configChanged(config);
}

void MainWindow::on_pushButton_stop_clicked()
{
    perfomeStop(false);
}

void MainWindow::on_radioButton_open_toggled(bool checked)
{
    config.default_cfg.trig_open = checked;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_checkBox_external_ttl_toggled(bool checked)
{
    config.default_cfg.trigSrc.bit.ttl = checked;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_checkBox_timer_toggled(bool checked)
{
    ui->spinBox_timer->setEnabled(checked);
    config.default_cfg.trigSrc.bit.fixed_pulser = checked;
    config.unify_config();
    emit configChanged( config);
}

void MainWindow::on_checkBox_random_toggled(bool checked)
{
    ui->spinBox_random->setEnabled(checked);
    ui->doubleSpinBox_dead_time->setEnabled(checked);
    ui->labelDeadTime->setEnabled(checked);
    config.default_cfg.trigSrc.bit.random_pulser = checked;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_spinBox_timer_valueChanged(int val)
{
    config.default_cfg.trigTimerFreq = val;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_spinBox_random_valueChanged(int val)
{
    config.default_cfg.trigRandomFreq = val;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_doubleSpinBox_dead_time_valueChanged(double val)
{
    config.default_cfg.trigRandomDeadTimeUs = val;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_checkBox_aux_cnt_for_trig_toggled(bool checked)
{
    config.default_cfg.aux_cnt_for_trig = checked;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_groupBoxStat_toggled(bool checked)
{
    config.default_cfg.stat_readout = checked;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_spinBoxStatPeriod_valueChanged(int period_ms)
{
    config.default_cfg.stat_period = static_cast<quint16>(period_ms);
    config.unify_config();
    emit configChanged(config);
}


void MainWindow::on_comboBoxPreset_currentIndexChanged(int index)
{
    StatPreset stat_preset;
    int preset_type = ui->comboBoxPreset->itemData(index).toInt();
    stat_preset = StatPresetFabric::getPreset(preset_type);

    config.default_cfg.stat_preset = stat_preset;
    config.unify_config();
    emit configChanged(config);
}

void MainWindow::on_tableWidgetDev_cellChanged(int row, int column)
{
    if(row == ROW_NAME) {
        auto table = ui->tableWidgetDev;
        DeviceIndex index = getDevByCol(column);
        if(index.isNull())
            return;
        auto it = config.devicesCfg.find(index);
        if(it == config.devicesCfg.end())
            return;
        it->custom_name = table->item(row, column)->text();
        emit configChanged(config);
        return;
    }
    if(row < ROW_CH_SLOT_FIRST || row > ROW_CH_SLOT_LAST) return;
    DeviceIndex index = getDevByCol(column);
    if(index.isNull()) return;
    const int slot = row-ROW_CH_SLOT_FIRST+1;
    if(isSlotBlocked(row)) return;
    const int ch = slotToCh(slot);

    Q_ASSERT(config.devicesCfg.contains(index));
    auto &cfg = config.devicesCfg[index];
    bool en = ui->tableWidgetDev->item(row, column)->checkState() == Qt::Checked;
    if(((cfg.vxsChMask & (1<<ch)) != 0) != en) {
        if(en)
            cfg.vxsChMask |= 1<<ch;
        else
            cfg.vxsChMask &= ~static_cast<quint32>(1<<ch);

        emit configChanged(config);
    }
}

void MainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void MainWindow::on_actionFeLinkStatusDialog_triggered()
{
    feLinkStatusDialog->show();
}

