#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QProcess>

#include "TtvxsAppConfig.h"

#include "app/BaseMainWindow.h"
#include "base/DeviceIndexEnabled.h"
#include "base/ProgramIndex.h"
#include "fsm/FsmState.h"
#include "pnp-server/PNPMessage.h"

namespace Ui {
class MainWindow;
}

class ConfigIndexName;
class ConfigSelector;
class ConfigurationManager;
class DeviceDescription;
class DeviceDispatcher;
class DeviceGlob;
class DeviceIndex;
class DiscoverDialog;
class EvNumChecker;
class FeLinkStatusDialog;
class LldpInfo;
class LldpInfoCache;
class MStreamStatus;
class PNPServer;
class QLabel;
class QTableWidget;
class QTableWidgetItem;
struct RcRunId;
class RemoteControlServer;
class RemoteControlServerStatus;
class TtvxsAppConfigDispatcher;
class TtvxsConfig;
class TtvxsConfigDb;
class TtvxsStatus;
class RcFsm;
class TrigDeviceMgrFsm;

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(const ConfigSelector &cs, QWidget *parent = nullptr);
    ~MainWindow();

signals:
    void switch_to_config(const ConfigIndexName &conf) const;
    void configChanged(const TtvxsAppConfig &config);

    //=== signals to dev
    void closeAll();
    void deleteHardware_(DeviceIndex);
    void resetTimestamp(DeviceIndex=DeviceIndex());
    void clearStat(DeviceIndex=DeviceIndex());
    void evNumReset(DeviceIndex=DeviceIndex());
    void softClear(DeviceIndex=DeviceIndex());
    void restartAll(const RcRunId &runId) const;

    void changeProgramDescription(const ProgramDescription &) const;
    void closeProgram(const ProgramDescription &) const;
    void deleteAll() const;
    void allowRemoteControl(bool) const;
    void sendRemoteState(QString) const;
    void sendRunError(const RcRunId &, QString) const;
    void sendRunWarn(const RcRunId &, QString) const;
    void sendBackPressure(const RcRunId &, bool) const;
    void write_history_config(const RcRunId &,
                              const QJsonObject &,
                              bool megamaster_flg) const;

    void config_transfer_to_rc(QByteArray data);
    void sendControlEnable(bool block) const;
    void stopAll(quint64 stopEvNum=0) const;
    void remoteStartCompleted(bool, const RcRunId &) const;
    void remoteStopCompleted(bool, const RcRunId &) const;
    void set_run_state(bool run);

public slots:
    void setConfig(int config_key, const TtvxsAppConfig &config_);
    void configSelectorChanged(const ConfigIndexName &conf);
    void lldpCacheUpdated(const LldpInfoCache &lldpCache);
    void onRcFsmStateChange(const FsmState &state, const FsmState &oldState);
    void onTrigFsmStateChange(const FsmState &state);
    void clientManagerRunStateChanged(const FsmState &state);

private:
    void init_GUI_style();
    void init_GUI_default_settings();
    void initTableDev();
    void initStatPresets();
    void writeSetup(DeviceIndex index=DeviceIndex());
    void addDevice(const DeviceIndex &index);
    void disconnectFromDev(const DeviceIndex &index);
    void sendLog(QString log);
    void updateGui();
    int getChCfgCol(const DeviceIndex &i) const;
    DeviceIndex getDevByTableItem(const QTableWidgetItem *item) const;
    DeviceIndex getDevByCol(int col) const;
    bool isSlotBlocked(int row);
    bool perfomeStart(const RcRunId &runId);
    void openGlobTrig();
    void closeGlobTrig();
    void setTrigState(bool open);
    void actionOnStateRun();
    void actionOnStateStop();
    void perfomeStop(bool isError);
    bool isReadForRun() const;
    void updateProgramDescription();
    void blockControl(bool block);
    void updateStatStopButton();
    void updateGroupBoxWarn();
    void logRcFsmStateChange(const FsmState &previousState,
                             const FsmState &state);
    void block_widgets_signal(bool block);

private slots:
    void statusUpdated(const TtvxsStatus &st, const TtvxsStatus &prev_st);
    void updateDeviceLldpInfo(const DeviceIndex &index, const LldpInfo &lldp);
    void mstream_status_updated(QString serial, const MStreamStatus &status);
    int get_col_by_serial(QString serial);
    static void update_table_size(QTableWidget * const table, bool widthOnly=false);
    void onRunError(const RcRunId &runId, QString errStr);
    void onRunWarning(const RcRunId &runId, QString warnStr);
    void onBackPressure(const RcRunId &runId, bool backPressure);
    void onEvNumCheckFail(QString mess);
    void client_config_receiver(QMap<int, ClientConfig> client_config_);
    void setTabState(bool isOk, int tabIndex);
    void config_transfer_from_clients(int id, const QJsonObject &json);
    void onRemoteStartRequest(const RcRunId &runId, quint64 startEvNumber);
    void onRemoteStopRequest(const RcRunId &runId, quint64 stopEvNumber);
    void onRemoteStatusUpdated(const RemoteControlServerStatus &status);
    void addLog(QtMsgType severity,
                const QString &log,
                const QString &thrName) override;

    //GUI slots
    void on_actionChange_devices_triggered();
    void on_actionReset_timestamp_triggered();
    void on_actionSoftclear_triggered();
    void on_actionClear_stats_triggered();
    void on_actionSNMP_dialog_triggered();
    void on_actionFeLinkStatusDialog_triggered();
    void on_actionEvent_number_reset_triggered();
    void on_actionAllow_Remote_Control_toggled(bool val);
    void on_pushButton_start_clicked();
    void on_pushButton_stop_clicked();
    void on_radioButton_open_toggled(bool checked);
    void on_checkBox_external_ttl_toggled(bool checked);
    void on_checkBox_timer_toggled(bool checked);
    void on_checkBox_random_toggled(bool checked);
    void on_spinBox_timer_valueChanged(int val);
    void on_spinBox_random_valueChanged(int val);
    void on_doubleSpinBox_dead_time_valueChanged(double val);
    void on_checkBox_aux_cnt_for_trig_toggled(bool checked);
    void on_groupBoxStat_toggled(bool checked);
    void on_spinBoxStatPeriod_valueChanged(int period_ms);
    void on_comboBoxPreset_currentIndexChanged(int index);
    void on_tableWidgetDev_cellChanged(int row, int column);
    void on_actionConfiguration_manager_triggered();

private:
    Ui::MainWindow *ui;
    ConfigurationManager * const configuration_manager;
    QString program_index;
    TtvxsAppConfig config;
    DeviceIndexEnabledMap devMap;
    ProgramIndex localProgramIndex;
    TtvxsConfigDb * const configDb;
    TtvxsAppConfigDispatcher * const configDispatcher;
    QMap<int, QJsonObject> clients_config;
    DiscoverDialog * const discoverDialog;
    FeLinkStatusDialog * const feLinkStatusDialog;
    DeviceDispatcher * const deviceDispatcher;
    PNPServer * const pnpServer;
    ProgramDescription progDesc;
    RemoteControlServer * const remoteControl;
    bool remoteConnected;
    EvNumChecker * const evNumChecker;
    QMap<DeviceIndex, DeviceGlob *> devices;
    QTimer * const writeRunHistoryTimer;
    QMap<int, bool> tabOk;
    bool megamaster_flg;

    RcFsm * const rc_fsm;
    TrigDeviceMgrFsm * const trig_fsm;

    QLabel *labelWarnTtvxs;
    QLabel *labelWarnClients;
    QLabel *labelWarnOther;

    QLabel * const labelFsmStateRc;
    QLabel * const labelFsmStateTrig;
    QLabel * const labelFsmStateClient;

    QList<QObject*> widgets_to_block;
};


#endif // MAINWINDOW_H
