/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TtvxsAppConfigConverter.h"

#include "TtvxsAppConfig.h"
#include "StatPresetConverter.h"

#include "base/ClientConfigConverter.h"
#include "base/ConfigConverterUtil.h"
#include "base/DeviceIndexConverter.h"
#include "base/RunIdConverter.h"

namespace {
// reg_trig_src
constexpr const char *k_ttl = "ttl";
constexpr const char *k_sfp = "sfp";
constexpr const char *k_random_pulser = "random_pulser";
constexpr const char *k_fixed_pulser = "fixed_pulser";

// TtvxsConfig
constexpr const char *k_device_index = "device_index";
constexpr const char *k_enabled = "enabled";
constexpr const char *k_run = "run";
constexpr const char *k_trig_open = "trig_open";
constexpr const char *k_trigSrc = "trigSrc";
constexpr const char *k_vxsChMask = "vxsChMask";
constexpr const char *k_trigTimerFreq = "trigTimerFreq";
constexpr const char *k_trigRandomFreq = "trigRandomFreq";
constexpr const char *k_trigRandomDeadTimeUs = "trigRandomDeadTimeUs";
constexpr const char *k_aux_cnt_for_trig = "aux_cnt_for_trig";
constexpr const char *k_stat_readout = "stat_readout";
constexpr const char *k_stat_period = "stat_period";
constexpr const char *k_stat_preset = "stat_preset";
constexpr const char *k_slot = "slot";
constexpr const char *k_custom_name = "custom_name";

// TtvxsAppConfig
constexpr const char *k_devicesCfg = "devicesCfg";
constexpr const char *k_default_cfg = "default_cfg";
constexpr const char *k_runEnable = "runEnable";
constexpr const char *k_runId = "runId";
constexpr const char *k_localRunNumber = "localRunNumber";
constexpr const char *k_allowRemoteControl = "allowRemoteControl";
constexpr const char *k_clients = "clients";
}

// ============ reg_trig_src
template<>
bool AbstractConfigConverter<reg_trig_src>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<reg_trig_src>::toJsonObject(const reg_trig_src &in)
{
    QVariantMap m;
    m[k_ttl] = QVariant::fromValue(in.bit.ttl);
    m[k_sfp] = QVariant::fromValue(in.bit.sfp);
    m[k_random_pulser] = QVariant::fromValue(in.bit.random_pulser);
    m[k_fixed_pulser] = QVariant::fromValue(in.bit.fixed_pulser);
    return fromVariantMap(m);
}

template<>
reg_trig_src AbstractConfigConverter<reg_trig_src>::fromJsonObject(const QJsonObject &in)
{
    reg_trig_src out;
    out.bit.ttl = in[k_ttl].toInt();
    out.bit.sfp = in[k_sfp].toInt();
    out.bit.random_pulser = in[k_random_pulser].toInt();
    out.bit.fixed_pulser = in[k_fixed_pulser].toInt();
    return out;
}

template<>
QJsonValue AbstractConfigConverter<reg_trig_src>::toJsonValue(const reg_trig_src &in)
{
    return toJsonObject(in);
}

template<>
reg_trig_src AbstractConfigConverter<reg_trig_src>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<reg_trig_src>::registerSubTypes()
{
}

// ============ TtvxsConfig
template<>
bool AbstractConfigConverter<TtvxsConfig>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<TtvxsConfig>::toJsonObject(const TtvxsConfig &in)
{
    QVariantMap m;
    m[k_device_index] = QVariant::fromValue(in.device_index);
    m[k_enabled] = QVariant::fromValue(in.enabled);
    m[k_run] = QVariant::fromValue(in.run);
    m[k_trig_open] = QVariant::fromValue(in.trig_open);
    m[k_trigSrc] = QVariant::fromValue(in.trigSrc);
    m[k_vxsChMask] = QVariant::fromValue(in.vxsChMask);
    m[k_trigTimerFreq] = QVariant::fromValue(in.trigTimerFreq);
    m[k_trigRandomFreq] = QVariant::fromValue(in.trigRandomFreq);
    m[k_trigRandomDeadTimeUs] = QVariant::fromValue(in.trigRandomDeadTimeUs);
    m[k_aux_cnt_for_trig] = QVariant::fromValue(in.aux_cnt_for_trig);
    m[k_stat_readout] = QVariant::fromValue(in.stat_readout);
    m[k_stat_period] = QVariant::fromValue(in.stat_period);
    m[k_stat_preset] = QVariant::fromValue(in.stat_preset);
    m[k_slot] = QVariant::fromValue(in.slot);
    m[k_custom_name] = QVariant::fromValue(in.custom_name);

    return fromVariantMap(m);
}

template<>
TtvxsConfig AbstractConfigConverter<TtvxsConfig>::fromJsonObject(const QJsonObject &in)
{
    TtvxsConfig out;
    update_value(in, k_device_index, &out.device_index);
    update_value(in, k_enabled, &out.enabled);
    update_value(in, k_run, &out.run);
    update_value(in, k_trig_open, &out.trig_open);
    update_value(in, k_trigSrc, &out.trigSrc);
    update_value(in, k_vxsChMask, &out.vxsChMask);
    update_value(in, k_trigTimerFreq, &out.trigTimerFreq);
    update_value(in, k_trigRandomFreq, &out.trigRandomFreq);
    update_value(in, k_trigRandomDeadTimeUs, &out.trigRandomDeadTimeUs);
    update_value(in, k_aux_cnt_for_trig, &out.aux_cnt_for_trig);
    update_value(in, k_stat_readout, &out.stat_readout);
    update_value(in, k_stat_period, &out.stat_period);
    update_value(in, k_stat_preset, &out.stat_preset);
    update_value(in, k_slot, &out.slot);
    update_value(in, k_custom_name, &out.custom_name);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<TtvxsConfig>::toJsonValue(const TtvxsConfig &in)
{
    return toJsonObject(in);
}

template<>
TtvxsConfig AbstractConfigConverter<TtvxsConfig>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<TtvxsConfig>::registerSubTypes()
{
    AbstractConfigConverter<reg_trig_src>::registerTypes();
    AbstractConfigConverter<StatPreset>::registerTypes();
}

// ============ TtvxsAppConfig
template<>
bool AbstractConfigConverter<TtvxsAppConfig>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<TtvxsAppConfig>::toJsonObject(const TtvxsAppConfig &in)
{
    QVariantMap m;
    m[k_devicesCfg] = ConfigConverterUtil::toJsonValue(in.devicesCfg);
    m[k_default_cfg] = QVariant::fromValue(in.default_cfg);
    m[k_runEnable] = QVariant::fromValue(in.runEnable);
    m[k_runId] = QVariant::fromValue(in.runId);
    m[k_localRunNumber] = QVariant::fromValue(in.localRunNumber);
    m[k_allowRemoteControl] = QVariant::fromValue(in.allowRemoteControl);
    m[k_clients] = ConfigConverterUtil::toJsonValue(in.clients);
    return fromVariantMap(m);
}

template<>
TtvxsAppConfig AbstractConfigConverter<TtvxsAppConfig>::fromJsonObject(const QJsonObject &in)
{
    TtvxsAppConfig out;
    out.devicesCfg = ConfigConverterUtil::toMap<DeviceIndex, TtvxsConfig>(in[k_devicesCfg]);
    update_value(in, k_default_cfg, &out.default_cfg);
    update_value(in, k_runEnable, &out.runEnable);
    update_value(in, k_runId, &out.runId);
    update_value(in, k_localRunNumber, &out.localRunNumber);
    update_value(in, k_allowRemoteControl, &out.allowRemoteControl);
    out.clients = ConfigConverterUtil::toMap<int, ClientConfig>(in[k_clients]);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<TtvxsAppConfig>::toJsonValue(const TtvxsAppConfig &in)
{
    return toJsonObject(in);
}

template<>
TtvxsAppConfig AbstractConfigConverter<TtvxsAppConfig>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<TtvxsAppConfig>::registerSubTypes()
{
    AbstractConfigConverter<TtvxsConfig>::registerTypes();
    AbstractConfigConverter<DeviceIndex>::registerTypes();
    AbstractConfigConverter<RcRunId>::registerTypes();
    AbstractConfigConverter<ClientConfig>::registerTypes();
}
