/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcFsm.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QTimer>

namespace {

Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.rc.fsm")
} // namespace

RcFsm::RcFsm(QObject *parent) : BaseFsm(parent)
{
    setObjectName("RcFSM");
    recreateInput(new RcFsmInput());
}

bool RcFsm::evaluate()
{
    if(BaseFsm::evaluate()) {
        if(getState() == FsmState::Error)
            printInputString(qDebug(fsm));
        return true;
    }

    if(getState() == FsmState::Error) {
        if (!anyError()) {
            setState(FsmState::Ready);
            return true;
        }
    } else {
        if (anyError()) {
            if(getState() == FsmState::Starting) {
                qDebug(fsm) << objectName() << "Got errr on StartRun"
                            << fsmStateToString(getInput()->clientManagerRunState)
                            << getInput()->vmeTrigsRunState;
            }
            setState(getState() == FsmState::Run ? FsmState::Stopping : FsmState::Error);
            return true;
        }
    }

    switch (getState()) {
    case FsmState::Ready:
        if (getRequestState() == FsmState::Starting ||
                getRequestState() == FsmState::CloseGlobTrig) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::CloseGlobTrig:
        if (getInput()->trigDeviceMgrState == FsmState::CloseGlobTrig &&
                (getRequestState() == FsmState::WaitSpillEnd ||
                 getRequestState() == FsmState::Starting)) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::WaitSpillEnd:
        if (getInput()->spillStateOld == true && getInput()->spillStateCur == false) {
            setState(FsmState::Starting);
            return true;
        }
        break;
    case FsmState::Starting:
        if(getRequestState() == FsmState::Run) {
            if (allInState(FsmState::Run)) {
                acceptRequestedState();
                return true;
            }
        }
        if(getRequestState() == FsmState::OpenGlobTrig) {
            if (getInput()->trigDeviceMgrState == FsmState::CloseGlobTrig &&
                    (getInput()->clientManagerRunState == FsmState::Run) &&
                    (getInput()->vmeTrigsRunState == FsmState::Run
                     || getInput()->vmeTrigsRunState == FsmState::Unknown)) {
                acceptRequestedState();
                return true;
            }
        }
        break;
    case FsmState::OpenGlobTrig:
        if(getRequestState() == FsmState::Run) {
            if (allInState(FsmState::Run)) {
                acceptRequestedState();
                return true;
            }
        }
        break;
    case FsmState::Run:
        if (false && getInput()->trigDeviceMgrState != FsmState::Run) { // optional
            setState(FsmState::Stopping);
            return true;
        }
        if (getRequestState() == FsmState::Stopping || getRequestState() == FsmState::PrepareStopping) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::PrepareStopping:
        if(getRequestState() == FsmState::Stopping)
            if (getInput()->trigDeviceMgrState == FsmState::Ready) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::Stopping:
//        if(getRequestState() == FsmState::Ready)
            if (allInState(FsmState::Ready)) {
//                acceptRequestedState();
                setState(FsmState::Ready);
                return true;
            }
        break;
    default:
    {};
    }
    return false;
}

void RcFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 RcFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::CloseGlobTrig:
    case FsmState::OpenGlobTrig:
    case FsmState::PrepareStopping:
        return 600;
    case FsmState::WaitSpillEnd:
        return 150000; // FIXME: read actual spill time from hardware
    case FsmState::Starting:
    case FsmState::Stopping:
        return 3000;
    default:
        return BaseFsm::stateTimeoutMs(state);
    }
}

void RcFsm::setInputState(FsmState RcFsmInput::*p, const char *fieldName, const FsmState &state)
{
    qDebug(fsm) << objectName() << fieldName << " change from: "
                << fsmStateToString(getInput()->*p)
                << " -> " << fsmStateToString(state);
    getInput()->*p = state;

    printInputString(qDebug(fsm));
    evaluate();
}

bool RcFsm::allInState(FsmState state) const
{
    return getInput()->trigDeviceMgrState == state &&
            getInput()->clientManagerRunState == state &&
            (getInput()->vmeTrigsRunState == state ||
             getInput()->vmeTrigsRunState == FsmState::Unknown);
}

bool RcFsm::anyError() const
{
    return getInput()->trigDeviceMgrState == FsmState::Error ||
            getInput()->clientManagerRunState == FsmState::Error ||
            getInput()->vmeTrigsRunState == FsmState::Error ||
            getInput()->trigDeviceMgrState == FsmState::Unknown ||
            getInput()->clientManagerRunState == FsmState::Unknown
            ;
}

void RcFsm::setTrigDeviceMgrState(const FsmState &state)
{
    setInputState(&RcFsmInput::trigDeviceMgrState, "trigDeviceMgrState", state);
}

void RcFsm::setClientManagerRunState(const FsmState &state)
{
    setInputState(&RcFsmInput::clientManagerRunState, "clientManagerRunState", state);
}

void RcFsm::setVmeTrigsRunState(const FsmState &state)
{
    setInputState(&RcFsmInput::vmeTrigsRunState, "vmeTrigsRunState", state);
}

void RcFsm::setSpillState(bool inSpillState)
{
    if(getInput()->spillStateCur == inSpillState &&
            getInput()->spillStateOld == getInput()->spillStateCur)
        return;

    getInput()->spillStateOld = getInput()->spillStateCur;
    getInput()->spillStateCur = inSpillState;
    if(getState() == FsmState::WaitSpillEnd)
        evaluate();
}

//QString RcFsmInput::toString(const RcFsmInput &r)
//{
//    QStringList sl;
//    sl += QString("readyForRun: %1").arg(r.readyForRun);
//    sl += QString("requestStartRun: %1").arg(r.requestStartRun);
//    sl += QString("requestStopRun: %1").arg(r.requestStopRun);
//    sl += QString("clientManagerRunState: %1").arg(r.clientManagerRunState);
//    sl += QString("vmeTrigsRunState: %1").arg(r.vmeTrigsRunState);
//    return sl.join(", ");
//}

QDebug operator<<(QDebug dbg, const RcFsmInput &in)
{
    dbg << BaseFsmInput(in);

    dbg << " trigDeviceMgrState:" << in.trigDeviceMgrState;
    dbg << " clientManagerRunState:" << fsmStateToString(in.clientManagerRunState);
    dbg << " vmeTrigsRunState:" << in.vmeTrigsRunState;

//    dbg << RcFsmInput::toString(r);
    return dbg;
}
