#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QAbstractButton>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QProcess>
#include <QMessageBox>
#include <QMetaType>
#include <QTime>
#include <QMap>

#include "base/app/BaseMainWindow.h"
#include "AfiViewerConfig.h"
#include "CalConfig.h"
#include "pnp-server/PNPMessage.h"
#include "AfiViewerStartupOptions.h"
#include "configuration-manager/ConfigurationManager.h"
#include "histogramm/HistMngrCfg.h"
#include "histogramm/DeviceData.h"

namespace Ui {
class MainWindow;
}

class QTableWidget;
class QSemaphore;
class QTreeWidgetItem;

class AfiViewerConfigDb;
class AfiViewerSignalRouter;
class AfiViewerConfigDispatcher;
class CalibrationDialog;
class FileReaderDialog;
class LogDialog;
class InlCheckDialog;
class TQDCAnalyzer;
class TcpClient;
class PNPServer;
class HistWorker;
class StatusDialog;
class ViewerZmqPublisher;
class ZmqConfig;
class ZmqDialog;

Q_DECLARE_METATYPE(QHostAddress)

class MainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit MainWindow(AfiViewerStartupOptions startup_options, QWidget *parent = nullptr);
    ~MainWindow();

signals:
    void deleteDevice(const DeviceIndex &);
    void closeAll();
    void addLog(QString);
    void resetAll();
    void requestRawData(int events);
    void tcpAddressChanged(QHostAddress addr, quint16 port);
    void rawDataEveryEvUpdated(bool dataFromFile);
    void closeFile();
    void invertUpdated(QMap<DeviceIndex, QVector<bool> > invertMap) const;
    void tdcTtvxMapUpdated(TdcTtvxsMapType ttMap) const;
    void inlConfigUpdated(InlResults) const;
    void requestInlResults(bool allRes) const;
    void set_can_skip(bool skip) const;
    void configChanged(const AfiViewerConfig &);
    void closeProgram(const ProgramDescription &) const;
    void changeProgramDescription(const ProgramDescription &) const;
    void switch_to_config(const ConfigIndexName &conf) const;

public slots:
    void gotProgramDescription(const ProgramDescription &newDescr);
    void gotClosedProgram(const ProgramDescription &newDescr);
    void setConfig(const AfiViewerConfig &newConfig);
    void configSelectorChanged(const ConfigIndexName& conf);
    void inl_corrections_updated(InlResults res, bool allRes);

private slots:
    void on_actionConfiguration_manager_triggered();
    void on_actionSet_source_file_triggered();
    void on_actionConnect_to_source_program_triggered();
    void on_actionDisconnect_from_program_triggered();

    void on_pushButtonResetAll_clicked();

    void tcpStatusUpdated(bool connOk, double dataRate);
    void updateHitStats(const AnalyzerStat &stats);
    void rawDataUpdated(const QStringList &data);

    void on_actionExit_triggered();
    void on_actionCalibrationData_triggered();
    void on_actionReset_all_triggered();
    void on_actionSave_INL_corrections_triggered();
    void on_actionShow_INL_check_triggered();
    void on_actionSet_INL_separate_location_toggled(bool checked);
    void on_actionHPTDC_vol_check_triggered();
    void on_actionSet_TDC_TTVXS_map_triggered();
    void on_actionSatistic_dialog_triggered();
    void on_action0MQ_publisher_triggered();
    void on_actionHPTDC_errors_triggered();

    void on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column);
    void on_radioButtonRawEvery_toggled(bool checked);
    void on_pushButtonRequestRaw_clicked();
    void on_spinBoxRawEvents_valueChanged(int limit);
    void on_checkBoxRawPeriodical_toggled(bool checked);
    void on_spinBoxRawTimer_valueChanged(int val);

    void sourceFileNameUpdated(QString fn);
    void tcpConnected(bool connected);
    void rawDataPeriodicRequest();
    void updateStatus();

    void write_config_request();
    void write_config();
    void waveform_config_updated(const WaveformViewerConfig &cfg);

public slots:
    void addLogFromSender(QtMsgType severity, const QString &log, const QString &thrName);

private:
    void updateGuiToConfig();
    void read_gui_config();
    void initTree();
    void updateModelList(const AnalyzerStat &analStat);
    void updateTreeWithStats();
    void updateTreeItemStates();
    void updateInlMarks();
    void updateTtvxsMarks();
    void setTableRowsCount(int nch);
    void updateControlState();
    void sendLog(QString l);
    void updateSourceLabel();
    void changeSourceState(AfiViewerConfig::SOURCE_DATA st);
    QVector<ProgramDescription> getFreeProgs() const;
    bool hasFreeProgs() const;
    bool isFreeProg(const ProgramDescription &pd) const;
    ProgramInterface getFreeProgramInterface(const ProgramDescription &pd) const;
    void updateTcpInterface(const ProgramDescription &descr);
    void updateRawDataControl();
    void disconnectFromProg();
    void setSourceColor(QString col);

private:
    Ui::MainWindow *ui;
    QSemaphore * const analyzerSem;// free space in analyzer queue
    QSemaphore * const histSem;// free space in histogram queue
    QSemaphore * const writerSem;
    QSemaphore * const zmqSem;

    AfiViewerConfig config;
    ConfigurationManager * const configuration_manager;
    AfiViewerConfigDb * const afiViewerConfigDb;
    AfiViewerSignalRouter * const afiViewerSignalRouter;
    AfiViewerConfigDispatcher * const configDispatcher;
    CalConfig calConfig;
    QTimer  *write_config_timer;
    bool write_config_request_flag;
    FileReaderDialog * const fileReaderDialog;
    PNPServer * const pnpServer;
    ProgramDescription selectedProgram;
    ProgramInterface selectedInterface;

    QMap<DeviceIndex, QTreeWidgetItem *> modTreeItem;
    QLabel * const dataSourceLabel;
    QLabel * const dataRateLabel;
    QLabel * const analyzerStatusLabel;
    QLabel * const semStatusLabel;
    QLabel * const vxsClockLabel;
    QLabel * const analyzerLabel;
    LogDialog *const logDialog;
    CalibrationDialog * const calibrationDialog;
    InlCheckDialog * const inlCheckDialog;

    TcpClient * const tcpClient;
    QThread * const tcpClientThread;
    TQDCAnalyzer * const analyzer;
    QThread * const analyzerThread;
    HistWorker * const worker;
    QThread * const worker_thread;

    QTreeWidgetItem * const rootTreeItem;
    QVector<ProgramDescription> progDescrs;
    QMap<QUuid, ProgramDescription> progDescrsMap;
    AnalyzerStat lastAnalStat;
    QTimer * const rawDataPeriodicTimer;
    QTimer * const statusTimer;
    bool rawDataReqDone;
    uint rawEventCnt;
    QElapsedTimer rawEventCntTimer;
    bool tcpConnOk;
    StatusDialog * const statusDialog;
    ViewerZmqPublisher * const zmqPublisher;
    QThread * zmq_thread;
    ZmqDialog * const zmqDialog;
};

#endif // MAINWINDOW_H
