#include "WaveformViewer.h"
#include "ui_WaveformViewer.h"

#include "base/DeviceIndexConverter.h"
#include "mregdev/qtmregdevice.h"

#include <QTimer>

int getDeviceChNum(const DeviceIndex &index)
{
    if(index.getDevId() == DEVICE_ID_PHOS_FEC)
        return 64;
    else
        return QtMregDevice::getDeviceChNum(index);

}

WaveformViewer::WaveformViewer(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::WaveformViewer),
    refreshTimer(new QTimer(this))
{
    ui->setupUi(this);
    AbstractConfigConverter<DeviceIndex>::registerTypes();

    connect(refreshTimer, &QTimer::timeout, [=](){this->refreshPlot();});
    refreshTimer->setSingleShot(true);
    reset();
}

WaveformViewer::~WaveformViewer()
{
    refreshTimer->stop();
    delete ui;
}

void WaveformViewer::deviceDataUpdated(const DecodedData &data, bool writeDecodedEnabled)
{
    Q_UNUSED(writeDecodedEnabled)
    pendingData = data;
    hasPendingData = true;
    if(!refreshTimer->isActive())
        refreshTimer->start(500);
}

void WaveformViewer::refreshPlot(bool force)
{
    if(!hasPendingData && !force)
        return;
    if(hasPendingData) {
        lastData = pendingData;
        hasPendingData = false;
    }
    if(lastData.events.isEmpty())
        return;
    showWaveform(pendingData.events.last());
    if(!force)
        refreshTimer->start(500);
}

void WaveformViewer::reset(const WaveformViewerConfig &config)
{
    cfg = config;
    lastData.clear();
    pendingData.clear();
    hasPendingData = false;
    addedDevices.clear();
    hasSeveralDev = false;
    hasSeveralTypes = false;
    initGui();
    ui->widgetWaveform->clear();
}

void WaveformViewer::on_checkBoxEn_toggled(bool checked)
{
    cfg.en = checked;
    ui->comboBoxDev->setEnabled(cfg.en);
    ui->spinBoxChMin->setEnabled(cfg.en);
    ui->spinBoxChMax->setEnabled(cfg.en);
    ui->widgetWaveform->setEnabled(cfg.en);
    emit configUpdated(cfg);
    refreshPlot(true);
}

void WaveformViewer::on_comboBoxDev_currentIndexChanged(int index)
{
    cfg.selectedDevice = ui->comboBoxDev->itemData(index).value<DeviceIndex>();
    emit configUpdated(cfg);
    refreshPlot();
}

void WaveformViewer::on_spinBoxChMin_valueChanged(int arg1)
{
    cfg.minCh = arg1;
    emit configUpdated(cfg);
    refreshPlot(true);
}

void WaveformViewer::on_spinBoxChMax_valueChanged(int arg1)
{
    cfg.maxCh = arg1;
    emit configUpdated(cfg);
    refreshPlot(true);
}

void WaveformViewer::initGui()
{
    ui->checkBoxEn->setEnabled(cfg.en);
    ui->comboBoxDev->clear();
    ui->comboBoxDev->addItem(" All ");
    if(!cfg.selectedDevice.isNull()) {
        addDevice(cfg.selectedDevice);
        ui->comboBoxDev->setCurrentIndex(1);
    }

    auto maxCh = qMax(qMax(cfg.minCh, cfg.maxCh), 0);
    ui->spinBoxChMin->setValue(0);
    ui->spinBoxChMin->setEnabled(false);
    ui->spinBoxChMin->setMaximum(maxCh);

    ui->spinBoxChMax->blockSignals(true);
    ui->spinBoxChMax->setValue(0);
    ui->spinBoxChMax->setEnabled(false);
    ui->spinBoxChMax->setMaximum(maxCh);
    ui->spinBoxChMax->blockSignals(false);

//    cfg.minCh = -1;
//    cfg.maxCh = 1000;
}

void WaveformViewer::addDevice(const DeviceIndex &index)
{
    ui->comboBoxDev->addItem(index.getIdent(), QVariant::fromValue(index));
    ui->comboBoxDev->model()->sort(0);

    auto maxCh = getDeviceChNum(index)-1;
    if(index.getDevId() == DEVICE_ID_PHOS_FEC)
        maxCh = 64-1;

    ui->spinBoxChMin->setMaximum(qMax(ui->spinBoxChMin->maximum(), maxCh));
    ui->spinBoxChMin->setEnabled(true);

    {
        auto sb = ui->spinBoxChMax;
        auto maxForMax = qMax(sb->maximum(), maxCh);
        if(sb->maximum() != maxForMax) {
            sb->blockSignals(true);
            sb->setMaximum(maxForMax);
            sb->setEnabled(true);
            if(cfg.maxCh == -1)
                sb->setValue(maxForMax);
            sb->blockSignals(false);
        }
    }

    addedDevices << index;
    DeviceIndex i;
    for(auto j : addedDevices) {
        if(i.isNull()) {
            i = j;
        } else {
            hasSeveralDev = true;
            if(i.getDevId() != j.getDevId())
                hasSeveralTypes = true;
        }
    }
}

void WaveformViewer::showWaveform(const CompleteEvent &ev)
{
    if(!cfg.en)
        return;

    for(auto devIt = ev.tqdcEv.cbegin(); devIt != ev.tqdcEv.cend(); ++devIt) {
        auto &index = devIt.key();
        if(ui->comboBoxDev->findData(QVariant::fromValue(index)) == -1) {
            addDevice(index);
        }
    }

    WaveformDataAll mvv;
    int chOffset = 0;
    for(auto devIt = ev.tqdcEv.cbegin(); devIt != ev.tqdcEv.cend(); ++devIt) {
        if(!cfg.selectedDevice.isNull() && devIt.key() != cfg.selectedDevice)
            continue;
        const auto &chHits = devIt.value().hits;

        for(auto chIt = chHits.cbegin(); chIt != chHits.cend(); ++chIt) {
            if((cfg.minCh!=-1 && chIt.key()<cfg.minCh) ||
                    (cfg.maxCh!=-1 && chIt.key()>cfg.maxCh))
                continue;
            const auto &adcHits = chIt.value().adcHits;
            for(auto &adc : adcHits) {
                cfg.selectedDevice.getIdent();
                QString chName = QString::number(chIt.key());
                if(hasSeveralDev)
                    chName.prepend(devIt.key().getSerialStr()+" ");
                if(hasSeveralTypes)
                    chName.prepend(devIt.key().getDeviceIdName()+" ");
//                        .arg((config.drawRelativeToBaseline && baseLine) ? "(b)": "")

                if(adc.adcTs) {
                    std::vector<double> vx = adc.wf;
                    for(uint i=0; i<adc.wf.size(); ++i)
                        vx[i] = i+adc.adcTs;
                    WaveformData wd(vx, adc.wf, chName.toStdString(), chIt.key()+chOffset);
                    mvv.push_back(wd);
                } else {
                    WaveformData wd(adc.wf, chName.toStdString(), chIt.key()+chOffset);
                    mvv.push_back(wd);
                }
            }
        }
        chOffset += getDeviceChNum(devIt.key());
    }

//    lastEventData = data;
//    EventData tmpData = data;

//    int chOffset=0;
//    for (quint32 serial : tmpData.devices.keys()) {
//        if(selectedSerial && selectedSerial != serial)
//            continue;
//        DeviceEvent &de = tmpData.devices[serial];
//        if(!adcCalData.contains(de.serial)){
//            updateAdc64Setup(de.devId, de.serial);
//        }
//        QMap<int, int> &baseLines = adcCalData[de.serial];
//        std::map<int, AdcChEvent>::iterator it=de.adcEvents.begin();
//        while(it!=de.adcEvents.end()) {
//            AdcChEvent &adcEv = it->second;
//            if(adcEv.ch<chDisplayLow || adcEv.ch>chDisplayHi){
//                it++;
//                continue;
//            }
//            const int &baseLine = baseLines[adcEv.ch];
//            const QString chName = QString("%1:%2%3")
//                    .arg(de.serial, 4, 16, QChar('0'))
//                    .arg(it->first+1)
//                    .arg((config.drawRelativeToBaseline && baseLine) ? "(b)": "");
//            std::vector<double>::iterator wfIt=adcEv.wf.begin();
//            while(wfIt!=adcEv.wf.end()){
//                if(config.drawRelativeToBaseline)
//                    (*wfIt) -= baseLine;
//                ++wfIt;
//            }
//            WaveformData wd(adcEv.wf, chName.toStdString(), it->first+chOffset);
//            mvv.push_back(wd);
//            it++;
//        }
//        chOffset += 64;
//    }

//    lastWaveformTime.start();
//    ui->widgetWaveform->replot();

    ui->widgetWaveform->setAllData(mvv);
//    ui->widget->setData(mvv);
//    statusUpdated(QString("Ev:%1").arg(data.evNum));
}
