﻿//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//    Copyright 2014-2015 Nikita Tarasov
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MainWindow.h"
#include "ui_MainWindow.h"

#include <math.h>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QStatusBar>
#include <QSettings>
#include <QThread>
#include <QTimer>
#include <QProgressBar>

#include "../version_str.h"
#include "PexmlFlashDev.h"
#include "flash-prog/DataBase.h"

namespace {
enum {COL_NAME, COL_SERIAL, COL_STATE, COL_DRV_FW, COL_DEV_FW, COL_TEMP, COL_PROGRESS, COL_NUM};
}

MainWindow::MainWindow(QWidget *parent):
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    taskId(100),
    devTaskId(0),
    connectedState(false),
    pexmlSerial(0),
    device(nullptr),
    devThread(nullptr),
    tempLabel(new QLabel(this)),
    fwLabel(new QLabel(this)),
    lastStatusLabel(new QLabel(this)),
    dataBase(new DataBase(this)),
    fileHeaderSize(-1),
    fileSize(-1),
    batchMode(false)
{
    qRegisterMetaType<Config>();

    ui->setupUi(this);
    ui->mainToolBar->hide();
    connect(ui->actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));
    setWindowTitle(QCoreApplication::applicationName());

    initTable();
    readConfig();

    ui->actionVerify_after_write->blockSignals(true);
    ui->actionVerify_after_write->setChecked(config.verifyAfterWrite);
    ui->actionVerify_after_write->blockSignals(false);
//    write65k();
    statusBar()->addPermanentWidget(lastStatusLabel);
    statusBar()->addPermanentWidget(tempLabel);
    statusBar()->addPermanentWidget(fwLabel);
}

MainWindow::~MainWindow()
{
    if(!batchMode)
        saveConfig();

    disconnectFromAll();
    delete ui;
}

void MainWindow::init(QString filePath, bool golden, int interfaceIndex)
{
    addDevice(interfaceIndex);
    batchMode = !filePath.isEmpty();

    if(batchMode) {
        if(device) {
            config.filePath = filePath;
            readFile();
            startWrite(golden);
        } else
            close();
    } else if(!config.filePath.isEmpty()) {
        ui->labelFile->setText(config.filePath);
        readFile();
    }
}

void MainWindow::on_actionWrite_Golden_Image_triggered()
{
    QMessageBox msgBox;
    msgBox.setText("Do you want to write \"Golden Image\"?");
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if(ret == QMessageBox::Yes)
        startWrite(true);
}

void MainWindow::on_actionGet_record_logs_triggered()
{
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    QSet<int> rows;
    foreach (const QTableWidgetSelectionRange &range, ranges) {
        for(int row = range.bottomRow(); row <= range.bottomRow(); ++row) {
            rows << row;
        }
    }

    qInfo() << QString("\t=== Log for %1 ===").arg(pexmlSerial, 8, 16, QChar('0'));
    dataBase->readInfo(pexmlSerial);
}

void MainWindow::on_actionVerify_after_write_toggled(bool doCheck)
{
    config.verifyAfterWrite = doCheck;
    config.saveSettings();
}

void MainWindow::on_actionVerify_Golden_image_triggered()
{
    startVerify(true);
}

void MainWindow::on_actionVerify_regular_image_triggered()
{
    startVerify(false);
}

void MainWindow::on_actionAbout_triggered()
{
    QString rev;
#ifdef SVN_REV
    rev = QString("%1").arg(SVN_REV);
    rev = rev.remove(QRegExp("^.*\\:"));
#endif
    if (!rev.isEmpty())
        rev = QString(", rev. %1").arg(rev);
    QString s = "<h1>"+qApp->applicationName()+"</h1><h2>Version "+QString(VERSION_STR)+rev+"</h2>";
    s += "<p>Copyright &copy; 2014-2015 <a href=\"http://afi.jinr.ru/\">AFI Electronics</a>";
    s += "<br>Joint Institute for Nuclear Research, Dubna";
    s += "<h3>Authors</h3>";
    s += "<p>Ivan Filippov";
    s += "<br>Alexey Shutov";
    s += "</p>";
    QMessageBox::about (this, "About", s);
}

void MainWindow::on_pushButtonFile_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this,
        tr("Open Bit file"), config.filePath, tr("Bit Files (*.bit)"));
    if(!fileName.isEmpty()) {
        fileHeaderSize = -1;
        fileSize = -1;
        config.filePath = fileName;
        ui->labelFile->setText(fileName);
        saveConfig();
        readFile();
    }
    updateWriteButtonState();
}

void MainWindow::on_pushButtonWrite_clicked()
{
    startWrite();
}

void MainWindow::deviceTempChanged(double temp)
{
    QString tempStr = QString("%1 %2C").arg(temp, 0, 'f', 1).arg(QChar(0x02da));
    ui->tableWidget->item(0, COL_TEMP)->setText(tempStr);
    updateTableSize();
}

void MainWindow::deviceStatusUpdated(QString s)
{
    ui->tableWidget->item(0, COL_STATE)->setText(s);
    updateTableSize();
}

void MainWindow::progressFormatChanged(QString format)
{
    QProgressBar *bar = qobject_cast<QProgressBar*>(ui->tableWidget->cellWidget(0, COL_PROGRESS));
    if(bar) {
        bar->setFormat(format);
        updateTableSize();
    }
}

void MainWindow::progressUpdated(int p)
{
    QProgressBar *bar = qobject_cast<QProgressBar*>(ui->tableWidget->cellWidget(0, COL_PROGRESS));
    if(bar) {
        if(bar->value() != p) {
            bar->setValue(p);
            updateTableSize();
        }
    }
}

void MainWindow::taskFinished(uint id, bool isWrite, bool ok, bool isGoldenImage)
{
    qInfo() << QString("%1 at %2 writeId:%3 status:%4")
               .arg(isWrite ? "dataWrote" : "dataVerified")
               .arg(pexmlSerial, 8, 16, QChar('0')).arg(id).arg(ok?"success":"fail");
    if(ok) {
        devTaskId = id;
        if(isWrite)
            dataBase->writeInfo(pexmlSerial, isGoldenImage, config.filePath);
    } else {
        devTaskId = -1;
    }

    updateTableSize();

    if(batchMode)
        close();
}

void MainWindow::startWrite(bool isGoldenImage)
{
    startTask(true, isGoldenImage);
}

void MainWindow::startVerify(bool isGoldenImage)
{
    startTask(false, isGoldenImage);
}

void MainWindow::startTask(bool isWrite, bool isGoldenImage)
{
    const QString taskStr = isWrite ? "write": "verify";
    qint64 taskTimeout = isWrite ? 5*60*1000 : 2*60*1000;
    if(device == nullptr) {
        qWarning() << QString("Can't %1. Device is disconnected").arg(taskStr);
        return;
    }
    qInfo() << "\n=== start ===";
    qInfo() << QString("%1 %2 image %3").arg(QDateTime::currentDateTime().toString())
               .arg(isGoldenImage?"golden": "second").arg(taskStr);
    setPushButtonEnabled(false);

    // Read data from file
    if(!readFile())
        return;
    bool allOk=true;

    ++taskId;
    if(taskId==0 || taskId>1000)
        taskId=100;
    devTaskId = 0;

    if(isWrite)
        emit startDevWrite(taskId, fileData, fileHeaderSize, isGoldenImage, config.verifyAfterWrite);
    else
        emit startDevVerify(taskId, fileData, fileHeaderSize, isGoldenImage);


    QElapsedTimer taskTimer;
    taskTimer.start();
    bool finished = false;
    bool someFailed = false;
    while(!finished) {
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);
        if(taskTimer.elapsed() > taskTimeout) {
            qWarning() << QString("%1 timeouted").arg(taskStr);
            break;
        }
        finished = devTaskId == taskId;
        if(devTaskId == -1) {
            finished = true;
            allOk = false;
        }
    }
    allOk &= finished && !someFailed;

    qInfo() << "==="<<(allOk?"end":"fail")<<"===";
    lastStatusLabel->setText(QString("%1 (%2)")
                             .arg(allOk?QString("ok in %1 sec").arg(taskTimer.elapsed()/1000):"fail")
                             .arg(QTime::currentTime().toString()));
    lastStatusLabel->setStyleSheet(QString("QLabel { color : %1; }").arg(allOk?"green":"red"));
    updateTableSize();
    setPushButtonEnabled(true);
}

void MainWindow::initTable()
{
    QTableWidget *t = ui->tableWidget;
    t->verticalHeader()->hide();
    t->setColumnCount(COL_NUM);
    t->setHorizontalHeaderItem(COL_SERIAL, new QTableWidgetItem("Serial"));
    t->setHorizontalHeaderItem(COL_NAME, new QTableWidgetItem("Device"));
    t->setHorizontalHeaderItem(COL_STATE, new QTableWidgetItem("State"));
    t->setHorizontalHeaderItem(COL_DRV_FW, new QTableWidgetItem("Driver ver."));
    t->setHorizontalHeaderItem(COL_DEV_FW, new QTableWidgetItem("Firmware"));
    t->setHorizontalHeaderItem(COL_TEMP, new QTableWidgetItem("Temp"));
    t->setHorizontalHeaderItem(COL_PROGRESS, new QTableWidgetItem("Progress"));
    t->hideColumn(COL_STATE);
    t->hideColumn(COL_TEMP);

    t->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    t->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize();
}

void MainWindow::updateTableSize()
{
    QTableWidget *table = ui->tableWidget;
    table->resizeRowsToContents();
     table->resizeColumnsToContents();

     int width = table->verticalHeader()->width()+5;
     for(int col=0; col<table->columnCount(); ++col) {
         if(table->isColumnHidden(col))
             continue;
         width += table->columnWidth(col);
     }
     if(width<20)
         width = 20;

     int height = table->horizontalHeader()->height()+5;
     for(int row=0; row<table->rowCount(); ++row) {
         if(table->isRowHidden(row))
             continue;
         height += table->rowHeight(row);
     }
     if(height<20)
         height = 20;
     table->setFixedSize(width, height);
     table->updateGeometry();
}

void MainWindow::addDevice(int interfaceIndex)
{
    if(device)
        disconnectFromAll();
    Q_ASSERT(device==nullptr);
    Q_ASSERT(devThread==nullptr);

    QTableWidget *table = ui->tableWidget;
    Q_ASSERT(table->rowCount() == 0);
    int row = -1;
    if(row==-1) {
        row = table->rowCount();
        table->setRowCount(row+1);
        table->setItem(row, COL_SERIAL, new QTableWidgetItem());
        table->setItem(row, COL_NAME, new QTableWidgetItem());
        table->setItem(row, COL_DRV_FW, new QTableWidgetItem());
        table->setItem(row, COL_DEV_FW, new QTableWidgetItem());
        table->setItem(row, COL_STATE, new QTableWidgetItem());
        table->setItem(row, COL_TEMP, new QTableWidgetItem());
        QProgressBar *bar = new QProgressBar(table);
        table->setCellWidget(row, COL_PROGRESS, bar);
        bar->setValue(0);
    }

    device = new PexmlFlashDev(nullptr, interfaceIndex);
    QString devName = QString::fromStdString(device->getDevName());
    if(!device->isOpened()) {
        qCritical() << "failed to open device:"<<devName;
        delete device;
        device = nullptr;
        return;
    }

    pexmlSerial = device->getDevSerial();
    QString drvFwStr = device->getDrvFwStr();
    QString devFwStr = device->getDevFwStr();
    devThread = new QThread(this);
    device->moveToThread(devThread);
    devThread->start();
    devThread->setObjectName(devName);

    connect(device, SIGNAL(setStatus(QString)), SLOT(deviceStatusUpdated(QString)));
    connect(device, SIGNAL(TempChanged(double)), SLOT(deviceTempChanged(double)));
    connect(this, SIGNAL(closeDevices()), device, SLOT(deleteLater()));
    connect(device, SIGNAL(destroyed(QObject*)), devThread, SLOT(quit()));

    connect(this, SIGNAL(startDevWrite(uint,QByteArray,int,bool,bool)),
            device, SLOT(write(uint,QByteArray,int,bool,bool)));
    connect(this, SIGNAL(startDevVerify(uint,QByteArray,int,bool)),
            device, SLOT(verify(uint,QByteArray,int,bool)));
    connect(device, SIGNAL(progressFormatChanged(QString)),
            SLOT(progressFormatChanged(QString)));
    connect(device, SIGNAL(progressUpdated(int)), SLOT(progressUpdated(int)));
    connect(device, SIGNAL(taskFinished(uint,bool,bool,bool)),
            SLOT(taskFinished(uint,bool,bool,bool)));

    table->item(0, COL_SERIAL)->setText(QString("%1-%2")
                                        .arg(pexmlSerial>>16, 4, 16, QChar('0'))
                                        .arg(pexmlSerial&0xFFFF, 4, 16, QChar('0')));
    table->item(0, COL_DRV_FW)->setText(drvFwStr);
    table->item(0, COL_DEV_FW)->setText(devFwStr);
    table->item(0, COL_NAME)->setText(devName);

    qInfo() << QString("Added pexml device:%1").arg(devName);
    updateTableSize();
}

void MainWindow::setPushButtonEnabled(bool enabled) const
{
    ui->pushButtonFile->setEnabled(enabled);
    ui->pushButtonWrite->setEnabled(enabled);
    ui->actionWrite_Golden_Image->setEnabled(enabled);
    ui->actionVerify_Golden_image->setEnabled(enabled);
    ui->actionVerify_regular_image->setEnabled(enabled);

    if(enabled) {
        updateWriteButtonState();
    } else {
        ui->pushButtonWrite->setEnabled(enabled);
        ui->actionWrite_Golden_Image->setEnabled(enabled);
    }
}

void MainWindow::disconnectFromAll()
{
    if(!device)
        return;

    emit closeDevices();

    QElapsedTimer timer;
    timer.start();

    while(true) {
        if(timer.elapsed()>2*60e3){
            qWarning() << "Timeout in disconnectFromAll";
            break;
        }
        QCoreApplication::processEvents(QEventLoop::ExcludeUserInputEvents, 100);

        if(devThread->isFinished()) {
            break;
        }
    }

    delete devThread;
    pexmlSerial = 0;
    devThread = nullptr;
    device = nullptr;
    ui->tableWidget->setRowCount(0);
}

void MainWindow::readConfig()
{
    config.loadSettings();

    QSettings settings;
    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
    settings.endGroup();
}

void MainWindow::saveConfig()
{
    config.saveSettings();

    QSettings settings;
    settings.beginGroup("MainWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();
}

void MainWindow::updateWriteButtonState() const
{
    bool enable = true;
    if(enable) {
        QFile file(config.filePath);
        enable = file.open(QIODevice::ReadOnly);
        file.close();
    }
    if(enable) {
        enable = fileHeaderSize!=-1;
    }
    ui->pushButtonWrite->setEnabled(enable);
    ui->actionWrite_Golden_Image->setEnabled(enable);
}

bool MainWindow::readFile()
{
    ui->labelFile->setStyleSheet("QLabel { color : black; }");
    QFile file(config.filePath);
    QFileInfo fileInfo(file);
    if(fileInfo.lastModified() == fileModified && fileInfo.size() == fileSize
            && fileData.size() == fileSize
            && fileHeaderSize>0
            && fileSize>fileHeaderSize) {
        qInfo() << "File wasn't modified. Use cached file data.";
        return true;
    }
    if(!fileInfo.exists()
            && fileData.size() == fileSize
            && fileHeaderSize>0
            && fileSize>fileHeaderSize
            ) {
        qWarning() << "File is no longer available, but we have a cache.";
        return true;
    }

    fileModified = fileInfo.lastModified();
    fileSize = fileInfo.size();
    fileHeaderSize = -1;

    if(!file.open(QIODevice::ReadOnly)){
            qWarning() << "Can't open bit file";
            return false;
    }
    fileData = file.readAll();
    file.close();
    if(fileData.isEmpty()) {
        qWarning() << "Read file is empty";
        return false;
    }
    if(fileData.size() != fileSize) {
        qWarning() << "data.size() != fileSize";
        return false;
    }

    qInfo() << QString("Read bit file '%1' (%2 bytes)").arg(config.filePath).arg(fileData.size());

    fileHeaderSize = FlashDev::findBitFileHeader(fileData.data(), fileData.size());
    ui->labelFile->setStyleSheet(QString("QLabel { color : %1; }")
                                 .arg((fileHeaderSize!=-1)?"green":"red"));
    if(fileHeaderSize == -1) {
        qWarning() << "Can't find header size of bit-file";
        return false;
    }

    qInfo() << "File header takes"<<fileHeaderSize<<"bytes of bit-file";
    return true;
}
