#include "PexmlFlashDev.h"

#include <QDebug>
#include <iostream>

#include "mlink/MlinkDeviceInterfaceDebug.h"
#include "mlink/regiomlink.h"

namespace {
//const int REV_PEXML_READOUT_SERIAL = 25674;
const int REV_FLASH_ID = 27846;
}


PexmlFlashDev::PexmlFlashDev(QObject * parent, int interfaceIndex):
    QObject(parent),
    mlinkDevice(new mlink::MlinkDeviceInterfaceDebug),
    isOpen(false),
    devSerial(0),
    drvVer(0)
{
    if(interfaceIndex==-1)
        devName = "";
    else
        devName = std::string("/dev/")+std::string(ML_DEVICE_NAME)+std::to_string(interfaceIndex);
    if (int ret = mlinkDevice->ml_open(devName.c_str())) {
        QString msg = QString("Failed to open device %1: %2")
                          .arg(devName.c_str())
                          .arg(strerror(-ret));
        const std::string str = msg.toStdString();
        qFatal("%s", str.c_str());
    } else {
        isOpen = true;
    }

    if(isOpen) {
        uint64_t data64 = 0;
        unsigned long data32 = 0;
        int ret;
        ret = mlinkDevice->ml_get_driver_version(&data32);
        if(ret==0)
            drvVer = data32;
        ret = mlinkDevice->ml_read_board_info(&board_info);
        if(ret==0) {
            devSerial = board_info.board_serial;
        }
        if(!devSerial) {
            ret = mlinkDevice->ml_get_board_serial_id(&data64);
            if (ret) {
                mlinkDevice->ml_close();
                isOpen = false;
                QString msg = QString("PEXML driver %1 does not support ML_IOC_GET_SERIAL_ID")
                                  .arg(drvVer);
                const std::string str = msg.toStdString();
                qFatal("%s", str.c_str());
            }
            devSerial = data64;
        }
        bool ok = configureFlash();
        if(!ok) {
            mlinkDevice->ml_close();
            isOpen = false;
            qFatal("Failed to configure flash.");
        }
    }
}

PexmlFlashDev::~PexmlFlashDev()
{
    if(isOpen)
        mlinkDevice->ml_close();
}

QString PexmlFlashDev::getDrvFwStr() const
{
    if(drvVer==0)
        return "unknown";
    return QString("%1.%2").arg(drvVer>>16).arg(drvVer&0xFFFF);
}

QString PexmlFlashDev::getDevFwStr() const
{
    if(board_info.board_serial)
        return QString("%1.%2:%3")
                .arg(board_info.fw_version.fw_ver_maj)
                .arg(board_info.fw_version.fw_ver_min)
                .arg(board_info.fw_version.fw_rev);
    else
        return "unknown";
}

void PexmlFlashDev::write(uint id, QByteArray fileData, int headerSize, bool isGoldenImage, bool verify)
{
    bool ok = startWrite(id, fileData.constData(), fileData.size(), headerSize, isGoldenImage, verify);
    emit taskFinished(id, true, ok, isGoldenImage);
}

void PexmlFlashDev::verify(uint id, QByteArray fileData, int headerSize, bool isGoldenImage)
{
    fileHeaderSize = headerSize;
    fullSize = fileData.size();
    bool ok = verifyFlash(isGoldenImage, fullSize, fileData.constData());
    emit taskFinished(id, false, ok, isGoldenImage);
}

bool PexmlFlashDev::sectorErase(bool isGoldenImage, uint32_t sizeToErase)
{
    emit progressFormatChanged("Erased %p%");
    return FlashDev::sectorErase(isGoldenImage, sizeToErase);
}

bool PexmlFlashDev::writeFlash(const char *fileData, bool isGoldenImage)
{
    bool ret = true;

    emit progressFormatChanged("Wrote %p%");
    ret = FlashDev::writeFlash(fileData, isGoldenImage);
    return ret;
}

bool PexmlFlashDev::verifyFlash(bool isGoldenImage, uint sizeToRead, const char *data)
{
    emit progressFormatChanged("Verify %p%");
    return FlashDev::verifyFlash(isGoldenImage, sizeToRead, data);
}

quint16 PexmlFlashDev::devReadReg(int regAddr)
{
    unsigned int data32 = 0;
    int ret = mlinkDevice->ml_read_flasher(regAddr, &data32);
    if(ret != 0) {
        qWarning() << __func__ << ":" << strerror(-ret);
    }
    return data32;
}

void PexmlFlashDev::devWriteReg(int regAddr, uint16_t data)
{
    int ret = mlinkDevice->ml_write_flasher(regAddr, data);
    if(ret != 0) {
        qWarning() << __func__ << ":" << strerror(-ret);
    }
}

void PexmlFlashDev::devWriteReg32(int regAddr, uint32_t data)
{
    devWriteReg(regAddr, data&0xFFFF);
    devWriteReg(regAddr+1, data>>16);
}


bool PexmlFlashDev::devBlkWriteReg(uint count, const uint *addr, uint16_t *data)
{
    for(uint i=0; i<count; ++i) {
        devWriteReg(addr[i], data[i]);
    }
    return true;
}

bool PexmlFlashDev::devBlkReadReg(uint count, const uint *addr, uint16_t *data)
{
    for(uint i=0; i<count; ++i) {
        data[i] = devReadReg(addr[i]);
    }
    return true;
}

int PexmlFlashDev::getMaxBlockSize() const
{
    return 1;
}

void PexmlFlashDev::sendProgress(int p)
{
    std::string s;
    s.append(std::to_string(p)).append(" %").append(p==100?"\n":"\r");
    std::printf("%s", s.c_str());
    std::cout<<std::flush;

    emit progressUpdated(p);
}

std::string PexmlFlashDev::getPrefix() const
{
    return QString("Pexml device(%1|%2)")
            .arg(devName.c_str()).arg(devSerial, 8, 16, QChar('0'))
            .toStdString();
}

uint PexmlFlashDev::getRegBase() const
{
    return 0x0;
}

bool PexmlFlashDev::canReadFlashId() const
{
    return board_info.fw_version.fw_rev >= REV_FLASH_ID;
//    return board_info.fw_version.fw_ver_maj > 3 ||
//            (board_info.fw_version.fw_ver_maj == 3 &&
//             board_info.fw_version.fw_ver_min >= 37);
}
