/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "BaseMainWindow.h"

#include <QCloseEvent>
#include <QCoreApplication>
#include <QDebug>
#include <QMenuBar>
#include <QSettings>
#include <QTimer>

#include "AboutBox.h"
#include "base/SyslogMessageSender.h"

BaseMainWindow::BaseMainWindow(QWidget *parent) :
    QMainWindow(parent)
{
    // setupMenu();

    // log message may be sent before inherited class constructor has finished,
    // should use QueuedConnection
    connect(&SyslogMessageSender::instance(), &SyslogMessageSender::newMsg,
            this, &BaseMainWindow::addLog,
            Qt::QueuedConnection);
    connect(&SyslogMessageSender::instance(), &SyslogMessageSender::logMessage,
            this, &BaseMainWindow::addLogMessage,
            Qt::QueuedConnection);

    QTimer::singleShot(0, this, &BaseMainWindow::delayedInit);
}

BaseMainWindow::~BaseMainWindow()
{
    SyslogMessageSender::instance().disconnect(this);
}

void BaseMainWindow::delayedInit()
{
    setWindowTitle(QCoreApplication::applicationName());
    loadMainWindowPosition();
}

void BaseMainWindow::closeEvent(QCloseEvent *event)
{
    saveMainWindowPosition();
    QMainWindow::closeEvent(event);
}

void BaseMainWindow::setupMenu(QMenu *file, QMenu *help, QMenu *options)
{
    menuFile = file;
    menuHelp = help;
    menuOptions = options;

    if (!menuFile)
        menuFile = menuBar()->addMenu(tr("&File"));

    QIcon iconQuit = QIcon::fromTheme("application-exit");
    menuFile->addAction(iconQuit,
                        tr("E&xit"),
                        this,
                        &BaseMainWindow::actionQuit,
                        QKeySequence::Quit);

    if (!menuHelp)
        menuHelp = menuBar()->addMenu(tr("&Help"));

    QIcon iconAbout = QIcon::fromTheme("help-about");
    menuHelp->addAction(iconAbout, tr("About..."), this, &BaseMainWindow::actionAbout);

    menuBar()->setVisible(true);
}

void BaseMainWindow::actionQuit()
{
    close();
}

void BaseMainWindow::actionAbout()
{
    AboutBox::about();
}

void BaseMainWindow::addLog(QtMsgType severity, const QString &log, const QString &thrName)
{
    Q_UNUSED(severity)
    Q_UNUSED(log)
    Q_UNUSED(thrName)
}

void BaseMainWindow::addLogMessage(QtMsgType severity, const QString &category, const QString &log, const QString &thrName)
{
    Q_UNUSED(severity)
    Q_UNUSED(category)
    Q_UNUSED(log)
    Q_UNUSED(thrName)
}

void BaseMainWindow::loadMainWindowPosition()
{
    QSettings settings;
    restoreGeometry(settings.value(SETTING_GROUP_MAINWINDOW + "/geometry")
                    .toByteArray());
    restoreState(settings.value(SETTING_GROUP_MAINWINDOW + "/windowState")
                 .toByteArray());
    loadAppGuiSettings(settings);
}

void BaseMainWindow::saveMainWindowPosition() const
{
    QSettings settings;
    settings.beginGroup(SETTING_GROUP_MAINWINDOW);
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    settings.endGroup();
    saveAppGuiSettings(settings);
}
