#include "RunGuard.h"

#include <QCryptographicHash>

#include "base/DaqModule.h"
#include "base/ProgramIndex.h"
#include "base/ConfigSelector.h"

namespace
{

QString generateKeyHash( const QString& key, const QString& salt )
{
    QByteArray data;

    data.append( key.toUtf8() );
    data.append( salt.toUtf8() );
    data = QCryptographicHash::hash( data, QCryptographicHash::Sha1 ).toHex();

    return data;
}

}

RunGuard::RunGuard(const DaqModule &daqModule, const QString &program_index)
    : RunGuard(QString("%1_%2").arg(daqModule.toString(), program_index))
{}

RunGuard::RunGuard( const QString& key )
    : key( key )
    , memLockKey( generateKeyHash( key, "_memLockKey" ) )
    , sharedmemKey( generateKeyHash( key, "_sharedmemKey" ) )
    , sharedMem( sharedmemKey )
    , memLock( memLockKey, 1 )
{
    memLock.acquire();
    {
        QSharedMemory fix( sharedmemKey );    // Fix for *nix: http://habrahabr.ru/post/173281/
        fix.attach();
    }
    memLock.release();
}

RunGuard::~RunGuard()
{
    release();
}

bool RunGuard::isAnotherRunning()
{
    if ( sharedMem.isAttached() )
        return false;

    memLock.acquire();
    const bool isRunning = sharedMem.attach();
    if ( isRunning )
        sharedMem.detach();
    memLock.release();

    return isRunning;
}

bool RunGuard::tryToRun()
{
    if ( isAnotherRunning() )   // Extra check
        return false;

    memLock.acquire();
    const bool result = sharedMem.create( sizeof( quint64 ) );
    memLock.release();
    if ( !result )
    {
        release();
        return false;
    }

    return true;
}

void RunGuard::release()
{
    memLock.acquire();
    if ( sharedMem.isAttached() )
        sharedMem.detach();
    memLock.release();
}

void RunGuard::runSingleOrExit(const ConfigSelector &cs)
{
    RunGuard::runSingleOrExit(cs.getType(), cs.getIndex());
}

void RunGuard::runSingleOrExit(const DaqModule &daqModule, const ProgramIndex &program_index)
{
    RunGuard::runSingleOrExit(daqModule, program_index.toString());
}

void RunGuard::runSingleOrExit(const DaqModule &daqModule, const QString &program_index)
{
    static RunGuard runGuard(daqModule, program_index);
    if (!runGuard.tryToRun()) {
        QString errStr = daqModule.toString() + " is already running; ";
        if (program_index.isEmpty())
            errStr += "in local run mode.";
        else
            errStr += QString("with run_index '%1'").arg(program_index);
        qFatal("%s", errStr.toStdString().c_str());
    }
}
