/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef ABSTRACTCONFIGCONVERTER_H
#define ABSTRACTCONFIGCONVERTER_H

#include <QJsonObject>
#include <QMetaType>
#include <QVariantMap>

template<class T>
class AbstractConfigConverter
{
public:
    static void registerTypes()
    {
        qRegisterMetaType<T>();
        registerSubTypes();
        if (classIsObject()) {
            if (! QMetaType::hasRegisteredConverterFunction<QJsonObject, T>()) {
                // run at least once, no thread-safety issue
                QMetaType::registerConverter<T, QJsonObject>(&toJsonObject);
                QMetaType::registerConverter<QJsonObject, T>(&fromJsonObject);
            }
        }
        if (! QMetaType::hasRegisteredConverterFunction<QJsonValue, T>()) {
            QMetaType::registerConverter<T, QJsonValue>(&toJsonValue);
            QMetaType::registerConverter<QJsonValue, T>(&fromJsonValue);
        }
    }

private:
    static void registerSubTypes();
    static bool classIsObject();

    static QJsonObject toJsonObject(const T &in);
    static T fromJsonObject(const QJsonObject &in);
    static QJsonValue toJsonValue(const T &in);
    static T fromJsonValue(const QJsonValue &in);
    static QJsonObject fromVariantMap(const QVariantMap &map) {
        QJsonObject obj;
        auto it = map.cbegin();
        while (it!=map.cend()) {
            obj[it.key()] = it.value().value<QJsonValue>();
            ++it;
        }
        return obj;
    }
    template<typename K>
    static bool update_value(const QJsonObject &in, const QString &key, K *val) {
        auto it = in.find(key);
        if(it==in.end())
            return false;
        Q_ASSERT(val);
        auto var = QVariant::fromValue(it.value());
        Q_ASSERT(var.canConvert<K>());
        *val = var.value<K>();
        return true;
    }

};

/*
namespace {
constexpr const char *k_zzzz = "zzzz";
}

template<>
bool AbstractConfigConverter<XXXX>::classIsObject()
{
    return true;
}

template<>
QJsonObject AbstractConfigConverter<XXXX>::toJsonObject(const XXXX &in)
{
    QVariantMap m;
    m[k_zzzz] = QVariant::fromValue(in.zzzz);
    return fromVariantMap(m);
}

template<>
XXXX AbstractConfigConverter<XXXX>::fromJsonObject(const QJsonObject &in)
{
    XXXX out;
    out.value = QVariant::fromValue(in.value).value<ValueType>();
    update_value(in, k_zzzz, &out.zzzz);
    return out;
}

template<>
QJsonValue AbstractConfigConverter<XXXX>::toJsonValue(const XXXX &in)
{
    return toJsonObject(in);
}

template<>
XXXX AbstractConfigConverter<XXXX>::fromJsonValue(const QJsonValue &in)
{
    return fromJsonObject(in.toObject());
}

template<>
void AbstractConfigConverter<XXXX>::registerSubTypes()
{
    CoreTypeConverter::registerTypes();
    AbstractConfigConverter<ValueType>::registerTypes();
}
*/

#endif // ABSTRACTCONFIGCONVERTER_H
