/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef DEVICEINDEX_H
#define DEVICEINDEX_H

#include <QHash>
#include <QMetaType>
#include <QSet>

#include "base/DeviceId.h"

class QDataStream;
class QDebug;

class DeviceIndex
{
public:
    friend class DeviceIndexConverter;

    DeviceIndex() = default;
    DeviceIndex(const DeviceIndex &index) = default;

    DeviceIndex(DeviceId device_id, quint64 serial_id, quint64 ipAddressHint = 0)
        : device_id(device_id),
        serial_id(serial_id),
        ipAddressHint(ipAddressHint) { }

    DeviceIndex &operator=(const DeviceIndex&) = default;

    friend bool operator ==(const DeviceIndex &l, const DeviceIndex &r);
    friend bool operator!=(const DeviceIndex &l, const DeviceIndex &r) { return !(l == r); }
    friend bool operator<(const DeviceIndex &l, const DeviceIndex &r);

    [[deprecated("Use = DeviceIndex()")]]
    void clear() { *this = DeviceIndex(); }
    bool isNull() const { return device_id==0 || serial_id==0; }

    inline DeviceId getDevId() const { return device_id; }
    inline quint64 getSerial() const { return serial_id; }

    QString getIdent() const;
    QString getDeviceIdName() const;
    QString getSerialStr() const;
    static QString getDeviceIdName(DeviceId device_id);
    static QString getSerialStr(quint64 serial_id);
    QString toString() const;
    static DeviceIndex fromString(const QString &r);

    friend uint qHash(const DeviceIndex &r, uint seed = 0) {
        return qHash(r.device_id ^ (r.serial_id << 8) ^ seed);
    }

    friend QDebug operator<<(QDebug dbg, const DeviceIndex &r);
    friend QDataStream &operator<<(QDataStream &out, const DeviceIndex &value);
    friend QDataStream &operator>>(QDataStream &in, DeviceIndex &value);

private:
    DeviceId device_id = DEVICE_ID_UNKNOWN;
    quint64 serial_id = 0;
    quint64 ipAddressHint = 0;
};

Q_DECLARE_METATYPE(DeviceIndex)

typedef QSet<DeviceIndex> DeviceIndexSet;
Q_DECLARE_METATYPE(DeviceIndexSet)

typedef QSet<DeviceId> DeviceIdSet;
Q_DECLARE_METATYPE(DeviceIdSet)

// required for Qt < 5.14
DeviceIndexSet makeDeviceIndexSet(const QList<DeviceIndex> &list);
DeviceIdSet makeDeviceIdSet(const QList<DeviceId> &list);

#endif // DEVICEINDEX_H
