#include "BaseFsm.h"

#include <QDebug>
#include <QLoggingCategory>
#include <QTimer>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm")
}

BaseFsm::BaseFsm(QObject *parent) : QObject(parent)
{
    qRegisterMetaType<FsmState>();
    auto t = new QTimer(this);
    connect(t, &QTimer::timeout, this, &BaseFsm::evaluate);
    t->start(100);
    setObjectName("BaseFSM");
    recreateInput(new BaseFsmInput);
}

void BaseFsm::goState(const FsmState &nextState)
{
    if (nextState == FsmState::Init || state == nextState)
        return;
    qDebug(fsm) << objectName() << "Request new state:" << nextState;
    input->requestState = nextState;
    evaluate();
}

void BaseFsm::printFsmInput() const
{
    auto qinfo = qInfo().noquote().nospace();
    qinfo << QString("[%1] ").arg(objectName());
    printInputString(qinfo);
}

bool BaseFsm::evaluate()
{
    if(input->requestState == FsmState::Error) {
        acceptRequestedState();
        return true;
    }
    if(stateTimeout()) {
        auto dbg = qDebug(fsm) << objectName() << "Timeout error. Input:";
        printInputString(dbg);
        setState(FsmState::Error);
        return true;
    }

    switch (state) {
    case FsmState::Init:
        setState(FsmState::Error);
        return true;
    case FsmState::Run:
        if (input->requestState == FsmState::Stopping) {
            acceptRequestedState();
            return true;
        }
        break;
    default:
    {}
    }
    return false;
}

void BaseFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 BaseFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Starting: return 3000;
    case FsmState::Stopping: return 3000;
    default: return -1;
    }
}

void BaseFsm::setState(const FsmState &nextState)
{
    if (nextState == state)
        return;
    qInfo(fsm) << objectName() << "state change " << state << " to " << nextState;
    qDebug(fsm) << objectName() << "state time:" << stateTime.elapsed() << (stateTimeout() ? "TIMEOUT" : "");
    auto dbg = qDebug(fsm) << objectName() << " input:";
    printInputString(dbg);

    const auto oldState = state;
    state = nextState;
    stateTime.restart();
    if(state == FsmState::Error && input->requestState != FsmState::Init) {
        qWarning(fsm) << objectName() << "RejectOnError requested state " << input->requestState;
        input->clearRequests();
    }
    emit stateChanged(state, oldState);
}

void BaseFsm::acceptRequestedState()
{
    const auto nextState = input->requestState;
    input->clearRequests();
    setState(nextState);
}

void BaseFsm::recreateInput(BaseFsmInput *p)
{
    input = p;
}

bool BaseFsm::stateTimeout() const
{
    const auto &timeout = stateTimeoutMs(state);
    if(timeout == -1)
        return false;
    const bool timeouted = stateTime.hasExpired(timeout);
    if(timeouted)
        qDebug(fsm) << objectName() << "Got timeout at state:" << state;
    return timeouted;
}

QDebug operator<<(QDebug dbg, const BaseFsmInput &r)
{
    dbg << "requestState:";
    if (r.requestState == FsmState::Init)
        dbg << "(none)";
    else
        dbg << r.requestState;

    dbg << "; ";
    return dbg;
}
