#include "DeviceMgrFsm.h"

#include "base/DeviceIndexEnabled.h"

#include <QDebug>
#include <QLoggingCategory>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.device")
}

DeviceMgrFsm::DeviceMgrFsm(QObject *parent):
    BaseFsm(parent)
{
    setObjectName("DevMgrFSM");
    recreateInput(new DeviceMgrFsmInput());
}

void DeviceMgrFsm::setDeviceList(const DeviceIndexEnabledMap &newList)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << newList;
    getInput()->devSet = newList.getEnabled();
    evaluate();
}

void DeviceMgrFsm::setMainSetupKey(int key)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << key;
    getInput()->setupKey = key;
    getInput()->lastSetupKeyUpdate.start();
    evaluate();
}

void DeviceMgrFsm::setDevSetupKey(const DeviceIndex &index, int key)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << index << key;
    getInput()->deviceSetupKeys[index] = key;
    evaluate();
}

bool DeviceMgrFsm::evaluate()
{
    if(BaseFsm::evaluate())
        return true;

    switch (getState()) {
    case FsmState::Error:
        if (allDevicesWritten()) {
            setState(FsmState::Ready);
            return true;
        }
        break;
    case FsmState::Ready:
        if (allDevicesWritten()) {
            if (getRequestState() == FsmState::Starting) {
                acceptRequestedState();
                return true;
            }
        } else if(setupKeyGuargExpired()) {
            qDebug(fsm) << __PRETTY_FUNCTION__ << "Go to Error state (some dev cfg is invalid)";
            setState(FsmState::Error);
            return true;
        }
        break;
    case FsmState::Starting:
        if(allDevicesWritten() && getRequestState() == FsmState::Run) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::Run:
        if (!allDevicesWritten() && setupKeyGuargExpired()) {
//            setState(FsmState::Stopping); // stop on error
            qDebug(fsm) << objectName() << "Go to Error state (some dev cfg is invalid)";
            setState(FsmState::Error); // stop on error
            return true;
        }
        if (getRequestState() == FsmState::Stopping) {
            acceptRequestedState();
            return true;
        }
        break;
    case FsmState::Stopping:
        if(getRequestState() == FsmState::Ready
                && allDevicesWritten()) {
            acceptRequestedState();
            return true;
        }
        break;
    default:
    {}
    }

    return false;
}

void DeviceMgrFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 DeviceMgrFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
    case FsmState::Starting:
    case FsmState::Stopping:
        return DEVICE_CONFIG_WRITE_TIMEOUT;
    default:
        return BaseFsm::stateTimeoutMs(state);
    }
}

bool DeviceMgrFsm::allDevicesWritten() const
{
    return checkDevKey(getInput()->setupKey, getInput()->deviceSetupKeys);
}

bool DeviceMgrFsm::setupKeyGuargExpired() const
{
    const auto &t = getInput()->lastSetupKeyUpdate;
    return !t.isValid() || t.elapsed() > DEVICE_CONFIG_WRITE_TIMEOUT;
}

bool DeviceMgrFsm::checkDevKey(int mainKey, const QMap<DeviceIndex, int> &keyMap) const
{
    for(auto key : getInput()->devSet)
        if(keyMap.value(key, 0) != mainKey)
            return false;
    return true;
}

QDebug operator<<(QDebug dbg, const DeviceMgrFsmInput &in)
{
    dbg << BaseFsmInput(in);

    dbg << "setupKey:" << in.setupKey << "; ";
    dbg << "devSet:" << in.devSet << "; ";
    dbg << "deviceSetupKeys:" << in.deviceSetupKeys << "; ";

    return dbg;
}
