/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MergeColumnsTableModel.h"

#include <QDebug>

MergeColumnsTableModel::MergeColumnsTableModel(QObject *parent)
    : QAbstractTableModel(parent)
{

}

int MergeColumnsTableModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    auto i = std::max((modelA ? modelA->rowCount() : 0),
                    (modelB ? modelB->rowCount() : 0));
    return i;
}

int MergeColumnsTableModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return
            (modelA ? modelA->columnCount() : 0) +
            (modelB ? modelB->columnCount() : 0);
}

QModelIndex MergeColumnsTableModel::mapFromSource(const QModelIndex &source) const
{
    if (source.model() == modelA)
        return index(source.row(), source.column());
    if (source.model() == modelB) {
        int columnOffsetB = modelA ? modelA->columnCount() : 0;
        return index(source.row(), source.column() + columnOffsetB);
    }
    return QModelIndex();
}

QModelIndex MergeColumnsTableModel::mapToSource(const QModelIndex & proxy) const
{
    int columnOffsetB = modelA ? modelA->columnCount() : 0;
    int row = proxy.row();
    int column = proxy.column();
    if (proxy.column() < columnOffsetB) {
        return modelA ?
                    modelA->index(row, column) :
                    QModelIndex();
    } else {
        return modelB ?
                    modelB->index(row, column - columnOffsetB) :
                    QModelIndex();
    }
}

QVariant MergeColumnsTableModel::data(const QModelIndex &index, int role) const
{
    return mapToSource(index).data(role);
}

QVariant MergeColumnsTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Vertical) {
        return modelA ? modelA->headerData(section, orientation, role) :
                        modelB ? modelB->headerData(section, orientation, role) :
                                 QVariant();
    }
    int columnOffsetB = modelA ? modelA->columnCount() : 0;

    if (section < columnOffsetB) {
        if (!modelA)
            return QVariant();
        return modelA->headerData(section, orientation, role);
    } else {
        if (!modelB)
            return QVariant();
        return modelB->headerData(section - columnOffsetB, orientation, role);
    }
}

void MergeColumnsTableModel::connectSourceModel(QAbstractTableModel *model)
{
    connect(model, &QAbstractItemModel::dataChanged,
            this, &MergeColumnsTableModel::onDataChanged);
    connect(model, &QAbstractItemModel::modelReset,
            this, &MergeColumnsTableModel::sourceModelReset);
}

void MergeColumnsTableModel::setSourceModelA(QAbstractTableModel *modelA)
{
    beginResetModel();
    this->modelA = modelA;
    connectSourceModel(modelA);
    endResetModel();
}

void MergeColumnsTableModel::setSourceModelB(QAbstractTableModel *modelB)
{
    beginResetModel();
    this->modelB = modelB;
    connectSourceModel(modelB);
    endResetModel();
}

void MergeColumnsTableModel::onDataChanged(const QModelIndex &topLeft,
                                           const QModelIndex &bottomRight,
                   const QVector<int> &roles)
{
    qDebug() << sender()->objectName() << topLeft << bottomRight << roles;
}

void MergeColumnsTableModel::sourceModelReset()
{
    emit modelReset();
}
