/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "LldpInfo.h"

#include <QHostAddress>
#include <QStringList>

#include "base/DeviceIndex.h"

bool LldpInfo::operator==(const LldpInfo &r) const
{
    return true &&
           haddr == r.haddr &&
           ip_addr == r.ip_addr &&
           ttl == r.ttl &&
           portID == r.portID &&
           hw_str == r.hw_str &&
           fw_str == r.fw_str &&
           sn_str == r.sn_str &&
           mn_str == r.mn_str &&
           modelname_str == r.modelname_str &&
           device_id == r.device_id &&
           serial_id == r.serial_id &&
           chassis_slot == r.chassis_slot &&
           master_mac == r.master_mac &&
           master_ip == r.master_ip &&
           master_udp_port == r.master_udp_port &&
           mstream_mac == r.mstream_mac &&
           mstream_ip == r.mstream_ip &&
           mstream_udp_port == r.mstream_udp_port &&
           mstream_locked == r.mstream_locked &&
           master_locked == r.master_locked &&
           valid == r.valid;
}

LldpInfo::operator DeviceIndex() const {
    return DeviceIndex(static_cast<DeviceId>(device_id), serial_id);
}

QString LldpInfo::formatVersion(const QString &input)
{
    auto list = input.split(".");
    QStringList sl;
    foreach (QString item, list) {
        bool ok = false;
        int n = item.toInt(&ok);
        if (ok)
            sl += QString::number(n);
        else
            sl += item;
    }
    if (sl.size() > 1 && sl.last() == "0")
        sl.removeLast();
    return sl.join(".");
}

QString LldpInfo::getIdent() const
{
    return DeviceIndex(*this).getIdent();
}

QString LldpInfo::getSerialStr() const
{
    return DeviceIndex(*this).getSerialStr();
}

QString LldpInfo::getModelStr() const
{
    QString s = modelname_str.trimmed();
    if(s.isEmpty())
        s = QString("0x%1").arg(device_id, 2, 16, QChar('0'));
    return s;
}

QString LldpInfo::getModelVerStr() const
{
    QString s = getModelStr();
    if (s.isEmpty())
        return s;
    if (!hw_str.isEmpty())
        s += "-" + hw_str;
    return s;
}

QString LldpInfo::getFirmWareStr() const
{
    return fw_str.trimmed();
}

int LldpInfo::getHwMajVer() const
{
    return getHwMajVer(hw_str);
}

int LldpInfo::getHwMajVer(const QString &str)
{
    if(str.isEmpty()) return 0;
    bool ok;
    int hwMajVer = str.leftRef(2).toInt(&ok);
    return ok ? hwMajVer : 0;
}

QPair<int, int> LldpInfo::getHwVer() const
{
    return getHwVer(hw_str);
}

QPair<int,int> LldpInfo::getHwVer(const QString &str)
{
    QPair<int,int> ver(0,0);
    QStringList list = str.split('.');
    if(list.size()==2) {
        int a,b;
        bool ok1,ok2;
        a = list.first().toInt(&ok1);
        b = list.last().toInt(&ok2);
        if(ok1 && ok2) {
            ver.first = a;
            ver.second = b;
        }
    }
    return ver;
}

QDebug operator<<(QDebug dbg, const LldpInfo &r)
{
    dbg << QString("LLDP: {'%1', ip: %2}")
           .arg(r.getIdent())
           .arg(QHostAddress(r.ip_addr).toString());
    return dbg;
}
