#include "AdcCoreModule.h"

#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

// ADC Core Registers
enum {
    REG_ADC_LAT                 = 0x00,
    REG_ADC_MATCH_WIN           = 0x01,
    REG_ADC_PREWIN              = 0x02,
    REG_ADC_WIN                 = 0x03, // Samples to readout
    REG_ADC_MEM_DEPTH           = 0x04, // in kilosamples
    REG_ADC_MODE                = 0x05, // 1:0 - mode (1-norm,2-baseline)
    REG_ADC_GAIN_CRTL           = 0x06, // 7:0 - for ch pairs
    REG64_ADC_CH_EN             = 0x08,

    REG_ADC_BUILD_PARAM_1       = 0x10, // [15:8] - ADC Bits
    REG_ADC_BUILD_PARAM_2       = 0x11, // [7:0] - ADC chips number; [15:8] - ADC chip channel number
    REG_ADC_BUILD_PARAM_3       = 0x12, // [7:0] - log2(max ADC latency); [15:8] - log2(max ADC match window)
//    REG_ADC_BUILD_PARAM_4   = 0x13, // reserved
    REG_ADC_PATTERN_TEST_CTRL   = 0x20, // [0] - enable, [1] - pause
    REG_ADC_TEST_PATTERN        = 0x21,
    REG64_ADC_TEST_COUNT        = 0x24,
    REG128_ADC_PATTERN_TEST_FAILED_MASK = 0x28, // RO 128-bit
    REG64_ADC_CH_HFIFO          = 0x40, // bit# = ch#, RO
    REG64_ADC_CH_MFIFO          = 0x48, // bit# = ch#, RO
    REG64_ADC_CH_EFIFO          = 0x50, // bit# = ch#, RO

    REG_ADC_CH_DATA_BASE        = 0x80, // RO, ADC Channel #0 Data
};

enum {
    REG_ADC_MODE_NORM = 1,
    REG_ADC_MODE_BASELINE = 2,
};

AdcCoreModule::AdcCoreModule(QtMregDevice * const dev)
    :BaseDeviceModule(dev)
{
    if(dev->getDeviceId() == DEVICE_ID_PHOS_FEC)
        setBaseAddr(0x300, 0xFF);
}

mlink::RegOpVector AdcCoreModule::getChannelDataRegOp() const
{
    RegOpVector regOp;
    for(int ch=0; ch<status.adcNCh; ++ch)
        regOp << RegRead(baseAddr()+REG_ADC_CH_DATA_BASE+ch);
    return regOp;
}

QVector<quint16> AdcCoreModule::getChannelData()
{
    const auto &regRes = dev->regOpExec(getChannelDataRegOp());
    QVector<quint16> res;
    res.reserve(regRes.size());
    for(int ch=0; ch<status.adcNCh; ++ch)
        res.push_back(regRes[ch].data);
    return res;
}

void AdcCoreModule::setTestPatternAndEnable(quint16 pattern)
{
    RegOpVector regOp;
    regOp << RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, 0)
          << RegWrite(baseAddr()+REG_ADC_TEST_PATTERN, pattern)
          << RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, 1)
//          << RegRead64(baseAddr+REG_ADC_PATTERN_TEST_CTRL)
//          << RegRead64(baseAddr+REG_ADC_PATTERN_TEST_FAILED_MASK)
             ;
    dev->regOpExec(regOp);
}

void AdcCoreModule::setPatternTestEn(bool en)
{
    dev->regOpExec(RegOpVector{RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, 0),
                               RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, en ? 1 : 0)});
}

quint64 AdcCoreModule::getPatternTestCount()
{
    return dev->regRead64(baseAddr()+REG64_ADC_TEST_COUNT);
}

quint64 AdcCoreModule::getPatternTestFailedChWithStop()
{
    auto regOp = RegOpVector{RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, 0)};
    regOp << RegRead64(baseAddr()+REG128_ADC_PATTERN_TEST_FAILED_MASK);
    const auto &regRes = dev->regOpExec(regOp);
    return regRes.size()==regOp.size() ? dev->get64(regRes, 1) : 0;
}

quint64 AdcCoreModule::getPatternTestFailedChWithPause()
{
    RegOpVector regOp;
    regOp << RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, 3)
          << RegRead64(baseAddr()+REG128_ADC_PATTERN_TEST_FAILED_MASK)
          << RegRead64(baseAddr()+REG128_ADC_PATTERN_TEST_FAILED_MASK+4) //ignored at this moment
          << RegWrite(baseAddr()+REG_ADC_PATTERN_TEST_CTRL, 1);
    const auto &regRes = dev->regOpExec(regOp);
    return regRes.size()==regOp.size() ? dev->get64(regRes, 1) : 0;
}

mlink::RegOpVector AdcCoreModule::setAdcChEnRegOp(quint64 mask)
{
    return RegWrite64(baseAddr()+REG64_ADC_CH_EN, mask);
}

mlink::RegOpVector AdcCoreModule::setModeRegOp(bool zsEn)
{
    const auto &mode = zsEn ? REG_ADC_MODE_NORM : REG_ADC_MODE_BASELINE;
    return RegOpVector{RegWrite(baseAddr()+REG_ADC_MODE, mode)};
}

mlink::RegOpVector AdcCoreModule::setLat(quint16 lat)
{
    return RegOpVector{RegWrite(baseAddr()+REG_ADC_LAT, lat)};
}

mlink::RegOpVector AdcCoreModule::setMatchWin(quint16 matchWin)
{
    return RegOpVector{RegWrite(baseAddr()+REG_ADC_MATCH_WIN, matchWin)};
}

mlink::RegOpVector AdcCoreModule::setPreWin(quint16 preWin)
{
    return RegOpVector{RegWrite(baseAddr()+REG_ADC_PREWIN, preWin)};
}

mlink::RegOpVector AdcCoreModule::setAdcWin(quint16 adcWin)
{
    return RegOpVector{RegWrite(baseAddr()+REG_ADC_WIN, adcWin)};
}

AdcPatternStatus AdcCoreModule::readPatternStatus()
{
    RegOpVector r;
    r << RegWrite(baseAddr() + REG_ADC_PATTERN_TEST_CTRL, 0x3);
    r << RegRead64(baseAddr()+REG64_ADC_TEST_COUNT);
    r << RegRead64(baseAddr()+REG128_ADC_PATTERN_TEST_FAILED_MASK);
    r << RegRead64(baseAddr()+REG128_ADC_PATTERN_TEST_FAILED_MASK+4);
    r << RegWrite(baseAddr() + REG_ADC_PATTERN_TEST_CTRL, 0x1);
    const auto &regRes = dev->regOpExec(r);
    AdcPatternStatus st;
    st.adcNCh = status.adcNCh;
    st.testCnt = QtMregDevice::get64(regRes, 1);
    st.failedChMask_lo = QtMregDevice::get64(regRes, 5);
    st.failedChMask_hi = QtMregDevice::get64(regRes, 9);
    return st;
}

void AdcCoreModule::readBuildParams()
{
    const RegOpVector regOp = {RegRead(baseAddr()+REG_ADC_BUILD_PARAM_1),
                               RegRead(baseAddr()+REG_ADC_BUILD_PARAM_2),
                               RegRead(baseAddr()+REG_ADC_BUILD_PARAM_3)};
    const auto &regRes = dev->regOpExec(regOp);
    if(regRes.size() == regOp.size()) {
        status.adcBits = regRes[0].data >> 8;

        status.adcChips = regRes[1].data & 0xFF;
        status.adcChPerChip = regRes[1].data >> 8;
        status.adcNCh = status.adcChips * status.adcChPerChip;

        status.adcMaxLat = 1 << (regRes[2].data & 0xFF);
        status.adcMaxMatchWin = 1 << (regRes[2].data >> 8);
    }
    if(dev->getDeviceId() == DEVICE_ID_TQDC16VS_E) {
        // remove it after introducing normal TADC module to TQDC16VS
        status.adcBits = 14;

        status.adcChips = 2;
        status.adcChPerChip = 8;
        status.adcNCh = status.adcChips * status.adcChPerChip;
    }
}
