#include "AdcSpiLtm9011Module.h"

#include "mregdev/qtmregdevice.h"

/*
[LTM9011_ADC_SPI_A0_REG]
Bit 7
Software Reset Bit
0 = Not Used
1 = Software Reset. All Mode Control Registers Are Reset to 00h. The ADC Is Momentarily Placed in SLEEP Mode.
After the Reset SPI Write Command Is Complete, Bit D7 Is Automatically Set Back to Zero. The Reset Register Is Write Only.
Bits 6-0
Unused, Don’t Care Bits.

[LTM9011_ADC_SPI_A1_REG]
Bit 7 DCSOFF
Clock Duty Cycle Stabilizer Bit
0 = Clock Duty Cycle Stabilizer On
1 = Clock Duty Cycle Stabilizer Off. This Is Not Recommended.
Bit 6 Data Output Randomizer Mode Control Bit
RAND
0 = Data Output Randomizer Mode Off
1 = Data Output Randomizer Mode On
Bit 5 TWOSCOMP Two’s Complement Mode Control Bit
0 = Offset Binary Data Format
1 = Two’s Complement Data Format
Bits 4-0 SLEEP: NAP_X Sleep/Nap Mode Control Bits
00000 = Normal Operation
0XXX1 = Channel 1 in Nap Mode
0XX1X = Channel 4 in Nap Mode
0X1XX = Channel 5 in Nap Mode
01XXX = Channel 8 in Nap Mode
1XXXX = Sleep Mode. Channels 1, 4, 5 and 8 Are Disabled
Note: Any Combination of Channels Can Be Placed in Nap Mode.

[LTM9011_ADC_SPI_A2_REG]
Bits 7-5 - ILVDS2:ILVDS0 LVDS Output Current Bits
Bit 4    - TERMON LVDS Internal Termination Bit
Bit 3    - OUTOFF Output Disable Bit
Bits 2-0 - OUTMODE2:OUTMODE0 Digital Output Mode Control Bits

[LTM9011_ADC_SPI_A3_REG]
Bit 7   - Digital Output Test Pattern Control Bit (0-off, 1-on)
Bit 6   - Unused
Bit 5-0 - TP13:TP8 Set the Test Pattern for Data Bit 13 (MSB) Through Data Bit 8.

[LTM9011_ADC_SPI_A5_REG]
Bit 7-0 - TP7:TP0 Set the Test Pattern for Data Bit 7 Through Data Bit 0 (LSB).
*/

namespace  {
enum {
    REG_ADC0_SPI_ADDR = 0x0,
    REG_ADC0_SPI_WRITE = 0x1,
    REG_ADC0_SPI_READ = 0x2,
    REG_ADS52J90_ACTIVE_ADC_SELECT = 0x4
};

// LTM9011
enum LTM9011_ADC_SPI_REG {
    LTM9011_ADC_SPI_A0_REG  = 0, // reset
    LTM9011_ADC_SPI_A1_REG  = 1, // format and power-down
    LTM9011_ADC_SPI_A2_REG  = 2, // output mode
    LTM9011_ADC_SPI_A3_REG  = 3, // test pattern msb
    LTM9011_ADC_SPI_A4_REG  = 4, // test pattern lsb
};
}

void AdcSpiLtm9011Module::setupAdc(quint8 adcBits)
{
    // Double loop for SPI1 and SPI2
    for(int adcId=0;adcId<2;adcId++) {
        // reset bit (auto set back to zero)
        spi_write(LTM9011_ADC_SPI_A0_REG, 1<<7, adcId);

        // randomizer
        // spi_write(LTM9011_ADC_SPI_A1_REG, 0x40, adcId);
        // 2's complement
        spi_write(LTM9011_ADC_SPI_A1_REG, 0x20, adcId);

        quint16 a2Val = 0;
        switch (adcBits) {
        case 12: a2Val = 0x2; break;
        case 14: a2Val = 0x1; break;
        case 16: a2Val = 0x0; break;
        default:
            Q_ASSERT(false);
            continue;
        }
        // assume that output is 2-lane
        spi_write(LTM9011_ADC_SPI_A2_REG, a2Val, adcId);
    }
}

void AdcSpiLtm9011Module::setCustomPatternMode(bool mode)
{
    if(mode) {
        setCustomPattern(0x200);
    } else {
        for(int adcId=0;adcId<2;adcId++) {
            spi_write(LTM9011_ADC_SPI_A3_REG, 0, adcId);
        }
    }
}

void AdcSpiLtm9011Module::setCustomPattern(quint16 data)
{
    quint8 dataLow = data & 0xff;
    quint8 dataHi = (data >> 8) & 0x3fu;
    dataHi |= 0x80; // enable test pattern

    for(int adcId=0;adcId<2;adcId++) {
        spi_write(LTM9011_ADC_SPI_A4_REG, dataLow, adcId);
        spi_write(LTM9011_ADC_SPI_A3_REG, dataHi, adcId);
    }
}

int AdcSpiLtm9011Module::getWriteReg(quint8 adcId) const
{
    return baseAddr() + REG_ADC0_SPI_WRITE + adcId*8;
}

int AdcSpiLtm9011Module::getAddrReg(quint8 adcId) const
{
    return baseAddr() + REG_ADC0_SPI_ADDR + adcId*8;
}

void AdcSpiLtm9011Module::spi_write(quint16 addr, quint16 data, quint8 adcId) const
{
    const int &writeReg = getWriteReg(adcId);
    const int &adrReg = getAddrReg(adcId);

    dev->regWrite(writeReg, data);
    dev->regWrite(adrReg, addr);
}

quint16 AdcSpiLtm9011Module::spi_read(quint16 addr, quint8 adcId) const
{
    Q_UNUSED(addr)
    Q_UNUSED(adcId)
    static bool warnFlag = true;
    if(warnFlag) {
        qWarning() << __PRETTY_FUNCTION__ << "Not implemented yet.";
        warnFlag = false;
    }
    return 0;
}
