#include "DeserializersControlModule.h"

#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"

// #include "DeserializeConrtolRegs.h" // toDel?

#define USE_REGOP_RESET_BIT

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

// Deserializers control
enum {
    REG_DES_CTRL                = 0x00,
    REG_DES_STATUS              = 0x01,
    REG_DES_IDELAY_TAP_VAL      = 0x03,
    REG_DES_IDELAY_LOAD_MASK    = 0x04,
    REG_DES_ADC_CH_INFO         = 0x05, // RO
};

enum {
    ADC_CLK_CTRL_BIT_ADC_RESET      = 0x8000,
    ADC_CLK_CTRL_BIT_DES_RESET      = 0x4000,
    ADC_CLK_CTRL_BIT_STATUS_RESET   = 0x2000,
    ADC_CLK_CTRL_BIT_ADC_SYNC       = 0x1000, // to sync ADC after reconfiguration
    ADC_CLK_CTRL_BIT_RESET          = 0x0100, // unused
//    ADC_CLK_CTRL_ALL_ADCS           = 0x00FF  // unused
};

DeserializersControlModule::DeserializersControlModule(QtMregDevice * const _dev)
    :BaseDeviceModule(_dev)
{
    if(dev->getDeviceId() == DEVICE_ID_PHOS_FEC)
        setBaseAddr(0x4000, 0xFF);
}

void DeserializersControlModule::post_init()
{
    quint16 val = dev->regRead(baseAddr() + REG_DES_ADC_CH_INFO);
    desCnt = val&0xFF;
    adcChPerChip = val>>8;
    if(dev->getDeviceId() == DEVICE_ID_TQDC16VS_E) {
        // remove it after introducing normal TADC module to TQDC16VS
        desCnt = 4;
        adcChPerChip = 4;
    }
    desSt.desCnt = desCnt;
    desSt.desCh = adcChPerChip;
}

mlink::RegOpVector DeserializersControlModule::getClockDelaySetRegOp(quint16 clockDelay) const
{
    RegOpVector regOp;
    regOp << RegWrite(baseAddr() + REG_DES_IDELAY_TAP_VAL, clockDelay);
    regOp << RegWrite(baseAddr() + REG_DES_IDELAY_LOAD_MASK, 0x5555);
    return regOp;
}

mlink::RegOpVector DeserializersControlModule::getDataDelaySetRegOp(quint16 dataDelay) const
{
    RegOpVector regOp;
    regOp << RegWrite(baseAddr() + REG_DES_IDELAY_TAP_VAL, dataDelay);
    regOp << RegWrite(baseAddr() + REG_DES_IDELAY_LOAD_MASK, 0xAAAA);
    return regOp;
}

mlink::RegOpVector DeserializersControlModule::getStatusRegOp() const
{
    return RegOpVector{RegRead(baseAddr() + REG_DES_STATUS)};
}

DeserializersControlStatus DeserializersControlModule::getStatus(mlink::RegOpVector regRes, int pos) const
{
    checkRegResValid(&DeserializersControlModule::getStatusRegOp, regRes, pos);

    DeserializersControlStatus desSt;
    desSt.desCnt = desCnt;
    desSt.desCh = adcChPerChip;
    desSt.setStatus(regRes[pos].data);
    return desSt;
}

DeserializersControlStatus DeserializersControlModule::readStatus()
{
    return desSt = getStatus(dev->regOpExec(getStatusRegOp()), 0);
}

void DeserializersControlModule::setDeserReset(bool en)
{
    updateDesCtrlVal();
    if(en)
        curDesCtrlVal |= ADC_CLK_CTRL_BIT_DES_RESET;
    else
        curDesCtrlVal &= ~ADC_CLK_CTRL_BIT_DES_RESET;
    dev->regWrite(baseAddr() + REG_DES_CTRL, curDesCtrlVal);
}

void DeserializersControlModule::ResetAdcPin()
{
    resetCtrlBit(ADC_CLK_CTRL_BIT_ADC_RESET);
}

void DeserializersControlModule::setAdcSync()
{
    resetCtrlBit(ADC_CLK_CTRL_BIT_ADC_SYNC);
}

void DeserializersControlModule::ResetDeser()
{
    resetCtrlBit(ADC_CLK_CTRL_BIT_DES_RESET);
    QThread::usleep(10);
}

void DeserializersControlModule::changeDesDelay(quint16 channel_mask, quint16 taps)
{
    dev->regWrite(baseAddr()+REG_DES_IDELAY_TAP_VAL, taps);
    dev->regWrite(baseAddr()+REG_DES_IDELAY_LOAD_MASK, channel_mask);
}

void DeserializersControlModule::ResetAdcClkDelay()
{
    // this bit is unused in f/w at least since 10 Jun 2021
    resetCtrlBit(ADC_CLK_CTRL_BIT_RESET);
}

quint16 DeserializersControlModule::lockTest()
{
    quint16 des_status = dev->regRead(baseAddr() + REG_DES_STATUS);
    desSt.setStatus(des_status);
    //    emit adcLockStatusUpdated(getDeviceIndex(), lock_ok, des_status);

    return des_status;

    //    if (!lock_ok) qCritical() << "ADC data lock error. " << adcStatusStr(des_status);
    //    else
    //        qDebug() << adcStatusStr(des_status);
}

void DeserializersControlModule::updateDesCtrlVal()
{
    curDesCtrlVal = dev->regRead(baseAddr() + REG_DES_CTRL);
}

void DeserializersControlModule::resetCtrlBit(quint16 bit)
{
    updateDesCtrlVal();
#ifdef USE_REGOP_RESET_BIT
    RegOpVector regOp;
    curDesCtrlVal |= bit;
    regOp <<RegWrite(baseAddr() + REG_DES_CTRL, curDesCtrlVal);
    curDesCtrlVal &= ~bit;
    regOp <<RegWrite(baseAddr() + REG_DES_CTRL, curDesCtrlVal);
    dev->regOpExec(regOp);
#else
    curDesCtrlVal |= bit;
    dev->regWrite(baseAddr + REG_DES_CTRL, curDesCtrlVal);
    curDesCtrlVal &= ~bit;
    dev->regWrite(baseAddr + REG_DES_CTRL, curDesCtrlVal);
#endif
}

void DeserializersControlModule::ResetAdcStatus()
{
    resetCtrlBit(ADC_CLK_CTRL_BIT_STATUS_RESET);
}

void DeserializersControlModule::setClockDelay(quint16 clkDelay)
{
    changeDesDelay(0x5555, clkDelay);
}

void DeserializersControlModule::setDataDelay(quint16 dataDelay)
{
    changeDesDelay(0xAAAA, dataDelay);
}

void DeserializersControlModule::setSerDesDelay(const DevSerDesDelay &delay)
{
    ResetAdcClkDelay();

    for(auto it=delay.cbegin(); it!=delay.cend(); ++it) {
        const auto &delay = it.value();
        quint16 mask = 1 << (it.key()*2);
        changeDesDelay(mask, delay.clock);
        mask <<= 1;
        changeDesDelay(mask, delay.data);
    }

    lockTest();

    ResetDeser();
    ResetAdcStatus();
    QThread::msleep(1);
}
