#include "HitCoreModule.h"

#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"

#include "HitCoreStatus.h"

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

// TADC Hit Core Registers
enum {
    REG_HIT_COUNTER_CTRL        = 0x00,
    REG64_HIT_CORE_TS           = 0x04,
    REG64_INPUT_SIGNAL_POLARITY = 0x08, // 1-negative, 0-positive
    REG64_HIT_CMP_STATE         = 0x10, // RO, Comparators State
    REG32_HIT_CNT_CH_BASE       = 0x80, // RO, 32-bit
};

enum {
    REG_HIT_COUNTER_LOCK_BIT = 0x0001
};

HitCoreModule::HitCoreModule(QtMregDevice * const dev)
    :BaseDeviceModule(dev)
{
    if(dev->getDeviceId() == DEVICE_ID_PHOS_FEC)
        setBaseAddr(0x600, 0xFF);
}

mlink::RegOpVector HitCoreModule::setChInvertRegOp(quint64 chInv) const
{
    return RegWrite64(baseAddr()+REG64_INPUT_SIGNAL_POLARITY, chInv);
}

mlink::RegOpVector HitCoreModule::getChHitCounterRegOp(int nCh) const
{
    RegOpVector regOp;
    regOp << RegWrite(baseAddr() + REG_HIT_COUNTER_CTRL, REG_HIT_COUNTER_LOCK_BIT);

    for(int ch=0; ch<nCh; ++ch)
        regOp << RegRead32(baseAddr() + REG32_HIT_CNT_CH_BASE + 2*ch);

    regOp << RegRead64(baseAddr() + REG64_HIT_CORE_TS);
    regOp << RegWrite(baseAddr() + REG_HIT_COUNTER_CTRL, 0);

    return regOp;
}

HitCoreStatus HitCoreModule::getHitStatus(const mlink::RegOpVector &regRes, int posOffset) const
{
    const int minAddr = baseAddr()+REG32_HIT_CNT_CH_BASE;
    const int maxAddr = baseAddr()+REG32_HIT_CNT_CH_BASE+0xFF;
    const int tsAddr = baseAddr() + REG64_HIT_CORE_TS;

    HitCoreStatus s;
    s.isPhos = dev->getDeviceId() == DEVICE_ID_PHOS_FEC;
    for(int i=posOffset+1; i<regRes.size()-1; i+=2) {
        const auto &regOp = regRes[i];
        if(regOp.addr == tsAddr) {
            s.ts = QtMregDevice::get64(regRes, i);
            break;
        }
#ifdef QT_DEBUG
        Q_ASSERT(regOp.addr >= minAddr && regOp.addr <= maxAddr &&
                 regRes[i+1].addr == (regOp.addr+1));
        const int &ch = (regOp.addr - minAddr)/2;
        Q_ASSERT(ch == s.chCnt.size());
#endif
        s.chCnt.push_back(QtMregDevice::get32(regRes, i));
    }
    return s;
}

mlink::RegOpVector HitCoreModule::getCmpStateRegOp() const
{
    return RegRead64(baseAddr()+REG64_HIT_CMP_STATE);
}
