#include "PllSi5326Module.h"

#include <QElapsedTimer>
#include <QThread>

#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"

using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

enum {
    REG_CMD_BASE        = 0x00,
    REG_FSM_CRTL        = 0x80,
    REG_FSM_STATE       = 0x81,
    REG_COMMANDS_NUM    = 0x82,
    REG_STATUS          = 0x83,
};

enum {
    BIT_CTRL_IDLE               = 0x00,
    BIT_CTRL_RECONFIG_REQUEST   = 0x01, // PLL reconfig request (start SPI sequence)
    BIT_CTRL_RAM_ACCESS_REQUEST = 0x02, // request to communicate with configuration RAM
};
enum {
    BIT_FSM_INIT            = 0x01,
    BIT_FSM_IDLE            = 0x02,
    BIT_FSM_SOFTWARE_ACCESS = 0x04, // software RAM read/write
    BIT_FSM_READ_RAM_CMD    = 0x08, // read command from RAM into SPI PHY
    BIT_FSM_SPI_ADDR_WR     = 0x10, // SPI PHY address write
    BIT_FSM_SPI_DATA_WR     = 0x20  // SPI PHY data write
};

enum {
    BIT_STATUS_LOL      = 0x1, // PLL LOL pin
    BIT_STATUS_INT_C1B  = 0x2, // PLL INT_C1B pin
    BIT_STATUS_C2B      = 0x4 // PLL C2B pin
};

PllSi5326Module::PllSi5326Module(QtMregDevice * const _dev)
    :BaseDeviceModule(_dev)
{
    if(dev->getDeviceId() == DEVICE_ID_PHOS_FEC)
        setBaseAddr(0x4400, 0xFF);
}

void PllSi5326Module::post_init()
{
    if(dev->getDeviceId() == DEVICE_ID_PHOS_FEC) {
//        patchPllConfig();
    }
}

bool PllSi5326Module::waitFsmState(quint16 state)
{
    if(!dev->isConnected())
        return false;
    QElapsedTimer timer;
    timer.start();
    while(timer.elapsed() < 1e3) {
        const auto &data = dev->regRead(baseAddr() + REG_FSM_STATE);
        if(data == state)
            return true;
    }
    qWarning("%s failed on timeout while waiting state: 0x%xd", __PRETTY_FUNCTION__, state);
    return false;
}

void PllSi5326Module::patchPllConfig()
{
    if(!dev->isConnected())
        return;
    dev->regWrite(baseAddr() + REG_FSM_CRTL, BIT_CTRL_RAM_ACCESS_REQUEST);
    if(!waitFsmState(BIT_FSM_SOFTWARE_ACCESS)) {
        return;
    }
    RegOpVector regOp;

    constexpr static const quint16 PLL_COMMAND_2 = 0x242;
    regOp << RegWrite(baseAddr() + REG_CMD_BASE + 2, PLL_COMMAND_2) // cmd#3 0x2a2 -> 0x242
          << RegRead (baseAddr() + REG_CMD_BASE + 2)
          << RegWrite(baseAddr() + REG_FSM_CRTL, 0)
          << RegWrite(baseAddr() + REG_FSM_CRTL, BIT_CTRL_RECONFIG_REQUEST)
          << RegWrite(baseAddr() + REG_FSM_CRTL, 0);

    const auto &regRes = dev->regOpExec(regOp);
    if(regRes[1].data != PLL_COMMAND_2)
        qCritical("Failed to write new PLL command to RAM regVal:0x%xd (expected 0x%xd)",
                  regRes[1].data, PLL_COMMAND_2);
    if(!waitFsmState(BIT_FSM_IDLE)) {
        return;
    }

    // Wait end of PLL internal calibration (the last PLL SPI command)
    QThread::sleep(1);
}
