#include "TdcCoreModule.h"

#include "TdcCoreConfig.h"
#include "TdcCoreStatus.h"

#include "mregdev/qtmregdevice.h"
#include "regio/regop.h"


using mlink::RegOpVector;
using mlink::RegRead;
using mlink::RegRead32;
using mlink::RegRead64;
using mlink::RegWrite;
using mlink::RegWrite32;
using mlink::RegWrite64;

// TDC Core Registers
enum {
    REG32_TDC_CH_EN_BASE    = 0x00,
    REG32_TDC_GEN_CTRL      = 0x20,
    REG32_TDC_READ_CTRL     = 0x22,
    REG32_TDC_MEASURE_CTRL  = 0x24,
    REG32_TDC_TRIG_WIN_SETUP= 0x26,
    REG32_TDC_GEN_STATUS    = 0x2c,
//    REG32_TDC_RC_ADJUST_SETUP = 0x28 //unused
};

// HPTDCs Status Registers
enum {
    REG_TDC_HPTDC_STATUS_BASE = 0x80,
//    REG_HPTDC1_STATUS_2 = 0x81,
//    REG_HPTDC1_STATUS_3 = 0x82,
//    REG_HPTDC1_STATUS_4 = 0x83,
    // Reg 0x0084-0x00FF - same for every other HPTDC
    // regs repeated for every HPTDC
};

enum {
    REG32_TDC_GEN_CTRL_BIT_STANDBY = 1
};

QDebug operator<<(QDebug dbg, const reg_measurement_ctrl &s)
{
    dbg << "reg_measurement_ctrl{"
        << "leading_resolution:" << s.bit.leading_resolution
        << "width_select:" << s.bit.width_select
        << "unused:" << s.bit.unused
        << "mode_rc_compression:" << s.bit.mode_rc_compression
        << "mode_rc:" << s.bit.mode_rc
        << "DLL_Mode:" << s.bit.DLL_Mode
        << "enable_leading:" << s.bit.enable_leading
        << "enable_trailing:" << s.bit.enable_trailing
        << "enable_pair:" << s.bit.enable_pair
        << "core_clock_source:" << s.bit.core_clock_source
        << "dll_clock_source:" << s.bit.dll_clock_source
        << "unused2:" << s.bit.unused2
        << "enable_matching:" << s.bit.enable_matching
        << "unused3:" << s.bit.unused3
        << "}";
    return dbg;
}

QDebug operator<<(QDebug dbg, const reg_read_ctrl &s)
{
    dbg << "reg_read_ctrl{"
        << "enable_local_header:" << s.bit.enable_local_header
        << "enable_local_trailer:" << s.bit.enable_local_trailer
        << "unused:" << s.bit.unused
        << "enable_error_mark:" << s.bit.enable_error_mark
        << "unused2:" << s.bit.unused2
        << "readout_fifo_size:" << s.bit.readout_fifo_size
        << "unused3:" << s.bit.unused3
        << "max_event_size:" << s.bit.max_event_size
        << "reject_readout_fifo_full:" << s.bit.reject_readout_fifo_full
        << "unused4:" << s.bit.unused4
        << "dead_time:" << s.bit.dead_time
        << "unused5:" << s.bit.unused5
        << "}";
    return dbg;
}

TdcCoreModule::TdcCoreModule(QtMregDevice * const dev)
    :BaseDeviceModule(dev)
{
    if(dev->getDeviceId() == DEVICE_ID_PHOS_FEC) {
        setBaseAddr(0x200, 0xFF);
        chipNum = 1;
    }

    restoreDefaultValues();
}

mlink::RegOpVector TdcCoreModule::setStandbyRegOp(bool standBy) const
{
    return RegWrite32(baseAddr()+REG32_TDC_GEN_CTRL, standBy ? REG32_TDC_GEN_CTRL_BIT_STANDBY : 0);
}

mlink::RegOpVector TdcCoreModule::setChEnMaskRegOp(std::vector<quint32> chMask) const
{
    RegOpVector v;
    int addr = baseAddr() + REG32_TDC_CH_EN_BASE;
    for(auto mask : chMask) {
        v << RegWrite32(addr, mask);
        addr += 2;
    }
    return v;
}

mlink::RegOpVector TdcCoreModule::setMatchWinAndLatRegOp(quint32 mW, quint32 lat) const
{
    return RegWrite32(baseAddr() + REG32_TDC_TRIG_WIN_SETUP, (mW&0xFFF)<<12 | (lat & 0xFFF));
}

void TdcCoreModule::setMeasurementCtrl(bool enVHR, bool enLead, bool enTrail)
{
    measurementCtrlVal.bit.mode_rc_compression = enVHR;
    measurementCtrlVal.bit.mode_rc = enVHR;

    measurementCtrlVal.bit.enable_leading = enLead;
    measurementCtrlVal.bit.enable_trailing = enTrail;

    dev->regWrite32(baseAddr() + REG32_TDC_GEN_CTRL, REG32_TDC_GEN_CTRL_BIT_STANDBY);
    if(!waitHptdcInit())
        return;

//    measureCtrl |= 1<<9; // mode_rc
    RegOpVector regOp;
    regOp << RegWrite32(baseAddr() + REG32_TDC_MEASURE_CTRL, measurementCtrlVal.all);
    regOp << RegWrite32(baseAddr() + REG32_TDC_GEN_CTRL, 0);
    dev->regOpExec(regOp);
}

mlink::RegOpVector TdcCoreModule::setHptdcAdjustConfigRegOp(const HptdcAdjustConfig &adjustConfig)
{
    readCtrlVal.bit.readout_fifo_size = adjustConfig.readout_fifo_size;

    measurementCtrlVal.bit.mode_rc_compression = adjustConfig.mode_rc_compression;
    measurementCtrlVal.bit.leading_resolution = adjustConfig.leading_resolution;
    measurementCtrlVal.bit.enable_pair = adjustConfig.enable_pair;
    measurementCtrlVal.bit.DLL_Mode = adjustConfig.dll_mode;
    measurementCtrlVal.bit.dll_clock_source = adjustConfig.dll_clock;
    measurementCtrlVal.bit.core_clock_source = adjustConfig.core_clock_source;

    RegOpVector regOp;
    regOp << RegWrite32(baseAddr() + REG32_TDC_READ_CTRL, readCtrlVal.all);
    regOp << RegWrite32(baseAddr() + REG32_TDC_MEASURE_CTRL, measurementCtrlVal.all);
//    qInfo("%s "
//          "\nREG32_TDC_READ_CTRL: 0x%08xd"
//          "\nREG32_TDC_MEASURE_CTRL: 0x%08xd",
//          __PRETTY_FUNCTION__,
//          readCtrlVal.all,
//          measurementCtrlVal.all);
//    qInfo() << readCtrlVal;
//    qInfo() << measurementCtrlVal;
    return regOp;
}

mlink::RegOpVector TdcCoreModule::getStatusRegOp() const
{
    RegOpVector v;
    v << RegRead32(baseAddr() + REG32_TDC_GEN_STATUS);
    for(int hptdcId=0; hptdcId<chipNum; ++hptdcId) {
        for(int reg=0; reg<4; ++reg) {
            v << RegRead(baseAddr() + REG_TDC_HPTDC_STATUS_BASE+(hptdcId*4)+reg);
        }
    }

    return v;
}

TdcCoreStatus TdcCoreModule::getStatus(const mlink::RegOpVector &regRes, int posOffset) const
{
    checkRegResValid(&TdcCoreModule::getStatusRegOp, regRes, posOffset);

    TdcCoreStatus st;
    int p = posOffset;
    st.generalStatus = QtMregDevice::get32(regRes, p); p += 2;
    st.hptdcStatus.reserve(chipNum);
    for(int hptdcId=0; hptdcId<chipNum; ++hptdcId) {
        HptdcStatus hSt;
        for(int reg=0; reg<4; ++reg) {
            hSt.regValues[reg] = regRes[p].data; p += 1;
        }
        st.hptdcStatus.push_back(hSt);
    }

    return st;
}

void TdcCoreModule::post_init()
{
    RegOpVector regOp;
    regOp << RegWrite32(baseAddr() + REG32_TDC_MEASURE_CTRL, measurementCtrlVal.all);
    regOp << RegWrite32(baseAddr() + REG32_TDC_READ_CTRL, readCtrlVal.all);
    dev->regOpExec(regOp);
}

bool TdcCoreModule::waitHptdcInit() const
{
    QElapsedTimer timer;
    timer.start();
    constexpr static const qint64 STANDBY_TIMEOUT = 200; // ~100ms for pll reset
    quint32 genCtrl;
    while(((genCtrl=dev->regRead32(baseAddr() + REG32_TDC_GEN_STATUS)) & 0x1) == 0){
        if(timer.elapsed() > STANDBY_TIMEOUT){
            qWarning() << "Failed to wait INIT state at device:" << dev->getIdent()
                       << "TDC_GEN_CTRL:" << hex << genCtrl;
            return false;
        }
    }
    return true;
}

void TdcCoreModule::restoreDefaultValues()
{
    reg_measurement_ctrl m;
    m.all = 0;
    m.bit.DLL_Mode = 2;
    m.bit.enable_leading = 1;
    m.bit.enable_trailing = 1;
    m.bit.core_clock_source = 1;
    m.bit.dll_clock_source = 3;
    m.bit.enable_matching = 1;

    reg_read_ctrl r;
    r.all = 0;
    r.bit.enable_local_header = 1;
    r.bit.enable_local_trailer = 1;
    r.bit.enable_error_mark = 1;
    r.bit.readout_fifo_size = 7;
    r.bit.max_event_size = 9;
    r.bit.reject_readout_fifo_full = 1;

    measurementCtrlVal = m;
    readCtrlVal = r;
}
