#include "AdcSerDesCalibration.h"

#include "device-modules/AdcSpiBaseModule.h"
#include "device-modules/AdcCore/AdcCoreModule.h"
#include "device-modules/DeserializersControl/DeserializersControlModule.h"
#include "mregdev/qtmregdevice.h"

#include <QDebug>
#include <QRandomGenerator>

AdcSerDesCalibration::AdcSerDesCalibration(AdcCoreModule *_adcCore,
                                     AdcSpiBaseModule *_spiAdc,
                                     DeserializersControlModule *_desCrtl,
                                     QtMregDevice *dev)
    :QObject(dev)
    ,adcCore(_adcCore)
    ,adcSpi(_spiAdc)
    ,desCtrl(_desCrtl)
{
}

void AdcSerDesCalibration::calibrate(bool hardMode, const QString &resDir)
{
    init();

    hardPatternTestEnabled = hardMode;
    QVector< QVector<bool> > ptr; //pattern test result
    QVector<int> adc_delay_value;

    desCtrl->ResetDeser();
    desCtrl->ResetAdcClkDelay();

    adcSpi->setCustomPatternMode(true);
    delayStatus = AdcClkStatus();

    generateSpiPatternSet();

    // [desIndex][clkDelay][dataDelay] = true/false
    QMap<int, QMap<int, QMap<int, bool > > > resultsPattern;
    QMap<int, QMap<int, QMap<int, bool > > > resultsAdcStatus;
    QElapsedTimer timer, startTime;
    timer.start();
    startTime = timer;
    int goodPoints=0, badPoints=0;
    for(int ch =0; ch<getNumberOfChannels();++ch) {
        chPassTestCh[ch] = 0;
    }
    for(clkDelay=0; clkDelay<32; clkDelay++) {
        desCtrl->setClockDelay(clkDelay);
        for(dataDelay=0; dataDelay<32; dataDelay++) {
            const int &taskProgres = 100*(clkDelay*32+dataDelay+1)/(32*32);
            emit progressUpdated(taskProgres);
            if(timer.elapsed() > 5000) {
                qInfo("%s - %3d %% Good-%d Bad-%d", __func__, taskProgres,
                      goodPoints, badPoints);
                timer.start();
            }
            passTestCh[clkDelay][dataDelay] = 0;
            failTestCh[clkDelay][dataDelay] = 0;

            // Set delays
            desCtrl->setDataDelay(dataDelay);
            desCtrl->ResetDeser();
            desCtrl->ResetAdcStatus();
            QThread::msleep(1);

            // Get des lock status
            const auto desSt = desCtrl->readStatus();
            QMap<int, bool> adcDesStatus;
            for(int desId=0; desId<desCnt; ++desId) {
                adcDesStatus[desId] = desSt.isDesOk(desId);
            }

            // Get channels pattern test
            QMap<int, bool> chPatternRes;
            for(int ch =0; ch<getNumberOfChannels();++ch) {
                chPatternRes[ch] = true;
            }
            chPatternRes = patternTestByChannel();

            // Save des lock status
            for(int desId=0; desId<desCnt; ++desId) {
                resultsAdcStatus[desId][clkDelay][dataDelay] = desSt.isDesOk(desId);
            }

            // Mask reults with ch. pattern test results
            bool allChRes=true;
            for(int ch : chPatternRes.keys()) {
                bool &v = chPatternRes[ch];
                allChRes &= v;
                resultsPattern[ch][clkDelay][dataDelay] = v;
                const int desId = ch/desCh;
                adcDesStatus[desId] &= v;
            }

            if(ptr.size()!=chPatternRes.size()) ptr.resize(chPatternRes.size());
            for(int ch=0; ch<ptr.size(); ch++) ptr[ch].push_back(chPatternRes[ch]);
            for(auto desId : adcDesStatus.keys()) {
                delayStatus.data[desId][clkDelay][dataDelay] = adcDesStatus[desId];
            }
            if(allChRes && desSt.isOk()) {
                ++goodPoints;
            } else {
                ++badPoints;
            }
        }
    }
    if(groupTwoDes) {
        auto groupedDelayStatus = AdcClkStatus();
        for(int desId=0; desId<desCnt/2; ++desId) {
            AdcClkStatus tmp;
            tmp.data[0] = delayStatus.data[desId*2];
            tmp.data[1] = delayStatus.data[desId*2+1];
            tmp = tmp.aggregate({tmp});
            groupedDelayStatus.data[desId*2] =
                    groupedDelayStatus.data[desId*2+1] = tmp.data[-1];
        }
        delayStatus = groupedDelayStatus;
    }

    qInfo("%s passed in %lldms", __func__, startTime.elapsed());

    qInfo().noquote() << AdcClkStatus::getLegend();
    qInfo().noquote() << delayStatus.toString();
    delayStatus.temp = readTemp();
    delayStatus.fwStr = getFirmwareVersionStr();

    adcSpi->setCustomPatternMode(false);
    saveResultsToFile(resDir);
}

void AdcSerDesCalibration::init()
{
    const auto &adcCoreStatus = adcCore->getStatus();
    adcBits = adcCoreStatus.adcBits;

    desCnt = desCtrl->readStatus().desCnt;
    desCh = desCtrl->readStatus().desCh;
}

void AdcSerDesCalibration::generateSpiPatternSet()
{
    spiPatternSet.clear();

    auto rndGen = QRandomGenerator::securelySeeded();
    const int &rndLimit = hardPatternTestEnabled ? 30 : 1;
    while(spiPatternSet.size()<rndLimit) {
        quint16 val = static_cast<quint16>(rndGen.generate());
        if(val!=0)
            spiPatternSet.push_back(val);
    }
    if(hardPatternTestEnabled) {
        spiPatternSet.push_back(0x5555);
        spiPatternSet.push_back(0xAAAA);

        spiPatternSet.push_back(0x0A95);
        spiPatternSet.push_back(0x056a);
    }
}

QMap<int, bool> AdcSerDesCalibration::patternTestByChannel()
{
    quint16 adc_pattern;
    QMap<int,bool> totalChRes;

    for(quint16 spi_pattern : spiPatternSet) {
        // Mask with adcBits
        spi_pattern &= (1<<adcBits)-1;
        // Arrange to 16bit data
        adc_pattern = (spi_pattern << (16-adcBits));
        //qInfo() << hex << "adc_pattern:" << adc_pattern;

        adcSpi->setCustomPattern(spi_pattern);
        const auto &chRes = TestAdcCustomPatternByChannel(adc_pattern);
        if(totalChRes.isEmpty()) {
            totalChRes = chRes ;
        } else {
            for(int ch : chRes .keys())
                totalChRes[ch]&=chRes [ch];
        }
    }
    return totalChRes;
}

//#define USE_CH_DATA_CHECK
QMap<int, bool> AdcSerDesCalibration::TestAdcCustomPatternByChannel(const quint16 &pattern)
{
    QMap<int, bool> test_ok;

    quint8 nChFailed = 0;
#ifndef USE_CH_DATA_CHECK
    adcCore->setTestPatternAndEnable(pattern);

    QThread::msleep(hardPatternTestEnabled ? 2 : 1);
    const auto &st = adcCore->readPatternStatus();
    for (int ch=0; ch<getNumberOfChannels(); ch++) {
        bool ok = (ch<64 ? (st.failedChMask_lo>>ch) : (st.failedChMask_hi>>(ch-64))) & 1;
        if(ok) {
            ++nChFailed;
            test_ok[ch] = false;
            ++failTestCh[clkDelay][dataDelay];
        } else {
            test_ok[ch] = true;
            ++passTestCh[clkDelay][dataDelay];
            ++chPassTestCh[ch];
        }
    }
    return test_ok;
#else
    const int pattern_cnt_reread_number = hardPatternTestEnabled ? 10 : 1;  //reading pattern mismatch counters 2 times
    static QSet<quint16> patterns;
    QVector <quint16> read_pattern_cnt;
    quint64 chFailed;
    static quint8 nChFailedMin = 110;
    if(nChFailed < nChFailedMin) {
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestFailedCh: 0x" << hex << chFailed << "(" << dec << nChFailed << ")";
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestCount: " << adcCore->getPatternTestCount();
        qInfo() << __PRETTY_FUNCTION__ << clkDelay << dataDelay << hex << pattern << "!!!!!!!!!!!!";

        nChFailedMin = nChFailed;
    }

    QVector<quint16> chData;
    for (int j=0; j<pattern_cnt_reread_number; j++) {
        chData = adcCore->getChannelData();
        for (int ch=0; ch<getNumberOfChannels(); ch++) {
            test_ok[ch] &= chData[ch] != pattern;
        }
    }
    auto chFailed2 = adcCore->getPatternTestFailedChWithStop();
    if(chFailed != chFailed2) {
        quint64 chFailedDiff = chFailed & (~chFailed2);
        if(chFailedDiff)
            qInfo() << __PRETTY_FUNCTION__ << "chFailedDiff: " << chFailedDiff;
        qInfo() << __PRETTY_FUNCTION__ << "chFailed!=chFailed2:" << hex << chFailed << hex << chFailed2 << hex << (chFailed^chFailed2);
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestCount: " << adcCore->getPatternTestCount();
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestCount: " << adcCore->getPatternTestCount();
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestCount: " << adcCore->getPatternTestCount();
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestCount: " << adcCore->getPatternTestCount();
        qInfo() << __PRETTY_FUNCTION__ << "getPatternTestCount: " << adcCore->getPatternTestCount();
        qInfo() << __PRETTY_FUNCTION__ << clkDelay << dataDelay << hex << pattern << "!!!!!!!!!!!!";
    }

    return test_ok;


    for (int ch=0; ch<getNumberOfChannels(); ch++) {
//        if(hasDeleteFlag())
//            return test_ok;
        read_pattern_cnt.clear();
        test_ok[ch] = true;

//        WriteChReg(ch, MEM_CH_ADC_PATTERN, pattern);
//        for (int j=0; j<pattern_cnt_reread_number+1; j++){
//            read_pattern_cnt.push_back(ReadChReg(ch, MEM_CH_ADC_PATTERN_MISMATCH_CNT));
//        }


        for (int j=1; j<read_pattern_cnt.size(); j++)
        {
            if (read_pattern_cnt[j]!=read_pattern_cnt[0])
            {
                //                QVector<quint16> adcVal;
                //                for(int k=0; k<100; ++k)
                //                    adcVal.push_back(ReadChReg(ch, MEM_CH_ADC_DATA));
                test_ok[ch] = false;
                patterns << pattern;
                break;
            } else {
                //                    qInfo()<<"test_ok[i] = true;"<<i;
            }
        }
    }
    //    qInfo() << __PRETTY_FUNCTION__ << hex << patterns;
    return test_ok;
#endif
}

double AdcSerDesCalibration::readTemp() const
{
    auto dev = qobject_cast<QtMregDevice*>(parent());
    return dev ? dev->getTemperature() : 0;
}

QString AdcSerDesCalibration::getFirmwareVersionStr() const
{
    auto dev = qobject_cast<QtMregDevice*>(parent());
    return dev ? dev->getFirmwareVersionStr() : QString();
}

void AdcSerDesCalibration::saveResultsToFile(const QString &resDir) const
{
    if(resDir.isEmpty())
        return;
    auto dev = qobject_cast<const QtMregDevice*>(parent());

    QFile f;
    int loopCnt = 0;
    do {
        QStringList fileNameParts{"devSerDesDelay"};
        if(dev) {
            fileNameParts << dev->getDeviceIndex().getIdent().replace(' ', '_');
            fileNameParts << QString("%1").arg(dev->getTemperature(), 2, 'f', 0, QChar('0'));
        }
        if(loopCnt)
            fileNameParts << QString::number(loopCnt);

        const auto &filePath = QString("%1/%2.txt")
                .arg(resDir).arg(fileNameParts.join('_'));

        f.setFileName(filePath);

        if(++loopCnt > 100) {
            qWarning() << "Failed to save adcClkDelay; last file attempt:"
                       << filePath;
            break;
        }
    }while(!f.open(QIODevice::WriteOnly|QIODevice::NewOnly));

    if(!f.isOpen())
        return;

    QTextStream stream(&f);
    stream << QString("=== %1 (f/w = %2; temp = %3) ===")
              .arg(dev ? dev->getDeviceIndex().getIdent() : "unknown device")
              .arg(delayStatus.fwStr)
              .arg(delayStatus.temp, 0, 'f', 1);
    stream << "\n";
    stream << delayStatus.toString();
    f.close();
}
