#include "FeLinkStatusModel.h"

#include <QColor>
#include <QDateTime>
#include <QTime>

#include "FeLink/FeLinkStatus.h"

namespace {
QString getStateTimeHr(quint32 time)
{
    const static quint32 DAY = 24*3600*1000;
    int days = time / DAY;
    return (days ? QString("%1d ").arg(days) : "") +
            QTime::fromMSecsSinceStartOfDay(time%DAY).toString("hh:mm:ss.zzz");
}

const int MAX_SLOT_NUMBER = 21;
const int CRATE_SLOT_ZERO = 10; // slot number for ch:0

int chToSlot(int ch)
{
    bool odd = ch&1;
    int slotBase = CRATE_SLOT_ZERO;
    if(odd)
        slotBase += 3;

    int slot = (ch&1) ? (slotBase+(ch>>1)) : (slotBase-(ch>>1));
    if(slot<0 || slot>MAX_SLOT_NUMBER) return -1;

    return slot;
}

bool sortFeLinkStatus(const FeLinkStatus &l, const FeLinkStatus &r)
{
    if(l.isMaster() != r.isMaster())
        return l.isMaster(); // Master to top
    const auto &lMasterIndex = l.getMasterIndex();
    const auto &rMasterIndex = r.getMasterIndex();

    if(lMasterIndex != rMasterIndex)
        return lMasterIndex<rMasterIndex;
    if(l.readerCh != r.readerCh) {
        if(lMasterIndex.getDevId() == DEVICE_ID_TTVXS)
            return chToSlot(l.readerCh) < chToSlot(r.readerCh);
        else
            return l.readerCh < r.readerCh;
    }

    return l.getSlaveIndex() < r.getSlaveIndex();
}

}
enum ROW_FEL_TRIGS {ROW_REMOTE_INDEX,
                    ROW_STATUS,
                    ROW_BITSLIDE,
                    ROW_REMOTE_VER,
                    ROW_TAI,
                    ROW_CUR_LINK_STATE_TIME,
                    ROW_CNT_RX_CODE_ERR,
                    ROW_CNT_RX_CRC_ERR,
                    ROW_CNT_TIME_FRAMES_SENT,
                    ROW_CNT_TRIG_FRAMES_SENT,
                    ROW_CNT_TIME_FRAMES_RECV,
                    ROW_CNT_TRIG_FRAMES_RECV,
                    ROW_CNT_UNKNOWN_FRAMES,
                    ROW_FEL_TOTAL  };

FeLinkStatusModel::FeLinkStatusModel(QObject *parent)
    :QAbstractTableModel(parent),
      statusList(new QList<FeLinkStatus>())
{
}

FeLinkStatusModel::~FeLinkStatusModel()
{
    delete statusList;
}

int FeLinkStatusModel::rowCount(const QModelIndex &) const
{
    return ROW_FEL_TOTAL;
}

int FeLinkStatusModel::columnCount(const QModelIndex &) const
{
    return statusList->size();
}

QVariant FeLinkStatusModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto row = index.row();
    if (statusList->size() < column)
        return QVariant();
    const auto &st = (*statusList)[column];

    switch (role) {
    case Qt::BackgroundRole: return data_BackgroundRole(st, row);
    case Qt::DisplayRole: return data_DisplayRole(st, row);
    case Qt::ToolTipRole: return data_ToolTipRole(st, row);
    }
    return QVariant();
}

QVariant FeLinkStatusModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (role != Qt::DisplayRole)
        return QVariant();
    if(orientation == Qt::Vertical) {
        switch (section) {
        case ROW_REMOTE_INDEX: return tr("Remote index");
        case ROW_STATUS: return tr("Status");
        case ROW_BITSLIDE: return tr("Bit slide");
        case ROW_REMOTE_VER: return tr("Remote ver.");
        case ROW_TAI: return tr("TAI");
        case ROW_CUR_LINK_STATE_TIME: return tr("Cur. link state time");
        case ROW_CNT_RX_CODE_ERR: return tr("RX_CODE_ERR");
        case ROW_CNT_RX_CRC_ERR: return tr("RX_CRC_ERR");
        case ROW_CNT_TIME_FRAMES_SENT: return tr("TIME_FRAMES_SENT");
        case ROW_CNT_TRIG_FRAMES_SENT: return tr("TRIG_FRAMES_SENT");
        case ROW_CNT_TIME_FRAMES_RECV: return tr("TIME_FRAMES_RECV");
        case ROW_CNT_TRIG_FRAMES_RECV: return tr("TRIG_FRAMES_RECV");
        case ROW_CNT_UNKNOWN_FRAMES: return tr("UNKNOWN_FRAMES");
        }
    } else {
        const auto &st = statusList->at(section);
        if(st.isMaster()) {
            if(st.readerIndex.getDevId() == DEVICE_ID_TTVXS)
                return QString("Slot:%1").arg(chToSlot(st.readerCh));
            else
                return QString("Ch:%1").arg(st.readerCh);
        } else
            return st.getSlaveIndex().getIdent();
    }
    return QVariant();
}

void FeLinkStatusModel::setStatusList(const QList<FeLinkStatus> &statusList_)
{
    auto sortedList = statusList_;
    std::sort(sortedList.begin(), sortedList.end(), &sortFeLinkStatus);

    auto ch_old = statusList->size();
    auto ch_new = sortedList.size();
    bool fullReset = ch_old != ch_new;

    QSet<int> colsToReset;
    if(fullReset) {
        beginResetModel();
    } else {
        int col=0;
        auto stOldIt=statusList->cbegin();
        for(auto stIt=sortedList.cbegin(); stIt!=sortedList.cend(); ++stIt) {
            if((*stIt) != (*stOldIt))
                colsToReset << col;
            ++stOldIt;
            ++col;
        }
    }

    (*statusList) = sortedList;

    if(fullReset) {
        endResetModel();
    } else {
        for(auto col : colsToReset)
            invalidateCol(col);
    }
}

void FeLinkStatusModel::invalidateCol(int col)
{
    if (col == -1)
        return;

    auto indexFrom = createIndex(0, col);
    auto indexTo = createIndex(rowCount()-1, col);
    emit dataChanged(indexFrom, indexTo);
    emit headerDataChanged(Qt::Horizontal, col,col);
}

QVariant FeLinkStatusModel::data_BackgroundRole(const FeLinkStatus &st, int row) const
{
    switch (row) {
    case ROW_STATUS: return st.status.allOk() ? QColor(Qt::darkGreen) : QColor(Qt::red);
    case ROW_TAI: return st.tai_valid ? QColor(Qt::darkGreen) : QColor(Qt::red);
    }
    return QVariant();
}

QVariant FeLinkStatusModel::data_DisplayRole(const FeLinkStatus &st, int row) const
{
    switch (row) {
    case ROW_REMOTE_INDEX: return st.remoteIndex.getIdent();
    case ROW_STATUS: return st.status.allOk() ? st.status.getInfoStr() : QString("status: 0x%1").arg(st.status.data.all, 0, 16);
    case ROW_BITSLIDE: return st.RX_Bitslide;
    case ROW_REMOTE_VER: return st.remoteVer.getInfoStr();
    case ROW_TAI: return QDateTime::fromSecsSinceEpoch(st.tai_sec).toString();
    case ROW_CUR_LINK_STATE_TIME: return getStateTimeHr(st.Current_link_state_time);
    case ROW_CNT_RX_CODE_ERR: return st.RX_code_error_counter;
    case ROW_CNT_RX_CRC_ERR: return st.RX_CRC_error_counter;
    case ROW_CNT_TIME_FRAMES_SENT: return st.Time_frames_sent_counter;
    case ROW_CNT_TRIG_FRAMES_SENT: return st.Trigger_frames_sent_counter;
    case ROW_CNT_TIME_FRAMES_RECV: return st.Time_frames_received_counter;
    case ROW_CNT_TRIG_FRAMES_RECV: return st.Trigger_frames_received_counter;
    case ROW_CNT_UNKNOWN_FRAMES: return st.Unknown_frames_received_counter;
    }

    return QVariant();
}

QVariant FeLinkStatusModel::data_ToolTipRole(const FeLinkStatus &st, int row) const
{

    switch (row) {
    case ROW_STATUS: return st.status.getInfoStr();
//    case ROW_TAI: return QDateTime::fromSecsSinceEpoch(st.tai_sec).toString();
    case ROW_CUR_LINK_STATE_TIME: return st.Current_link_state_time;
    }

    return QVariant();
}
