#ifndef DEVICEDATA
#define DEVICEDATA

#include "base/DeviceIndex.h"
#include "lib-common/MpdRawTypes.h"

#include <QElapsedTimer>
#include <QMap>
#include <QVariant>
#include <QVector>

namespace  {
const double HPTDC_TS_STEP_NS = 0.024/1.024;
}


/*
// ======= Data types =======
TAI

BaseEvent
    TAI

// --- TTVXS ---

ttvxs_trig_src_t


TtvxsEvent : public BaseEvent
    ttvxs_trig_src_t

// --- TDC ---

TdcEdge

TdcChHit

// --- ADC ---

AdcChHit

TqdcChHits
    QVector<TdcChHit>
    QVector<AdcChHit>

TqdcEvent : public BaseEvent
    QMap<int, TqdcChHits>

// --- DevStat ---

DevStat

CompleteEvent
    QMap<DeviceIndex, TqdcEvent> // for: TDC72-VXS TDC64-VHLE TQDC16-VS ADC64-VE
    QMap<DeviceIndex, TtvxsEvent> // for: TTVXS-RC
    QMap<DeviceIndex, DevStat>

DecodedData
    QMap<quint32, CompleteEvent>

// ======= Statistic types =======

AnalyzerChStat

AnalyzerDevStat
    QMap<int, AnalyzerChStat>

AnalyzerStat
    QMap<DeviceIndex, AnalyzerDevStat>

struct WorkerStat
*/


// ======= Data types =======

struct  TAI
{
    TAI(const TaiTimeBlock *tsTAI) {
        flags = tsTAI->taiFlags;
        sec = tsTAI->wfTsHi;
        nSec = tsTAI->wfTsLow;
    }
    TAI(){ clear(); }
    void clear() {
        sec=0;
        flags=0;
        nSec=0;
    }

    quint32 sec;
    quint32 flags:2;
    quint32 nSec:30;
};

struct  BaseEvent
{
    BaseEvent(){ clear(); }
    virtual void clear() {
        index.clear();
        evNum=0;
        tai.clear();
    }

    DeviceIndex index;
    quint32 evNum;
    TAI tai;
};

// --- TTVXS ---

typedef union {
    unsigned int all;
    struct {
        unsigned int pulser: 1; // no connection
        unsigned int random_pulser: 1; // link data error
        unsigned int external_ttl: 1; // pci bus signaled error
        unsigned int digital_sfp: 1; // error propagation
        unsigned int reserved: 28; // no clock from remote
    } bit;
} ttvxs_trig_src_t;


struct  TtvxsEvent : public BaseEvent
{
    ttvxs_trig_src_t trig_src;
};

// --- TDC ---

struct  TdcEdge
{
    TdcEdge(quint16 ts, quint32 t, bool l):time(t),tdcTs(ts),rising(l){}
    TdcEdge(){ clear(); }
    void clear() {
        time=0;
        tdcTs=0;
        rising=true;
    }

    quint32 time;
    quint16 tdcTs;
    bool rising;
};

struct  TdcChHit
{
    TdcChHit(){ clear(); }
    void clear() {
        ch=0;
        tdcTs=0;
        leadingTime=0;
        trailingTime=0;
        hasLeadTimeInl=false;
        hasTrailTimeInl=false;
        hasWidthTimeInl=false;
    }

    double getLeadTimeNs(const QVector<double> &inlCorr) {
        if(hasLeadTimeInl)
            return leadTimeInl;
        hasLeadTimeInl = true;
        return leadTimeInl=HPTDC_TS_STEP_NS*(inlCorr[leadingTime%1024]+leadingTime);
    }
    double getTrailTimeNs(const QVector<double> &inlCorr) {
        if(hasTrailTimeInl)
            return trailTimeInl;
        hasTrailTimeInl = true;
        return trailTimeInl = HPTDC_TS_STEP_NS*(inlCorr[trailingTime%1024]+trailingTime);
    }
    double getHitTime(const QVector<double> &inlCorr) {
        return leadingTime ? getLeadTimeNs(inlCorr) : getTrailTimeNs(inlCorr);
    }
    double getWidthNs(const QVector<double> &inlCorr) {
        if(hasWidthTimeInl)
            return widthTimeInl;
        hasWidthTimeInl=true;
        return widthTimeInl=(getTrailTimeNs(inlCorr) - getLeadTimeNs(inlCorr));
    }

    quint8 ch;
    quint16 tdcTs;
    quint32 leadingTime; // in HPTDC ticks
    quint32 trailingTime; // in HPTDC ticks
    double leadTimeInl; // in ns
    bool hasLeadTimeInl;
    double trailTimeInl; // in ns
    bool hasTrailTimeInl;
    double widthTimeInl; // in ns
    bool hasWidthTimeInl;
};

// --- ADC ---

struct AdcChHit
{
    AdcChHit () { clear(); }
    void clear() {
        ch = 0;
        adcTs = 0;
        taiTime.clear();
        mode = 0;
        wf.clear();
    }

    quint8 ch;
    quint16 adcTs; // 8ns intervals from event TS
    TAI taiTime;
    uint mode; // 1-cal; 2-samp; 3-sumQ
    std::vector<double> wf;
//    quint32 sumQ; // 18-bit, sum of adc samples
};

struct  TqdcChHits
{
    QVector<TdcChHit> tdcHits;
    QVector<AdcChHit> adcHits;
};

struct  TqdcEvent : public BaseEvent
{
    TqdcEvent(){ clear(); }
    void clear() {
        BaseEvent::clear();
        hits.clear();
        temp.clear();
        fw_rev.clear();
        bmc_rev.clear();
        errors.clear();
        chMask = 0;
    }

    QMap<int, TqdcChHits> hits; // per channel
    QMap<int, double> temp; // per tempId
    QVariant fw_rev;
    QVariant bmc_rev;
    QVector<quint32> errors; // full 32-bit hptdc words (0x<HptdId><HptdcErr>)
    quint64 chMask;
};

// --- DevStat ---

struct DevStat {
    DevStat() { clear(); }
    void clear() {
        index.clear();
        regIoTimeout = 0;
        regIoError = 0;
        stats.clear();
    }

    DeviceIndex index;
    bool regIoTimeout;
    bool regIoError;
    QMap<quint16, quint16> stats; // stat from regs: <regAdr, regVal>
};

struct CompleteEvent
{
    void clear() {
        evNum = 0;
        tqdcEv.clear();
        ttvxsEv.clear();
        devStats.clear();
    }

    quint32 evNum;
    QMap<DeviceIndex, TqdcEvent> tqdcEv; // for: TDC72-VXS TDC64-VHLE TQDC16-VS ADC64-VE
    QMap<DeviceIndex, TtvxsEvent> ttvxsEv; // for: TTVXS-RC
    QMap<DeviceIndex, DevStat> devStats;
};

struct DecodedData
{
    DecodedData () { clear(); }
    void clear() {
        events.clear();
        event_cnt = 0;
    }

    QMap<quint32, CompleteEvent> events;
    quint32 event_cnt;
};
Q_DECLARE_METATYPE(DecodedData)

// ======= Statistic types =======

struct AnalyzerChStat
{
    void clear() { adc_hits = 0; tdc_hits = 0;}
    quint32 adc_hits;
    quint32 tdc_hits;
};

struct AnalyzerDevStat
{
    void clear() { dev_ch_hits.clear(); err_cnt.clear(); }
    QMap<int, AnalyzerChStat> dev_ch_hits; // per channel
    QMap<quint32, QList<QPair<quint32,quint32> > > err_cnt; //stores evNum ranges per error word (<0x<HptdId><HptdcErr>, {<100,150>, {200,250}, ...}>)
};

struct AnalyzerStat
{
    AnalyzerStat ():
        lastEvNum(0)
    {
        clear();
    }
    void clear() {
        eventDecoded=0;
        sizeDecoded=0;
        decodeTimeNs=0;
        decodeLoop=0;
        workerLoop=0;
        formatErrorCnt=0;
        dev_stat.clear();
        ttvxsList.clear();

        timer.invalidate();
        skip=0;
        pass=0;
    }

    qint64 eventDecoded;
    qint64 sizeDecoded;
    quint32 lastEvNum;
    qint64 decodeTimeNs; // decode time 200-300 us per max adcWin
    qint64 workerLoop;
    qint64 decodeLoop;
    int formatErrorCnt;
    QMap<DeviceIndex, AnalyzerDevStat> dev_stat;
    QList<DeviceIndex> ttvxsList;

    QElapsedTimer timer;
    qint64 skip;
    qint64 pass;
};
Q_DECLARE_METATYPE(AnalyzerStat)

struct WorkerStat
{
    WorkerStat ()
    {
        clear();
    }
    void clear() {
        workerTimeMs=0;
        decodeLoop=0;

    }
    qint64 workerTimeMs;
    qint64 decodeLoop;
};
Q_DECLARE_METATYPE(WorkerStat)

#endif // DEVICEDATA

