#include "HistWorker.h"
#include "DeviceData.h"
#include "DeviceHelper.h"
#include "HistMngrCfg.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QSemaphore>
#include <QTimer>

#include "mregdev/qtmregdevice.h"

HistWorker::HistWorker(QSemaphore *_histSem) :
    QObject(nullptr),
    histSem(_histSem),
    refreshHistTimer(new QTimer(this))
{
    qRegisterMetaType<WorkerStat>();
    qRegisterMetaType<InlResults>();
    connect(refreshHistTimer, &QTimer::timeout, [=](){ refreshCurrentPlot(); });
    refreshHistTimer->start(500);
}

HistWorker::~HistWorker()
{
    destroy_flg = true;
    trashAllData();
}

void HistWorker::requestInlResults(bool allRes) const
{
    if(allRes){
        emit inlResultsUpdated(inlResults, allRes);
        return;
    }
    InlResults inlRes;
    if(!config.is_inl_enabled) return;

    QVector<double> codeDensData(1024);

    QMap<DeviceIndex, DeviceCalibration> devicesCal;
    for(const DeviceIndex &index : knownDevices) {
        devicesCal[index] = inlResults.value(index, DeviceCalibration(index));
    }

    for(HistKey hist_key: histogramms.keys()){
        if(hist_key.hist_id == HistId::inl){
            DeviceIndex index = hist_key.index;
            int ch = hist_key.ch;

            const Histogramm &curHist = histogramms[hist_key];
            const int nBins = curHist.n_bins;
            if(nBins != 1024) continue;
            for (int bin = 0; bin < nBins; ++bin) {
                codeDensData[bin] = curHist.bins[bin];
            }
            // inlRes[index][ch] = makeDNLCorrectionTable(codeDensData);
            devicesCal[index].inl[ch] = codeDensData;
            codeDensData.clear();
            codeDensData.resize(1024);

        }
    }

    for(const DeviceIndex &index : knownDevices) {
        if(!devicesCal[index].inl.isEmpty()) {
            if(devErrors.contains(index))
                devicesCal[index].errMask = devErrors[index];
            inlRes[index] = devicesCal[index];
        }
    }

    emit inlResultsUpdated(inlRes);
}

void HistWorker::tdcTtvxMapUpdated(TdcTtvxsMapType ttMap)
{
    tdcTtvxsMap = ttMap;
}

void HistWorker::inlConfigUpdated(InlResults inl)
{
    fullDevCorrMap = inl;
    corrFillCurDevMap();
}

void HistWorker::corrFillCurDevMap()
{
    auto devList = knownDevices;
    if(!config.stop_dev_index.isNull()) {
        devList << config.stop_dev_index;
    }
    // Fill curDevCorrMap from fullDevCorrMap according to devList
    for(auto index : devList) {
        auto &dev_cal = curDevCorrMap[index];
        const int nch = QtMregDevice::getDeviceChNum(index);
        if(nch<=0) {
            qCritical() << __PRETTY_FUNCTION__
                        << "Can't detect nch for dev:"
                        << index.getDeviceIdName();
            Q_ASSERT(false);
        }
        for(int ch=0; ch<nch; ++ch) {
            auto &ch_cal = dev_cal.inl[ch];
            if(ch_cal.isEmpty()){
                ch_cal = fullDevCorrMap.value(index).inl
                    .value(ch, QVector<double>(1024, 0));
            }
        }
    }
}

void HistWorker::clearAll()
{
    knownDevices.clear();
    curDevCorrMap.clear();
    dev_ch_offset = config.dev_ch_offset;

    stat.clear();
    for(auto key : histogramms.keys()) {
        if(HistGroupCfg::isBaseHist(key.hist_id)) {
            histogramms[key].reset();
        } else {
            histogramms.remove(key);
        }
    }

    clearTaiHist();
    inlResults.clear();
}

void HistWorker::clearHist(const HistKey &hist_key)
{
    for(auto it=histogramms.begin(); it!=histogramms.end(); ++it) {
        if(!hist_key.isEmpty() && !(it.key() == hist_key))
            continue;
        if(it.key().hist_id == HistId::tdcTaiTime)
            clearTaiHist();
        else
            it->reset();
    }

    refreshCurrentPlot(true);
    inlResults.clear();
}

void HistWorker::handle_data()
{
    emit workerStatUpdated(stat);
    data_processing = true;
    while(!data_list.isEmpty()) {
        if(destroy_flg || !config.global_enabled) {
            trashAllData();
            break;
        }
        DecodedData data = data_list.takeFirst();
        histSem->release();

        int tdcHits;
        int adcHits;
        int ev_cnt=0;
        auto evIt = data.events.begin();
        while(evIt != data.events.end()) {
            CompleteEvent &ev = evIt.value();
            if(++ev_cnt>1000) {
                ev_cnt = 0;
                // for quick handle clean & destroy signals
                QCoreApplication::processEvents(QEventLoop::AllEvents, 10);
            }
            if(destroy_flg)
                break;
            tdcHits=0;
            adcHits=0;
            dataProcessingSingleEv(ev, tdcHits, adcHits);

            auto tdcHitEvIt = histogramms.find(HistKey(HistId::tdcHitEv));
            if(tdcHitEvIt != histogramms.end()) {
                if(isHistGroupEn(HistId::tdcHitEv)) {
                    tdcHitEvIt->fill(tdcHits);
                }
            }
            auto adcHitEvIt = histogramms.find(HistKey(HistId::adcHitEv));
            if(adcHitEvIt != histogramms.end()) {
                if(isHistGroupEn(HistId::adcHitEv)) {
                    adcHitEvIt->fill(adcHits);
                }
            }
            ++evIt;
        }
    }
    if(!destroy_flg){
        refreshCurrentPlot();
    }
    data_processing = false;
}

void HistWorker::dataUpdated(const DecodedData &data)
{
    if(destroy_flg) {
        histSem->release();
        return;
    }

    data_list << data;
    if(data_processing)
        return;

    QElapsedTimer worker_timer;
    worker_timer.start();

    handle_data();

    stat.decodeLoop++;
    stat.workerTimeMs += worker_timer.elapsed();
}

void HistWorker::setHistConfig(HistMngrCfg config_)
{
    if(config.hist_groups == config_.hist_groups &&
            config.dev_ch_offset == config_.dev_ch_offset &&
            config.tai_ts_window == config_.tai_ts_window &&
            config.mw_min == config_.mw_min &&
            config.mw_max == config_.mw_max) {
        config = config_;
        return;
    }

    // At this moment nBins/min/max can't be changed (no need to resize)
    // for all histograms except tdcTime/tdcCh
//    for(auto hCfgIt=config_.hist_groups.cbegin(); hCfgIt!=config_.hist_groups.cbegin(); ++hCfgIt) {
//        const auto &histId = hCfgIt.key();
//        if(histId == HistId::tdcCh)
//            continue;
//        auto hCfgOldIt=config.hist_groups.constFind(histId);
//        if(hCfgOldIt==config.hist_groups.cend())
//            continue;
//        const auto &oldInfo = config.getHistGroupInfo(histId);
//        const auto &newInfo = config_.getHistGroupInfo(histId);
//        if((newInfo.n_bins == oldInfo.n_bins) &&
//                (newInfo.min == oldInfo.min) &&
//                (newInfo.max == oldInfo.max))
//            continue;
//        for(auto hIt=histogramms.begin(); hIt!=histogramms.end(); ++hIt) {
//            if(hIt.key().hist_id == histId)
//                hIt->resize(newInfo.n_bins, newInfo.min, newInfo.max);
//        }
//    }

    // Save state
    auto devChCnt = getDevChCntMap();

    config = config_;
    //todo: merge knownDevices with config.dev_ch_offset / this->dev_ch_offset

    for(auto hCfg : config.hist_groups) {
        if(hCfg.isBaseHist()) {
            addBaseHist(hCfg.hist_id);
        } else {
            for(auto devIndex : knownDevices)
                for(int ch=0; ch<QtMregDevice::getDeviceChNum(devIndex); ++ch)
                    addDeviceChHist(hCfg.hist_id, devIndex, ch);
        }
    }
    for(auto histKey : histogramms.keys()) {
        if(!config.hist_groups.contains(histKey.hist_id))
            histogramms.remove(histKey);
    }

    // Restore state
    auto chHistIt = histogramms.find(HistKey(HistId::tdcCh));
    if(chHistIt != histogramms.end()) {
        int maxChNum = HistMngrCfg::getMaxChNum(dev_ch_offset);
        chHistIt->resize(maxChNum, 0, maxChNum);

        fillDevChHist(devChCnt);
    }
    auto taiHistIt = histogramms.find(HistKey(HistId::tdcTaiTime));
    if(taiHistIt != histogramms.end()) {
        if(taiHistIt->n_bins)
            recreateTaiHist(taiReperSec);
    }
    auto tdcTimeIt = histogramms.find(HistKey(HistId::tdcTime));
    if(tdcTimeIt != histogramms.end()) {
        const auto &hInfo = config.getHistGroupInfo(HistId::tdcTime);
        tdcTimeIt->resize(hInfo.n_bins, hInfo.min, hInfo.max);
    }
}

void HistWorker::refreshCurrentPlot(bool force)
{
    emit hist_updated(histogramms, force);
}

void HistWorker::addDevice(const DeviceIndex &index)
{
    if(knownDevices.contains(index)) return;

    const int nch = QtMregDevice::getDeviceChNum(index);
    auto devChCnt = getDevChCntMap();

     //Add INL/INL-check hist
    for(auto hCfg : config.hist_groups) {
        if(hCfg.isBaseHist())
            continue;
        for(int ch=0; ch<nch; ++ch)
            addDeviceChHist(hCfg.hist_id, index, ch);
    }

    //TODO:: find-out this asserrtt
    Q_ASSERT(!dev_ch_offset.contains(index));
    knownDevices << index;
    corrFillCurDevMap();
    int maxChNum = HistMngrCfg::getMaxChNum(dev_ch_offset);

    auto it = dev_ch_offset.find(index);
    if(it == dev_ch_offset.end()) {
        dev_ch_offset[index] = maxChNum;
        maxChNum += nch;
    }

    auto chHistIt = histogramms.find(HistKey(HistId::tdcCh));
    if(chHistIt != histogramms.end()) {
        chHistIt->resize(maxChNum, 0, maxChNum);
        fillDevChHist(devChCnt);
    }

    auto hitHistIt = histogramms.find(HistKey(HistId::tdcHitEv));
    if(hitHistIt != histogramms.end()) {
        auto &bins = hitHistIt->bins;
        auto old_bins = bins;
        int maxHit = maxChNum ? maxChNum*100 : 100;
        hitHistIt->resize(maxHit, 0, maxHit);
        old_bins.resize(bins.size());
        bins = old_bins;
        // Remember that hitHistIt->over_flow can be not empty
    }

    emit devListUpdated(dev_ch_offset, histogramms);
}

void HistWorker::addBaseHist(HistId id)
{
    const auto &key = HistKey(id);
    if(histogramms.contains(key))
        return;

    auto cfgIt = config.hist_groups.constFind(id);
    if(cfgIt == config.hist_groups.cend())
        return;

    const auto &info = config.getHistGroupInfo(id);
    histogramms[key] = Histogramm(info.title, info.n_bins, info.min, info.max);
}

void HistWorker::addDeviceChHist(HistId id, const DeviceIndex &index, int ch)
{
    const auto &key = HistKey(id, index, ch);
    if(histogramms.contains(key))
        return;

    auto cfgIt = config.hist_groups.constFind(id);
    if(cfgIt == config.hist_groups.cend())
        return;

    const auto &info = config.getHistGroupInfo(id);
    const QString &title = QString("%1 %2 ch:%3")
            .arg(info.title).arg(index.getIdent()).arg(ch);
    histogramms[key] = Histogramm(title, info.n_bins, info.min, info.max);
}

void HistWorker::dataProcessingSingleEv(CompleteEvent &ev,
                                        int &tdcHitsPerEv,
                                        int &adcHitsPerEv)
{
    curEv = &ev;
    TqdcEvent tqdcStopEvent;

    if(!config.stop_dev_index.isNull() && ev.tqdcEv.contains(config.stop_dev_index)) {
        TqdcEvent tqdcEvent = ev.tqdcEv[config.stop_dev_index];
        if(tqdcEvent.hits.contains(config.stop_ch) &&
                !tqdcEvent.hits[config.stop_ch].tdcHits.isEmpty()) {
            tqdcStopEvent = tqdcEvent;
        }
    }

    TqdcChHits &stopHits = tqdcStopEvent.hits[config.stop_ch];
    auto tqdcEvIt = ev.tqdcEv.begin();
    while(tqdcEvIt != ev.tqdcEv.end()) {
        TqdcEvent &tqdcEv = tqdcEvIt.value();
        if(!knownDevices.contains(tqdcEv.index)) {
            addDevice(tqdcEv.index);
        }
        dataProcessingTqdc(tqdcEv,
                           tdcHitsPerEv,
                           adcHitsPerEv,
                           tqdcStopEvent,
                           stopHits.tdcHits);
        ++tqdcEvIt;
    }
    curEv = nullptr;
}

void HistWorker::dataProcessingTqdc(TqdcEvent &ev,
                                    int &tdcHitsPerEv,
                                    int &adcHitsPerEv,
                                    TqdcEvent &tqdcStopEvent,
                                    QVector<TdcChHit> &stopHits)
{
    // TQDC processing
    auto tqdcChIt = ev.hits.begin();
    while(tqdcChIt != ev.hits.end()) {
        const auto &ch = tqdcChIt.key();
        auto &tdcHits = tqdcChIt.value().tdcHits;
        if(isHistGroupEn(HistId::inlCheck) &&
           (ev.index != config.stop_dev_index || ch != config.stop_ch))
            fillDoublePulse(HistKey(HistId::tdcDoublePulseTime),
                            tqdcStopEvent,
                            stopHits,
                            ev,
                            tdcHits);
        if(tdcHits.size() != 0)
            dataProcessingTdc(ev, ch, tdcHits, tdcHitsPerEv);

        auto &adcHits = tqdcChIt.value().adcHits;
        if(adcHits.size() != 0)
            dataProcessingAdc(ev, ch, adcHits, adcHitsPerEv);

        ++tqdcChIt;
    }

    dataProcessingTai(ev);
    grabStatInfo(ev);
    errorsProcessing(ev);
}

void HistWorker::dataProcessingTdc(TqdcEvent &ev,
                                   const int &ch,
                                   QVector<TdcChHit> &tdcHits,
                                   int &tdcHitsPerEv)
{
    tdcHitsPerEv += tdcHits.size();

    for (auto hitIt = tdcHits.begin(); hitIt != tdcHits.end(); ++hitIt) {
        TdcChHit &tdcHit = *hitIt;

        if(isHistGroupEn(HistId::tdcCh))
            histogramms[HistKey(HistId::tdcCh)]
                .fill(ch + dev_ch_offset[ev.index]);
        if(isHistGroupEn(HistId::tdcTime)) {
            if(tdcHit.hasLeadTimeInl)
                histogramms[HistKey(HistId::tdcTime)]
                    .fill(tdcHit.leadTimeInl);
            else if(tdcHit.hasTrailTimeInl)
                histogramms[HistKey(HistId::tdcTime)]
                    .fill(tdcHit.trailTimeInl);
            else
                histogramms[HistKey(HistId::tdcTime)]
                    .fill(tdcHit.getHitTime(getInl(ev, tdcHit)));
        }

        if(isHistGroupEn(HistId::tdcWidth) &&
           tdcHit.leadingTime &&
           tdcHit.trailingTime)
            histogramms[HistKey(HistId::tdcWidth)]
                .fill(tdcHit.hasWidthTimeInl ?
                      tdcHit.widthTimeInl :
                      tdcHit.getWidthNs(getInl(ev, tdcHit)));

        if(isHistGroupEn(HistId::inl))
            histogramms[HistKey(HistId::inl, ev.index, ch)]
                .fill(tdcHit.leadingTime%1024);
    }
}

void HistWorker::dataProcessingAdc(TqdcEvent &ev,
                                   const int &ch,
                                   QVector<AdcChHit> &adcHits,
                                   int &adcHitsPerEv)
{
    adcHitsPerEv += adcHits.size();

    Histogramm *adcChHist = isHistGroupEn(HistId::adcCh)
            ? &histogramms[HistKey(HistId::adcCh)]
            : nullptr;
    Histogramm *adcTimeHist = isHistGroupEn(HistId::adcTime)
            ? &histogramms[HistKey(HistId::adcTime)]
            : nullptr;
    Histogramm *adcChSamplesHist = isHistGroupEn(HistId::chSamples)
            ? &histogramms[HistKey(HistId::chSamples, ev.index, ch)]
            : nullptr;
    Histogramm *adcChAmplHist = isHistGroupEn(HistId::chAmpl)
            ? &histogramms[HistKey(HistId::chAmpl, ev.index, ch)]
            : nullptr;

    for(auto hitIt = adcHits.begin(); hitIt != adcHits.end(); ++hitIt) {
        AdcChHit &adcChHit = *hitIt;
        if(adcChHist)
            adcChHist->fill(ch + dev_ch_offset[ev.index]);
        if(adcTimeHist)
            adcTimeHist->fill(adcChHit.adcTs*8);

        double ampl =0;
        bool has_any_sample = false;
        for(const auto &sample : adcChHit.wf) {
            if(has_any_sample) {
                ampl = config.ampl_hist_with_max
                    ? std::max(sample, ampl) : std::min(sample, ampl);
            } else {
                has_any_sample = true;
                ampl = sample;
            }
            if(adcChSamplesHist) {
                adcChSamplesHist->fill(sample);
            }
        }
        if(has_any_sample && adcChAmplHist)
            adcChAmplHist->fill(ampl);
    }
}


void HistWorker::dataProcessingTai(TqdcEvent &ev)
{
    // TAI processing
    if(ev.tai.flags == 2 &&
        histogramms.contains(HistId::tdcTaiTime) &&
        isHistGroupEn(HistId::tdcTaiTime)){
        Histogramm &hist = histogramms[HistId::tdcTaiTime];

        if(hist.n_bins == 0) {
            recreateTaiHist(ev.tai.sec);
        }
        double dt = ev.tai.sec-taiReperSec;
        dt += ev.tai.nSec*1e-9;
        hist.fill(dt);
    }
}

void HistWorker::grabStatInfo(TqdcEvent &ev)
{
    // Grab stat info for INL results
    if(config.is_inl_enabled) {
        for (auto errWord: ev.errors) {
            auto &hptdcMap = inlResults[ev.index].errMask;
            const uint hptdcId = (errWord>>24) & 0xF;
            const quint16 errMask = errWord & 0x3FFF;
            hptdcMap[hptdcId] |= errMask;
        }
        if(!ev.fw_rev.isNull()) {
            inlResults[ev.index].fw_versions.insert(ev.fw_rev.toUInt());
        }
        if(!ev.bmc_rev.isNull()) {
            inlResults[ev.index].bmc_versions.insert(ev.bmc_rev.toUInt());
        }
        auto &minTemp = inlResults[ev.index].minTemp;
        auto &maxTemp = inlResults[ev.index].maxTemp;
        auto &tempStat = inlResults[ev.index].tempStat;
        for (auto temp_id: ev.temp.keys()) {
            const auto &temp = ev.temp[temp_id];
            auto &tempStatOne = tempStat[temp_id];
            tempStatOne.addTemp(temp);

            if(!tempStatOne.initialized)
                continue;
            if(!tempStatOne.temp_valid(temp)) {
                qWarning().noquote() <<
                    QString("[%1] Bad temp: temp_id=%2 mean=%3 temp=%4")
                        .arg(ev.index.getSerialStr())
                        .arg(temp_id)
                        .arg(tempStatOne.mean, 0, 'f', 2)
                        .arg(temp, 0, 'f', 2);
                continue;
            }
            if(minTemp.contains(temp_id))
                minTemp[temp_id] = std::min(minTemp[temp_id], temp);
            else
                minTemp[temp_id] = temp;
            if(maxTemp.contains(temp_id))
                maxTemp[temp_id] = std::max(maxTemp[temp_id], temp);
            else
                maxTemp[temp_id] = temp;
        }
    }
}

void HistWorker::errorsProcessing(TqdcEvent &ev)
{
    // Errors processing
    if(!ev.errors.empty() &&
       histogramms.contains(HistKey(HistId::errors)) &&
       isHistGroupEn(HistId::errors)) {
        auto &histErr = histogramms[HistKey(HistId::errors)];
        std::vector<int> errorsBits;
        errorsBits.resize(16, 0);
        for (auto errWord: ev.errors) {
            errWord &= 0x3FFF;
            for(ulong bit=0; errWord; ++bit){
                if(errWord&1){
                    ++errorsBits[bit];
                }
                errWord = (errWord>>1);
            }
        }
        for(ulong bit=0; bit<16; ++bit){
            if(errorsBits[bit])
                histErr.fill(bit, errorsBits[bit]);
        }
    }
}

void HistWorker::fillDoublePulse(HistKey hist_key,
                                 TqdcEvent &tdcStopEvent,
                                 QVector<TdcChHit> &stopHits,
                                 TqdcEvent &ev,
                                 QVector<TdcChHit> &tdcEvents)
{
    const TAI *stopTai = &tdcStopEvent.tai;
    const TAI *evTai = &ev.tai;
    TdcTtvxsMapType::const_iterator it;
    // TODO: Move look for stopTai to upper method
    QMap<DeviceIndex, TtvxsEvent>::const_iterator it1;

    if((it=tdcTtvxsMap.constFind(tdcStopEvent.index)) != tdcTtvxsMap.constEnd()) {
        it1 = curEv->ttvxsEv.find(it.value());
        if(it1 != curEv->ttvxsEv.constEnd()){
            stopTai = &it1.value().tai;
        }
    }
    if((it=tdcTtvxsMap.constFind(ev.index)) != tdcTtvxsMap.constEnd()) {
        it1 = curEv->ttvxsEv.find(it.value());
        if(it1 != curEv->ttvxsEv.constEnd()) {
            evTai = &it1.value().tai;
        }
    }

    if(evTai->flags != 2 || stopTai->flags!=2)
        return;

    auto stopHitIt = stopHits.begin();
    while(stopHitIt != stopHits.end()) {
        TdcChHit &stopHit = *stopHitIt;
        const double stopTimeNs =
            stopHit.hasLeadTimeInl ?
            stopHit.leadTimeInl :
            stopHit.getLeadTimeNs(getInl(tdcStopEvent, stopHit));

        double dtTai = 0; // in ns
        dtTai += (evTai->sec - stopTai->sec)*1e9;
        dtTai += evTai->nSec - stopTai->nSec;
        auto tdcHitIt = tdcEvents.begin();
        while(tdcHitIt != tdcEvents.end()) {
            TdcChHit &tdcHit = *tdcHitIt;
            const double hitTimeNs =
                tdcHit.hasLeadTimeInl ?
                tdcHit.leadTimeInl :
                tdcHit.getLeadTimeNs(getInl(ev, tdcHit));

            double dt = dtTai+hitTimeNs-stopTimeNs;
            histogramms[hist_key].fill(dt);
            HistKey inl_check_key = HistKey(HistId::inlCheck, ev.index, tdcHit.ch);
            histogramms[inl_check_key].fill(dt);

            ++tdcHitIt;
        }
        ++stopHitIt;
    }
}

QVector<double> &HistWorker::getInl(const TqdcEvent &ev, const TdcChHit &tdcEv)
{
    return getInl(ev.index, tdcEv.ch);
}

QVector<double> &HistWorker::getInl(const DeviceIndex &index, uint ch)
{
    return curDevCorrMap[index].inl[ch];
}

void HistWorker::recreateTaiHist(int reper)
{
    taiReperSec = reper;

    Histogramm &hist = histogramms[HistKey(HistId::tdcTaiTime)];
    const auto &winSec = config.tai_ts_window;
    hist.resize((winSec ? winSec*10 : 0), 0, winSec);

    refreshCurrentPlot(true);
    emit taiReperUpdated(taiReperSec);
}

void HistWorker::clearTaiHist()
{
    taiReperSec = 0;
    auto taiHistIt = histogramms.find(HistKey(HistId::tdcTaiTime));
    if(taiHistIt != histogramms.end())
        taiHistIt->resize(0, 0, 0);
    emit taiReperUpdated(taiReperSec);
}

QMap<DeviceIndex, QMap<uint, uint> > HistWorker::getDevChCntMap()
{
    QMap<DeviceIndex, QMap<uint, uint>> devChCnt;
    auto chHistIt = histogramms.constFind(HistKey(HistId::tdcCh));
    if(chHistIt == histogramms.constEnd())
        return devChCnt;

    for (const DeviceIndex &i: dev_ch_offset.keys()) {
        const uint &chOffset = dev_ch_offset[i];
        const uint nch = QtMregDevice::getDeviceChNum(i);
        for(uint ch=0; ch<nch; ++ch) {
            const int globCh = static_cast<int>(chOffset+ch);
            const uint chCnt = static_cast<uint>(chHistIt->bins.value(globCh));
            if(chCnt)
                devChCnt[i][ch] = chCnt;
        }
    }
    return devChCnt;
}

void HistWorker::fillDevChHist(QMap<DeviceIndex, QMap<uint, uint> > map)
{
    auto hIt = histogramms.find(HistKey(HistId::tdcCh));
    if(hIt == histogramms.end())
        return;

    for (const DeviceIndex &i: map.keys()) {
        Q_ASSERT(dev_ch_offset.contains(i));
        auto it = dev_ch_offset.constFind(i);
        if(it==dev_ch_offset.cend())
            continue;
        const uint &chOffset = it.value();
        const auto &devChCnt = map[i];
        for(uint ch : devChCnt.keys()) {
            hIt->fill(chOffset+ch, devChCnt[ch]);
        }
    }
}

void HistWorker::trashAllData()
{
    histSem->release(data_list.size());
    data_list.clear();
}

bool HistWorker::isHistGroupEn(HistId groupId) const
{
    return config.hist_groups.value(groupId).hist_group_en;
}
