#include "TQDCAnalyzer.h"

#include <iostream>

#include <QByteArray>
#include <QDateTime>
#include <QDebug>
#include <QSemaphore>
#include <QTimer>

#include "base/bmc/fpga_mcu_regs_addr_tdc.h"
#include "mregdev/device_id.h"
#include "mregdev/mregdevice.h"

#include "DeviceHelper.h"
#include "HistParams.h"
#include "HptdcError.h"

namespace {
// 14-bit signed value, 1/16 degrees Celsius
double decodeTemp_v1(qint32 val)
{
    val &= 0x1FFF;
    val = (val << 19) >> 19;
    return val / 32.0;
}
QString get_raw_data_prefix(const quint32 *word,const void *firstWord)
{
    const long ev_hdr_offset = word - reinterpret_cast<const quint32*>(firstWord);
    return QString("[%1] %2").arg(ev_hdr_offset,3,10,QChar('0')).arg(*word, 8, 16, QChar('0'));
}

const int STATISTIC_TIMEOUT = 500;
const int CLEAR_ERR_STR_STACK_TIMEOUT = 1000;



enum {
    REG_TEMPERATURE = mlink::MregDeviceAdcm::ADCM16_TEMPERATURE_REG, //0x4A, // 74
    REG_FW_REV = mlink::MregDeviceAdcm::ADCM16_FW_REV_REG, //0x4A, // 74
};
}

TQDCAnalyzer::TQDCAnalyzer(QSemaphore *_analyzerSem,
                           QSemaphore *_histSem,
                           QSemaphore *_writerSem, QSemaphore *_zmqSem,
                           QObject *parent) :
    QObject(parent),
    analyzerSem(_analyzerSem),
    histSem(_histSem),
    writerSem(_writerSem),
    zmqSem(_zmqSem),
    statTimer(new QTimer(this)),
    readTimer(new QTimer(this)),
    prevEventNum(0),
    enabled(true),
    signalInverted(false),
    writeDecodedEnabled(false),
    rawDataEvents(0),
    printBadEvent(true),
    rawDataEveryEv(false),
    clear_err_stack(0),
    can_skip(true),
    lookForStop(false),
    stopEvNum(0)
{
    connect(statTimer, &QTimer::timeout, this, &TQDCAnalyzer::sendStat);
    connect(readTimer, &QTimer::timeout, this, &TQDCAnalyzer::read);
    statTimer->start(STATISTIC_TIMEOUT);
}

void TQDCAnalyzer::dataAccuired(QByteArray newData)
{
    if(data.size())
        data.append(newData);
    else
        data = newData;
    if(enabled){
        read();
    }

    analyzerSem->release();
}

void TQDCAnalyzer::read()
{
    curPos = 0;
    curEventStartPos = 0;

    QElapsedTimer decodeTime;
    decodeTime.start();
    quint32 eventDecoded=0;
    while(true) {
        try{
            if(!readEvent()) {
//                qInfo()<<"End of data";
                break;
            }
            ++eventDecoded;
            if(decodedData.event_cnt > DECODED_EVENTS_LIMIT)
                flushData();
        } catch(const char* message) {
            ++analyzerStats.formatErrorCnt;

            QString mess(message);
            if(findNextEventHeader()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
                sendLog(mess);
            } else{
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            //            if(printBadEvent)
        } catch(const QString message) {
            ++analyzerStats.formatErrorCnt;
            QString mess(message);
            if(findNextEventHeader()) {
                qInfo()<<message;
                if(mess!=message)
                    qInfo()<<mess;
                sendLog(message);
            } else {
                prevEventNum=lastEventNumAtPrevFrame;
                break;
            }
            //            if(printBadEvent)
        }
    }

    analyzerStats.decodeTimeNs += decodeTime.nsecsElapsed();
    analyzerStats.eventDecoded += eventDecoded;
    analyzerStats.sizeDecoded += curEventStartPos;
    ++analyzerStats.decodeLoop;

    if(curEventStartPos) {
        if(curEventStartPos==static_cast<uint>(data.size()))
            data.clear();
        else
            data = data.right(data.size()-static_cast<int>(curEventStartPos));
    }

    flushData();
}

void TQDCAnalyzer::reset()
{
    analyzerStats.clear();
    decodedData.clear();
    lastErrStr.clear();
    rawDataTxt.clear();
}

void TQDCAnalyzer::sendStat()
{
    if(CLEAR_ERR_STR_STACK_TIMEOUT < STATISTIC_TIMEOUT * (++clear_err_stack)==0) {
        clear_err_stack = 0;
        lastErrStr.clear();
    }
    printBadEvent = true;
    analyzerStats.timer.start();
    emit deviceStatsUpdated(analyzerStats);
}

void TQDCAnalyzer::startWrite(QString, bool rawData)
{
    writeDecodedEnabled = !rawData;
}

void TQDCAnalyzer::stopWrite()
{
    writeDecodedEnabled = false;
}

void TQDCAnalyzer::requestRawData(int events)
{
    rawDataTxt.clear();
    rawDataEvents = events;
}

void TQDCAnalyzer::setRawDataEveryEv(bool everyEv)
{
    rawDataEveryEv = everyEv;
}

void TQDCAnalyzer::invertUpdated(QMap<DeviceIndex, QVector<bool> > _invertMap)
{
    invertMap = _invertMap;
}

void TQDCAnalyzer::set_can_skip(bool skip)
{
    can_skip = skip;
}

void TQDCAnalyzer::prepareForStop(quint32 evNum)
{
    if(analyzerStats.lastEvNum == evNum){
        emit gotStopEvent(evNum);
    } else {
        lookForStop = true;
        stopEvNum = evNum;
    }
}

bool TQDCAnalyzer::findNextEventHeader()
{
    uint shift = 0;
    const uint dataSize = static_cast<uint>(data.size());
    const char * dataStart = data.data();
    uint pos;
    while(true) {
        shift += 2;
        pos = curEventStartPos+shift;
        if(pos+sizeof(MpdEventHdr) > dataSize) {
            break;
        }
        mpdEventHdr = reinterpret_cast<const MpdEventHdr *>(dataStart+pos);
        if(mpdEventHdr->sync == MPD_SYNC_NORM || mpdEventHdr->sync == MPD_SYNC_EOB) {
            curPos = curEventStartPos + shift;

            QString mess = QString("Syncword was found at 0x%1; Shift:0x%2")
                    .arg(curEventStartPos,0,16).arg(shift,0,16).toStdString().c_str();
            qInfo()<<mess;

            return true;
        }
    }
    curEventStartPos = dataSize;
    return false;
}

bool TQDCAnalyzer::readEvent()
{
    curEventStartPos = curPos;
    quint32 bRead;
    quint32 leftSize;
    if(static_cast<uint>(data.size())<curPos+sizeof(MpdEventHdr))
        return false;

    mpdEventHdr = reinterpret_cast<MpdEventHdr *>(data.data()+curPos);

    if(mpdEventHdr->sync != MPD_SYNC_NORM) {
        if(mpdEventHdr->sync == MPD_SYNC_EOB ||
                mpdEventHdr->sync == MPD_SYNC_RUN_START ||
                mpdEventHdr->sync == MPD_SYNC_RUN_STOP){
            curPos += mpdEventHdr->length + sizeof(MpdEventHdr);
            return true;
        }
        QString mess = QString("Format error: wrong MPD_SYNC_WORD: %1; at file pos:%2")
                .arg(mpdEventHdr->sync, 0, 16).arg(curPos);
        throw(mess.toStdString().c_str());
    }

    if(static_cast<uint>(data.size()) < curPos+mpdEventHdr->length+sizeof(MpdEventHdr))
        return false;

    if(rawDataEveryEv)
        rawDataEvents = 1;
    if(can_skip && (rawDataEveryEv || rawDataEvents==0) && skip_work(analyzerSem)) {
        ++analyzerStats.skip;
        curPos += mpdEventHdr->length + sizeof(MpdEventHdr);
        return true;
    } else {
        ++analyzerStats.pass;
        previousProceed=true;
    }

    if(rawDataEvents) {
//        printRawTxt(sizeof(MpdEventHdr), "MpdEventHdr");
        rawDataTxt << QString("\t%1:").arg("MpdEventHdr");
        printRawWord(&mpdEventHdr->sync,":\t\tSyncWord");
        printRawWord(&mpdEventHdr->length, QString("\t\tlen:%1")
                     .arg(mpdEventHdr->length));
        printRawWord(&mpdEventHdr->evNum, QString("\t\tevNum:%1")
                     .arg(mpdEventHdr->evNum));
    }

    curPos += sizeof(MpdEventHdr);
    leftSize = mpdEventHdr->length;

//    qInfo()<<"Event number is "<<mpdEventHdr->evNum;
    curCompleteEv.clear();
    curCompleteEv.evNum = mpdEventHdr->evNum;
    while(leftSize) {
        bRead = readDevice(leftSize);
        if(leftSize < bRead)
            throw("Format error: Sum of MpdDevice blocks is lager then declared in MpdEvent header");
        leftSize -= bRead;
    }
    decodedData.events[curCompleteEv.evNum] = curCompleteEv;
    ++decodedData.event_cnt;
    curCompleteEv.clear();

    prevEventNum = mpdEventHdr->evNum;
    lastEventNumAtPrevFrame = prevEventNum;
    analyzerStats.lastEvNum = prevEventNum;

    if(rawDataEvents) {
        if(--rawDataEvents == 0) {
            emit rawDataUpdated(rawDataTxt);
            rawDataTxt.clear();
        } else
            rawDataTxt << "==============\n";
    }
    if(lookForStop &&  stopEvNum == prevEventNum) {
        emit gotStopEvent(stopEvNum);
        lookForStop = false;
        stopEvNum = 0;
    }

    return true;
}

void TQDCAnalyzer::flushData()
{
    if(decodedData.events.empty())
        return;

    while (!histSem->tryAcquire(1, 100)) {
        if(!noHistSpace.isValid())
            qInfo()<<"There is no free space for histogram queue";
        noHistSpace.restart();
    }
    if(writeDecodedEnabled) {
        Q_ASSERT(writerSem);
        while (!writerSem->tryAcquire(1, 100)) {
            if(!noWriterSpace.isValid())
                qInfo()<<"There is no free space for writer queue";
            noWriterSpace.restart();
        }
    }
    while (!zmqSem->tryAcquire(1, 100)) {
        if(!noZmqSpace.isValid())
            qInfo()<<"There is no free space for zmq queue";
        noZmqSpace.restart();
    }
    if(noHistSpace.isValid()) {
        if(noHistSpace.elapsed()>5000) {
            qInfo()<<"Restore free space for histogram queue";
            noHistSpace.invalidate();
        }
    }
    if(noWriterSpace.isValid()) {
        if(noWriterSpace.elapsed()>5000) {
            qInfo()<<"Restore free space for writer queue";
            noWriterSpace.invalidate();
        }
    }
    if(noZmqSpace.isValid()) {
        if(noZmqSpace.elapsed()>5000) {
            qInfo()<<"Restore free space for zmq queue";
            noZmqSpace.invalidate();
        }
    }

    emit deviceDataUpdated(decodedData, writeDecodedEnabled);
    decodedData.clear();
}

void TQDCAnalyzer::decodeTdcBlock(const quint32 *payloadPtr)
{
    quint16 tdcWord = 0;
    int tdcWordCnt=0;
    quint16 tdcTS = 0;
    QString wordStr;
    bool hasHptdcError = false;
    QMap<quint8, QVector<TdcEdge> > chEdgeMap;
    const DeviceIndex &index = curTqdcEv.index;
    if(rawDataEvents) {
        printRawTxt(0, "DataBlockHeader (HPTDC)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%1; len:%2")
                     .arg(dataBlockHdr->type()).arg(dataBlockHdr->len));
        if(dataBlockHdr->specific.tdc.fifio_overflow)
            rawDataTxt.last().append(" (event FIFO overflow)");
    }
    while(dataBlockHdr->len > tdcWord*sizeof(quint32)){
        const quint32 &word = *(payloadPtr+tdcWord);
        ++tdcWordCnt;
        const int dt = word >> 28;
//        qInfo() << tdcWord<< ":  " << hex << word << dec;
        wordStr = QString(":\t\t");
        switch (dt) {
        case 2: // tdc ev hdr
        {
            tdcTS = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF; // evNum in TdcHeader
            if(rawDataEvents)
                wordStr.append(QString("ev hdr evNum:%1").arg(tdcEv));
//            qInfo() << "tdc ev hdr: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (mpdEventHdr->evNum & 0xFFF));
        }
            break;
        case 3: // tdc ev trl
        {
            quint16 wordCnt = word & 0xFFF;
            quint16 tdcEv = (word>>12) & 0xFFF;
            if(rawDataEvents)
                wordStr.append(QString("ev trl evNum:%1; words:%2").arg(tdcEv).arg(wordCnt));
//            qInfo() << "tdc ev trl: evNum="<<tdcEv;
//            Q_ASSERT(tdcEv == (mpdEventHdr->evNum & 0xFFF));
            if(tdcWordCnt != wordCnt) {
                int wc = 0;
                const quint32 *devPtr = reinterpret_cast<const quint32 *>(mpdDeviceHdr);
                while(wc*sizeof(quint32) < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length){
                    const quint32 *wordPtr = devPtr+wc;
                    qWarning()  << get_raw_data_prefix(wordPtr, devPtr);
                    ++wc;
                }
            }
            Q_ASSERT(tdcWordCnt == wordCnt);
            tdcWordCnt=0;
        }
            break;
        case 4: // tdc leading edge
//            qInfo() << tdcWord<< ":  " << hex << word << dec;
        case 5: // tdc trl edge
        {
            const quint8 ch = (word>>21)&0x7F;
            const quint32 time = word&0x1FFFFF;
            const bool risening = (dt==5)==signalInverted;
            if(rawDataEvents)
                wordStr.append(QString("%1 ch:%2; t(ps):%L3")
                               .arg(risening?"leading":"trailing").arg(ch,2).arg(24000*time/1024));
            chEdgeMap[ch].push_back(TdcEdge(tdcTS,time, risening));
            break;
        }
        case 6:
        {
            hasHptdcError = true;
            curTqdcEv.errors.push_back(word);
            const int tdcId = (word>>24)&0xF;
            QStringList errList;
            for(int errBit=0; errBit< 15; ++errBit){
                if((word & 0x7FFF) & 1<<errBit) {
                    errList << QString("%1\t\t%2").arg(1<<errBit, 14, 16).arg(HPTDC_ERR_STR[errBit]);
                }
            }
            QString errStr = QString("Got HPTDC error (0x%1 tdcId=%2):\n%3")
                    .arg(mpdDeviceHdr->deviceSerial, 8, 16, QChar('0'))
                    .arg(tdcId).arg(errList.join("\n"));
            if(rawDataEvents)
                wordStr.append(errStr);
            if(!lastErrStr.contains(errStr)) {
                qWarning().noquote() << errStr;
                lastErrStr << errStr;
                if(lastErrStr.size() > 10)
                    lastErrStr.removeFirst();
            }

            break;
        }
        default:
            break;
        }
        if(rawDataEvents)
            printRawWord(&word, wordStr);
        ++tdcWord;
    }
    if(hasHptdcError) {
        qWarning() << "Event header:";
        quint16 wordCnt = 0;
        const quint32 *evPtr = reinterpret_cast<const quint32 *>(mpdEventHdr);
        while(wordCnt*sizeof(quint32) < sizeof(MpdEventHdr)){
            const quint32 *wordPtr = evPtr+wordCnt;
            qWarning()  << get_raw_data_prefix(wordPtr, evPtr);
            ++wordCnt;
        }

        qWarning() << "Device event payload:";
        wordCnt = 0;
        const quint32 *devPtr = reinterpret_cast<const quint32 *>(mpdDeviceHdr);
        while(wordCnt*sizeof(quint32) < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length){
            const quint32 *wordPtr = devPtr+wordCnt;
            qWarning()  << get_raw_data_prefix(wordPtr, devPtr);
            ++wordCnt;
        }
    }


    const auto &evNum = mpdEventHdr->evNum;
    auto &errWordMap = analyzerStats.dev_stat[index].err_cnt;
    for(quint32 word : curTqdcEv.errors) {
        auto &errEvRangesList = errWordMap[word];
        if(errEvRangesList.isEmpty() || errEvRangesList.last().second+1 != evNum) {
            errEvRangesList<< QPair<quint32, quint32>(evNum,evNum);
        } else {
            ++errEvRangesList.last().second;
        }
    }


    // Combine TdcEdge into TdcChHit and fill into
    bool hasTdcChHit = false;
    auto &devInvMap = invertMap[index];
    auto &devHits = analyzerStats.dev_stat[index].dev_ch_hits;
    for (quint8 ch: chEdgeMap.keys()) {
        for (const TdcEdge &edge: chEdgeMap[ch]) {
            if(edge.rising != devInvMap.value(ch, false)) {
                if(hasTdcChHit) {
                    curTqdcEv.hits[ch].tdcHits.push_back(curTdcHit);
                    ++devHits[ch].tdc_hits;
                    curTdcHit.clear();
                } else {
                    hasTdcChHit = true;
                }
                curTdcHit.ch = ch;
                curTdcHit.leadingTime = edge.time;
                curTdcHit.tdcTs = edge.tdcTs;
            } else {
                curTdcHit.trailingTime = edge.time;
                if(!hasTdcChHit)
                    curTdcHit.tdcTs = edge.tdcTs;
                curTqdcEv.hits[ch].tdcHits.push_back(curTdcHit);
                ++devHits[ch].tdc_hits;
                curTdcHit.clear();
                hasTdcChHit = false;
            }
        }
        if(hasTdcChHit) {
            curTqdcEv.hits[curTdcHit.ch].tdcHits.push_back(curTdcHit);
            ++devHits[curTdcHit.ch].tdc_hits;
            curTdcHit.clear();
            hasTdcChHit = false;
        }
    }
}

void TQDCAnalyzer::decodeAdcBlock(const quint32 *payloadPtr)
{
    int blockWord = 0;
    int totHitLen = -1;
    int curHitLen =0;
    const int &dataBlockCh = mpdDeviceHdr->deviceId == DEVICE_ID_PHOS_FEC ?
                dataBlockHdr->specific.phose_adc.ch : dataBlockHdr->specific.adc.ch;
    qint16 sample1 = 0, sample2 = 0;
    std::vector<double> &waveForm = curAdcHit.wf;
    if(rawDataEvents) {
        printRawTxt(0, "DataBlockHeader (ADC)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%1; len:%2 ch:%3")
                     .arg(dataBlockHdr->type()).arg(dataBlockHdr->len)
                     .arg(dataBlockCh));
    }

    while(dataBlockHdr->len > blockWord*sizeof(quint32)) {
        const quint32 &word = *(payloadPtr+blockWord);
        ++blockWord;
        if(totHitLen==-1) {
            curAdcHit.ch = dataBlockCh;
            curAdcHit.adcTs = word & 0xFFFF;
            totHitLen = word >> 16;
            if(totHitLen%sizeof(quint16))
                throw(QString("Format error: bad size of adc data len: %1").arg(totHitLen));
            if(dataBlockHdr->len < blockWord*sizeof(quint32)+totHitLen) {
                throw(QString("Format error:there is no enough data "
                              "in adc data block for hit with length %1")
                        .arg(totHitLen));
            }
            totHitLen /= sizeof(qint16);

            waveForm.reserve(totHitLen);
            if(rawDataEvents) {
                rawDataTxt << QString("%1:\t\tlen:%2; ts:%3")
                              .arg(*(payloadPtr+blockWord-1), 8, 16, QChar('0'))
                              .arg(totHitLen).arg(curAdcHit.adcTs);
            }
            if(totHitLen==0) {
                totHitLen = -1;
            }
        } else {
            sample1 = static_cast<qint16>( word      & 0xFFFF);
            sample2 = static_cast<qint16>((word>>16) & 0xFFFF);
            waveForm.push_back(sample1);
            ++curHitLen;
            // skip sample2 in case totHitLent is odd
            if(curHitLen != totHitLen) {
                waveForm.push_back(sample2);
                ++curHitLen;
            }
            if(curHitLen == totHitLen) {
                curTqdcEv.hits[curAdcHit.ch].adcHits.push_back(curAdcHit);
                ++analyzerStats.dev_stat[curTqdcEv.index].dev_ch_hits[curAdcHit.ch].adc_hits;
                curAdcHit.clear();
                totHitLen = -1;
                curHitLen=0;
            }
            if(rawDataEvents) {
                rawDataTxt << QString("%1:\t\t%2 %3")
                              .arg(*(payloadPtr+blockWord-1), 8, 16, QChar('0'))
                              .arg(sample2, 10).arg(sample1, 10);
            }
        }
    }
}

void TQDCAnalyzer::decodePureAdcBlock(const quint32 *payloadPtr)
{
    curAdcHit.ch = mstreamHdr->usrDefBits;
    curAdcHit.taiTime = reinterpret_cast<const AdcDataType1 *>(payloadPtr);

    if(rawDataEvents) {
        const auto &tai = curAdcHit.taiTime;
        QString suffix = QString(":\t\t%2")
                .arg(QDateTime::fromTime_t(tai.sec).toString());
        printRawWord(payloadPtr, suffix);
        suffix = QString(":\t\tnsec:%L1 %2")
                .arg(tai.nSec,9,10,QChar('0'))
                .arg(tai.flags == 2 ? "ok" : QString(" f:%1").arg(tai.flags));
        printRawWord(payloadPtr+1, suffix);
    }

    const quint16 adcOffset = adcRawDataSigned ? 0 : 0x8000;
    for(int wordIndex=2; wordIndex<mstreamHdr->words32b;++wordIndex) {
        quint32 *word = (quint32 *)(data.data()+curPos+wordIndex*sizeof(quint32));
        const qint16 &veven = (qint16)((*word >> 16) & 0xFFFF) - adcOffset;
        const qint16 &vodd  = (qint16)((*word >> 0)  & 0xFFFF) - adcOffset;
        curAdcHit.wf.push_back(veven);
        curAdcHit.wf.push_back(vodd);
        if(rawDataEvents) {
            printRawWord(word, QString(":\t\t%1 %2").arg(veven).arg(vodd));
        }
    }
    curTqdcEv.hits[curAdcHit.ch].adcHits.push_back(curAdcHit);
    ++analyzerStats.dev_stat[curTqdcEv.index].dev_ch_hits[curAdcHit.ch].adc_hits;
    curAdcHit.clear();
}

void TQDCAnalyzer::decodeTtvxsBlock(const quint32 *payloadPtr)
{
    if(dataBlockHdr->len < sizeof (curTtvxsEv.trig_src))
        throw(QString("Format error: TTVXS Trigger Data Block length (%1) should contains at lest 'Trigger word'")
            .arg(dataBlockHdr->len));

    curTtvxsEv.trig_src.all = *payloadPtr;
    if(!analyzerStats.ttvxsList.contains(curTtvxsEv.index))
        analyzerStats.ttvxsList << curTtvxsEv.index;
    if(rawDataEvents) {
        QStringList trigList;
        if(curTtvxsEv.trig_src.bit.pulser) trigList << "pulser";
        if(curTtvxsEv.trig_src.bit.random_pulser) trigList << "random_pulser";
        if(curTtvxsEv.trig_src.bit.external_ttl) trigList << "external_ttl";
        if(curTtvxsEv.trig_src.bit.digital_sfp) trigList << "digital_sfp";
        printRawWord(payloadPtr, QString("\t\ttrigSrc: %1").arg(trigList.join(',')));
        for(uint word=1; word<dataBlockHdr->len/sizeof (quint32); ++word) {
            QString suffix;
            switch (word) {
            case 1:
                suffix = QString("\t\ttrigCandid: %L1").arg(*(payloadPtr+word));
                break;
            case 2:
                suffix = QString("\t\ttrigAccept: %L1").arg(*(payloadPtr+word));
                break;
            }
            printRawWord(payloadPtr+word, suffix);
        }
    }
}

void TQDCAnalyzer::decodeDevStat(const quint32 *payloadPtr)
{
    curTdcStat.index = DeviceIndex(static_cast<DeviceId>(mpdDeviceHdr->deviceId),
                                   mpdDeviceHdr->deviceSerial);
    curTdcStat.regIoTimeout = dataBlockHdr->specific.stat.regio_timeout;
    curTdcStat.regIoError = dataBlockHdr->specific.stat.regio_error;
    if(rawDataEvents) {
        printRawTxt(0, "DataBlockHeader (Stats)");
        printRawWord(dataBlockHdr, QString(":\t\ttype:%2; len:%3")
                     .arg(dataBlockHdr->type()).arg(dataBlockHdr->len));

        if(curTdcStat.regIoTimeout)
            rawDataTxt.last().append(" (has RegIO timeout)");
        if(curTdcStat.regIoError)
            rawDataTxt.last().append(" (has RegIO error)");
    }

    int parsed_len = 0;
    while(parsed_len < dataBlockHdr->len) {
        const quint32 &word32 = *payloadPtr;
        quint16 addr = word32>>16;
        quint16 val = word32 & 0xFFFF;
        curTdcStat.stats[addr] = val;
        if(rawDataEvents) {
            printRawWord(payloadPtr, QString(":\t\tAddr:%2\tval:%3")
                          .arg(addr, 4, 16, QChar('0'))
                          .arg(val, 4, 16, QChar('0')));
        }
        switch (addr) {
        case REG_TEMPERATURE:
        case REG_BMC_TEMP_1:
        case REG_BMC_TEMP_2:
        case REG_BMC_TEMP_3:
        case REG_BMC_TEMP_4: {
            double temp = addr==REG_TEMPERATURE ? mlink::MregDevice::convertOneWireTemp(val)
                                                : decodeTemp_v1(val);
            int temp_id = addr==REG_TEMPERATURE ? 0
                                                : addr-REG_BMC_TEMP_1+1;
            if(val)
                curTqdcEv.temp[temp_id] = temp;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" temp[%1]:%2").arg(temp_id).arg(temp));
            }
            break;
        case REG_FW_REV:
            if(val)
                curTqdcEv.fw_rev = val;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" fw_rev:%1").arg(val));
            break;
        case REG_BMC_REVERSION:
            if(val)
                curTqdcEv.bmc_rev = val;
            if(rawDataEvents)
                rawDataTxt.last().append(QString(" bmc_rev:%1").arg(val));
            break;
        }
        ++payloadPtr;
        parsed_len += sizeof(quint32);
    }
}

void TQDCAnalyzer::printRawTxt(uint size, QString header, int shift)
{
    if(!rawDataEvents) return;
    if(!header.isEmpty())
        rawDataTxt << QString("\t%1:").arg(header);
    quint32 *word = reinterpret_cast<quint32*>(data.data()+curPos+shift);
    for(uint w=0; w<(size/sizeof(quint32)); ++w){
        printRawWord(word+w);
    }
}

void TQDCAnalyzer::printRawWord(const void *word, QString suffix)
{
    rawDataTxt << get_raw_data_prefix(reinterpret_cast<const quint32*>(word),
                                      mpdEventHdr).append(suffix);
}

quint32 TQDCAnalyzer::readDevice(uint upperLimitToDevLength)
{
    quint32 byteRead = 0;

    if(upperLimitToDevLength < sizeof(MpdDeviceHdr))
        throw("Format error: There is no data for MpdDeviceHdr");

    mpdDeviceHdr = reinterpret_cast<MpdDeviceHdr *>(data.data()+curPos);
    if(rawDataEvents) {
        printRawTxt(0, QString("MpdDeviceHdr (0x%1").arg(mpdDeviceHdr->deviceSerial, 8, 16, QChar('0')));
        quint32 *word = reinterpret_cast<quint32 *>(data.data()+curPos);
        printRawWord(word, ":\t\tSerial");
        printRawWord(word+1, QString(":\t\tDevId:0x%1 Len:%2")
                     .arg(mpdDeviceHdr->deviceId, 2, 16, QChar('0'))
                     .arg(mpdDeviceHdr->length));
    }

//    qInfo()<<" Device:"<<hex<<mpdDeviceHdr->deviceSerial<<dec;
    curPos += sizeof(MpdDeviceHdr);
    byteRead += sizeof(MpdDeviceHdr);

    if(upperLimitToDevLength < sizeof(MpdDeviceHdr) + mpdDeviceHdr->length)
        throw("Format error: There is no data for MpdDevice payload");

    const bool &isPureAdcDev =
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64VE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64 ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64S2 ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64VE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64VE_XGE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_TDC64VHLE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_ADC64WR;

    const bool &isTadcDev =
            mpdDeviceHdr->deviceId == DEVICE_ID_TDC72VXS ||
            mpdDeviceHdr->deviceId == DEVICE_ID_TDC64VHLE ||
            mpdDeviceHdr->deviceId == DEVICE_ID_TQDC16VS_E ||
            mpdDeviceHdr->deviceId == DEVICE_ID_PHOS_FEC;
    const bool &isTtvxsDev =
            mpdDeviceHdr->deviceId != DEVICE_ID_TTVXS;

    if(!(isPureAdcDev || isTadcDev || isTtvxsDev)) {
        if(rawDataEvents) {
            printRawTxt(mpdDeviceHdr->length, "Skipped device block");
        }

        curPos += mpdDeviceHdr->length;
        byteRead += mpdDeviceHdr->length;
        return byteRead;
    }
//    qInfo()<<"mpdDevice: length:"<<mpdDeviceHdr->length<<" serial:"<<hex<<mpdDeviceHdr->deviceSerial;

    quint32 devPayloadLen=0;
    curTqdcEv.clear();
    curTtvxsEv.clear();
    curTdcStat.clear();
    curTtvxsEv.evNum = curTqdcEv.evNum = mpdEventHdr->evNum;
    curTtvxsEv.index = curTqdcEv.index = DeviceIndex(static_cast<DeviceId>(mpdDeviceHdr->deviceId),
                                                     mpdDeviceHdr->deviceSerial);
    while(mpdDeviceHdr->length > devPayloadLen) {
        if(devPayloadLen+sizeof(MpdMStreamHdr)>mpdDeviceHdr->length) {
            throw(QString("Format error: rest size of devPayloadLen is not enough for MStream header(%1):"
                          "devPayloadLen-parsed=%2-%3")
                .arg(sizeof(MpdMStreamHdr))
                .arg(mpdDeviceHdr->length)
                .arg(devPayloadLen));
        }
        mstreamHdr = reinterpret_cast<MpdMStreamHdr *>(data.data()+curPos);


//        qInfo()<<"mstreamHdr: length:"<<mstreamHdr->words32b*sizeof(quint32);
        if(rawDataEvents && mstreamHdr->subtype !=0xF) {
            printRawTxt(0, "MStreamHdr");
            QString suffix = QString(":\t\tsubtype:%1; words:%2")
                    .arg(mstreamHdr->subtype).arg(mstreamHdr->words32b);
            if(mstreamHdr->subtype && isPureAdcDev)
                suffix += QString(" ch:%1").arg(mstreamHdr->usrDefBits);
            printRawWord(mstreamHdr, suffix);
        }

        curPos += sizeof(MpdMStreamHdr);
        devPayloadLen += sizeof(MpdMStreamHdr);
        const quint32 mStreamWholeSize = mstreamHdr->words32b*sizeof(quint32);
        if(devPayloadLen+mStreamWholeSize>mpdDeviceHdr->length) {
            throw(QString("Format error: rest size of devPayloadLen is not enough for MStream payload(%1 32-bit words):"
                          "devPayloadLen-parsed=%2-%3")
                .arg(mstreamHdr->words32b)
                .arg(mpdDeviceHdr->length)
                .arg(devPayloadLen));
        }
        switch (mstreamHdr->subtype) {
        case 0:
        {
//            if(mstreamHdr->words32b*sizeof(quint32) != sizeof(AdcDataType1))
//                throw( QString("Format error: Invalid size(%1 Bytes) of MStream subtype=0 block; should be %2 Bytes.")
//                    .arg(mstreamHdr->words32b*sizeof(quint32)).arg(sizeof(AdcDataType0)).toStdString().c_str());
            tsTAI = reinterpret_cast<TaiTimeBlock *>(data.data()+curPos);
            quint32 blockLen = sizeof(TaiTimeBlock);
            if(rawDataEvents) {
                printRawTxt(0, "TAI");
                QString suffix = QString(":\t\t%2")
                        .arg(QDateTime::fromTime_t(tsTAI->wfTsHi).toString());
                printRawWord(&tsTAI->wfTsHi, suffix);
                suffix = QString(":\t\tnsec:%L1")
                        .arg(tsTAI->wfTsLow,9,10,QChar('0'));
                printRawWord(&tsTAI->wfTsHi+1, suffix);
            }

            if(isPureAdcDev) {
                const quint64 *chMask = reinterpret_cast<quint64*>(data.data()+curPos+blockLen);
                curTqdcEv.chMask = *chMask;
                curTqdcEv.tai = TAI(tsTAI);
                break;
            }
            while(mStreamWholeSize > blockLen) {
                dataBlockHdr = reinterpret_cast<DataBlockHeader *>(data.data()+curPos+blockLen);
                if(blockLen+sizeof(DataBlockHeader)+dataBlockHdr->len > mStreamWholeSize)
                    throw(QString("Format error: TDC Data Block length (%1) is not equal to MStream length:"
                                  "MStream-tsTAI-DataBlockHeader=%2-%3-%4")
                        .arg(dataBlockHdr->len)
                        .arg(mStreamWholeSize)
                        .arg(blockLen)
                        .arg(sizeof(DataBlockHeader)));
                blockLen += sizeof(DataBlockHeader);

                const quint32* payloadPtr =
                    reinterpret_cast<const quint32*>(data.data()+curPos+blockLen);

                switch(dataBlockHdr->type()) {
                case 0:
                    curTqdcEv.tai = TAI(tsTAI);
                    decodeTdcBlock(payloadPtr);
                    break;
                case 1:
                    curTqdcEv.tai = TAI(tsTAI);
                    decodeAdcBlock(payloadPtr);
                    break;
                case 0xA:
                    curTtvxsEv.tai = TAI(tsTAI);
                    decodeTtvxsBlock(payloadPtr);
                    break;
                case 0xF:
                    decodeDevStat(payloadPtr);
                    break;
                default:
                    if(rawDataEvents) {
                        printRawTxt(dataBlockHdr->len+sizeof(DataBlockHeader),
                                    QString("Unknown data block:%1")
                                    .arg(dataBlockHdr->type()),
                                    blockLen-sizeof(DataBlockHeader));
                    }

//                    throw(QString("Format error: unknown TQQDC Data type: %1").arg(dataBlockHdr->type));
                }
                blockLen += dataBlockHdr->len;
            }
            break;
        }
        case 1:
        {
            if(isPureAdcDev) {
                const quint32* payloadPtr =
                        reinterpret_cast<const quint32*>(data.data()+curPos);
                decodePureAdcBlock(payloadPtr);
                break;
            }
        }
        default:
            if(rawDataEvents) {
                printRawTxt(mstreamHdr->words32b*sizeof(quint32), "MStreamPayload");
            }
//            throw( QString("Format error: Invalid subtype number in MStream header: %1.")
//                .arg(mstreamHdr->subtype).toStdString().c_str());
        }

        curPos += mstreamHdr->words32b*sizeof(quint32);
//        qInfo()<<"MS block:ev="<<mpdEventHdr->evNum<<"; subtype="<<mstreamHdr->subtype<<"; curpos="<<curpos;
        devPayloadLen += mstreamHdr->words32b*sizeof(quint32);
    }
    if(mpdDeviceHdr->length != devPayloadLen)
        throw(QString("Format error: Sum of sizes of MStream blocks(%1) isn't equal to one declared in MpdDevice header(%2)")
            .arg(devPayloadLen).arg(mpdDeviceHdr->length).toStdString().c_str());

    if(isPureAdcDev || isTadcDev) {
        curCompleteEv.tqdcEv[curTqdcEv.index] = curTqdcEv;
        curTqdcEv.clear();
    } else if(isTtvxsDev) {
        curCompleteEv.ttvxsEv[curTtvxsEv.index] = curTtvxsEv;
        curTtvxsEv.clear();
    } else {
        Q_ASSERT(false);
    }

    if(!curTdcStat.index.isNull()) {
        curCompleteEv.devStats[curTdcStat.index] = curTdcStat;
        curTdcStat.clear();
    }

    if(rawDataEvents) {
        rawDataTxt << "";
    }

    byteRead += devPayloadLen;
    return byteRead;
}

void TQDCAnalyzer::sendLog(QString s) const
{
    emit addLog(s);
}
