//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef QTMREGDEVICE_H
#define QTMREGDEVICE_H

#include <vector>

#include <QMap>
#include <QObject>
#include <QVector>

#include "device_id.h"
#include "base/fw_version.h"
#include "base/DeviceIndex.h"
#include "sdb/AfiSdbId.h"

namespace mlink {
class DeviceAttr;
class MregDevice;
class RegIOBase;
class RegOpVector;
class RegOp;
}

class BMStatus;
struct FeLinkStatus;
class QHostAddress;
class QTimer;
class WrEpStatus;

class QtMregDevice : public QObject
{
    Q_OBJECT
public:

    Q_DISABLE_COPY(QtMregDevice)

    explicit QtMregDevice(const DeviceIndex &index, QObject *parent = nullptr);
    ~QtMregDevice() override;
    void detect();
    bool isConnected() const;
    mlink::DeviceAttr getAttr() const;
    DeviceId getDeviceId() const;
    fw_version_t getDeviceFwVersion() const;
    bool hasMultiAckSupport() const;
    quint8 getHwBufSize() const;
    quint64 getDeviceSerialNumber() const;
    DeviceIndex getDeviceIndex() const { return deviceIndex; }
    QString getIdent() const;
    QString getDeviceIdStr() const;
    QString getSerialIdStr() const;
    static int getDeviceChNum(const DeviceIndex &index);
    QString messagePrefix() const;
    QString getFirmwareVersionStr() const;
    QHostAddress getAddress() const;
    QString getAddressStr() const;
    QString getDeviceInfoStr() const;
//    QPair<int,int> getHwVer() const;
    double getTemperature() const;
    void setBlockLiveMagic(bool block=true);
    bool writeLiveMagic();
    virtual bool checkLiveMagic();
    bool hasWrEpStatus() const;
    bool hasSdbDevice(AFI_SDB_ID id) const;
    quint16 getBaseAddr(AFI_SDB_ID id) const;
    QPair<quint16, quint16> getAddrRange(AFI_SDB_ID id) const;

    /// read 16-bit register
    quint16 regRead(int regNumber);

    /// write 16-bit register
    void regWrite(int regNumber, quint16 data);

    //    void regWrite(const QMap<quint16, quint16> &data);

    /// read 32-bit register
    quint32 regRead32(int regNumber);

    /// write 32-bit register
    void regWrite32(int regNumber, quint32 data);

    /// read 64-bit register
    quint64 regRead64(int regNumber);

    /// write 64-bit register
    void regWrite64(int regNumber, quint64 data);

    /// read 16-bit registers
    std::vector<quint16> regReadBlk(const std::vector<int> &registers, bool checkAccess = true);

    /// evaluate register operations
    mlink::RegOpVector regOpExec(const mlink::RegOpVector &data);

    /// read 32-bit word from RAM at word offset wordNumber
    quint32 memRead(int wordNumber);

    /// write 32-bit word to RAM at word offset wordNumber
    void memWrite(int wordNumber, quint32 data);

    /// read wordCount 32-bit words from RAM starting from startWord offset
    std::vector<quint32> memReadBlk(int startWord, size_t wordCount);

    /// write 32-bit words to RAM starting from startWord offset
    void memWriteBlk(int wordNumber, const std::vector<quint32> &data);

    static quint32 get32(const mlink::RegOpVector &rx, int startPos);
    static quint64 get64(const mlink::RegOpVector &rx, int startPos);
    static quint32 get32(QVector<mlink::RegOp>::const_iterator &regOpIt, bool *ok=nullptr);
    static quint64 get64(QVector<mlink::RegOp>::const_iterator &regOpIt, bool *ok=nullptr);


Q_SIGNALS:
    void deviceConnected(const DeviceIndex &);
    void deviceDisconnected(const DeviceIndex &);
    void liveMagicFailed(const DeviceIndex &);
    void setStatus(const DeviceIndex &, const QString &);
    void TempChanged(const DeviceIndex &, double);
    void feLinkStatusUpdated(const DeviceIndex &index, const QMap<int,FeLinkStatus> &stMap) const;

protected Q_SLOTS:
    virtual void statusPoll();
    virtual void on_deviceConnected();
    virtual void on_deviceDisconnected();

public Q_SLOTS:
    virtual void deleteHardware(const DeviceIndex &index);
    void setHostAddress(const DeviceIndex &index, const QHostAddress &address);

protected:
    mlink::MregDevice *createMregDevice(mlink::RegIOBase &p, int address, int id) Q_REQUIRED_RESULT;
    double readTemp();
    void readFeLinkStat();
    quint64 readTimestampCurrent();
    quint64 readTimestampRun();
    quint64 readTimestampTrig();
    WrEpStatus readWrEpStatus();
    BMStatus readBMStatus();
    virtual int getFeLinkCnt() const { return 1; }

    void enable();
    void disable();
    quint16 readLiveMagic();

    mlink::RegIOBase * const link;

private:
    const DeviceIndex deviceIndex;
    mlink::MregDevice * const mregDevice; // dev must be after link
    QTimer * const timerStatusPoll;
    QTimer * const startupOfflineTimer;
    bool connectedState = false;
    double currentTemp = 0;
    quint32 liveMagic = 0;
    bool supportLiveMagic = false;
    bool blockLiveMagic = false;
};


#endif // QTMREGDEVICE_H
