#include "ConfigurationManager.h"
#include "ui_ConfigurationManager.h"

#include <QDebug>
#include <QMessageBox>

#include "base/ConfigSelector.h"
#include "json-tree-model/JsonTreeModel.h"
#include "lib-common/UIStyles.h"

#include "CMSubDialog.h"
#include "ConfigListModel.h"


ConfigurationManager::ConfigurationManager(QWidget *parent, const DaqModule &program_type) :
    QDialog(parent),
    ui(new Ui::ConfigurationManager),
    program_type(program_type),
    model_config_info(new JsonTreeModel(this)),
    model_configs_list(new ConfigListModel(this))
{
    ui->setupUi(this);

    auto f = QFont();
    f.setBold(true);
    for(auto t : QSet<QTreeView*>{ui->treeViewCfgList, ui->treeViewCfgInfo}) {
        t->header()->setFont(f);
        t->header()->setDefaultAlignment(Qt::AlignCenter);
        t->header()->sectionResizeMode(QHeaderView::ResizeToContents);
    }

    ui->treeViewCfgList->setModel(model_configs_list);
    ui->treeViewCfgInfo->setModel(model_config_info);

    if(program_type.isEmpty())
        return;

    connect(ui->treeViewCfgList, &ConfigListTreeView::currentCfgIndexChanged,
            this, &ConfigurationManager::currentCfgIndexChanged);

    ui->splitter->setStretchFactor(ui->splitter->indexOf(ui->treeViewCfgList),6);
    ui->splitter->setStretchFactor(ui->splitter->indexOf(ui->treeViewCfgInfo),5);

    ui->treeViewCfgList->header()->setStyleSheet(UIStyles::TREE_HEADER_STYLE);
    ui->treeViewCfgList->setStyleSheet(UIStyles::TREE_STYLE);
}

ConfigurationManager::~ConfigurationManager()
{
    delete ui;
}

void ConfigurationManager::setVisible(bool visible)
{
    QDialog::setVisible(visible);
    if(visible && !initialized)
        reset();
}

void ConfigurationManager::reset()
{
    initialized = true;
    const ConfigSelector cs(program_type, configuration);
    selected_json = base_config.read(cs);

    //--- mongo status
    update_label_db_status();

    model_config_info->loadJson(selected_json);
    ui->treeViewCfgInfo->expandAll();
    populate_config_list_tree();
}

void ConfigurationManager::configSelectorChanged(const ConfigIndexName &cin)
{
    if(configuration == cin)
        return;

    configuration = cin;
    model_configs_list->set_current_cin(configuration);
    if(isVisible())
        reset();
    else
        initialized = false;
}

void ConfigurationManager::populate_config_list_tree()
{
    QJsonObject jsonCfgMap;
    config_map = base_config.get_config_map(program_type);

    for(auto piIt = config_map.cbegin(); piIt != config_map.cend(); ++piIt){
        QJsonObject jsonProgIndexMap;
        auto &programIndex = piIt.key();

        for(auto cnIt = piIt->cbegin(); cnIt != piIt->cend(); ++cnIt) {
            jsonProgIndexMap[cnIt.key()] = cnIt.value().toString(Qt::ISODate);
        }
        jsonCfgMap[programIndex.toString()] = jsonProgIndexMap;
    }

    model_configs_list->loadJson(jsonCfgMap);
}

void ConfigurationManager::update_label_db_status()
{
    auto label = ui->label_db_status;
    bool mongoOk = false;
#ifdef MONGO_DRIVERS
    mongoOk = base_config.mongo_db.try_connection();
#endif

    label->setText(mongoOk ? "Mongo connected" : "Mongo disconnected");
    auto palette = label->palette();
    palette.setColor(label->foregroundRole(), mongoOk ?
                         QColor(95, 156, 95) : QColor(Qt::red));
    label->setPalette(palette);
}

void ConfigurationManager::on_pushButton_new_clicked()
{
    exec_sub_dialog(CMDialog_Type_New);
}

void ConfigurationManager::on_pushButton_rename_clicked()
{
    exec_sub_dialog(CMDialog_Type_Rename);
}

void ConfigurationManager::on_pushButton_clone_clicked()
{
    exec_sub_dialog(CMDialog_Type_Clone);
}

void ConfigurationManager::exec_sub_dialog(int type)
{
    CMSubDialog sub_dialog(static_cast<CMDialogType>(type),
                           type==CMDialog_Type_New ? default_json : selected_json,
                           program_type, configuration_selected, this);
    connect(&sub_dialog, &CMSubDialog::switch_to_config,
            this, &ConfigurationManager::switch_to_config);
    connect(&sub_dialog, &CMSubDialog::switch_to_config,
            this, &QDialog::accept);
    if(sub_dialog.exec() == QDialog::Accepted) {
        on_pushButtonRefresh_clicked();
    }
}

void ConfigurationManager::on_pushButton_delete_clicked()
{
    int ret = QMessageBox::question(this,
                                    "Delete configuration",
                                    QString("Delete configuration '%1'?")
                                    .arg(configuration_selected.toString()));
    if(ret == QMessageBox::Yes){
        base_config.delete_config(ConfigSelector(program_type, configuration_selected));
        populate_config_list_tree();
    }
}

//---button switch clicked---
void ConfigurationManager::on_pushButton_switch_to_clicked()
{
    emit switch_to_config(configuration_selected);
    QDialog::accept();
}

void ConfigurationManager::on_pushButtonRefresh_clicked()
{
    const ConfigSelector cs(program_type, configuration_selected);
    selected_json = base_config.read(cs);
    model_config_info->loadJson(selected_json);
    populate_config_list_tree();
}


void ConfigurationManager::on_treeViewCfgList_doubleClicked(const QModelIndex &index)
{
    auto cin_var = model_configs_list->data(index, Qt::UserRole);
    if(cin_var.isValid()) {
        auto cin = cin_var.value<ConfigIndexName>();
        emit switch_to_config(cin);
        QDialog::accept();
    }
}

void ConfigurationManager::currentCfgIndexChanged(const QModelIndex &current)
{
    bool sub_item = current.isValid() && current.parent().isValid();
    if(sub_item) {
        auto cin_var = model_configs_list->data(current, Qt::UserRole);
        if(cin_var.isValid()) {
            auto cin = cin_var.value<ConfigIndexName>();
            configuration_selected = cin;
        }
    }

    bool this_item = (configuration_selected == configuration);

    ui->pushButton_rename->setEnabled(sub_item);
    ui->pushButton_clone->setEnabled(sub_item);
    ui->pushButton_delete->setEnabled(sub_item && !this_item);
    ui->pushButton_switch_to->setEnabled(sub_item && !this_item);

    if(sub_item) {
        const ConfigSelector cs(program_type, configuration_selected);
        selected_json = base_config.read(cs);
        model_config_info->loadJson(selected_json);
        ui->treeViewCfgInfo->expandAll();
    }
}

void ConfigurationManager::on_checkBox_stateChanged(int arg1)
{
    if(arg1)
        ui->treeViewCfgInfo->expandAll();
    else
        ui->treeViewCfgInfo->collapseAll();
}
