#include "BaseConfig.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QJsonDocument>
#include <QHostInfo>

#include "MongoConsts.h"
#include "RootConfigUtil.h"

#include "base/CoreTypeConverter.h"
#include "base/ConfigSelector.h"
#include "base/ConfigMetaConverter.h"
#include "base/ConfigDocumentConverter.cpp"
#include "base/RcRunId.h"

#ifdef MONGO_DRIVERS
#include "mongoqt/BsonQtConverter.h"
#endif

QJsonObject getRecordMetaInfo() {
    QJsonObject meta;

    RootConfigUtil::setModificationTimestamp(&meta);
    meta[k_current_host] = QHostInfo::localHostName();
    meta[k_phantom] = false;

    return meta;
}


BaseConfig::BaseConfig() :
    mongo_alive(false)
{
    CoreTypeConverter::registerTypes();
#ifdef MONGO_DRIVERS
    mongo_alive = mongo_db.try_connection();
#endif
}

void BaseConfig::init(const DaqModule &program_type)
{
#ifdef MONGO_DRIVERS
    if(mongo_alive){
        mongo_db.mongo_init();
        if (!synchronization(program_type)){
            qWarning() << "Synchronize Mongo saved failed!";
        }
    }
#endif
}

ConfigDocument BaseConfig::read_(const ConfigSelector &cs)
{
    if(cs.getType().isEmpty()){
        qCritical() << "Can't read config with empty program type";
        return QJsonObject();
    }
    ConfigDocument doc;

#ifdef MONGO_DRIVERS
    if(mongo_alive && !cs.getIndex().isDefault()){
        //---read settings from mongo
        doc = mongo_db.read_(cs);
    }
#endif
    if(doc.isEmpty()){
        doc = json_db.read_(cs);
    }
    return doc.getConfigMeta().isPhantom() ?
           ConfigDocument() : doc;
}

QJsonObject BaseConfig::read(const ConfigSelector &cs)
{
    if(cs.getType().isEmpty()){
        qCritical() << "Can't read config with empty program type";
        return QJsonObject();
    }
    QJsonObject cfg_db_record;

#ifdef MONGO_DRIVERS
    if(mongo_alive && !cs.getIndex().isDefault()){
        //---read settings from mongo
        cfg_db_record = mongo_db.read(cs);
    }
#endif
    if(cfg_db_record.isEmpty()){
        cfg_db_record = json_db.read(cs);
    }

    return RootConfigUtil::is_phantom(cfg_db_record)
            ? QJsonObject() : cfg_db_record[k_config].toObject();
}

bool BaseConfig::write(ConfigDocument &cd)
{
    if (cd.isEmpty()) {
        qCritical() << "Can't write config with empty program type";
        return false;
    }

    bool ok = true;
#ifdef MONGO_DRIVERS
    const bool program_index_default = cd.getConfigSelector().getIndex().isDefault();
    if(mongo_alive && !program_index_default){
        ok &= mongo_db.write(cd);
    }
#endif
    ok &= json_db.write(cd);
    return ok;
}

bool BaseConfig::write(const ConfigSelector &cs, const QJsonObject &jsonCfg)
{
    if (cs.getType().isEmpty()) {
        qCritical() << "Can't write config with empty program type";
        return false;
    }

    QJsonObject recordJson;
    const auto cs_json = QVariant::fromValue(cs).value<QJsonObject>();
    recordJson[k_config_selector] = cs_json;
    recordJson[k_config] = jsonCfg;
    recordJson[k_meta] = getRecordMetaInfo();


    bool ok = true;
#ifdef MONGO_DRIVERS
    const bool program_index_default = cs.getIndex().isDefault();
    if(mongo_alive && !program_index_default){
        ok &= mongo_db.write(cs, recordJson);
    }
#endif
    ok &= json_db.write(cs, recordJson);
    return ok;
}

QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > BaseConfig::get_config_map(const DaqModule &program_type)
{
    if(program_type.isEmpty()) {
        return QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> >();
    }

    QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > toReturn;
#ifdef MONGO_DRIVERS
    if(mongo_alive)
        toReturn = mongo_db.get_config_map(program_type);
#endif

    auto local = json_db.get_config_map(program_type);
    if(toReturn.isEmpty()) {
        return local;
    }

    auto it = local.find(ProgramIndex());
    if(it != local.end()) {
        toReturn.insert(it.key(), *it);
    }
    return toReturn;
}

bool BaseConfig::delete_config(const ConfigSelector &cs)
{
    if(cs.getType().isEmpty()){
        qCritical() << "Can't delete config with empty program type";
        return false;
    }

    QJsonObject metaJson;
    metaJson[k_phantom] = true;
    RootConfigUtil::setDeletionTimestamp(&metaJson);
    metaJson[k_delete_host] = QHostInfo::localHostName(); // FIXME: is it necessary?
    QJsonObject updateJson({{k_meta, metaJson}});

    bool ok = true;
#ifdef MONGO_DRIVERS
    if(mongo_alive){
        if(!mongo_db.delete_config(cs)){
            qWarning() << "Remove configuration from Mongo failed. Collection:"
                       << COLLECTION_SAVED;
        }
        ok &= mongo_db.write(cs, metaJson);
    }
#endif

    ok &= json_db.delete_config(cs);
    if(!mongo_alive) {
        ok &= json_db.write(cs, metaJson);
    }
    return true;
}

bool BaseConfig::write(const RcRunId &rc, const QJsonObject &jsonCfg)
{
    if (! rc.runIndex.isEmpty()) {
        qWarning() << "write history failed. Can't find program index";
        return false;
    }

    if (rc.runNumber != 0) {
        qWarning() << "write history failed. Can't find run number";
        return false;
    }

    QJsonObject recordJson;
    const auto rc_json = QVariant::fromValue(rc).value<QJsonObject>();
    recordJson[k_run_id] = rc_json;
    recordJson[k_config] = jsonCfg;
    recordJson[k_meta] = getRecordMetaInfo();

    bool ok = true;
#ifdef MONGO_DRIVERS
    if(mongo_alive) {
        ok &= mongo_db.write(rc, recordJson);
    }
#endif
    ok &= json_db.write(rc, recordJson);
    return ok;
}

RcRunId BaseConfig::check_run_number(const RcRunId &rc)
{
    if(mongo_alive){
        #ifdef MONGO_DRIVERS
        return mongo_db.check_run_number(rc);
        #endif
    }
//    return local_db.check_run_number(program_type, run_index, run_number);
    //TODO
    return RcRunId();
}

//------------- LOCAL ----- MONGO ---
// case 1     [normal]    [normal]
// case 2     [normal]    [fantom]
// case 3     [normal]    [      ]
// case 4     [fantom]    [normal]
// case 5     [fantom]    [fantom]
// case 6     [fantom]    [      ]
// case 7     [      ]    [normal]
// case 8     [      ]    [fantom]

#ifdef MONGO_DRIVERS
bool BaseConfig::synchronization(const DaqModule &program_type)
{
    return false;
//    Q_ASSERT(!program_type.isEmpty());
//    qInfo() << "Mongo synchronization to local";
//    if(!local_db.local_db_chekup()){
//        qWarning() << QString("Mongo synchronization failed");
//        return false;
//    }
//    auto list_local = local_db.get_rootconfig_map(program_type);
//    auto list_mongo = mongo_db.get_json_map(program_type);

//    for(auto pi : list_local.keys()) {
//        auto &list_local_pi = list_local[pi];
//        for(auto cn : list_local_pi.keys()) {
//            const auto cs = ConfigSelector(program_type, pi, cn);
//            auto &rc_local = list_local_pi[cn];
//            QDateTime local_date = RootConfigUtil::getDeletionTimestamp(rc_local);
//            bool local_dead = !local_date.isNull();
//            if(!local_dead)
//                local_date = RootConfigUtil::getModificationTimestamp(rc_local);

//            bool contains = list_mongo.contains(pi) && list_mongo[pi].contains(cn);
//            if(!contains) {
//                if(local_dead){
//                    local_db.delete_config(cs); // case 6
//                }
//                else{
//                    mongo_db.write_config(rc_local.get_json().object()); // case 3
//                }
//                continue;
//            }

//            auto &rc_mongo = list_mongo[pi][cn];
//            QDateTime mongo_date = RootConfigUtil::getDeletionTimestamp(rc_mongo);
//            bool mongo_dead = !mongo_date.isNull();
//            if(!mongo_dead)
//                mongo_date = RootConfigUtil::getModificationTimestamp(rc_mongo);
//            bool local_is_newer = local_date > mongo_date;
//            if(local_is_newer) {
//                if(local_dead) {
//                    mongo_db.delete_config(cs); // case 4 and 5
//                    local_db.delete_config(cs);
//                }
//                else{
//                    mongo_db.write_config(rc_local.get_json().object()); // case 1 and 2
//                }

//            } else {
//                if(mongo_dead) {
//                    local_db.delete_config(cs); // case 2 and 5
//                    list_local_pi.remove(cn);
//                }
//                else{
//                    if(rc_local != BsonQtConverter::to_rootconfig(rc_mongo))
//                        local_db.write_config(BsonQtConverter::to_rootconfig(rc_mongo), false); // case 1 and 4
//                }
//            }
//        }
//    }
//    local_db.sync();

//    for(auto pi : list_mongo.keys()) {
//        auto &list_mongo_pi = list_mongo[pi];
//        for(auto cn : list_mongo_pi.keys()) {
//            auto &rc_mongo = list_mongo_pi[cn];
//            bool contains = list_local.contains(pi) && list_local[pi].contains(cn);
//            if(contains) continue; // case 1, 2, 4, 5
//            if(RootConfigUtil::is_phantom(rc_mongo)) continue; // case 8
//            local_db.write_config(BsonQtConverter::to_rootconfig(rc_mongo)); // case 7
//        }
//    }
//    return true;
}
#endif
