#include "JsonDB.h"
#include "base/ConfigSelector.h"
#include "base/ConfigDocument.h"
#include "base/RcRunId.h"
#include "MongoConsts.h"

#include <QJsonObject>
#include <QJsonDocument>
#include <QJsonParseError>
#include <QFile>
#include <QDateTime>
#include <QDir>
#include <QStandardPaths>
#include <QCoreApplication>
#include <QDirIterator>

namespace {
QString get_application_path(){
    return QString(CONFIG_PATH +  "/" +
                   QCoreApplication::organizationName());
}

QString get_type_path(ConfigSelector cs){
    return  QString(get_application_path() + "/" +
                    cs.getType().toString());
}

QString get_index_path(ConfigSelector cs){
    return QString(get_type_path(cs) + "/" +
                   cs.getIndex().toString());
}

QString get_config_file(ConfigSelector cs){
    return  QString(get_index_path(cs) + "/" +
                    cs.getConfigurationName() + ".json");
}

QString get_history_path(){
    return QString(get_application_path() + "/" +
                   HISTORY_RUN);
}

QString get_history_file(RcRunId run_id){
    QString hist_file_name = QString("%1_%2_%3.json")
        .arg(run_id.runIndex)
        .arg(run_id.runNumber)
        .arg(run_id.uuid.toString());

    return  QStringList({get_history_path(), run_id.runIndex, hist_file_name}).join("/");
}

bool json_to_file(QString file_name, QJsonObject obj){
    QFile config_file(file_name);
    if(!config_file.exists()) {
        QFileInfo fi(config_file);
        auto dirPath = fi.dir().absolutePath();
        QDir().mkpath(dirPath);
    }
    if(!config_file.open(QIODevice::WriteOnly)){
        qWarning() << QString("Can't open config file at %1")
            .arg(get_application_path());
        return false;
    }
    const auto data = QJsonDocument(obj).toJson();
    const auto ret = config_file.write(data);
    if(ret != data.size()){
        auto errStr = ret == -1 ? config_file.errorString() :
                                  QString("wrote %1/%2 bytes").arg(ret)
                                  .arg(data.size());
        qWarning() << QString("Can't write config in %1 (%2)")
            .arg(get_application_path()).arg(errStr);
        return false;
    }
    return true;
}
}

ConfigDocument JsonDB::read_(const ConfigSelector &cs)
{
    return ConfigDocument();
}

QJsonObject JsonDB::read(const ConfigSelector &cs)
{
    if(cs.getType().isEmpty()){
        qWarning() << "Read config failed. Set program type";
        return QJsonObject();
    }

    QFile config_file(get_config_file(cs));
    if(!config_file.open(QIODevice::ReadOnly)){
        qWarning() << QString("Can't find or open json config file in %1")
            .arg(get_config_file(cs));
        return QJsonObject();
    }

    QJsonParseError json_err;
    QByteArray data = config_file.readAll();
    QJsonDocument json(QJsonDocument::fromJson(data, &json_err));
    if (QJsonParseError::NoError != json_err.error) {
        qWarning() << QString("ERROR: Parsing json data: %1 from file %2")
            .arg(json_err.errorString()).arg(get_config_file(cs));
        return QJsonObject();
    }
    //TODO: check that record is not a phantom
    if(!json.isObject()){
        return QJsonObject();
    }

    return json.object();
}

bool JsonDB::write(const ConfigDocument &cd)
{
    if(cd.isEmpty()) {
        qWarning() << "write config failed. Set program type";
        return false;
    }

    return json_to_file(get_config_file(cd.getConfigSelector()), cd.getConfig());
}

bool JsonDB::write(const ConfigSelector &cs, const QJsonObject &obj)
{
    if(cs.getType().isEmpty()) {
        qWarning() << "write config failed. Set program type";
        return false;
    }

    return json_to_file(get_config_file(cs), obj);
}

bool JsonDB::write(const RcRunId &rc, const QJsonObject &jsonCfg)
{
    return json_to_file(get_history_file(rc), jsonCfg);
}

int JsonDB::delete_config(const ConfigSelector &cs)
{
    QFile config_file(get_config_file(cs));
    QDir index_path(get_index_path(cs));

    bool ok = config_file.remove();

    if(index_path.isEmpty(QDir::NoDotAndDotDot|QDir::Files))
        index_path.removeRecursively();

    if(!ok){
        qInfo() << QString("Can't remove config file in %1")
            .arg(get_application_path());
        return false;
    }
    return ok;
}

QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > JsonDB::get_config_map(const DaqModule &daq_module)
{
    QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > config_map;
    if(daq_module.isEmpty())
        return config_map;

    QString type_path  = get_type_path(ConfigSelector(daq_module));
    QList<QFileInfo> indexes = QDir(type_path).entryInfoList(QDir::NoDotAndDotDot|QDir::AllEntries);
    for(auto &fia : indexes) {
        if(!fia.isDir())
            continue;

        //TODO: check that it is not phantom and get modDateTime
        ProgramIndex pi(fia.baseName());
        QString index_path = type_path + "/" + fia.baseName();
        QList<QFileInfo> configs = QDir(index_path)
                .entryInfoList(QDir::NoDotAndDotDot|QDir::AllEntries);
        for(auto &fi : configs) {
            if(fi.isFile())
                config_map[pi][fi.baseName()] = fi.lastModified();
        }
    }
    return config_map;
}
