#include "StartupOptions.h"

#include <QCoreApplication>
#include <QDebug>

#include "BaseConfig.h"
#include "base/Globals.h"

const int CELL_WIDTH = 25;

namespace {
QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > getConfigMap(const DaqModule &program_type)
{
    return BaseConfig().get_config_map(program_type);
}

void show_config_list(const DaqModule &program_type, const QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > &config_map)
{
    QString output_str;
    output_str = ":\n";
    const auto typeName = program_type.toString();
    if(config_map.isEmpty()){
        output_str.append(QString("No saved %1 configurations. \n").arg(typeName));
    }
    else{
        output_str.append(QString("List of saved %1 configurations: \n").arg(typeName));
    }

    QString  separator;
    for(int i=0; i<(2*CELL_WIDTH+1); ++i){
        separator.append("-");
    }

    separator.prepend("+");
    separator.append("+");
    separator.append("\n");

    int tail_1 = CELL_WIDTH - QString(" Program index").length();
    QString tail_1_str;
    for(int i=0; i<tail_1; ++i){
        tail_1_str.append(" ");
    }

    int tail_2 = CELL_WIDTH - QString(" Configuration name").length();
    QString tail_2_str;
    for(int i=0; i<tail_2; ++i){
        tail_2_str.append(" ");
    }

    output_str.append(separator);
    QString labels = QString("|") + "\033[91m" + " Program index" + "\033[0m" + tail_1_str +
                     "|" + "\033[91m" + " Configuration name" + "\033[0m" + tail_2_str +  "|" + "\n";
    output_str.append(labels);
    output_str.append(separator);

    for(auto it_pi=config_map.begin(); it_pi!=config_map.end(); ++it_pi) {
        auto &program_index = it_pi.key();
        for(auto it_cn=it_pi->begin(); it_cn!=it_pi->end(); ++it_cn) {
            auto &configuration_name = it_cn.key();
            int tail_1 = CELL_WIDTH - program_index.getIndex().length() -1;
            int tail_2 = CELL_WIDTH - configuration_name.length() -1;

            QString tail_1_str;
            for(int i=0; i<tail_1; ++i){
                tail_1_str.append(" ");
            }

            QString tail_2_str;
            for(int i=0; i<tail_2; ++i){
                tail_2_str.append(" ");
            }

            QString table_data = QString("| ") + program_index.getIndex() + tail_1_str + "| " +
                                 configuration_name + tail_2_str +  "|" + "\n";
            output_str.append(table_data);
            output_str.append(separator);
        }
    }

    output_str.append("\n");
    qInfo().noquote() << output_str;
}

}
StartupOptionsParser::StartupOptionsParser(const QList<QCommandLineOption> &initialOptions)
{
    QString rev;
    #ifdef GIT_DESCR
        rev = QString("%1").arg(GIT_DESCR);
        rev = rev.remove(QRegExp("^.*\\:"));
    #endif
    program_version = rev;
    setApplicationDescription(QCoreApplication::applicationName());
    addHelpOption();

    addPositionalArgument("program_index",
    QCoreApplication::translate("main","with which index to run the program"));

    addOptions(initialOptions);
    addOptions(createOptions());
}

QList<QCommandLineOption> StartupOptionsParser::createOptions()
{
    QList<QCommandLineOption> result;
    result << QCommandLineOption( QStringList()  << "l" <<  "list", "List available configurations");

    result << QCommandLineOption( QStringList()  << "v" <<  "version", "Current program version");

    result << QCommandLineOption(
        QStringList() << "config",
        QCoreApplication::translate("main", "Startup with certain configuration"),
        "config"
        );

    result << QCommandLineOption( QStringList()  << "d" <<  "debug", "Run in debug mode");
    return result;
}

void StartupOptionsParser::parse_options()
{
    process(QCoreApplication::arguments());

    list = isSet("list");
    version = isSet("version");
    Globals::setDebugMode(isSet("debug"));
}

StartupOptions::StartupOptions(const DaqModule &program_type) :
    cs(program_type)
{
}

void StartupOptions::showVersion()
{
    if (parser.program_version.isEmpty()) {
        qInfo().noquote() << QString("\033[91m")
                + QString("Unknown program version")
                + QString("\033[0m");

    } else {
        qInfo().noquote() << QString("\033[32m")
                + QString("program version: %1").arg(parser.program_version)
                + QString("\033[0m");
    }
    QCoreApplication::exit();
    Q_UNREACHABLE();
}

void StartupOptions::setConfig(const QString &configuration_name)
{
    auto config_map = getConfigMap(cs.getType());
    if (!config_map[cs.getIndex()].contains(configuration_name)) {
        qCritical().noquote() << QString("No configuration called \"%1\"").arg(configuration_name);
        show_config_list(cs.getType(), config_map);
        parser.showHelp();
    }
    if (configuration_name.isEmpty() || (configuration_name.startsWith('-'))) {
        parser.showHelp();
    }
    cs.setConfigurationName(configuration_name);
}

void StartupOptions::parse_options()
{
    parser.parse_options();
    if(parser.version)
        showVersion();

    const QStringList args = parser.positionalArguments();
    if (!args.empty()) {
        cs.setIndex(args[0]);
    }

    if (parser.list)
        show_config_list(cs.getType(), getConfigMap(cs.getType()));

    if (parser.isSet("config"))
        setConfig(parser.value("config"));
}
