#include "RemoteControlServer.h"

#include <QCoreApplication>

#include "RemoteControlProtocol.h"
#include "RemoteControlTlvServer.h"
#include "RunControlCommands.h"
#include "base/RcRunId.h"
#include "fsm/BaseFsm.h"
#include "pnp-server/PNPMessage.h"

RemoteControlServer::RemoteControlServer(QObject *parent, quint16 preferedTcpPort, QString _descr) :
    QObject(parent),
    descr(_descr.isEmpty() ? QCoreApplication::applicationName() : _descr),
    server(new RemoteControlTlvServer(preferedTcpPort, this))
{
    connect(server, &RemoteControlTlvServer::connectionAccepted,
            this, &RemoteControlServer::connectionAccepted);

    connect(server, &RemoteControlTlvServer::tlvDataReceived,
            this, &RemoteControlServer::tlvDataReceived);

    connect(server, &RemoteControlTlvServer::connectionStatusUpdated,
            this, &RemoteControlServer::connectionStatusUpdated);
}

RemoteControlServer::~RemoteControlServer()
{
}

void RemoteControlServer::startListen(bool state)
{
    server->startListen(state);
}

void RemoteControlServer::connectionAccepted()
{
    sendDescription();
    sendFsmState();
}

void RemoteControlServer::tlvDataReceived(const TypeLengthHeader &hdr, const QByteArray &data)
{
    command_line_parse(data);
}

void RemoteControlServer::config_transfer(QByteArray data) const
{
    server->send_byte_array(RC_MSG_CONFIG, data);
}

void RemoteControlServer::send_command_message(QString mess) const
{
    if (!server->isListening())
        return;
    QByteArray mess_ba = mess.toUtf8();
    server->send_byte_array(RC_MSG_CMD, mess_ba);
}

void RemoteControlServer::command_line_parse(QByteArray data)
{
    QString mess = QString::fromUtf8(data.data(), data.size());
    for (QString command : mess.split(RC_CMD_END)) {
        command = command.trimmed();
        qDebug() << QString("Remote command:").append(command);
        QStringList commandParts = command.split(" ", QString::SkipEmptyParts);
        if(commandParts.isEmpty())
            continue;
        if(commandParts.first() == RC_CMD_START
                || commandParts.first() == RC_CMD_STOP) {
            // start <globIndexStr> <runNumber> [startEvNumber]
            // stop <globIndexStr> <runNumber> [stopEvNumber]

            qInfo() << QString("RemoteControlServer received command:").append(commandParts.first());
            bool ok = commandParts.size()>=4;
            RcRunId runId;
            if(ok)
                runId.runNumber = commandParts[2].toInt(&ok);
            if(ok) {
                runId.uuid = QUuid::fromString(commandParts[3]);
                ok = !runId.uuid.isNull();
            }
            quint64 evNumber=0;
            if(ok && commandParts.size()>4)
                evNumber = commandParts[4].toULongLong(&ok);
            if(ok) {
                runId.runIndex = commandParts[1];
                if(commandParts.first() == RC_CMD_START)
                    emit startRequest(runId, evNumber);
                else
                    emit stopRequest(runId, evNumber);
            }
            else{
                qWarning() << "RemoteControlServer can't define evNumber in command:"
                           << command;
            }
        } else if(commandParts.first() == RC_CMD_STATE) {
            // state
            qInfo() << QString("RemoteControlServer received command:").append(RC_CMD_STATE);
            if(commandParts[1] != ""){
                emit timeToStart(commandParts[1]);
            } else{
                emit stateRequest();
            }
        }else if(commandParts.first() == RC_CMD_EV_NUM_CHECK) {
            // EvNumCheck
//            qInfo() << QString("Received command:").append(RC_CMD_EV_NUM_CHECK);
            bool ok = commandParts.size()==2;

            int checkIndex = 0;
            if(ok)
                checkIndex = commandParts[1].toInt(&ok);
            if(ok){
                emit evNumCheckRequest(checkIndex);
            }
            else
                qWarning() << "RemoteControlServer can't define checkIndex in command:"
                           << command;
        }
    }
}

void RemoteControlServer::sendState(const QString &state)
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_STATE)
            .arg(state)
            .arg(RC_CMD_END);
    send_command_message(mess);
}

void RemoteControlServer::onFsmStateChange(const FsmState &state)
{
    fsmState = state;
    sendFsmState();
}

void RemoteControlServer::onRunError(const RcRunId &runId, QString runError)
{
    QString mess = QString("%1 %2 %3%4")
            .arg(RC_CMD_ERROR)
            .arg(QString(runId))
            .arg(runError.replace("\r", "").replace("\n","").trimmed())
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::onRunWarning(const RcRunId &runId, QString warnStr)
{
    QString mess = QString("%1 %2 %3%4")
            .arg(RC_CMD_WARN)
            .arg(QString(runId))
            .arg(warnStr.replace("\r", "").replace("\n","").trimmed())
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::onBackPressure(const RcRunId &runId, bool backPressure)
{
    if(runId.runIndex.isEmpty())
        return;
    QString mess = QString("%1 %2 %3%4")
            .arg(RC_CMD_BACK_PRESSURE)
            .arg(QString(runId))
            .arg(backPressure ? "1" : "0")
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::onVmeFlag(bool hasVme)
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_VME_FLAG)
            .arg(hasVme ? "1" : "0")
            .arg(RC_CMD_END);

    send_command_message(mess);
}

void RemoteControlServer::sendDescription() const
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_DESCR).arg(descr).arg(RC_CMD_END);
    send_command_message(mess);
}

void RemoteControlServer::sendFsmState() const
{
    QString mess = QString("%1 %2 %3")
            .arg(RC_CMD_FSM_STATE)
            .arg(fsmStateToString(fsmState))
            .arg(RC_CMD_END);
    send_command_message(mess);
}

void RemoteControlServer::changeProgramDescription(const ProgramDescription &pd)
{
    descr = pd.getName();
    sendDescription();
}

void RemoteControlServer::evNumCheckResponse(int checkIndex, const EvNumSet &evNums) const
{
    QString evNumsStr;
    for (quint64 evNum : evNums) {
        evNumsStr += QString::number(evNum)+" ";
    }
    QString mess = QString("%1 %2 %3 %4")
            .arg(RC_CMD_EV_NUM_CHECK)
            .arg(checkIndex)
            .arg(evNumsStr)
            .arg(RC_CMD_END);
    send_command_message(mess);
}
