/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RemoteControlTlvServer.h"

#include <QHostInfo>
#include <QTcpServer>
#include <QTcpSocket>

#include <stdexcept>
#include <unistd.h>

#include "RemoteControlProtocol.h"
#include "RemoteControlServerStatus.h"
#include "RunControlCommands.h"

namespace {
const qint32 MAX_BLOCK_SIZE = 1e5;
}

RemoteControlTlvServer::RemoteControlTlvServer(quint16 preferedTcpPort, QObject *parent) :
    QObject(parent),
    tcpSocket(nullptr),
    qTcpServer(new QTcpServer(this)),
    status(new RemoteControlServerStatus(preferedTcpPort))
{
    qRegisterMetaType<RemoteControlServerStatus> ();
    qRegisterMetaType<TypeLengthHeader>();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &RemoteControlTlvServer::updateConnections);
}

RemoteControlTlvServer::~RemoteControlTlvServer()
{
    blockSignals(true);
    startListen(false);
    blockSignals(false);
    delete qTcpServer;
    delete status;
}

int RemoteControlTlvServer::getTcpPort() const
{
    return status->tcpPort;
}

bool RemoteControlTlvServer::isFree() const
{
    return !status->portInUse;
}

bool RemoteControlTlvServer::isListening() const
{
    return status->listening;
}

void RemoteControlTlvServer::updateConnections()
{
    while (QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if (tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)) {
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            status->portInUse = false;
        }
        if (tcpSocket == nullptr) {
            tcpSocket = pendingConnection;
            status->portInUse = true;
            status->peerHost = tcpSocket->peerAddress();
            status->peerPort = tcpSocket->peerPort();

            connect(tcpSocket, &QTcpSocket::readyRead,
                    this, &RemoteControlTlvServer::readData);

            connect(tcpSocket, &QTcpSocket::disconnected,
                    this, &RemoteControlTlvServer::onDisconnected);

            QHostInfo info = QHostInfo::fromName(status->peerHost.toString());
            qInfo() << QString("RemoteControlServer accepted connection from %1:%2")
                           .arg(info.hostName())
                           .arg(tcpSocket->peerPort());
            updateStatus();
            emit connectionAccepted();
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if (tcpSocket == nullptr) {
        qDebug() << "There are no connections.";
    }
}

void RemoteControlTlvServer::updateStatus() const
{
    emit connectionStatusUpdated(*status);
}

void RemoteControlTlvServer::startListen(bool state)
{
    if (status->listening != state) {
        if (state) {
            listen();
        } else {
            close();
        }
        status->listening = state;
    }
    updateStatus();
}

void RemoteControlTlvServer::listen()
{
    while (!qTcpServer->listen(QHostAddress::Any, status->preferedTcpPort)) {
        //        QAbstractSocket::SocketError errNum = qTcpServer->serverError();
        //        if(errNum==QAbstractSocket::AddressInUseError){
        //            if(!portInUse){
        //                QString mess = tr("Unable to start the remote control server on port %1: %2, restart in 5 s")
        //                        .arg(tcpPort).arg(qTcpServer->errorString());
        //                sendLog(mess, Qt::red);
        //                portInUse=true;
        //                qFatal(mess.toStdString().c_str());
        //            }
        //            exit(1);
        //            sleep(5);
        //            continue;
        //        }
        QString s = tr("RemoteControlServer unable to listen on TCP port %1: %2")
                        .arg(status->preferedTcpPort)
                        .arg(qTcpServer->errorString());
        qCritical() << s;
        throw std::runtime_error(s.toStdString());
    }
    status->tcpPort = qTcpServer->serverPort();

    qInfo() << QString("RemoteControlServer listening on %1:%2")
                   .arg(qTcpServer->serverAddress().toString())
                   .arg(status->tcpPort);
}

void RemoteControlTlvServer::close()
{
    qTcpServer->close();
    if (tcpSocket) {
        tcpSocket->abort();
        delete tcpSocket;
        tcpSocket = nullptr;
        status->portInUse = false;
    }
    qDebug() << QString("TcpServer closed on %1:%2")
                    .arg(qTcpServer->serverAddress().toString())
                    .arg(status->tcpPort);
}

void RemoteControlTlvServer::onDisconnected()
{
    status->portInUse = false;
    updateStatus();
    qInfo() << "RemoteControlServer disconnected";
}

bool RemoteControlTlvServer::isConnectionAlive() const
{
    return tcpSocket != nullptr && tcpSocket->isValid()
        && (tcpSocket->state() == QTcpSocket::ConnectedState);
}

void RemoteControlTlvServer::send_byte_array(quint32 type, const QByteArray &payload) const
{
    if (!isConnectionAlive()) {
        return;
    }
    QByteArray data = QByteArray(sizeof(TypeLengthHeader), Qt::Uninitialized);
    data.append(payload);
    quint16 check = 0;

    TypeLengthHeader *hdr = reinterpret_cast<TypeLengthHeader *>(data.data());
    hdr->type = type;
    hdr->len = static_cast<quint32>(data.size()) + sizeof(check);

    check = qChecksum(data.data(), static_cast<uint>(data.size()), Qt::ChecksumIso3309);
    data.append(reinterpret_cast<char *>(&check), sizeof(check));

    int written = static_cast<int>(tcpSocket->write(data));

    if (written == data.size()) {
        //        if(type == RC_MSG_CONFIG || type == RC_MSG_CONFIG_OLD) {
        //            qInfo()<< "RemoteControlServer sent configuration";
        //        }
    } else {
        qWarning() << "RemoteControlServer failed to send info string";
    }
}

void RemoteControlTlvServer::readData()
{
    try{
        while(tcpSocket->bytesAvailable()){
            data.append(tcpSocket->readAll());

            while(!data.isEmpty()) {
                bool type_warn=true, size_warn=true;
                TypeLengthHeader *hdr=nullptr;
                while(static_cast<uint>(data.size()) >= sizeof(TypeLengthHeader)) {
                    hdr= reinterpret_cast<TypeLengthHeader*>(data.data());
                    if((hdr->type == RC_MSG_CONFIG) || (hdr->type == RC_MSG_CONFIG_OLD)
                        || (hdr->type == RC_MSG_CMD) || (hdr->type == RC_MSG_CMD_OLD)) {
                        if(hdr->len <= MAX_BLOCK_SIZE) {
                            break;
                        } else if(size_warn) {
                            size_warn = false;
                            qWarning() << QString("RemoteControlServer message error: block size is %1. Permissible maximum is %2")
                                              .arg(hdr->len).arg(MAX_BLOCK_SIZE);
                        }
                    } else if(type_warn) {
                        qWarning() << "RemoteControlServer message error: unknown message type";
                        type_warn = false;
                    }

                    data.remove(0, 1);
                    hdr = nullptr;
                }
                if(!hdr)
                    continue;

                if(data.size() < static_cast<int>(hdr->len))
                    continue;

                quint32 tsd_size = hdr->len - sizeof(quint16);
                QByteArray tsd_data = data.left(static_cast<int>(tsd_size));
                quint16 check_calc = qChecksum(tsd_data.data(), tsd_size, Qt::ChecksumIso3309);

                quint16 *check = reinterpret_cast<quint16*>(data.data()+tsd_size);
                if(*check != check_calc){
                    qWarning() << QString("RemoteControlServer message error: checksum is incorrect.");
                    data.remove(0, 1);
                    continue;
                }
                else{
                    switch (hdr->type) {
                    case RC_MSG_CONFIG_OLD :
                    case RC_MSG_CONFIG :
                        qWarning() << "RemoteControlServer message warning: unexpected type of message";
                        break;
                    case RC_MSG_CMD_OLD :
                    case RC_MSG_CMD :
                        emit tlvDataReceived(*hdr, tsd_data.mid(2*sizeof (qint32)));
                        // command_line_parse(tsd_data.mid(2*sizeof (qint32)));
                        break;
                    }
                    data.remove(0, static_cast<int>(hdr->len));
                }
            }
        }
    }
    catch(...) {
        qWarning()<<"Exception in" << __PRETTY_FUNCTION__;
    }
}
